/****************************************************************************
avwhdr.h - Analyze View header file

Author:  Paul Hughett

Copyright (c) 2003 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

The structure definitions for the header appear to be Copyright (c)
1986-1995 by Biomedical Imaging Resource, Mayo Foundation, under the
file name dbh.h.


This header file defines the image file header used in the AnalyzeView
software by Mayo Clinic and in other later programs.  Its definition and
meaning has been integrated from several sources including the following:

   1. The AnalyzeFromMayo section of David Clunie's Medical Image Format FAQ.
   2. The analyze_hdr.h file in the NRIA program suite.
   3. An anonymous PDF document entitled ANALYZE(tm) 7.5 File Format.
   4. Personal communications from Eugene Gualtieri.

None of these sources can be regarded as truly authoritative, although
1 comes reasonably close; in any case, programs using this format have
often modified it to their own special needs.  Thus, deviations from
the description here are to be expected.

An image or image set in AnalyzeView format is defined by two files:
a header file as described here which carries the hdr extension; and
an image file which carries the img extension.


DATA TYPES

The following types of image data are supported.

     type                code    bpp      description
     DT_BINARY              1      1      binary
     DT_UNSIGNED_CHAR       2      8      unsigned char, gray scale
     DT_SIGNED_SHORT        4     16      signed short
     DT_SIGNED_INT          8     32      signed int
     DT_FLOAT              16     32      float
     DT_COMPLEX            32     64      floats in (real, imag) pairs
     DT_DOUBLE             64     64      double
     DT_RGB               128     24      unsigned char (red, green, blue)


COMMENTS ON SPECIFIC FIELDS

The header_key struct is required and indicates how the file is laid
out on disk.  The sizeof_header field is almost always 348 and can be
used both as a magic number to identify the AnalyzeView header and to
determine the byte order used for both the header and image files.
All the documentation I've seen states that the extents field must be
16384, but I've never seen any explanation given for what the field
means.  Similarly, I've never seen an explanation of how a 'regular'
file differs from other files.

	int sizeof_header   // must indicate size of header file
	int extents;        // should be 16384 
	char regular;       // 'r' 
		
		
The image_dimension struct decribes the organization and size of
images. These elements enable IO routines to reference images by
volume and slice number.
	
	short int dim[]         // array of image dimensions 
                                // dim[0]  = number of dimensions; usually 4 
		                // dim[1]  = image width 
		                // dim[2]  = image height 
		                // dim[3]  = volume depth 
		                // dim[4]  = volumes in file 
			
	char vox_units[4]       // labels voxerl spatial unit 
	char cal_units[4]       // labels voxel calibration unit 
	short int datatype      // Acceptable values are 

	short int bitpix        // bits per pixel 
	float pixdim[]          // parallel array to dim giving voxel
	                        // dimensions in each dimension;
                                // 0.0 implies value is unknown.
	                        // pixdim[1] = voxel width 
	                        // pixdim[2] = voxel height 
	                        // pixdim[3] = voxel depth or slice thickness 
		         
	float vox_offset;       // byte offset in the .img file at which
	                        // voxels start. If value is negative
	                        // specifies that the absolute value
	                        // is applied for every image in the file. 
		                   
	float cal_max, cal_min  // specify range of calibration values 
	int glmax, glmin        // the max and min values for entire data set 
		
	
The data_history substructure is not required, but the 'orient' element
is used to indicate individual slice orientation and determines whether
the ANALYZE 'Movie' program will attempt to flip the images before 
displaying a movie sequence.

	orient: 
		0 - transverse unflipped
		1 - coronal unflipped
		2 - sagittal unflipped
		3 - transverse flipped
		4 - coronal flipped
		5 - sagittal flipped



COMMENTS ON SPECIFIC APPLICATION PROGRAMS

Analyze(tm) apparently expects that all images are 4-dimensional and
thus that dim[0] == 4.  These dimensions are column, row, slice, and
instant.  Furthermore, regular == 'r' is expected and means that the
file is regular; it does not seem to be defined what 'regular' means.

SPM99 and SPM2 use a variant of this format, in which the originator
field is repurposed to record the origin and the funused1 and funused2
to record a scale factor and image intensity zero intercept.

Programs from the NRIA package are rather casual about which fields
they actually set in this header.  About you can count on are
sizeof_hdr, dim, pixdim, bitpix, and perhaps glmin and glmax.  The
datatype field appears to be set to random garbage and cannot be
trusted to specify the data format; in any case, the allowed values in
that field don't allow length and signedness to be set independently.

# $Id: avwhdr.h,v 1.3 2004/05/13 14:54:51 hughett Exp $  $Name: pyvox-0-72 $
*****************************************************************************/


#ifndef AVWHDR_H
#define AVWHDR_H

struct header_key          		/*      header_key       */
    {                                  		/* off + size*/
        int sizeof_hdr;                		/* 0 + 4     */
        char data_type[10];            		/* 4 + 10    */
        char db_name[18];              		/* 14 + 18   */
        int extents;                   		/* 32 + 4    */
        short int session_error;       		/* 36 + 2    */
        char regular;                  		/* 38 + 1    */
        char hkey_un0;                 		/* 39 + 1    */
    };           				/* total=40  */

struct image_dimension      		/*      image_dimension  */
    {                               		/* off + size*/
        short int dim[8];           		/* 0 + 16    */
        char vox_units[4];			/* 16 + 4    */
        char cal_units[8];			/* 20 + 4    */
        short int unused1;			/* 24 + 2    */
        short int datatype;			/* 30 + 2    */
        short int bitpix;                   	/* 32 + 2    */
        short int dim_un0;                      /* 34 + 2    */
        float pixdim[8];                        /* 36 + 32   */
        float vox_offset;                    	/* 68 + 4    */
        float roi_scale;                       	/* 72 + 4    */
        float funused1;                      	/* 76 + 4    */
        float funused2;                      	/* 80 + 4    */
        float cal_max;                    	/* 84 + 4    */
        float cal_min;                     	/* 88 + 4    */
        int compressed;                     	/* 92 + 4    */
        int verified;                     	/* 96 + 4    */
        int glmax, glmin;                 	/* 100 + 8   */
    };          				/* total=108 */
         
struct data_history          		/*      data_history     */
    {                                    	/* off + size*/
        char descrip[80];                	/* 0 + 80    */
        char aux_file[24];               	/* 80 + 24   */
        char orient;                     	/* 104 + 1   */
        char originator[10];             	/* 105 + 10  */
        char generated[10];              	/* 115 + 10  */
        char scannum[10];                	/* 125 + 10  */
        char patient_id[10];             	/* 135 + 10  */
        char exp_date[10];               	/* 145 + 10  */
        char exp_time[10];               	/* 155 + 10  */
        char hist_un0[3];                	/* 165 + 3   */
        int views;                       	/* 168 + 4   */
        int vols_added;                  	/* 172 + 4   */
        int start_field;                 	/* 176 + 4   */
        int field_skip;                  	/* 180 + 4   */
        int omax,omin;                   	/* 184 + 8   */
        int smax,smin;                   	/* 192 + 8   */
    };                     			/* total=200 */

struct dsr                 		/*      dsr              */
    {                                  		/* off + size*/
        struct header_key hk;          		/* 0 + 40    */
        struct image_dimension dime;   		/* 40 + 108  */
        struct data_history hist;      		/* 148 + 200 */
    };                     			/* total=348 */
     

/* Permitted values for the datatype field */
#define DT_NONE			0
#define DT_UNKNOWN		0
#define DT_BINARY		1
#define DT_UNSIGNED_CHAR	2
#define DT_SIGNED_SHORT		4
#define DT_SIGNED_INT		8
#define DT_FLOAT	       16
#define DT_COMPLEX	       32
#define DT_DOUBLE	       64
#define DT_RGB		      128
#define DT_ALL		      255



#endif
