/****************************************************************************
bips.h - Public header file for the Basic Image Processing Subroutines

Author:  Paul Hughett

Copyright (c) 1999-2004 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging applications.

$Id: bips.h,v 1.47 2005/12/03 19:07:57 hughett Exp $  $Name: pyvox-0-72 $
*****************************************************************************/


#ifndef BIPS_H
#define BIPS_H

#ifdef HAVE_COMPLEX_H
#include <complex.h>
#endif

#include "errm.h"
#include "exim.h"

#ifdef __cplusplus          /* Allow to be called from C++ */
extern "C" {
#endif



/*-------------------------------------------------------------------------
Unary point functions in standard form

These functions all take one source and one destination array, both of
which are the same type.  Some take input from the destination, thus
updating it in place.
-------------------------------------------------------------------------*/

typedef int bips_unop(long nelem, int type, void *dest, long ds,
			void *src1, long ss1);

extern bips_unop bips_not;
extern bips_unop bips_copy;
extern bips_unop bips_minus;
extern bips_unop bips_real;
extern bips_unop bips_imag;
extern bips_unop bips_carg;
extern bips_unop bips_conj;
extern bips_unop bips_abs;
extern bips_unop bips_floor;
extern bips_unop bips_ceil;
extern bips_unop bips_square;
extern bips_unop bips_sqrt;
extern bips_unop bips_cbrt;
extern bips_unop bips_exp;
extern bips_unop bips_exp2;
extern bips_unop bips_expm1;
extern bips_unop bips_log;
extern bips_unop bips_log10;
extern bips_unop bips_log1p;
extern bips_unop bips_log2;
extern bips_unop bips_sin;
extern bips_unop bips_cos;
extern bips_unop bips_tan;
extern bips_unop bips_asin;
extern bips_unop bips_acos;
extern bips_unop bips_atan;
extern bips_unop bips_sinh;
extern bips_unop bips_cosh;
extern bips_unop bips_tanh;
extern bips_unop bips_erf;
extern bips_unop bips_erfc;
extern bips_unop bips_lgamma;
extern bips_unop bips_tgamma;

extern bips_unop bips_and_set;
extern bips_unop bips_or_set;
extern bips_unop bips_xor_set;
extern bips_unop bips_min_set;
extern bips_unop bips_max_set;
extern bips_unop bips_nzmin_set;
extern bips_unop bips_add_set;
extern bips_unop bips_sub_set;
extern bips_unop bips_mul_set;
extern bips_unop bips_div_set;



/*--------------------------------------------------------------------------
Type conversion functions in standard form

These functions all take a source argument of any type and convert it to
a destination array of some given type.
---------------------------------------------------------------------------*/

extern int bips_uchar(long nelem, unsigned char *dst, long ds,  
    int stype, void *src, long ss);
extern int bips_schar(long nelem, signed char *dst, long ds,  
    int stype, void *src, long ss);
extern int bips_ushort(long nelem, unsigned short *dst, long ds,  
    int stype, void *src, long ss);
extern int bips_short(long nelem, short *dst, long ds,  
    int stype, void *src, long ss);
extern int bips_uint(long nelem, unsigned int *dst, long ds,  
    int stype, void *src, long ss);
extern int bips_int(long nelem, int *dst, long ds,  
    int stype, void *src, long ss);
extern int bips_ulong(long nelem, unsigned long *dst, long ds,  
    int stype, void *src, long ss);
extern int bips_long(long nelem, long *dst, long ds,  
    int stype, void *src, long ss);
extern int bips_float(long nelem, float *dst, long ds,  
    int stype, void *src, long ss);
extern int bips_double(long nelem, double *dst, long ds,  
    int stype, void *src, long ss);
#ifdef HAVE_COMPLEX_H
extern int bips_fcomplex(long nelem, float complex *dst, long ds,  
    int stype, void *src, long ss);
extern int bips_dcomplex(long nelem, double complex *dst, long ds,  
    int stype, void *src, long ss);
#endif



/*--------------------------------------------------------------------------
Type conversion functions of non-standard form
---------------------------------------------------------------------------*/

extern int bips_cast(long nelem, int dtype, void *dst, long ds,  
    int stype, void *src, long ss);
extern int bips_round_cast(long nelem, int dtype, void *dst, long ds,  
    int stype, void *src, long ss);
extern int bips_scale_to_uchar(long nelem, unsigned char *dst, long ds,
    int stype, void *src, long ss,
    double offset, double gain);
#ifdef HAVE_COMPLEX_H
extern int bips_fcomplex2(long nelem, float complex *dst, long ds,  
    int stype, void *src1, long ss1, void *src2, long ss2);
extern int bips_dcomplex2(long nelem, double complex *dst, long ds,  
    int stype, void *src1, long ss1, void *src2, long ss2);
#endif




/*---------------------------------------------------------------------------
Binary point operations in standard form

These functions each take two source arguments and one destination argument,
all of which are the same type.  Some take input from the destination, thus
updating it in place.
---------------------------------------------------------------------------*/

typedef int bips_binop(long nelem, int type, void *dest, long ds,
		       void *src1, long ss1, void *src2, long ss2);

extern bips_binop bips_min;
extern bips_binop bips_max;
extern bips_binop bips_and;
extern bips_binop bips_or;
extern bips_binop bips_xor;
extern bips_binop bips_add;
extern bips_binop bips_sub;
extern bips_binop bips_mul;
extern bips_binop bips_div;
extern bips_binop bips_rem;
extern bips_binop bips_pow;
extern bips_binop bips_atan2;
extern bips_binop bips_hypot;



/*---------------------------------------------------------------------------
Ternary point operations in standard form

These functions each take three source arguments and one destination argument,
all of which are the same type.
---------------------------------------------------------------------------*/

typedef int bips_ternop(long nelem, int type, void *dest, long ds,
		        void *src1, long ss1, void *src2, long ss2,
			void *src3, long ss3);

extern bips_ternop bips_fma;



/*----------------------------------------------------------------------------
Functions with an unusual form
-----------------------------------------------------------------------------*/

extern int bips_cmove(long nelem, int type, void *dst, long ds,   
    unsigned char *sel1, long ss1, void *src2, long ss2);
extern int bips_compare(long nelem, unsigned char *dst, long ds,   
    int stype, void *src1, long ss1, void *src2, long ss2,  
    unsigned char lsval, unsigned char eqval, unsigned char gtval);
extern int bips_eq(long nelem, unsigned char *dst, long ds,   
    int stype, void *src1, long ss1, void *src2, long ss2,  
    unsigned char eqval, unsigned char neval);
extern int bips_gt(long nelem, unsigned char *dst, long ds,   
    int stype, void *src1, long ss1, void *src2, long ss2,  
    unsigned char gtval, unsigned char leval);
extern int bips_lt(long nelem, unsigned char *dst, long ds,   
    int stype, void *src1, long ss1, void *src2, long ss2,  
    unsigned char ltval, unsigned char geval);

extern void *bips_locate(int type, void *src, long i);

extern int bips_zero(long nelem, int type, void *dst, long ds);
extern int bips_one (long nelem, int type, void *dst, long ds);
extern int bips_ffff(long nelem, int type, void *dst, long ds);
extern int bips_ramp(long nelem, int type, void *dst, long ds);
extern int bips_fill_max(long nelem, int type, void *dst, long ds);
extern int bips_fill_min(long nelem, int type, void *dst, long ds);

extern int bips_zero_range(long first, long limit, int type, void *dst,
    long ds);
extern int bips_fill_range(long first, long limit, int type, void *dst,
    long ds, double value);

extern int bips_fma_set(long nelem, int type, void *dest, long ds,
    int stype1, void *src1, long ss1, void *src2, long ss2);

extern int bips_gather(long nelem, int  type, void *dst, long ds, 
    void *src, long *index);
extern int bips_scatter(long nelem, int  type, void *dst, long *index, 
    void *src, long ss);

extern int bips_lookup(long nelem, int type, int xtype,   
    void *dst, long ds, void *src, long ss,
    void *table, long ts, long ntable);

extern int bips_alpha(long nelem, long *dsti, long dsi,
    int type, void *dstf, long dsf, double base, double incr);
extern int bips_lerp(long nelem, int datype, void *dst, long ds,
    void *alf, long as, int stype, void *src, long ss);
extern int bips_glerp(long nelem, int dtype, void *dst, long ds,
    void *alf, long as, int stype, void *src, long *index);


extern int bips_absdiff(long nelem, double *dst, long ds,
    int stype, void *src1, long ss1, void *src2, long ss2);
extern int bips_sqrdiff(long nelem, double *dst, long ds,
    int stype, void *src1, long ss1, void *src2, long ss2);

extern int bips_chamfer(long nelem, int dtype, void *dst, long ds, long doff,
    int stype, void *src, long ss);

extern int bips_moments(long nelem, double dst[3],
    int stype, void *src, long ss);

extern int bips_conv1(long nelem, int dtype, void *dst, long ds,
    int stype, void *src, long ss, int ctype, void *coef);

extern int bips_lostat1(long nelem, double *dst0, void *dst1, double *dst2,
    int stype, void *src, long ss, double coef);
extern int bips_lostat2(long nelem, int dtype, void *dst1, void *dst2, long ds,
    double *s0, void *s1, double *s2, double scale);

#ifdef __cplusplus
}
#endif

#endif /* BIPS_H */
