/****************************************************************************
clap.c - C wrapper functions for the LAPACK library

Author:  Paul Hughett

Copyright (c) 2000-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This header file provides prototypes for the BLAS and LAPACK subroutines
used by the Voxel Kit.

$Id: clap.h,v 1.15 2005/12/07 19:27:44 hughett Exp $  $Name: pyvox-0-72 $
****************************************************************************/

#ifndef CLAP_H
#define CLAP_H

#include "config.h"         /* For platform-specific C-Fortran interface */


#ifdef __cplusplus          /* Allow to be called from C++ */
extern "C" {
#endif



/*------------------------------------------------------------------------
Prototypes for BLAS subroutines
-------------------------------------------------------------------------*/

/* Matrix multiplication */
extern void sgemm_(fchar1 *transa, fchar1 *transb, fint *m, fint *n, fint *k,
    freal *alpha, freal *a, fint *lda, freal *b, fint *ldb,
    freal *beta, freal *c, fint *ldc);
extern void dgemm_(fchar1 *transa, fchar1 *transb, fint *m, fint *n, fint *k,
    fdouble *alpha, fdouble *a, fint *lda, fdouble *b, fint *ldb,
    fdouble *beta, fdouble *c, fint *ldc);
extern void cgemm_(fchar1 *transa, fchar1 *transb, fint *m, fint *n, fint *k,
    fscomplex *alpha, fscomplex *a, fint *lda, fscomplex *b, fint *ldb,
    fscomplex *beta, fscomplex *c, fint *ldc);
extern void zgemm_(fchar1 *transa, fchar1 *transb, fint *m, fint *n, fint *k,
    fdcomplex *alpha, fdcomplex *a, fint *lda, fdcomplex *b, fint *ldb,
    fdcomplex *beta, fdcomplex *c, fint *ldc);


/*-------------------------------------------------------------------------
Prototypes for LAPACK subroutines
-------------------------------------------------------------------------*/

/* LU (triangular) factorization */
extern void sgetrf_(fint *m, fint *n, freal   *a, fint *lda, fint *ipiv,
    fint *info);
extern void dgetrf_(fint *m, fint *n, fdouble *a, fint *lda, fint *ipiv,
    fint *info);
extern void cgetrf_(fint *m, fint *n, fscomplex *a, fint *lda, fint *ipiv,
    fint *info);
extern void zgetrf_(fint *m, fint *n, fdcomplex *a, fint *lda, fint *ipiv,
    fint *info);

/* Inversion using LU factorization */
extern void sgetri_(fint *n, freal *a, fint *lda, fint *ipiv,
    freal *work, fint *lwork, fint *info);
extern void dgetri_(fint *n, fdouble *a, fint *lda, fint *ipiv,
    fdouble *work, fint *lwork, fint *info);
extern void cgetri_(fint *n, fscomplex *a, fint *lda, fint *ipiv,
    fscomplex *work, fint *lwork, fint *info);
extern void zgetri_(fint *n, fdcomplex *a, fint *lda, fint *ipiv,
    fdcomplex *work, fint *lwork, fint *info);

/* Solution of linear system by LU factorization */
extern void sgetrs_(fchar1 *trans, fint *n, fint *nrhs, freal *a, fint *lda,
    fint *ipiv, freal *b, fint *ldb, fint *info);
extern void dgetrs_(fchar1 *trans, fint *n, fint *nrhs, fdouble *a, fint *lda,
    fint *ipiv, fdouble *b, fint *ldb, fint *info);
extern void cgetrs_(fchar1 *trans, fint *n, fint *nrhs, fscomplex *a,
    fint *lda, fint *ipiv, fscomplex *b, fint *ldb, fint *info);
extern void zgetrs_(fchar1 *trans, fint *n, fint *nrhs, fdcomplex *a,
    fint *lda, fint *ipiv, fdcomplex *b, fint *ldb, fint *info);

/* Cholesky factorization */
extern void spotrf_(fchar1 *uplo, fint *n, freal     *a, fint *lda,
    fint *info);
extern void dpotrf_(fchar1 *uplo, fint *n, fdouble   *a, fint *lda,
    fint *info);
extern void cpotrf_(fchar1 *uplo, fint *n, fscomplex *a, fint *lda,
    fint *info);
extern void zpotrf_(fchar1 *uplo, fint *n, fdcomplex *a, fint *lda,
    fint *info);

/* Singular value decomposition */
/* NOTE that the complex functions have an extra argument */
extern void sgesvd_(fchar1 *jobu, fchar1 *jobvt, fint *m, fint *n,
    freal *a, fint *lda, freal *s, freal *u, fint *ldu,
    freal *vt, fint *ldvt, freal *work, fint *lwork, fint *info);
extern void dgesvd_(fchar1 *jobu, fchar1 *jobvt, fint *m, fint *n,
    fdouble *a, fint *lda, fdouble *s, fdouble *u, fint *ldu,
    fdouble *vt, fint *ldvt, fdouble *work, fint *lwork, fint *info);
extern void cgesvd_(fchar1 *jobu, fchar1 *jobvt, fint *m, fint *n,
    fscomplex *a, fint *lda, freal *s, fscomplex *u, fint *ldu,
    fscomplex *vt, fint *ldvt, fscomplex *work, fint *lwork, freal *rwork,
    fint *info);
extern void zgesvd_(fchar1 *jobu, fchar1 *jobvt, fint *m, fint *n,
    fdcomplex *a, fint *lda, fdouble *s, fdcomplex *u, fint *ldu,
    fdcomplex *vt, fint *ldvt, fdcomplex *work, fint *lwork, fdouble *rwork,
    fint *info);

/* Symmetric/Hermitian eigenvalue problem */
extern void ssyev_(fchar1 *jobz, fchar1 *uplo, fint *n, freal *a,
    fint *lda, freal *w, freal *work, fint *lwork, fint *info);
extern void dsyev_(fchar1 *jobz, fchar1 *uplo, fint *n, fdouble *a,
    fint *lda, fdouble *w, fdouble *work, fint *lwork, fint *info);
extern void cheev_(fchar1 *jobz, fchar1 *uplo, fint *n, fscomplex *a,
    fint *lda, fscomplex *w, fscomplex *work, fint *lwork, freal *rwork,
    fint *info);
extern void zheev_(fchar1 *jobz, fchar1 *uplo, fint *n, fdcomplex *a,
    fint *lda, fdcomplex *w, fdcomplex *work, fint *lwork, fdouble *rwork,
    fint *info);


#ifdef __cplusplus
}
#endif

#endif  /* CLAP_H */
