/****************************************************************************
errm.h - Prototypes for error management functions

Author:  Paul Hughett

Copyright (c) 1998-2002 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This public header file contains prototypes for various error management
functions used within bblimage (but probably useful elsewhere as well).

FIXME: Provide clean methods for handling errors in BIPS and LAPACK/BLAS.

FIXME: Add function to translate the error code into a readable string?

FIXME: Should the returned error flag be the numeric error code, or just
"non-zero"?

FIXME: Provide longer descriptions of the error codes and classify fuzzy
cases.

$Id: errm.h,v 1.19 2005/12/29 18:33:13 hughett Exp $  $Name: pyvox-0-72 $
****************************************************************************/


#ifndef ERRM_H
#define ERRM_H

#ifdef __cplusplus          /* Allow to be called from C++ */
extern "C" {
#endif

/* Generic error codes */
#define ERR_INAPP    -1      /* Algorithm not applicable */
#define ERR_OKAY      0      /* No error, operation successful */
#define ERR_MISC      1      /* Miscellaneous or unspecified error */
#define ERR_UNIMP     2      /* Not yet implemented */
#define ERR_ILLOP     3      /* Illegal operation */
#define ERR_INCOMP    4      /* Incompatible operands */
#define ERR_MALLOC    5      /* Unable to allocate sufficient memory */
#define ERR_EOF       6      /* Unexpected end of file */
#define ERR_INIT      7      /* Uninitialized or invalid operand */
#define ERR_DOMAIN    8      /* Operand not valid */
#define ERR_RANGE     9      /* Result is not representable */
#define ERR_FILE     10      /* Input/output error */
#define ERR_ALGOR    11      /* Algorithm error (e.g. non-convergence) */
#define ERR_TYPE     12      /* Value or argument of invalid type */
#define ERR_UNDEF    13      /* Undefined type, attribute, ... */
#define ERR_ASSERT   14      /* Assertion failure or other internal error */
#define ERR_INDEX    15      /* Subscript is invalid or out of range */


/* Public global variables */

/* The following global variable was made private on 2002-08-18
   to force programmers to use the set_error_exit() function instead,
   which can be more easily made thread-safe, and also to change the
   encoding to permit continue, exit, and abort options. */
#if 0
extern int errm_abort_flag;       /* Abort (and core dump) on error? */
#endif

/* Public functions of the new interface */
extern void set_error_context(const char *context);
extern void clear_error_context(void);
extern void set_error_action  (int action);
extern void set_warning_action(int action);
extern int preset_error(const int errcode, const char *message, ...);
extern int set_error   (const int errcode, const char *message, ...);
extern int set_warning (const int errcode, const char *message, ...);
extern int get_error(char **message);
extern const char *get_error_message(void);
extern void errm_panic(const char *message, ...);


/* Public macros for the new interface */

/* Set the error code and message; continue at next line */
#define SetError(errcode, message) \
    set_error((errcode), (message))

/* Set the error code and message; return the non-zero error code */
#define NZReturnError(errcode, message) \
    return set_error(errcode, message)

/* Set the error code and message; return the non-zero error code */
#define NZReturnError1(errcode, message, arg1) \
    return set_error(errcode, message, arg1)

/* Set the error code and message; return zero */
#define ZReturnError(errcode, message) \
    do {set_error((errcode), (message)); return 0; } while (0)

/* Set the error code and message; return zero */
/* FIXME: C99 allows optional args in macros; does that help here? */
#define ZReturnError1(errcode, message, arg1) \
    do {set_error((errcode), (message), (arg1)); return 0; } while (0)

/* Report a warning; then convert to NZ error or continue. */
#define NZWarning(errcode, message)            \
    do {                                       \
        int flag;                              \
        flag = set_warning(errcode, message);  \
        if (flag)    return flag;              \
    } while (0)

/* Report a warning; then convert to Z error or continue. */
#define ZWarning(errcode, message)             \
    do {                                       \
        int flag;                              \
        flag = set_warning(errcode, message);  \
        if (flag)    return 0;                 \
    } while (0)


/* Return non-zero for an error already set at a deeper level */
#define NZReturnLibError() \
    return get_error(0)

/* Return zero for an error already set at a deeper level */
#define ZReturnLibError() \
    return 0


#ifdef __cplusplus
}
#endif

#endif /* ERRM_H */
