/****************************************************************************
exim.h - Public header file defining export/import types

Author:  Paul Hughett

Copyright (c) 2000-2004 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file defines various internal and external data types and
declares functions for converting between the internal and external
representations; it is part of the Voxel Kit, a set of general-purpose
functions for volume image processing, with a bias toward medical
imaging applications.  

FIXME: The exim code probably has many other applications and should
be made easy to use independently.

FIXME: A plain char type might be useful for importing or exporting header
files and similar stuff even though it's useless for image data.  In any
case, decide how to handle non-native text data.

$Id: exim.h,v 1.36 2006/02/23 21:49:57 hughett Exp $  $Name: pyvox-0-72 $
*****************************************************************************/


#ifndef EXIM_H
#define EXIM_H

#include <stdio.h>

#ifdef HAVE_COMPLEX_H
#include <complex.h>
#endif

#ifdef __cplusplus          /* Allow to be called from C++ */
extern "C" {
#endif



/*------------------------------------------------------------------------
External data representations

The following numeric codes (extypes) may be used to indicate the
external (in-file) representation for a datum. The length of a datum
is given by the trailing digits of the type name; the endianness is
recorded separately.  All of these codes (except EXTP_NONE) are
positive to distinguish them for the type codes for the internal data
representations.  The numeric value of the code has no significance,
but is used as an array index into the external_types table in exim.c.
The EXTP_NONE code is sometimes used to request a default external
type.

NOTE: If you change these defines, you *must* change the external_types
table in exim.c to match!

Prior to 2000/01/21, there was no EXTP_NONE type and EXTP_UCHAR
occupied code 0; among other things, this means that all the valid
codes were non-negative rather than positive.  In addition, the codes
were not contiguous.  I *think* that I've made all the necessary
changes but that means I've probably actually missed a few, so keep
your eyes open.

FIXME: Add other floating point types as needed, e.g. EXTP_VAXG8 and
EXTP_CVAXG16 for real and complex data in the VAX G floating point
format.  IEEE format gets the simple FLOAT and COMPLEX specifiers
because it is both common and a preferred representation.
--------------------------------------------------------------------------*/

#define EXTP_NONE        0       /* Undefined or unspecified type */

#define EXTP_UCHAR       1       /* (Deprecated) Alias for EXTP_UINT1 */
#define EXTP_UINT1       1       /* Unsigned 1-byte int */
#define EXTP_UINT2       2       /* Unsigned 2-byte int */
#define EXTP_UINT4       3       /* Unsigned 4-byte int */
#define EXTP_INT1        4       /* Signed two's complement 1-byte int */
#define EXTP_INT2        5       /* Signed two's complement 2-byte int */
#define EXTP_INT4        6       /* Signed two's complement 4-byte int */
#define EXTP_FLOAT4      7       /* 4-byte IEEE-754 floating point */
#define EXTP_FLOAT8      8       /* 8-byte IEEE-754 floating point */
#define EXTP_COMPLEX8    9       /* Real, imag pair of float4's */
#define EXTP_COMPLEX16  10       /* Real, imag pair of float8's */



/*---------------------------------------------------------------------------
Limits of external data representations

FIXME: Define all of them, not just those currently needed.
----------------------------------------------------------------------------*/

#define UINT1_MAX      255U
#define UINT2_MAX    65535UL
#define INT1_MIN      -128
#define INT1_MAX       127
#define INT2_MIN    -32768L
#define INT2_MAX     32767L



/*-----------------------------------------------------------------------
Internal data representations

The following numeric codes (intypes) may be used to indicate the
internal (in-core) representation of a datum; the length is implicit in
the type and defined by the native architecture.  All of these codes
(except INTP_NONE) are negative, to distinguish them from the codes
which denote external representations.  Each of these types
corresponds directly to a native type in the C language, but note that
processing functions may not be provided for all of these types.  Two
different numeric codes may refer to the same underlying
representation; for example INTP_INT and INTP_LONG will usually be the
same on a 32-bit platform (but are considered distinct types anyway).
The numeric value of the code has no significance, but is used as an
array index into the internal_types table in exim.c.  The INTP_NONE
code is sometimes used to request a default internal type.

The INTP_SCHAR type has an extra S to indicate explicitly that it is 
signed, since char does not default to signed in standard C, unlike 
all the other integral types.  It may not matter much; this type is 
arguably worthless for imaging work.  A char type with unknown signedness
is so obviously useless that I haven't even defined a type for it.

Prior to 2000/01/21, there was no EXTP_NONE type and EXTP_UCHAR
occupied code 0; among other things, this means that all the valid
codes were non-negative rather than positive.  In addition, the codes
were not contiguous.  I *think* that I've made all the necessary
changes but that means I've probably actually missed a few, so keep
your eyes open.

NOTE: If you change these defines, you *must* also change the internal_types
table in exim.c to match!
-------------------------------------------------------------------------*/

#define INTP_NONE        0      /* Undefined or unspecified type */

#define INTP_UCHAR      -1      /* Unsigned char */
#define INTP_USHORT     -2      /* Unsigned short */
#define INTP_UINT       -3      /* Unsigned int */
#define INTP_ULONG      -4      /* Unsigned long */

#define INTP_SCHAR      -5      /* Signed character */
#define INTP_SHORT      -6      /* Signed short */
#define INTP_INT        -7      /* Signed int */
#define INTP_LONG       -8      /* Signed long */

#define INTP_FLOAT      -9      /* Float */
#define INTP_DOUBLE    -10      /* Double */
#define INTP_FCOMPLEX  -11      /* Float complex */
#define INTP_DCOMPLEX  -12      /* Double complex */



/*------------------------------------------------------------------------
anytype - Generic type for a single BIPS datum

It is occasionally necessary to store a single value of unknown
internal type in a fashion compatible with exim and BIPS, which
usually refer to their arguments via unsigned char pointers which are
cast to the type defined by a numeric code.  This typedef creates a
union large enough to contain a single datum of any type supported by
BIPS and exim, and can be used as a BIPS source or destination.  The
pyvox_set_value() and pyvox_get_value() can be used with index = 0
to convert between anytype and Python numbers.  The various union
elements can be used to convert between anytype and the usual C types.

A typical usage of this type is shown below.  This code fragment fills
an n-element array dst of exim type vtype with the value 2.0.

    int vtype;            /# Exim type code #/
    anytype value;        /# The value itself #/
    ...
    exim_set_value(&value, vtype, 1, 2.0);        /# Convert to dest type #/
    bips_copy(nelem, vtype, dst, 1, &value, 0);   /# Fill destination #/

NOTE: Prior to 2002-08-26, this type was defined as an unsigned char
array as shown below; this change requires that all uses change.

      typedef unsigned char anytype[32];

FIXME: Modify so that the vxl_set_voxel() and vxl_get_voxel()
functions can be used to convert between anytype and voxel array
elements.  Or keep these functions for compatibility and add some
other functions for the more general function.

FIXME: Add long long and long double types if and when supported.
-------------------------------------------------------------------------*/

typedef union {
    unsigned char *as_ucptr;         /* As unsigned char pointer */
    unsigned char  as_uchar;
    unsigned short as_ushort;
    unsigned int   as_uint;
    unsigned long  as_ulong;
    signed char    as_schar;
    signed short   as_short;
    signed int     as_int;
    signed long    as_long;
    float          as_float;
    double         as_double; 
#if HAVE_COMPLEX_H
    float  complex as_fcomplex;
    double complex as_dcomplex;
#endif
}
    anytype;



/*-------------------------------------------------------------------------
Exim mode structure

This structure contains the parameters for exim_scalar and relatives
that do not depend on the individual field being converted.  The codes
EXIM_IMPORT, etc. are used in the dir field to specify the direction of
data transfer; not all of these have been implemented yet.

FIXME: Add fields to accumulate error flags, and to control the handling
of errors.
--------------------------------------------------------------------------*/

struct exim_mode {
    int dir;             /* Import, export, or dump? */
    FILE *exfile;        /* File containing external data */
    FILE *dump;          /* File to print values to */
    int bigend;          /* Big endian? */
};

/* Possible values for the dir field */
#define EXIM_IMPORT   1          /* Read from file into memory */
#define EXIM_EXPORT   2          /* Write to file from memory */
#define EXIM_IMDUMP   3          /* Print values from file */
#define EXIM_EXDUMP   4          /* Print values from memory */  



/*----------------------------------------------------------------------
SCALAR & VECTOR - Package the processing of a field

These macros encapsulate calls to exim_scalar and exim_vector into a
format which can be written on a single line and read as a table; they
also arrange to detect errors and call exim_error when necessary.

NOTE: These macros were deprecated on 2002-09-10; use ZSCALAR and ZVECTOR
in new code.
-----------------------------------------------------------------------*/

#if 0
#define SCALAR(field,intype,extype,format)                     \
    if (( err = exim_scalar(mode, &dsr->field, INTP_##intype,  \
                      EXTP_##extype, format) ))                \
        exim_error(#field, err)

#define VECTOR(nelem,field,intype,extype,format)                     \
    if (( err = exim_vector(mode, nelem, &dsr->field, INTP_##intype, \
  		       EXTP_##extype, format) ))                     \
        exim_error(#field, err)
#endif



/*----------------------------------------------------------------------
ZSCALAR & ZVECTOR - Package the processing of a field

These macros encapsulate calls to exim_scalar and exim_vector into a
format which can be written on a single line and read as a table; they
also arrange to detect errors and call ZReturnLibError when necessary.

FIXME: Modify so that the field name appears in the error message?
-----------------------------------------------------------------------*/

#define ZSCALAR(field,intype,extype,format)                    \
    if (( err = exim_scalar(mode, &dsr->field, INTP_##intype,  \
                      EXTP_##extype, format) ))                \
        ZReturnLibError()

#define ZVECTOR(nelem,field,intype,extype,format)                    \
    if (( err = exim_vector(mode, nelem, &dsr->field, INTP_##intype, \
  		       EXTP_##extype, format) ))                     \
        ZReturnLibError()



/*--------------------------------------------------------------------------
Public functions
---------------------------------------------------------------------------*/

extern int exim_all_extypes(int *array, int size);
extern int exim_all_intypes(int *array, int size);
extern int exim_sizeof_type(int type);
extern int exim_sizeof_extype(int type);
extern int exim_sizeof_intype(int type);
extern int exim_natural_intype(int extype);
extern int exim_natural_extype(int intype);
extern int exim_isint(int type);
extern int exim_isunsigned(int type);
extern int exim_isfloat(int type);
extern int exim_iscomplex(int type);
extern int exim_isreal(int type);
extern int exim_complex_type(int type);
extern int exim_real_type(int type);
extern int exim_sum_type(int type);
extern double exim_epsilon(int type);
extern int exim_max(int type, anytype *value);
extern int exim_min(int type, anytype *value);
extern const char *exim_nameof_type(int type);
extern const char *exim_describe_type(int type);
extern int exim_get_value(double *value, void *src, int stype, long n);
extern int exim_set_value(void *dst, int dtype, long n, double value);
extern void exim_error(char *field, int err);

extern void exim_swap_bytes(long count, unsigned char *dest,
    unsigned char *src, int nbout, int nbin, int *inbyte, unsigned char *ctab);
extern int exim_ulong_to_bytes(long count, unsigned char *dest,
    unsigned long *src, int nbytes, int bigendian);

extern int exim_export(long nelem, unsigned char *dst, int dtype, long ds,
    int bigend, void *src, int stype, long  ss);
extern int exim_import(long nelem, void *dst, int dtype, long ds, 
    unsigned char *src, int stype, long ss, int bigend);

extern struct exim_mode *exim_create_mode(int dir, FILE *exfile,
    FILE *dump, int bigend);
extern void exim_destroy_mode(struct exim_mode *mode);
extern int exim_scalar(struct exim_mode *mode, void *indata, int intype,
    int extype, char *format);
extern int exim_vector(struct exim_mode *mode, size_t nelem, void *indata,
    int intype, int extype, char *format);
extern void *exim_alloc(struct exim_mode *mode, void *indata,
    size_t nelem, int intype);

#ifdef __cplusplus
}
#endif

#endif /* EXIM_H */
