/****************************************************************************
pyvox.h - Header file to define shared objects for Pyvox

Author:  Paul Hughett

Copyright (c) 1999-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This header file defines the objects that must be shared among the different
files that define different components of the Pyvox module.

$Id: pyvox.h,v 1.46 2006/02/21 19:35:39 hughett Exp $  $Name: pyvox-0-72 $
******************************************************************************/


#ifndef PYVOX_H
#define PYVOX_H


#ifdef HAVE_COMPLEX_H
#include <complex.h>
#endif


#include "tk.h"             /* Needed for Tkinter PhotoImage */
#include "voxel.h"          /* Needed for voxel_array */
#include "errm.h"

/*** FIXME: Does this do any good?  At least, it shouldn't do any harm. ***/
#ifdef __cplusplus          /* Allow to be called from C++ */
extern "C" {
#endif



/*-----------------------------------------------------------------------
pyvox_array - Python object struct to represent a voxel array

This struct forms a Python wrapper around the voxel_array struct used
in the Voxel Kit.  The voxel array itself is kept elsewhere, which
complicates memory management but makes it possible to use the voxel
kit as a C library independently of Python.
-----------------------------------------------------------------------*/

typedef struct {
    PyObject_HEAD            /* Python header for all objects */
    voxel_array *varray;     /* Pointer to the voxel array itself */
    PyObject *header;        /* Python dict with info from image file header */
    PyObject *metadata;      /* Python dict with other metadata */
    PyObject *userdata;      /* Python dict for the user's use */
}  pyvox_array;



/*------------------------------------------------------------------------
Global variables and objects for the Pyvox module and friends
------------------------------------------------------------------------*/

extern PyTypeObject parray_type;
extern PyTypeObject kernel_type;
extern PyTypeObject voxtype_type;

extern PyObject *pyvox_dict;
extern PyObject *PyvoxError;



/*------------------------------------------------------------------------
ReturnPyvoxError - Set the error string and return a value

These macros set the error string in the error object PyvoxError and
return either zero or one.  The ReturnPyvoxError version should be
used in methods that normally return a Python object but return zero
in case of an error; the XReturnPyvoxError version is used in some
other contexts where a side effect is normal and the function returns
non-zero to indicate an error.  The SetPyvoxError version simply sets
the error message and leaves it to the user to return the appropriate
error value.

On the other hand, the macros SetPyvoxLibError, ReturnPyvoxLibError,
and XReturnPyvoxLibError use the error message previously set by some
library function called by Pyvox via the set_error function of errm.c
rather than using a explicit message.

The proper usage of these macros has been a bit confused but is
gradually becoming clear.  A top-level function in Pyvox that is
called directly by the Python interpreter should use set_error_context
and clear_error_context to define the error context; use
ReturnPyvoxError or XReturnPyvoxError to return errors that it finds
itself; and use ReturnPyvoxLibError or XReturnPyvoxLibError to return
errors found by one of the functions that it calls.  The internal
functions, whether part of Pyvox, the Voxel Kit, or whatever, should
use NZReturnError or ZReturnError (from errm.h) to declare errors. 
It's not clear that any of the other macros serve any useful purpose.

FIXME: Go back through the code and check that these rules are followed.

FIXME: This code duplicates error handling code in pyexim.c except that
it names Exim rather than Pyvox.  Figure out how to use the code only
once and avoid having to maintain both.

ALGORITHM: The do {...} while 0 format is used so that this macro can
be used wherever a statement is acceptable.

These macros call clear_error_context to make sure that the context is
cleared even on an immediate error return from a function.

FIXME: Convert the error codes into specific Python exceptions rather
than the generic PyvoxError?
--------------------------------------------------------------------------*/

#define ReturnPyvoxError(errcode, message)                 \
    do { set_error(errcode, message);                      \
         PyErr_SetString(PyvoxError, get_error_message()); \
         clear_error_context();                            \
         return 0; }                                       \
    while (0)

#define XReturnPyvoxError(errcode, message)                \
    do { set_error(errcode, message);                      \
         PyErr_SetString(PyvoxError, get_error_message()); \
         clear_error_context();                            \
         return 1; }                                       \
    while (0)

/* FIXME: Arrange for the errcode to do something useful */
/* FIXME: Use Py_FatalError instead? */
#define FatalPyvoxError(errcode, message) \
    errm_panic(message)


#define ReturnPyvoxLibError()                              \
    do { PyErr_SetString(PyvoxError, get_error_message()); \
         clear_error_context();                            \
         return 0; }                                       \
    while (0)

#define XReturnPyvoxLibError()                             \
    do { PyErr_SetString(PyvoxError, get_error_message()); \
         clear_error_context();                            \
         return 1; }                                       \
    while (0)


/* The following macros are of dubious utility and have been commented
   out until I find a real use for them or kill them entirely. */

#if 0

#define SetPyvoxError(errcode, message)           \
    do { set_error(errcode, message);             \
         PyErr_SetString(PyvoxError, message);    \
         clear_error_context(); }                 \
     while (0)

#define SetPyvoxLibError()                                 \
    do { PyErr_SetString(PyvoxError, get_error_message()); \
         clear_error_context(); }                          \
    while (0)

#endif


/*----------------------------------------------------------------------
pyvox_type - Python object struct to represent a Pyvox type

This struct provides a place to record information about a Pyvox
internal or external type.

FIXME: Consider providing slots here for all the fields currently
recorded in exim.c.  This would be faster to access, but might require
more maintenance.
-----------------------------------------------------------------------*/

typedef struct {
    PyObject_HEAD            /* Python header for all objects */
    int code;                /* Type code number from exim.h */
    char *repr;              /* String to return for repr() */
} pyvox_type;



/*-----------------------------------------------------------------------
kernel_object - Python object struct to represent a kernel

This struct forms a Python wrapper around the vxl_kernel struct used
in the Voxel Kit to represent a neighborhood or convolution kernel.
The vxl_kernel struct itself is kept elsewhere, which complicates
memory management but makes it possible to use the Voxel Kit as a C
library independent of Python.
-----------------------------------------------------------------------*/

typedef struct {
    PyObject_HEAD               /* Header for all Python objects */
    struct vxl_kernel *kern;    /* Pointer to the kernel itself */
} kernel_object;




/*---------------------------------------------------------------------------
Prototypes of shared functions 
----------------------------------------------------------------------------*/

extern int PyvoxValue_Check(PyObject *ob, const char *name,
    int *vtype, anytype *value);
extern int PyvoxLong_Check(PyObject *ob, long *value);
extern int PyvoxLongSeq_Check(PyObject *ob, long minlen, long maxlen,
    long *length, long *values);
extern int PyvoxDouble_Check(PyObject *ob, double *value);
extern int PyvoxDoubleSeq_Check(PyObject *ob, char *name, 
    long minlen, long maxlen, long *length, double *values);
extern double *PyvoxDoubleSeq_AsDouble(PyObject *ob, char *name,
    long minlen, long maxlen, long *length);
#ifdef HAVE_COMPLEX_H
extern int PyvoxComplex_Check(PyObject *ob, double complex *value);
extern int PyvoxComplexSeq_Check(PyObject *ob, char *name, 
    long minlen, long maxlen, long *length, double complex *values);
#endif

extern int PyvoxIntype_Check(PyObject *ob, char *name, int *code);
extern int PyvoxExtype_Check(PyObject *ob, char *name, int *code);
extern PyObject * PyvoxType_FromInt(int code);

extern void PyvoxArray_Dump    (PyObject *ob, char *name, int die);
extern int PyvoxArray_Check    (PyObject *ob, const char *name,
    voxel_array **vob);
extern int PyvoxArray_CheckSrc (PyObject *ob, const char *name,
    voxel_array **vob);
extern int PyvoxArray_CheckDest(PyObject *ob, const char *name,
    voxel_array **vob);

extern int PyvoxKernel_Check(PyObject *ob, const char *name,
    struct vxl_kernel **vob);

extern int Pyvox_ParseShape    (PyObject *shape, int *rank, long *dimen);

extern PyObject *pyvox_get_value(int intype, void *datum, long index);
extern int pyvox_set_value(int intype, void *data, long index,
    PyObject *value);

extern PyObject *pyvox_list_from_array(long nelem, int type, void *data,
    long stride);

extern PyObject *parray_get_voxel(voxel_array *src, long *coord);


/* FIXME: Shared functions that should probably be made private */
extern pyvox_array *parray_alloc (voxel_array **varray,
    int type, int rank, long *dimen);
extern pyvox_array *parray_create(voxel_array **varray);
extern pyvox_array *parray_wrap  (voxel_array  *varray);
extern PyObject    *pyvox_from_temp_varray (voxel_array *varray);
    // extern voxel_array *PyVoxel_GetArray   (PyObject *ob);
    // extern voxel_array *PyVoxel_GetArraySrc(PyObject *ob);



/*----------------------------------------------------------------------------
Python Module Initializers
----------------------------------------------------------------------------*/

void myinitexim(void);
void myinitpyvoxC(void);



#ifdef __cplusplus
}
#endif

#endif /* PYVOX_H */
