/****************************************************************************
voxel.h - Public header file for the Voxel Kit

Author:  Paul Hughett

Copyright (c) 1998-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging applications.

$Id: voxel.h,v 1.139 2006/02/21 16:16:59 hughett Exp $  $Name: pyvox-0-72 $
*****************************************************************************/


#ifndef VOXEL_H
#define VOXEL_H

#include <stddef.h>
#include <stdio.h>

#include "errm.h"
#include "exim.h"



/*--------------------------------------------------------------------------
Capacity parameters:  
These parameters set the capacity limits of the voxel kit; they should 
be redefinable by modifying these definitions and recompiling, within the
limits of the machine, OS, and compiler.  They are also used as a sanity
check.
--------------------------------------------------------------------------*/

#define  VXL_MAX_RANK  8    /* Maximum allowed number of dimensions */
#define  VXL_MAX_RGBA  4    /* Number of echoes in RGBA image */


#ifdef __cplusplus          /* Allow to be called from C++ */
extern "C" {
#endif



/*--------------------------------------------------------------------------
voxel_array - Descriptor for a multi-dimensional data array

This structure defines the current internal representation of a voxel
data set, typically read from a voxel data file or computed in core.
Note that this structure, while similar to the external representation,
is NOT guaranteed to be identical and must be properly converted from
the external representation.  In particular, the file may have been written
on a machine with different architecture and byte order, or the file may
have been written in an older version of the format.  Modulo that caveat,
the contents of this structure are essentially those of the external
header, except that all fields are in the native machine format.

A voxel array is designed to represent data on a rectangular grid of
points.  Two-dimensional images and medical volume images are typical
examples.  Up to 8-dimensional data may be represented, including
mutiple spectral bands.

The magic field is set to VXL_MAGIC to indicate that this structure is
a voxel array and has been initialized.

The type field indicates the format of each datum, and is selected
from one of the constants defined in exim.h; the nbytes field
indicates the number of bytes allocated to each datum.

The length field contains the number of bytes (storage units) actually
allocated for the data itself.  It must be equal to nbytes times the
product of the dimensions.  (Prior to 2000-11-25, it was permitted to
be greater; as far as I know, no code actually took advantage of this
freedom.)

The dimen array contains the number of elements per dimension of the
data.  The most slowly varying coordinates are given first; thus, for
example, the dimensions of a three-dimensional image scanned in the
usual raster order would be slice, row, and column.

The origin array contains the physical coordinate corresponding to
index zero on each coordinate; the spacing array contains the physical
spacing between successive planes along each coordinate, in unstated
units.  If the spacing is zero, it means that both the origin and
spacing are unknown or not meaningful.  (For example, suppose that
successive slices are T1-, T2-, and proton-weighted MRI images.)

The data field points to the actual voxel data, or is zero to indicate
that no storage has been allocated.  By convention, voxels are listed
starting with the upper left corner of the first slice, then right
across the slice, then down successive scanlines, then forward through
successive slices, and so on through all the dimensions.  By an
independent convention, the data are stored in row-major order; that
is, the last coordinate varies most rapidly as the data are acessed
sequentially in memory.  The two conventions combined dictate that
(for example) the first coordinate is slice number, the second row
number, the third column number, and the fourth the echo number.  The
coordinate indices start at 0 for each dimension.  If data is
non-zero, then type, rank, dimen, nbytes, and length must be set to
the correct values

FIXME: Add a vector of strides along each dimension, which will facilitate
some access operations.

FIXME: Modify to support array views which share the same data but differ
in rank and dimensions.

Basic states of a voxel_array:  A voxel array is uninitialized if the
magic word is not the correct magic value; it is initialized if the
the magic word is correct, and other fields have been set to
reasonable values.  It is unallocated if data == 0, and allocated
otherwise; the type, rank, dimen, nbytes, and length of an allocated
array must be set to the correct values, but other fields (except
magic) need not have any specific values.  

It is required that the data field MUST either be zero or a value
returned by malloc or its friends.  Furthermore, the lifetime of a
pointer read from data is not guaranteed to last beyond the next Voxel
Kit call that writes to that voxel array.


There are various special applications of the voxel_array structure,
which make additional assumptions about the format of the data
described; two of these special formats are described next.


RAW DATA FORMAT

It is sometimes necessary to read image data from an external file
without committing to any particular external representation.  For
example, an image display program might want to allow the user to
dynamically change the assumed external format while searching for the
proper interpretation of an image data file abandoned on his doorstep.
The ``raw data format'' is used for this purpose and essentially
treats the data as simply a long stream of 8-bit unsigned chars.  The
last (innermost) dimension of the voxel array is assumed to index over
the bytes in each datum.  The outer dimensions may be used to structure
the data but this is not required; it is permissible to have just a
1-d structure in which the contents are considered as simply a stream
of bytes.


RGB2D DATA FORMAT

There are those who claim that image data is meant to be seen, not
simply computed.  Preparing volume data for display usually involves
extracting a 2d section through the multidimensional data set, scaling
or otherwise manipulating the intensity range of the data to fit
within the limits of a display, and sometimes other operations; the
``rgb2d format'' is intended to capture the result of these
manipulations without making any commitment to a particular display
device.  An rgb2d image is always a 3-dimensional unsigned char array
in which the last dimension has depth three or four, representing in
order the red, green, blue, and optionally alpha channels of the
displayable image.  (The use of unsigned char limits the intensity
range to 0..255.  This will be adequate for most but not all display
devices; it may someday be necessary to provide a similar format with
longer pixels.)

FIXME: Generalize to specify a "band axis" which contains RGB or other
spectral bands and is treated specially by some operations?  The tricky
part will be defining exactly which operations use this to define bands
and which just treat it as an additional axis.

FIXME:  Should there also be a mono2d format for gray scale images?
It's probably unnecessary for display systems, but it might be
useful for making screen dumps of monochrome images as monochrome
rather than rgb images three times larger.
---------------------------------------------------------------------------*/

#define VXL_MAGIC    19182   /* The magic number */

/* Multi-dimensional array for image data */
struct vxl_array_struct {
    long magic;                    /* Magic number = VXL_MAGIC */
    long rank;                     /* Number of dimensions >= 0 */
    long type;                     /* Architecture-independent type */
    long nbytes;                   /* Number of bytes per datum */
    long length;                   /* Length of data area, in bytes */
    long   dimen  [VXL_MAX_RANK];  /* Array of dimensions */
    double origin [VXL_MAX_RANK];  /* Array of origin values */
    double spacing[VXL_MAX_RANK];  /* Array of spacings/intervals */
    void *data;                    /* The actual data */
};

typedef struct vxl_array_struct voxel_array;



/*--------------------------------------------------------------------------
struct vxl_raw_format - Descriptor for the external (raw) format of an image

Thus structure is used primarily (if not only) by the
vxl_raw_to_rgba2d function and describes how the external (raw) format
of data comprising a voxel_array is to be converted to a 2D monochrome
or color image and displayed.  It is used within an image display
program to permit changing the data intepretation on the fly while
interactively searching for the right format.  To support this, it
also includes intensity windowing and lookup table transformation
(although the latter has yet to be implemented).

Raw image data refers to image data stored in the external (file)
format, which need not be the same as the internal (native) format of
the processor.  Within the file itself, raw image data is simply a
stream of (at least 8-bit) unsigned char bytes.  Raw image data is
stored internally as an unsigned char voxel array in which the last
(innermost) dimension is assumed to be an index over the bytes
comprising a voxel.  In the usual case, the voxel array has only one
(very long) dimension and appears as simply a byte stream.

The axes array defined in this struct indicates which dimensions of the
data are to be extracted as the y, x, and channel coordinates of the
display.  The values of the axes[0] and axes[1] elements respectively
indicate the y and x dimensions; the value of axes[2] indicates the
dimension that is to be mapped to the red, green, blue, and alpha channels
of the display.  If axes[2] is set to a value in the range 0...rank-1, then 
the chan[0..3] fields indicate the specific coordinate values along this 
dimension that should be used for red, green, blue, and alpha; setting 
them all equal yields a monochrome image, of course.  On the other hand, 
if axes[2] is set to the rank of the data array, then there is only one 
channel and it is not explicitly indexed along any dimension; the chan 
fields are ignored and a monochrome image is seen.

Strictly speaking, this function does not enforce any interpretation of
the output channels, but the following will be correct in many cases:

   nchan   channels
       1   gray
       2   gray, alpha
       3   red, green, blue
       4   red, green, blue, alpha

Related functions (even if not yet implemented):

   vxl_raw_to_rgba2d - Convert raw image data to 2d rgba section
   vxl_voxel_to_rgb2d - Convert voxel array to 2d rgb section
   vxl_rgb2d_to_XImage - Convert 2d rgb image data to XImage format
   vxl_raw_to_voxel - Convert raw image data to voxel array
                         [from file or data already in core?]
   vxl_voxel_to_raw - Convert voxel array to raw image data
   vxl_read_raw_data - Read raw image data from a file
   vxl_write_image_data - Write raw image data to a file

NOTE: Prior to 2002-03-18 the chan array consisted of three separate
fields recho, gecho, and becho and the nchan field was absent.  Some
other fields were changed in type.  The magic number was incremented
to reflect these incompatible changes.

FIXME: Description of external raw format vs reformatting for display?
Dynamic section extraction from an open file (to save memory)?

FIXME:  Define ``natural'' translations between native and raw format
that preserve as much information as possible?
--------------------------------------------------------------------------*/

#define VXL_RAWFORM   19186     /* Magic number for a raw format struct */

struct vxl_raw_format {
    long  magic;                  /* Magic number = VXL_RAWFORM */

    /* Parameters describing the external (raw) format */
    long  type;                   /* Code for external data format */
    long  bigend;                 /* 1 => big-endian; 0 => little-endian */
    long  rank;                   /* Number of dimensions >= 0 */
    long  dimen[VXL_MAX_RANK];    /* Dimensions of the voxel array */

    /* Parameters describing how to reformat for 2d RGB display */
    long   coords[VXL_MAX_RANK];  /* Coordinates of current section */
    long   direct[VXL_MAX_RANK];  /* Forward (>=0) or reverse (<0) scan */
    double minint[VXL_MAX_RGBA];  /* Minimum intensity value for window */
    double maxint[VXL_MAX_RGBA];  /* Maximum intensity value for window */
    unsigned char 
         *colormap[VXL_MAX_RGBA]; /* Color map for each channel of RGBA */
    long  axes[3];                /* Input axes to map to Y, X, Channel */
    long  nchan;                  /* Number of channels in output RGBA array */
    long  chan[VXL_MAX_RGBA];     /* Channel number to map onto RGBA */
};



/*----------------------------------------------------------------------------
voxel_object_data - Description of an object

This structure contains some useful information on a single object
(a connected region of non-zero voxels) found in a volume image.

Most fast algorithms for finding objects in an image have the 
characteristic that they sometimes assign different id numbers to 
two apparently distinct objects that are later found to be parts 
the same object.  When this happens, one object id is taken as canonical 
and the "canon" field of the other is set to this canonical id; 
furthermore, the count field of the non-canonical object is added
to the canonical object and then set to zero.  The canonical object
number is distinguished by the fact that its "canon" field is equal to
its "ident" field.  The "point" coordinates of the non-canonical object 
are left unchanged, although it's not clear that they are useful for 
anything.  The ident (and canon) fields are limited to the size of 
unsigned short to facilitate later table lookup on object idents.  If 
you have more than 65535 objects in an image, you're out of luck, at 
least with this version of the software.  Ident 0 is always reserved for 
the background.
----------------------------------------------------------------------------*/

struct voxel_object_data {
    unsigned short ident;       /* Object identification number */
    unsigned short canon;       /* Canonical object id number */
    long point[3];              /* Coordinates of first point in the object */
    long count;                 /* Count of voxels in this object */
};



/*-----------------------------------------------------------------------
vxl_kernel - Define a neighborhood or convolution kernel

This structure defines a neighborhood of a voxel by giving the
coordinate offsets (or deltas) to each neighbor relative to the
coordinates of the central voxel; it optionally gives a coefficient to
go with each neighbor, as for a convolution kernel, as well as a
constant bias term for the convolution.

If coef is zero, then there are no coefficients, and the structure
merely defines the set of voxels which constitute the neighborhood;
the bias should be ignored in this case.  Otherwise the structure
defines a convolution kernel, for which the bias and coefficients are
always of type double (type = INTP_DOUBLE) or double complex (type
= INTP_DCOMPLEX).

A kernel must have rank greater than 0, since convolution is not
defined for scalars.

The macro VXL_MAX_NEIGH is an upper limit on the number of neighbors
that may be included in a kernel.  The true maximum value is probably
INT_MAX, but we use a smaller value to help catch blunders.  The
choice of 1024 for the number of neighbors is more or less arbitrary
but should suffice for most applications; if you have more than that
many neighbors, you probably need a better algorithm anyway.

NOTE: The magic and alloc fields were added on 2006-02-12; old kernel
code will need to be recompiled, and perhaps modified.

FIXME: Consider adding an overall scale factor to this struct.  It's
unneccessary for convolution, where it can be combined with the
coefficients, but would be useful for local statistics.
-----------------------------------------------------------------------*/

#define VXL_MAX_NEIGH     1024     /* Maximum number of neighbors */
#define VXL_KERN_MAGIC   14810     /* Magic number to identify kernels */

struct
vxl_kernel {
    int magic;                    /* Magic kernel number = VXL_KERN_MAGIC */
    int rank;                     /* Number of dimensions */
    int alloc;                    /* Number of neighbors allocated */
    int count;                    /* Number of neighbors in this kernel */
    int type;                     /* Common type of coefs and bias */
    long (*delta)[VXL_MAX_RANK];  /* Coordinate deltas for each neighbor */
    void *coef;                   /* Coefficient array, or zero */
    anytype bias;                 /* Constant bias term for convolution */
};



/*----------------------------------------------------------------------
Kernel manipulation
-----------------------------------------------------------------------*/

extern struct vxl_kernel *vxl_create_kernel(int rank, int count, int type,
    long (*delta)[VXL_MAX_RANK], void *coef, void*bias);
extern int vxl_destroy_kernel(struct vxl_kernel *kern);
extern int vxl_set_kernel_bias(struct vxl_kernel *kern, int type, void *bias);
extern int vxl_set_kernel_coef(struct vxl_kernel *kern, long *delta,
    int type, void *coef);
extern int vxl_del_kernel_coef(struct vxl_kernel *kern, long *delta);

extern struct vxl_kernel *vxl_lowpass_kernel(int rank);



/*------------------------------------------------------------------------
Standard neighborhoods and convolution kernels

FIXME:  The better way to define all these kernels would be to write
functions that create kernels with desired characteristics on demand. 
-------------------------------------------------------------------------*/

extern struct vxl_kernel neigh3d27;



/*-----------------------------------------------------------------------
Array creation, destruction, and metainformation

SEE ALSO vxli.h for some alternate versions for internal use only.
------------------------------------------------------------------------*/

extern voxel_array *vxl_create_array(void);
extern voxel_array *vxl_copy_array(voxel_array *src);
extern void vxl_destroy_array(voxel_array *dest);
extern void vxl_init_array(voxel_array *dest);
extern int  vxl_alloc_array(voxel_array *dest, int type, int rank, 
    long *dimen);
extern voxel_array *vxl_create_scalar(int type, double value);
extern long vxl_count(voxel_array *src);
extern unsigned char *vxl_locate(voxel_array *src, long *coords);
extern int  vxl_store_scalar(voxel_array *dest, int type, double value);
extern void vxl_free_array(voxel_array *dest);
extern int  vxl_reshape_array(voxel_array *dest, int rank, long *dimen);


/*------------------------------------------------------------------------
Array initialization, including random number generation
------------------------------------------------------------------------*/

extern int vxl_const  (voxel_array *dest, int vtype, anytype *value);
extern int vxl_fftramp(voxel_array *dest, int axis);
extern int vxl_ramp   (voxel_array *dest, int axis);
extern int vxl_randi  (voxel_array *dest, void *state, int n);
extern int vxl_randn  (voxel_array *dest, void *state,
    double mean, double sdev);
extern int vxl_randu  (voxel_array *dest, void *state, double a, double b);



/*-----------------------------------------------------------------------
Pixel and slice access
------------------------------------------------------------------------*/

extern int vxl_get_voxel(double *value, voxel_array *src, long *coords);
extern int vxl_set_voxel(voxel_array *dst, long *coords, double value);
extern int  vxl_get_slice(voxel_array *dst, voxel_array *src,
    long *nidx, long **index);
extern int  vxl_set_slice(voxel_array *dst, long *nidx, long **index,
    voxel_array *src);



/*-----------------------------------------------------------------------
Standard unary operations on voxel arrays

The following functions all take two voxel arrays as arguments, the
first as destination and the second as source.  Some other code,
notably Pyvox, exploits this parallelism to simplify its code; the
typedef for vxl_unop specifies the interface and is used by Pyvox.
Most, and perhaps all, are elementwise operations.
-----------------------------------------------------------------------*/

typedef int  vxl_unop (voxel_array *dest, voxel_array *src);

/* Type casts */
extern vxl_unop vxl_uchar;
extern vxl_unop vxl_schar;
extern vxl_unop vxl_ushort;
extern vxl_unop vxl_short;
extern vxl_unop vxl_uint;
extern vxl_unop vxl_int;
extern vxl_unop vxl_ulong;
extern vxl_unop vxl_long;
extern vxl_unop vxl_float;
extern vxl_unop vxl_double;
extern vxl_unop vxl_fcomplex;
extern vxl_unop vxl_dcomplex;

/* Arithmetic and bitwise operations */
extern vxl_unop vxl_copy;
extern vxl_unop vxl_not;
extern vxl_unop vxl_minus;
extern vxl_unop vxl_real;
extern vxl_unop vxl_imag;
extern vxl_unop vxl_conj;
extern vxl_unop vxl_carg;
extern vxl_unop vxl_abs;
extern vxl_unop vxl_floor;
extern vxl_unop vxl_ceil;
extern vxl_unop vxl_sqrt;
extern vxl_unop vxl_cbrt;

/* Elementary functions */
extern vxl_unop vxl_exp;
extern vxl_unop vxl_exp2;
extern vxl_unop vxl_expm1;
extern vxl_unop vxl_log;
extern vxl_unop vxl_log10;
extern vxl_unop vxl_log1p;
extern vxl_unop vxl_log2;
extern vxl_unop vxl_sin;
extern vxl_unop vxl_cos;
extern vxl_unop vxl_tan;
extern vxl_unop vxl_asin;
extern vxl_unop vxl_acos;
extern vxl_unop vxl_atan;
extern vxl_unop vxl_sinh;
extern vxl_unop vxl_cosh;
extern vxl_unop vxl_tanh;
extern vxl_unop vxl_erf;
extern vxl_unop vxl_erfc;
extern vxl_unop vxl_lgamma;
extern vxl_unop vxl_tgamma;



/*------------------------------------------------------------------------
Standard binary operations on voxel arrays

Most but not quite all of these are elementwise operations.
-------------------------------------------------------------------------*/

typedef int vxl_binop(voxel_array *dest,
                       voxel_array *src1, voxel_array *src2);

extern vxl_binop vxl_min;
extern vxl_binop vxl_max;
extern vxl_binop vxl_and;
extern vxl_binop vxl_or;
extern vxl_binop vxl_xor;
extern vxl_binop vxl_add;
extern vxl_binop vxl_sub;
extern vxl_binop vxl_mul;
extern vxl_binop vxl_div;
extern vxl_binop vxl_rem;
extern vxl_binop vxl_pow;
extern vxl_binop vxl_atan2;
extern vxl_binop vxl_hypot;

extern vxl_binop vxl_cross;



/*------------------------------------------------------------------------
Standard ternary operations on voxel arrays

Most, and perhaps all, are elementwise operations.
-------------------------------------------------------------------------*/

typedef int vxl_ternop(voxel_array *dest, voxel_array *src1,
                       voxel_array *src2, voxel_array *src3);

extern vxl_ternop vxl_fma;



/*------------------------------------------------------------------------
Array input and output
-------------------------------------------------------------------------*/

extern int  vxl_read_uchar_data(voxel_array *dest, FILE *src, int rank,
    long *dimen);
extern int vxl_read_raw_data(voxel_array *dest, FILE *src, 
    int rank, long *dimen, int intype, int extype, int bigend);
extern int vxl_write_uchar_data(FILE *dest, voxel_array *src);
extern int vxl_write_raw_data(FILE *dest, int extype, int bigend, 
    voxel_array *src);

extern int  vxl_read_pnm_header(FILE *src, char *magic,
     long *width, long *height, long *maxval);
extern int  vxl_read_pgm_image(voxel_array *dest, char *filename);
extern int  vxl_read_ppm_image(voxel_array *dest, char *filename);
extern int  vxl_write_pgm_image(const char *fname, voxel_array *src);
extern int  vxl_write_ppm_image(const char *fname, voxel_array *src);



/*----------------------------------------------------------------------
Array reduction operations
----------------------------------------------------------------------*/

extern int vxl_amax (voxel_array *dest, voxel_array *src, 
    int naxes, int *axes);
extern int vxl_amin (voxel_array *dest, voxel_array *src, 
    int naxes, int *axes);
extern int vxl_aprod(voxel_array *dest, voxel_array *src, 
    int naxes, int *axes);
extern int vxl_asum (voxel_array *dest, voxel_array *src, 
    int naxes, int *axes);
extern int vxl_asum0(voxel_array *dest, voxel_array *src, 
    int naxes, int *axes);

extern int vxl_minmax(voxel_array *minres, voxel_array *maxres,
    voxel_array *src, int naxes, int *axes);

extern int vxl_dot(voxel_array *result, voxel_array *src1, voxel_array *src2,
    voxel_array *weight, int naxes, int *axes);
extern int vxl_norm1(voxel_array *result, voxel_array *src1, voxel_array *src2,
    voxel_array *weight, int naxes, int *axes);
extern int vxl_norm2(voxel_array *dest, voxel_array *src1, voxel_array *src2,
    voxel_array *weight, int naxes, int *axes);
extern int vxl_normsup(voxel_array *result, voxel_array *src1,
    voxel_array *src2, voxel_array *weight, int naxes, int *axes);

extern int vxl_mean(voxel_array *dest, voxel_array *src1, voxel_array *src2,
    int naxes, int *axes);



/*---------------------------------------------------------------------------
Fourier and other orthogonal transforms
----------------------------------------------------------------------------*/

extern int vxl_fft(voxel_array *dest, int dtype, voxel_array *src,
    int inverse, int naxes, int *axes);



/*-----------------------------------------------------------------------
Specialized type casting operations
------------------------------------------------------------------------*/

extern int vxl_fcomplex2(voxel_array *dest, voxel_array *src1,
     voxel_array *src2);
extern int vxl_dcomplex2(voxel_array *dest, voxel_array *src1,
     voxel_array *src2);
extern int vxl_cast (voxel_array *dest, int dtype, voxel_array *src);
extern int vxl_scale_to_uchar(voxel_array *dest, voxel_array *src,
    double minval, double maxval);
extern int vxl_scale(int dtype, voxel_array *dest, voxel_array *src,
    int gtype, void *gain, int btype, void *bias);



/*------------------------------------------------------------------------
Two-way and three-way comparison operations on voxel arrays
------------------------------------------------------------------------*/

typedef int vxl_comp2(voxel_array *dest, voxel_array *src1, 
    voxel_array *src2, unsigned char iftrue, unsigned char iffalse);

extern vxl_comp2 vxl_equal;
extern vxl_comp2 vxl_greater;
extern vxl_comp2 vxl_less;

extern int vxl_compare(voxel_array *dest, voxel_array *src1, 
    voxel_array *src2, unsigned char lsval, unsigned char eqval,
    unsigned char gtval);



/*---------------------------------------------------------------------
Conditional move and other selection operations
----------------------------------------------------------------------*/

extern int vxl_cmove(voxel_array *dest, voxel_array *sel1, voxel_array *src2);



/*-----------------------------------------------------------------------
Graphics functions 
-------------------------------------------------------------------------*/

extern int  vxl_fill_2d_contour(voxel_array *canvas, long npoint,
    double *yx, double fill);



/*-----------------------------------------------------------------------
Coordinate transformations, interpolation, and resampling 
-------------------------------------------------------------------------*/

/* Interpolation at a single point */
extern int vxl_nearest_point(int dtype, void *dest, voxel_array *src,
    double *point);
extern int vxl_linear_point(int dtype, void *dest, voxel_array *src,
    double *point);
extern int vxl_cubic_point(int dtype, void *dest, voxel_array *src,
    double *point);

/* Generation of coordinate maps */
extern int vxl_nearest_map(voxel_array *dest, int dtype, voxel_array *src,
    voxel_array *map);
extern int vxl_linear_map(voxel_array *dest, int dtype, voxel_array *src,
    voxel_array *map);
extern int vxl_cubic_map(voxel_array *dest, int dtype, voxel_array *src,
    voxel_array *map);
extern int vxl_affine_map(voxel_array *result, int rank, long *dimen,
    double *matrix, double *offset);
extern int vxl_polynomial_map(voxel_array *dest, int rank, long *dimen,
    int nterm, int *outcor, int *power, double *coef);

/* Reference implementations for combined transforms and resampling */
extern int vxl_affine_nearest0(voxel_array *dest, int type, int rank,    
    long *dimen, voxel_array *src, double *matrix, double *offset);
extern int vxl_affine_linear0(voxel_array *dest, int type, int rank,    
    long *dimen, voxel_array *src, double *matrix, double *offset);
extern int vxl_polynomial_linear0(voxel_array *dest, 
    int type, int rank, long *dimen, voxel_array *src,  
    int nterm, int *outcor, int *power, double *coef);

/* Fast implementations for combined transforms and resampling */
extern int vxl_affine_nearest(voxel_array *dest, int type, int rank,    
    long *dimen, voxel_array *src, double *matrix, double *offset);
extern int vxl_affine_linear(voxel_array *dest, int type, int rank,    
    long *dimen, voxel_array *src, double *matrix, double *offset);



/*-------------------------------------------------------------------------
Convolution and convolution-like operations 
-------------------------------------------------------------------------*/

extern int vxl_convolve(voxel_array *dest, voxel_array *src,
    struct vxl_kernel *neigh, long *shrink);
extern int vxl_dilabel(voxel_array *dest, voxel_array *src,
    struct vxl_kernel *neigh);
extern int vxl_dilate(voxel_array *dest, voxel_array *src,
    struct vxl_kernel *neigh);
extern int vxl_erode(voxel_array *dest, voxel_array *src,
    struct vxl_kernel *neigh);
extern int vxl_lostat(voxel_array *mean, voxel_array *sdev,
    voxel_array *src, struct vxl_kernel *kern, double scale);



/*--------------------------------------------------------------------------
Statistical operations 
--------------------------------------------------------------------------*/

extern int vxl_minmax_ulong(unsigned long *minval, unsigned long *maxval,
    voxel_array *src);
extern int vxl_stat(voxel_array *mean, voxel_array *stddev, 
    voxel_array *source, voxel_array *weight, int naxes, int *axes);
extern int vxl_moments(double *mass, double *center, double *moments,
    voxel_array *src);
extern int vxl_mop(voxel_array *result, voxel_array *moments, 
    voxel_array *array1, voxel_array *array2, voxel_array *array3);
extern int  vxl_histo_uchar(voxel_array *hist, voxel_array *src,
    voxel_array *weight);
extern int  vxl_bihist_uchar(voxel_array *hist, voxel_array *mar1, 
    voxel_array *mar2, voxel_array *src1, voxel_array *src2,
    voxel_array *weight);
extern int  vxl_kmeans1_histo(voxel_array *hist, int nclass,
    double *cent);
extern int  vxl_kmeans2_histo(voxel_array *hist, int nclass,
    double *cent1, double *cent2);
extern int  vxl_nnclass1_uchar(voxel_array *dest, voxel_array *src,
    int nclas, unsigned char *clas, double *cent);
extern int  vxl_nnclass2_uchar(voxel_array *dest, voxel_array *src1,
    voxel_array *src2, int nclas,  unsigned char *clas,
    double *cent1, double *cent2);



/*------------------------------------------------------------------------
Some miscellaneous operations 
-------------------------------------------------------------------------*/

extern int  vxl_chamfer(voxel_array *dest, voxel_array *src, int dtype);
extern int  vxl_find_objects_uchar3d(voxel_array *dest,
    long *nobjects, struct voxel_object_data *oblist[], voxel_array *src);
extern int vxl_logcomp_ulong(voxel_array *dest, voxel_array *src);
extern int vxl_lookup(voxel_array *dest, voxel_array *src, 
    voxel_array *table);



/*------------------------------------------------------------------------
Image formatting for display 
-------------------------------------------------------------------------*/

extern int  vxl_2d_section(voxel_array *dest, voxel_array *src, int dim0,
    int dim1, long *slice);
extern int  vxl_2d_sectwin(voxel_array *dest, voxel_array *src, int dim0,
    int dim1, long *slice, int bigend, int gmin, int gmax);
extern int  vxl_init_raw_format(struct vxl_raw_format *fmt);
extern void vxl_print_raw_format(FILE *stream, struct vxl_raw_format *fmt);
extern int  vxl_raw_to_rgba2d(voxel_array *dst, voxel_array *src,
    struct vxl_raw_format *fmt);



/*------------------------------------------------------------------------
Matrix operations on voxel arrays 
--------------------------------------------------------------------------*/

extern int matrix_determinant(int *type, void *value, voxel_array *matrix);
extern int matrix_multiply(voxel_array *product,
    voxel_array *amatrix, voxel_array *bmatrix);
extern int matrix_cholesky(voxel_array *factor, voxel_array *matrix);
extern int matrix_svd(voxel_array *left, voxel_array *values,
    voxel_array *right, voxel_array *matrix);
extern int matrix_eigen_symmetric(voxel_array *values, 
    voxel_array *vectors, voxel_array *matrix);
extern int matrix_get_diag(voxel_array *diagonal,voxel_array *matrix);
extern int matrix_inverse(voxel_array *inverse, voxel_array *matrix);
extern int matrix_principal_axes(voxel_array *values, 
    voxel_array *vectors, voxel_array *matrix);
extern int matrix_solve(voxel_array *soln, voxel_array *matrix,
    voxel_array *rhs);
extern int matrix_transpose(voxel_array *trans, voxel_array *matrix, int conj);


#ifdef __cplusplus
}
#endif

#endif /* VOXEL_H */
