#############################################################################
# pyvox.py - Python extension module for Voxel Kit bindings
#
# Author:  Paul Hughett
#
# Copyright (c) 2001-2005 Trustees of the University of Pennsylvania
# See the file COPYING for license terms.
#
#
# This file contains the Python-language part of the pyvox extension to
# Python, and imports the appropriate objects from the C language part.
#
# FIXME: Provide a way of getting the error object created by pyvoxC
#
# FIXME: Generalize here or in mmul to allow rank-1 vectors to be treated
# as row or column vectors as appropriate.
#
# FIXME: Check for proper shape and type of arguments.  Throw an exception
# on mismatch?
#
# FIXME: This file is getting too long for comfort.  Look into ways of
# splitting into smaller files.
#############################################################################

rcsid = "$Id: pyvox.py,v 1.108 2006/02/24 20:47:38 hughett Exp $ $Name: pyvox-0-72 $"

import exceptions
import os
import re
import stat
import string
import tempfile
import types
#import warnings        # FIXME: Requires Python 2.1 or better

from copy import copy
from math import sin, cos
from exceptions import Exception
from types import IntType, FloatType, ListType, TupleType, ComplexType
import exim
import pyvoxC



#--------------------------------------------------------------------------
# Import some specific symbols from the C language part of the extension
#--------------------------------------------------------------------------

array   = pyvoxC.array
const   = pyvoxC.const
fftramp = pyvoxC.fftramp
lowpass = pyvoxC.lowpass
ramp    = pyvoxC.ramp
randi   = pyvoxC.randi
randn   = pyvoxC.randn
randu   = pyvoxC.randu
refcnt  = pyvoxC.refcnt

ArrayType  = pyvoxC.ArrayType
KernelType = pyvoxC.KernelType

error_action = pyvoxC.error_action



#------------------------------------------------------------------------
# Define symbols for presence/absence of optional features
#
# FIXME: Add X, tkphoto, gzip, ... ?
#-----------------------------------------------------------------------

have_complex = pyvoxC.feature("have_complex")



#--------------------------------------------------------------------------
# Pass pyvox dictionary to pyvox C
#
# This dictionary is stored within PyvoxC and permits the C level of pyvox 
# to call functions and methods defined in this Python module.
#--------------------------------------------------------------------------

pyvoxC.set_pyvox_dict(globals())



#------------------------------------------------------------------------
# pyvox.Error - Exception object to report errors
#
# FIXME: Pyvox and PyvoxC should use the same exception class to report
# errors.
#------------------------------------------------------------------------

class Error(exceptions.Exception) : 
    def __init__(self, args=None) :
        self.args = args



#-------------------------------------------------------------------------
# pyvox.metacopy - Copy selected metadata
#
# Implements the array.metacopy() method.
#
# FIXME: Should not try to copy from a non-existent dictionary attribute.
#
# FIXME: Uninformative error message if ranks do not match in attempt to
# copy origin and spacing.
#-------------------------------------------------------------------------

def metacopy(self, other, **kwargs) :

    all = (len(kwargs) == 0)

    if all or kwargs.get("filedata", 0) or kwargs.get("metadata", 0) :
        for key in other.metadata.keys() :
            self.metadata[key] = other.metadata[key]

    if all or kwargs.get("header", 0) :
        for key in other.header.keys() :
            self.header[key] = other.header[key]

    if all or kwargs.get("physdata", 0) :
        self.origin  = other.origin
        self.spacing = other.spacing

    if all or kwargs.get("userdata", 0) :
        for key in other.userdata.keys() :
            self.userdata[key] = other.userdata[key]

    return self



#-------------------------------------------------------------------------
# Conversions to special types of Pyvox arrays
#
# These functions are made available to the Pyvox user to convert generic
# array representations into the specific representations assumed by
# particular functions within Pyvox.  They are also used within Pyvox
# itself to convert generic representations to specific ones.  These
# functions may also be used to check the size of the converted data
# and will throw an exception if the size is wrong; if no expected size
# is specified, then the size is determined by the data provided.  The
# type of the returned value is double complex if any element of data
# is complex; otherwise it is double.
#
# If the data is already in the correct format, it is NOT copied but is
# returned just as it is.  (Except for diag, which always creates a new
# array.)
#
# FIXME:  Allow the user to specify the desired type of the result?
#
# FIXME: Always create a new array,to avoid unexpected problems with
# sharing?  Particularly when the original array is reshaped.
#
# FIXME: Can some of this code be simplified by sharing common parts?
#-------------------------------------------------------------------------

# Return exim.dcomplex if any element is complex; otherwise exim.double.
# FIXME: Get the shape of the data at the same time?
def datatype(data) :

    if isinstance(data, IntType) or \
       isinstance(data, FloatType) :
        return exim.double

    elif isinstance(data, ComplexType) :
        return exim.dcomplex

    elif isinstance(data, ArrayType) :
        if data.type.iscomplex :
            return exim.dcomplex
        else :
            return exim.double

    elif isinstance(data, ListType) or \
         isinstance(data, TupleType) :
        for item in data :
            if datatype(item) == exim.dcomplex :
                return exim.dcomplex
        return exim.double


# Convert to rank-1 vector (n) of type double
def vector(data, n=None) :

    dtype = datatype(data)
    if n :
        try :
            if isinstance(data, ArrayType) and data.size == [n] and \
               data.type in (exim.double, exim.dcomplex) : 
                return data
            elif isinstance(data, ArrayType) and len(data) == n :
                return data.cast(dtype).reshape([n])
            elif isinstance(data, TupleType) and len(data) == n :
                return array([n], dtype, data)
            elif isinstance(data, ListType) and len(data) == n :
                return array([n], dtype, data)
            elif isinstance(data, IntType) or \
                 isinstance(data, FloatType) or \
                 isinstance(data, ComplexType) :
                return array([n], dtype, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a %s-element numeric vector" \
                  % (repr(data), n)
            raise Error, msg
    else :
        try :
            if isinstance(data, ArrayType) and data.rank == 1 and \
               data.type in (exim.double, exim.dcomplex) :
                return data
            elif isinstance(data, ArrayType) :
                return data.cast(dtype).reshape([len(data)])
            elif isinstance(data, TupleType) :
                return array([len(data)], dtype, data)
            elif isinstance(data, ListType) :
                return array([len(data)], dtype, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a numeric vector" % repr(data)
            raise Error, msg


# Convert to row vector (1 * n) of type double
def row(data, n=None) :

    dtype = datatype(data)
    if n :
        try :
            if isinstance(data, ArrayType) and data.size == [1, n] \
               and data.type in (exim.double, exim.dcomplex) :
                return data
            elif isinstance(data, ArrayType) and len(data) == n :
                return data.cast(dtype).reshape([1,n])
            elif isinstance(data, TupleType) and len(data) == n :
                return array([1,n], dtype, data)
            elif isinstance(data, ListType) and len(data) == n :
                return array([1,n], dtype, data)
            elif isinstance(data, IntType) or \
                 isinstance(data, FloatType) or \
                 isinstance(data, ComplexType) :
                return array([1,n], dtype, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a %s-element numeric column vector" \
                  % (repr(data), n)
            raise Error, msg
    else :
        try :
            if isinstance(data, ArrayType) and \
               data.rank == 2 and \
               data.size[0] == 1 and \
               data.type in (exim.double, exim.dcomplex) :
                return data
            elif isinstance(data, ArrayType) :
                return data.cast(dtype).reshape([1, len(data)])
            elif isinstance(data, TupleType) :
                return array([1, len(data)], dtype, data)
            elif isinstance(data, ListType) :
                return array([1, len(data)], dtype, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a numeric column vector" % repr(data)
            raise Error, msg


# Convert to column vector (n * 1) of type double
def column(data, n=None) :

    dtype = datatype(data)
    if n :
        try :
            if isinstance(data, ArrayType) and data.size == [n,1] \
               and data.type in (exim.double, exim.dcomplex) :
                return data
            elif isinstance(data, ArrayType) and len(data) == n :
                return data.cast(dtype).reshape([n,1])
            elif isinstance(data, TupleType) and len(data) == n :
                return array([n,1], dtype, data)
            elif isinstance(data, ListType) and len(data) == n :
                return array([n,1], dtype, data)
            elif isinstance(data, IntType) or \
                 isinstance(data, FloatType)  or \
                 isinstance(data, ComplexType) :
                return array([n,1], dtype, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a %s-element numeric column vector" \
                  % (repr(data), n)
            raise Error, msg
    else :
        try :
            if isinstance(data, ArrayType) and data.rank == 2 and \
               data.size[1] == 1 and \
               data.type in (exim.double, exim.dcomplex) :
                return data
            elif isinstance(data, ArrayType) :
                return data.cast(dtype).reshape([len(data), 1])
            elif isinstance(data, TupleType) :
                return array([len(data), 1], dtype, data)
            elif isinstance(data, ListType) :
                return array([len(data), 1], dtype, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a numeric column vector" % repr(data)
            raise Error, msg


# Convert to coordinate point
# Just an alias for column() but clearer to user in many contexts
point = column


# Convert to diagonal matrix (n by n) of type double or double complex
# FIXME: Generalize to generate nonsquare matrices of any type
def diag(data, n=None) :

    dtype = datatype(data)
    if n :
        try :
            if isinstance(data, ArrayType) and data.size == [n] and \
               data.type in (exim.double, exim.dcomplex) : 
                vector = data
            elif isinstance(data, ArrayType) and len(data) == n :
                vector = data.cast(dtype).reshape([n])
            elif isinstance(data, TupleType) and len(data) == n :
                vector = array([n], dtype, data)
            elif isinstance(data, ListType) and len(data) == n :
                vector = array([n], dtype, data)
            elif isinstance(data, IntType) or isinstance(data, FloatType) :
                vector = array([n], dtype, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a %s-dimensional diagonal matrix" \
                  % (repr(data), n)
            raise Error, msg
    else :
        try :
            if isinstance(data, ArrayType) and data.rank == 1 and \
               data.type in (exim.double, exim.dcomplex) :
                vector = data
            elif isinstance(data, ArrayType) :
                vector = data.cast(dtype).reshape([len(data)])
            elif isinstance(data, TupleType) :
                vector = array([len(data)], dtype, data)
            elif isinstance(data, ListType) :
                vector = array([len(data)], dtype, data)
            else :
                raise Error
        except :
            msg = "Cannot convert %s to a diagonal matrix" % repr(data)
            raise Error, msg
    n = len(vector)
    result = array([n,n], dtype, 0)
    for i in range(n) :
        result[i,i] = vector[i]
    return result


# Convert to a rank-2 matrix; returns *diagonal* matrix for scalar data
# FIXME: Allow data to be a flat list if (nr, nc) specified?
# FIXME: Allow nr to be the list (nr, nc)?
def matrix(data, nr=None, nc=None) :

    dtype = datatype(data)
    if nr and not nc :
        nc = nr
    try :
        if isinstance(data, ArrayType) :
            if nr and data.size == [nr,nc] and \
               data.type in (exim.double, exim.dcomplex) :
                return data
            elif nr and len(data) == nr*nc :
                return data.cast(dtype).reshape([nr,nc])
            elif not nr and data.rank == 2 :
                return data.cast(dtype)
            else :
                raise Error
        elif isinstance(data, IntType) or \
             isinstance(data, FloatType) or \
             isinstance(data, ComplexType) :
            if not nr :
                raise Error
            else :
                result = array([nr, nc], dtype, 0)
                if nr < nc : nn = nr
                else:        nn = nc
                for n in range(nn):
                    result[n,n] = data
                return result
        elif isinstance(data, ListType ) or isinstance(data, TupleType) :
            nr1 = len(data)
            nc1 = len(data[0])
            for r in range(nr1) :
                if len(data[r]) != nc1 :
                    raise Error
            if (nr and nr != nr1) or (nc and nc != nc1) :
                raise Error
            result = array([nr1, nc1], dtype, 0)
            for r in range(nr1) :
                for c in range(nc1) :
                    result[r,c] = data[r][c]
            return result
        else :
            raise Error
    except :
        if nr :
            msg = "Cannot convert %s to a %s by %s numeric matrix" % \
                  (repr(data), nr, nc)
        else :
            msg = "Cannot convert %s to a numeric matrix" % \
                  repr(data)
        raise Error, msg


                        
#--------------------------------------------------------------------------
# Type checking and conversion
#
# If x can be converted to type foo, do so and return it; else raise an
# exception.  Unadvertized and intended for internal use.
#
# FIXME: These may eventually get replaced by the conversion functions
# given just above.
#--------------------------------------------------------------------------


# Check if x can reasonably be interpreted as an integer
def check_int(x):
    if isinstance(x, IntType):
        return x
    elif isinstance(x, FloatType):
        return int(x)
    else:
        raise Error, 'Cannot interpret arg as an integer'


# Check if x can reasonably be interpreted as a float value
def check_float(x):
    if isinstance(x, FloatType):
        return x
    elif isinstance(x, IntType):
        return float(x)
    else:
        raise Error, 'Cannot interpret arg as a float value'


# Check if x can reasonably be interpreted as a rank-1 vector of n elements
# (A scalar will be interpreted as a constant vector.)
def check_vector(x, n) :
    if isinstance(x, ArrayType) and x.size == [n] :
        return x
    if isinstance(x, ArrayType) and x.size == [1, n] :
        return x.double().reshape([n])
    try :
        return array([n], exim.double, x)
    except :
        raise Error, "Cannot interpret as a numeric vector"


# Check if x can reasonably be interpreted as an n x 1 column vector
# (A scalar will be interpreted as a constant vector.)
def check_column(x, n) :
    if isinstance(x, ArrayType) and x.size == [n,1] :
        return x
    if isinstance(x, ArrayType) and x.size == [n] :
        return x.double().reshape([n,1])
    try :
        return array([n,1], exim.double, x)
    except :
        raise Error, 'Cannot interpret as an n x 1 column vector'


# Check if x can reasonably be interpreted as an nr x nc matrix
def check_matrix(x, nr, nc) :
    if isinstance(x, ArrayType) and x.size == [nr, nc] :
        return x
    try :
        return array([nr,nc], exim.double, x)
    except :
        raise Error, 'Cannot interpret as a matrix of right size'



#-------------------------------------------------------------------------
# pyvox.kernel - Create convolution kernels
#
# This function will create a new convolution kernel with specified
# properties.
#
# FIXME: Directly support old calling sequence here?
#
# FIXME: Add error checks for type and format of arguments?
#
# FIXME: The rank could be determined from a nonempty data argument, except
# that it is needed to distingish the old and new calling sequences.  Once
# the old calling sequence is removed, consider making rank optional.
#--------------------------------------------------------------------------

def kernel(rank, data=None, arg3=[], **keywords) :

    # Support old calling sequence
    if type(rank) in (types.ListType, types.TupleType) :
        if len(keywords) != 0 :
            raise Exception, "Cannot use keywords with old calling sequence"
        return pyvoxC.old_kernel(rank, data, arg3)

    # Extract option values from dict
    alloc  = keywords.get("alloc",  0)
    bias   = keywords.get("bias",   0.0)
    ktype  = keywords.get("type",   exim.double)
    center = keywords.get("center", None)
    size   = keywords.get("size",   None)
    thresh = keywords.get("thresh", ktype.min)

    # Create kernel; set bias
    # FIXME: alloc will be too large if data contains subthreshold elements
    if data and not alloc :
        alloc = len(data)
    kern = pyvoxC.kernel(rank, ktype, alloc)
    kern.bias = bias

    # Case 0: No data was supplied
    if not data :
        pass

    # Case 1: Populate kernel from a dict
    # FIXME: Default rank from dict keys?
    elif type(data) == types.DictType :
        for delta in data.keys() :
            coef = data[delta]
            if abs(coef) >= thresh :
                kern[delta] = coef

    # Case 2: Populate kernel from a list or tuple
    # FIXME: Default rank from value of size?
    elif type(data) in (types.ListType, types.TupleType) :

        # Verify or default the size argument
        if not size :
            if rank == 1 :
                size = [len(data)]
            else :
                raise Exception, "Size is required when rank > 1"
        else :
            if rank == 1 and type(size) == types.IntType :
                size = [size]
            if rank != len(size) :
                raise Exception, "Size is inconsistent with rank"

        # Compare number of elements from size and list
        count = 1
        for i in range(rank) :
            count = count * size[i]
        if count != len(data) :
            raise Exception, "Size is inconsistent with number of coefs"

        # Construct default center value, if needed
        if not center :
            center = rank * [0]
            for i in range(rank) :
                center[i] = (size[i]-1)/2

        # Loop over the elements in the list
        for n in range(count) :

            # Convert flat index to multi-index, accounting for center
            index = rank * [0]    # Index into the array
            ikern = rank * [0]    # Index into the kernel
            n2 = n
            for i in range(rank) :
                i2 = rank - 1 -i
                n2, index[i2] = divmod(n2, size[i2])
                ikern[i2] = index[i2] - center[i2]

            # Copy the corresponding matrix element into the kernel
            if abs(data[n]) >= thresh :
                kern[ikern] = data[n]

    # Case 3: Populate kernel from an array
    # FIXME: Default rank and size from array parameters?
    elif type(data) == ArrayType :

        # Count elements in the array
        count = 1
        for i in range(data.rank) :
            count = count * data.size[i]

        # Construct default center value, if needed
        if not center :
            center = data.rank * [0]
            for i in range(data.rank) :
                center[i] = (data.size[i]-1)/2

        # Loop over the elements in the array
        for n in range(count) :

            # Convert flat index to multi-index, accounting for center
            index = data.rank * [0]    # Index into the array
            ikern = data.rank * [0]    # Index into the kernel
            n2 = n
            for i in range(data.rank) :
                i2 = data.rank - 1 - i
                n2, index[i2] = divmod(n2, data.size[i2])
                ikern[i2] = index[i2] - center[i2]

            # Copy the corresponding matrix element into the kernel
            if abs(data[index]) >= thresh :
                kern[ikern] = data[index]

    else :
        raise Exception, "Data option must be dict or array"

    return kern

    


#--------------------------------------------------------------------------
# pyvox.monomials - Compute monomials of degree <= n in k literals
#
# This function lists the monomials of total degree less than or equal to
# n in k literals, as an array of k columns by the number of distinct
# monomials; each row of the table gives the power to be applied to the
# corresponding literal.  For example, evaluating monomials(2,2) might
# give the following array and corresponding monomials (with literals y,x):
#
#       0   0    =>  1
#       0   1    =>  x
#       0   2    =>  x^2
#       1   0    =>  y
#       1   1    =>  xy
#       2   0    =>  y^2
#
# The order of the rows in the table is not guaranteed to be consistent
# from one version of pyvox to another.
#
# FIXME: Define standard order by degree and powers?
#
# One application for this function is to list the moments of order n or
# less in k dimensions, and to provide a defined order for listing the
# moments in a generic moment calculating function.
#--------------------------------------------------------------------------

def monorec(n, k, prefix) :

    if k == 0 :
        return prefix
    elif n == 0 :
        return prefix + k*[0,]
    else :
        ans = []
        for p in range(n+1) :
            ans = ans + monorec(n-p, k-1, prefix+[p,])
        return ans


def monomials(n, k) :

    # Return None if n or k is too small
    if n < 0 or k < 0:
        return None

    # Call recursive function to compute the monomials
    ans = monorec(n, k, [])

    # Reformat as a pyvox array
    nrows = len(ans) / k
    ans = array([nrows, k], exim.int, ans)
    return ans



#-------------------------------------------------------------------------
# truefile - Get true name of a possibly compressed file
#
# This function checks for a readable file with the given filename, or
# a compressed copy; it returns the filename of the original or compressed
# copy found, or None if none is found.
#-------------------------------------------------------------------------

def truefile(filename) :
    if os.access(filename, os.R_OK) :
        return filename
    elif re.search(r'\.gz$', filename) or re.search(r'\.Z$', filename) :
        return None
    elif os.access(filename+".gz", os.R_OK) :
        return filename + ".gz"
    elif os.access(filename+".Z",  os.R_OK) :
        return filename + ".Z"
    else :
        return None
    


#-------------------------------------------------------------------------
# rawread - Read a possibly compressed raw image file
#
# This function reads an image file of some given name, type, and size 
# into a Pyvox array, and returns the array.  If the given filename
# ends in ".gz" or ".Z", it is assumed to be a compressed file and is
# uncompressed into a temporary file and then read. If there is no image
# file with the given name, but there is a file with either ".gz" or ".Z"
# appended, then that file is assumed to be the image file compressed,
# which is uncompressed on the fly and read instead.  The first element
# of the dimen array may be zero, in which case the number of slices
# is determined by the size of the (uncompressed) image file.
#
# FIXME: The seek option is not yet supported by array.write(); and its
# value is not stored in the metadata, because I'm not yet convinced that
# it should be.
#
# FIXME:  The configure script checks if this platform supports gzip;
# is it possible to use that knowledge here?  Support the platform supports
# only compress/uncompress but not gzip/gunzip?
#
# FIXME: Is there any way to uncompress/gunzip/etc entirely in core, without
# needing to read/write the disk?  It's not difficult to uncompress on the
# fly using os.popen(); the problem is that we don't know how big to make
# the destination array without knowing how big the uncompressed file is,
# which the uncompression routines don't tell us.  We don't want to read
# the whole thing in, because some multi-dimensional image files are huge.
#
# FIXME: Add support for bzip2?
#--------------------------------------------------------------------------

def rawread(filename, dimen, extype=exim.uint1, bigend=1, seek=0) :

    # Check first for an uncompressed file; then for a compressed file
    if os.access(filename, os.R_OK) :
        if re.search(r'\.gz$', filename) :
            gzfile = filename
            gzprog = 'gzip'
        elif re.search(r'\.Z$', filename) :
            gzfile = filename
            gzprog = 'compress'
        else :
            gzfile = None
            gzprog = None
    elif os.access(filename+".gz", os.R_OK) :
        gzfile = filename + ".gz"
        gzprog = 'gzip'
    elif os.access(filename+".Z",  os.R_OK) :
        gzfile = filename + ".Z"
        gzprog = 'compress'
    else :
        raise Exception, "Unable to read image file %s" % (filename,)

    # Uncompress the file into a temporary, if necessary
    if gzfile :
        tmpfile = tempfile.mktemp('.img')
        imgfile = tmpfile
        cmd = "gunzip -c %s > %s" % (gzfile, tmpfile)
        os.system(cmd)
    else :
        tmpfile = None
        imgfile = filename

    # Determine number of bytes per pixel
    nbytes = extype.nbytes
    if not nbytes :
        raise Exception, "Unable to interpret extype %s" % (extype,)

    # If necessary, determine number of slices from size of the file
    if isinstance(dimen, IntType) :
        dimen = [dimen]
    if dimen[0] == 0 :
        if seek != 0 :
            errmsg = "Automatic dimensioning may not be combined with seek"
            raise Exception, errmsg
        npix = 1          # Pixels per slice
        for i in range(len(dimen)-1) :
            npix = npix * dimen[i+1]
        imgstat = os.stat(imgfile)
        nslc = imgstat[stat.ST_SIZE] / (nbytes * npix)
        if nslc < 1 :
            errmsg = "Image file '%s' contains less than one slice" % filename
            raise Exception, errmsg
        dim2 = [nslc,] + list(dimen[1:])
    else :
        dim2 = list(dimen)

    # Actually read the file 
    image = pyvoxC.rawread(imgfile, dim2, extype, bigend, seek)

    # Create and fill in file metadata
    image.metadata["filename"] = filename
    image.metadata["format"]   = 'raw'
    image.metadata["bigend"]   = bigend
    image.metadata["compress"] = gzprog

    # Record header information (pretending that there was a header)
    image.header["extype"] = extype
    image.header["dimen"]  = dim2

    # Remove the temporary file, if used
    if tmpfile :
        os.remove(tmpfile)

    # Return the array object
    return image



#--------------------------------------------------------------------------
# AnalyzeVW Headers
#
# This code block contains information about the format of the AnalyzeVW
# header.  Specifically, there is a dict avwimg_types which describes how
# Pyvox array types are to be mapped to AnalyzeVW datatypes, and there is
# a table describing each field in the AnalyzeVW header.  Note that taking
# the default on some fields (e.g. datatype, dim) can produce garbage headers.
#
# This definition currently supports the origin field of the SPM variant
# rather than the originator field of strict Analyze 7.5.  This field is
# not documented anywhere that I have been able to find, but Gene Gualtieri
# and I have done some testing and believe that it contains the 1-based index
# coordinates of the voxel that corresponds to the physical origin (0,...).
# The code to read and write the headers translates this to and from the
# origin as defined by Pyvox.
#
# FIXME: One possible approach to supporting multiple variants is to add
# to exim.unpack a command to backup some number of bytes and reread the
# data.
#
# FIXME: Check the field definitions below against the original definition,
# and against some typical (and bbl) actual header files.  Could also
# check against the Insight implementation.
#
# FIXME: Write function to read and dump a header.  It will be useful in its
# own right, as well as for the testing just mentioned.
#
# FIXME: Strings mostly default to zero rather than blank.  Is this consistent
# with actual practice?  The sample code from Mayo uses blanks for
# cal_units and vox_units, but zero for all other default strings.
#
# FIXME: Graham Wideman's web site claims that the SPM version of this
# header also includes a scale factor and image intensity zero-intercept.
# Track down exactly what these mean and consider supporting them.
#
# FIXME: Add a means to support either strict AnalyzeView 7.5 or SPM under
# user control.
#---------------------------------------------------------------------------

# Map internal type -> avw datatype, bits/pix, exim external type
avwimg_types = { exim.uchar    : ( 2,  8, exim.uint1),
                 exim.short    : ( 4, 16, exim.int2),
                 exim.int      : ( 8, 32, exim.int4),
                 exim.float    : (16, 32, exim.float4),
                 exim.fcomplex : (32, 64, exim.complex8),
                 exim.double   : (64, 64, exim.float8),
                 }

# Describe the fields in the header: name, default, extype, and nelem
avwhdr_fields = (
    # Header key
    ("sizeof_hdr",    348,   exim.int4),
    ("data_type",     "",    exim.string,   10),
    ("db_name",       "",    exim.string,   18),
    ("extents",       16384, exim.int4),
    ("session_error", 0,     exim.int2),
    ("regular",       "r",   exim.string),
    ("hkey_un0",      "",    exim.string),
    
    # Image dimension
    ("dim",           1,     exim.int2,      8),
    ("vox_units",     " ",   exim.string,    4),
    ("cal_units",     " " ,  exim.string,    8),
    ("unused1",       0,     exim.int2),
    ("datatype",      0,     exim.int2),
    ("bitpix",        0,     exim.int2),
    ("dim_un0",       0,     exim.int2),
    ("pixdim",        0.0,   exim.float4,    8),
    ("vox_offset",    0.0,   exim.float4),
    ("funused1",      0.0,   exim.float4),
    ("funused2",      0.0,   exim.float4),
    ("funused3",      0.0,   exim.float4),
    ("cal_max",       0.0,   exim.float4),
    ("cal_min",       0.0,   exim.float4),
    ("compressed",    0,     exim.int4),
    ("verified",      0,     exim.int4),
    ("glmax",         0,     exim.int4),
    ("glmin",         0,     exim.int4),
    
    # Data history
    ("descrip",       "",    exim.string,   80),
    ("aux_file",      "",    exim.string,   24),
    ("orient",        "",    exim.string),
    # Analyze 7.5 uses originator; SPM uses origin
    #("originator",    "",    exim.string,   10),
    ("origin",        0,     exim.int2,      5),
    ("generated",     "",    exim.string,   10),
    ("scannum",       "",    exim.string,   10),
    ("patient_id",    "",    exim.string,   10),
    ("exp_date",      "",    exim.string,   10),
    ("exp_time",      "",    exim.string,   10),
    ("hist_un0",      "",    exim.string,    3),
    ("views",         0,     exim.int4),
    ("vols_added",    0,     exim.int4),
    ("start_field",   0,     exim.int4),
    ("field_skip",    0,     exim.int4),
    ("omax",          0,     exim.int4),
    ("omin",          0,     exim.int4),
    ("smax",          0,     exim.int4),
    ("smin",          0,     exim.int4),
    )



#--------------------------------------------------------------------------
# read_avwhdr - Read an AnalyzeView header file
#
# This function reads an AnalyzeView format header from a file and converts
# it to a dict whose keys are the field names and corresponding values the
# field contents.  Whether the header is big-endian or not is saved as the
# value of dict["bigend"] and can be used to decide whether or not the
# image file itself is big-endian.
#
# FIXME: Save the name of the header file in the dict?
#
# FIXME: Is there a way to automatically identify bblanz headers (which are
# a BBL-specific perversion of the Analyze View headers), and read them
# correctly?
#---------------------------------------------------------------------------

def read_avwhdr(filename) :
    
    hdrin = open(filename, "rb")
    data  = hdrin.read()
    if data[0:4] == '\000\000\001\134' :
        bigend = 1
    elif data[0:4] == '\134\001\000\000' :
        bigend = 0
    else :
        raise Exception, "Incorrect magic number in AnalyzeView header"
    dict = exim.unpack(data, avwhdr_fields, bigend)
    dict["bigend"]   = bigend
    return dict



#---------------------------------------------------------------------------
# read_avwimg - Read an AnalyzeView header file and its associated image file
#
# This function reads an AnalyzeView format header from the named file,
# determines the image characteristics from that header, and reads the
# associated image file.
#
# FIXME: Handle compressed header files; compressed images should already
# be handled
#
# FIXME: Test this against a variety of genuine AnalyzeView headers, both
# Analyze 7.5 and the SPM variant.
#---------------------------------------------------------------------------

def read_avwimg(filename) :

    # Read the header file
    hdr = read_avwhdr(filename)

    # Determine rank, dimensions, pixel size, origin
    dimen  = hdr["dim"]
    rank   = dimen[0]
    dimen  = dimen[1:(rank+1)]
    dimen.reverse()               # Reverse coordinate order
    bigend   = hdr["bigend"] 
    datatype = hdr["datatype"] & 255
    bitpix   = hdr["bitpix"]

    # Determine origin and pixel size; convert SPM origin to Pyvox origin
    pixdim = hdr["pixdim"][1:(rank+1)]
    pixdim.reverse()              # Reverse coordinate order
    origin   = hdr["origin"][0:rank]
    origin.reverse()              # Reverse coordinate order
    origin = (-vector(origin)+1) * vector(pixdim)
    # origin = rank * [0,]        # For Analyze 7.5; we use SPM variant instead

    # Determine data type for a correct AnalyzeView header
    if datatype == 1 :
        raise Exception, "Binary AnalyzeView images are not supported"
    elif datatype == 2 and bitpix == 8 :
        extype = exim.uint1
    elif datatype == 4 and bitpix == 16 :
        extype = exim.int2
    elif datatype == 8 and bitpix == 32 :
        extype = exim.int4
    elif datatype == 16 and bitpix == 32 :
        extype = exim.float4
    elif datatype == 32 :
        extype = exim.complex8
    elif datatype == 64 and bitpix == 64 :
        extype = exim.float8
    elif datatype == 128 and bitpix == 24 :
        extype = exim.uint1
        rank = rank + 1
        dimen.append(3)
        pixdim.append(0)
        origin.append(0)

    # Determine data type for defective AnalyzeView headers produced by
    # certain BBL software
    elif bitpix == 8 :
        extype = exim.uint1
    elif bitpix == 16 :
        extype = exim.int2

    # Failed to determine data type, so fail
    else :
        raise Exception, "Cannot identify AnalyzeView data type"

    # Read the image file
    imgfile = re.sub(r'\.hdr$', '.img', filename)
    image   = rawread(imgfile, dimen, extype, bigend)    

    # Record image metadata;
    # rawread should already have set the compression here
    # FIXME: This replaces some metadata already set by rawread,
    # which is inelegant, though effective.
    image.header  = hdr
    image.origin  = origin
    image.spacing = pixdim
    image.metadata["filename"] = filename
    image.metadata["format"]   = "avw"
    image.metadata["bigend"]   = hdr["bigend"]

    return image    
    


#---------------------------------------------------------------------------
# pyvox.read - Generic array read function for pyvox
#
# This function reads data from an external file, converts it to a Pyvox
# array and returns the array.  Several external file formats are supported,
# including raw, analyze, pnm, tiff?, and jpeg?.  In the usual case, it is
# sufficient to specify the filename, and this function will obtain the
# image dimensions and so on from the header.  In the cases where the necessary
# information is not given in the header, or there is not header (e.g. raw),
# optional arguments are used to specify them.  The information extracted
# from the header or the arguments is saved in the attribute array.header,
# which is a dict containing the header information.  (The attribute
# array.userdata is another dict reserved for the user.)
#
# If the header and array data are in separate files, the filename of the
# header file should be provided; the filename of the data file will be
# automatically extracted from the header or header filename.
#
# Compressed files (*.Z, *.gz, and someday *.bz2) will be automatically
# uncompressed, provided that the uncompression programs are installed;
# the extension need not be specified, only the name of the
# file before compression.  The special attribute array.truename gives the
# true filename under which the data was found.
#
# The keyword option format specifies the format of the file, but in
# most cases the format can be determined from the file extension or from
# magic numbers in the file header.  The value supplied for this argument
# is either a string naming the format or an exim external type or raw data.
# (Or the format-specific function itself?).
#
# The keyword option dimen is a list or tuple which gives the number of
# pixels along each dimension of the image, with the most slowly varying first.
# In most cases, the array dimensions are specified in the header or header
# file.  If the first element is zero, the function reads as much data as is
# available from the file.  A naked number may be used instead of a
# singleton list or tuple; 0 means to read as many pixels as possible from
# the file and return them as a one-dimensional array.
#
# The keyword option bigend should be 1 for big-endian data, or 0 for
# little-endian data.  (Support native endian?)
#
# ALGORITHM PLANNED: After uncompressing the file, this function reads the
# first 512 bytes of the file and passes that plus the given filename and any
# optional arguments to all known format-specific readers.  The first
# one that claims to know how to read the file gets it.
#
# FIXME: The format option, if explicitly specified, should override the
# filename extension.
#
# FIXME: It might be useful to default otherwise-unspecified parameters to
# uncompressed, big-endian, uint1, raw, and 256 x 256 pixel slices.  But
# it will require some care to get it to work right.
#
# FIXME: Provide header information for pgm and ppm images.  Perhaps
# pseudo-headers for raw formats?
#
# FIXME: Add support for the seek, origin, spacing, and intype options.
# Also modify the format-specific functions to use keyword arguments.
#
# FIXME: Arrange to support compression wherever reasonable: e.g. raw image
# data, PGM and PPM format, ... but not JPEG (which is already compressed).
#
# FIXME: Allow format readers to check their headers before becoming
# committed to read a file.
#
# FIXME: Add readers for jpeg, tiff, dicom, meta-images?
#
# FIXME: Provide a method for loading and registering format-specific readers
# after this module is loaded?
#
# FIXME: The (de)compression program is determined by the extension of the
# filename (or, in a future version) by the magic number in the file.  Is
# there any serious utility for a compress option to override these?
#
# FIXME: Support DICOM here or though some special function?  MetaImage format?
#
# NOTE: pyvox.read formerly meant something else (raw read, as I recall), but
# has probably been deprecated for long enough that we can redefine it.
#---------------------------------------------------------------------------

readdepwarn = "pyvox.read() now wants '%s' as a keyword rather " + \
              "than a positional argument"

def read(filename, xdimen='xxx', xformat='xxx', xbigend='xxx', **kwargs) :

    # Extract positional or keyword arguments
    # FIXME: The positional arguments were deprecated on 2003-12-02.
    # FIXME: Warnings require Python 2.1 
    # Wait a decent interval and then support only keyword arguments.
    if xdimen != 'xxx' :
        #warnings.warn(readdepwarn % 'dimen', exceptions.DeprecationWarning)
        dimen = kwargs.get('dimen', xdimen)
    else :
        dimen = kwargs.get('dimen',   None)

    if xformat != 'xxx' :
        #warnings.warn(readdepwarn % 'format', exceptions.DeprecationWarning)
        format = kwargs.get('format', xformat)
    else :
        format = kwargs.get('format', '')
    if type(format) == types.StringType :
        format = string.lower(format)

    if xbigend != 'xxx' :
        #warnings.warn(readdepwarn % 'bigend', exceptions.DeprecationWarning)
        bigend = kwargs.get('bigend', xbigend)
    else :
        bigend = kwargs.get('bigend', None)
    
    # Extract keyword arguments
    seek     = kwargs.get('seek',    None)       # Works only for raw format
    origin   = kwargs.get('origin',  None)
    spacing  = kwargs.get('spacing', None)
    intype   = kwargs.get('intype',  None)
    extype   = kwargs.get('extype',  None)

    # Handle special defaults and errors for 'raw' format
    if not format and extype in exim.extypes :
        format = 'raw'
    elif format in exim.extypes :
        extype = format
        format = 'raw'
    elif format in exim.intypes or extype in exim.intypes :
        raise Exception, \
              "The extype/format argument supports only *external* datatypes"

    # DEBUG: Report arguments (to verify kwargs mechanism)
    if 0 :
        print "filename =", filename
        print "dimen    =", dimen
        print "format   =", format
        print "bigend   =", bigend
        print "seek     =", seek
        print "origin   =", origin
        print "spacing  =", spacing
        print "intype   =", intype
        print "extype   =", extype

    # Read a raw format file
    if format == 'raw' :
        if bigend == None :
            bigend = 1
        if dimen == None :
            raise Exception, "dimen argument required for raw data format"
        if not extype :
            raise Exception, "extype argument required for raw data format"
        if seek == None :
            seek = 0
        return rawread(filename, dimen, extype, bigend, seek)

    # Read an AnalyzeView header and image
    elif re.search(r'^analyze$', format,   re.I) or \
         re.search(r'^avw$',     format,   re.I) or \
         re.search(r'\.hdr$',    filename, re.I) :
        return read_avwimg(filename)

    # Read a PGM image file
    elif re.search(r'^pgm$',  format,   re.I) or \
         re.search(r'\.pgm$', filename, re.I) :
        image = pyvoxC.read_pgm(filename)
        image.metadata["filename"] = filename
        image.metadata["format"]   = "pgm"
        image.metadata["bigend"]   = 1
        return image

    # Read a PPM image file
    elif re.search(r'^ppm$',  format,   re.I) or \
         re.search(r'\.ppm$', filename, re.I) :
        image = pyvoxC.read_ppm(filename)
        image.metadata["filename"] = filename
        image.metadata["format"]   = "ppm"
        image.metadata["bigend"]   = 1
        return image

    # No other formats are defined yet
    else :
        raise Exception, "Format '%s' not recognized as valid" % format



#-------------------------------------------------------------------------
# pyvox.write_avwhdr - Write an AnalyzeView header file
#
# This function writes an AnalyzeView header to a specified file.
#-------------------------------------------------------------------------

def write_avwhdr(header, filename, bigend='NS') :

    # Take byte order from header if not specified by caller
    if bigend == 'NS' :
        if header.has_key("bigend") :
            bigend = header["bigend"]
        else :
            raise Exception, "Byte order (bigend) not specified"

    # Write the header file
    hdr    = exim.pack(header, avwhdr_fields, bigend)
    hdrout = open(filename, 'wb')
    hdrout.write(hdr)
    hdrout.close()    



#-------------------------------------------------------------------------
# pyvox.write_avwimg - Write AnalyzeVW header and image files
#
# Given an image, this function writes an AnalyzeVW header file and
# an image file.  The filename given should be that of the header
# file; the name of the image file will be constructed automatically.
# The values of the fields dim, datatype, bitpix, pixdim, glmin, and
# glmax are all taken from the array itself, not from any values that
# may have been stored in the header dictionary.
#
# FIXME: The correct definition of glmin and glmax for complex images is
# not known; we just set them to 0.
#-------------------------------------------------------------------------

def write_avwimg(image, filename, bigend=1) :

    # Copy some header attributes from the image itself
    datatype, bitpix, extype = avwimg_types.get(image.type, [None,0,0])
    # print image.type, datatype, bitpix, extype
    if datatype == None :
        type   = image.type
        errmsg = "Array type '%s' not supported as AnalyzeVW image" % type
        raise Exception, errmsg
    rank = image.rank
    dim  = image.size
    dim.reverse()
    header = image.header.copy()
    header["dim"]      = [rank] + dim
    header["datatype"] = datatype
    header["bitpix"]   = bitpix
    if image.type.iscomplex :
        header["glmin"] = 0
        header["glmax"] = 0
    else :
        header["glmin"] = image.amin()
        header["glmax"] = image.amax()

    # Save origin and spacing, converting Pyvox origin to SPM origin
    # FIXME: The origin is undefined if the pixdim is zero; I have patched
    # this for now, but a definitive solution is needed.
    pixdim = image.spacing
    pixdim.reverse()
    header["pixdim"]   = [rank] + pixdim
    origin = image.origin
    origin.reverse()
    origin = -vector(origin) / vector(pixdim) + 1
    for i in range(len(origin)) :
        if pixdim[i] == 0.0 :
            origin[i] = 0
    header["origin"] = origin.list()
    # origin = rank * [0,]        # For Analyze 7.5; we use SPM variant instead

    # Write the header file
    write_avwhdr(header, filename, bigend)

    # Write the image file
    imgfile = re.sub(r'\.(hdr|avw)$', '.img', filename)
    image.writeraw(imgfile, extype, bigend)

    return



#-------------------------------------------------------------------------
# pyvox.write_eps - Write a Pyvox array as an Encapsulated PostScript file
#
# This function writes a 2-D unsigned char image as an Encapsulated
# PostScript file.  The height and width options specify the height and
# width of the output image in inches.  If only one of these is specified,
# then the other will be chosen to maintain the aspect ratio of the image;
# if neither is specified, then they will be chosen to maintain the aspect
# ratio and set the resolution to 72 samples/inch.
#
# This function should normally be called through the array.write() method
# rather than being called directly by the user.
#
# FIXME: Center the image on a standard page, just for convenience in printing
# and using ghostscript.
#
# FIXME: Do the inner (intensity to hex) loop in C for speed.
#
# FIXME: Upgrade to level 2 PostScript and support color images.
#
# FIXME: Allow any image with exactly two non-unit dimensions.
#-------------------------------------------------------------------------

def write_eps(image, filename, **kwargs) :

    # Verify image rank and type
    if image.type != exim.uchar or image.rank != 2 :
        errmsg = "Only a 2-D unsigned char image can be written as PostScript"
        raise Exception, errmsg

    # Open image file
    outfile = open(filename, 'w')
    if not outfile :
        errmsg = "Unable to open file '%s' for write" % filename
        raise Exception, errmsg

    # Compute dimensions
    height = 72.0 * kwargs.get("height", 0)     # Height in points
    width  = 72.0 * kwargs.get("width",  0)     # Width in points
    nrows, ncols = image.size
    if height and not width :
        width = height * ncols / nrows
    elif width and not height :
        height = width * nrows / ncols
    elif not width and not height :
        width  = ncols
        height = nrows

    # Write prolog
    outfile.write("%!PS-Adobe_2.0 EPSF-2.0\n")
    outfile.write("%%%%BoundingBox: 0 0 %d %d\n" % (width, height))
    outfile.write("%%Pages: 1\n")
    outfile.write("%%EndComments\n")
    outfile.write("%%EndProlog\n\n")
    outfile.write("%%Page: 1 1\n\n")
    outfile.write("/state save def           % Save current state\n\n")
    outfile.write("16 dict begin             % Build local dictionary\n\n")
    line = "/scanline %d string def   %% One scanline of pixels\n\n" % ncols
    outfile.write(line)
    line = "%f %f scale    %% Scale image to desired size\n\n"
    outfile.write(line % (width, height))
    outfile.write("%d %d 8   %% Columns, rows, bits\n" % (ncols, nrows))
    line = "[ %d 0 0 %d 0 %d ]   %% Transform matrix\n"
    outfile.write(line % (ncols, -nrows, nrows))
    outfile.write("{currentfile scanline readhexstring pop}\nimage\n\n")    

    # Write image data
    # FIXME: Add a newline after each 40 pixels
    scanline = (ncols+1) * [0,]
    scanline[-1] = '\n'
    for irow in range(nrows) :
        for icol in range(ncols) :
            scanline[icol] = "%02x" % image[irow, icol]
        line = '' . join(scanline)
        outfile.write(line)            

    # Write postlog
    outfile.write("\nshowpage            % Draw image\n\n")
    outfile.write("end  state restore  % Close block and restore state\n\n")
    outfile.write("%%Trailer\n")

    # Close image file
    outfile.close()
    


#-------------------------------------------------------------------------
# pyvox.write_array - Write a pyvox array in specified external format
#
# This function is called from the array.write() method so that its
# capabilities can be implemented in Python rather than C.  It is not
# intended for general use.  The format used for the external file is
# determined by looking first at the format argument, second at the
# filename, and finally defaults to the most natural raw external format
# corresponding to internal format of the array.  For backward compatibility,
# omitting the format, or specifying 0 or None, takes the default.
#
# FIXME: Some of image writers called from here take positional arguments,
# and some keyword arguments; some of the arguments are pulled out into
# explicit variables and some are left in the kw dict.  There are many
# opportunities for confusion and error here!
#
# FIXME: Implement the compress, exact, seek, and extype arguments.
#
# FIXME: It is permitted for filename to be an open file, but this is not
# currently supported by the writeppm, writepgm, or writeavw methods.
# Once it is supported, should the name associated with the open file
# be examined to guess the format?
#
# FIXME: Arrange to support compression wherever reasonable: e.g. raw image
# data, PGM and PPM format, ... but not JPEG (which is already compressed).
# The right way is probably to provide another argument compress: if 1,
# it picks the best compression available on your machine; if 'gzip',
# 'compress', or 'bzip2', uses the specified compression program.
#--------------------------------------------------------------------------

def write_array(parray, filename, format=None, bigend=-1, **kwargs) :

    # Extract and merge keyword arguments
    if format == None and kwargs.has_key('format') :
        format = kwargs['format']
    if bigend == -1 and kwargs.has_key('bigend') :
        bigend = kwargs['bigend']
    elif bigend == -1 :
        bigend = 1
    compress = kwargs.get('compress', None)
    exact    = kwargs.get('exact',    None)
    seek     = kwargs.get('seek',     None)
    extype   = kwargs.get('extype',   None)

    # DEBUG: Report positional and keyword arguments
    if 0 :
        print 'filename =', filename
        print 'format   =', format
        print 'bigend   =', bigend
        print 'compress =', compress
        print 'exact    =', exact
        print 'seek     =', seek
        print 'extype   =', extype

    # If format omitted, try to guess it from the filename
    if not format and type(filename) == types.StringType :
        if re.search(r'\.hdr$', filename) :
            format = 'analyze'
        elif re.search(r'\.pgm', filename) :
            format = 'pgm'
        elif re.search(r'\.ppm$', filename) :
            format = 'ppm'
        elif re.search(r'\.eps$', filename) :
            format = 'eps'

    # Write an Analyze header and image
    if format == 'analyze' or format == 'avw' :
        write_avwimg(parray, filename, bigend)

    # Write an Encapsulated PostScript image file
    elif format == 'eps' :
        write_eps(parray, filename, **kwargs)

    # Write a PGM image file
    elif format == 'pgm' :
        parray.writepgm(filename)

    # Write a PPM image file
    elif format == 'ppm' :
        parray.writeppm(filename)

    # Write a raw format file with specified external type
    elif format in exim.extypes :
        parray.writeraw(filename, format, bigend)

    # Write a raw format file with default external type
    elif format == 'raw' or format == None or format == 0 :
        format = parray.type.extype
        parray.writeraw(filename, format, bigend)

    # Sorry, that's all the formats for now
    else :
        raise Exception, "Image format '%s' not recognized" % format



#---------------------------------------------------------------------------
# pyvox.affine - Affine transform
#
# This class allows the representation and manipulation of affine
# transforms in any number of dimensions.  With a few exceptions,
# operations such as translate, scale, and rotate operate in place
# rather than creating a new transform.
#
# In compositions, the order of composition can be selected with the
# optional pre argument, which takes the following possible values:
#
#   1 => precompose  (first position)
#   2 => postcompose (second position)
#   0 => postcompose (for compatibility only; pre used to be a boolean
#        meaning precompose)
#  -1 => precompose with inverse
#  -2 => postcompose with inverse
#
# NOTE: To avoid a conflict with the matrix() function, we use matrox
# as the name of the local variable containing the matrix part of the
# affine transform.
#
# FIXME: Add the capability to initialize from a poly transform of degree
# one.
#
# FIXME: Add repr() and str() methods.  Add a prettyprint method.
#
# FIXME: Add __copy__() and __deepcopy__() methods.
#
# FIXME: Add __getitem__() and __setitem__() methods so that subscripting
# with one or two subscripts accesses the corresponding parameter of the
# transform.  Check that the implied order for single subscripting is
# both sensible and consistent with the convention for arrays.
#
# FIXME: Add method to return index to params.  Arrange for this class and
# afflst to use the same parameter order.
#---------------------------------------------------------------------------

class affine:
    "Affine transform"

    #------------------------------------------------------------------------
    # Initialize an affine transform to given values:
    #
    # A single scalar argument yields the identity transform for
    # that number of dimensions.
    #------------------------------------------------------------------------
    
    def __init__(self, ndim, matrox=None, offset=None):
        self.ndim = ndim
        if matrox :
            self.matrix = matrix(matrox, ndim, ndim)
        else :
            self.matrix = array([ndim,ndim], exim.double, 0)
            for i in range(ndim):
                self.matrix[i,i] = 1.0
        if offset :
            self.offset = point(offset, ndim)
        else :
            self.offset = array([ndim,1], exim.double, 0)


    #-------------------------------------------------------------------------
    # Add affine transforms
    #-------------------------------------------------------------------------

    def __add__(self, other) :

        # Delegate affine + poly
        if isinstance(other, poly) :
            return other + self

        # Handle affine + affine here
        if not isinstance(other, affine) :
            raise Exception, "Other argument is not an affine transform"
        if self.ndim != other.ndim :
            raise Exception, "Dimensions do not match"

        newmat = self.matrix + other.matrix
        newoff = self.offset + other.offset

        return affine(self.ndim, newmat, newoff)

 
    #--------------------------------------------------------------------------
    # Copy and deep copy
    #
    # Always copies both top level and the arrays that define the transform
    #-------------------------------------------------------------------------
    
    def __copy__(self) :
        
        return affine(self.ndim, self.matrix.copy(), self.offset.copy())


    def __deepcopy__(self, visit) :
        
        return affine(self.ndim, self.matrix.copy(), self.offset.copy())


    #-----------------------------------------------------------------------
    # Get coefficient specified by subscripts
    #
    # FIXME:  Add better error checking and reporting
    #-----------------------------------------------------------------------

    def __getitem__(self, subscr) :

        row, col = subscr
        if col == self.ndim :
            return self.offset[row, 0]
        else :
            return self.matrix[row, col]


    #-----------------------------------------------------------------------
    # Set coefficient specified by subscripts
    #
    # FIXME:  Add better error checking and reporting
    #-----------------------------------------------------------------------

    def __setitem__(self, subscr, value) :

        row, col = subscr
        if col == self.ndim :
            self.offset[row, 0] = value
        else :
            self.matrix[row, col] = value
        return


    #-----------------------------------------------------------------------
    # Generate plain string representation
    #-----------------------------------------------------------------------

    def __str__(self) :

        result = ""
        for i in range(self.ndim) :
            for j in range(self.ndim) :
                result = result + "%12.6f " % self.matrix[i,j]
            result = result + "%12.6f\n" % self.offset[i,0]

        return result


    #-------------------------------------------------------------------------
    # Subtract affine transforms
    #-------------------------------------------------------------------------

    def __sub__(self, other) :

        # Delegate affine - poly
        if isinstance(other, poly) :
            return other.__rsub__(self)

        # Handle affine - affine here
        if not isinstance(other, affine) :
            raise Exception, "Other argument is not an affine transform"
        if self.ndim != other.ndim :
            raise Exception, "Dimensions do not match"

        newmat = self.matrix - other.matrix
        newoff = self.offset - other.offset

        return affine(self.ndim, newmat, newoff)


    #--------------------------------------------------------------------------
    # Print self in readable format
    #
    # FIXME: Add a stream to print to
    #--------------------------------------------------------------------------
    
    def write(self):
        for i in range(self.ndim):
            for j in range(self.ndim):
                print self.matrix[i,j], 
            print self.offset[i,0]
        print

 
    #--------------------------------------------------------------------------
    # Deep copy
    #
    # FIXME: Use copy(affine) from the copy module in new code.
    #-------------------------------------------------------------------------
    
    def copy(self) :

        return self.__copy__()


    #--------------------------------------------------------------------------
    # Return a vector of parameters defining the transform
    #
    # For compatibility with optimizers of various flavors
    #--------------------------------------------------------------------------
    
    def param(self) :
        ndim   = self.ndim
        nparm  = ndim*(ndim+1)
        result = array(nparm, exim.double, 0)
        for irow in range(ndim) :
            for icol in range(ndim) :
                iparm = irow*(ndim+1) + icol
                result[iparm] = self.matrix[irow, icol]
            result[irow*(ndim+1)+ndim] = self.offset[irow, 0]
        return result


    #--------------------------------------------------------------------------
    # Initialize from a vector of parmeters defining the transform
    #
    # For compatibility with optimizers of various flavors
    #--------------------------------------------------------------------------
    
    def setparam(self, parm) :
        ndim  = self.ndim
        nparm = ndim*(ndim+1)
        parm  = vector(parm, nparm)
        for irow in range(ndim) :
            for icol in range(ndim) :
                iparm = irow*(ndim+1) + icol
                self.matrix[irow, icol] = parm[iparm]
            self.offset[irow, 0] = parm[irow*(ndim+1)+ndim]
        return self


    #-------------------------------------------------------------------------
    # Add an increment to the parameters defining the transform
    #
    # For compatibility with optimizers of various flavors
    #-------------------------------------------------------------------------
    
    def addparam(self, parm) :
        ndim  = self.ndim
        nparm = ndim*(ndim+1)
        parm  = vector(parm, nparm)
        for irow in range(ndim) :
            for icol in range(ndim) :
                iparm = irow*(ndim+1) + icol
                self.matrix[irow, icol] = self.matrix[irow, icol] + parm[iparm]
            ioff = irow*(ndim+1)+ndim
            self.offset[irow, 0] = self.offset[irow, 0] + parm[ioff]
        return self


    #--------------------------------------------------------------------------
    # Compose with another affine transform
    #-------------------------------------------------------------------------
    
    def compose(self, other, pre=0):
        if not isinstance(other, affine) :
            raise Error, 'Other is not an affine transform'
        if self.ndim != other.ndim :
            raise Error, 'Self and other have different number of dimensions'
        if pre == 0 or pre == 2 :
            self.offset = other.matrix.mmul(self.offset) + other.offset
            self.matrix = other.matrix.mmul(self.matrix)
        elif pre == 1 :
            self.offset = self.matrix.mmul(other.offset) + self.offset
            self.matrix = self.matrix.mmul(other.matrix)
        elif pre == -1 :
            matrox = other.matrix.inverse()
            offset = matrox.mmul(-other.offset)
            self.offset = self.matrix.mmul(offset) + self.offset
            self.matrix = self.matrix.mmul(matrox)
        elif pre == -2 :
            matrox = other.matrix.inverse()
            self.offset = matrox.mmul(self.offset - other.offset)
            self.matrix = matrox.mmul(self.matrix)
        else :
            raise Exception, "Invalid value of pre"
        return self


    #-------------------------------------------------------------------------
    # Compose with given matrix and offset of an affine transform
    #-------------------------------------------------------------------------
    
    def compose2(self, matrox=1, offset=0, pre=0):
        matrox = matrix(matrox, self.ndim, self.ndim)
        offset = point(offset, self.ndim)
        if pre == 0 or pre == 2 :
            self.offset = matrox.mmul(self.offset) + offset
            self.matrix = matrox.mmul(self.matrix)
        elif pre == 1 :
            self.offset = self.matrix.mmul(offset) + self.offset
            self.matrix = self.matrix.mmul(matrox)
        elif pre == -1 :
            matrox = matrox.inverse()
            offset = matrox.mmul(-offset)
            self.offset = self.matrix.mmul(offset) + self.offset
            self.matrix = self.matrix.mmul(matrox)
        elif pre == -2 :
            matrox = matrox.inverse()
            self.offset = matrox.mmul(self.offset - offset)
            self.matrix = matrox.mmul(self.matrix)
        else :
            raise Exception, "Invalid value of pre"            
        return self


    #-------------------------------------------------------------------------
    # Compose with matrix of an affine transform
    #-------------------------------------------------------------------------
    
    def composematrix(self, matrox, pre=0):
        matrox = matrix(matrox, self.ndim, self.ndim)
        if pre == 0 or pre == 2 :
            self.offset = matrox.mmul(self.offset)
            self.matrix = matrox.mmul(self.matrix)
        elif pre == 1:
            self.matrix = self.matrix.mmul(matrox)
        elif pre == -1 :
            self.matrix = self.matrix.mmul(matrox.inverse())
        elif pre == -2 :
            matrox = matrox.inverse()
            self.offset = matrox.mmul(self.offset)
            self.matrix = matrox.mmul(self.matrix)
        else:
            raise Exception, "Invalid value of pre"  
        return self


    #-------------------------------------------------------------------------
    # Compose with index-to-physical transform given by an origin and scale
    #-------------------------------------------------------------------------
    
    def i2p(self, origin, scale, pre=0) :
        matrox = self.scalematrix(scale)
        offset = point(origin, self.ndim)
        return self.compose2(matrox, offset, pre)


    #-------------------------------------------------------------------------
    # Compose with physical-to-index transform given by an origin and scale
    #-------------------------------------------------------------------------
    
    def p2i(self, origin, scale, pre=0) :
        matrox = self.scalematrix(scale)
        offset = point(origin, self.ndim)
        matrox = matrox.inverse()
        offset = matrox.mmul(-offset)
        return self.compose2(matrox, offset, pre)


    #-------------------------------------------------------------------------
    # Invert the transformation (in place)
    #-------------------------------------------------------------------------
    
    def invert(self):
        self.matrix = self.matrix.inverse()
        self.offset = self.matrix.mmul(-self.offset)
        return self


    #------------------------------------------------------------------------
    # Compute new transform which is the inverse of the old
    #------------------------------------------------------------------------
    
    def inverse(self):
        matrox = self.matrix.inverse()
        offset = matrox.mmul(-self.offset)
        return affine(self.ndim, matrox, offset)


    #-------------------------------------------------------------------------
    # Compose with a translation
    #-------------------------------------------------------------------------
    
    def translate(self, delta, pre=0):
        delta = point(delta, self.ndim)
        if pre == 0 or pre == 2:
            self.offset = self.offset + delta
        elif pre == 1 :
            self.offset = self.offset + self.matrix.mmul(delta)
        elif pre == -1 :
            self.offset = self.offset - self.matrix.mmul(delta)
        elif pre == -2 :
            self.offset = self.offset - delta
        else :
            raise Exception, "Invalid value of pre"
        return self


    #-------------------------------------------------------------------------
    # Generate scaling matrix from scalar or vector
    #-------------------------------------------------------------------------
    
    def scalematrix(self, scale):
        scale = vector(scale, self.ndim)
        matrox = array([self.ndim,self.ndim], exim.double, 0)
        for i in range(self.ndim):
                matrox[i,i] = scale[i]
        return matrox


    #-------------------------------------------------------------------------
    # Compose with a scaling, isotropic or not
    #-------------------------------------------------------------------------
    
    def scale(self, scale, pre=0):
        matrox = self.scalematrix(scale)
        return self.composematrix(matrox, pre)


    #-------------------------------------------------------------------------
    # Generate an elementary rotation matrix
    #-------------------------------------------------------------------------
    
    def rotationmatrix(self, axis1, axis2, angle):
        matrox = array([self.ndim,self.ndim], exim.double, 0)
        for i in range(self.ndim):
            matrox[i,i] = 1.0
        matrox[axis1, axis1] =  cos(angle)
        matrox[axis1, axis2] =  sin(angle)
        matrox[axis2, axis1] = -sin(angle)
        matrox[axis2, axis2] =  cos(angle)
        return matrox


    #--------------------------------------------------------------------------
    # Compose with an elementary rotation matrix
    #--------------------------------------------------------------------------
    
    def rotate(self, axis1, axis2, angle, pre=0):
        d = self.rotationmatrix(axis1, axis2, angle)
        return self.composematrix(d, pre)

    #--------------------------------------------------------------------------
    # Compose with a rotation around an axis in 3D
    #--------------------------------------------------------------------------
    
    def rotate3d(self, axis, angle, pre=0) :
        if self.ndim != 3 :
            raise Error, 'rotate3d only works in 3 dimensions'
        axis = point(axis, 3)
        axis = axis / axis.norm2()
        q0 = cos(angle/2.0)
        q1 = axis[0,0] * sin(angle/2.0)
        q2 = axis[1,0] * sin(angle/2.0)
        q3 = axis[2,0] * sin(angle/2.0)
        matrox = array([3,3], exim.double, 0)
        matrox[0,0] = q0**2 + q1**2 - q2**2 - q3**2
        matrox[0,1] = 2.0*(q1*q2 - q0*q3)
        matrox[0,2] = 2.0*(q1*q3 + q0*q2)
        matrox[1,0] = 2.0*(q1*q2 + q0*q3)
        matrox[1,1] = q0**2 + q2**2 - q1**2 - q3**2
        matrox[1,2] = 2.0*(q2*q3 - q0*q1)
        matrox[2,0] = 2.0*(q1*q3 - q0*q2)
        matrox[2,1] = 2.0*(q2*q3 + q0*q1)
        matrox[2,2] = q0**2 + q3**2 - q1**2 - q2**2
        return self.composematrix(matrox, pre)


    #-------------------------------------------------------------------------
    # Generate an elementary shear matrix
    #-------------------------------------------------------------------------
    
    def shearmatrix(self, axis1, axis2, coef):
        matrox = array([self.ndim,self.ndim], exim.double, 0)
        for i in range(self.ndim):
            matrox[i,i] = 1.0
        matrox[axis1, axis2] =  coef
        return matrox


    #--------------------------------------------------------------------------
    # Compose with an elementary shear matrix
    #--------------------------------------------------------------------------
    
    def shear(self, axis1, axis2, coef, pre=0):
        d = self.shearmatrix(axis1, axis2, coef)
        return self.composematrix(d, pre)


    #-------------------------------------------------------------------------
    # Resample with linear interpolation (Fast implementation)
    #-------------------------------------------------------------------------
    
    def linear(self, source, dimen):
        matrox = self.matrix
        offset = self.offset
        result = pyvoxC.affine_linear(matrox, offset, source, dimen)
        result.origin  = self.point (source.origin)
        result.spacing = self.vector(source.spacing)
        return result


    #-------------------------------------------------------------------------
    # Resample with linear interpolation (Reference implementation)
    #-------------------------------------------------------------------------
    
    def linear0(self, source, dimen):
        matrox = self.matrix
        offset = self.offset
        result = pyvoxC.affine_linear0(matrox, offset, source, dimen)
        result.origin  = self.point (source.origin)
        result.spacing = self.vector(source.spacing)
        return result
    

    #--------------------------------------------------------------------------
    # Generate coordinate map
    #--------------------------------------------------------------------------

    def map(self, dimen) :

        matrox = self.matrix
        offset = self.offset
        return pyvoxC.affine_map(matrox, offset, dimen)


    #--------------------------------------------------------------------------
    # Resample with nearest neighbor interpolation (Fast implementation)
    #--------------------------------------------------------------------------
    
    def nearest(self, source, dimen):
        matrox = self.matrix
        offset = self.offset
        result = pyvoxC.affine_nearest(matrox, offset, source, dimen)
        result.origin  = self.point (source.origin)
        result.spacing = self.vector(source.spacing)
        return result


    #------------------------------------------------------------------------
    # Resample with nearest neighbor interpolation (Reference implementation)
    #------------------------------------------------------------------------
    
    def nearest0(self, source, dimen):
        matrox = self.matrix
        offset = self.offset
        result = pyvoxC.affine_nearest0(matrox, offset, source, dimen)
        result.origin  = self.point (source.origin)
        result.spacing = self.vector(source.spacing)
        return result


    #-------------------------------------------------------------------------
    # Transform a point
    #
    # FIXME: Use plain call to transform a point?
    #--------------------------------------------------------------------------
    
    def point(self, x):
        x = point(x, self.ndim)
        return self.matrix.mmul(x) + self.offset


    #------------------------------------------------------------------------
    # Transform a vector
    #------------------------------------------------------------------------
    
    def vector(self, x):
        x = point(x, self.ndim)
        return self.matrix.mmul(x)


    #--------------------------------------------------------------------------
    # Compute a norm on affine transforms
    #--------------------------------------------------------------------------
    
    def norm(self, other=None, length=100.0) :
        ndim = self.ndim
        if other == None or other == 0 :
            matrox = self.matrix
            offset = self.offset
        elif other == 1 :
            matrox = self.matrix
            offset = self.offset
            for n in range(ndim) :
                matrox[n,n] = matrox[n,n] - 1.0
        else :
            matrox = self.matrix - other.matrix
            offset = self.offset - other.offset
        result = 0.0
        for r in range(ndim) :
            res = 0.0
            for c in range(ndim) :
                res = res + abs(matrox[r,c])
            res = length*res + abs(offset[r,0])
            if res > result :
                result = res
        return result


#-----------------------------------------------------------------------------
# Provide a class object similar to the other Pyvox type objects
#------------------------------------------------------------------------------

AffineType = affine



#---------------------------------------------------------------------------
# pyvox.poly - Polynomial transform
#
# This class allows the representation and manipulation of polynomial
# transforms in any number of dimensions.  With a few exceptions,
# operations such as translate, scale, and rotate operate in place
# rather than creating a new transform.
#
# In compositions, the order of composition can be selected with the
# optional pre argument, which takes the following possible values
# given below.  Note that polynomial transforms are not (currently)
# invertible and compositions that require them will generate errors.
#
#   1 => precompose  (first position)
#   2 => postcompose (second position)
#  -1 => precompose with inverse
#  -2 => postcompose with inverse
#
# ALGORITHM:  The self.keys list is needed by many methods and is updated
# by any method that could change the terms that are present; changes
# only to the coeficients, however, do not require an update.
#
# FIXME: There is a tension here that needs to be considered.  On one hand,
# operations on poly transforms should not include any more terms than
# are needed, and arguably should drop terms with coefficient zero.  On
# the other hand, the list of terms determines what parameters are available
# for optimization, and this should not be gratuitously modified.
#
# FIXME: Add a method keys() to replace the self.keys attribute, for
# consistency with other mapping objects, and to ensure that the keys
# are sorted.  Ditto terms() and perhaps items().
#
# FIXME: Implement a method poly.map() to generate a coordinate map,
# similar to affine.map().
#
# FIXME: Add __copy__ and __deepcopy__ methods?  Add __getitem__() and
# __setitem__()?
#
# FIXME: Add a prettyprint method?
#
# FIXME: Add methods for differential form and Jacobian.
#
# FIXME: Allow input and output dimension to differ?
#
# FIXME: Add code to check that the various arguments are consistent and
# to generate comprehensible errors if not.
#
# FIXME: Add code to compose poly transforms with each other.
#
# FIXME: Add code to truncate a poly transform to a specified maximum degree.
#
# FIXME: Add method to return index to params.  Arrange for this class and
# afflst to use the same parameter order.
#---------------------------------------------------------------------------

class poly :

    "Polynomial transform in n dimensions"

    #-----------------------------------------------------------------------
    # Initialize a polynomial transform
    #
    # Instance variables:
    #
    #    ndim     Number of dimensions
    #    terms    Dict of terms:  (degree, outcor, power) -> coef
    #    keys     Dict keys in sorted order
    #    scalen   Scale length for preconditioning params
    #
    # FIXME: Setting degree=0 currently creates an identity transform plus
    # translation, with all coefficients optimizable.  It is arguable that
    # the linear coefficients should be excluded from optimization.
    #
    # FIXME: What should the minimal generator poly(n) produce?  The obvious
    # possibilities are the zero and identity transforms, either with all
    # terms or only the minimum necessary terms.
    #-----------------------------------------------------------------------

    def __init__(self, ndim, **options) :

        # Extract options and values from the call
        self.ndim = ndim
        degree = options.get('degree', None)
        init   = options.get('init',   None)

        # Initialize instance variables
        self.terms  = {}
        self.scalen = 100.0

        # Construct terms for the requested degree, if provided
        # FIXME: Modify array.list() to generate nested lists, then
        # simplify the code below.
        if degree is not None :
            monos = monomials(degree, ndim)
            for irow in range(monos.size[0]) :
                mono = monos[irow,:].tuple()
                degr = 0
                for pow in mono :
                    degr = degr + pow
                for iout in range(ndim) :
                    key = (degr, iout, mono)
                    self.terms[key] = 0.0

        # Initialize from given data
        if init is not None :
            self.init(init)

        return


    #------------------------------------------------------------------
    # Coefficientwise sum of two polynomial transforms, or poly + affine
    #
    # NOTE: The affine class detects the case affine + poly and delegates
    # it to this method.
    #------------------------------------------------------------------

    def __add__(self, other) :

        if isinstance(other, affine) :
            if self.ndim != other.ndim :
                raise Exception, "Addends differ in number of dimensions"
            return self + poly(self.ndim, init=other)
        elif isinstance(other, poly) :
            if self.ndim != other.ndim :
                raise Exception, "Addends differ in number of dimensions"
            sum = self.terms.copy()
            oth = other.terms
            for key, coef in oth.items() :
                sum[key] = sum.get(key, 0.0) + oth[key]
            return poly(self.ndim, init=sum)
        else :
            raise exception, "Addend must be a poly or affine instance"

        return


    #-----------------------------------------------------------------
    # Scalar multiplication of a polynomial transform
    #
    # FIXME: Support elementwise product of output points?
    #----------------------------------------------------------------

    def __mul__(self, other) :

        if type(other) in (types.IntType, types.FloatType) :
            result = {}
            terms  = self.terms
            for key in terms.keys() :
                result[key] = other * terms[key]
        else :
            raise Exception, "Multiplier must be a number"

        return poly(self.ndim, init=result)       


    #-----------------------------------------------------------------
    # repr() - Computer-readable string representation
    #-----------------------------------------------------------------

    def __repr__(self) :

        self.sortkeys()
        result = "pyvox.poly(%d, init=[" % self.ndim
        for key in self.keys :
            degr, outcor, power = key
            coef = self.terms[key]
            result = result + "[%d, [" % outcor 
            for j in range(self.ndim) :
                result = result +  "%d, " % power[j]
            result = result + "],  %f], " % coef
        result = result + "])"

        return result


    #-----------------------------------------------------------------
    # Number * poly - Scalar multiplication
    #----------------------------------------------------------------

    def __rmul__(self, other) :

        return self.__mul__(other)


    #------------------------------------------------------------------
    # Coefficientwise difference of affine and polynomial transforms
    #
    # NOTE: The affine class detects the case affine - poly and delegates
    # to this method.
    #------------------------------------------------------------------

    def __rsub__(self, other) :

        if isinstance(other, affine) :
            if self.ndim != other.ndim :
                raise Exception, "Addends differ in number of dimensions"
            return poly(self.ndim, init=other) - self
        elif isinstance(other, poly) :
            if self.ndim != other.ndim :
                raise Exception, "Addends differ in number of dimensions"
            sum = self.terms.copy()
            oth = other.terms
            for key, coef in oth.items() :
                sum[key] = oth[key] - sum.get(key, 0.0)
            return poly(self.ndim, init=sum)
        else :
            raise exception, "Addend must be a poly or affine instance"

        return
    

    #-----------------------------------------------------------------
    # str() - Human-readable string representation
    #-----------------------------------------------------------------

    def __str__(self) :

        self.sortkeys()
        result = "Polynomial transform in %d dimensions:\n" % self.ndim
        for key in self.keys :
            degr, outcor, power = key
            coef = self.terms[key]
            result = result + "   %3d,  [" % outcor 
            for j in range(self.ndim) :
                result = result +  "%2d, " % power[j]
            result = result + "],  %12.8f\n" % coef
        result = result + "\n"

        return result
    

    #------------------------------------------------------------------
    # Coefficientwise difference of two polynomial transforms
    #------------------------------------------------------------------

    def __sub__(self, other) :

        if isinstance(other, affine) :
            if self.ndim != other.ndim :
                raise Exception, "Addends differ in number of dimensions"
            return self - poly(self.ndim, init=other)
        elif isinstance(other, poly) :
            if self.ndim != other.ndim :
                raise Exception, "Addends differ in number of dimensions"
            sum = self.terms.copy()
            oth = other.terms
            for key, coef in oth.items() :
                sum[key] = sum.get(key, 0.0) - oth[key]
            return poly(self.ndim, init=sum)
        else :
            raise exception, "Addend must be a poly or affine instance"

        return
    

    #-----------------------------------------------------------------
    # Deep copy of a polynomial transform
    #-----------------------------------------------------------------

    def copy(self) :

        return poly(self.ndim, init=self.terms)


    #----------------------------------------------------------------------
    # Initialize from another transform, list/tuple, etc
    #---------------------------------------------------------------------

    def init(self, data) :

        ndim = self.ndim

        # Clear all existing coeficients to zero
        for key in self.terms.keys() :
            self.terms[key] = 0.0

        # If data == 0, then you're already done
        if data == 0 :
            pass
        
        # If data == 1, create an identity transform
        elif data == 1 :
            for outcor in range(ndim) :
                power = ndim*[0,]
                power[outcor] = 1
                key = (1, outcor, tuple(power))
                self.terms[key] = 1.0

        # Initialize from a dict
        elif isinstance(data, types.DictType) :
            for key, coef in data.items() :
                d, i, k = key
                if len(k) != self.ndim :
                    errmsg = "Dict key '%s' is wrong number of dimensions" \
                             % key
                    raise Exception, errmsg
                self.terms[key] = coef

        # Initialize from another polynomial transform
        elif isinstance(data, PolyType) :
            if data.ndim != self.ndim :
                raise Exception, "Data argument has wrong number of dimensions"
            for key, coef in data.terms.items() :
                self.terms[key] = coef

        # Initialize from a nested list or tuple
        # FIXME: Check for repeated entries in the argument
        elif isinstance(data, types.ListType) or \
             isinstance(data, types.TupleType) :
            for outcor, power, coef in data :
                if len(power) != self.ndim :
                    errmsg = "Datum '%s' has wrong number of dimensions" % \
                    ((outcor, power, coef),)
                    raise Exception, errmsg
                degr = 0
                for pow in power :
                    degr = degr + pow
                key = (degr, outcor, tuple(power))
                self.terms[key] = coef

        # Initialize from an affine transform
        elif isinstance(data, AffineType) :
            if data.ndim != self.ndim :
                raise Exception, "Data argument has wrong number of dimensions"
            ndim  = self.ndim
            mat   = data.matrix
            off   = data.offset
            for outcor in range(ndim) :
                key = (0, outcor, ndim*(0,))
                self.terms[key] = off[outcor,0]
                power = (1,) + (ndim-1)*(0,)
                for incor in range(ndim) :
                    key = (1, outcor, power)
                    self.terms[key] = mat[outcor, incor]
                    power = (0,) + power[:-1]

        # Complain about anything else
        else :
            raise Exception, "Don't know how to initialize from given data"

        # Sort terms into canonical order
        self.sortkeys()
        
        return


    #------------------------------------------------------------------
    # Compute norm of polynomial transform
    #------------------------------------------------------------------

    def norm(self, other=None, length=100.0) :

        # Create list for the sum on each output coordinate
        sums = self.ndim * [0]

        # If other is omitted, compute the norm on self alone
        if not other :
            terms = self.terms

        # If other is 1, subtract an identity transform from self
        elif other == 1 :
            ident = self.dictident()
            terms = self.dictdiff(self.terms, ident)

        # Otherwise compute the norm of self - other
        else :
            terms = self.dictdiff(self.terms, other.terms)

        # Compute sums over the norms of the terms
        for key, coef in terms.items() :
            degr, outcor, power = key
            sums[outcor] = sums[outcor] + abs(coef)*length**degr

        # Take the largest sum as the result
        res = sums[0]
        for i in range(1,self.ndim) :
            res = max(res, sums[i])

        return res


    #-----------------------------------------------------------------
    # Scale a polynomial transform
    #-----------------------------------------------------------------

    def scale(self, coef, pre=2) :

        # Construct affine transform for the scaling
        aff = affine(self.ndim)
        aff.scale(coef)

        # Compose with self
        self.compose(aff, pre)

        return self


    #-----------------------------------------------------------------
    # Translate a polynomial transform
    #-----------------------------------------------------------------

    def translate(self, delta, pre=2) :

        # Construct affine transform for the translation
        aff = affine(self.ndim)
        aff.translate(delta)

        # Compose with self
        self.compose(aff, pre)

        return self


    #--------------------------------------------------------------------
    # Truncate to specified maximum degree
    #
    # FIXME: Since the basis functions are not orthogonal, simple
    # truncation will NOT provide the nearest transform of the
    # given degree.  I think using Chebychev truncation will, so
    # look into it.
    #-------------------------------------------------------------------

    def truncate(self, maxdegree) :

        keys = self.terms.keys()
        for key in keys :
            degr, outcor, power = key
            if degr > maxdegree :
                del self.terms[key]

        self.sortkeys()
        return


    #------------------------------------------------------------------
    # Add second polynomial to another; both expressed as dicts
    #------------------------------------------------------------------

    def dictadd(self, poly1, poly2) :

        # Add each term of poly2
        for key, coef in poly2.items() :
            poly1[key] = poly1.get(key, 0) + coef

        return poly1


    #-----------------------------------------------------------------
    # Compose two poly transforms given as dictionaries
    #
    # The result() is poly2(poly1()) or poly1() o poly2().
    #-----------------------------------------------------------------

    def dictcomp(self, ndim, result, poly1, poly2) :

        # Separate poly1 by output coordinates
        alist = []
        for i in range(ndim) :
            alist.append({})
        for key, coef in poly1.items() :
            degr, outcor, power = key
            alist[outcor][key]  = coef

        # Loop over terms in second polynomial transform
        for key, coef in poly2.items() :

            # Recompute terms in poly2 by substituting the terms in poly1
            if coef != 0 :
                degr, outcor, power = key
                pterm = {(0, outcor, ndim*(0,)) : coef}
                for incor in range(len(power)) :
                    for j in range(power[incor]) :
                        pterm = self.dictmul(pterm, alist[incor], outcor)
                self.dictadd(result, pterm)

        return result        
        

    #------------------------------------------------------------------
    # Compute difference of two transforms expressed as dicts
    #------------------------------------------------------------------

    def dictdiff(self, poly1, poly2) :

        # Initialize result to poly1
        result = poly1.copy()

        # Subtract each term of poly2
        for key, coef in poly2.items() :
            result[key] = result.get(key, 0) - coef

        return result


    #----------------------------------------------------------------
    # Create identity transform as a dict
    #----------------------------------------------------------------

    def dictident(self) :

        result = {}
        for i in range(self.ndim) :
            power = self.ndim * [0]
            power[i] = 1
            key = (1, i, tuple(power))
            result[key] = 1.0

        return result


    #-----------------------------------------------------------------
    # Compute product of two transforms expressed as dicts
    #
    # Specialized for use in precompose; not for general use
    #
    # FIXME: Should check that number of input coordinates is the same
    #-----------------------------------------------------------------

    def dictmul(self, poly1, poly2, outcor) :

        # Initialize result to empty dict
        result = {}

        # Loop over all terms in first operand
        for key1, coef1 in poly1.items() :
            degr1, out1, pow1 = key1

            # Loop over all terms in second operand
            for key2, coef2 in poly2.items() :
                degr2, out2, pow2 = key2

                # Compute product and add to result
                # FIXME: Fails without useful message if operands
                # have different number of input coordinates
                if coef1 != 0 and coef2 != 0 :
                    pow3  = len(pow1) * [0]
                    degr3 = 0
                    for i in range(len(pow1)) :
                        pow3[i] = pow1[i] + pow2[i]
                        degr3   = degr3 + pow3[i]
                    key3 = (degr3, outcor, tuple(pow3))
                    result[key3] = coef1 * coef2 + result.get(key3, 0.0)

        return result                


    #--------------------------------------------------------------------
    # Sort terms into canonical order (by degree, outcor, and power)
    #--------------------------------------------------------------------

    def sortkeys(self) :

        keys = self.terms.keys()
        keys.sort()
        self.keys = keys

        return


    #--------------------------------------------------------------------
    # Generate arrays for interface to C code  (Internal)
    #--------------------------------------------------------------------
                
    def make_arrays(self) :

        # Allocate space for outcor, power, and coef arrays
        nterms = len(self.terms)
        outcar = array( nterms, exim.int, 0)
        powar  = array([nterms, self.ndim], exim.int, 0)
        coefs  = array( nterms, exim.double, 0)

        # Copy indices and coefs into their arrays
        for iterm in range(len(self.keys)) :
            key = self.keys[iterm]
            degr, outcor, power = key
            outcar[iterm] = outcor
            coefs [iterm] = self.terms[key]
            for j in range(self.ndim) :
                powar[iterm, j] = power[j]
        
        return outcar, powar, coefs


    #----------------------------------------------------------------------
    # Print self in readable format
    #
    # FIXME: Should this be str() instead?
    #----------------------------------------------------------------------

    def write(self) :

        print "Polynomial transform in %d dimensions" % self.ndim
        for key in self.keys :
            degr, outcor, power = key
            coef = self.terms[key]
            print "   %3d,  [" % outcor ,
            for j in range(self.ndim) :
                print "%2d, " % power[j] ,
            print "],  %12.8f" % coef
        print

        return


    #----------------------------------------------------------------------
    # Get coefficients
    #
    # This method returns the polynomial coefficients as a Pyvox vector
    # that exactly corresponds to the current keys and terms.  It is used
    # by some optimizers that prefer to work directly with the coefficients
    # rather than preconditioned parameters.
    #
    # FIXME: Currently used only by pollst.  Decide if this is really the
    # right way to handle that code's needs.  An alternative approach would
    # be to specify the preconditioning parameters in the param methods.
    #----------------------------------------------------------------------

    def getcoefs(self) :

        nkeys = len(self.keys)
        data  = array(nkeys, exim.double)
        for ikey in range(nkeys) :
            key = self.keys[ikey]
            data[ikey] = self.terms[key]

        return data
    

    #----------------------------------------------------------------------
    # Set coefficients
    #
    # This method sets the polynomial coefficients from a vector or list
    # that exactly corresponds to the current keys and terms.  It is used
    # by some optimizers that prefer to work directly with the coefficients
    # rather than preconditioned parameters.
    #
    # FIXME: Currently used only by pollst.  Decide if this is really the
    # right way to handle that code's needs.
    #----------------------------------------------------------------------

    def setcoefs(self, data) :

        for ikey in range(len(self.keys)) :
            key = self.keys[ikey]
            self.terms[key] = data[ikey]

        return
    

    #---------------------------------------------------------------------
    # Return vector of parameters defining the transform
    #
    # The parameters are preconditioned, which will hopefully speed
    # convergence; in any case, it makes the parameters all about the
    # same size, which simplifies writing the optimization algorithms.
    # Specifically, the optimization parameters returned by this method
    # and accepted by the setparam and addparam methods are len ** deg
    # times the difference from the identity transform when len is the
    # value of self.scalen and deg is the total degree of the term; len
    # should be set to approximate the radius of the image, or the maximum
    # absolute coordinate value.
    #
    # FIXME: Check someday whether preconditioning actually provides the
    # benefits claimed above.
    #---------------------------------------------------------------------

    def param(self) :

        params = array(len(self.keys), exim.double)
        for i in range(len(self.keys)) :
            key = self.keys[i]
            degr, outcor, power = key
            coef = self.terms[key]
            if degr == 1 and power[outcor] == 1 :
                coef = coef - 1.0
            coef = coef * self.scalen**degr
            params[i] = coef

        return params


    #---------------------------------------------------------------------
    # Set parameters from a vector
    #---------------------------------------------------------------------

    def setparam(self, parm) :

        for i in range(len(self.keys)) :
            key = self.keys[i]
            degr, outcor, power = key
            coef = parm[i] / self.scalen**degr
            if degr == 1 and power[outcor] == 1 :
                coef = coef + 1.0
            self.terms[key] = coef

        return self


    #----------------------------------------------------------------------
    # Increment parameters from a vector
    #----------------------------------------------------------------------

    def addparam(self, delta) :

        for i in range(len(self.keys)) :
            key = self.keys[i]
            degr, outcor, power = key
            coef  = self.terms[key] + delta[i] / self.scalen**degr
            self.terms[key] = coef

        return self


    #--------------------------------------------------------------------
    # Compose with an affine transform
    #
    # FIXME: This and related code could probably be refactored to good
    # effect.
    #--------------------------------------------------------------------

    def compose(self, other, pre=2) :

        # Check for valid arguments
        if pre not in (-2, -1, 0, 1, 2) :
            raise Exception, "The pre argument may only be -2, -1, 1, or 2"
        if not isinstance(other, affine) and not isinstance(other, poly) :
            errmsg = "Composition valid only for affine or poly transform"
            raise Exception, errmsg
        if self.ndim != other.ndim :
            raise Exception, "Args differ in number of dimensions"

        # Compose with an affine transform
        if isinstance(other, affine) :
            if pre == 0 or pre == 2 :
                self.postcompose(other)
            elif pre == -2 :
                self.postcompose(other.inverse())
            elif pre == 1 :
                self.precompose(other)
            elif pre == -1 :
                self.precompose(other.inverse())

        # Compose with another poly transform
        elif isinstance(other, poly) :
            
            # Preserve all terms but set coefs to zero
            result = {}
            for key in self.terms.keys() :
                result[key] = 0.0

            # Abbreviate some variables
            ndim = self.ndim
            slf  = self.terms
            oth  = other.terms

            # Do the actual composition
            if pre == 0 or pre == 2 :
                self.terms = self.dictcomp(ndim, result, slf, oth)
            elif pre == 1 :
                self.terms = self.dictcomp(ndim, result, oth, slf)
            elif pre in (-1, -2) :
                raise Exception, "Poly transforms are not invertible"

        # Finish up
        self.sortkeys()
        return self
    

    #---------------------------------------------------------------------
    # precompose - Precompose (in place) with an affine transform
    #
    # For internal use
    #
    # FIXME: Check that affine transform has right number of coords
    #
    # FIXME: This method and its subsidiaries seem to do the right thing,
    # but is neither clear nor efficient.  Improve it.
    #---------------------------------------------------------------------

    def precompose(self, aff) :

        # Lift affine transform into a list of dicts, one per outcor
        alist = []
        mat   = aff.matrix
        off   = aff.offset
        ndim  = self.ndim
        for outcor in range(self.ndim) :
            aft = {}
            key = (0, outcor, ndim*(0,))
            aft[key] = off[outcor,0]
            power = (1,) + (ndim-1) * (0,)
            for incor in range(ndim) :
                key  = (1, outcor, power)
                aft[key] = mat[outcor, incor]
                power = (0,) + power[:-1]
            alist.append(aft)

        # Copy all terms but set coefs to zero
        result = {}
        for key in self.terms.keys() :
            result[key] = 0.0

        # Loop over terms in polynomial transform
        for key, coef in self.terms.items() :

            # Recompute terms by substituting the affine transform
            if coef != 0 :
                degr, outcor, power = key
                pterm = {(0, outcor, ndim*(0,)) : coef}
                for incor in range(len(power)) :
                    for j in range(power[incor]) :
                        pterm = self.dictmul(pterm, alist[incor], outcor)
                self.dictadd(result, pterm)

        # Save as new definition and sort terms
        self.terms = result
        self.sortkeys()

        return self
        

    #----------------------------------------------------------------------
    # postcompose - Postcompose (in place) with an affine transform
    #
    # This method is intended to be called by the compose method rather
    # than the general public.
    #
    # FIXME: This method and its subsidiaries seems to do the right thing,
    # but is neither clear nor efficient.  Improve it.
    #----------------------------------------------------------------------

    def postcompose(self, aff) :

        # Get parameters of the affine transpose; verify compatibility
        matrox = aff.matrix
        offset = aff.offset
        affdim = aff.ndim
        if affdim != self.ndim :
            raise Error, "Affine transform has incompatible dimensions"

        # Initialize dictionary of terms for result; the keys are in the
        # form (degree, outcor, power) and the values are the coefs.
        terms = {}
        zeros = affdim * (0,)
        for i in range(affdim) :
            key = (0, i, zeros)
            terms[key] = offset[i,0]

        # Loop over terms
        for iterm in range(len(self.terms)) :
            key  = self.keys[iterm]
            coef = self.terms[key]
            degree, outcor, power = key

            # Loop over rows of the affine matrix
            for irow in range(affdim) :
                affcoef = matrox[irow, outcor]
                key     = (degree, irow, power)
                terms[key] = terms.get(key, 0.0) + coef*affcoef

        # Save as new definition and sort terms
        self.terms = terms
        self.sortkeys()

        return self
       

    #------------------------------------------------------------------------
    # Evaluate transform at a point
    #
    # FIXME: Convert to a C function for performance?  Modify to accept
    # 1-d, row, or column vectors?
    #
    # FIXME: Would this be faster to use the dict directly?
    #------------------------------------------------------------------------

    def point(self, x) :

        y = array(self.nout, exim.double, 0)
        for key, val in self.terms.items() :
            if val != 0 :
                degree, outcor, power = key
                for pow in power :
                    for k in range(pow) :
                        val = val * x[j]
                y[outcor] = y[outcor] + val

        return y


    #---------------------------------------------------------------------
    # Resample image through a polynomial transform - Reference Version
    #----------------------------------------------------------------------

    def linear0(self, source, dimen) :

        outcor, power, coefs = self.make_arrays()
        image = pyvoxC.poly_linear0(outcor, power, coefs, source, dimen)
        return image


    #---------------------------------------------------------------------
    # Resample image through a polynomial transform
    #
    # This is the fastest available verified implementation.  At the moment,
    # however, that is only the reference implementation.
    #----------------------------------------------------------------------

    def linear(self, source, dimen) :

        outcor, power, coefs = self.make_arrays()
        image = pyvoxC.poly_linear0(outcor, power, coefs, source, dimen)
        return image


    #---------------------------------------------------------------------
    # Generate coordinate map
    #----------------------------------------------------------------------

    def map(self, dimen) :

        outcor, power, coefs = self.make_arrays()
        image = pyvoxC.poly_map(outcor, power, coefs, dimen)
        return image



# Provide a class object similar to the other Pyvox type objects
PolyType = poly
         



#---------------------------------------------------------------------------
# spline - Cubic spline transform
#
# This class allows the definition of coordinate transforms (or other
# functions) defined by cubic splines.
#
# Instance variables are: nin = number of dimensions for input; nout =
# number of dimensions for output; ngrid = number of nodes in grid;
# intrans = affine transform for input; outtrans = affine transform for
# output; offset = array of affine offsets; matrix = array of affine matrices.
#
# FIXME: Incomplete; this is just a tentative outline of the interface.
#---------------------------------------------------------------------------

class spline :
    "Cubic spline transform in n dimensions"

    # Initialize a cubic spline transform
    def __init__(self, nin, nout, ngrid) :
        self.nin   = nin
        self.nout  = nout
        self.ngrid = ngrid
        self.intrans  = affine(nin)
        self.outtrans = affine(nout)
        # FIXME: Decide format and initial values for offset and matrix

    # Evaluate at a point
    def point(self, x) :
        pass     # FIXME: Not yet implemented

    # Return affine approximation at a point
    def local(self, x) :
        pass     # FIXME: Not yet implemented

    # Set the local affine approximation at at node
    def setlocal(self, node, affine) :
        pass     # FIXME: Not yet implemented

    # Compose with an affine transform
    def compose(self, affine, pre=1) :
        pass     # FIXME: Not yet implemented

    # Resample an image using nearest neighbor interpolation
    def nearest(self, source, dimen) :
        pass     # FIXME: Not yet implemented

    # Resample an image using linear interpolation
    def linear(self, source, dimen) :
        pass     # FIXME: Not yet implemented



#-----------------------------------------------------------------------------
# ccmap - Connected component map
#
# This class represents the connected components found in an image, and
# allows the user to get descriptions of the components (usually just called
# regions for brevity) or maps showing which pixels belong to which components.
#
# FIXME: Check for pathological cases, like an image with no objects.
# Add a regression test.
#----------------------------------------------------------------------------

class ccmap :

    "Connected component map"


    #-------------------------------------------------------------------------
    # Initialize a connected component map
    #
    # Most of the real work is done in a C function.
    #-------------------------------------------------------------------------

    def __init__(self, image, **kwdict) :

        # Get the keyword arguments
        minsize = kwdict.get("minsize", 0)
        maxreg  = kwdict.get("maxreg",  1000000)

        # Create instance variables; values are only temporary
        self.nregions = 0      # Number of regions, including backgroun
        self.reglist  = []     # List of [regno, count, seed] triples
        self.regmap   = 0      # Image with regno for each pixel
        
        # Scan the image to find connected components.
        # Rawlist is a list of (regno, canon, count, seed) tuples; rawmap
        # is an unsigned char or short image that shows the region number
        # for each pixel.
        rawlist, rawmap = pyvoxC.ccmap(image)

        # Sort nonbackground regions by decreasing number of pixels
        back = rawlist[0]      # Background region
        rest = rawlist[1:]     # All the other regions
        rest.sort(lambda x,y : y[2] - x[2])

        # Exclude regions according to minsize and maxreg; add their counts
        # to the background region.
        nreg = 1        # Number of regions after exclusions, with background
        for n in range(len(rest)) :
            region = rest[n]
            if region[2] < minsize or n >= maxreg :
                region[1] = 0           # Reassign to background
                back[2] += region[2]    # Include count in background
            else :
                nreg += 1

        # Reassign region numbers in order of decreasing pixel counts.
        type = exim.uchar                   # Type for region map
        if nreg > 256 :
            type = exim.ushort
        nold = len(rawlist)                 # Number of old region numbers
        old2new = array(nold, type, 0)      # Map old to new region nmbrs
        newlist = [[0, back[2], None]]      # Initialize new list of regions
        for n in range(len(rest)) :
            regno, canon, count, seed = rest[n]
            if regno == canon :
                old2new[regno] = n + 1
                newlist.append([n+1, count, seed])
            else :
                old2new[regno] = old2new[canon]
        newmap = rawmap.lookup(old2new)

        # Save the results for future reference
        self.reglist  = newlist
        self.regmap   = newmap
        self.nregions = len(newlist)

        return


    #------------------------------------------------------------------------
    # bigob - Return map for the largest nonbackground region
    #
    # This is provided mainly to assist those who used the now deprecated
    # function array.bigob().  Returns None if there are no objects.
    #-----------------------------------------------------------------------

    def bigob(self) :
        
        if self.nregions > 0 :
            return self.map(1)
        else :
            return None
        

    #-------------------------------------------------------------------------
    # count - Report number of pixels in a region
    #-------------------------------------------------------------------------

    def count(self, region=1) :

        if region >= 0 and region < self.nregions :
            return self.reglist[region][1]
        else :
            return 0


    #-----------------------------------------------------------------------
    # map - Show which pixel belongs to which region
    #
    # FIXME: Throw an error if there is no such region?
    #
    # FIXME: Add an option to specify the code number to use instead on 255?
    # Add an option to specify the type of the map?
    #----------------------------------------------------------------------

    def map(self, region=None) :

        # Case 1: Report on all regions
        if region == None :
            return self.regmap.copy()

        # Case 2: Report on a specified region
        else :
            return self.regmap.compare(region, 0, 255, 0)
            

    #----------------------------------------------------------------------
    # regions - Return list of nonbackground regions
    #----------------------------------------------------------------------

    def regions(self) :

        return self.reglist[1:]
    

    #---------------------------------------------------------------------
    # seed - Return a point contained in some specified region
    #
    # Note that the background (region 0) need not be connected and thus
    # does not have a seed point.
    #---------------------------------------------------------------------

    def seed(self, region=1) :

        if region > 0 and region < self.nregions :
            return self.reglist[region][2]
        else :
            return None
