##############################################################################
# reged - Generic region editor and image viewer
#
# Author:  Paul Hughett
#
# Copyright (c) 2002-2005 Trustees of the University of Pennsylvania
# See the file COPYING for license terms.
#
#
# This Python module implements an interactive image viewer and region
# editor for brain image analysis.
#
# A note on terminology: the "map" image generally defines multiple regions;
# the "mask" image derived from the map image defines only a single region.
# But that distinction may not be observed consistently below.
#
# FIXME: Complex types are not supported, and won't be until someone comes
# up with a use case for them.
#
# FIXME: The sliders for intensity windowing do not work right for float
# images where the range of values is small; they can assume integer values
# only, which has definite problems when the iamge intensity range is [0,1].
#
# FIXME: Show the current base, resliced, and map image filenames.
#
# FIMXE: Rewrite in a more modern window toolkit, of which PyGTK seems
# to be a good choice, if I can ever get the time to learn it.
# 
# FIXME: Define the menus in tables or a config file; and work out a better
# mechanism for enabling and disabling menu items.
#
# FIXME: Add the capability to read a tissue map and compute region/tissue
# volumes.  Or should that be a separate command line program?
#
# FIXME: Add a menu command to edit the header file, after the fashion of
# editavw.
#
# FIXME: Write a separate mode or program to support interactive parameter
# determination for raw images without headers.  The parameters to be
# found should include byte order; data type; number of rows, columns and
# echoes; intensity range(?); and perhaps others.  This program should write
# an AnalyzeView header to save the parameters found.  There is some relic
# code in this file that might be useful for this purpose.  Allow the user
# (that is, the task-specific editor) to specify default values for nrows,
# ncols, extype, and bigend.  Note that array.rgba2d() is probably the
# right tool for this purpose; but note that there is no tool for the
# reverse conversion.
# 
# FIXME: Now that image headers are supported, images should know their own
# pixel size, etc.  Remove the instance variables that are duplicating these.
#
# FIXME: Generalize to also handle 2D and 4D images without hacks.
#
# FIXME: Show images with anisotropic sampling in their correct aspect.
#
# FIXME: Would it make sense to refactor this into image engine and
# GUI components?
#
# FIXME: Add support for RGB images, at least in the image viewer.
#
# FIXME: Allow the region definitions to be read from a file?
#
# FIXME: Test carefully in 'index' mode; I don't think that has been
# seriously tested recently.
#
# FIXME: Catch exceptions in main loop or just below, and try to recover
# and produce a useful error message.  Convert as many exceptions as 
# possible into a displayed error message.
#
# FIXME: Add the capability to show the mask image by itself; this should
# be useful for catching stray masked/unmasked pixels.
#
# FIXME:  Add a "smart brush" that fills only up to an edge in the base
# image.
#
# FIXME: Refactor to support the following viewers and editors: (1) raw
# image viewer with fast format determination and header creation;
# (2) plain image viewer; (3) brain mask editor; (4) region map editor;
# and (5) image reslicer.  Consider renaming the refactored modules with
# some more appropriate name.  Note that these refactored classes will
# often be further specialized to specific analysis protocols.
#
# FIXME: A raw image viewer should allow the user to specify any of the
# following parameters while viewing an image: image rank and dimensions,
# pixel size, data type, byte order, intensity range, axes, and number of
# bands/echoes.
#  
# FIXME: Add menu function to open an existing map file, or read from an
# SROI file.  Add fill within a polygon.
#
# FIXME: "Edit/Remove All" should be disabled if no map is present.  Gather
# the no map enables/disables into a single function.  Note also the case when
# the map exists but is unnamed.
#
# FIXME: Work on the esthetics of the status window.
#
# FIXME: Brush update needs to be faster, to make drawing smoother without
# skipping.  Interpolate linearly between event positions.
#
# FIXME: Limit the zoom for large images and small screens; or provide a
# scrolling window for the image display.
#
# FIXME: Add the capability to adjust the relative salience of the base
# and mask images; a fixed value doesn't seem good enough.  
#
# FIXME: Label the axes with the anatomical directions, including pageup
# and down.
#
# FIXME: Define shift-PageUp and shift-PageDn to move by increments of 10
# slices.  Ditto arrow keys.
#
# FIXME: Add support for RGB images?  Possibly not useful in light of
# intended purpose.
#
# FIXME: Detect and report errors in some useful fashion, beginning with
# size mismatches between base and map images.  This includes handling the
# various exceptions that get generated in subroutines.
#
# FIXME: Generalize to provide all the capabilities of qdv, ultimately
# replacing it.  The trickiest capability to replace will probably be to
# quickly flip through types, byte orders, and image sizes to identify
# a mystery file.
#
##############################################################################

rcsid = "$Id: reged.py,v 1.39 2005/11/03 19:36:41 hughett Exp $  $Name: pyvox-0-72 $"

import copy
import re, string
import Tkinter
import tkMessageBox
import Pmw             # Python MegaWidgets

from exceptions import Exception

import pyvox, exim
import tkphoto


# Abbreviate some Tkinter methods
Canvas = Tkinter.Canvas
Entry  = Tkinter.Entry
Frame  = Tkinter.Frame
Label  = Tkinter.Label
Menu   = Tkinter.Menu
Menubutton = Tkinter.Menubutton
Scale  = Tkinter.Scale
IntVar = Tkinter.IntVar
StringVar  = Tkinter.StringVar
Toplevel   = Tkinter.Toplevel
HORIZONTAL = Tkinter.HORIZONTAL
LEFT   = Tkinter.LEFT
TOP    = Tkinter.TOP
NW     = Tkinter.NW
W      = Tkinter.W



#----------------------------------------------------------------------------
# canonaxes - Validate and canonicalize anatomical axes description
#
# A valid axes description is exactly three characters long, all upper case,
# and contains exactly one of AP, one of LR, and one of IS.  This function
# returns a canonicalized descriptor if possible and None otherise.
#----------------------------------------------------------------------------

def canonaxes(axes) :
    
    axes = string.upper(axes[:3])
    if not re.search('A', axes) and not re.search('P', axes) :
        return None
    elif not re.search('L', axes) and not re.search('R', axes) :
        return None
    elif not re.search('I', axes) and not re.search('S', axes) :
        return None
    else :
        return axes

    

#---------------------------------------------------------------------------
# findaxes - find axes and direct for given src and dest description
#
# This function determines the axes and direct values implied by given
# pixel orders for the source and destination.  A pixel order is defined
# by a three-character string similar to "SPL", where the first character
# (in this example) means that the outermost index increases from inferior
# to Superior, the middle index increases from anterior to Posterior, and
# the last index increases from right to Left.  The possible letters are
#
#     A  = anterior  vs   P = posterior
#     I  = inferior  vs   S = superior
#     L  = left      vs   R = right
#----------------------------------------------------------------------------

def findaxes (src, dest) :

    axes   = 3 * [None]
    direct = 3 * [None]

    # Handle anterior/posterior axes
    asrc = string.find(src,  'A')
    adst = string.find(dest, 'A')
    psrc = string.find(src,  'P')
    pdst = string.find(dest, 'P')
    if asrc >= 0 and adst >= 0 :
        axes  [adst] = asrc
        direct[adst] = 1
    elif asrc >= 0 and pdst >= 0 :
        axes  [pdst] = asrc
        direct[pdst] = -1
    elif psrc >= 0 and pdst >= 0 :
        axes  [pdst] = psrc
        direct[pdst] = 1
    elif psrc >= 0 and adst >= 0 :
        axes  [adst] = psrc
        direct[adst] = -1
    else :
        raise Exception, "Anterior/posterior axis not specified"
    
    # Handle inferior/superior axes
    isrc = string.find(src,  'I')
    idst = string.find(dest, 'I')
    ssrc = string.find(src,  'S')
    sdst = string.find(dest, 'S')
    if isrc >= 0 and idst >= 0 :
        axes  [idst] = isrc
        direct[idst] = 1
    elif isrc >= 0 and sdst >= 0 :
        axes  [sdst] = isrc
        direct[sdst] = -1
    elif ssrc >= 0 and sdst >= 0 :
        axes  [sdst] = ssrc
        direct[sdst] = 1
    elif ssrc >= 0 and idst >= 0 :
        axes  [idst] = ssrc
        direct[idst] = -1
    else :
        raise Exception, "Inferior/superior axis not specified"
    
    # Handle left/right axes
    lsrc = string.find(src,  'L')
    ldst = string.find(dest, 'L')
    rsrc = string.find(src,  'R')
    rdst = string.find(dest, 'R')
    if lsrc >= 0 and ldst >= 0 :
        axes  [ldst] = lsrc
        direct[ldst] = 1
    elif lsrc >= 0 and rdst >= 0 :
        axes  [rdst] = lsrc
        direct[rdst] = -1
    elif rsrc >= 0 and rdst >= 0 :
        axes  [rdst] = rsrc
        direct[rdst] = 1
    elif rsrc >= 0 and ldst >= 0 :
        axes  [ldst] = rsrc
        direct[ldst] = -1
    else :
        raise Exception, "Inferior/superior axis not specified"

    # Check that everything got specified somewhere
    for i in range(3) :
        if axes[i] == None or direct[i] == None :
            raise Exception, "Hunh?  Some axis got missed"

    return axes, direct



#---------------------------------------------------------------------------
# viewer class - Core capabilities for an image viewer/editor
#
# ALGORITHM:  Coords z,y,x refer to slices, rows, and columns within the
# image file; coords w,v,u refer to screen depth, down, and right within
# the image display window.
#
# FIXME: Incompletely separated from reged and not yet usable alone or as a
# base class for some other type of image editor.
#
# FIXME: Support for reslicing arguably belongs in a separate class,
# assuming I can find some reasonable way of separating the functions.
#---------------------------------------------------------------------------

class viewer :
    
    #------------------------------------------------------------------------
    # Instance variables to be directly configured by the user:
    #
    #  parent     Parent window, if any
    #  fileaxes   Order of pixel axes in the file
    #
    #
    # Other instance variables:
    #
    #  zoom       Current image zoom factor
    #  minzoom    Minimum image zoom factor; depends on image size
    #  maxzoom    Maximum image zoom factor; depends on image size
    #  status     Status window
    #  canvas     Canvas on which images are displayed
    #  basefile   Filename from which base image was read
    #  base       Base image, as Pyvox image; None if no base image defined
    #  dimen      Dimensions of base image, as a volume in 3D
    #  tkbase     Tkphoto to contain the base plus mask image
    #  imgitem    Canvas item containing the base image
    #  noimgitem  Canvas text item explaining that no image is open
    #  horitem    Canvas line item for the horizontal crosshair
    #  veritem    Canvas line item for the vertical crosshair
    #  coords     Coordinates of displayed position
    #  axes       File axes to map to screen depth, down, and right
    #  direct     Direction on each screen axis: 1 => fwd; -1 => rev
    #  cbyte      Current selection of byte order: extype, bigend, and chan
    #  extype     Assumed external type of base imge
    #  bigend     Assumed byte order of base image
    #  nrows      Assumed number of rows per slice
    #  ncols      Assumed number of columns per slice
    #  viewaxes   Order of pixel axes in the view
    #  minwin     Minimum value (from slider) for intensity window
    #  maxwin     Maximum value (from slider) for intensity window
    #  invert     Invert the intensity window?
    #  chan       Mapping to RGBA channels
    #  cross      Display crosshairs?
    #  format     StringVar containing current pixel format string
    #  czsvar     StringVar containing current z value
    #  cysvar     StringVar containing current y value
    #  cxsvar     StringVar containing current x value
    #
    #
    #  Following are for reslicing:
    #
    #  acpoint    Coordinates of the anterior commissure
    #  pcpoint    Coordinates of the posterior commissure
    #  mspoint    Coordinates of a mid-sagittal point
    #  rawbase    Base image before resampling
    #  rawfile    File that unresampled base image was read from
    #  resliced   Base image after resampling
    #  retrani    Transform from new index coords to old index coords
    #  retranp    Transform from new physical coords to old physical coords
    #------------------------------------------------------------------------


    #------------------------------------------------------------------------
    # Class variables
    #------------------------------------------------------------------------

    # FIXME: Not required for AnalyzeView headers, but may be useful
    # later for an interactive raw image viewer/specifier.
    #dataformats = [          # Possible formats of base image file
    #    # label                extype    bigend   chan
    #    ['Unsigned Char',     exim.uint1,  0,    [0,0,0] ],
    #    ['Unsigned Short BE', exim.uint2,  1,    [0,0,0] ],
    #    ['Unsigned Short LE', exim.uint2,  0,    [0,0,0] ],
    #    ]


    #------------------------------------------------------------------------
    # Initialize
    #
    # Order of arguments is not guaranteed.
    #
    # FIXME: Allow less finicky alternatives to the baseformat parameter;
    # e.g. exim.uint2BE?
    #------------------------------------------------------------------------
    def __init__(self, parent=None, fileaxes='SPL',
                 baseformat='Unsigned Char') :

        # Remember permanent configuration parameters
        self.parent   = parent
        self.fileaxes = fileaxes
        
        # Set initial data format assumptions
        self.nrows = 256
        self.ncols = 256

        # Set base format
        # FIXME: This code not needed to support AnalyzeView headers
        #self.setbaseformat(baseformat)

        # Set initial display parameters
        self.minwin   =   0
        self.maxwin   = 255
        self.invert   = 0
        self.zoom     = 2   
        self.coords   = [128, 128, 128]
        self.viewaxes = 'SPL'
        self.cross    = 1

        # Initialize reslicing parameters
        self.rawbase  = None
        self.rawfile  = None
        self.retrani  = None
        self.retranp  = None
        self.acpoint  = None
        self.pcpoint  = None
        self.mspoint  = None

        # Setup file to screen mapping
        self.axes, self.direct = findaxes(self.fileaxes, self.viewaxes)

        # Initialize the status window
        self.initstatwind()
        
        # Initialize items on the viewing window canvas
        self.noimgitem = None
        self.imgitem = None
        self.horitem = None
        self.veritem = None
        
        # Create and configure the image viewing window
        self.viewwin = Toplevel(self.status)
        self.viewwin.title('Image Viewer')
        self.bindcommon(self.viewwin)
        self.viewwin.bind('<ButtonRelease-3>', self.setcoords)
        self.canvas = Canvas(self.viewwin, height=self.nrows, width=self.ncols)
        self.canvas.pack()

        # Set up and display the base and mask images
        self.tkbase  = tkphoto.tkphoto()
        self.base    = None

        return


    #-------------------------------------------------------------------------
    # Initialize the status window
    #
    # FIXME: Add display for ac, pc, and ms points?
    #-------------------------------------------------------------------------

    def initstatwind(self) :

        # Create and configure the status window
        # FIXME: Add more useful status fields to the window
        # FIXME: Report the current, min, and max zoom factor
        if self.parent :
            self.status = Toplevel(self.parent)
        else :
            self.status = Tkinter.Tk()
        self.status.title('Status Window')
        self.bindcommon(self.status)

        # Create menu bar in the status window
        menubar = Frame(self.status, relief=Tkinter.RAISED, borderwidth=2)
        menubar.pack(fill=Tkinter.X)

        # Create File menu button
        filebtn = Menubutton(menubar, text='File')
        filebtn.pack(side=LEFT, padx=5)
        filemenu = Menu(filebtn)
        filebtn['menu'] = filemenu
        filemenu.add_command(label='Quit',     command=self.quit)
        self.filemenu = filemenu

        # Create Edit menu button
        editbtn = Menubutton(menubar, text='Edit')
        editbtn.pack(side=LEFT, padx=5)
        editmenu = Menu(editbtn)
        editbtn['menu'] = editmenu
        self.editmenu = editmenu

        # Create and populate Reslice menu button
        # FIXME: Reslice action should be inactive until all points defined
        # FIXME: Save Resliced should be inactive until resliced is done
        reslbtn = Menubutton(menubar, text='Reslice')
        reslbtn.pack(side=LEFT, padx=5)
        reslmenu = Menu(reslbtn)
        reslbtn['menu'] = reslmenu
        self.reslmenu = reslmenu
        reslmenu.add_command(label='Set Antr Comm',  command=self.setac)
        reslmenu.add_command(label='Set Post Comm',  command=self.setpc)
        reslmenu.add_command(label='Set Mid Sag Pt', command=self.setms)
        reslmenu.add_command(label='Reslice Image',  command=self.reslice)
        reslmenu.add_command(label='Save Resliced As...',
                             command=self.save_resliced_as)

        # Create Compute menu button
        compbtn = Menubutton(menubar, text='Compute')
        compbtn.pack(side=LEFT, padx=5)
        compmenu = Menu(compbtn)
        compbtn['menu'] = compmenu
        self.compmenu = compmenu

        # Add menu buttons to the menu bar
        menubar.tk_menuBar(filebtn)
        menubar.tk_menuBar(editbtn)
        menubar.tk_menuBar(compbtn)

        # Create Option Menu for data format
        # FIXME: Not required to support AnalyzeView headers but may
        # be useful for a raw image viewer
        #fmtoptions = []
        #for i in range(len(self.dataformats)) :
        #    fmtoptions.append(self.dataformats[i][0])
        #initfmt = self.dataformats[self.cbyte][0]
        #entry = Pmw.OptionMenu(self.status, labelpos=W,
        #                       label_text='Format:',
        #                       items=fmtoptions,
        #                       initialitem = initfmt,
        #                       menubutton_width=20,
        #                       command=self.newformat)
        #entry.pack(anchor=W, padx=5, pady=5)
        #self.formatmenu = entry

        # Start frame to display coordinates
        frame = Frame(self.status)
        Label(frame, text='Index and physical coordinates').pack(side=TOP)

        # Display index and physical coordinates for Z
        # FIXME: Add capability to set physical coords
        frame1 = Frame(frame)
        Label(frame1, text='Z =').pack(side=LEFT)
        self.czsvar = StringVar()      # Index coord
        entry = Entry(frame1, textvariable=self.czsvar, width=4)
        entry.pack(side=LEFT)
        entry.bind('<Return>', self.newz)
        Label(frame1, text='     ').pack(side=LEFT)
        self.pzsvar = StringVar()      # Physical coord
        entry = Entry(frame1, textvariable=self.pzsvar, width=7)
        entry.pack(side=LEFT)
        #entry.bind('<Return>', self.newz)
        frame1.pack(side=TOP)

        # Display index and physical coordinates for Y
        frame1 = Frame(frame)
        Label(frame1, text='Y =').pack(side=LEFT)
        self.cysvar = StringVar()      # Index coord
        entry = Entry(frame1, textvariable=self.cysvar, width=4)
        entry.pack(side=LEFT)
        entry.bind('<Return>', self.newy)
        Label(frame1, text='     ').pack(side=LEFT)
        self.pysvar = StringVar()      # Physical coord
        entry = Entry(frame1, textvariable=self.pysvar, width=7)
        entry.pack(side=LEFT)
        #entry.bind('<Return>', self.newy)
        frame1.pack(side=TOP)

        # Display index and physical coordinates for X
        frame1 = Frame(frame)
        Label(frame1, text='X =').pack(side=LEFT)
        self.cxsvar = StringVar()      # Index coord
        entry = Entry(frame1, textvariable=self.cxsvar, width=4)
        entry.pack(side=LEFT)
        entry.bind('<Return>', self.newx)
        Label(frame1, text='     ').pack(side=LEFT)
        self.pxsvar = StringVar()      # Physical coord
        entry = Entry(frame1, textvariable=self.pxsvar, width=7)
        entry.pack(side=LEFT)
        #entry.bind('<Return>', self.newx)
        frame1.pack(side=TOP)

        # Finish frame to display coordinates
        frame.pack(side=TOP, anchor=W, padx=5, pady=5)

        # Create scale for minimum of the intensity window
        scale = Scale(self.status, label="Minimum intensity",
                      orient=HORIZONTAL, length=128,
                      from_=0, to=512,
                      command=self.newintensity)
        scale.set(self.minwin)
        self.minwind = scale
        scale.pack(side=TOP, anchor=W)

        # Create scale for maximum of the intensity window
        scale = Scale(self.status, label="Maximum intensity",
                      orient=HORIZONTAL, length=128,
                      from_=0, to=512,
                      command=self.newintensity)
        scale.set(self.maxwin)
        self.maxwind = scale
        scale.pack(side=TOP, anchor=W)

        # Create Option Menu for possible views
        self.viewoptions = ('Axial', 'Coronal', 'Sagittal')
        entry = Pmw.OptionMenu(self.status, labelpos = W,
                               label_text='View:',
                               items=self.viewoptions,
                               menubutton_width = 20,
                               command = self.newview)
        entry.pack(anchor=W, padx=5, pady=5)
        self.viewmenu = entry

        return


    #---------------------------------------------------------------------
    # Define common bindings for a toplevel window
    #
    # (Tkinter apparently does not propagate key events from one toplevel
    # window (view) to the other (status), so we bind keys to each).
    #
    # FIXME: Binding for space bar is specific to region editor.
    #
    # FIXME: Bindings for +/= and - are relics of qdv; remove them?
    #---------------------------------------------------------------------

    def bindcommon(self, widget) :
        widget.bind('=', self.incraxis)        # +/=  => incr depth axis
        widget.bind('-', self.decraxis)        # - => decr depth axis
        widget.bind('a', self.nextview)        # a => next view
        # FIXME: Following not needed to support AnalyzeView headers
        #widget.bind('b', self.nextformat)      # b => next format/byte order
        widget.bind('c', self.nextcross)       # c => toggle crosshairs on/off
        widget.bind('i', self.nextinvert)      # i => toggle intensity invert
        if not self.parent :
            widget.bind('q', self.quit)        # q => quit
        widget.bind('z', self.nextzoom)        # z => increase zoom
        widget.bind('<space>', self.nextmask)  # Space => toggle mask on/off
        widget.bind('<Down>',  self.down)      # Down  => crosshair down
        widget.bind('<Left>',  self.left)      # Left  => crosshair left
        widget.bind('<Right>', self.right)     # Right => crosshair right
        widget.bind('<Up>',    self.up)        # Up    => crosshair up
        widget.bind('<Prior>', self.decraxis)  # Page up => previous slice
        widget.bind('<Next>',  self.incraxis)  # Page dn => next slice


    #-----------------------------------------------------------------------
    # Set data format for the base image
    #
    # FIXME: This function is not needed to support AnalyzeView headers but
    # might be useful for a raw image viewer.
    #-----------------------------------------------------------------------

    #def setbaseformat(self, format) :
    #    
    #    # Set selected initial format of base image file
    #    self.cbyte = 0
    #    for i in range(len(self.dataformats)) :
    #        if format == self.dataformats[i][0] :
    #            self.cbyte = i
    #    ignore, self.extype, self.bigend, self.chan = \
    #            self.dataformats[self.cbyte]
    #
    #    return

    
    #-----------------------------------------------------------------------
    # Get previous slice along current axis
    #
    # Or: Decrement coordinate on current axis
    #
    # FIXME: Holding down the key does not work right; it looks like
    # Tkinter holds redrawing until the key is released; will calling
    # widget.update() in redraw fix this?
    #-------------------------------------------------------------------------

    def decraxis(self, event=None) :
        if not self.base :
            return
        self.saveslice()
        caxis = self.axes[0]
        coord = self.coords[caxis]
        limit = self.dimen [caxis]
        step  = self.direct[0]
        self.coords[caxis] = self.incr(coord, -step, 0, limit)
        self.redraw()


    #------------------------------------------------------------------------
    # Move crosshairs down
    #-----------------------------------------------------------------------

    def down(self, event=None) :
        if not self.base :
            return
        iaxis = self.axes[1]
        step  = self.direct[1]
        limit = self.dimen[iaxis]
        self.coords[iaxis] = self.incr(self.coords[iaxis], step, 0, limit)
        self.redraw()


    #------------------------------------------------------------------------
    # Extract 2-D slice from a 3-D or 4-D image
    #
    # FIXME: Views may be a better way to handle this
    #------------------------------------------------------------------------

    def getslice(self, array, coords, axes, direct) :
        vaxis = axes[1]      # Screen down
        uaxis = axes[2]      # Screen right
        vcor = slice(None, None, direct[1])
        ucor = slice(None, None, direct[2])
        subs = coords[0:3]
        subs[vaxis] = vcor
        subs[uaxis] = ucor
        if array.rank == 4 :
            subs = [0] + subs
        result = array[subs].reshape()
        if (vaxis, uaxis) in ((1,0), (2,0), (2,1)) :
            result = result.trans()
        return result
        

    #----------------------------------------------------------------------
    # Get next slice along current depth axis
    # Or: Increment coordinate on current axis
    #----------------------------------------------------------------------
    
    def incraxis(self, event=None) :
        if not self.base :
            return
        self.saveslice()
        caxis = self.axes[0]
        coord = self.coords[caxis]
        limit = self.dimen [caxis]
        step  = self.direct[0]
        self.coords[caxis] = self.incr(coord, step, 0, limit)
        self.redraw()


    #-----------------------------------------------------------------------
    # Increment/decrement and wraparound within limits [lower, upper)
    #-----------------------------------------------------------------------
    
    def incr(self, value, amount, lower, upper) :
        value = value + amount
        while value < lower :
            value = value + (upper - lower)
        while value >= upper :
            value = value - (upper - lower)
        return value


    #-------------------------------------------------------------------------
    # Move crosshairs left
    #-------------------------------------------------------------------------
    
    def left(self, event=None) :
        if not self.base :
            return
        iaxis = self.axes[2]
        step  = self.direct[2]
        limit = self.dimen[iaxis]
        self.coords[iaxis] = self.incr(self.coords[iaxis], -step, 0, limit)
        self.redraw()


    #------------------------------------------------------------------------
    # Event loop for an unparented region editor
    #------------------------------------------------------------------------
    
    def mainloop(self) :
        if self.parent :
            raise Exception, "You must use the root mainloop"
        else :
            self.status.mainloop()


    #-----------------------------------------------------------------------
    # Select pixel format from option menu in status window
    #
    # FIXME: This code is not needed to support AnalyzeView headers but
    # might be useful for a raw image viewer.
    #-----------------------------------------------------------------------
    
    #def newformat(self, format) :
    #    
    #    if not self.base :
    #        return
    #    self.saveslice()
    #
    #    # Set the format that was just selected
    #    self.setbaseformat(format)
    #
    #    # Reread the base image file using the new format;
    #    # display "Processing..." window during reread
    #    procwin = Toplevel(self.status)
    #    Label(procwin, text="Processing...").pack(padx=50,pady=50)
    #    procwin.transient(self.status)
    #    self.status.update()
    #    self.readbase(self.basefile)
    #    self.status.config(cursor='top_left_arrow')
    #    procwin.destroy()
    #    self.redraw()


    #----------------------------------------------------------------------
    # Redraw with new limits for intensity window
    #----------------------------------------------------------------------
    
    def newintensity(self, newvalue=None) :
        self.minwin = self.minwind.get()
        self.maxwin = self.maxwind.get()
        self.redraw()

            
    #----------------------------------------------------------------------
    # Select a new view
    #----------------------------------------------------------------------
    
    def newview(self, view=None) :
        if not self.base :
            return
        self.saveslice()
        if view == None :
            pass
        elif view == self.viewoptions[0] :
            self.viewaxes = 'SPL'               # Axial view
        elif view == self.viewoptions[1] :
            self.viewaxes = 'PIL'               # Coronal
        elif view == self.viewoptions[2] :
            self.viewaxes = 'RIP'               # Sagittal
        else :
            raise Exception, "Unknown view %s selected" % view
        self.axes, self.direct = findaxes(self.fileaxes, self.viewaxes)
        self.redraw()


    #-----------------------------------------------------------------------
    # New x coord entered in status window
    #-----------------------------------------------------------------------
    
    def newx(self, event=None) :
        if not self.base :
            return
        self.saveslice()
        self.coords[0] = string.atoi(self.cxsvar.get())
        self.redraw()


    #------------------------------------------------------------------------
    # New y coord entered in status window
    #------------------------------------------------------------------------
    
    def newy(self, event=None) :
        if not self.base :
            return
        self.saveslice()
        self.coords[0] = string.atoi(self.cysvar.get())
        self.redraw()


    #------------------------------------------------------------------------
    # New z coord entered in status window
    #------------------------------------------------------------------------
    
    def newz(self, event=None) :
        if not self.base :
            return
        self.saveslice()
        self.coords[0] = string.atoi(self.czsvar.get())
        self.redraw()


    #------------------------------------------------------------------------
    # Next axis selection
    #-----------------------------------------------------------------------
    
    def nextview(self, event=None) :
        if not self.base :
            return
        caxis = self.incr(self.axes[0], 1, 0, 3)
        self.viewmenu.invoke(self.viewoptions[caxis])


    #------------------------------------------------------------------------
    # Next byte order selection
    #
    # FIXME: This code is not needed to support AnalyzeView headers but
    # may be useful for a raw image viewer.
    #------------------------------------------------------------------------
    
    #def nextformat(self, event=None) :
    #    if not self.base :
    #        return
    #    self.cbyte = self.incr(self.cbyte, 1, 0, len(self.dataformats))
    #    self.formatmenu.invoke(self.dataformats[self.cbyte][0])


    #------------------------------------------------------------------------
    # Next possible crosshairs state
    #------------------------------------------------------------------------
    
    def nextcross(self, event=None) :
        if not self.base :
            return
        self.cross = not self.cross
        self.redraw()


    #------------------------------------------------------------------------
    # i => invert intensity window
    #------------------------------------------------------------------------
    
    def nextinvert(self, event=None) :
        self.invert = not self.invert
        self.redraw()


    #------------------------------------------------------------------------
    # Next zoom selection
    #------------------------------------------------------------------------
    
    def nextzoom(self, event=None) :

        if not self.base :
            return
        self.zoom = self.incr(self.zoom, 1, self.minzoom, self.maxzoom+1)
        self.redraw()


    #------------------------------------------------------------------------
    # Insert a 2-D slice into a 3-D or 4-D image
    #
    # FIXME: Views might be a better way to do this
    #------------------------------------------------------------------------
    
    def putslice(self, array, slicedata, coords, axes, direct) :
        vaxis = axes[1]      # Screen down
        uaxis = axes[2]      # Screen right
        vcor = slice(None, None, direct[1])
        ucor = slice(None, None, direct[2])
        subs = array.rank * [0]
        subs = coords[0:3]
        subs[vaxis] = vcor
        subs[uaxis] = ucor
        if array.rank == 4 :
            subs = [0] + subs
        if (vaxis, uaxis) in ((1,0), (2,0), (2,1)) :
            slicedata = slicedata.trans()
        array[subs] = slicedata

        
    def oldputslice(self, array, slicedata, coords, axes, direct) :
        vaxis = axes[1]      # Screen down
        uaxis = axes[2]      # Screen right
        vcor = slice(None, None, direct[1])
        ucor = slice(None, None, direct[2])
        if vaxis == 0 and uaxis == 1 :
            array[vcor, ucor, coords[2]] = slicedata
        elif vaxis == 1 and uaxis == 0 :
            array[ucor, vcor, coords[2]] = slicedata.trans()
        elif vaxis == 0 and uaxis == 2 :
            array[vcor, coords[1], ucor] = slicedata
        elif vaxis == 2 and uaxis == 0 :
            array[ucor, coords[1], vcor] = slicedata.trans()
        elif vaxis == 1 and uaxis == 2 :
            array[coords[0], vcor, ucor] = slicedata
        elif vaxis == 2 and uaxis == 1 :
            array[coords[0], ucor, vcor] = slicedata.trans()
        else :
            raise Exception, "Unknown axes requested in putslice"

    
    #------------------------------------------------------------------------
    # Quit command for unparented region editor only
    #
    # FIXME: Should offer to save the resliced image, if not already saved.
    #------------------------------------------------------------------------
    
    def quit(self, event=None) :
        if self.parent :
            raise Exception, "Attempt to quit a parented region editor"
        else :
            self.saveslice()
            if self.querysavemap() :
                return               # Cancelled by user
            self.status.quit()


    #------------------------------------------------------------------------
    # Read a base image; does NOT redraw the image
    #
    # FIXME: This code is not needed to support AnalyzeView headers but
    # may be useful for a raw image viewer.
    #------------------------------------------------------------------------
    
    def oldreadbase(self, filename) :

        # Allow user to save old base image first
        self.saveslice()
        if self.querysavemap() :
            return                  # If user presses cancel

        # Check that the new base image file exists
        self.basefile = pyvox.truefile(filename)
        if not self.basefile :
            tkMessageBox.showerror("Invalid Base File",
                                   "The requested base image file does not " \
                                   "exist or is not readable.")
            return

        # Read a raw base image
        dimen = [0, self.nrows, self.ncols]
        self.base = pyvox.rawread(self.basefile, dimen,
                                  self.extype, self.bigend)

        # Remember dimensions and set the cursor position
        self.dimen  = self.base.size
        self.coords = [self.dimen[0] / 2, self.dimen[1] / 2, self.dimen[2] / 2]

        # Set max gray level in the intensity windowing widget
        maxgray = self.base.amax()
        self.minwind.config(to=maxgray)
        self.maxwind.config(to=maxgray)
        self.maxwind.set(maxgray)
        
        return


    #------------------------------------------------------------------------
    # Read a base image; does NOT redraw the image
    #
    # FIXME: This contains some hacks to handle 2-D and 4-D AnalyzeView images.
    # A 2-D image will be promoted to a 3-D image that contains only a single
    # slice.  A 4-D image for which one or more dimensions has length one
    # will have the enough unit-length dimensions throw away to make it into
    # a 3-D image.  Note that this is a HACK, and can be expected to produce
    # erroneous results on occasion.  NOTE also that this should NOT be done
    # for the map image, which can be written back out and should still have
    # the same number of dimensions it started with; the base image cannot
    # be written out, so this issue does not arise.  One-dimensional images
    # will not be handled correctly, since there is no reasonable way to
    # present one as a 2-D slice.
    #------------------------------------------------------------------------
    
    def readbase(self, filename) :

        # Allow user to save old map image first
        self.saveslice()
        if self.querysavemap() :
            return                  # If user presses cancel

        # Force the user to supply a header file rather than a raw image
        if re.search(r'\.img(\.gz|\.Z)?$', filename) :
            tkMessageBox.showerror("Raw Image Files Not Supported",
                                   "The current version of this editor " \
                                   "requires AnalyzeView headers rather " \
                                   "than raw image files.")
            return

        # Check that the requested base image file exists
        self.basefile = pyvox.truefile(filename)
        if not self.basefile :
            tkMessageBox.showerror("Invalid Base File",
                                   "The requested base image file does not " \
                                   "exist or is not readable.")
            return

        # Read the header file and its image file; convert to 3-D
        newbase = pyvox.read(filename)
        newbase = newbase.reshape()
        if len(newbase.size) == 2 :
            newbase.reshape( [1,] + newbase.size )
        if len(newbase.size) != 3 :
            tkMessageBox.showerror("Invalid Base File",
                                   "The requested base image file could not" \
                                   "be converted to 3-D.")
            return
        self.base = newbase
        
        # Remember dimensions and set the cursor position
        self.dimen = self.base.size
        self.coords = [self.dimen[0] / 2, self.dimen[1] / 2, self.dimen[2] / 2]

        # Set default and max gray level in the intensity windowing widgets
        # NOTE: The dictionary cnf is used because "from" is a reserved word
        # in Python and thus cannot be used as the name of a keyword arg.
        mingray = self.base.amin()
        maxgray = self.base.amax()
        cnf = {"from" : mingray, "to" : maxgray}
        self.minwind.configure(cnf)
        self.minwind.set(mingray)
        self.maxwind.configure(cnf)
        self.maxwind.set(maxgray)

        # Set the minimum, maximum, and initial zoom factors
        maxdim = max(self.dimen[0], self.dimen[1], self.dimen[2])
        self.minzoom = max( 128 / maxdim, 1)
        self.maxzoom = max(1024 / maxdim, self.minzoom)
        self.zoom    = max( 512 / maxdim, 1)
        
        return

    
    #------------------------------------------------------------------------
    # redraw - Redraw the image and status windows
    #
    # FIXME: Implement makeimage for plain image viewer
    #
    # FIXME: Don't redo steps unnecessarily
    #------------------------------------------------------------------------
    
    def redraw(self) :

        # Display a message and return if no base image defined
        if not self.base :
            text = 'No base image defined'
            self.canvas.configure(height=256, width=256)
            if self.imgitem :
                self.canvas.delete(self.imgitem)
            if not self.noimgitem :
                self.noimgitem = self.canvas.create_text(128, 128, text=text)
            return

        # Determine dimensions of the current image
        mag    = self.zoom
        dimen  = self.dimen
        height = mag * dimen[self.axes[1]]
        width  = mag * dimen[self.axes[2]]
        self.canvas.configure(height=height, width=width)

        # Determine coordinates of the current slice
        lcoords = copy.copy(self.coords)
        for i in range(len(self.coords)) :
            if lcoords[i] < 0 :
                lcoords[i] = 0
            elif lcoords[i] >= dimen[i] :
                lcoords[i] = dimen[i] - 1

        # Construct and display the current image
        bslice = self.makeimage(lcoords, mag)
        self.tkbase.putimage(bslice, None, None, mag, None)
        if self.noimgitem :
            self.canvas.delete(self.noimgitem)
        if not self.imgitem :
            self.imgitem = self.canvas.create_image(0, 0, anchor=NW,
                                                    image=self.tkbase)

        # Update the crosshairs
        if self.cross :
            sydimen = dimen[self.axes[1]]  # Length of screen y axis
            sxdimen = dimen[self.axes[2]]  # Length of screen x axis
            if self.direct[1] < 0 :
                sy = mag * (sydimen-1 - lcoords[self.axes[1]])
            else :
                sy = mag * lcoords[self.axes[1]]
            if self.direct[2] < 0 :
                sx = mag * (sxdimen-1 - lcoords[self.axes[2]])
            else :
                sx = mag * lcoords[self.axes[2]]
            if not self.horitem :
                self.horitem = self.canvas.create_line(0,0,1,1, fill='white')
            if not self.veritem :
                self.veritem = self.canvas.create_line(0,0,1,1, fill='white')
            self.canvas.coords(self.horitem, 0, sy, width-1, sy)
            self.canvas.coords(self.veritem, sx, 0, sx, height-1)
        else :
            if self.horitem :
                self.canvas.delete(self.horitem)
                self.horitem = None
            if self.veritem :
                self.canvas.delete(self.veritem)
                self.veritem = None

        # Update index coords in the status window
        self.czsvar.set("%4d" % lcoords[0])
        self.cysvar.set("%4d" % lcoords[1])
        self.cxsvar.set("%4d" % lcoords[2])

        # Update physical coords in the status window
        i2p = self.base.i2p()
        pcoords = i2p.point(pyvox.column(lcoords))
        self.pzsvar.set("%7.3f" % pcoords[0,0])
        self.pysvar.set("%7.3f" % pcoords[1,0])
        self.pxsvar.set("%7.3f" % pcoords[2,0])

        return

    
    #------------------------------------------------------------------------
    # Move crosshairs right
    #------------------------------------------------------------------------
    
    def right(self, event=None) :
        if not self.base :
            return
        iaxis = self.axes[2]
        step  = self.direct[2]
        limit = self.dimen[iaxis]
        self.coords[iaxis] = self.incr(self.coords[iaxis], step, 0, limit)
        self.redraw()


    #-----------------------------------------------------------------------
    # saveslice - Do any necessary cleanup before moving to a new slice
    #-----------------------------------------------------------------------

    def saveslice(self) :
        pass


    #------------------------------------------------------------------------
    # Set slice coordinates from mouse position at right click
    #------------------------------------------------------------------------
    
    def setcoords(self, event) :
        mag = self.zoom
        dimen = self.dimen
        sydimen = dimen[self.axes[1]]  # Length of screen y axis
        sxdimen = dimen[self.axes[2]]  # Length of screen x axis
        syaxis = self.axes[1]
        sxaxis = self.axes[2]
        if self.direct[1] < 0 :
            self.coords[syaxis] = dimen[syaxis]-1 - event.y / mag
        else :
            self.coords[syaxis] = event.y / mag
        if self.direct[2] < 0 :
            self.coords[sxaxis] = dimen[sxaxis]-1 - event.x / mag
        else :
            self.coords[sxaxis] = event.x / mag
        self.cross = 1
        self.redraw()


    #------------------------------------------------------------------------
    # Move crosshairs up
    #------------------------------------------------------------------------
    
    def up(self, event=None) :
        if not self.base :
            return
        iaxis = self.axes[1]
        step  = self.direct[1]
        limit = self.dimen[iaxis]
        self.coords[iaxis] = self.incr(self.coords[iaxis], -step, 0, limit)
        self.redraw()
        return


    #-------------------------------------------------------------------------
    # Set position of anterior commissure for reslicing
    #
    # FIXME: Show position on status window
    #-------------------------------------------------------------------------

    def setac(self, event=None) :

        if not self.base :
            return
        self.acpoint = pyvox.point(self.coords)
        if self.retrani :
            self.acpoint = self.retrani.point(self.acpoint)
        return


    #-------------------------------------------------------------------------
    # Set position of posterior commissure for reslicing
    #
    # FIXME: Show position on status window
    #-------------------------------------------------------------------------

    def setpc(self, event=None) :

        if not self.base :
            return
        self.pcpoint = pyvox.point(self.coords)
        if self.retrani :
            self.pcpoint = self.retrani.point(self.pcpoint)
        return


    #-------------------------------------------------------------------------
    # Set position of (any) mid-sagittal point for reslicing
    #
    # FIXME: Show position on status window
    #-------------------------------------------------------------------------

    def setms(self, event=None) :

        if not self.base :
            return
        self.mspoint = pyvox.point(self.coords)
        if self.retrani :
            self.mspoint = self.retrani.point(self.mspoint)
        return


    #-------------------------------------------------------------------------
    # Reslice image along AC-PC axis
    #
    # The midpoint of the AC-PC line is moved to the middle of the image, and
    # the new axes are defined as follows, where AP is the vector from the
    # AC point to the PC point, and AM is the vector from the AC point to
    # the midsagittal point.  The new y axis is parallel to AP and positive
    # posterior.  The new z axis is parallel to the component of AM
    # perpendicular to AP and is positive superior.  The new x axis is
    # parallel to the cross product of AP and AM and is positive left.
    #
    # FIXME: Support file axes other than SPL.
    #
    # FIXME: To prevent errors, disallow large rotations and any reflections.
    #-------------------------------------------------------------------------

    def reslice(self, event=None) :

        if not self.base :
            return

        # Save away the original base image
        if not self.rawbase :
            self.rawbase = self.base
            self.rawfile = self.basefile

        # DEBUG: Report the coordinates
        debug = 0
        if debug :
            print
            print "AC: ", self.acpoint
            print "PC: ", self.pcpoint
            print "MS: ", self.mspoint

        # Check that AC, PC, and MS points are defined
        msg = []
        if not self.acpoint :
            msg.append('Anterior Commissure')
        if not self.pcpoint :
            msg.append('Posterior Commissure')
        if not self.mspoint :
            msg.append('Mid-sagittal Point')
        if msg :
            msg = "The following reference points have not been defined: " + \
                  string.join(msg, ', ')
            tkMessageBox.showerror("Undefined Reference Points", msg)
            return

        # Check that file axes are SPL
        if self.fileaxes != 'SPL' :
            tkMessageBox.showerror("Not An Axial Image", \
                                   "Sorry, but the current software only " \
                                   "supports certain (SPL) axial images.")
            return

        # Compute difference vectors; correct for anisotropic sampling
        # FIXME: Modify affine.vector to preserve shape of vector
        ap = pyvox.vector(self.pcpoint) - pyvox.vector(self.acpoint)
        ap = self.rawbase.i2p().vector(ap).reshape()
        am = pyvox.vector(self.mspoint) - pyvox.vector(self.acpoint)
        am = self.rawbase.i2p().vector(am).reshape()

        # Compute the new x axis (+left); check for collinearity
        # FIXME: Implement cross product in pyvox and use here
        nv = [ ap[1]*am[2] - ap[2]*am[1],
               ap[2]*am[0] - ap[0]*am[2],
               ap[0]*am[1] - ap[1]*am[0] ]
        nv = pyvox.vector(nv)
        lnv = nv.norm2()
        if lnv < 2.0 :
            tkMessageBox.showerror("Refs Not Independent", \
                                   "Reference points are coincident or "
                                   "collinear.")
            return
        newx = nv / lnv
        if newx[2] < 0 :
            newx = -newx

        # Compute the new y axis (+posterior)
        newy = ap / ap.norm2()
        if newy[1] < 0 :
            newy = -newy

        # Compute the new z axis (+superior)
        newz = [ newx[1]*newy[2] - newx[2]*newy[1],
                 newx[2]*newy[0] - newx[0]*newy[2],
                 newx[0]*newy[1] - newx[1]*newy[0] ]
        newz = pyvox.vector(newz)
        if newz[0] < 0 :
            newz = -newz

        # DEBUG: Report new axes
        if debug :
            print "New X: ", newx.list()
            print "New Y: ", newy.list()
            print "New Z: ", newz.list()

        # Set parameters for the transform
        # FIXME: Image size and pixel dimensions should not be hardwired
        oldorg = (pyvox.point(self.acpoint) + pyvox.point(self.pcpoint)) / 2.0
        olddim = self.rawbase.spacing
        newsiz = [256, 256, 256]
        neworg = [128, 128, 128]
        newdim = [0.9375, 0.9375, 0.9375]
        rotate = pyvox.array([3,3], exim.double, 0)
        rotate[:,0] = newz
        rotate[:,1] = newy
        rotate[:,2] = newx

        # Construct transform to map new index coords into old index coords
        # FIXME: Compute physical transform for output to parm file
        retrani = pyvox.affine(3)
        retrani.translate(neworg, -2)    # Move new center to origin
        retrani.scale(newdim)            # Convert to physical coords
        retrani.compose2(rotate)         # Rotate to old axes
        retrani.scale(olddim, -2)        # Convert back to index coords
        retrani.translate(oldorg)        # Move origin to mid point
        self.retrani = retrani
        if debug :
            print "Reslicing transform (index coords)"
            print retrani

        # Resample the base image and make it the new base image
        # FIXME: Create setbase() method for this and readbase to share?
        # FIXME: Transform current position into the resliced image
        resliced = retrani.linear(self.rawbase, newsiz)
        resliced.spacing = newdim
        resliced.origin  = -0.9375 * pyvox.point(neworg)
        self.resliced    = resliced
        self.base        = self.resliced
        self.basefile    = None
        self.dimen       = self.base.size
        self.coords      = [self.dimen[0]/2, self.dimen[1]/2, self.dimen[2]/2]

        # Redraw the image
        self.redraw()
        
        return


    #-----------------------------------------------------------------------
    # Save Reslice As...
    #
    # FIXME: Provide a reasonable default filename
    #-----------------------------------------------------------------------

    def save_resliced_as(self, event=None) :

        # Verify that there is a resliced image to save
        if not hasattr(self, "resliced") or not self.resliced :
            errmsg = "You must reslice the image before you can " \
                     "save the resliced image"
            tkMessageBox.showerror("No Resliced Image To Save", errmsg)
            return

        # Query user for filename; prohibit raw image files
        while 1 : 
            dialog = Pmw.PromptDialog(self.status,
                                      title="Save Resliced As...",
                                      label_text="New Resliced File Name",
                                      entryfield_labelpos=Tkinter.N,
                                      defaultbutton=0,
                                      buttons=("OK", "Cancel"))
            result = dialog.activate()
            value  = dialog.get()
            if result == "Cancel" :
                return
            elif result == "OK" :
                reslfile = dialog.get()
                if not re.search(r'\.hdr(\.gz|\.Z)?$', reslfile) :
                    errmsg = "Raw image files are no longer supported. " \
                             "Use a .hdr extension to get AnalyzeView " \
                             "images and headers."
                    tkMessageBox.showerror("Raw Images Not Supported", errmsg)
                    continue
                                          
                self.reslfile = reslfile
                self.saveresl()
                break
            else :
                errmsg = "Invalid result %s in Save Resliced As..." % result
                raise Exception, errmsg

        self.redraw()
        return


    #------------------------------------------------------------------------
    # Save resliced image and parameters to files
    #
    # FIXME: Provide some array and affine methods (repr, str, list?) to
    # generate the desired formats below, or something equally useful.
    #
    # FIXME: Support as an independent command?
    #
    # FIXME: Add the physical transform
    #------------------------------------------------------------------------
    
    def saveresl(self, event=None) :
        if not self.resliced :
            return

        # Save the resliced image
        self.resliced.write(self.reslfile)

        # Generate text representations of landmark points
        actext = "AC point: "
        pctext = "PC point: "
        mstext = "MS point: "
        for i in range(3) :
            actext += "  %10.3f" % self.acpoint[i,0]
            pctext += "  %10.3f" % self.pcpoint[i,0]
            mstext += "  %10.3f" % self.mspoint[i,0]

        # Generate text representation of the index transform
        # FIXME: Modify affine class to return the appropriate element
        # when called with subscripts; or perhaps have affine.coefs
        # return the n *(n+1) array of coefficients.
        intext = "Index transform:: \n"
        matrix = self.retrani.matrix
        offset = self.retrani.offset.reshape()
        for i in range(3) :
            for j in range(3) :
                intext += "  %10.6f" % matrix[i,j]
            intext += "  %10.6f" % offset[i]
            if i < 2 :
                intext += '\n'
        alltext = string.join([actext, pctext, mstext, intext, "::\n"], '\n')

        # Save the reslicing parameter file
        rparfile = re.sub(r'\.hdr(\.Z|\.gz)?$', '.par', self.reslfile)
        rparout  = open(rparfile, 'w')
        rparout.write(alltext)
        rparout.close()
        
        self.redraw()
        return



#----------------------------------------------------------------------------
# reged class - Region editor windows and widgets
#----------------------------------------------------------------------------

class reged (viewer) :

    #------------------------------------------------------------------------
    # Instance variables to be directly configured by the user:
    #
    #  maptype    Format used in the map image file
    #             'index'  Pixel contains index to a region; always uint1
    #             'mask'   Pixel contains value of image, or zero; 'mask1'
    #                      and 'mask2' are deprecated aliases
    #  regions    List of (name, number, (r,g,b)) tuples defining regions
    #  fileaxes   Same as in viewer class
    #
    #
    # Other instance variables:
    #
    #  mapfile    Filename from which map image was read
    #  mapalt     Has map been altered since last written to disk?
    #  map        Map image, as Pyvox image; None if no map image defined.
    #             Always uchar.  Contains index for 'index' map, and 0/255
    #             for a 'mask' map.  Translated at input/output.
    #  mslice     Currently selected slice and region of the map image
    #  region     Current region selection, 'All',  or None
    #  maskon     Display mask?
    #  painting   Is button-1 down (and painting active)?
    #  brushxy    List of relative coordinates in brush footprint
    #  brushval   Value that brush deposits on the mask
    #  regsvar    StringVar containing current region selection, or blank
    #  redlut     Red lookup table for color shading of regions
    #  grnlut     Green lookup table for color shading of regions
    #  blulut     Blue lookup table for color shading of regions
    #------------------------------------------------------------------------


    #------------------------------------------------------------------------
    # Initialize
    #
    # Order of arguments is not guaranteed.
    #------------------------------------------------------------------------
    
    def __init__(self, parent=None, maptype='mask', fileaxes='SPL',
                 regions=None, baseformat = 'Unsigned Char') :

        # Initialize base class
        viewer.__init__(self, parent=parent, fileaxes=fileaxes,
                        baseformat=baseformat)

        # Remember configuration parameters; support old mask1 & mask2
        if maptype in ('mask1', 'mask2') :
            maptype = 'mask'
        self.maptype  = maptype
        self.regions  = regions

        # Set initial status information
        self.mapalt   = 0
        self.region   = None
        self.maskon   = 0

        # Configure the image viewing window
        self.viewwin.bind('<ButtonRelease-1>', self.paintoff)
        self.viewwin.bind('<ButtonPress-1>',   self.painton)
        self.viewwin.bind('<Motion>',          self.paint)

        # Set current painting parameters
        self.painting = 0
        self.brushxy  = ((0,0),)
        self.brushval = 255
        
        # Set up the mask images
        self.map     = None
        self.mapfile = None
        self.mslice  = None
        self.setmap(None, None)

        # Add items to the file menu
        # FIXME: The index values below depend on the number of items
        # put into the menu by the viewer class and keeping them coordinated
        # is a real pain; figure out a better way.  Also it would be nice
        # to keep the Quit item always at the bottom.
        self.filemenu.add_command(label="New Map",  command=self.newmap)
        self.filemenu.add_command(label='Save Map', command=self.savemap)
        self.savemapindex = 3
        self.filemenu.entryconfig(self.savemapindex, state='disabled')
        self.filemenu.add_command(label="Save Map As...",
                                  command=self.savemapas)
        self.savemapasindex = 4
        self.filemenu.entryconfig(self.savemapasindex, state='disabled')

        # Add items to the edit menu
        self.editmenu.add_command(label='Remove All', command=self.paintall)
        self.removeallindex = 1
        self.editmenu.entryconfig(self.removeallindex, state='disabled')

        # Add items to the compute menu
        self.compmenu.add_command(label='Volumes', command=self.compvolumes)
        self.compvolindex = 1
        self.compmenu.entryconfig(self.compvolindex, state='disabled')
        
        # Construct list of regions, rgb luts, and a dictionary by name
        if not self.regions :
            raise Exception, "No regions have been defined"
        self.regionoptions = ['None', 'All']
        self.regionbyname = {}
        self.redlut = pyvox.array(256, exim.ushort, 255)
        self.grnlut = pyvox.array(256, exim.ushort, 255)
        self.blulut = pyvox.array(256, exim.ushort, 255)
        for region in self.regions :
            regname = region[0]
            regnum  = region[1]
            self.regionbyname[regname] = regnum
            self.regionoptions.append(regname)
            if len(region) >= 3 :
                red, grn, blu = region[2]
                self.redlut[regnum] = red
                self.grnlut[regnum] = grn
                self.blulut[regnum] = blu

        # Create Option Menu to select current region for painting;
        # disabled if no current map
        regopts = self.regionoptions
        if not self.map :
            regopts = ['None']
        entry = Pmw.OptionMenu(self.status, labelpos = W,
                               label_text = 'Region:',
                               items = regopts,
                               menubutton_width = 20,
                               command = self.newregion)
        entry.pack(anchor=W, padx=5, pady=5)
        self.regionmenu = entry
        
        # Create Option Menu for brush mode (paint/erase)
        self.modeoptions = ('Add', 'Remove')
        self.mode = StringVar()
        self.mode.set(self.modeoptions[0])
        entry = Pmw.OptionMenu(self.status, labelpos = W,
                               label_text = 'Mode:',
                               items = self.modeoptions,
                               menubutton_width = 20,
                               command = self.newmode)
        entry.pack(anchor=W, padx=5, pady=5)

        # Create Option Menu for brush size
        self.sizeoptions = ('1 x 1', '3 x 3', '5 x 5', '7 x 7', '9 x 9',
                            '11 x 11', '15 x 15', '19 x 19', '25 x 25' )
        entry = Pmw.OptionMenu(self.status, labelpos = W,
                               label_text = 'Size:',
                               items = self.sizeoptions,
                               menubutton_width = 20,
                               command = self.newsize)
        entry.pack(anchor=W, padx=5, pady=5)


    #--------------------------------------------------------------------
    # compvolumes - Compute region volumes
    #--------------------------------------------------------------------

    def compvolumes(self, event=None) :

        self.saveslice()

        # Check for prerequisites
        # FIXME: Provide better error reporting
        if not self.map or not self.regions :
            return

        # Compute volume of a pixel
        # FIXME: Assumes spacing is in mm; actually get units from header
        # FIXME: Modify pyvox to support None as the default spacing;
        # then omit pixel volume and units here.
        pixsize = self.base.spacing
        pixvol  = pixsize[0] * pixsize[1] * pixsize[2] / 1000.0

        # Determine where user wants the report saved
        # FIXME: Provide a reasonable default filename
        dialog = Pmw.PromptDialog(self.status,
                                  title="Compute Volumes...",
                                  label_text="Filename for volumes report",
                                  entryfield_labelpos=Tkinter.N,
                                  defaultbutton=0,
                                  buttons=("OK", "Cancel"))
        result = dialog.activate()
        value  = dialog.get()
        if result == "Cancel" :
            return
        elif result == "OK" :
            volfile = dialog.get()
        else :
            raise Exception, "Invalid result %s in savemapas" % result

        # Open report file and write header lines
        volout = open(volfile, 'w')
        line = "\nRegional Volumes:\n"
        volout.write(line)
        if self.basefile :
            line = "Base image = %s\n" % self.basefile
            volout.write(line)
        if self.mapfile :
            line = "Map image  = %s\n" % self.mapfile
            volout.write(line)
        if pixsize :
            line = "Pixel size = %.3f x %.3f x %.3f [mm3]\n" % tuple(pixsize)
            volout.write(line)
            line = "Pixel volume = %11.6f [cm3]\n" % pixvol
            volout.write(line)        
        line = "\nRegion                          Pixels    Vol (cm3)" \
               "\n---------------------------     ------    ---------\n"
        volout.write(line)
        
        # Compute and display region volumes
        histo  = self.map.histo()
        for name, num, rgb in self.regions :
            regvol = histo[num] * pixvol
            line = "%-30s  %6d     %8.3f\n" % (name, histo[num], regvol)
            volout.write(line)

        # Close output file and return
        volout.close()
        return        
        

    #---------------------------------------------------------------------
    # makeimage - Construct the image to be displayed in the image window
    #
    # FIXME: Do smoother handling of base/map size mismatch
    # FIXME: Size mismatch is currently ignored
    #---------------------------------------------------------------------

    def makeimage(self, lcoords, mag) :
        
        # Create the base image
        # FIXME: Try gamma transformation instead of linear
        gray = self.getslice(self.base, lcoords, self.axes, self.direct)
        if self.invert :
            gray = gray.uchar(self.maxwin, self.minwin)
        else :
            gray = gray.uchar(self.minwin, self.maxwin)

        # Overlay all regions on the current image
        if self.map and self.region == 'All' and self.maskon :
            #gray = ((gray.float()/255.0).pow(0.5)*255.0).ushort()
            gray = gray.ushort()
            regm = self.getslice(self.map, lcoords, self.axes, self.direct)
            rgb  = pyvox.array(gray.size+[3], exim.uchar, 0)
            rgb[...,0] = (regm.lookup(self.redlut) * gray).uchar(0, 65535)
            rgb[...,1] = (regm.lookup(self.grnlut) * gray).uchar(0, 65535)
            rgb[...,2] = (regm.lookup(self.blulut) * gray).uchar(0, 65535)

        # Overlay current region on the base image
        elif self.map and self.region and self.maskon :
            # and  self.map.size == self.base.size :
            lut = pyvox.ramp([256,3], exim.uchar, 0)
            rgb = gray.lookup(lut)            
            # print "Getting ready for mask slice"    # DEBUG
            if not self.mslice :
                # print "Getting mask slice"          # DEBUG
                # print lcoords, self.map.size        # DEBUG
                self.mslice = self.getslice(self.map, lcoords,
                                            self.axes, self.direct)
                self.mslice = self.mslice.compare(self.region, 0, 255, 0)
            # print "Overlaying mask on base image"   # DEBUG
            rgb[...,1] = rgb[...,1].reshape() & self.mslice

        # Show base image without any region overlay
        else :
            lut = pyvox.ramp([256,3], exim.uchar, 0)
            rgb = gray.lookup(lut)            

        return rgb


    #----------------------------------------------------------------------
    # New map - Create a new, blank map image
    #----------------------------------------------------------------------
    
    def newmap(self, event=None) :

        # Offer user chance to save old modified map first
        self.saveslice()
        if self.querysavemap() :
            return                  # Cancelled by user

        # A base image is required to create a new map
        if not self.base :
            tkMessageBox.showerror("No Base Image",
                                   "A base image is required to create a " \
                                   "new map image.")
            return
        
        # Create the new map
        map = pyvox.array(self.base.size, exim.uchar, 0)
        map.metacopy(self.base, physdata=1)
        self.setmap(map, None)
        self.redraw()

        return


    #-----------------------------------------------------------------------
    # Select new brush mode (paint/erase)
    #-----------------------------------------------------------------------
    
    def newmode(self, mode) :
        if mode == self.modeoptions[0] :
            self.brushval = 255
        else :
            self.brushval = 0


    #-----------------------------------------------------------------------
    # New region selection entered in its Option Menu
    #-----------------------------------------------------------------------
    
    def newregion(self, region) :
        self.saveslice()
        if region == 'None' :
            self.region = None
        elif region == 'All' :
            self.region = 'All'
        elif self.maptype == 'index' :
            self.region = self.regionbyname[region]
        elif self.maptype == 'mask' :
            self.region = 255
        else :
            raise Exception, "Undefined map and region types"
        if not self.base :
            return
        if self.map and self.region :
            self.maskon = 1
        self.redraw()


    #------------------------------------------------------------------------
    # Select new brush size from option menu
    #------------------------------------------------------------------------
    
    def newsize(self, size) :

        mat = re.search('[0-9]+', size)
        if not mat :
            raise Exception, "Invalid brush size %s" % size
        diam = string.atoi(mat.group())
        irad = (diam+1) / 2
        rad2 = (diam / 2.0) ** 2
        brush = []
        for ix in range(-irad, irad+1) :
            for iy in range(-irad, irad+1) :
                if ix*ix + iy*iy < rad2 :
                    brush = brush + [[iy, ix]]
        if brush :
            self.brushxy = brush
        else :
            self.brushxy = ((0,0),)


    #------------------------------------------------------------------------
    # Next possible mask state; toggle on/off
    #------------------------------------------------------------------------
    
    def nextmask(self, event=None) :
        if not self.base or not self.map or not self.region :
            return
        self.maskon = not self.maskon
        self.redraw()


    #------------------------------------------------------------------------
    # Motion: If enabled, clear or set mask under pointer
    # FIXME: Write array method paint() that takes a kernel
    #------------------------------------------------------------------------
    
    def paint(self, event) :
        if self.painting :
            self.paintxy(event.x, event.y)


    #-----------------------------------------------------------------------
    # Remove All : Make entire slice NOT part of the current region
    #-----------------------------------------------------------------------
    
    def paintall(self, event=None) :
        if self.mslice :
            self.mslice[...] = 0
            self.mapalt = 1
            self.redraw()


    #------------------------------------------------------------------------
    # ButtonRelease-1 : Turn painting off
    #------------------------------------------------------------------------
    
    def paintoff(self, event) :
        self.painting = 0


    #------------------------------------------------------------------------
    # ButtonPress-1: Turn painting on
    #------------------------------------------------------------------------
    
    def painton(self, event) :
        if self.map and self.maskon and self.region and self.region != 'All' :
            self.painting = 1
            self.paintxy(event.x, event.y)
        else :
            self.painting = 0


    #------------------------------------------------------------------------
    # Paint at a given location
    #------------------------------------------------------------------------
    
    def paintxy(self, x, y) :
        self.mapalt = 1
        mag = self.zoom
        y   = y / mag
        x   = x / mag
        # print "Painting at", x, y    # DEBUG
        sy, sx = self.mslice.size
        for dx, dy in self.brushxy :
            y1 = y + dy
            x1 = x + dx
            if y1 >= 0 and y1 < sy and x1 >= 0 and x1 < sx :
                self.mslice[y1, x1] = self.brushval
        self.redraw()


    #------------------------------------------------------------------------
    # Offer to save a modified map
    #
    # FIXME: This does not work properly if called after a new map has been
    # created but not yet named by savemapas.
    #------------------------------------------------------------------------
    
    def querysavemap(self) :
        if not self.map or not self.mapalt :
            return
        message = "The map has been modified but not saved.  Save changes?"
        query = Pmw.MessageDialog(self.status, title='Save map file?',
                                  defaultbutton = 0,
                                  buttons=('Save', 'Discard', 'Cancel'),
                                  message_text=message)
        query.iconname('Save map file?')
        result = query.activate()
        if result == 'Save' :
            self.savemap()
            return 0
        elif result == 'Discard' :
            return 0
        elif result == 'Cancel' :
            return 1
        else :
            raise Exception, "Invalid result %s from save file query" % result


    #------------------------------------------------------------------------
    # Read a map image; does NOT redraw the image
    #------------------------------------------------------------------------
    
    def readmap(self, filename) :

        # Allow user to save the current map first
        self.saveslice()
        if self.querysavemap() :
            return                  # Cancelled by user

        # Verify the map type
        if self.maptype not in ('index', 'mask') :
            raise Exception, "Map type %d not recognized" % self.maptype

        # Require a header file rather than a raw image
        if re.search(r'\.img(\.gz|\.Z)?$', filename) :
            tkMessageBox.showerror("No More Raw Image Files",
                                   "The current version of this editor " \
                                   "requires AnalyzeView headers rather " \
                                   "than raw image files.")
            return

        # Determine true name of the map file and verify its existence
        mapfile = pyvox.truefile(filename)
        if not mapfile :
            tkMessageBox.showerror("Invalid Map File",
                                   "The requested map image file does not " \
                                   "exist or is not readable.")
            return

        # Read the image file, including any header file
        map = pyvox.read(filename)

        # Verify that base and map images have same dimensions
        # FIXME: Allow trival 4D image to match a 3D image
        if not self.base.size == map.size :
            tkMessageBox.showerror("Mismatched Base and Map Images",
                                   "The dimensions of the requested map " \
                                   "image do not match the base image.")
            return

        # If the map type is 'index', the map file must be unsigned char
        if self.maptype == 'index' and map.type != exim.uchar :
            tkMessageBox.showerror("Bad Map Type",
                                   "The map image file must be type " \
                                   "unsigned char.")
            return

        # If the map type is 'mask', the map and base must be the same type
        if self.maptype == 'mask' and map.type != self.base.type :
            tkMessageBox.showerror("Bad Map Type",
                                   "The base and map image files must have " \
                                   "the same data type.")
            return
        
        # Convert a 'mask' map into an index map
        if self.maptype == 'mask' :
            map = map.compare(0, 255, 0, 255)

        # Install the new map image
        self.setmap(map, mapfile)

        return


    #------------------------------------------------------------------------
    # Save map to file
    #------------------------------------------------------------------------
    
    def savemap(self, event=None) :
        if not self.map :
            return
        self.saveslice()
        if self.maptype == 'index' :
            self.map.write(self.mapfile)
        elif self.maptype == 'mask' :
            # FIXME: Following code assumes two's complement signed integral
            # types, so that  -1 & foo == foo for any foo.
            lut = pyvox.array(256, self.base.type, -1)
            lut[0] = 0
            newmap = self.map.lookup(lut) & self.base
            newmap.write(self.mapfile)
        else :
            raise Exception, "Unknown map type %s" % self.maptype
        self.mapalt = 0

        self.redraw()
        return


    #------------------------------------------------------------------------
    # Save map as new file
    #
    # FIXME: Provide a reasonable default filename
    #------------------------------------------------------------------------
    
    def savemapas(self, event=None) :
        
        if not self.map :
            return
        self.saveslice()          # Redundant but harmless

        # Query user for filename; prohibit raw image files
        while 1 : 
            dialog = Pmw.PromptDialog(self.status,
                                      title="Save Map As...",
                                      label_text="New Map File Name",
                                      entryfield_labelpos=Tkinter.N,
                                      defaultbutton=0,
                                      buttons=("OK", "Cancel"))
            result = dialog.activate()
            value  = dialog.get()
            if result == "Cancel" :
                return
            elif result == "OK" :
                mapfile = dialog.get()
                if re.search(r'\.img(\.gz|\.Z)?$', mapfile) :
                    errmsg = "Raw image files are no longer supported. " \
                             "Use a .hdr extension to get AnalyzeView " \
                             "images and headers."
                    tkMessageBox.showerror("Raw Images Not Supported", errmsg)
                    continue
                                          
                self.setmap(self.map, mapfile)
                self.savemap()
                break
            else :
                raise Exception, "Invalid result %s in savemapas" % result

        self.redraw()
        return

    
    #------------------------------------------------------------------------
    # Save the updated map slice, if any
    #
    # FIXME: Should really use the same coords as did redraw!
    #------------------------------------------------------------------------
    
    def saveslice (self) :
        if not self.mslice :
            return
        bslice = self.getslice(self.map, self.coords, self.axes, self.direct)
        omask  = bslice.compare(self.region, 0, 255, 0)
        diff   = self.mslice.compare(omask, 255, 0, 255)
        bslice.cmove(diff, self.region & self.mslice)
        self.putslice(self.map, bslice, self.coords, self.axes, self.direct)
        self.mslice = None

        return


    #----------------------------------------------------------------------
    # Set map and mapfile
    #
    # Setting self.map and self.mapfile should always be done through this
    # method to ensure than the appropriate state is modified.
    #
    # FIXME: Save the old mslice and map here?
    #----------------------------------------------------------------------
    
    def setmap(self, map, mapfile) :

        # No map means no mapfile
        if not map :
            mapfile = None

        # Change configuration when map state changes
        if map and not self.map :
            self.filemenu.entryconfig(self.savemapasindex, state='active')
            self.editmenu.entryconfig(self.removeallindex, state='active')
            self.compmenu.entryconfig(self.compvolindex,   state='active')
        elif not map and self.map :
            self.filemenu.entryconfig(self.savemapasindex, state='disabled')
            self.editmenu.entryconfig(self.removeallindex, state='disabled')
            self.compmenu.entryconfig(self.compvolindex,   state='disabled')
            self.mslice = None
            self.mapalt = 0
        self.map = map

        # Enable or disable selection of a current region
        if self.map :
            self.regionmenu.setitems(self.regionoptions, 0)

        # Change configuration when mapfile state changes
        if mapfile and not self.mapfile :
            self.filemenu.entryconfig(self.savemapindex,   state='active')
        elif not mapfile and self.mapfile :
            self.filemenu.entryconfig(self.savemapindex,   state='disabled')
        self.mapfile = mapfile
            
        return  
