#############################################################################
# regis.py - Pyvox library for image registration
#
# Copyright (c) 2000-2005 Trustees of the University of Pennsylvania
# See the file COPYING for license terms.
#
# This Python module includes various tools for image registration.
#############################################################################

rcsid = "$Id: regis.py,v 1.39 2005/08/07 19:44:35 hughett Exp $ $Name: pyvox-0-72 $"

import sys, os, getopt
import stat
from math import sqrt, log
from exceptions import Exception

import pyvox, exim
import optim


class RegistrationError (Exception) :
    def __init__(self, args=None) :
        self.args = args



#--------------------------------------------------------------------------
# correl - Weighted correlation of two arrays
#
# This function computes the weighted Pearson product moment correlation of
# the given arrays.
#
# FIXME: Move to a stat module, or redefine as a parray method.
# FIXME: Is this done correctly for an unbiased estimator?
# FIXME: Add functions for mean, var, covar.
# FIXME: Generalize to operate only over specified axes.
#---------------------------------------------------------------------------

def correl(x, y, w=None) :
    if w :
        sumw   = w.asum()         # Sum of the weights
        sumwx  = x.dot(w)         # Weighted sum of first array x
        sumwy  = y.dot(w)         # Weighted sum of second array y
        sumwxx = x.dot(x, w)      # Weighted sum of x*x
        sumwyy = y.dot(y, w)      # Weighted sum of y*y
        sumwxy = x.dot(y, w)      # Weighted sum of x*y
    else :
        sumw   = x.count()
        sumwx  = x.asum()
        sumwy  = y.asum()
        sumwxx = x.dot(x)
        sumwyy = y.dot(y)
        sumwxy = x.dot(y)
    varx   = sumwxx - sumwx * sumwx / sumw
    vary   = sumwyy - sumwy * sumwy / sumw
    result = sumwxy - sumwx * sumwy / sumw
    result = result / sqrt(varx*vary)
    return result



#---------------------------------------------------------------------------
# info - Information content of an unsigned char image
#
# This function computes the information content, in bits per pixel, of an 
# unsigned char image.  The image may be of any rank and shape.  Zero voxels
# are assumed to be background and are ignored.
#
# ALGORITHM:  The various games below limiting probabilities to some value
# strictly greater than zero are intended to avoid either dividing by zero
# or taking the log of zero; they should have no effect on the result
# otherwise.
#
# FIXME: This would go more sensibly into a stat module, if we had a stat
# module.
#
# FIXME: This gets a divide by zero error if all the voxels are zero.
#---------------------------------------------------------------------------

def info(img, mask=None) :

    # Compute the histogram and discard zero intensities
    hist = img.histo(mask)
    hist[0] = 0

    # Convert to probabilities
    pmin = 1.0 / hist.asum()
    hist = hist.double() * pmin

    # Compute information
    bits = hist * hist.max(pmin).log()
    bits = bits.asum() / log(2.0)
    return -bits



#----------------------------------------------------------------------------
# mutinfo - Mutual information of two unsigned char images
#
# This function computes the mutual information of two unsigned char images,
# in bits per pixel.  The two images may be of any rank and shape but must
# be the same rank and shape.  An optional mask of per-voxel weights may
# also be provided, and defaults to 1. 
#
# FIXME: Convert to the standard form array.mutinfo?
#
# FIXME: This would go more sensibly into a stat module, if we had a stat
# module.
#
# FIXME: It would be useful to also accept the arguments bi, mar1, mar2 for
# applications in which the bivariate histogram has already been computed.
# This can be distinguished from img1, img2, mask by the fact that the
# arguments in the latter case must all have the same shape.
#
# ALGORITHM:  The various games below limiting probabilities to some value
# strictly greater than zero are intended to avoid either dividing by zero
# or taking the log of zero; they should have no effect on the result
# otherwise.
#
# FIXME: Should zero voxels be excluded, as in the info() function?  If so,
# the case that all voxels are zero must be handled without crashing.
#----------------------------------------------------------------------------

def mutinfo(img1, img2, mask=None) :

    # Get the bivariate histogram
    (hist, mar1, mar2) = img1.bihist(img2, mask)
    mar1.reshape([256,1])
    mar2.reshape([1,256])

    # Convert to probabilities
    mar1 = mar1.double() / mar1.asum()
    mar2 = mar2.double() / mar2.asum()
    pmin = 1.0 / hist.asum()
    hist = hist.double() * pmin

    # Compute mutual information
    bits = hist.max(pmin) / mar1.mmul(mar2).max(pmin)
    bits = hist * bits.log()
    return bits.asum() / log(2.0)



#---------------------------------------------------------------------------
# obregis - Objective function for image registration
#
# FIXME: It would be useful to have an objective function that could be
# dynamically upgraded from rigid to affine to spline transforms, for
# registration protocols that use progressively more flexible transforms.
# As specialist obfunctions proliferate (obrigid, obaffine, obpoly,...),
# this begins to look more and more like a good ideas, even if it will
# take a substantial amount of work.
#
# FIXME: Make sure that this works for transforms more general than affine.
# On first glance, it ought to work for any transform which is applied
# uniformly over the whole image, such as polynomial or perspective; but
# it won't handle transforms such as splines which want to optimize
# locally.
#
# FIXME: Make it possible to plug in different metrics rather than having
# them hardwired into this code.  One possible approach is to define a
# generic regmetric class with methods scale() and eval(); a init() class
# might be necessary for efficiency.
#
# FIXME: It might be prudent to use a method setmask() to set the mask,
# in case some initialization needs to be done.
#
# FIXME: Most current image registrators are limited to unsigned char
# images only, or even more specifically to 3D unsigned char images.
#
# FIXME: Allow setting the size of the resampled test image.
#
# FIXME: Provide for an intensity mapping as well.
#
# FIXME: Provide a means for controlling how much progress information is
# printed out, and to what device.  Perhaps an iteration hook which calls
# a user-specified function?
#
# FIXME: Implement additional transforms: similarity (rigid plus uniform
# scale); restricted affine (rigid plus scale along each principal axis).
#---------------------------------------------------------------------------

class obregis (optim.obfunction) :


    #----------------------------------------------------------------------
    # Initialize an instance, including optimization parameters
    # Assumes that the derived class has already defined self.npar and
    # that zero is a reasonable initial guess for all parameters.
    #----------------------------------------------------------------------
    
    def __init__(self, source, target) :
        optim.obfunction.__init__(self, self.npar)

        # Initialize attributes intended to be set by the user
        self.source = source       # Source image to register
        self.target = target       # Target to register to
        self.scenter = None        # Physical "center" of source image
        self.tcenter = None        # Physical "center" of target image
        self.mask   = None         # Image mask/weights in target coords
        self.metric = "mutinfo"    # Metric to use for similarity
        self.interp = "linear"     # Interpolation type to use
        self.niter  = 10           # Number of iterations

        # Initialize attributes intended to be read by the user
        self.xopt   = self.npar * [0,]   # Init/final optimum found
        self.fopt   = None               # Value at optimum

        # Initialize other attributes
        self.funct  = None
        self.src2   = source     # Shrunk images for multi-scale
        self.tgt2   = target
        self.mask2  = self.mask  # Shrunk mask for multi-scale
        self.tmean  = None       # Norms for correl, if needed
        self.tnorm  = None
        self.shrink = 0
        self.smooth = 0

        return

    
    #----------------------------------------------------------------------
    # Evaluate discrepancy for given params
    #---------------------------------------------------------------------
    
    def __call__(self, param) :
        size = self.tgt2.size
	if self.mask2 :
            npix = self.mask2.asum()
        else :
            npix = self.tgt2.count()
        aff  = self.stgt2src(param, self.shrink)
        if self.interp == "linear" :
            test = aff.linear(self.src2, size)
        elif self.interp == "nearest" :
            test = aff.nearest(self.src2, size)
	else :
	    raise Error, "Interpolation %s not recognized" % self.interp
        test.metacopy(self.tgt2, physdata=1)
        # FIXME: The following is a hack to support intensity scaling for
        # obaffinegb; either do it right or remove it entirely
        if (0) :    
            lower = param[1]
            upper = param[0] * 255.0 + param[1]
            test = test.uchar(lower, upper)
        if self.metric == "mutinfo" :
            # Mutual information metric, in bits per pixel and
            # inverted so minimization optimizes
            fnew = -mutinfo(test, self.tgt2, self.mask2)
        elif self.metric == "correl" :
            # Correlation between two images, scaled to 0,1 range and
            # inverted so that minimization optimizes.
            # FIXME: Arrange to precompute the target mean and variance
            # Compute averages and norms to scale the correlation
            #if not self.tnorm :
            #    self.tnorm = self.tgt2.norm2(0, self.mask2)
            #    self.tmean = self.tgt2.dot(self.mask2) / npix
            #ntst = test.norm2(0, self.mask2)
            #if self.mask2 :
            #    mtst = test.dot(self.mask2) / npix
            #else :
            #    mtst = test.asum() / npix
            # Compute the Pearson product moment correlation
            #fnew = test.dot(self.tgt2, self.mask2) - mtst*self.tmean
            #fnew = -fnew / (ntst * self.tnorm)
            fnew = -correl(test, self.tgt2, self.mask2)
        elif self.metric == "norm2" :
            # L2 metric (akin to ANOVA), expressed per pixel
            fnew = test.norm2(self.tgt2, self.mask2)
            fnew = fnew / sqrt(npix)
        else :
            raise RegistrationError, "Metric %s not recognized" % self.metric
        line = "Error at "
        for i in range(len(param)) :
            line = line + "%8.4f" % param[i]
        line = line + " is %7.4f" % fnew
        if not self.fopt or fnew < self.fopt :
            self.xopt = param
            self.fopt = fnew
            line = line +  " *\n"
        else :
            line = line + "\n"
        if self.logout :
            self.logout.write(line)

        return fnew


    #----------------------------------------------------------------------
    # Set scale and smoothing levels for test image
    # FIXME: Preserve old shrunk images if possible
    # FIXME: For mutinfo metric, do quantization rather than smoothing;
    #    find cleaner algorithm for shrinking (e.g. image.shrink()!)
    # FIXME: Choose a better control parameter than metric == 'mutinfo'
    #    to decide whether to use linear or NN interpolation in shrink
    #----------------------------------------------------------------------
    
    def scale(self, shrink, smooth = 0) :
        self.tmean = None     # Forget summary value for old shrink
        self.tnorm = None
        self.src2  = self.source
        self.tgt2  = self.target
        self.mask2 = self.mask
        if self.metric == 'mutinfo' :
            sf = 2.0**shrink
            aff = pyvox.affine(self.ndim)
            aff.scale(sf)
            ssz = self.source.size
            tsz = self.target.size
            for i in range(self.ndim) :
                ssz[i] = ssz[i] / sf
                tsz[i] = tsz[i] / sf
            self.src2 = aff.nearest(self.source, ssz)
            self.tgt2 = aff.nearest(self.target, tsz)
            if self.mask :
                self.mask2 = aff.nearest(self.mask, tsz)
        else :
            self.src2 = self.source
            self.tgt2 = self.target
            for i in range(shrink) :
                self.src2 = self.src2.lowpass(2)
                self.tgt2 = self.tgt2.lowpass(2)
                if self.mask2 :
                    self.mask2 = self.mask2.lowpass(2)
            for i in range(smooth) :
                self.src2 = self.src2.lowpass()
                self.tgt2 = self.tgt2.lowpass()
                # Assume mask does not need to be smoothed
        self.shrink = shrink
        self.smooth = smooth
        self.fopt = None

        return



#---------------------------------------------------------------------------
# obaffine - Objective function for affine registration
#
# Parameters consist of N*N matrix elements, and N translations, where N
# is the number of image dimensions.  A gain or bias difference between
# the two images should not matter for either correl or mutinfo metric.
#
# This class is specialized below to define rigid and translation-only
# transforms.  The methods that must ordinarily be redefined are __init__
# and ctgt2src.
#
# FIXME: Currently supports unsigned char images only.
#
# FIXME: Need a way to scale an uchar image by a float value.  Array.uchar
# with limits works but seems sort of inelegant.
#
# FIXME: Provide a method to print out the parameters with labels for
# human consumption?  Make the stepwise printouts optional?
#--------------------------------------------------------------------------

class obaffine (obregis) :


    #----------------------------------------------------------------------
    # Initialize instance, including optimization parameters
    #---------------------------------------------------------------------
    
    def __init__(self, source, target, init=None) :
        if source.rank != target.rank :
            raise RegistrationError, \
                  "source and target have different number of dimensions"
        ndim = source.rank
        self.ndim = ndim
        self.npar = ndim*(ndim+1)
        obregis.__init__(self, source, target)
        if init == None or isinstance(init, pyvox.AffineType) :
            self.init = init
        elif isinstance(init, obrigid) or isinstance(init, obaffine) :
            self.init = init.ctgt2src()
            self.metric  = init.metric
            self.scenter = init.scenter
            self.tcenter = init.tcenter
        else :
            raise RegistrationError, \
                  "Cannot translate init argument to an initial transform"
        for i in range(ndim):
            self.xopt[i*(ndim+1)] = 1.0
        self.xtol = ndim*ndim * [0.001,] + ndim * [0.1,]
        self.step = ndim*ndim * [0.01,]  + ndim * [1.0,]
        self.ftol = 0.001       # Assumes per-voxel discrepancy

        return
    

    #------------------------------------------------------------------------
    # Initialize transform from an obrigid precursor
    # FIXME: This method is ugly!  Consider defining methods set_itgt2src()
    # and set_ptgt2src() instead; then only the names are ugly.
    # FIXME: Remove this method and use the init argument instead?
    #------------------------------------------------------------------------
    
    def initrigid(self, rigid) :
        self.init     = rigid.ctgt2src()
        self.scenter  = rigid.scenter
        self.tcenter  = rigid.tcenter

        return
    

    #------------------------------------------------------------------------
    # Construct centered target-to-source transform; this is the
    # transform directly derived from the transform parameters and
    # thus the form on which the optimizer works.  It is also the
    # affine transform that should be used to initialize a more
    # general type of transform.
    #-----------------------------------------------------------------------
    
    def ctgt2src(self, param=None) :
        if not param :
            param = self.xopt
        if self.init :
            aff = self.init.copy()
        else :
            aff = pyvox.affine(self.ndim)
        n0 = self.ndim
        n1 = n0*n0
        n2 = self.npar
        mat = pyvox.array([n0,n0], exim.double, param[ 0:n1].list())
        off = pyvox.array([n0, 1], exim.double, param[n1:n2].list())
        aff = aff.compose2(mat, off)
        return aff


    #-----------------------------------------------------------------------
    # Construct physical target-to-source transform
    #-----------------------------------------------------------------------
    
    def ptgt2src(self, param=None) :
        aff = self.ctgt2src(param)
        if self.tcenter :
            aff.translate(self.tcenter, -1)
        if self.scenter :
            aff.translate(self.scenter)
        return aff


    #------------------------------------------------------------------------
    # Construct index target-to-source transform
    #------------------------------------------------------------------------
    
    def itgt2src(self, param=None) :
        aff = self.ptgt2src(param)
        aff.translate(self.target.origin,   1)
        aff.scale    (self.target.spacing,  1)
        aff.translate(self.source.origin,  -2)
        aff.scale    (self.source.spacing, -2)
        return aff


    #------------------------------------------------------------------------
    # Construct shrunk target-to-source transform
    # For internal use; maps source onto shrunk target image
    #------------------------------------------------------------------------
    
    def stgt2src(self, param=None, shrink=None) :
        aff = self.itgt2src(param)
        if shrink :
            aff.scale(2.0**shrink, 1)
            aff.scale(1.0/(2.0**shrink))
        return aff


    #-----------------------------------------------------------------------
    # Set parameters from an affine transform
    # FIXME: Deprecated
    #-----------------------------------------------------------------------
    
    def setaffine (self, aff) :
        self.xopt = aff.matrix.list() + aff.offset.list()
        return


    #------------------------------------------------------------------------
    # Construct target-to-source transform for given param values
    # FIXME: Deprecated
    #------------------------------------------------------------------------
    
    def tgt2src(self, param=None, shrink=None) :
        raise Exception, \
              'tgt2src method is deprecated; use itgt2src or ptgt2src instead'
        if not param :
            param = self.xopt
        aff = pyvox.affine(self.ndim)
        if shrink :
            aff.scale(2.0**shrink)
        if self.tcenter :
            aff.translate(-self.tcenter)
        n0 = self.ndim

        n1 = n0*n0
        n2 = self.npar
        mat = pyvox.array([n0,n0], exim.double, param[ 0:n1].list())
        off = pyvox.array([n0, 1], exim.double, param[n1:n2].list())
        aff.compose2(mat, off)
        if self.scenter :
            aff.translate(self.scenter)
        if shrink :
            aff.scale(1.0/(2.0**shrink))
        # aff.write()   # DEBUG
        return aff



#----------------------------------------------------------------------------
# obtrans - Objective function for registration by translation only
#
# The parameters to be optimized consist of one translation for each
# dimension of the source and target.
#
# FIXME: This class has not been tracking changes to the obaffine and
# obrigid classes and may be obsolete now.
#----------------------------------------------------------------------------

class obtrans (obaffine) :


    #-----------------------------------------------------------------------
    # Initialize instance, including optimization parameters
    #----------------------------------------------------------------------
    
    def __init__(self, source, target) :
        if source.rank != target.rank :
            raise RegistrationError, "Source and target have different rank"
        self.ndim = source.rank    # Image dimension
        self.npar = self.ndim      # Nmbr of params to optimize
        obregis.__init__(self, source, target)
        self.xtol   = self.ndim * [ 0.1, ]
        self.ftol   = 0.001      # Assumes per-voxel discrepancy
        self.step   = self.ndim * [ 1.0, ]
        return


    #------------------------------------------------------------------------
    # Construct centered target-to-source transform; this is the
    # transform directly derived from the transform parameters and
    # thus the form on which the optimizer works.  It is also the
    # affine transform that should be used to initialize a more
    # general type of transform.
    #------------------------------------------------------------------------
    
    def ctgt2src(self, param=None) :
        if not param :
            param = self.xopt
        aff = pyvox.affine(self.ndim)
        aff.translate(param)
        return aff



#----------------------------------------------------------------------------
# obrigid - Objective function for rigid registration
#
# The parameters to be optimized consist of three translations and three
# rotation angles in 3D, or two translations and one rotation angle in 2D.
#
# FIXME: Currently works for unsigned char only.
#
# FIXME: Risk of gimbal lock in 3D if angles get near pi/2
#----------------------------------------------------------------------------

class obrigid (obaffine) :


    #----------------------------------------------------------------------
    # Initialize instance, including optimization parameters
    # FIXME: It seems rather awkward that this class is derived from
    # obaffine but calls obregis.__init__ for initialization.
    #----------------------------------------------------------------------
    
    def __init__(self, source, target, init=None) :
        if source.rank != target.rank :
            raise RegistrationError, "Source and target ranks do not match"
        if source.rank == 3 :
            self.ndim = 3         # Image dimension
            self.npar = 6         # Nmbr of params to optimize
        elif source.rank == 2 :
            self.ndim = 2         # Image dimension
            self.npar = 3         # Nmbr of params to optimize
        else :
            raise RegistrationError, "obrigid supports only 2D or 3D images"
        obregis.__init__(self, source, target)
        if init == None or isinstance(init, pyvox.AffineType) :
            self.init = init
        elif isinstance(init, obrigid) or isinstance(init, obaffine) :
            self.init = init.ctgt2src()
            self.metric  = init.metric
            self.scenter = init.scenter
            self.tcenter = init.tcenter
        else :
            raise RegistrationError, \
                  "Cannot translate init argument to an initial transform"
        if self.ndim == 3 :
            self.step   = [ 1.0, 1.0, 1.0, 0.01, 0.01, 0.01 ]
            self.xtol   = [ 0.1, 0.1, 0.1, 0.001, 0.001, 0.001 ]
            self.ftol   = 0.001      # Assumes per-voxel discrepancy
        else :
            self.step   = [ 1.0, 1.0, 0.01 ]
            self.xtol   = [ 0.1, 0.1, 0.001 ]
            self.ftol   = 0.001      # Assumes per-voxel discrepancy
            return


    #------------------------------------------------------------------------
    # Construct centered target-to-source transform
    # Use this one to initialize an affine registration
    #------------------------------------------------------------------------
    
    def ctgt2src(self, param=None) :
        if not param :
            param = self.xopt
        if self.init :
            aff = self.init.copy()
        else :
            aff = pyvox.affine(self.ndim)
        if self.ndim == 3 :
            aff.translate(param[0:3])
            aff.rotate(1, 2, param[3])
            aff.rotate(0, 2, param[4])
            aff.rotate(0, 1, param[5])
        else :
            aff.translate(param[0:2])
            aff.rotate(0, 1, param[2])
        return aff



#---------------------------------------------------------------------------
# obaffinegb - Objective function for affine registration, plus gain and bias
#
# Parameters consist of 2 intensity gain/offsets, 9 matrix elements, and 3 
# translations.
#
# FIXME: This class is arguably obsolete and should be deleted.  If not
# deleted, it needs to be made consistent with obaffine.
#
# FIXME: Currently for 3D unsigned char images only.
#
# FIXME: Need a way to scale an uchar image by a float value.  Array.uchar
# with limits works but seem sort of inelegant.
#
# FIXME: Provide a method to print out the parameters with labels for
# human consumption?  Make the stepwise printouts optional?
#--------------------------------------------------------------------------

class obaffinegb (obregis) :


    #-----------------------------------------------------------------------
    # Initialize instance, including optimization parameters
    #-----------------------------------------------------------------------
    
    def __init__(self, source, target) :
        if source.rank != 3 or target.rank != 3 :
            raise RegistrationError, "obaffine supports only 3D images"
        self.ndim =  3         # Image dimension
        self.npar = 14         # Nmbr of params to optimize
        obregis.__init__(self, source, target)
        self.step = [0.01, 5.0] + 9 * [0.01,]  + 3 * [1.0,]
        self.xtol = [0.05, 1.0] + 9 * [0.001,] + 3 * [0.1,]
        self.ftol = 0.05       # Assumes per-voxel discrepancy

        return


    #-----------------------------------------------------------------------
    # Set parameters from an affine transform
    #----------------------------------------------------------------------
    
    def setaffine (self, aff) :
        self.xopt = [1.0, 0.0] + aff.matrix.list() + aff.offset.list()
        return

    #-----------------------------------------------------------------------
    # Construct target-to-source transform for given param values
    #----------------------------------------------------------------------
    
    def tgt2src(self, param=None, shrink=None) :
        if not param :
            param = self.xopt
        aff = pyvox.affine(self.ndim)
        if shrink :
            aff.scale(2.0**shrink)
        if self.tcenter :
            aff.translate(-self.tcenter)
        mat = pyvox.array([3,3], exim.double, param[ 2:11].list())
        off = pyvox.array([3,1], exim.double, param[11:14].list())
        aff.compose2(mat, off)
        if self.scenter :
            aff.translate(self.scenter)
        if shrink :
            aff.scale(1.0/(2.0**shrink))
        return aff



#---------------------------------------------------------------------------
# obpoly - Objective function for polynomial registration
#
# The init argument must be a polynomial transform and the coefficients
# that it provides will be the parameters to be optimized.  A gain or bias
# difference between the two images should not matter for either correl or
# mutinfo metric.
#
# FIXME: The init argument is rather limited, but it cannot currently be
# generalized because of the way that superclass initialization is done.
#
# FIXME: Currently supports unsigned char images only.
#
# FIXME: Provide a method to print out the parameters with labels for
# human consumption?  Make the stepwise printouts optional?
#--------------------------------------------------------------------------

class obpoly (obregis) :


    #----------------------------------------------------------------
    # Initialize instance, including optimization parameters
    #
    # FIXME: The notions of the base and init transforms seem to be
    # not really understood here.  Work on them further.
    #----------------------------------------------------------------
    
    def __init__(self, source, target, init) :

        # Check that source and target are consistent
        if source.rank != target.rank :
            raise RegistrationError, \
                  "source and target have different number of dimensions"

        # Initialize various parameters and the base class
        self.ndim = source.rank
        self.init = init
        self.npar = len(init.terms)
        obregis.__init__(self, source, target)

        # Initialize optimization parameters
        # FIXME: Is there a reasonable theory for setting these
        # parameters?  Or perhaps they should be provided by the
        # transform, which might be in a better position to pick
        # good values.
        nterms = len(self.init.keys)
        self.ftol = 0.0001        # Assumes per-voxel discrepancy
        self.xopt = nterms * [0.0]
        self.xtol = nterms * [0.01]
        self.step = nterms * [1.0]

        return


    #------------------------------------------------------------------
    # Construct centered target-to-source transform
    #
    # The centered target-to-source transform is the init transform composed
    # with a perturbation defined by the optimization parameters.  It is thus
    # the transform directly derived from the transform parameters and the
    # form on which the optimizer works.  It is also the transform that should
    # be used to initialize a more general type of transform.
    #
    # FIXME: Think much more carefully about how the base and init transforms
    # work.
    #-------------------------------------------------------------------
    
    def ctgt2src(self, param=None) :
        if not param :
            param = self.xopt

        # Save poly coefs in the transform
        poly = self.init.copy()
        poly.setparam(param)

        return poly


    #--------------------------------------------------------------------
    # Construct physical target-to-source transform
    #--------------------------------------------------------------------
    
    def ptgt2src(self, param=None) :
        poly = self.ctgt2src(param)
        if self.tcenter :
            poly.translate(self.tcenter, -1)
        if self.scenter :
            poly.translate(self.scenter, 2)

        return poly


    #----------------------------------------------------------------------
    # Construct index target-to-source transform
    #----------------------------------------------------------------------
    
    def itgt2src(self, param=None) :
        poly = self.ptgt2src(param)
        poly.translate(self.target.origin,   1)
        poly.scale    (self.target.spacing,  1)
        poly.translate(self.source.origin,  -2)
        poly.scale    (self.source.spacing, -2)

        return poly


    #---------------------------------------------------------------------
    # Construct shrunk target-to-source transform
    #
    # For internal use; maps source onto shrunk target image
    #---------------------------------------------------------------------
    
    def stgt2src(self, param=None, shrink=None) :
        poly = self.itgt2src(param)
        if shrink :
            poly.scale(2.0**shrink, 1)
            poly.scale(1.0/(2.0**shrink))

        return poly



#---------------------------------------------------------------------------
# affreg class - Affine or rigid registration of volume images
#
# This provides a generic turnkey mechanism to register one volume image to
# another, using either an affine or rigid transformation and any of the
# mean square, correlation or mutual information metrics.  The general
# pattern of use is:
#
#      foo = regis.affreg(options...)     # Create instance; set parameters
#      foo.register()                     # Compute registration transform
#      result  = foo.resample()           # Resample source per transform
#      result2 = foo.resample(image)      # Resample any image per transform
#
# FIXME: Provide adequate documentation for this class, particularly including
# the settable options.
#
# FIXME: Add support for cubic, sinc, etc interpolation.
#
# FIXME: It might make good sense to move the functionality here into the
# obaffine class.
#
# FIXME: AnalyzeView images are prone to have extra dimensions of length
# one, which really confuses the process of figuring out the "real" number
# of dimensions to register in, particularly since the coordinate
# system of the resampling image should replicate the extra dimensions.
# This is not handled well in this code; try to think of a better approach.
# Note also that if the calling program does any resampling, etc, on its
# own, it may also be confused by the extra dimensions.
#
# FIXME: Handle intensity ranges other than 0-255; note that in the norm2
# metric, both images must be scaled the same.  Rather than min and max,
# it might work better to scale to 1st and 99th percentiles, or mean plus
# and minus 3 sigma.  Considering that zero is sometimes used as background,
# it might be useful to always preserve it.
#
# FIXME: Add options to match datatype and intensity range of the output
# image to the target image?
#
# FIXME: Show image datatypes.
#
# FIXME: Noticed for slab to whole-brain EPI on Kosha's phantom1 data that
# mutinfo gives more accurate results than norm2.  Suspect that this is
# due to the lack of a mask for norm2.  Investigate and fix if needed.
#
# FIXME: Found for whole-brain EPI to structural on Kosha's phantom1 data
# that Brent's method is *slower* than golden section.  This suggests either
# that there is a lot of noise in the metric values near the optimum or that
# Brent's method is not implemented quite right.
#----------------------------------------------------------------------------

class affreg :


    #----------------------------------------------------------------------
    # Class variables
    #----------------------------------------------------------------------

    # Attributes that may be set by the user
    # FIXME: Some of these should depend on the program stage
    attrset = ['centering', 'diag', 'lineopt', 'metric', 'outfile',
               'program', 'reverse', 'source', 'srcfile', 'srcfile2',
               'target', 'tgtfile', 'trafile', 'trtype',
               'xyz',]

    # Settable parameters for stage in multi-scale program
    stageset = []

    # Standard program #1: Rigid transform, not mutinfo metric
    # FIXME: The trtype must be first.
    stdprog1 = [{"trtype": 'rigid', "shrink": 2, "smooth": 1},
                {"trtype": 'rigid', "shrink": 1, "smooth": 1},
                {"trtype": 'rigid', "shrink": 1, "smooth": 0},]
    
    # Standard program #2: Affine transform, not mutinfo metric
    stdprog2 = [{"trtype": 'rigid' , "shrink": 2, "smooth": 1},
                {"trtype": 'rigid' , "shrink": 1, "smooth": 1},
                {"trtype": 'affine', "shrink": 2, "smooth": 1},
                {"trtype": 'affine', "shrink": 1, "smooth": 0},]
    
    # Standard program #3: Rigid transform, mutinfo metric
    stdprog3 = [{"trtype": 'rigid', "shrink": 2, "smooth": 1},
                {"trtype": 'rigid', "shrink": 0, "smooth": 0},]
    
    
    #----------------------------------------------------------------------
    # Initialize an instance
    #----------------------------------------------------------------------

    def __init__(self, **dict) :

        # Set standard image parameters and defaults
        self.trtype  = "affine"         # Transform type: rigid or affine
        self.metric  = "correl"         # Correlation metric
        self.centering = "mass"         # How to determine centers of rotation
        self.reverse = 0                # Reverse roles of source and target?
        self.diag    = 0                # Write intermediate image files?
        self.outfile = ['affreg.hdr',   # Filename for primary resampled image
                        'affreg2.hdr']  # ...and for secondary
        self.trafile = 'affreg.dat'     # Name of transformation file
        self.xyz     = 0                # Use XYZ rather than normal ZYX order
        self.program = None             # Sequence of multi-scale stages
        self.lineopt = 'brent'          # One-dimensional line search optimizer
        #self.lineopt = 'golden'        # ...faster than Brent?!

        # Initialize from keyword dict, if provided
        for key, value in dict.items() :
            if key in self.attrset :
                setattr(self,key, value)
            else :
                raise Exception, "Unknown keyword '%s'" % key

        return


    #------------------------------------------------------------------------
    # affine2text - Convert affine transform to text
    #
    # FIXME: This should probably become a method of the affine class, as
    # should its inverse.
    #------------------------------------------------------------------------

    def affine2text(self, aff, xyz) :

        result = ""
        n1     = aff.ndim - 1
        for i0 in range(aff.ndim) :
            for j0 in range(aff.ndim) :
                i, j = i0, j0
                if xyz :
                    i = n1 - i0
                    j = n1 - j0
                result = result + "%12.6f " % aff.matrix[i,j]
            result = result + "%12.6f\n" % aff.offset[i,0]

        return result
    

    #------------------------------------------------------------------------
    # Rescale source and target intensities; cast to unsigned char
    #
    # This method is called by the register method and only rarely needs to
    # be directly called by the user.  The unscaled source image provided by
    # the user is retained as self.source; the rescaled image is retained as
    # self.source1.
    #
    # FIXME: Add option to specify type of scaling: None, max, stdev, ...
    #
    # FIXME: The norm2 metric requires that both source and target be rescaled
    # by the same factor.  Or does it?
    #------------------------------------------------------------------------

    def rescale(self) :

        # Rescale source image
        self.source1 = self.source.reshape()
        self.source1  = self.source1.uchar(0, self.source1.amax())
        print "   Source dimensions: %s\n" % self.source1.size
        print "   Source origin    : %s" % self.source1.origin
        print "   Source spacing   : %s" % self.source1.spacing

        # Rescale target image; save orig dimensions for later resampling
        self.tgt_siz = self.target.size 
        self.tgt_org = self.target.origin
        self.tgt_spc = self.target.spacing
        self.target  = self.target.uchar(0, self.target.amax()).reshape()
        self.tgt_i2p = self.target.i2p()
        print
        print "   Target dimensions: %s" % self.target.size
        print "   Target origin    : %s" % self.target.origin
        print "   Target spacing   : %s" % self.target.spacing

        return        


    #------------------------------------------------------------------------
    # Compute centers of rotation
    #
    # This method is called by the register method and only rarely needs to
    # be directly called by the user.
    #------------------------------------------------------------------------

    def center(self) :

        # Centering method 1: Set centers of rotation to centers of mass.
        # This method should generally be suitable for images that cover 
        # about the same volume but do not share a common coordinate system.
        if self.centering == 'mass' :
            (srcm, srccg, srccm) = self.source1.moments()
            srccg.reshape([3,1])
            (tgtm, tgtcg, tgtcm) = self.target.moments()
            tgtcg.reshape([3,1])

        # Centering method 2: Set centers of rotation to common center of
        # overlap region.
        # This method is recommended for images that share a common
        # coordinate system, such as structural and function images
        # taken in the same session; it does not require that the images
        # cover the same volume and is recommended for registering slabs to
        # whole-brain images.
        elif self.centering == 'coord' :
            srcmin = pyvox.vector(self.source1.origin)
            srcmax = pyvox.vector(self.source1.origin) + \
                     pyvox.vector(self.source1.spacing) * \
                     (pyvox.vector(self.source1.size) - 1)
            tgtmin = pyvox.vector(self.target.origin)
            tgtmax = pyvox.vector(self.target.origin) + \
                     pyvox.vector(self.target.spacing) * \
                     (pyvox.vector(self.target.size) - 1)
            minmin = srcmin.max(tgtmin)
            maxmax = srcmax.min(tgtmax)
            self.srccg = (minmin + maxmax) / 2.0
            self.tgtcg = self.srccg

        # Report centers of rotation
        print "\nCentering method is '%s'" % self.centering
        print "   Source center    : %s" % self.srccg.list()
        print "   Target center    : %s" % self.tgtcg.list()

        # FIXME: Add code to set the initial transforms here?

        return


    #-------------------------------------------------------------------------
    # Register source image to target image
    #-------------------------------------------------------------------------

    def register(self) :

        # Set logout file according to diag option
        if self.diag :
            self.logout = sys.stderr
        else :
            self.logout = None

        # Rescale images to unsigned char
        self.rescale()
        
        # Compute centers of rotation
        self.center()
        
        # Choose default multi-scale program, if not already specified
        # FIXME: Work out a better method for handling restricted affine
        # transforms; then add program for affine registration.
        if self.trtype == 'rigid' and self.metric != 'mutinfo' :
            self.program = self.stdprog1
        elif self.trtype == 'rigid' and self.metric == 'mutinfo' :
            self.program = self.stdprog3
        elif self.trtype == 'affine' :
            self.program = self.stdprog2
        elif not self.program :
            errmsg = "No default multi-scale program defined for given params"
            raise Exception, errmsg

        # Initialize the objective function
        trtype1 = 'rigid'
        if not self.reverse :
            obfun = obrigid(self.source1, self.target)
            obfun.scenter = self.srccg
            obfun.tcenter = self.tgtcg
        else :
            obfun = obrigid(self.target, self.source1)
            obfun.scenter = self.tgtcg
            obfun.tcenter = self.srccg
        obfun.metric  = self.metric
        obfun.logout  = self.logout

        # If requested, resample source image prior to registration
        if self.diag :
            itrans = obfun.itgt2src()
            oldimg = itrans.linear(self.source1, self.target.size)
            oldimg.spacing = self.target.spacing
            oldimg.origin  = self.target.origin
            oldimg.write('affreg0.hdr')

        # Loop over the stages of the multi-scale program
        for stage in self.program :

            # Update objective function with any new parameters
            # FIXME: Add support for per-stage step, ftol, xtol, ltol
            for key, value in stage.items() :
                if key == 'trtype' :
                    if trtype1 == 'rigid' and value == 'affine' :
                        trtype1 = 'affine'
                        if not self.reverse :
                            obfun = obaffine(self.source1, self.target, obfun)
                        else :
                            obfun = obaffine(self.target, self.source1, obfun)
                        obfun.logout = self.logout
                elif key in self.stageset :
                    setattr(obfun, key, value)
                elif key == 'shrink' :
                    shrink1 = value
                elif key == 'smooth' :
                    smooth1 = value
                else :
                    errmsg = "Unknown stage parameter '%s'" % key

            # Optimize as required for this stage
            # FIXME: Report other params, including trtype
            # FIXME: Make niter, lineopt settable?
            # FIXME: Add tolerances to the program parameters
            # FIXME: Understand and document how the tolerances work, and
            #        how the values should be chosen.
            if self.logout :
                msg = "\n%s registration at scale (%d,%d)\n" % \
                      (trtype1.capitalize(),  shrink1, smooth1)
                self.logout.write(msg)
            obfun.scale(shrink1, smooth1)
            obfun.niter = 10
            if shrink1 == 0 and smooth1 == 0 :      # DEBUG
                obfun.ltol = 0.01
                obfun.step = [0.1,  0.1,  0.1,  0.001,  0.001,  0.001 ]
                obfun.xtol = [0.01, 0.01, 0.01, 0.0001, 0.0001, 0.0001]
                obfun.ftol = 0  # 0.000001   DEBUG
            optim.powell(obfun, self.lineopt)

            # FIXME: Possibly useful value to finish rigid registration
            # obfun2.step = [0.05, 0.05, 0.05, 0.0001, 0.0001, 0.0001] # DEBUG
            # obfun2.ftol = 0.005   # DEBUG

            # FIXME: Possibly useful final stage for rigid registration
            # print "\nRigid registration at scale (0,1)"
            # obfun2.scale(0,1)
            # obfun2.ftol  = 0.001
            # obfun2.niter = 10
            # optim.powell(obfun2, self.lineopt)


        # Print out the transform giving the best match
        print
        self.ptgt2src = obfun.ptgt2src()
        if self.reverse :
            self.ptgt2src = self.ptgt2src.inverse()
        print "Final target-to-source transform in physical coordinates" + \
              "(ZYX order):"
        print self.ptgt2src
        self.itgt2src = self.ptgt2src.copy()
        self.itgt2src.compose(self.target.i2p(),  1)
        self.itgt2src.compose(self.source1.p2i(), 2)
        print "Final target-to-source transform in index coordinates" + \
              "(ZYX order):"
        print self.itgt2src

        # Write the best transform to a file
        trout = open(self.trafile, 'w')
        line = "# affreg: target-to-source transform, physical coordinates, "
        if self.xyz :
            line = line + "XYZ order\n"
        else :
            line = line + "ZYX order\n"
        trout.write(line)
        line = self.affine2text(self.ptgt2src, self.xyz)
        trout.write(line)
        trout.close()

        return


    #------------------------------------------------------------------
    # Old version of registration
    #------------------------------------------------------------------
    
    def old_register(self) :

        # Always use multiscale rigid transform for initial registration
        obfun1 = obrigid(self.source1, self.target)
        self.obfun1    = obfun1
        obfun1.metric  = self.metric
        obfun1.scenter = self.srccg
        obfun1.tcenter = self.tgtcg
        if self.diag :
            # Write resampled source image prior to registration
            itrans = obfun1.itgt2src()
            oldimg = itrans.linear(self.source1, self.target.size)
            oldimg.spacing = self.target.spacing
            oldimg.origin  = self.target.origin
            oldimg.write('affreg0.hdr')
            # Print each step of optization
            obfun1.logout = sys.stderr
        print "\nRigid registration at scale (2,1)"
        obfun1.scale(2, 1)
        obfun1.niter = 10
        optim.powell(obfun1, self.lineopt)
        print "\nRigid registration at scale (1,1)"
        obfun1.scale(1, 1)
        obfun1.niter = 10
        optim.powell(obfun1, self.lineopt)

        # Case 1: Finish up using rigid registration
        if self.trtype == 'rigid' :
            obfun2 = obfun1
            self.obfun2 = obfun1
            print "\nRigid registration at scale (1,0)"
            obfun2.scale(0,0)    # DEBUG
            obfun2.niter = 10
            # obfun2.step  = [0.05, 0.05, 0.05, 0.0001, 0.0001, 0.0001]   # DEBUG
            # obfun2.ftol  = 0.005   # DEBUG
            if self.diag :
                obfun2.logout = sys.stderr
            optim.powell(obfun2, self.lineopt)

            # print "\nRigid registration at scale (0,1)"
            # obfun2.scale(0,1)
            # obfun2.ftol  = 0.001
            # obfun2.niter = 10
            # optim.powell(obfun2, self.lineopt)

        # Case 2: Finish up using affine registration
        if self.trtype == 'affine' :
            obfun2 = obaffine(self.source1, self.target, obfun1)
            print "\nAffine registration at scale (2,1)"
            obfun2.scale(2,1)
            obfun2.niter = 10
            if diag :
                obfun2.logout = sys.stderr
            optim.powell(obfun2, self.lineopt)
            print "\nAffine registration at scale (1,0)"
            obfun2.scale(1,0)
            obfun2.niter = 10
            optim.powell(obfun2, lineopt)

        # Print out the best match found
        print
        self.ptgt2src = obfun2.ptgt2src()
        print "Final target-to-source transform in physical coordinates" + \
              "(ZYX order):"
        print self.ptgt2src
        self.itgt2src = obfun2.itgt2src()
        print "Final target-to-source transform in index coordinates" + \
              "(ZYX order):"
        print self.itgt2src

        # Write the best transform to a file
        trout = open(self.trafile, 'w')
        line = "# affreg: target-to-source transform, physical coordinates, "
        if self.xyz :
            line = line + "XYZ order\n"
        else :
            line = line + "ZYX order\n"
        trout.write(line)
        line = self.affine2text(self.ptgt2src, self.xyz)
        trout.write(line)
        trout.close()

        return


    #----------------------------------------------------------------------
    # Resample an image into target coordinates
    #
    # Resamples the given image into the target's coordinate system, using
    # the transform computed by the register method.  The image defaults to
    # the original, unscaled source image.
    #---------------------------------------------------------------------

    def resample(self, image=None) :

        # Generate informative error if register() hasn't run yet
        if not self.ptgt2src :
            raise Exception, "The register() method failed or wasn't called"

        # Default to original source image
        if not image :
            image = self.source
            
        # Compute the transform in index coordinates
        trans = self.ptgt2src.copy()
        trans.compose(self.tgt_i2p, 1)
        trans.compose(image.p2i(),  2)

        # Resample and copy metadata from original target image
        result = trans.linear(image, self.target.size)
        result.reshape(self.tgt_siz)
        result.origin  = self.tgt_org
        result.spacing = self.tgt_spc
        self.result = result

        return result


    #----------------------------------------------------------------------
    # Compute bivariate histogram
    #
    # Must be run after resample().
    #----------------------------------------------------------------------

    def bihist(self) :

        res2 = self.result.uchar(0, self.result.amax())
        bihist, mar1, mar2 = res2.bihist(self.target)

        return bihist.logcomp()        
        

    #-----------------------------------------------------------------------
    # Run as a command line program
    #
    # This method is called from the affreg script and does all the real
    # work of that program.  It should not usually need to be called
    # directly by the user.
    #-----------------------------------------------------------------------

    def command(self, argv) :

        # Collect options from the command line
        nout = 0                       # Nr times -o option seen so far
        opts, args = getopt.getopt(argv[1:], 'C:dM:o:rT:t:x')
        for opt in opts :
            if opt[0] == '-C' :
                self.centering = opt[1]
                if self.centering not in ('coord', 'mass') :
                    raise Exception, "Centering '%s' not recognized" % \
                          self.centering
            elif opt[0] == '-d' :
                self.diag = 1
            elif opt[0] == '-M' :
                self.metric = opt[1]
                if self.metric not in ('correl', 'norm2', 'mutinfo') :
                    raise Exception, "Metric '%s' not recognized" % metric
            elif opt[0] == '-o' :
                if nout >= 2 :
                    errmsg = "The -o option may not be used more than twice"
                    raise Exception, errmsg
                self.outfile[nout] = opt[1]
                nout = nout + 1
            elif opt[0] == '-r' :
                self.reverse = 1
            elif opt[0] == '-T' :
                self.trtype = opt[1]
                if self.trtype not in ('affine', 'rigid') :
                    errmsg = "Transform type '%s' not recognized" % trtype
                    raise Exception, errmsg
            elif opt[0] == '-t' :
                self.trafile = opt[1]
            elif opt[0] == '-x' :
                self.xyz = 1

        # Get arguments from the command line
        argc = len(args)
        if argc < 2 :
            print "Command format is: <program> srcfile tgtfile [srcfile2]"
        self.srcfile  = args[0]
        self.tgtfile  = args[1]
        self.srcfile2 = None
        if len(args) > 2 :
            self.srcfile2 = args[2]

        # Read and prepare the source image; remember scource before scaling
        print "\nReading the source image"
        print "   Source file      : %s" % self.srcfile
        self.source = pyvox.read(self.srcfile)
        
        # Read and prepare the target image; remember original target
        # parameters to apply to resampled source image(s)
        print "\nReading the target image"
        print "   Target file      : %s" % self.tgtfile
        self.target  = pyvox.read(self.tgtfile)

        # Register the source to the target image
        self.register()

        # Transform primary source according to the transform found; write out
        if self.outfile[0] :

            # Resample the source image
            print "Mapping primary source per final transform"
            resamp = self.itgt2src.linear(self.source, self.target.size)

            # Compute and output the bivariate histogram
            if self.diag :
                print "Computing the final bivariate histogram"
                resamp2 = resamp.uchar(0, resamp.amax())
                (bihist, mar1, mar2) = resamp2.bihist(self.target)
                bihist.logcomp().write("bihist.hdr")

            # Set parameters of resampled source to match the target
            resamp.reshape(self.tgt_siz)
            resamp.spacing = self.tgt_spc
            resamp.origin  = self.tgt_org

            # Write out the resampled image
            resamp.write(self.outfile[0])

        # Transform secondary source according to the transform found;
        # write out
        if self.srcfile2 and self.outfile[1] :
            print "Mapping secondary source per final transform"
            source2 = pyvox.read(self.srcfile2).reshape()
            strans = self.ptgt2src.copy()
            strans.compose(self.tgt_i2p, 1)
            strans.compose(source2.p2i())
            resamp2 = strans.linear(source2, self.target.size)
            resamp2.reshape(self.tgt_siz)
            resamp2.spacing = self.tgt_spc
            resamp2.origin  = self.tgt_org
            resamp2.write(self.outfile[1])

        return
