/*****************************************************************************
slamch.c - Report machine parameters for single precison floating point

This function reports various machine parameters for single precision
floating point, mostly as determined from the C include file float.h.
It replaces the version supplied by LAPACK, which attempted to compute
these parameters automatically but is apparently optimized incorrectly
by gcc 4.0.  In any case, this code is faster and simpler.


    CMACH   (input) CHARACTER*1   
            Specifies the value to be returned by DLAMCH:   
            = 'E' or 'e',   DLAMCH := eps   
            = 'S' or 's ,   DLAMCH := sfmin   
            = 'B' or 'b',   DLAMCH := base   
            = 'P' or 'p',   DLAMCH := eps*base   
            = 'N' or 'n',   DLAMCH := t   
            = 'R' or 'r',   DLAMCH := rnd   
            = 'M' or 'm',   DLAMCH := emin   
            = 'U' or 'u',   DLAMCH := rmin   
            = 'L' or 'l',   DLAMCH := emax   
            = 'O' or 'o',   DLAMCH := rmax   

            where   

            eps   = relative machine precision   
                  = FLT_EPSILON
            sfmin = safe minimum, such that 1/sfmin does not overflow   
                  = FLT_MIN, but fudged if needed to prevent overflow
            base  = base of the machine   
                  = FLT_RADIX
            prec  = eps*base 
                  = FLT_EPSILON * FLT_RADIX
            t     = number of (base) digits in the mantissa   
                  = FLT_MANT_DIG
            rnd   = 1.0 when rounding occurs in addition, 0.0 otherwise   
                  Not provided by C; we always return 1.0
            emin  = minimum exponent before (gradual) underflow   
                  = FLT_MIN_EXP
            rmin  = underflow threshold = base**(emin-1)   
                  = FLT_MIN
            emax  = largest exponent before overflow   
                  = FLT_MAX_EXP
            rmax  = overflow threshold  = (base**emax)*(1-eps)   
                  = FLT_MAX


FIXME: The rnd parameter (rounding during addition, rather than truncation) 
is not defined in float.h.  Here we assume that it is 1 (which I think is 
required by IEEE 754).  This may break on platforms that use other than IEEE
or that provide less than fully compliant IEEE floating point.
*****************************************************************************/

#include <float.h>
#include "f2c.h"

doublereal                   /* Returns requested machine parameter */
slamch_(
    char *cmach)             /* Specifies desired parameter; see above */
{
    doublereal sfmin;

    if (*cmach == 'E' || *cmach == 'e')
        return FLT_EPSILON;
    else if (*cmach == 'S' || *cmach == 's') {
        /* We start off with the smallest representable normalized value,
           and fudge to make sure that it can be inverted without overflow. */
        sfmin = FLT_MIN;
        if (sfmin <= 1.0/FLT_MAX)
            sfmin = (1.0/FLT_MAX)*(1.0+FLT_EPSILON);
        return sfmin;
    }
    else if (*cmach == 'B' || *cmach == 'b')
        return FLT_RADIX;
    else if (*cmach == 'P' || *cmach == 'p')
        return FLT_RADIX * FLT_EPSILON;
    else if (*cmach == 'N' || *cmach == 'n')
        return FLT_MANT_DIG;
    else if (*cmach == 'R' || *cmach == 'r')
        return 1;
    else if (*cmach == 'M' || *cmach == 'm')
        return FLT_MIN_EXP;
    else if (*cmach == 'U' || *cmach == 'u')
        return FLT_MIN;
    else if (*cmach == 'L' || *cmach == 'l')
        return FLT_MAX_EXP;
    else if (*cmach == 'O' || *cmach == 'o')
        return FLT_MAX;
    else
        return 0;
}

            
