/****************************************************************************
display.c - Display preparation functions for the Voxel Kit

Author:  Paul Hughett

Copyright (c) 1998-2003 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  This particular file defines various optimized
special-purpose functions for preparing selected slices of a voxel
array for display on the X Window system or related display software.
They are optimized so that a program using them can provide apparently
instant display of image data.
*****************************************************************************/

static const char rcsid[] = "$Id: display.c,v 1.13 2003/08/20 22:20:50 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include <float.h>
#include <limits.h>
#include <stdlib.h>

#include "errm.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"
#include "bips.h"



/*-------------------------------------------------------------------------
vxl_2d_section - Extract a full 2D slice from a voxel array

Given a voxel array and two coordinate axes, this function extracts
the 2d slice of the voxel array specified by dim0 and dim1 and stores
it as a two-dimensional voxel array.  This is useful for image display.

The slice argument is an array of length equal to the rank of the
source; each element corresponds to one source dimension and gives the
desired coordinate value in that dimension.  The elements
corresponding to the values of dim0 and dim1 are ignored.

FIXME:  No attempt has been made to optimize the cache hit ratio.
--------------------------------------------------------------------------*/

int                               /* Returns non-zero on error */
vxl_2d_section(
    voxel_array *dest,            /* Where to put the 2D section */
    voxel_array *src,             /* Array to extract section from */
    int dim0,                     /* Src dimen to become dest dimen 0 */
    int dim1,                     /* Src dimen to become dest dimen 1 */
    long *slice)                  /* Coordinates in other src dimens */
{
    long sbuff[VXL_MAX_RANK];     /* Working copy of dimensions/slice coords */
    long soff;                    /* Offset to start of data in src */
    long n0, n1;                  /* Nmbr elements in dest dimens 0 and 1 */
    long sd0, sd1;                /* Step sizes in source */
    long dd0;                     /* Step size in destination */
    long i0, i1;                  /* Index over voxels in destination */
    int i, err;

    /* Verify the arguments provided.  Note that the validity of the
       values in slice will be verified in vxli_offset. */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    if (src->rank < 2)
	NZReturnError(ERR_ILLOP, "Rank of source is less than 2");
    if (dim0 < 0 || dim0 >= src->rank)
	NZReturnError(ERR_ILLOP, "dim0 specifies invalid slice");
    if (dim1 < 0 || dim1 >= src->rank)
	NZReturnError(ERR_ILLOP, "dim1 specifies invalid slice");

    /* Set up the destination voxel array */
    n0  = sbuff[0] = src->dimen[dim0];
    n1  = sbuff[1] = src->dimen[dim1];
    err = vxl_alloc_array(dest, src->type, 2, sbuff);
    if (err)    NZReturnLibError();

    /* Compute location of (0,0) and step sizes for indexing */
    for (i = 0; i < src->rank; i++)
	sbuff[i] = slice[i];
    sbuff[dim0] = 0;
    sbuff[dim1] = 0;
    soff = vxli_offset(src, sbuff);
    if (soff < 0)    NZReturnLibError();
    sd0 = vxli_step(src, dim0);
    sd1 = vxli_step(src, dim1);
    dd0 = dest->dimen[1];

    /* The following code is the inner loop of vxl_2d_section.  It
       is written as a macro, since it must be repeated for each
       data type.  The do {...} while (0) construction is an ugly
       artifice that lets the macro look like a function call in
       any context. */
    #define vxl_2d_section_loop(type)                       \
    do {                                                    \
	type *dbase, *sbase;    /* Location of (0,0) */     \
	dbase = (type *)dest->data;                         \
	sbase = (type *)src->data + soff;                   \
	for (i0 = 0; i0 < n0; i0++)                         \
	    for (i1 = 0; i1 < n1; i1++)                     \
	        dbase[i0*dd0 + i1] = sbase[i0*sd0+i1*sd1];  \
    } while (0)

    /* Copy the requested slice, using the correct C type for data */
    switch (dest->type) {

    case INTP_UCHAR: 
	vxl_2d_section_loop(unsigned char);  
	break;

    case INTP_USHORT: 
	vxl_2d_section_loop(unsigned short);
	break;

    default:
	NZReturnError1(ERR_UNIMP, "Voxel type %s not yet implemented",
                       exim_nameof_type(dest->type));

    }   /* End of switch by voxel type */

    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_2d_sectwin - Extract and intensity window a 2D slice from a voxel array

Given a voxel array and two coordinate axes, this function extracts
the 2d slice of the voxel array specified by dim0 and dim1, rescales
the intensity window given by gmin and gmax to the range [0..255], and
stores the result as a two-dimensional unsigned char voxel array.  The
last dimension of the voxel array runs over the bytes of a voxel; this
is to permit on-the-fly conversion from unsigned char, unsigned short
big-endian, and unsigned short little-endian in a display program
without forcing the user to know the data format in advance.

The slice argument is an array of length equal to the rank of the
source; each element corresponds to one source dimension and gives the
desired coordinate value in that dimension.  The elements
corresponding to the values of dim0 and dim1 are ignored.

FIXME:  No attempt has been made to optimize the cache hit ratio.
--------------------------------------------------------------------------*/

int                               /* Returns non-zero on error */
vxl_2d_sectwin(
    voxel_array *dest,            /* Where to put the 2D section */
    voxel_array *src,             /* Array to extract section from */
    int dim0,                     /* Src dimen to become dest dimen 0 */
    int dim1,                     /* Src dimen to become dest dimen 1 */
    long *slice,                  /* Coordinates in other src dimens */
    int bigend,                   /* 1 => big-endian, 0 => little-endian */
    int gmin,                     /* Gray level that maps to 0 */
    int gmax)                     /* Gray level that maps to 255 */
{
    long sbuff[VXL_MAX_RANK];     /* Working copy of dimensions/slice coords */
    long soff;                    /* Offset to start of data in src */
    long n0, n1;                  /* Nmbr elements in dest dimens 0 and 1 */
    long sd0, sd1;                /* Step sizes in source */
    long dd0;                     /* Step size in destination */
    long i0, i1;                  /* Index over voxels in destination */
    unsigned char *dbase;         /* Address of (0,0) in the destination */
    unsigned char *sbase;         /* Address of that voxel in the source */
    int ldim;                     /* Index of last (byte) dimension */
    int nbytes;                   /* Number of bytes per voxel */
    unsigned int  sgray;          /* Intensity of a source voxel */
    unsigned char dgray;          /* Intensity of a destination voxel */
    float scale;                  /* Source-to-dest conversion factor */
    int i, err;

    /* Verify the arguments provided.  Note that the validity of the
       values in slice will be verified in vxli_offset. */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    if (src->rank < 3)
	NZReturnError(ERR_ILLOP, "Rank of source is less than 3");
    if (dim0 < 0 || dim0 >= src->rank)
	NZReturnError(ERR_ILLOP, "dim0 specifies invalid slice");
    if (dim1 < 0 || dim1 >= src->rank)
	NZReturnError(ERR_ILLOP, "dim1 specifies invalid slice");
    ldim = src->rank - 1;
    nbytes = src->dimen[ldim];
    if (nbytes < 1 || nbytes > 2)
	NZReturnError(ERR_ILLOP, "nbytes is neither 1 nor 2");

    /* Set up the destination voxel array */
    n0  = sbuff[0] = src->dimen[dim0];
    n1  = sbuff[1] = src->dimen[dim1];
    err = vxl_alloc_array(dest, INTP_UCHAR, 2, sbuff);
    if (err)    NZReturnLibError();

    /* Compute location of (0,0) and step sizes for indexing */
    for (i = 0; i < src->rank; i++)
	sbuff[i] = slice[i];
    sbuff[dim0] = 0;
    sbuff[dim1] = 0;
    sbuff[ldim] = 0;        /* Ignore any byte coord from user */
    soff = vxli_offset(src, sbuff);
    if (soff < 0)    NZReturnLibError();
    sd0   = vxli_step(src, dim0);
    sd1   = vxli_step(src, dim1);
    dd0   = dest->dimen[1];
    dbase = (unsigned char *)dest->data;
    sbase = (unsigned char *)src->data + soff;
    scale = 255.0 / (gmax - gmin);

    /* Copy and window the requested slice, if source is char */
    if (nbytes == 1) {
	for (i0 = 0; i0 < n0; i0++)
	    for (i1 = 0; i1 < n1; i1++) {
	        sgray = sbase[i0*sd0+i1*sd1];
		if (sgray <= gmin)
		    dgray = 0;
		else if (sgray >= gmax)
		    dgray = 255;
		else
		    dgray = scale*(sgray - gmin) + 0.5;
		dbase[i0*dd0 + i1] = dgray;
	    }
    }

    /* Copy and window the requested slice, if source is big-endian */
    else if (nbytes == 2 && bigend) {
	for (i0 = 0; i0 < n0; i0++)
	    for (i1 = 0; i1 < n1; i1++) {
	        sgray  = sbase[i0*sd0+i1*sd1] << 8;
		sgray += sbase[i0*sd0+i1*sd1+1];
		if (sgray <= gmin)
		    dgray = 0;
		else if (sgray >= gmax)
		    dgray = 255;
		else
		    dgray = scale*(sgray - gmin) + 0.5;
		dbase[i0*dd0 + i1] = dgray;
	    }
    }

    /* Copy and window the requested slice, if source is little-endian */
    else  {  /* nbytes == 2 && !bigend */
	for (i0 = 0; i0 < n0; i0++)
	    for (i1 = 0; i1 < n1; i1++) {
	        sgray  = sbase[i0*sd0+i1*sd1];
		sgray += sbase[i0*sd0+i1*sd1+1] << 8;
		if (sgray <= gmin)
		    dgray = 0;
		else if (sgray >= gmax)
		    dgray = 255;
		else
		    dgray = scale*(sgray - gmin) + 0.5;
		dbase[i0*dd0 + i1] = dgray;
	    }
    }


    return ERR_OKAY;
}



/*------------------------------------------------------------------------
vxl_init_raw_format - Initialize a raw format struct with default values

This function initializes a raw format struct with the default values
for each field.  This makes it easier to add new fields later without
having to update every user of this structure (assuming, of course,
that all users are wise enough to use this function rather than doing
it themselves).  The default rank is 0 for an empty array, so you must
at least specify the rank and dimensions to get a sensible display.
------------------------------------------------------------------------*/

int
vxl_init_raw_format(
    struct vxl_raw_format *fmt)    /* Format struct to be initialized */
{
    int i;

    if (fmt == 0)
	NZReturnError(ERR_INIT, "Null pointer to vxl_raw_format struct");

    /* Describe the assumed format of the image data */
    fmt->magic  = VXL_RAWFORM;
    fmt->type   = INTP_UCHAR;
    fmt->bigend = 1;
    fmt->rank   = 0;
    for (i = 0; i < VXL_MAX_RANK; i++) {
	fmt->dimen [i] = 1;
	fmt->coords[i] = 0;
	fmt->direct[i] = 1; 
    }

    /* Describe its translation into an RGBA image */
    fmt->nchan = 3;
    for (i = 0; i < VXL_MAX_RGBA; i++) {
	fmt->minint[i]   =   0;
	fmt->maxint[i]   = 255;
	fmt->colormap[i] =   0;
        fmt->chan[i]     =   0;
    }
    fmt->axes[0] = 1;
    fmt->axes[1] = 0;
    fmt->axes[2] = 2;

    return 0;
}



/*---------------------------------------------------------------------------
vxl_print_raw_format - Print out contents of a vxl_raw_format struct

This function print outs the contents of a vxl_raw_format struct in
human-readable format on the specified stream.  It is intended for
debugging.
----------------------------------------------------------------------------*/

void 
vxl_print_raw_format(
    FILE *stream,                  /* Stream to print out on */
    struct vxl_raw_format *fmt)    /* Format struct to be printed */
{
    int i;

    fprintf(stream, "\nContents of vxl_raw_format struct at %p:", fmt);
    
    if (fmt == 0) {
        fprintf(stream, "\n   (Null pointer to struct)\n");
        return;
    }
    if (fmt->magic != VXL_RAWFORM) {
        fprintf(stream, "\n   (Invalid magic number %ld in struct)\n",
                fmt->magic);
        return;
    }

    fprintf(stream, "\n   type    = %ld", fmt->type);
    fprintf(stream, "\n   bigend  = %ld", fmt->bigend);
    fprintf(stream, "\n   rank    = %ld", fmt->rank);
    fprintf(stream, "\n   dimen   = ");
    for (i = 0; i < fmt->rank && i < VXL_MAX_RANK; i++)
        fprintf(stream, " %ld", fmt->dimen[i]);
    fprintf(stream, "\n   coords  = ");
    for (i = 0; i < fmt->rank && i < VXL_MAX_RANK; i++)
        fprintf(stream, " %ld", fmt->coords[i]);
    fprintf(stream, "\n   direct  = ");
    for (i = 0; i < fmt->rank && i < VXL_MAX_RANK; i++)
        fprintf(stream, " %ld", fmt->direct[i]);
    fprintf(stream, "\n   axes <Y,X,E> = %ld, %ld, %ld",
            fmt->axes[0], fmt->axes[1], fmt->axes[2]);
    fprintf(stream, "\n   nchan   = %ld", fmt->nchan);
    fprintf(stream, "\n   chan    = ");
    for (i=0; i < fmt->nchan && i < VXL_MAX_RGBA; i++)
        fprintf(stream, " %ld", fmt->chan[i]);
    fprintf(stream, "\n   minint  = ");
    for (i=0; i < fmt->nchan && i < VXL_MAX_RGBA; i++)
        fprintf(stream, " %f", fmt->minint[i]);
    fprintf(stream, "\n   maxint  = ");
    for (i=0; i < fmt->nchan && i < VXL_MAX_RGBA; i++)
        fprintf(stream, " %f", fmt->maxint[i]);
    fprintf(stream, "\n   colormap = ");
    for (i=0; i < fmt->nchan && i < VXL_MAX_RGBA; i++)
        fprintf(stream, " %p", fmt->colormap[i]);
    fprintf(stream, "\n");

    return;
}



/*-------------------------------------------------------------------------
vxl_raw_to_rgba2d - Convert external (raw) image format to 2d rgba section

This function takes image data in its external (raw) format and
extracts the red, green, blue, and alpha channels of an unsigned char
2d section, using intensity windowing to adjust the intensity range of
the raw data to the intensity range of the display.  The current
version will support 1-4 output channels; see the comments near the
definition vxl_raw_format for the usual interpretation of the
different channels.

The primary purpose of this function is to make it easy to write an
image display program that does not have to commit to an image data
format before it reads the data into memory but that can let the user
interactively try different data formats while trying to determine the
real format of an image data file abandoned on his doorstep.

NOTE: Prior to 2002-03-18, this function was named vxl_raw_to_rgb2d and
did not support an alpha channel.  

FIXME:  Implement color lookup tables for each channel.  Allow reversal
of specified coordinate directions.

FIXME: Is it useful to also support internal types for the source in this
function?
-------------------------------------------------------------------------*/

int                                 /* Returns non-zero on error */
vxl_raw_to_rgba2d(
    voxel_array *dst,               /* Output image, as rgba2d data */
    voxel_array *src,               /* Input  image, as raw data */
    struct vxl_raw_format *fmt)     /* Descriptor of external format */
{
    long xdim, ydim, edim;          /* Axes for x, y, echo/channel */
    int nchan;                      /* Number of channels in output */
    long sbuff[VXL_MAX_RANK];       /* Working copy of dimens/slice coords */
    int st;                         /* Type of the source data */
    int ft = INTP_FLOAT;            /* Type of the intermediate data */
    unsigned char *sdata;           /* Location of the raw (source) data */
    int snbytes;                    /* Number of bytes per source datum */
    float *fbuff;                   /* Buffer for intermediate data */
    int   bigend;                   /* Is source data bigendian? */
    unsigned char *dst0;            /* Start of a destination scanline */
    unsigned char *src0[VXL_MAX_RGBA];   /* Start of source scanline */
    float off[VXL_MAX_RGBA];        /* Intensity offsets for RGBA */
    float mul[VXL_MAX_RGBA];        /* Intensity multipliers for RGBA */
    long   nrows, ncols;            /* Size of the section */
    long   ssx, ssy;                /* Source strides in dest's x and y */
    long   dsx, dsy;                /* Destination strides in x and y */
    int err;                        /* Error in called function? */
    int ic;                         /* Index over channels */
    long i, del;

    /* Verify the arguments */
    if (dst == 0 || dst->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    if (fmt == 0 || fmt->magic != VXL_RAWFORM)
	NZReturnError(ERR_INIT, "Missing or invalid raw format specifier");
    if (fmt->rank < 2)
	NZReturnError(ERR_INIT, "Assumed rank of source is less than 2");

    /* Get and verify the transformation parameters */
    bigend = fmt->bigend;
    st   = fmt->type;
    ydim = fmt->axes[0];
    xdim = fmt->axes[1];
    edim = fmt->axes[2];
    nchan = fmt->nchan;
    if (ydim < 0 || ydim >= fmt->rank)
	NZReturnError(ERR_ILLOP, "Invalid choice of source Y coordinate");
    if (xdim < 0 || xdim >= fmt->rank)
	NZReturnError(ERR_ILLOP, "Invalid choice of source X coordinate");
    if (edim < 0 || edim > fmt->rank)  /* NOTE: edim MAY equal rank */
	NZReturnError(ERR_ILLOP, "Invalid choice of source channel coord");
    if ( nchan < 1 || nchan > 4 )
	NZReturnError(ERR_ILLOP, "Invalid number of channels in output");
    if ( edim == xdim || edim == ydim || xdim == ydim )
        NZReturnError(ERR_ILLOP, "Channels are not distinct");

    /* Set up the destination voxel array */
    nrows = sbuff[0] = fmt->dimen[ydim];
    ncols = sbuff[1] = fmt->dimen[xdim];
    sbuff[2] = nchan;
    err = vxl_alloc_array(dst, INTP_UCHAR, 3, sbuff);
    if (err)    NZReturnLibError();
    dst0 = dst->data;
    dsx  = nchan;
    dsy  = nchan * ncols;

    /* Set up indexing for the source array */
    snbytes = exim_sizeof_extype(fmt->type);
    if (snbytes == 0)
        NZReturnError(ERR_ILLOP, "Invalid data type");
    ssx = snbytes * vxli_step2(fmt->rank, fmt->dimen, xdim);
    ssy = snbytes * vxli_step2(fmt->rank, fmt->dimen, ydim);
    sdata = src->data;
    for (i = 0; i < VXL_MAX_RANK; i++)
	sbuff[i] = fmt->coords[i];
    sbuff[xdim] = 0;
    sbuff[ydim] = 0;
    if (0 <= edim && edim < fmt->rank)  {                   
	/* Channel is an explicit dimension */
	for (ic = 0; ic < nchan; ic++) {
	    sbuff[edim] = fmt->chan[ic];
            del = vxli_offset2(fmt->rank, fmt->dimen, sbuff);
            if (del < 0)    NZReturnLibError();
	    src0[ic] = sdata + snbytes * del;
	}
    }
    else {
	/* Channel is an implicit dimension */
        del = vxli_offset2(fmt->rank, fmt->dimen, sbuff);
        if (del < 0)    NZReturnLibError();
        src0[0]= sdata + snbytes * del;
	for (ic = 1; ic < nchan; ic++)
	    src0[ic]= src0[0];
    }

    /* Set up an intermediate float buffer */
    fbuff = mallock(ncols*sizeof(float));
    if (fbuff == 0)    NZReturnLibError();

    /* DEBUG:  Print out some possibly useful values */
    if (0) {
	fprintf(stderr, "rank = %ld,  xdim = %ld  ydim = %ld  edim = %ld\n",
		fmt->rank, xdim, ydim, edim);
	fprintf(stderr, "dimen =  %ld, %ld, %ld, %ld\n",
		fmt->dimen[0], fmt->dimen[1], fmt->dimen[2], fmt->dimen[3]);
	fprintf(stderr, "sbuff = %ld, %ld, %ld, %ld\n",
		sbuff[0], sbuff[1], sbuff[2], sbuff[3]);
	fprintf(stderr, "nbytes = %d ssx = %ld  ssy = %ld  src0[0] = %ld\n", 
		snbytes, ssx, ssy, (long)(src0[0] - (unsigned char *)(sdata)));
    }

    /* Set up intensity windowing for each color channel */
    for (ic = 0; ic < nchan; ic++) {
	off[ic] = fmt->minint[fmt->chan[ic]];
	mul[ic] = 255.0 / 
	    (fmt->maxint[fmt->chan[ic]] - fmt->minint[fmt->chan[ic]]);
    }
    
    /* Loop over scanlines in the destination */
    for (i = 0; i < nrows; i++) {

	/* Loop over the output channels */
	for (ic = 0; ic < nchan; ic++) {

	    /* Convert input channel from raw to float */
	    err = exim_import(ncols, fbuff, ft, 1, 
			      src0[ic], st, ssx, bigend);
	    if (err)
		NZReturnError(ERR_MISC, "Error importing image data");

	    /* Scale input channel and cast to unsigned char */
	    err = bips_scale_to_uchar(ncols, dst0+ic, dsx, 
				      ft, fbuff, 1, off[ic], mul[ic]);
	    if (err)	NZReturnLibError();

	    /* Find beginning of next source scanline */
	    src0[ic] += ssy;
	}

	/* Find beginning of next destination scanline */
	dst0  += dsy;
    }

    /* Free the intermediate buffer */
    free(fbuff);

    return ERR_OKAY;
}
