/****************************************************************************
element.c - Elementwise functions for the Voxel Kit

Author: Paul Hughett

Copyright (c) 1998-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  This particular file defines elementwise operations on
voxel arrays.

FIXME: Add voxel kit functions for compound assignments, or updating
in place, to follow the BIPS functions for this purpose.  Maybe also
extend upwards into Pyvox as well (since they will save storage in
many applications).

FIXME: Add the new C99 math functions isinf, isnan, and so on when
supported by the host platform.
*****************************************************************************/

static const char rcsid[] = "$Id: element.c,v 1.46 2005/12/29 18:33:14 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include <float.h>
#include <limits.h>
#include <stdlib.h>
#include <time.h>

#ifdef HAVE_TGMATH_H
#include <tgmath.h>
#else
#include <math.h>
#endif

#include "rand.h"
#include "errm.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"
#include "bips.h"



/*------------------------------------------------------------------------
vxli_ran1state - Get state vector for random number generation

This function takes a pointer to a state vector for the ran1 generator
and checks whether or not it is zero.  If it is non-zero, this
function simply returns it.  Otherwise, it returns the address of the
default state vector, initializing it if necessary from the current
time.

FIXME: This code will break in 2038, when the time_t overflows.
-------------------------------------------------------------------------*/

static void *ran1state = 0;   /* State vector for the default RNG */

static void *                 /* Returns zero on error */
vxli_ran1state(
    void *state)              /* State offered as argument */
{
    long seed;

    /* If caller provided a valid state, just return that */
    if (state != 0)
        return state;

    /* If necessary, initialize the default random generator from
       the current time. */
    if (ran1state == 0) {
        seed = time(0);
        ran1state = ran1_init(seed);
	if (ran1state == 0)    ZReturnLibError();
    }

    /* Return the default random number generator state */
    return ran1state;
}



/***************************************************************************
                   Standard Type Casting Operations

The following functions for elementwise type casting of voxel arrays fit
a common pattern and are implemented using shared code.
***************************************************************************/


/*------------------------------------------------------------------------
StandardCast - Common core for a "standard" cast operation

Most of the typecasting functions in the Voxel Kit follow the same
pattern; here we create a macro for that pattern to shorten the code
and simplify modifications.  This is done as a macro rather than an
inline function since the type of the ddata pointer varies.

The macro arguments are the C type of the destination (e.g. short),
the numeric type code (e.g. INTP_SHORT), and the BIPS function to
actually do the cast (e.g. bips_short).
-------------------------------------------------------------------------*/

#define StandardCast(dsttype, dstcode, bips_cast)                  \
    long count;            /* Total number of voxels */            \
    dsttype  *ddata;       /* Destination value array */           \
    void     *sdata;       /* Source value array */                \
    int stype;             /* Type of the source data */           \
    int err;                                                       \
                                                                   \
    /* Validate the source argument */                             \
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)     \
	NZReturnError(ERR_INIT, "Invalid or empty source array");  \
    stype = src->type;                                             \
                                                                   \
    /* Allocate the destination; copy origin and spacing */        \
    err = vxl_alloc_array(dest, dstcode, src->rank, src->dimen);   \
    if (err)    NZReturnLibError();                                \
    vxli_copy_physical(dest, src);                                 \
                                                                   \
    /* Scale and convert the image */                              \
    count = vxli_count(src);                                       \
    ddata = dest->data;                                            \
    sdata = src->data;                                             \
    err = bips_cast(count, ddata, 1, stype, sdata, 1);             \
    if (err != 0)    NZReturnLibError();                           \
                                                                   \
    return 0



/*------------------------------------------------------------------------
vxl_uchar - Convert to unsigned char

This function converts a voxel array of any (non-complex) type into a
voxel array of type unsigned char.
-------------------------------------------------------------------------*/

int
vxl_uchar(
    voxel_array *dest,          /* Destination array; always uchar */
    voxel_array *src)           /* Source array */
{
    StandardCast(unsigned char, INTP_UCHAR, bips_uchar);
}



/*------------------------------------------------------------------------
vxl_schar - Convert to signed char

This function converts a voxel array of any (non-complex) type into a
voxel array of type signed char.
-------------------------------------------------------------------------*/

int
vxl_schar(
    voxel_array *dest,          /* Destination array; always signed char */
    voxel_array *src)           /* Source array */
{
    StandardCast(signed char, INTP_SCHAR, bips_schar);
}



/*------------------------------------------------------------------------
vxl_ushort - Convert to unsigned short

This function converts a voxel array of any (non-complex) type into a
voxel array of type unsigned short.
-------------------------------------------------------------------------*/

int
vxl_ushort(
    voxel_array *dest,          /* Destination array; always ushort */
    voxel_array *src)           /* Source array */
{
    StandardCast(unsigned short, INTP_USHORT, bips_ushort);
}



/*------------------------------------------------------------------------
vxl_short - Convert to short

This function converts a voxel array of any (non-complex) type into a
voxel array of type short.
-------------------------------------------------------------------------*/

int
vxl_short(
    voxel_array *dest,          /* Destination array; always short */
    voxel_array *src)           /* Source array */
{
    StandardCast(short, INTP_SHORT, bips_short);
}



/*------------------------------------------------------------------------
vxl_uint - Convert to unsigned int

This function converts a voxel array of any (non-complex) type into a
voxel array of type unsigned int.
-------------------------------------------------------------------------*/

int
vxl_uint(
    voxel_array *dest,          /* Destination array; always uint */
    voxel_array *src)           /* Source array */
{
    StandardCast(unsigned int, INTP_UINT, bips_uint);
}



/*------------------------------------------------------------------------
vxl_int - Convert to type int

This function converts a voxel array of any (non-complex) type into a
voxel array of type int.
-------------------------------------------------------------------------*/

int
vxl_int(
    voxel_array *dest,          /* Destination array; always int */
    voxel_array *src)           /* Source array */
{
    StandardCast(int, INTP_INT, bips_int);
}



/*------------------------------------------------------------------------
vxl_ulong - Convert to unsigned long

This function converts a voxel array of any (non-complex) type into a
voxel array of type unsigned long.
-------------------------------------------------------------------------*/

int
vxl_ulong(
    voxel_array *dest,          /* Destination array; always ulong */
    voxel_array *src)           /* Source array */
{
    StandardCast(unsigned long, INTP_ULONG, bips_ulong);
}



/*------------------------------------------------------------------------
vxl_long - Convert to type long

This function converts a voxel array of any (non-complex) type into a
voxel array of type long.
-------------------------------------------------------------------------*/

int
vxl_long(
    voxel_array *dest,          /* Destination array; always long */
    voxel_array *src)           /* Source array */
{
    StandardCast(long, INTP_LONG, bips_long);
}



/*------------------------------------------------------------------------
vxl_float - Convert to type float

This function converts a voxel array of any type into a voxel array of
type float.
-------------------------------------------------------------------------*/

int
vxl_float(
    voxel_array *dest,          /* Destination array; always float */
    voxel_array *src)           /* Source array */
{
    StandardCast(float, INTP_FLOAT, bips_float);
}



/*------------------------------------------------------------------------
vxl_double - Convert to type double

This function converts a voxel array of any type into a voxel array of
type double.
-------------------------------------------------------------------------*/

int
vxl_double(
    voxel_array *dest,          /* Destination array; always double */
    voxel_array *src)           /* Source array */
{
    StandardCast(double, INTP_DOUBLE, bips_double);
}



/*------------------------------------------------------------------------
vxl_fcomplex - Convert to type float complex

This function converts a voxel array of any type into a voxel array of
type float complex.
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H
int
vxl_fcomplex(
    voxel_array *dest,          /* Destination array; always float */
    voxel_array *src)           /* Source array */
{
    StandardCast(float complex, INTP_FCOMPLEX, bips_fcomplex);
}
#endif



/*------------------------------------------------------------------------
vxl_dcomplex - Convert to type double complex

This function converts a voxel array of any type into a voxel array of
type double complex.
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H
int
vxl_dcomplex(
    voxel_array *dest,          /* Destination array; always double */
    voxel_array *src)           /* Source array */
{
    StandardCast(double complex, INTP_DCOMPLEX, bips_dcomplex);
}
#endif



/*-----------------------------------------------------------------------
vxl_fcomplex2 - Convert real and imaginary parts to float complex

This function computes two voxel arrays of any single type into the
real and imaginary parts of a float complex voxel array.

FIXME: Handle type mismatches by just silently casting the lesser one?
------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H

int                             /* Returns nonzero on error */
vxl_fcomplex2(
    voxel_array *dest,          /* Destination array */
    voxel_array *src1,          /* Source array 1: real part */
    voxel_array *src2)          /* Source array 2: imaginary part */
{
    long count;                 /* Total number of voxels */            
    float complex *ddata;       /* Destination value array */           
    int stype;                  /* Type of both source arrays */           
    void  *s1data;              /* Source 1 data */                
    void  *s2data;              /* Source 2 data */                
    int err;                                                       
                                                                   
    /* Validate the source arguments */                             
    if (src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0)     
	NZReturnError(ERR_INIT, "Invalid or empty source 1 array");  
    stype = src1->type;                                              
    if (src2 == 0 || src2->magic != VXL_MAGIC || src2->data == 0)     
	NZReturnError(ERR_INIT, "Invalid or empty source 2 array");
    if (src2->type != stype)
        NZReturnError(ERR_INCOMP, "Source arrays have different types");
    
    /* Check that source shapes are compatible */
    if ( ! vxli_same_physical(src1, src2) )    NZReturnLibError();
                                                                       
    /* Allocate the destination; copy origin and spacing */        
    err = vxl_alloc_array(dest, INTP_FCOMPLEX, src1->rank, src1->dimen);   
    if (err)    NZReturnLibError();                                
    vxli_copy_physical(dest, src1);                                 
                                                                   
    /* Scale and convert the image */                              
    count  = vxli_count(src1);                                       
    ddata  = dest->data;                                            
    s1data = src1->data;                                             
    s2data = src2->data;                                             
    err = bips_fcomplex2(count, ddata, 1, stype, s1data, 1, s2data, 1);
    if (err != 0)    NZReturnLibError();                           
                                                                   
    return 0;
}

#endif



/*-----------------------------------------------------------------------
vxl_dcomplex2 - Convert real and imaginary parts to double complex

This function computes two voxel arrays of any single type into the
real and imaginary parts of a double complex voxel array.

FIXME: Handle type mismatches by just silently casting the lesser one?
------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H

int                             /* Returns nonzero on error */
vxl_dcomplex2(
    voxel_array *dest,          /* Destination array */
    voxel_array *src1,          /* Source array 1: real part */
    voxel_array *src2)          /* Source array 2: imaginary part */
{
    long count;                 /* Total number of voxels */            
    double complex *ddata;      /* Destination value array */           
    int stype;                  /* Type of both source arrays */           
    void  *s1data;              /* Source 1 data */                
    void  *s2data;              /* Source 2 data */                
    int err;                                                       
                                                                   
    /* Validate the source arguments */                             
    if (src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0)     
	NZReturnError(ERR_INIT, "Invalid or empty source 1 array");  
    stype = src1->type;                                              
    if (src2 == 0 || src2->magic != VXL_MAGIC || src2->data == 0)     
	NZReturnError(ERR_INIT, "Invalid or empty source 2 array");
    if (src2->type != stype)
        NZReturnError(ERR_INCOMP, "Source arrays have different types");
                          
    /* Check that source shapes are compatible */
    if ( ! vxli_same_physical(src1, src2) )    NZReturnLibError();

    /* Allocate the destination; copy origin and spacing */        
    err = vxl_alloc_array(dest, INTP_DCOMPLEX, src1->rank, src1->dimen);   
    if (err)    NZReturnLibError();                                
    vxli_copy_physical(dest, src1);                                 
                                                                   
    /* Scale and convert the image */                              
    count  = vxli_count(src1);                                       
    ddata  = dest->data;                                            
    s1data = src1->data;                                             
    s2data = src2->data;                                             
    err = bips_dcomplex2(count, ddata, 1, stype, s1data, 1, s2data, 1);
    if (err != 0)    NZReturnLibError();                           
                                                                   
    return 0;
}

#endif



/*------------------------------------------------------------------------
vxl_cast - Convert to any specified type

This function converts a voxel array of any type into a voxel array of
any type specified by the caller.
-------------------------------------------------------------------------*/

int
vxl_cast(
    voxel_array *dest,          /* Destination array */
    int dtype,                  /* Destination type */
    voxel_array *src)           /* Source array */
{
    /* Switch according to the destination type */
    switch (dtype) {
	
    case INTP_UCHAR:
	return vxl_uchar(dest, src);

    case INTP_SCHAR:
	return vxl_schar(dest, src);

    case INTP_USHORT:
	return vxl_ushort(dest, src);

    case INTP_SHORT:
	return vxl_short(dest, src);

    case INTP_UINT:
	return vxl_uint(dest, src);

    case INTP_INT:
	return vxl_int(dest, src);

    case INTP_ULONG:
	return vxl_ulong(dest, src);

    case INTP_LONG:
	return vxl_long(dest, src);

    case INTP_FLOAT:
	return vxl_float(dest, src);

    case INTP_DOUBLE:
	return vxl_double(dest, src);

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	return vxl_fcomplex(dest, src);

    case INTP_DCOMPLEX:
	return vxl_dcomplex(dest, src);
#endif

    default:
	NZReturnError(ERR_UNDEF, "Invalid destination type specified");

    }

    return 0;    /* Keeps gcc compiler happy */

}



/*-------------------------------------------------------------------------
vxl_scale_to_uchar - Scale and convert to unsigned char

This function scales a image of any non-complex type and converts it
to an unsigned character image.  The caller specifies lower and upper
limits which are converted to 0 and 255 respectively; input values
outside these limits are silently converted to 0 and 255.

FIXME: The vxl_scale() function does scaling, limiting, and casting to
a new type.  Does it replace this function?
-------------------------------------------------------------------------*/

int
vxl_scale_to_uchar(
    voxel_array *dest,          /* Destination array; always uchar */
    voxel_array *src,           /* Source array; any non-complex type */
    double minval,              /* Input value corresponding to 0 */
    double maxval)              /* Input value corresponding to 255 */
{
    double offset, gain;        /* Scaling parameters for bips */
    long count;                 /* Total number of voxels */
    unsigned char *ddata;       /* Destination value array */
    unsigned char *sdata;       /* Source value array */
    int stype;                  /* Type of the source data */
    int err;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    stype = src->type;

    /* Allocate the destination; copy origin and spacing */
    err = vxl_alloc_array(dest, INTP_UCHAR, src->rank, &(src->dimen[0]));
    if (err)    NZReturnLibError();
    vxli_copy_physical(dest, src);

    /* Scale and convert the image */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;
    offset = minval;
    gain   = 255.0 / (maxval - minval);
    err = bips_scale_to_uchar(count, ddata, 1, stype, sdata, 1, offset, gain);
    if (err != 0)    NZReturnLibError();

    return 0;
}



/***************************************************************************
                   Standard Unary Point Operations

The following unary point operations on voxel arrays fit a common pattern
and are implemented with shared code.
***************************************************************************/


/*-------------------------------------------------------------------------
vxli_standard_unop - Common core for a "standard" unary point operation

A "standard" unary point operation in BIPS is one whose calling sequence
is consistent with the bips_unop prototype shown in bips.h; the following
function is used to invoke a standard unary point operation to do the
defined operation on two voxel_arrays.  For an example of its use, see
the function vxl_not and its friends below.
-------------------------------------------------------------------------*/

static inline 
int                                /* Returns non-zero on error */
vxli_standard_unop(
    bips_unop *bips_fcn,           /* Standard BIPS function to use */
    voxel_array *dest,             /* Destination array */
    voxel_array *src1)             /* Source array */
{
    int type = 0;                  /* Internal type code for data */
    long ds = 1, ss1 = 1;          /* Stride in each array */
    long count;                    /* Total number of elements to process */
    int err;

    /* Check validity of source array; count number of elements */
    if ( src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0
	 || exim_sizeof_intype((type = src1->type)) == 0 )
	NZReturnError(ERR_INIT, "Invalid source array");
    count = vxli_count(src1);

    /* Allocate the destination; copy origin and spacing */
    err = vxl_alloc_array(dest, type, src1->rank, src1->dimen);
    if (err)    NZReturnLibError();
    vxli_copy_physical(dest, src1);

    /* Call the specified BIPS routine to do the actual computation */
    err = bips_fcn(count, type, dest->data, ds, src1->data, ss1);
    if (err != 0)    NZReturnLibError();

    return 0;
}



/*-------------------------------------------------------------------------
vxli_creal_unop - Common core for a complex->real unary point operation

This is a variant of vxli_standard_unop for those BIPS functions that
map real types into the same real types but convert complex types into
the corresponding real types; for instance, the real(), imag(), abs(),
and arg() functions.  

FIXME: Consider changing the calling sequence of the BIPS functions to
put the (source) type next to the source argument.
-------------------------------------------------------------------------*/

static inline 
int                                /* Returns non-zero on error */
vxli_creal_unop(
    bips_unop *bips_fcn,           /* Standard BIPS function to use */
    voxel_array *dest,             /* Destination array */
    voxel_array *src1)             /* Source array */
{
    int dtype = 0, stype = 0;      /* Internal type code for dest and src */
    long ds = 1, ss1 = 1;          /* Stride in each array */
    long count;                    /* Total number of elements to process */
    int err;

    /* Check validity of source array; count number of elements */
    if ( src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0
	 || exim_sizeof_intype((stype = src1->type)) == 0 )
	NZReturnError(ERR_INIT, "Invalid source array");
    count = vxli_count(src1);
    
    /* Allocate the destination; copy origin and spacing */
    dtype = exim_real_type(stype);
    err = vxl_alloc_array(dest, dtype, src1->rank, src1->dimen);
    if (err)    NZReturnLibError();
    vxli_copy_physical(dest, src1);

    /* Call the specified BIPS routine to do the actual computation */
    err = bips_fcn(count, stype, dest->data, ds, src1->data, ss1);
    if (err != 0)    NZReturnLibError();

    return 0;
}



/*-------------------------------------------------------------------------
Standard unary point operations on voxel arrays

These functions implement various unary point operations on voxel
arrays that follow a standard pattern

ALGORITHM: These definitions could be reduced to one line apiece using
an appropriate macro, but that would make them too hard for etags (and
the ordinary programmer) to find and understand.
--------------------------------------------------------------------------*/

/* See also vxl_copy_array, which does something slightly different. */
int vxl_copy(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_copy, dest, src1);
}

int vxl_not(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_not, dest, src1);
}

int vxl_minus(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_minus, dest, src1);
}

#ifdef HAVE_COMPLEX_H
int vxl_real(voxel_array *dest, voxel_array *src1)
{   return vxli_creal_unop(bips_real, dest, src1);
}

int vxl_imag(voxel_array *dest, voxel_array *src1)
{   return vxli_creal_unop(bips_imag, dest, src1);
}

int vxl_conj(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_conj, dest, src1);
}

int vxl_carg(voxel_array *dest, voxel_array *src1)
{   return vxli_creal_unop(bips_carg, dest, src1);
}
#endif

int vxl_abs(voxel_array *dest, voxel_array *src1)
{   return vxli_creal_unop(bips_abs, dest, src1);
}

int vxl_floor(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_floor, dest, src1);
}

int vxl_ceil(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_ceil, dest, src1);
}

int vxl_sqrt(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_sqrt, dest, src1);
}

#ifdef HAVE_CBRT
int vxl_cbrt(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_cbrt, dest, src1);
}
#endif

int vxl_exp(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_exp, dest, src1);
}

#ifdef HAVE_EXP2
int vxl_exp2(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_exp2, dest, src1);
}
#endif

#ifdef HAVE_EXPM1
int vxl_expm1(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_expm1, dest, src1);
}
#endif

int vxl_log(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_log, dest, src1);
}

int vxl_log10(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_log10, dest, src1);
}

#ifdef HAVE_LOG1P
int vxl_log1p(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_log1p, dest, src1);
}
#endif

#ifdef HAVE_LOG2
int vxl_log2(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_log2, dest, src1);
}
#endif

int vxl_sin(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_sin, dest, src1);
}

int vxl_cos(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_cos, dest, src1);
}

int vxl_tan(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_tan, dest, src1);
}

int vxl_asin(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_asin, dest, src1);
}

int vxl_acos(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_acos, dest, src1);
}

int vxl_atan(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_atan, dest, src1);
}

int vxl_sinh(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_sinh, dest, src1);
}

int vxl_cosh(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_cosh, dest, src1);
}

int vxl_tanh(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_tanh, dest, src1);
}

#ifdef HAVE_ERF
int vxl_erf(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_erf, dest, src1);
}
#endif

#ifdef HAVE_ERFC
int vxl_erfc(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_erfc, dest, src1);
}
#endif

#ifdef HAVE_LGAMMA
int vxl_lgamma(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_lgamma, dest, src1);
}
#endif

#ifdef HAVE_TGAMMA
int vxl_tgamma(voxel_array *dest, voxel_array *src1)
{   return vxli_standard_unop(bips_tgamma, dest, src1);
}
#endif




/***************************************************************************
                  Standard Binary Point Operations

The following binary point operations on voxel arrays all fit a single
pattern and are implemented using a shared code for that pattern.
***************************************************************************/


/*------------------------------------------------------------------------
vxli_setup_binpt - Setup a binary point operation on voxel arrays

This function does some setup processing that is common to all binary
point operations on voxel arrays (well, at least those that will
actually be done within bips), and is intended to make it easier to
permit one of the operands to be a scalar.  Specifically, this
function verifies that both operands are valid, non-empty, and have
the same type and that they either have the same shape or that one is
a scalar (rank 0) voxel array; it further determines the total count
of elements to be processed and determines the stride in each source
array, the rank of the result (of operating on the two arrays), and
the dimensions of the result.  The value returned is the total number
of elements to process, or zero if the arrays are invalid or
incompatible.  The case of two scalars is supported for uniformity,
but not efficient compared to the direct C calculation.

FIXME: Generalize to convert mixed types where possible, or to implement
operations that are defined with mixed types.
-------------------------------------------------------------------------*/

static inline
long                            /* Returns nmbr of elements, or zero */
vxli_setup_binpt(
    voxel_array *src1,          /* First array */
    voxel_array *src2,          /* Second array */
    int  *rank,                 /* (Out) Rank of result */
    long **dimen,               /* (Out) Dimensions of result */
    double **origin,            /* (Out) Origin of result */
    double **spacing,           /* (Out) Spacing of result */
    long *ss1,                  /* (Out) Stride in first array */
    long *ss2)                  /* (Out) Stride in second array */
{
    long count = 0;             /* Number of elements to be processed */
    voxel_array *srcx = 0;      /* First non-scalar array, or src1 */

    /* Check validity of first source array */
    if (src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0)
	ZReturnError(ERR_INIT, "Invalid or empty source array src1");

    /* Check validity of second source array */
    if (src2 == 0 || src2->magic != VXL_MAGIC || src2->data == 0)
	ZReturnError(ERR_INIT, "Invalid or empty source array src2");

    /* Check that source types are the same */
    if (src1->type != src2->type)
	ZReturnError(ERR_INCOMP, "Source types are different");

    /* Check if source shapes are compatible */
    if ( ! vxli_same_physical(src1, src2) )    ZReturnLibError();
    
    /* Find the non-scalar source, if any */
    srcx = src1;
    if ( src1->rank == 0 && src2->rank > 0)   srcx = src2;

    /* Set result parameters from the non-scalar source */
    *rank    = srcx->rank;
    *dimen   = srcx->dimen;
    *origin  = srcx->origin;
    *spacing = srcx->spacing;
    count    = vxli_count(srcx);

    /* Set strides to 0 / 1 for scalar / non-scalar source */
    *ss1 = (src1->rank == 0) ? 0 : 1;
    *ss2 = (src2->rank == 0) ? 0 : 1;

    return count;
}



/*----------------------------------------------------------------------------
vxli_standard_binop - Common core for a standard binary point operation

This function implements a common core for those binary point operations
which are consistent with the vxl_binop typedef and which are implemented
by BIPS functions consistent with the bips_binop typedef.
----------------------------------------------------------------------------*/

static inline 
int                                /* Returns non-zero for error */
vxli_standard_binop(
    bips_binop  *bips_fcn,         /* BIPS function to use */
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2)             /* Second source array */
{
    int rank;                      /* Rank of result */
    long *dimen = 0;               /* Dimensions of result */
    double *origin = 0;            /* Origin of result */
    double *spacing = 0;           /* Spacing of result */
    int type;                      /* Common type of source arrays */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    type = src1->type;
    if ( exim_sizeof_intype(type) == 0 )
	NZReturnError(ERR_UNDEF, "Invalid internal type");
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &origin, &spacing,
			     &ss1, &ss2);
    if (count <= 0)    NZReturnLibError();

    /* Allocate the destination image; copy origin and spacing */
    err = vxl_alloc_array(dest, type, rank, dimen);
    if (err)    NZReturnLibError();
    vxli_set_physical(dest, origin, spacing);

    /* Do the actual binary operation */
    err = bips_fcn(count, type, dest->data, ds, 
		   src1->data, ss1, src2->data, ss2);
    if (err != 0)    NZReturnLibError();

    return 0;
}



/*-------------------------------------------------------------------------
Standard binary point operations on voxel arrays

These functions implement various binary point operations on voxel
arrays that follow a standard pattern.  Each function computes the
elementwise element of two voxel arrays of the same type and generates
an output array of the same size and type; that is, dest = funct(src1,
src2) is computed for each voxel.  Either array may be a scalar array,
in which case that scalar value is used for all computations.  The
handling of domain and range errors is inherited from the underlying C
implementation.

ALGORITHM: These definitions could be reduced to one line apiece using
an appropriate macro, but that would make them too hard for etags (and
the ordinary programmer) to find and understand.
--------------------------------------------------------------------------*/

int vxl_and(voxel_array *dest, voxel_array *src1, voxel_array *src2)
{   return vxli_standard_binop(bips_and, dest, src1, src2);
}

int vxl_or(voxel_array *dest, voxel_array *src1, voxel_array *src2)
{   return vxli_standard_binop(bips_or, dest, src1, src2);
}

int vxl_xor(voxel_array *dest, voxel_array *src1, voxel_array *src2)
{   return vxli_standard_binop(bips_xor, dest, src1, src2);
}

int vxl_min(voxel_array *dest, voxel_array *src1, voxel_array *src2)
{   return vxli_standard_binop(bips_min, dest, src1, src2);
}

int vxl_max(voxel_array *dest, voxel_array *src1, voxel_array *src2)
{   return vxli_standard_binop(bips_max, dest, src1, src2);
}

int vxl_add(voxel_array *dest, voxel_array *src1, voxel_array *src2)
{   return vxli_standard_binop(bips_add, dest, src1, src2);
}

int vxl_sub(voxel_array *dest, voxel_array *src1, voxel_array *src2)
{   return vxli_standard_binop(bips_sub, dest, src1, src2);
}

int vxl_mul(voxel_array *dest, voxel_array *src1, voxel_array *src2)
{   return vxli_standard_binop(bips_mul, dest, src1, src2);
}

int vxl_div(voxel_array *dest, voxel_array *src1, voxel_array *src2)
{   return vxli_standard_binop(bips_div, dest, src1, src2);
}

int vxl_rem(voxel_array *dest, voxel_array *src1, voxel_array *src2)
{   return vxli_standard_binop(bips_rem, dest, src1, src2);
}

int vxl_pow(voxel_array *dest, voxel_array *src1, voxel_array *src2)
{   return vxli_standard_binop(bips_pow, dest, src1, src2);
}

int vxl_atan2(voxel_array *dest, voxel_array *src1, voxel_array *src2)
{   return vxli_standard_binop(bips_atan2, dest, src1, src2);
}

#ifdef HAVE_HYPOT
int vxl_hypot(voxel_array *dest, voxel_array *src1, voxel_array *src2)
{   return vxli_standard_binop(bips_hypot, dest, src1, src2);
}
#endif




/***************************************************************************
                  Standard Ternary Point Operations

The following ternary point operations on voxel arrays all fit a single
pattern and are implemented using a shared code for that pattern.

FIXME: Add other ternary operations such as mask and mix.
***************************************************************************/


/*------------------------------------------------------------------------
vxli_setup_terpt - Setup a binary point operation on voxel arrays

This function does some setup processing that is common to all ternary
point operations on voxel arrays (well, at least those that will
actually be done within bips), and is intended to make it easier to
permit one of the operands to be a scalar.  Specifically, this
function verifies that all three operands are valid, non-empty, and
have the same type and that they either have the same shape or that
one is a scalar (rank 0) voxel array; it further determines the total
count of elements to be processed and determines the stride in each
source array, the rank of the result (of operating on the three
arrays), and the dimensions of the result.  The value returned is the
total number of elements to process, or zero if the arrays are invalid
or incompatible.  The case of three scalars is supported for
uniformity, but not efficient compared to the direct C calculation.

FIXME: Generalize to convert mixed types where possible, or to implement
operations that are defined with mixed types.
-------------------------------------------------------------------------*/

static inline
long                            /* Returns nmbr of elements, or zero */
vxli_setup_terpt(
    voxel_array *src1,          /* First array */
    voxel_array *src2,          /* Second array */
    voxel_array *src3,          /* Third array */
    int  *rank,                 /* (Out) Rank of result */
    long **dimen,               /* (Out) Dimensions of result */
    double **origin,            /* (Out) Origin of result */
    double **spacing,           /* (Out) Spacing of result */
    long *ss1,                  /* (Out) Stride in first array */
    long *ss2,                  /* (Out) Stride in second array */
    long *ss3)                  /* (Out) Stride in third array */
{
    long count = 0;             /* Number of elements to be processed */
    voxel_array *srcx = 0;      /* First non-scalar array, or src1 */

    /* Check validity of first source array */
    if (src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0)
	ZReturnError(ERR_INIT, "Invalid or empty source array src1");

    /* Check validity of second source array */
    if (src2 == 0 || src2->magic != VXL_MAGIC || src2->data == 0)
	ZReturnError(ERR_INIT, "Invalid or empty source array src2");

    /* Check validity of third source array */
    if (src3 == 0 || src3->magic != VXL_MAGIC || src3->data == 0)
	ZReturnError(ERR_INIT, "Invalid or empty source array src3");

    /* Check that source types are the same */
    if ( src1->type != src2->type || src1->type != src3->type )
	ZReturnError(ERR_INCOMP, "Source types are different");

    /* Find the non-scalar source, if any */
    srcx = src1;
    if ( src2->rank > 0 )   srcx = src2;
    if ( src3->rank > 0 )   srcx = src3;

    /* Check that source shapes are compatible */
    if ( src1 != srcx && ! vxli_same_physical(src1, srcx) )
        ZReturnLibError();
    if ( src2 != srcx && ! vxli_same_physical(src2, srcx) )
        ZReturnLibError();
    if ( src3 != srcx && ! vxli_same_physical(src3, srcx) )
        ZReturnLibError();

    /* Set destination parameters from the non-scalar source */
    *rank    = srcx->rank;
    *dimen   = srcx->dimen;
    *origin  = srcx->origin;
    *spacing = srcx->spacing;
    count    = vxli_count(srcx);

    /* Set strides to 0 / 1 for scalar / non-scalar source */
    *ss1 = (src1->rank == 0) ? 0 : 1;
    *ss2 = (src2->rank == 0) ? 0 : 1;
    *ss3 = (src3->rank == 0) ? 0 : 1;

    return count;
}



/*----------------------------------------------------------------------------
vxli_standard_ternop - Common core for a standard ternary point operation

This function implements a common core for those ternary point operations
which are consistent with the vxl_ternop typedef and which are implemented
by BIPS functions consistent with the bips_ternop typedef.
----------------------------------------------------------------------------*/

static inline 
int                                /* Returns non-zero for error */
vxli_standard_ternop(
    bips_ternop *bips_fcn,         /* BIPS function to use */
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2,             /* Second source array */
    voxel_array *src3)             /* Third source array */
{
    int rank;                      /* Rank of result */
    long *dimen = 0;               /* Dimensions of result */
    double *origin = 0;            /* Origin of result */
    double *spacing = 0;           /* Spacing of result */
    int type;                      /* Common type of source arrays */
    long ds = 1, ss1, ss2, ss3;    /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    type = src1->type;
    if ( exim_sizeof_intype(type) == 0 )
	NZReturnError(ERR_UNDEF, "Invalid internal type");
    count = vxli_setup_terpt(src1, src2, src3, 
                             &rank, &dimen, &origin, &spacing,
			     &ss1, &ss2, &ss3);
    if (count <= 0)
        NZReturnLibError();

    /* Allocate the destination image; set origin and spacing */
    err = vxl_alloc_array(dest, type, rank, dimen);
    if (err)    NZReturnLibError();
    vxli_set_physical(dest, origin, spacing);

    /* Do the actual operation in BIPS */
    err = bips_fcn(count, type, dest->data, ds, 
		   src1->data, ss1, src2->data, ss2, src3->data, ss3);
    if (err != 0)    NZReturnLibError();

    return 0;
}


/* Fused multiply-add, or elementwise w = x*y + z */
int vxl_fma(voxel_array *dest, voxel_array *src1, voxel_array *src2,
            voxel_array *src3)
{   return vxli_standard_ternop(bips_fma, dest, src1, src2, src3);
}




/************************************************************************
                 Miscellaneous Point Operations

The following point (or elementwise) operations on voxel arrays do not
seem to follow any standard pattern.
************************************************************************/



/*-------------------------------------------------------------------------
vxl_cmove - Conditional move from one array to another

This function takes a destination, selector, and source array as
operands.  For each element of the selector that is non-zero, the
corresponding element of the source is copied to the corresponding
element of the destination.  The destination, selector, and source
must have the same rank, dimensions, origin, and spacing.  The
selector must be unsigned char; the destination and source may be any
type but must be the same type.  Note that this function is atypical
in that the destination must already exist and be compatible with the
source array.
------------------------------------------------------------------------*/

int                                /* Returns non-zero on error */
vxl_cmove(
    voxel_array *dest,             /* Destination array */
    voxel_array *sel1,             /* First source array: selector */
    voxel_array *src2)             /* Second source array: source */
{
    long ds = 1, ss1 = 1, ss2;     /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity of destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC || dest->data == 0)
	ZReturnError(ERR_INIT, "Invalid or empty destination array");
    if (dest->rank == 0)
        ZReturnError(ERR_INIT, "Operation invalid for a scalar destination");

    /* Check validity of first source array */
    if (sel1 == 0 || sel1->magic != VXL_MAGIC || sel1->data == 0)
	ZReturnError(ERR_INIT, "Invalid or empty selector array");
    if (sel1->type != INTP_UCHAR)
	ZReturnError(ERR_INIT, "Selector array must be unsigned char");
    if (sel1->rank == 0)
        ZReturnError(ERR_INIT, "Operation invalid for a scalar selector");

    /* Check validity of second source array */
    if (src2 == 0 || src2->magic != VXL_MAGIC || src2->data == 0)
	ZReturnError(ERR_INIT, "Invalid or empty source array");

    /* Check that arrays are compatible */
    if ( dest->type != src2->type )
	ZReturnError(ERR_INCOMP,
		     "Destination and source must be the same type");
    if ( ! vxli_same_physical(dest, sel1) )    ZReturnLibError();
    if ( ! vxli_same_physical(dest, src2) )    ZReturnLibError();

    /* Call BIPS to do the real work */
    count = vxli_count(dest);
    ss2 = (src2->rank > 0) ? 1 : 0;
    err = bips_cmove(count, dest->type, dest->data, ds,
		     sel1->data, ss1, src2->data, ss2);
    if (err != 0)    NZReturnLibError();

    return 0;
}



/*-------------------------------------------------------------------------
vxl_compare - Compare two voxel arrays

This function compares two voxel arrays element by element and generates
an usigned char voxel array of the same shape, each element of which is
one of three given values according to whether an element of the first
array is less than, equal to, or greater than the corresponding element
of the second array.  The two arrays may be of any non-complex type; 
either one may be a scalar, in which case that scalar is used in all the
comparisons.

See also vxl_equal, which provides only equality comparisons but
support complex types.

FIXME: Write a more general function which allows lsval, eqval, and
gtval to be arrays.
------------------------------------------------------------------------*/

int                                /* Returns non-zero on error */
vxl_compare(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2,             /* Second source array */
    unsigned char ltval,           /* Output value if less than */
    unsigned char eqval,           /* Output value if equal */
    unsigned char gtval)           /* Output value if greater than */
{
    int rank;                      /* Rank of result */
    long *dimen = 0;               /* Dimensions of result */
    double *origin = 0;            /* Origin of result */
    double *spacing = 0;           /* Spacing of result */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &origin, &spacing,
			     &ss1, &ss2);
    if (count <= 0)	NZReturnLibError();

    /* Check that operands are not complex */
    if ( exim_iscomplex(src1->type) || exim_iscomplex(src2->type) )
        NZReturnError(ERR_ILLOP, "Not defined for complex types");
    
    /* Allocate the destination image; copy origin and spacing */
    err = vxl_alloc_array(dest, INTP_UCHAR, rank, dimen);
    if (err)    NZReturnLibError();
    vxli_set_physical(dest, origin, spacing);

    /* Do the actual comparison */
    err = bips_compare(count, dest->data, ds, 
		       src1->type, src1->data, ss1, src2->data, ss2,
		       ltval, eqval, gtval);
    if (err != 0)    NZReturnLibError();

    return 0;
}



/*-------------------------------------------------------------------------
vxl_const - Fill voxel array with a constant value

This function sets all elements of a voxel array to a constant value.
The rank, dimensions, and type of the array must have been set
previously.  The value is specified as an anytype value and will be
converted to the type of the array.
--------------------------------------------------------------------------*/

int                                  /* Returns non-zero on error */
vxl_const(
    voxel_array *dest,               /* Destination */
    int vtype,                       /* Internal type of value to set to */
    anytype *value)                  /* Value to set to */
{
    int dtype = 0;                   /* Type of the destination array */
    long count;                      /* Number of elements in destination */
    int flag;

    /* Check validity of the destination */
    if ( dest == 0 || dest->magic != VXL_MAGIC || dest->data == 0 ||
	 exim_sizeof_intype((dtype = dest->type)) == 0 )
	NZReturnError(ERR_INIT, "Invalid destination array");
    count = vxli_count(dest);

    /* Cast and copy the converted value to each element of the destination */
    flag = bips_cast(count, dtype, dest->data, 1, vtype, value, 0);
    if (flag)    NZReturnLibError();

    return 0;
}



/*-------------------------------------------------------------------------
vxl_equal - Compare two voxel arrays for equality

This function compares two voxel arrays element by element and
generates an usigned char voxel array of the same shape, each element
of which is one of two given values according to whether an element of
the first array is equal or not equal to the corresponding element of
the second array.  The two arrays may be of any type, including
complex types; either one may be a scalar, in which case that scalar
is used in all the comparisons.

See also vxl_compare(), which provides less/greater comparisons but
does not support complex types.

FIXME: Write a more general function which allows eqval and neval to
be arrays?
------------------------------------------------------------------------*/

int                                /* Returns non-zero on error */
vxl_equal(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2,             /* Second source array */
    unsigned char eqval,           /* Output value if equal */
    unsigned char neval)           /* Output value if not equal */
{
    int rank;                      /* Rank of result */
    long *dimen = 0;               /* Dimensions of result */
    double *origin = 0;            /* Origin of result */
    double *spacing = 0;           /* Spacing of result */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Check validity and compatibility of source images */
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &origin, &spacing,
			     &ss1, &ss2);
    if (count <= 0)	NZReturnLibError();

    /* Allocate the destination image; copy origin and spacing */
    err = vxl_alloc_array(dest, INTP_UCHAR, rank, dimen);
    if (err)    NZReturnLibError();
    vxli_set_physical(dest, origin, spacing);

    /* Do the actual comparison */
    err = bips_eq(count, dest->data, ds, src1->type, src1->data, ss1,
                  src2->data, ss2, eqval, neval);
    if (err != 0)    NZReturnLibError();

    return 0;
}



/*------------------------------------------------------------------------
vxl_fftramp - Fill voxel array with a frequency ramp

This function sets each element of a voxel array equal to the FFT
frequency corresponding to one of its coordinates.  If the coordinate
cannot be represented in the voxel type, it will be converted from
(signed) long to the voxel type according to the usual C conversion
rules.  This operation is not permitted for empty or rank-0 arrays.
(It could be, but it's not particularly useful and would have to be
treated as a special case.)
-------------------------------------------------------------------------*/

int                            /* Returns non-zero on error */
vxl_fftramp(
    voxel_array *dest,         /* Destination array */
    int axis)                  /* Coordinate axis to store */
{
    int rank;                  /* Rank of the array */
    int type = 0;              /* Type of the destination */
    long *dimen;               /* Dimensions of the array */
    long nelem;                /* Number of elements per row */
    long nelem2;               /* Number of elements on selected axis */
    long coord[VXL_MAX_RANK];  /* Starting voxel of each row */
    unsigned char *row;        /* Location of a row */
    long n;                    /* Index over elements */
    long freq;                 /* FFT frequency */
    int  flag;
    
    /* Check validity of the destination */
    if ( dest == 0 || dest->magic != VXL_MAGIC || dest->data == 0 ||
	 dest->rank == 0 ||  vxli_count(dest) == 0 ||
	 exim_sizeof_intype((type = dest->type)) == 0 )
	NZReturnError(ERR_INIT, "Invalid destination array");
    rank  = dest->rank;
    dimen = dest->dimen;
    nelem = dimen[rank-1];

    /* Check that the requested type is not an unsigned type */
    if ( exim_isunsigned(type) )
        NZReturnError(ERR_INIT, "Not valid for unsigned types");

    /* Check validity of the axis */
    if ( axis < 0 || axis >= rank )
	NZReturnError(ERR_ILLOP, "Invalid axis specified");
    nelem2 = dimen[axis];

    /* Case 1: Ramping on the last axis */
    if (axis == rank-1) {

        /* Compute the first row */
	for (n = 0; n < nelem; n++) {
   	    freq = n;
	    if (freq > nelem/2)
  	        freq = n - nelem;
	    row  = bips_locate(type, dest->data, n);
	    flag = bips_cast(1, type, row, 0, INTP_LONG, &freq, 0);
	    if (flag)    NZReturnLibError();
	} 

        /* Copy to the *remaining* rows */
        vxli_firstrow(rank, coord);
	while ( vxli_nextrow(rank, dimen, coord) ) {
	    row = vxli_locate(dest, coord);
	    if (row == 0)    NZReturnLibError();
	    flag = bips_copy(nelem, type, row, 1, dest->data, 1);
	    if (flag)    NZReturnLibError();
	}
    }

    /* Case 2: Ramping on any other axis */
    else {

        /* Loop over the rows of the voxel array */
        vxli_firstrow(rank, coord);
	do {

	    /* Compute the appropriate value */
   	    freq = coord[axis];
	    if (freq > nelem2/2)
  	        freq = freq - nelem2;

	    /* Copy into each element of this row */
	    row = vxli_locate(dest, coord);
	    if (row == 0)    NZReturnLibError();
	    flag = bips_cast(nelem, type, row, 1, INTP_LONG, &freq, 0);
	    if (flag)    NZReturnLibError();
	
	} while ( vxli_nextrow(rank, dimen, coord) );
    } 

    return 0;
}



/*-------------------------------------------------------------------------
vxl_greater - Compare two voxel arrays for greater than

This function compares two voxel arrays element by element and
generates an usigned char voxel array of the same shape, each element
of which is one of two given values according to whether an element of
the first array is greater than or not greater than the corresponding
element of the second array.  The two arrays may be of any type,
except complex types; either one may be a scalar, in which case that
scalar is used in all the comparisons.

See also vxl_compare(), which provides three-way comparisons.

FIXME: Write a more general function which allows eqval and neval to
be arrays?
------------------------------------------------------------------------*/

int                                /* Returns non-zero on error */
vxl_greater(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2,             /* Second source array */
    unsigned char gtval,           /* Output value if greater than */
    unsigned char leval)           /* Output value if not greater than */
{
    int rank;                      /* Rank of result */
    long *dimen = 0;               /* Dimensions of result */
    double *origin = 0;            /* Origin of result */
    double *spacing = 0;           /* Spacing of result */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Not defined for complex types */
    if ( exim_iscomplex(src1->type) )
        NZReturnError(ERR_ILLOP, "Not defined for complex types");

    /* Check validity and compatibility of source images */
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &origin, &spacing,
			     &ss1, &ss2);
    if (count <= 0)	NZReturnLibError();

    /* Allocate the destination image; copy origin and spacing */
    err = vxl_alloc_array(dest, INTP_UCHAR, rank, dimen);
    if (err)    NZReturnLibError();
    vxli_set_physical(dest, origin, spacing);

    /* Do the actual comparison */
    err = bips_gt(count, dest->data, ds, src1->type, src1->data, ss1,
                  src2->data, ss2, gtval, leval);
    if (err != 0)    NZReturnLibError();

    return 0;
}



/*-------------------------------------------------------------------------
vxl_less - Compare two voxel arrays for less than

This function compares two voxel arrays element by element and
generates an usigned char voxel array of the same shape, each element
of which is one of two given values according to whether an element of
the first array is less than or not less than the corresponding
element of the second array.  The two arrays may be of any type,
except complex types; either one may be a scalar, in which case that
scalar is used in all the comparisons.

See also vxl_compare(), which provides three-way comparisons.

FIXME: Write a more general function which allows eqval and neval to
be arrays?
------------------------------------------------------------------------*/

int                                /* Returns non-zero on error */
vxl_less(
    voxel_array *dest,             /* Destination array */
    voxel_array *src1,             /* First source array */
    voxel_array *src2,             /* Second source array */
    unsigned char ltval,           /* Output value if less than */
    unsigned char geval)           /* Output value if not less than */
{
    int rank;                      /* Rank of result */
    long *dimen = 0;               /* Dimensions of result */
    double *origin = 0;            /* Origin of result */
    double *spacing = 0;           /* Spacing of result */
    long ds = 1, ss1, ss2;         /* Stride in each array */
    long count;                    /* Total elements to process */
    int err;

    /* Not defined for complex types */
    if ( exim_iscomplex(src1->type) )
        NZReturnError(ERR_ILLOP, "Not defined for complex types");

    /* Check validity and compatibility of source images */
    count = vxli_setup_binpt(src1, src2, &rank, &dimen, &origin, &spacing,
			     &ss1, &ss2);
    if (count <= 0)	NZReturnLibError();

    /* Allocate the destination image; copy origin and spacing */
    err = vxl_alloc_array(dest, INTP_UCHAR, rank, dimen);
    if (err)    NZReturnLibError();
    vxli_set_physical(dest, origin, spacing);

    /* Do the actual comparison */
    err = bips_lt(count, dest->data, ds, src1->type, src1->data, ss1,
                  src2->data, ss2, ltval, geval);
    if (err != 0)    NZReturnLibError();

    return 0;
}



/*-------------------------------------------------------------------------
vxl_logcomp_ulong - Log-compress an unsigned long volume image

This function compresses the dynamic range of an unsigned long volume
image using the transformation y = A*log(1+x), where A is chosen such
that the largest voxel value actually present in the image is converted
to the value 255; the compressed image is saved as unsigned char.

FIXME: Generalize to other input types.  Since the usual usage is to
prepare a wide-dynamic-range image for display, leaving the destination
as uchar should be fine.  But note that you must define what to do with
negative values in the source.
--------------------------------------------------------------------------*/

int                            /* Returns non-zero on error */
vxl_logcomp_ulong(
    voxel_array *dest,         /* Destination array; always uchar */
    voxel_array *src)          /* Source array; always ulong */
{
    unsigned long maxv;        /* Maximum value in source */
    double A;                  /* Scaling factor for compressed values */
    unsigned char *ddata;      /* Actual array of destination values */
    unsigned long *sdata;      /* Actual array of source values */
    long count;                /* Number of voxels */
    long i, err;

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || 
        src->data == 0 || src->type != INTP_ULONG)
	NZReturnError(ERR_INIT, "Invalid or empty source array");

    /* Allocate the destination argument; copy origin and spacing */
    err = vxl_alloc_array(dest, INTP_UCHAR, src->rank, &(src->dimen[0]));
    if (err)    NZReturnLibError();
    vxli_copy_physical(dest, src);

    /* Locate source and destination */
    count = vxli_count(src);
    ddata = dest->data;
    sdata = src->data;

    /* Determine the appropriate value of A */
    maxv = 0;
    for (i = 0; i < count; i++)
        if (maxv < sdata[i])
            maxv = sdata[i];
    A = 255.99/log(1.0+maxv);

    /* Transform the image */
    for (i = 0; i < count; i++)
	ddata[i] = floor(A*log(1.0001+(double)sdata[i]));

    return 0;
}



/*------------------------------------------------------------------------
vxl_lookup - Process volume image through a lookup table

This function takes each voxel in the source image and uses it as the
index into a lookup table, saving the result of the lookup as the
destination image.  The source image must be either unsigned char or
unsigned short but may be of any shape.  The table image must be rank
1 and contain at least as many entries as the largest voxel value
within the source image; but it (and thus the destination) may be of
any type.
-------------------------------------------------------------------------*/

int                            /* Returns non-zero on error */
vxl_lookup(
    voxel_array *dest,         /* Destination array */
    voxel_array *src,          /* Source array */
    voxel_array *table)        /* Lookup table */
{
    long count;                /* Total number of voxels per channel */
    long ncol = 1;             /* Number of channels */
    int nbytes;                /* Bytes per destination element */
    int rank = 1;              /* Rank of the destination */
    long dimen[VXL_MAX_RANK];  /* Size of the destination */
    long dts = 1;              /* Destination and table stride */
    unsigned char *dcol;       /* Destination column */
    unsigned char *tcol;       /* Table column */
    int i, err;

    /* Validate the source argument */
    if ( src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    if ( src->type != INTP_UCHAR && src->type != INTP_USHORT )
	NZReturnError(ERR_TYPE, 
                      "Source must be unsigned char or unsigned short");

    /* Validate the table array */
    if ( table == 0 || table->magic != VXL_MAGIC || table->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty lookup table");
    nbytes = exim_sizeof_intype(table->type);
    if (nbytes == 0)
	NZReturnError(ERR_UNDEF, "Invalid type code in lookup table");

    /* Determine destination dimensions */
    for (i = 0; i < src->rank; i++)
	dimen[i] = src->dimen[i];
    if ( table->rank == 1 ) {
	rank = src->rank;
	dts = ncol = 1;
    }
    else if ( table->rank == 2 ) {
	rank = src->rank + 1;
	dts = ncol = dimen[rank-1] = table->dimen[1];
    }
    else 
	NZReturnError(ERR_ILLOP, "Lookup table must be rank 1 or rank 2");

    /* Allocate the destination; copy origin and spacing */
    count = vxli_count(src);
    err = vxl_alloc_array(dest, table->type, rank, dimen);
    if (err)    NZReturnLibError();
    vxli_copy_physical(dest, src);

    /* Process in BIPS */
    for (i = 0; i < ncol; i++) {
	dcol = (unsigned char *)dest ->data + i*nbytes;
	tcol = (unsigned char *)table->data + i*nbytes;
	err  = bips_lookup(count, table->type, src->type, dcol, dts, 
			   src->data, 1, tcol, dts, table->dimen[0]);
	if (err != 0)    NZReturnLibError();
    }

    return 0;
}



/*------------------------------------------------------------------------
vxl_ramp - Fill voxel array with a linear ramp

This function sets each element of a voxel array equal to one of its
coordinates.  If the coordinate cannot be represented in the voxel
type, it will be converted from (signed) long to the voxel type
according to the usual C conversion rules.  This operation is not
permitted for empty or rank-0 arrays.  (It could be, but it's not
particularly useful and would have to be treated as a special case.)
-------------------------------------------------------------------------*/

int                            /* Returns non-zero on error */
vxl_ramp(
    voxel_array *dest,         /* Destination array */
    int axis)                  /* Coordinate axis to store */
{
    int rank;                  /* Rank of the array */
    int type = 0;              /* Type of the destination */
    long *dimen;               /* Dimensions of the array */
    long nelem;                /* Number of elements per row */
    long coord[VXL_MAX_RANK];  /* Starting voxel of each row */
    unsigned char *row;        /* Location of a row */
    anytype val;               /* Axis coord converted to dest type */
    int err;
    
    /* Check validity of the destination */
    if ( dest == 0 || dest->magic != VXL_MAGIC || dest->data == 0 ||
	 dest->rank == 0 ||  vxli_count(dest) == 0 ||
	 exim_sizeof_intype((type = dest->type)) == 0 )
	NZReturnError(ERR_INIT, "Invalid destination array");
    rank  = dest->rank;
    dimen = dest->dimen;
    nelem = dimen[rank-1];

    /* Check validity of the axis */
    if ( axis < 0 || axis >= rank )
	NZReturnError(ERR_ILLOP, "Invalid axis specified");

    /* Loop over the rows of the voxel array */
    vxli_firstrow(rank, coord);
    do {
	row = vxli_locate(dest, coord);
        if (row == 0)    NZReturnLibError();

	/* Case 1: Ramping on the last axis */
	if (axis == rank-1) {
	    err = bips_ramp(nelem, type, row, 1);
	}

	/* Case 2: Ramping on any other axis */
	else {
	    err  = bips_cast(1, type, &val, 0, INTP_LONG, &coord[axis], 0);
	    err += bips_copy(nelem, type, row, 1, &val, 0);
	}
	
	if (err)    NZReturnLibError();

    } 
    while ( vxli_nextrow(rank, dimen, coord) );

    return 0;
}



/*-------------------------------------------------------------------------
vxl_randi - Fill voxel array with random integers

This function fills a voxel array with newly generated random choices
from the integers 0, ..., N-1.  The rank, dimensions, and type of the
array must have been set previously.  This function is NOT thread-safe
if the default generator is used.
--------------------------------------------------------------------------*/

int                                  /* Returns non-zero on error */
vxl_randi(
    voxel_array *dest,               /* Destination */
    void *state,                     /* State vector of random generator */
    int n)                           /* Number of available choices */
{
    int type = 0;                    /* Type of the destination array */
    long count;                      /* Number of elements in destination */
    int nbytes = 0;                  /* Number of bytes per element */

    /* Check validity of the destination */
    if ( dest == 0 || dest->magic != VXL_MAGIC || dest->data == 0 ||
	 (nbytes = exim_sizeof_intype((type = dest->type))) == 0 )
	NZReturnError(ERR_INIT, "Invalid destination array");
    count = vxli_count(dest);

    /* FIXME: For now, support only arrays of type int */
    if ( dest->type != INTP_INT )
        NZReturnError(ERR_UNIMP, "Only int is currently supported");

    /* Fill the destination with random values */
    state = vxli_ran1state(state);
    if (state == 0)    NZReturnLibError();
    ran1_choose_array(count, dest->data, 1, state, n);

    return 0;
}



/*-------------------------------------------------------------------------
vxl_randn - Fill voxel array with normal random variates

This function fills a voxel array of type double with newly generated
normal random variates with a given mean and variance.  The type,
rank, and dimensions of the array must have been set previously.  This
function is NOT thread-safe if the default generator is used.

FIXME: Compute by rows for better locality.
--------------------------------------------------------------------------*/


int                                  /* Returns non-zero on error */
vxl_randn(
    voxel_array *dest,               /* Destination */
    void *state,                     /* State vector of random generator */
    double mean,                     /* Desired mean */
    double sdev)                     /* Desired standard deviation */
{
    long count;                      /* Number of elements in destination */
    void *data;                      /* Data location in the destination */
    int flag;

    /* Check validity of the destination */
    if ( dest == 0 || dest->magic != VXL_MAGIC || dest->data == 0 )
	NZReturnError(ERR_INIT, "Invalid destination array");
    if ( dest->type != INTP_DOUBLE )
        NZReturnError(ERR_INIT, "Only type double is permitted");
    count = vxli_count(dest);
    data  = dest->data;

    /* Fill the destination with standard normal random variates */
    /* FIXME: Add error check on ran1_normal_array? */
    state = vxli_ran1state(state);
    if (state == 0)    NZReturnLibError();
    ran1_normal_array(count, data, 1, state);

    /* Adjust to desired mean and standard deviation */
    if (sdev != 1.0) {
        flag = bips_mul_set(count, INTP_DOUBLE, data, 1, &sdev, 0);
        if (flag)    NZReturnLibError();
    }
    if (mean != 0.0) {
        flag = bips_add_set(count, INTP_DOUBLE, data, 1, &mean, 0);
        if (flag)    NZReturnLibError();
    }

    return 0;
}



/*-------------------------------------------------------------------------
vxl_randu - Fill voxel array with uniform random variates

This function fills a voxel array of type double with newly generated
uniform random variates in the open interval (a,b).  The rank,
dimensions, and type of the array must have been set previously.  Note
that the lower and upper limits of the range could possibly be reached
by roundoff error, although this should never happen for the default
values a = 0.0 and b = 1.0; This function is NOT threadsafe if the
default generator is used.

FIXME: Compute by rows for better locality.
--------------------------------------------------------------------------*/

int                                  /* Returns non-zero on error */
vxl_randu(
    voxel_array *dest,               /* Destination */
    void *state,                     /* State vector of random generator */
    double a,                        /* Lower limit of range */
    double b)                        /* Upper limit of range */
{
    long count;                      /* Number of elements in destination */
    void *data=0;                    /* Data location in destination */
    double width;                    /* Width of desired range */
    int flag;

    /* Check validity of the destination */
    if ( dest == 0 || dest->magic != VXL_MAGIC || dest->data == 0 )
	NZReturnError(ERR_INIT, "Invalid destination array");
    if ( dest->type != INTP_DOUBLE )
        NZReturnError(ERR_INIT, "Only type double is permitted");
    count = vxli_count(dest);
    data  = dest->data;

    /* Fill the destination with random values */
    state = vxli_ran1state(state);
    if (state == 0)    NZReturnLibError();
    ran1_double_array(count, data, 1, state);

    /* Adjust to desired range */
    width = b - a;
    if (width != 1.0) {
        flag = bips_mul_set(count, INTP_DOUBLE, data, 1, &width, 0);
        if (flag)    NZReturnLibError();
    }
    if (a != 0.0) {
        flag = bips_add_set(count, INTP_DOUBLE, data, 1, &a, 0);
        if (flag)    NZReturnLibError();
    }

    return 0;
}



/*-------------------------------------------------------------------------
vxl_scale - Scale and offset an image of any type

This function casts each element of a voxel array to double or double
complex, multiplies it by a constant gain, adds a constant bias, and
casts it to some specified type, rounding and limiting as necessary.
The algorithm used here makes better use of cache and is more
efficient than the separate calculations for large arrays of integral
type.  If either of the gtype or gain arguments is zero, or the gain
constant is 1.0, the multiplication is skipped.  If either of the
btype or bias arguments is zero, or then the bias constant is 0.0, the
addition is skipped.

FIXME: A common application is to scale to the range of an integral
type, in which case specifying min and max would be more convenient.
Write a separate function to support this application?

FIXME: Does this replace the vxl_scale_to_uchar() function?  Or the
bips_scale_to_uchar() function?
---------------------------------------------------------------------------*/

int                               /* Returns non-zero on error */
vxl_scale(
    int dtype,                    /* Desired type for result */
    voxel_array *dest,            /* Destination array */
    voxel_array *src,             /* Source array */
    int gtype,                    /* Type of gain constant */
    void *gain,                   /* Constant gain to multiply by */
    int btype,                    /* Type of bias constant */
    void *bias)                   /* Constant bias to add */
{
    int stype;                    /* Type of source data */
    int wtype = INTP_DOUBLE;      /* Working type for computations (default) */
    int rank;                     /* Rank of source */
    long *dimen;                  /* Dimensions of source */
    long coord[VXL_MAX_RANK];     /* Coordinates of current row */
    long nelem;                   /* Number of elements in a row */
    unsigned char *srow;          /* Addr of current row in source */
    unsigned char *drow;          /* Addr of current row in destination */
    unsigned char *buffer;        /* Working memory for one row */
    anytype gv, bv;               /* Gain and bias in working type */
    int flag;                     /* Error flag */

    /* Validate the source argument */
    if ( src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    stype = src->type;
    rank  = src->rank;
    dimen = src->dimen;
    nelem = dimen[rank-1];

    /* Work in complex if source, gain, or bias is complex */
    wtype = INTP_DOUBLE;
    if ( exim_iscomplex(stype) || 
	 exim_iscomplex(gtype) ||
	 exim_iscomplex(btype) )
        wtype = INTP_DCOMPLEX;

    /* Convert gain and bias to working type */
    if ( gtype && gain ) {
        flag  = bips_cast(1, wtype, &gv, 0, gtype, gain, 0);
	if (flag)    NZReturnLibError();
    }
    if ( btype && bias ) {
        flag = bips_cast(1, wtype, &bv, 0, btype, bias, 0);
	if (flag)    NZReturnLibError();
    }

    /* Treat gain == 1.0 and bias == 0.0 as no-ops */
    if (wtype == INTP_DOUBLE) {
        if ( gtype && gain && gv.as_double == 1.0 )
	    gain = 0;
	if ( btype && bias && bv.as_double == 0.0 )
	    bias = 0;
    }
#ifdef HAVE_COMPLEX_H
    else {
        if ( gtype && gain && gv.as_dcomplex == 1.0 )
	    gain = 0;
	if ( btype && bias && bv.as_dcomplex == 0.0 )
	    bias = 0;
    }
#endif

    /* Allocate the destination; copy origin and spacing */
    flag = vxl_alloc_array(dest, dtype, rank, dimen);
    if (flag)    NZReturnLibError();
    vxli_copy_physical(dest, src);

    /* Allocate buffer for one row in working type */
    buffer = mallock(nelem*exim_sizeof_intype(wtype));
    if (buffer == 0)    NZReturnLibError();

    /* Loop over rows of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Locate source and destination rows */
	srow = vxli_locate(src,  coord);
	drow = vxli_locate(dest, coord);
	if (srow == 0 || drow == 0)    NZReturnLibError();

	/* Cast source to working type */
	flag = bips_cast(nelem, wtype, buffer, 1, stype, srow, 1);
	if (flag != 0)    NZReturnLibError();

	/* Multiply by gain and add bias */
	if (gain) {
	    flag = bips_mul_set(nelem, wtype, buffer, 1, &gv, 0);
	    if (flag != 0)    NZReturnLibError();
	}
	if (bias) {
	    flag = bips_add_set(nelem, wtype, buffer, 1, &bv, 0);
	    if (flag != 0)    NZReturnLibError();
	}

	/* Cast to destination type */
	flag = bips_round_cast(nelem, dtype, drow, 1, wtype, buffer, 1);
	if (flag != 0)    NZReturnLibError();

    }
    while( vxli_nextrow(rank, dimen, coord) );

    /* Free working memory and return */
    free(buffer);
    return 0;
}
