/****************************************************************************
errm.c - Error manager

Author:  Paul Hughett

Copyright (c) 1998-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file provides a standard interface for reporting program errors
within the Pyvox package written at the Brain Behavior Laboratory of
the University of Pennsylvania.

The current error manager interface uses the functions preset_error()
or set_error() to declare an error in the function that detects it,
and get_error() to interpret that error in the calling function.  An
error description includes a numeric code specifying the general type
of error, and a string describing the error in more details.  The
numeric code and descriptive string are stored in thread-local
buffers, so that the caller can ignore the messages without causing a
memory leak.  (BUT NOTE that the thread-safe version has not yet been
implemented.)  The routine detecting the error is expected to call
preset_error() or set_error() AND return a flag of some sort to alert
the caller to the error.  Two macros ReturnError() and XReturnError()
are provided to set the error number and string and return either the
error code or zero.

FIXME: Generalize for use in other packages and programs?

FIXME: Some other reasonable responses to an error are to do a
longjump to the main command loop to try another interactive command,
or to convert to a Python exception and unwind the stack.  Consider
adding the capability for these.  More generally, allow the programmer
to define his own error handlers.

FIXME:  Include a machine-interpretable error tag to indicate the error
type to some recovery mechanism?

FIXME: A thought for obviating the distinction between ordinary and Pyvox
errors.  Provide a hook here for Pyvox to set the error message (and code)
appropriately for themselves.
*****************************************************************************/

static const char rcsid[] = "$Id: errm.c,v 1.12 2005/12/29 18:33:14 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "errm.h"


#define MAXERRMSG  255          /* Maximum length of error message */



/************************************************************************
              New Error Management Interface
*************************************************************************/


/*------------------------------------------------------------------------
error buffer - Storage for reported errors

The errm_action flag indicates what should happen when an error is reported
via set_error.  The defined options are shown below; the default is 1.
(But note that Pyvox changes the action to 0 so that it can convert all
errors to Python exceptions.)

    0 => set error code and message; then continue
    1 => set error code and message; then call exit()
    2 => set error code and message; then call abort()

The errm_warning flag indicates what should be done when an error is
reported via set_warning.  The defined options are shown below; the default
is 1.

    0 => silently ignore warning
    1 => print message on stderr and continue
    2 => convert to an error

FIXME: Make thread-local (except the action flags).
------------------------------------------------------------------------*/

static const char *errm_context = 0;      /* Error context */
static int  errm_code = 0;                /* Numeric error code */
static char errm_msg[MAXERRMSG+1] = "";   /* Error message */
static int  errm_action = 1;              /* Action to take on error? */
static int  errm_warning = 1;             /* Action to take on warning? */



/*-----------------------------------------------------------------------
errm_panic - Report a irrecoverable error

This function reports an irrecoverable error and either aborts the
program or calls exit(EXIT_FAILURE).  It should be used only for
errors from which it is not possible to safely continue (e.g. a buffer
overrun), or for assertion failures in functions that have no other
way to signal an error.

The name errm_panic() is used in preference to plain panic() because
the latter is used in Tcl and perhaps other packages and we don't want
to collide with it.
-----------------------------------------------------------------------*/

void
errm_panic(
   const char *message,       /* Message as a format string */
   ...)                       /* Any arguments needed */
{
    va_list ap;

    va_start(ap, message);
    fprintf(stderr, "PANIC: ");
    vfprintf(stderr, message, ap);
    fprintf(stderr, "\n");
    va_end(ap);
    if (errm_action == 2)
	abort();
    else
	exit(EXIT_FAILURE);
}



/*------------------------------------------------------------------------
set_error_action - Enable or disable abort on error

This function sets an internal flag indicating what action should be taken
when set_error() is called.  The supported options are

   0  =>  Set error code and message, then continue;
   1  =>  Print error message and call exit(EXIT_FAILURE);
   2  =>  Print error message and call abort() to dump core.

Codes outside this range are treated as if 1 was used.  The internal
flag is shared over all threads and is initialized to 1; this
reproduces the existing exit-on-failure behavior in the command-line
programs.

Calling abort() will normally cause a core dump for debugging, if this
is enabled in the shell.
-------------------------------------------------------------------------*/

void
set_error_action(
    int action)               /* Desired error response */
{
    if (action < 0 || action > 2)
        errm_action = 1;
    else
        errm_action = action;

    return;
}



/*------------------------------------------------------------------------
set_warning_action - Specify action to take on a warning

This function sets an internal flag indicating what action should be taken
when set_warning() is called.  The supported options are

   0  =>  Silently ignore warning
   1  =>  Print error message on stderr and continue
   2  =>  Convert to an error

Codes outside this range are treated as if 1 was used.  The internal
flag is shared over all threads and is initialized to 1
-------------------------------------------------------------------------*/

void
set_warning_action(
    int action)               /* Desired error response */
{
    if (action < 0 || action > 2)
        errm_warning = 1;
    else
        errm_warning = action;

    return;
}



/*------------------------------------------------------------------------
set_error_context - Remember the error context

This function sets an internal variable to the address of a string
giving the error context; this context string will be printed out as
part of the error message if set_error is invoked.  The pointer
provided must remain valid until either clear_error_context() or
set_error() is called.

As an example of why the context string is useful, the Pyvox extension
to Python sets the error context to the name of the Pyvox function
before it calls the Voxel Kit library; thus, error messages generated
by the Voxel Kit are labelled with the Pyvox function known to the
Python programmer, rather than the Voxel Kit function which is of
interest only to the developers of the Pyvox extension.
------------------------------------------------------------------------*/

void
set_error_context(
    const char *context)            /* Error context */
{
    errm_context = context;
    return;
}



/*------------------------------------------------------------------------
clear_error_context - Forget the error context

This function clears the internal variable which gives the error
context.
--------------------------------------------------------------------------*/

void
clear_error_context(void)
{
    errm_context = 0;
    return;
}



/*-----------------------------------------------------------------------
preset_error - Set error code and message before continuing

This function allows the function detecting an error to set the
numeric error code and descriptive string for later inspection by the
calling function.  It returns the numeric code that has been set.  The
recommended set of codes are defined in errm.h.  Unlike set_error(),
this function always continues and thus is suitable for argument
checking functions that find potential errors when multiple types of
arguments are suitable.

NOTE: The vsnprintf() function is used if available to prevent
overruning the error message buffer; if it's not available, we use
plain vsprintf and panic afterwards if the buffer has been overrun.

FIXME: Current version is not thread-safe.
------------------------------------------------------------------------*/

int                              /* Returns the numeric error code */
preset_error(
    int errcode,                 /* Numeric error code */
    const char *message,         /* Error message, as a format string */
    ...)                         /* Any values needed for format string */
{
    va_list ap;
    int nchar = 0;

    /* Record the error code */
    errm_code = errcode;

    /* Report the context (if any) in the error message */
    if (errm_context != 0)
        nchar = sprintf(errm_msg, "%s: ", errm_context);

    /* Record the error message provided by the caller */
    va_start(ap, message);
#ifdef HAVE_VSNPRINTF
    nchar += vsnprintf(errm_msg+nchar, MAXERRMSG-nchar-1, message, ap);
#else
    nchar += vsprintf(errm_msg+nchar, message, ap);
    if (nchar > MAXERRMSG-1)
        errm_panic("Overrun in error message buffer");
#endif
    va_end(ap);

    /* Add an end of line */
    strcat(errm_msg, "\n");
    nchar += 1;

    /* Then return the error code */
    return errcode;
}




/*------------------------------------------------------------------------
set_error - Set error code and message

This function allows the function detecting an error to set the
numeric error code and descriptive string for later inspection by the
calling function.  If enabled by an earlier call to
set_error_action(), it will then immediately exit the program, or
abort and dump core.  Otherwise, it returns the numeric code that has
been set.  See also preset_error() for a version that never exits or
aborts.

The recommended set of error codes are defined in errm.h.

NOTE: The vsnprintf() function is used if available to prevent
overruning the error message buffer; if it's not available, we use
plain vsprintf and panic afterwards if the buffer has been overrun.

FIXME: Current version is not thread-safe.
-------------------------------------------------------------------------*/

int                              /* Returns the numeric error code */
set_error(
    int errcode,                 /* Numeric error code */
    const char *message,         /* Error message, as a format string */
    ...)                         /* Any values needed for format string */
{
    va_list ap;
    int nchar = 0;

    /* Record the error code */
    errm_code = errcode;

    /* Report the context (if any) in the error message */
    if (errm_context != 0)
        nchar = sprintf(errm_msg, "%s: ", errm_context);

    /* Record the error message provided by the caller */
    va_start(ap, message);
#ifdef HAVE_VSNPRINTF
    nchar += vsnprintf(errm_msg+nchar, MAXERRMSG-nchar-1, message, ap);
#else
    nchar += vsprintf(errm_msg+nchar, message, ap);
    if (nchar > MAXERRMSG-1)
        errm_panic("Overrun in error message buffer");
#endif
    va_end(ap);

    /* Add an end of line */
    strcat(errm_msg, "\n");
    nchar += 1;

    /* Take the previously requested error action, if any */
    if (errcode != 0 && errm_action != 0) {
        fprintf(stderr, errm_msg);
        if (errm_action == 1)
            exit(EXIT_FAILURE);
        else if (errm_action == 2)
            abort();
    }

    /* Otherwise just return the error code */
    return errcode;
}



/*------------------------------------------------------------------------
set_warning - Set code and message for a warning

This function allows the function detecting an warning to set the numeric
error code and descriptive string for later inspection by the calling
function.  It returns the numeric code that has been set.  The recommended
set of codes are defined in errm.h.

This function returns zero for a silent or simply reported warning; it
returns the error code for a warning converted to an error and the caller
must be prepared to handle this case.

NOTE: The vsnprintf() function is used if available to prevent
overruning the error message buffer; if it's not available, we use
plain vsprintf and panic afterwards if the buffer has been overrun.

FIXME: Current version is not thread-safe.
-------------------------------------------------------------------------*/

int                              /* Returns the numeric error code */
set_warning(
    int errcode,                 /* Numeric error code */
    const char *message,         /* Error message, as a format string */
    ...)                         /* Any values needed for format string */
{
    va_list ap;
    int nchar = 0;

    /* Record the error code */
    errm_code = errcode;

    /* Report the context (if any) in the error message */
    if (errm_context != 0)
        nchar = sprintf(errm_msg, "%s: ", errm_context);

    /* Record the error message provided by the caller */
    va_start(ap, message);
#ifdef HAVE_VSNPRINTF
    nchar += vsnprintf(errm_msg+nchar, MAXERRMSG-nchar-1, message, ap);
#else
    nchar += vsprintf(errm_msg+nchar, message, ap);
    if (nchar > MAXERRMSG-1)
        errm_panic("Overrun in error message buffer");
#endif
    va_end(ap);

    /* Add an end of line */
    strcat(errm_msg, "\n");
    nchar += 1;

    /* Take the action previously specified for warnings */
    if (errcode == 0 || errm_warning == 0)
        return 0;                      /* 0 => silently ignore warnings */
    else if (errm_warning == 1) {
        fprintf(stderr, errm_msg);
        return 0;                      /* 1 => print message and continue */
    }
    else 
        return errcode;                /* 2 => return as error */
}



/*-----------------------------------------------------------------------
get_error - Get error code and message

This function returns the error code and sets a pointer to the message
that were set by set_error.  The pointer to the message is guaranteed
to remain valid only until set_error is called again (within the same
thread).  Trying to free or modify the message string itself is
forbidden.

FIXME: Make thread-safe.
-------------------------------------------------------------------------*/

int                             /* Returns error code */
get_error(
    char **message)             /* Pointer to error message */
{
    if (message != 0)
        *message = errm_msg;
    return errm_code;
}



/*-----------------------------------------------------------------------
get_error_message - Get error message

This function returns a pointer to the error message set by set_error.
The pointer to the message is guaranteed to remain valid only until
set_error is called again (within the same thread).  Trying to free or
modify the message string itself is forbidden.

FIXME: Make thread-safe.
------------------------------------------------------------------------*/

const char *get_error_message(void)
{
    return errm_msg;
}
