/****************************************************************************
matrix.c - Matrix calculations on voxel arrays

Author:  Paul Hughett

Copyright (c) 2001-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file contains functions for doing matrix calculations on voxel
arrays, mostly using the LAPACK and BLAS Fortran libraries.  It is
part of the Voxel Kit, a set of general-purpose functions for volume
image processing, with a bias toward medical imaging applications.

The intention is to provide a set of matrix operations which wrap the
BLAS and LAPACK subroutines at a level similar to Matlab; that is, we
emphasize ease of use rather than blazing performance.  The assumption
here is that matrix operations are used relatively little compared to
image processing operations and thus do not require high performance;
users for whom this assumption is not true may wise to write their 
own Pyvox methods to call BLAS and LAPACK directly.

The contents of this file are still somewhat experimental.  It's not
yet entirely clear how to portably interface between C and Fortran and
how to package the numerical linear algebra operations for use by
Voxel Kit users.

Tentative policy: Make these functions simple and natural to use, even
if it takes extra time to transpose or otherwise convert from C to
Fortran.  

Tentative policy: Assume that long <-> integer, float <-> real, and
double <-> double precision, but arrange for compile errors if these
are false.

FIXME: Complete the minimal set of matrix operations: transpose
(generalized for n-dimen arrays); least squares; pseudoinverse; 
and ???

FIXME: Include these functions in voxel.h or add a new header matrix.h?
For the moment, leave them in voxel.h while I decide if this is the
right approach.

FIXME:  Identify common validity checking steps and isolate as functions
or macros.  Q:  What are the efficiency tradeoffs between the two?
*****************************************************************************/

static const char rcsid[] = "$Id: matrix.c,v 1.27 2006/01/23 21:22:54 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_TGMATH_H
#include <tgmath.h>
#else
#include <math.h>
#endif

#include "errm.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"
#include "bips.h"
#include "clap.h"          /* Needed only for BLAS and LAPACK */



/*-------------------------------------------------------------------------
Dummy MAIN__

On some platforms, the native LAPACK and BLAS libraries have been
translated by f2c and require libf2c (sometimes libg2c), which in turn
expects to find a function MAIN__ which is the Fortran main program.
We have the main program in C, but must still provide this function; but
it aborts the program if it is ever called.
---------------------------------------------------------------------------*/

int MAIN__(void);       /* A prototype, just to placate fussy compilers */

int
MAIN__()
{
    fprintf(stderr, "Called the dummy Fortran MAIN program\n");
    abort();
}



/*-------------------------------------------------------------------------
matrix_check - Verify that voxel array represents a matrix
-------------------------------------------------------------------------*/

static inline int
matrix_check(
    voxel_array *matrix)
{
    if ( matrix == 0 || matrix->magic != VXL_MAGIC ||
	 matrix->rank != 2 || matrix->data == 0 )
	return 0;
    if ( matrix->type != INTP_FLOAT    &&
         matrix->type != INTP_DOUBLE   &&
         matrix->type != INTP_FCOMPLEX &&
         matrix->type != INTP_DCOMPLEX )
	return 0;
    return 1;
}



/*-------------------------------------------------------------------------
matrix_check_square - Verify that voxel array represents a square matrix
-------------------------------------------------------------------------*/

static inline int
matrix_check_square(
    voxel_array *matrix)
{
    if ( matrix == 0 || matrix->magic != VXL_MAGIC ||
	 matrix->rank != 2 || matrix->data == 0 )
	return 0;
    if ( matrix->type != INTP_FLOAT    &&
         matrix->type != INTP_DOUBLE   &&
         matrix->type != INTP_FCOMPLEX &&
         matrix->type != INTP_DCOMPLEX )
	return 0;
    if ( matrix->dimen[0] != matrix->dimen[1] )
	return 0;
    return 1;
}



/*-------------------------------------------------------------------------
Generate single-precision real functions from generic functions
--------------------------------------------------------------------------*/

#define   ITYPE 1
#include "matrix1.c"



/*--------------------------------------------------------------------------
Generate double-precision real functions from generic functions
---------------------------------------------------------------------------*/

#undef  ITYPE
#define ITYPE 2
#include "matrix1.c"



/*--------------------------------------------------------------------------
Generate single-precision complex functions from generic functions
---------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H
#undef  ITYPE
#define ITYPE 3
#include "matrix1.c"
#endif


/*--------------------------------------------------------------------------
Generate double-precision complex functions from generic functions
---------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H
#undef  ITYPE
#define ITYPE 4
#include "matrix1.c"
#endif


/*--------------------------------------------------------------------------
matrix_determinant - Determinant of a voxel array

This function computes the determinant of a voxel array considered as a
matrix.
---------------------------------------------------------------------------*/

int                            /* Returns nonzero on error */
matrix_determinant(
    int *type,                 /* Data type of determinant */
    void *value,               /* Value of determinant */
    voxel_array *matrix)       /* Matrix to take determinant of */
{
    int flag;

    /* Switch on array type to computational function */
    switch (matrix->type) {

    case INTP_FLOAT:
        flag = matrix_sdet(type, value, matrix);
        break;

    case INTP_DOUBLE:
        flag = matrix_ddet(type, value, matrix);
        break;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
        flag = matrix_cdet(type, value, matrix);
        break;

    case INTP_DCOMPLEX:
        flag = matrix_zdet(type, value, matrix);
        break;
#endif

    default:
	NZReturnError(ERR_ILLOP, "Matrix type must be floating-point");
    }
    
    return flag;
}



/*--------------------------------------------------------------------------
matrix_multiply - Product of matrices

This function computes the product of two voxel arrays considered as
matrices.  A voxel array of rank 1 given as the second factor is treated
as a column vector.

FIXME: Does not support mixed types.  See matrix_xmul for other issues.
---------------------------------------------------------------------------*/

int                            /* Returns error flag */
matrix_multiply(
    voxel_array *product,      /* Product of matrices */
    voxel_array *amatrix,      /* Left factor */
    voxel_array *bmatrix)      /* Right factor */
{
    int flag = 0;

    /* Switch on array type to computational function */
    switch (amatrix->type) {

    case INTP_FLOAT:
        flag = matrix_smul(product, amatrix, bmatrix);
        break;

    case INTP_DOUBLE:
        flag = matrix_dmul(product, amatrix, bmatrix);
        break;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
        flag = matrix_cmul(product, amatrix, bmatrix);
        break;

    case INTP_DCOMPLEX:
        flag = matrix_zmul(product, amatrix, bmatrix);
        break;
#endif

    default:
	flag = set_error(ERR_ILLOP, "Matrix type must be floating-point");
        break;
    }
    
    return flag;
}



/*-----------------------------------------------------------------------
matrix_cholesky - Cholesky decomposition of a matrix

This function computes the Cholesky decomposition of a symmetric or
Hermitian positive definite square matrix represented as a voxel array
of floating-point type.  That is, given a matrix A, it returns the
lower diagonal matrix L such that A = L*L**T.  If the matrix is not
positive definite, a non-zero value is returned and the destination
matrix is unchanged.  The symmetric of the input matrix is NOT
verified.
-------------------------------------------------------------------------*/

int                              /* Returns non-zero if error */
matrix_cholesky(
    voxel_array *factor,         /* Lower triangular factor of matrix */
    voxel_array *matrix)         /* Matrix to find factor of */
{
    int flag;

    /* Switch on matrix data type */
    switch (matrix->type) {

    case INTP_FLOAT:
        flag = matrix_schol(factor, matrix);
        break;

    case INTP_DOUBLE:
        flag = matrix_dchol(factor, matrix);
        break;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
        flag = matrix_cchol(factor, matrix);
        break;

    case INTP_DCOMPLEX:
        flag = matrix_zchol(factor, matrix);
        break;
#endif

    default: 
	flag = set_error(ERR_ILLOP, "Matrix type must be floating-point");
        break;
    }

    return flag;
}



/*-----------------------------------------------------------------------
matrix_svd - singular value decomposition of a matrix

This function computes the singular value decomposition of a matrix
represented as a voxel array of floating-point type.  That is, given a
matrix A with dimensions (m,n), it returns an m*m orthogonal or
Hermitian matrix U, an n*n orthogonal or Hermitian matrix VT, and a
vector S of length min(m,n) such that A = U*diag(S)*VT.
-------------------------------------------------------------------------*/

int                           /* Returns non-zero if error */
matrix_svd(
    voxel_array *left,        /* Matrix of left singular vectors */
    voxel_array *values,      /* Vector of singular values */
    voxel_array *right,       /* Transposed matrix of right singular vectors */
    voxel_array *matrix)      /* Matrix to decompose */
{
    int flag;

    /* Switch on matrix data type */
    switch (matrix->type) {

    case INTP_FLOAT:
        flag = matrix_ssvd(left, values, right, matrix);
        break;

    case INTP_DOUBLE:
        flag = matrix_dsvd(left, values, right, matrix);
        break;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
        flag = matrix_csvd(left, values, right, matrix);
        break;

    case INTP_DCOMPLEX:
        flag = matrix_zsvd(left, values, right, matrix);
        break;
#endif

    default: 
	flag = set_error(ERR_ILLOP, "Matrix type must be floating-point");
        break;
    }

    return flag;
}



/*--------------------------------------------------------------------------
matrix_eigen_symmetric - Solve real symmetric eigensystem

This function computes the eigenvalues and eigenvectors of a voxel
array considered as a real symmetric matrix or a complex Hermitian
matrix.  The eigenvalues are returned as a rank-1 voxel array.  The
eigenvectors are returned as a square rank-2 voxel array, the rows of
which are the eigenvectors.  The matrix of eigenvectors is guaranteed
to be orthogonal but not necessarily to be proper. (If a proper
rotation is necessary, use matrix_principal_axes instead.)
---------------------------------------------------------------------------*/

int                               /* Returns non-zero on error */
matrix_eigen_symmetric(
    voxel_array *values,          /* Vector of eigenvalues */
    voxel_array *vectors,         /* Matrix of eigenvectors */
    voxel_array *matrix)          /* Matrix to solve */
{
    /* Switch on array type to computational function */
    switch (matrix->type) {

    case INTP_FLOAT:
        matrix_ssyev(values, vectors, matrix);
        break;

    case INTP_DOUBLE:
        matrix_dsyev(values, vectors, matrix);
        break;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
        matrix_csyev(values, vectors, matrix);
        break;

    case INTP_DCOMPLEX:
        matrix_zsyev(values, vectors, matrix);
        break;
#endif

    default:
	NZReturnError(ERR_ILLOP, "Matrix type must be floating-point");
    }

    return 0;
}



/*----------------------------------------------------------------------
matrix_get_diag - Extract diagonal elements of a matrix

This function extracts the diagonal elements of a rectangular matrix
and returns them as a vector of the same type as the matrix.
-----------------------------------------------------------------------*/

int                               /* Returns non-zero if error */
matrix_get_diag(
    voxel_array *diagonal,        /* Diagonal elements */
    voxel_array *matrix)          /* Matrix to get diagonal of */
{
    int type;                     /* Type of data in matrix */
    long *dimen;                  /* Dimensions of matrix */
    long ss;                      /* Stride between diagonal elements */
    long n;                       /* Dimension of vector result */
    int flag;                     /* Error flag */

    /* Check rank, type, and dimensions of the matrix */
    if ( matrix == 0 || matrix->magic != VXL_MAGIC ||
	 matrix->rank != 2 || matrix->data == 0 )
	NZReturnError(ERR_ILLOP, "Argument is not a valid matrix");
    type  = matrix->type;
    dimen = matrix->dimen;

    /* Allocate voxel array for the vector result */
    n = dimen[0];
    if ( n > dimen[1] )
	n = dimen[1];
    flag = vxl_alloc_array(diagonal, type, 1, &n);
    if (flag)    NZReturnLibError();

    /* Copy diagonal elements to result */
    ss = dimen[1] + 1;
    flag = bips_copy(n, type, diagonal->data, 1, matrix->data, ss);
    if (flag)     NZReturnLibError();

    return 0;
}



/*-----------------------------------------------------------------------
matrix_inverse - Inverse of a matrix

This function computes the inverse of a non-singular square matrix
represented as a voxel array of type float or double.  If the matrix
is singular, a non-zero value is returned and the destination matrix
is unchanged.
-------------------------------------------------------------------------*/

int                              /* Returns non-zero if error */
matrix_inverse(
    voxel_array *inverse,        /* Inverse of matrix */
    voxel_array *matrix)         /* Matrix to find inverse of */
{
    int flag;

    /* Switch on matrix data type */
    switch (matrix->type) {

    case INTP_FLOAT:
        flag = matrix_sinv(inverse, matrix);
        break;

    case INTP_DOUBLE:
        flag = matrix_dinv(inverse, matrix);
        break;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
        flag = matrix_cinv(inverse, matrix);
        break;

    case INTP_DCOMPLEX:
        flag = matrix_zinv(inverse, matrix);
        break;
#endif

    default: 
	flag = set_error(ERR_ILLOP, "Matrix type must be floating-point");
        break;
    }

    return flag;
}



/*--------------------------------------------------------------------------
matrix_principal_axes - Principal axes decomposition of symmetric matrix

This function computes the principal axes decomposition of a voxel array
considered as a real symmetric matrix; it is equivalent to solving the
real symmetric eigensystem using matrix_eigen_symmetric except that
the matrix of eigenvectors is adjusted if necessary to ensure that it
defines a proper rotation.  Note that this operation is not defined for
complex Hermitian matrices, the determinant of which may lie anywhere
on the unit circle.

FIXME: Perhaps define row operations, to avoid that ugly jump into
type-specific operations at the end.

FIXME: Does this operation fail if the matrix has a zero eigenvalue?
---------------------------------------------------------------------------*/

int                           /* Returns non-zero if error */
matrix_principal_axes(
    voxel_array *values,      /* Eigenvalues, as vector */
    voxel_array *vectors,     /* Proper orthogonal matrix of eigenvectors */
    voxel_array *matrix)      /* Matrix to decompose */
{
    anytype det;       /* Determinant of matrix of eigenvectors */
    int dtype;         /* Data type of determinant */
    int type;          /* Type of matrix */
    int flag = 0;      /* Error flag */
    long n, i;

    /* Solve the eigensystem */
    flag = matrix_eigen_symmetric(values, vectors, matrix);
    if (flag)    return flag;

    /* Compute determinant of rotation matrix */
    flag = matrix_determinant(&dtype, &det, vectors);
    if (flag)    return flag;

    /* If det == -1, invert last row of rotation matrix */
    n = matrix->dimen[0];
    type = matrix->type;
    switch (type) {

    case INTP_FLOAT:
        if (*(float *)&det < 0) {    /* FIXME: Ugly, ugly! */
            float *data;
            data = vectors->data;
            for (i = 0; i < n; i++) {
                data[n*(n-1)+i] = -data[n*(n-1)+i];
            }
        }
	break;

    case INTP_DOUBLE:
        if (*(double *)&det < 0) {    /* FIXME: Ugly, ugly! */
            double *data;
            data = vectors->data;
            for (i = 0; i < n; i++) {
                data[n*(n-1)+i] = -data[n*(n-1)+i];
            }
        }
	break;
        
    default:
        NZReturnError(ERR_ILLOP, "Matrix type must be float or double");

    } /* end of switch */

    return 0;
}



/*-----------------------------------------------------------------------
matrix_solve - Solve a linear system AX = B

This function uses the LU decomposition to solve the linear system AX
= B for compatible float or double matrices A and B.  If the matrix A
is singular, the destination matrix is unchanged and a non-zero value
is returned.
-------------------------------------------------------------------------*/

int                              /* Returns non-zero if error */
matrix_solve(
    voxel_array *soln,           /* Solution X of system */
    voxel_array *matrix,         /* Matrix A on left-hand side */
    voxel_array *rhs)            /* Matrix B on right-hand side */
{
    int flag;

    /* Switch on matrix data type */
    switch (matrix->type) {

    case INTP_FLOAT:
        flag = matrix_ssolv(soln, matrix, rhs);
        break;

    case INTP_DOUBLE:
        flag = matrix_dsolv(soln, matrix, rhs);
        break;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
        flag = matrix_csolv(soln, matrix, rhs);
        break;

    case INTP_DCOMPLEX:
        flag = matrix_zsolv(soln, matrix, rhs);
        break;
#endif

    default: 
	flag = set_error(ERR_ILLOP, "Matrix type must be floating-point");
	break;
    }

    return flag;
}



/*--------------------------------------------------------------------------
matrix_transpose - Ordinary or Hermitian transpose of a matrix

This function computes the ordinary or Hermitian transpose of a matrix.

FIXME: Generalize to do any permutation of coordinates on an array of
any rank, and to reverse the order of indices on an axis.  Also
generalize to work on arrays of any type.

FIXME: Block for better cache performance.
----------------------------------------------------------------------------*/

int                         /* Returns non-zero if error */
matrix_transpose(
    voxel_array *trans,     /* Transposed matrix */
    voxel_array *matrix,    /* Matrix to be transposed */
    int conj)               /* Take complex conjugate? */
{
    int type;               /* Type of data in matrix */
    int nbytes;             /* Bytes per data element */
    long tdimen[2];         /* Dimensions of transposed matrix */
    long nsrows;            /* Src rows = dest cols, in elements */
    long nscols;            /* Src cols = dest rows, in elements */
    unsigned char *src;     /* Source row addr, in bytes */
    unsigned char *dst;     /* Destination column addr, in bytes */
    int info=0;             /* Status return from BIPS */
    long i, flag;

    /* Verify rank and type */
    if ( matrix == 0 || matrix->magic != VXL_MAGIC ||
	 matrix->rank != 2 || matrix->data == 0 )
	NZReturnError(ERR_ILLOP, "Argument is not a matrix");
    type   = matrix->type;
    nbytes = matrix->nbytes;   

    /* Allocate voxel array for transposed matrix */
    tdimen[1] = nsrows = matrix->dimen[0];
    tdimen[0] = nscols = matrix->dimen[1]; 
    flag = vxl_alloc_array(trans, type, 2, tdimen);
    if (flag)    NZReturnLibError();

    /* Copy rows into columns */
    dst = trans ->data;
    src = matrix->data;
    for (i = 0; i < nsrows; i++) {
        if (conj == 0)
            info = bips_copy(nscols, type, dst, nsrows, src, 1);
#ifdef HAVE_COMPLEX_H
        else
            info = bips_conj(nscols, type, dst, nsrows, src, 1);
#endif
	if (info != 0)    NZReturnLibError();
	dst += nbytes;
	src += nbytes * nscols;
    }

    return ERR_OKAY;
}
