/****************************************************************************
matrix1.c - Generic matrix functions for the Voxel Kit

Author:  Paul Hughett

Copyright (c) 2002-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file contains generic matrix functions for the Voxel Kit, many of
which are wrapper functions for LAPACK and BLAS Fortran subroutines.
It is included in matrix.c once for each data type to be supported, with
macro magic to specify the data types.  Apart from the macros, the
letter x is used below in function names to mean any of:

         s   single-precision real
         d   double-precision real
         c   single-precision complex
         z   double-precision complex

These functions have been gathered into a separate file so that only
one generic version needs to be maintained for multiple real and
complex types and since calling Fortran from C is rather messy,
especially when supporting multiple platforms.  Having all the mess in
the same place should make porting to additional platforms a little
bit easier.

For purposes of describing the C functions, we will suppose that the
successive elements of a matrix are written out from left to right,
and then top to bottom; typographical rows are called "rows", and
columns are called "columns."  Thus the usual C statement for
initializing a matrix will show rows and columns in the natural way
simply by inserting returns at the end of each row.  Upper and lower
triangle matrices are defined in the usual fashion relative to this
typographical representation.

The nomenclature "leading dimension" is retained, even though it
appears as a trailing dimension in C.  Actually, we could change that
to "last dimension" and retain the acronym; there are now a few
instances of that usage in the code below.

This initial implementation is designed to support the LAPACK and BLAS
functions as translated into C by the f2c program; this is known to
work with the RedHat rpm packages for blas and lapack.  Other
platforms will be added as they become necessary.

Likewise, the selection of LAPACK subroutines supported is driven by
current needs of the Pyvox project; other subroutines will be added
as needed.

The likely issues to deal with in porting this to another platform are
the following:

1. The Fortran implementation may mangle subroutine names.  For example,
the LAPACK subroutine DSYEV may appear to C as dsyev_, with an added
trailing underscore.  Other implementations are alleged to use a leading
underscore, or do no mangling.  Capitialization may also vary between
platforms.  Modifying the FMANGLE macro in config.h is the recommended
way to do this.

2.  The sizes of integer and floating point types may differ between C
and Fortran.  Here we assume (temporarily) that int <-> integer, float
<-> real, and double <-> double precision.  Scalars are no problem;
arrays are harder because they may require the allocation of temporary
storage for the conversion.  The typedefs freal, fdouble, fscomplex,
and fdcomplex in config.h define the C types equivalent to these
Fortran types.

3.  The semantics of strings differ between C and Fortran, and suitable
mangling may be required to pass the CHARACTER*1 values that LAPACK
interprets to control its operation.  Here we assume (temporarily) that
it suffices for C to pass a pointer to char.  There is no support as yet
in config.h to handle this problem.

4.  LAPACK and BLAS do not use COMMON, so we don't have to figure out
how to translate _that_ between C and Fortran.


Some other issues for interfacing C and Fortran are, as far as I know,
independent of the platform.

1.  Fortran stores arrays in column-major order, with the first subscript
varying most rapidly as you consider successive elements in memory.  C
stores arrays in row-major order, with the last subscript varying most
rapidly.  In effect, passing an array between languages transposes it.

2.  Fortran passes arguments by reference (with the possible exception
of string arguments, which are sometimes passed by descriptor); C
normally passes arguments by value, but can achieve the effect of
passing by reference by passing a pointer to the actual value.

4. Some implementations of Fortran (e.g. f2c) expect a Fortran main
program to link to.  If the main program is in C, it may be necessary
to provide a dummy Fortran main program.

5. Different implementations of Fortran may require that different
libraries be linked into the program.


Some issues concerning the wrappers and how thick they are:

1.  Fortran name mangling.

2.  Differences between C and Fortran types, and the required conversions
(which may need to be applied to arrays and strings).

3.  Hiding the allocation of work storage from the C programmer, without
being too inefficient.

4.  Hiding transposition of arrays.  (Rewriting descriptions and doing limited
adaptations within the functions.)

5.  Arrays taken raw, or with voxel array headers.

6.  Type of data encoded in function name or passed as a parameter a la BIPS?

7.  Provide raw LAPACK subroutines or nicer but less efficient versions?

8.  Use special compressed forms like LU in a single array or return separate
arrays?  Support special matrix formats like symmetric, banded, etc, or 
just plain arrays?

9.  Fortran origin is 1, C origin is zero.


There are also some LAPACK-specific issues:

1.  On encountering an error, most LAPACK subroutines call the subroutine
XERBLA.  The standard library implementation of this subroutine prints
an error message and aborts the program.  This is not acceptable for
many LAPACK users, so it may be necessary to override the normal XERBLA
implementation; whether or not this is possible, and how it can be done
may be platform-dependent.

2.  Many LAPACK functions require working storage; FORTRAN's limitations
require that this storage be allocated by the caller and passed to the
LAPACK subroutine.  Implementing this in C requires that the C wrapper
allocate the necessary storage via malloc, or automatic allocation, which
is a nuisance and possibly inefficient.


FIXME: Other functions to add: condition estimates, svd, LU, cholesky,
...

FIXME: Consider special interpretation of arrays with rank != 2.
Perhaps rank 0 arrays are scalars, rank 1 arrays are column vectors,
and rank 3 and up arrays preserve the last dimension and stack the
others?

FIXME: Try this code on the Suns and Tru64 and make any changes needed
to support those platforms as well.  Don't forget to install LAPACK
and BLAS first.  Modify the configure script to detect the platform
and do the right thing.
*****************************************************************************/

#if ITYPE == 1
static const char rcsid1[] = "$Id: matrix1.c,v 1.18 2005/10/28 17:02:16 hughett Exp $ Release $Name: pyvox-0-72 $";
#endif



/*-------------------------------------------------------------------------
Setup definitions for requested numeric type
--------------------------------------------------------------------------*/

#undef  FTYPE       /* A Fortran data type */
#undef  CTYPE       /* The equivalent C type */
#undef  RTYPE       /* Equivalent Fortran real type (for complex types) */
#undef  VTYPE       /* The equivalent type code from exim */
#undef  FNAME       /* A BLAS or LAPACK function */
#undef  CNAME       /* The C function that wraps it */
#undef  xsyev

#if ITYPE == 1            /* Single-precision real */
#define FTYPE freal
#define CTYPE float
#define RTYPE freal
#define VTYPE INTP_FLOAT
#define CNAME(name)  matrix_s##name
#define FNAME(name)  FMANGLE(s##name)
#define xsyev    FNAME(syev)

#elif ITYPE == 2          /* Double-precision real */
#define FTYPE fdouble
#define CTYPE double
#define RTYPE fdouble
#define VTYPE INTP_DOUBLE
#define CNAME(name)  matrix_d##name
#define FNAME(name)  FMANGLE(d##name)
#define xsyev    FNAME(syev)

#elif ITYPE == 3          /* Single-precision complex */
#define FTYPE fscomplex
#define CTYPE float complex
#define RTYPE freal
#define VTYPE INTP_FCOMPLEX
#define CNAME(name)  matrix_c##name
#define FNAME(name)  FMANGLE(c##name)
#define xsyev    FNAME(heev)

#elif ITYPE == 4          /* Double-precision complex */
#define FTYPE fdcomplex
#define CTYPE double complex
#define RTYPE fdouble
#define VTYPE INTP_DCOMPLEX
#define CNAME(name)  matrix_z##name
#define FNAME(name)  FMANGLE(z##name)
#define xsyev    FNAME(heev)

#endif


/*--------------------------------------------------------------------------
Define mangled names for BLAS and LAPACK subroutines

NOTE: A few subroutines change names between the real and complex versions
and are handled in the previous section.
---------------------------------------------------------------------------*/

#define xgemm    FNAME(gemm)
#define xgesvd   FNAME(gesvd)
#define xgetrf   FNAME(getrf)
#define xgetri   FNAME(getri)
#define xgetrs   FNAME(getrs)
#define xpotrf   FNAME(potrf)



/*-------------------------------------------------------------------------
matrix_xmul - Compute matrix product of two matrices

This function computes the product of two matrices represented as
voxel arrays.  A rank 1 array used for the second factor will be
treated as a column vector.

FIXME: If factors are not the same type, coerce to the higher type?

FIXME:  Allow the first matrix to have rank > 2; this can be used for
color matrixing and similar operations.  Also support arrays of rank 0
as scalars.
--------------------------------------------------------------------------*/

#define matrix_xmul  CNAME(mul)

static inline 
int                              /* Returns non-zero if error */
matrix_xmul(
    voxel_array *result,         /* Product of matrices */
    voxel_array *amatrix,        /* First matrix:  A */
    voxel_array *bmatrix)        /* Second matrix: B */
{
    int type1, type2;            /* Types of the two voxel arrays */
    long dimen1[2];              /* Dimensions of matrix A */
    long dimen2[2];              /* Dimensions of matrix B */
    long dimen3[2];              /* Dimensions of result */
    fint m, n, k;                /* Matrix dimensions */
    fint lda, ldb, ldc;          /* Matrix last dimensions */
    fchar1 ftra, ftrb;           /* Transpose A?  B? */
    FTYPE alpha=1.0, beta=0.0;   /* Coefficients for xGEMM */         
    FTYPE *a=0, *b=0, *c=0;      /* Matrix data */
    int flag;

    /* Check that matrix A is valid */
    if ( amatrix == 0 ||
         amatrix->magic != VXL_MAGIC ||
         amatrix->data == 0 ||
	 amatrix->rank != 2 ||
         amatrix->type != VTYPE )
        NZReturnError(ERR_DOMAIN, "Left factor is not a valid matrix");
    type1 = amatrix->type;
    dimen1[0] = amatrix->dimen[0];
    dimen1[1] = amatrix->dimen[1];

    /* Check that matrix B is valid */
    if ( bmatrix == 0 ||
         bmatrix->magic != VXL_MAGIC ||
         bmatrix->data == 0 ||
	 ( bmatrix->rank != 1 && bmatrix->rank != 2 ) || 
         bmatrix->type != VTYPE )
        NZReturnError(ERR_DOMAIN, "Right factor is not a valid matrix");
    type2 = bmatrix->type;

    /* Check that A and B matrices are compatible */
    if (type1 != type2)
	NZReturnError(ERR_INCOMP, "Matrix types are not compatible");
    if ( bmatrix->rank == 2 && bmatrix->dimen[0] == dimen1[1] ) {
	dimen2[0] = bmatrix->dimen[0];
	dimen2[1] = bmatrix->dimen[1];
    }
    else if ( bmatrix->rank == 1 && bmatrix->dimen[0] == dimen1[1] ) {
	dimen2[0] = bmatrix->dimen[0];
	dimen2[1] = 1;
    }
    else
	NZReturnError(ERR_INCOMP, "Matrix shapes are not compatible");
    
    /* Allocate voxel array for result */
    dimen3[0] = dimen1[0];
    dimen3[1] = dimen2[1];
    flag = vxl_alloc_array(result, type1, 2, dimen3);
    if (flag)    NZReturnLibError();

    /* DEBUG: Show dimensions */
    if (0) {
	fprintf(stderr, "Dimen1: %ld %ld\n", dimen1[0], dimen1[1]);
	fprintf(stderr, "Dimen2: %ld %ld\n", dimen2[0], dimen3[1]);
	fprintf(stderr, "Dimen3: %ld %ld\n", dimen3[0], dimen3[1]);
    }

    /* Multiply the matrices */
    a = amatrix->data;
    b = bmatrix->data;
    c = result ->data;
    m = ldc = ldb = dimen2[1];
    k = lda = dimen2[0];
    n = dimen1[0];
    ftra = 'N';
    ftrb = 'N';
    if (0)  {        /* DEBUG */
	fprintf(stderr, "lda = %ld, ldb = %ld, ldc = %ld\n", 
		(long)lda, (long)ldb, (long)ldc);
        fprintf(stderr, "m = %ld, n = %ld, k = %ld\n",
		(long)m, (long)n, (long)k);
    }
    xgemm(&ftrb, &ftra, &m, &n, &k, &alpha, b, &ldb, a, &lda, &beta, c, &ldc);

    return 0;
}



/*--------------------------------------------------------------------------
matrix_xdet - Determinant of a voxel array of type x

This function computes the determinant of a voxel array considered as a
matrix.

ALGORITHM: The documentation for the pivot vector returned by xGETRF
is somewhat confusing.  It appears that it is NOT a permutation per
se, but a list of the row interchanges that were done.  Thus the
vector 1 3 3 indicates that row 1 was untouched, row 2 was
interchanged with 3, and the new(!) row 3 was untouched.  Assuming
that this interpretation is correct, the sign of the permutation can
be determined simply by counting the number of interchanges; how to
compute the permutation matrix is less clear.

FIXME: Allocating new storage each time for mdata and permute is
inefficient if many determinants must be calculated.  Storage allocated
on the stack for sufficiently small arrays would be better.
---------------------------------------------------------------------------*/

#define matrix_xdet  CNAME(det)

static inline 
int                            /* Returns nonzero on error */
matrix_xdet(
    int *type,                 /* Data type of determinant */
    CTYPE *det,                /* Value of determinant */
    voxel_array *matrix)       /* Matrix to take determinant of */
{
    CTYPE value;               /* Determinant value */
    fint  *pivot  = 0;         /* Pivot vector */
    fint m, n, lda;            /* Dimensions of matrix */
    CTYPE  *data  = 0;         /* Data from the voxel array */
    FTYPE  *mdata = 0;         /* Matrix data */
    fint info;                 /* Info return from LAPACK */
    int i;

    /* Verify rank, shape, and type of array */
    if ( !matrix_check_square(matrix) )
	NZReturnError(ERR_ILLOP, "Not a matrix or not square");

    /* Do set up for LAPACK */
    m = n = lda = matrix->dimen[0];
    pivot = mallock(m*sizeof(fint));
    if (pivot == 0)    NZReturnLibError();

    /* Copy matrix data into working array */
    data  = matrix->data;
    mdata = mallock(m*n*sizeof(FTYPE));
    if (mdata == 0)    NZReturnLibError();
    for (i = 0; i < m*n; i++)
        mdata[i] = data[i];
    
    /* Call LAPACK for LU decomposition */
    xgetrf(&m, &n, mdata, &lda, pivot, &info);
    if (info < 0)
        NZReturnError(ERR_MISC, "Error calling xGETRF");

    /* Case 1: Matrix is exactly singular */
    if (info > 0)
        value = 0.0;

    /* Case 2: Matrix is not singular */
    else {

        /* Compute determinant from diagonal of LU decomposition */
        value = 1.0;
        for (i = 0; i < m*n; i += n+1)
            value *= mdata[i];

        /* Multiply determinant by sign of permutation */
        for (i = 0; i < m; i++) {
            if (pivot[i] != i+1)          /* Fortran origin is 1 */
                value = -value;
        }
    }

    /* Release temporary storage */
    free(mdata);
    free(pivot);

    /* Store type and value of determinant */
    *type = matrix->type;
    *det  = value;

    return 0;
}



/*-----------------------------------------------------------------------
matrix_xchol - Cholesky factorization of a matrix

This function computes the cholesky factorization of a symmetric or
Hermitian positive definite square matrix represented as a voxel
array.  If the matrix is not positive definite, the destination matrix
is unchanged and a non-zero value is returned.

FIXME: There is a possible error here on platforms for which sizeof(long)
> sizeof(int) = 4.  On such platforms, the variable n is cast from long
to int and information might be lost.
-------------------------------------------------------------------------*/

#define matrix_xchol  CNAME(chol)

static inline 
int                              /* Returns non-zero if error */
matrix_xchol(
    voxel_array *factor,         /* Cholesky factor of matrix */
    voxel_array *matrix)         /* Matrix to factor */
{
    int type;                    /* Type of matrix */
    long dimen[2];               /* Matrix dimensions */
    fchar1 uplo = 'U';           /* Get upper triangular factor in Fortran */
    fint n;                      /* First dimension */
    fint lda;                    /* Last dimension of matrix */
    CTYPE *mdata = 0;            /* Matrix data */
    CTYPE *rdata = 0;            /* Result data */
    fint info;                   /* Info return from LAPACK */
    long i, j, flag;

    /* Verify rank, shape, and type of matrix */
    if ( !matrix_check_square(matrix) )
	NZReturnError(ERR_ILLOP, "Not a matrix or not square");
    type = matrix->type;
    n    = dimen[0] = dimen[1] = matrix->dimen[0];
    lda  = matrix->dimen[0];

    /* Allocate destination array */
    flag = vxl_alloc_array(factor, type, 2, dimen);
    if (flag)    NZReturnLibError();

    /* Copy matrix data */
    mdata = matrix ->data;
    rdata = factor->data;
    for (i = 0; i < n*n; i++)
        rdata[i] = mdata[i];

    /* Compute Cholesky factorization */
    xpotrf(&uplo, &n, rdata, &lda, &info);
    if ( info < 0 )
        NZReturnError(ERR_MISC, "Error calling xPOTRF");
    if ( info > 0 )
        NZReturnError(ERR_DOMAIN, "Matrix is not positive definite");

    /* Fill upper triangle with zeros */
    for (i = 0; i < n-1; i++)
        for (j = i+1; j < n; j++)
            rdata[i*n+j] = 0.0;

    return ERR_OKAY;
}



/*--------------------------------------------------------------------------
matrix_xsvd - Singular value decomposition of a matrix

This function computes the singular value decomposition of a real or
complex matrix.  That is, A = U*D*V**T where U and V are orthogonal or
Hermitian, and D is diagonal.

NOTE: The equivalent Fortran problem is A**T = (V)*(D)*(U**T).

Dimensions of the various arrays, in the C world:

   A    m,n
   U    m,m
   D    min(m,n)     # A vector
   VT   n,n

--------------------------------------------------------------------------*/

#define matrix_xsvd  CNAME(svd)

static inline
int                               /* Returns nonzero on error */
matrix_xsvd(
    voxel_array *left,            /* Matrix of left singular vectors */
    voxel_array *values,          /* Vector of singular values */
    voxel_array *right,           /* Transposed right singular vector */
    voxel_array *matrix)          /* Matrix to be factored */
{
    int type, rtype;              /* Matrix type and corresponding real type */
    long udim [2];                /* Dimensions of U */
    long vtdim[2];                /* Dimensions of VT */
    fchar1 jobu  = 'A';           /* Compute all left singular vectors */
    fchar1 jobvt = 'A';           /* Compute all right singular vectors */
    fint m, n;                    /* Dimensions of A */
    fint lda, ldu, ldvt;          /* Last dimension of A, U, VT */
    long  ns, mn, i;              /* min(m,n), m*n, and a loop index */
    fint info;                    /* Information flag returned */
    fint lwork;                   /* Length of work array */
    FTYPE *a=0, *mdata=0;         /* Matrix data */
    FTYPE *u=0, *vt=0;            /* Singular vector data */
    RTYPE *s=0;                   /* Singular value data */
    FTYPE work1, *work=0;         /* Work array of matrix type */
    RTYPE *rwork=0;               /* Work array of real type */
    int flag;

    /* Verify rank and type of matrix; extract dimensions */
    if ( ! matrix_check(matrix) )
        NZReturnError(ERR_INCOMP, "Not a matrix");
    type  = matrix->type;
    rtype = exim_real_type(type);
    m = matrix->dimen[0];
    n = matrix->dimen[1];
    lda = n;
    udim [0] = udim [1] = ldu  = m;
    vtdim[0] = vtdim[1] = ldvt = n;
    ns = m;
    if (n <ns)    ns = n;

    /* Make a copy of the input matrix */
    mdata = matrix->data;
    mn    = m*n;
    a     = mallock(m*n*sizeof(FTYPE));
    if (a == 0)    NZReturnLibError();
    for (i = 0; i < mn; i++)
        a[i] = mdata[i];

    /* Allocate destination arrays */
    flag = vxl_alloc_array(left,   type,  2, udim) ||
           vxl_alloc_array(values, rtype, 1, &ns)  ||
           vxl_alloc_array(right,  type,  2, vtdim);
    if (flag)    NZReturnLibError();
    u  = left  ->data;
    s  = values->data;
    vt = right ->data;

    /* Allocate work space */
    lwork = -1;
#if ITYPE==1 || ITYPE==2
    xgesvd(&jobu, &jobvt, &n, &m, a, &lda, s, vt, &ldvt, u, &ldu,
           &work1, &lwork, &info);
#else
    xgesvd(&jobu, &jobvt, &n, &m, a, &lda, s, vt, &ldvt, u, &ldu,
           &work1, &lwork, rwork, &info);
#endif
    if (info != 0)
        NZReturnError(ERR_MALLOC, "Failed to size workspace for SVD");
    lwork = (fint)work1;
    work  = mallock(lwork*sizeof(FTYPE));
    if (work==0)    NZReturnLibError();
#if  ITYPE==1 || ITYPE==2
    rwork = 0;        /* Prevents a gcc warning message */
#else
    rwork = mallock(5*ns*sizeof(RTYPE));
    if (rwork == 0)    NZReturnLibError();
#endif

    /* Call LAPACK to compute the SVD */
#if ITYPE==1 || ITYPE==2
    xgesvd(&jobu, &jobvt, &n, &m, a, &lda, s, vt, &ldvt, u, &ldu,
           work, &lwork, &info);
#else
    xgesvd(&jobu, &jobvt, &n, &m, a, &lda, s, vt, &ldvt, u, &ldu,
           work, &lwork, rwork, &info);
#endif
    if (info < 0)
        NZReturnError(ERR_INCOMP, "Bad argument to xGESVD");
    if (info > 0)
        NZReturnError(ERR_ALGOR, "SVD algorithm failed to converge");

    /* Free working storage and exit */
#if  ITYPE==3 || ITYPE==4
    free(rwork);
#endif    
    free(work);
    free(a);

    return 0;
}



/*--------------------------------------------------------------------------
matrix_xsyev - Solve real symmetric eigensystem

This function computes the eigenvalues and eigenvectors of a voxel
array considered as a real symmetric or complex Hermitian matrix.  The
eigenvalues are returned as a rank-1 voxel array.  The eigenvectors
are returned as a square rank-2 voxel array, the rows of which are the
eigenvectors.  The matrix of eigenvectors is guaranteed to be
orthogonal but not necessarily to be proper. (If a proper rotation is
necessary, use matrix_principal_axes instead.)
---------------------------------------------------------------------------*/

#define matrix_xsyev   CNAME(syev)

static inline 
int                               /* Returns non-zero if error */
matrix_xsyev(
    voxel_array *values,          /* Vector of eigenvalues */
    voxel_array *vectors,         /* Matrix of eigenvectors */
    voxel_array *matrix)          /* Matrix to solve */
{
    int type;                     /* Type of matrix */
    long dimen[2];                /* Dimensions of matrix */
    fint n;                       /* Dimension of matrix */
    fint lda;                     /* Last dimension of matrix */
    fchar1 jobz;                  /* Job selector */
    fchar1 uplo;                  /* Upper or lower triangle? */
    FTYPE *work=0;                /* Working storage for LAPACK */
#if ITYPE == 3 || ITYPE == 4
    RTYPE *rwork=0;               /* More working storage */
#endif
    FTYPE wsize;                  /* Recommended size of working storage */
    CTYPE *matdata=0;             /* Matrix data */
    CTYPE *valdata=0;             /* Eigenvalue data */
    CTYPE *vecdata=0;             /* Eigenvector data */
    fint info;                    /* Status returned from LAPACK */
    fint lwork;                   /* Size of working storage */
    long i, flag;

    /* Verify rank, shape, and type of matrix */
    if ( !matrix_check_square(matrix) )
	NZReturnError(ERR_ILLOP, "Not a matrix or not square");
    type = matrix->type;
    dimen[0] = dimen[1] = matrix->dimen[0];

    /* Allocate destination arrays */
    flag = vxl_alloc_array(values,  type, 1, dimen);
    if (flag)    NZReturnLibError();
    flag = vxl_alloc_array(vectors, type, 2, dimen);
    if (flag)    NZReturnLibError();

    /* Set up type-independent arguments for LAPACK */
    n = lda = matrix->dimen[0];
    jobz = 'V';           /* Both eigenvalues and vectors */
    uplo = 'U';           /* Lower triangle (in C!) */

    /* Set up remaining LAPACK arguments */
    matdata = matrix->data;
    valdata = values->data;
    vecdata = vectors->data;

    /* Copy matrix into eigenvector array for computation in place */
    for ( i = 0; i < n*n; i++)
        vecdata[i] = matdata[i];

    /* Allocate working storage */
    lwork = -1;
#if ITYPE == 3 || ITYPE == 4
    xsyev(&jobz, &uplo, &n, vecdata, &lda, valdata, &wsize, 
          &lwork, rwork, &info);
#else
    xsyev(&jobz, &uplo, &n, vecdata, &lda, valdata, &wsize, 
          &lwork, &info);
#endif
    lwork = (fint)wsize;
    work  = mallock(lwork*sizeof(FTYPE));
    if (work == 0)    NZReturnLibError();
#if ITYPE == 3 || ITYPE == 4
    rwork = mallock((3*n-2)*sizeof(RTYPE));
#endif

    /* Call LAPACK to do the work */
#if ITYPE == 3 || ITYPE == 4
    xsyev(&jobz, &uplo, &n, vecdata, &lda, valdata, work, 
           &lwork, rwork, &info);
#else
    xsyev(&jobz, &uplo, &n, vecdata, &lda, valdata, work, 
          &lwork, &info);
#endif

    /* Free allocated storage */
#if ITYPE == 3 || ITYPE == 4
    free(rwork);
#endif
    free(work);

    /* Check for LAPACK errors */
    if (info < 0)
	NZReturnError(ERR_MISC, "Error calling xSYEV or xHEEV");
    if (info > 0)
	NZReturnError(ERR_ALGOR, "Eigenvalue calculation failed to converge");

    return 0;
}



/*-----------------------------------------------------------------------
matrix_xinv - Inverse of a matrix

This function computes the inverse of a non-singular square matrix
represented as a voxel array.  If the matrix is singular, the
destination matrix is unchanged and a non-zero value is returned.

FIXME: There is a possible error here on platforms for which sizeof(long)
> sizeof(int) = 4.  On such platforms, the variable n is cast from long
to int and information might be lost.
-------------------------------------------------------------------------*/

#define matrix_xinv  CNAME(inv)

static inline 
int                              /* Returns non-zero if error */
matrix_xinv(
    voxel_array *inverse,        /* Inverse of matrix */
    voxel_array *matrix)         /* Matrix to find inverse of */
{
    int type;                    /* Type of matrix */
    long dimen[2];               /* Matrix dimensions */
    fint n;                      /* First dimension */
    fint *pivot = 0;             /* Pivot vector */
    fint lda;                    /* Last dimension of matrix */
    fint lwork;                  /* Length of working storage */
    CTYPE *mdata = 0;            /* Matrix data */
    CTYPE *idata = 0;            /* Inverse data */
    FTYPE *work  = 0;            /* Working storage */
    FTYPE wsize;                 /* Recommended size */
    fint info;                   /* Info return from LAPACK */
    long i, flag;

    /* Verify rank, shape, and type of matrix */
    if ( !matrix_check_square(matrix) )
	NZReturnError(ERR_ILLOP, "Not a matrix or not square");
    type = matrix->type;
    n = dimen[0] = dimen[1] = matrix->dimen[0];

    /* Allocate destination array */
    flag = vxl_alloc_array(inverse, type, 2, dimen);
    if (flag)    NZReturnLibError();

    /* Do set up for LAPACK */
    lda   = matrix->dimen[0];
    pivot = mallock(n*sizeof(fint));
    if (pivot == 0)    NZReturnLibError();

    /* Copy matrix data */
    mdata = matrix ->data;
    idata = inverse->data;
    for (i = 0; i < n*n; i++)
        idata[i] = mdata[i];

    /* Compute LU decomposition */
    xgetrf(&n, &n, idata, &lda, pivot, &info);
    if ( info < 0 )
        NZReturnError(ERR_MISC, "Error calling xGETRF");
    if ( info > 0 )
        NZReturnError(ERR_DOMAIN, "Matrix is singular");

    /* Allocate working storage for inverse */
    lwork = -1;
    xgetri(&n, idata, &lda, pivot, &wsize, &lwork, &info);
    lwork = wsize;
    work = mallock(lwork*sizeof(FTYPE));
    if (work == 0)    NZReturnLibError();

    /* Compute inverse */
    xgetri(&n, idata, &lda, pivot, work, &lwork, &info);
    if ( info != 0 )
        NZReturnError(ERR_MISC, "Error calling xGETRI");
    free(work);

    /* Free temporary storage and return */
    free(pivot);
    return ERR_OKAY;
}



/*-----------------------------------------------------------------------
matrix_xsolv - Solve a linear system AX = B

This function uses the LU decomposition to solve the linear system AX
= B for compatible matrices A and B.  If the matrix is singular, a
non-zero value is returned and the destination matrix is unchanged.


ALGORITHM: Both A and B must be transposed to Fortran conventions before
being passed to LAPACK; the solution X must be transposed back to C
conventions before being returned.

FIXME: There is a possible error here on platforms for which
sizeof(long) > sizeof(int) = 4.  On such platforms, the variables lda,
ldb, and n are cast from long to int and information might be lost.
-------------------------------------------------------------------------*/

#define matrix_xsolv  CNAME(solv)

static inline 
int                              /* Returns non-zero if error */
matrix_xsolv(
    voxel_array *soln,           /* Solution X of system */
    voxel_array *matrix,         /* Matrix A on left-hand side */
    voxel_array *rhs)            /* Matrix B on right-hand side */
{
    int type;                    /* Commmon type of matrices */
    long adim[2], bdim[2];       /* Dimensions of A and B matrices */
    fint *pivot = 0;             /* Pivot vector */
    fint lda, ldb;               /* Last dimension of each matrix */
    fint n;                      /* Value of bdim[1] */
    fint info;                   /* Info return from LAPACK */
    fchar1 tsel = 'N';           /* Do not transpose A in LAPACK */
    CTYPE *mdata = 0;            /* Matrix data A */
    CTYPE *rdata = 0;            /* RHS data B */
    CTYPE *sdata = 0;            /* Solution data X */
    FTYPE *adata = 0;            /* Transposed copy of A */
    FTYPE *bdata = 0;            /* Transposed copy of B or X */
    long i, j, flag;

    /* Verify shape and type of input matrices */
    if ( !matrix_check_square(matrix) )
	NZReturnError(ERR_ILLOP, "Left-hand side not a square matrix");
    lda = adim[0] = adim[1] = matrix->dimen[0];
    type = matrix->type;
    if ( !matrix_check(rhs) )
	NZReturnError(ERR_ILLOP, "Right-hand side not a matrix");
    if ( type != rhs->type )
	NZReturnError(ERR_UNIMP, "Types of A and B matrices do not match");
    ldb = bdim[0] = rhs->dimen[0];
    n   = bdim[1] = rhs->dimen[1];
    if ( lda != ldb )
        NZReturnError(ERR_INCOMP, "A and B matrices have incompatible shapes");
        
    /* Allocate destination array */
    flag = vxl_alloc_array(soln, type, 2, bdim);
    if (flag)    NZReturnLibError();

    /* Do set up for LAPACK */
    pivot = mallock(lda*sizeof(fint));
    if (pivot == 0)    NZReturnLibError();

    /* Copy/transpose A to temporary storage */
    adata = mallock(lda*lda*sizeof(FTYPE));
    if (adata == 0)    NZReturnLibError();
    mdata = matrix->data;
    for (i = 0; i < lda; i++)
        for (j = 0; j < lda; j++)
            adata[j*lda+i] = mdata[i*lda+j];

    /* Copy/transpose B to temporary storage */
    bdata = mallock(ldb*n*sizeof(FTYPE));
    if (bdata == 0)    NZReturnLibError();
    rdata = rhs->data;
    for (i = 0; i < ldb; i++)
        for (j = 0; j < n; j++)
            bdata[j*ldb+i] = rdata[i*n+j];

    /* Compute LU decomposition */
    xgetrf(&lda, &lda, adata, &lda, pivot, &info);
    if ( info < 0 )
        NZReturnError(ERR_MISC, "Error calling xGETRF");
    if ( info > 0 )
        NZReturnError(ERR_DOMAIN, "Matrix is singular");

    /* Use LU decomposition to solve system */
    xgetrs(&tsel, &lda, &n, adata, &lda, pivot, bdata, &ldb, &info);
    if ( info < 0 )
        NZReturnError(ERR_MISC, "Error calling xGETRS");

    /* Copy/transpose the solution X from temporary storage */
    sdata = soln->data;
    for (i = 0; i < ldb; i++)
        for (j = 0; j < n; j++)
            sdata[i*n+j] = bdata[j*lda+i];

    /* Free temporary storage */
    free(bdata);
    free(adata);
    free(pivot);

    return ERR_OKAY;
}
