/****************************************************************************
memm.c - Memory management

Author:  Paul Hughett

Copyright (c) 1998-2003 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file provides a standard interface for memory allocation and
deallocation within software written at the Brain Behavior Laboratory
at the University of Pennsylvania.  The current version consists of versions
of malloc, calloc, and realloc which check whether memory was successfully
allocated and report an error if not.  Others may be added in the future,
such as the ability to maintain a fast pool of small blocks of fixed size.

NOTE that this file differs from the same-named file in segm and some other
BBL software; the difference is that it uses the new error management code
to report errors rather than just calling fatal or warning.
*****************************************************************************/

static const char rcsid[] = "$Id: memm.c,v 1.3 2003/08/20 22:20:51 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include <stdlib.h>

#include "errm.h"
#include "memm.h"



/*-------------------------------------------------------------------------
callock - Allocate and clear memory, with error checking

This function calls the standard library routine calloc and generates
a fatal error if it cannot reallocate a block of the requested size.
If the requested block size is non-positive, a warning is generated and
zero is returned.
---------------------------------------------------------------------------*/

void *                     /* Returns zero on error */
callock( 
    size_t nelem,          /* Number of elements */
    size_t elsize )        /* Size of each element */
{
  void * retval;

  if (nelem <= 0 || elsize <= 0)
      ZWarning(ERR_ASSERT, "Block requested with non-positive length.");

  retval = calloc(nelem, elsize);
  if (retval == NULL) 
      ZReturnError1(ERR_MALLOC, "Unable to allocate %d bytes.", nelem*elsize);

  return retval;
}



/*---------------------------------------------------------------------------
mallock - Allocate memory, with error message if not available.

This function calls the standard library routine malloc and generates
a fatal error if it cannot allocate a block of the requested size.
To help catch memory errors, attempting to allocate a block with
non-positive length generates a warning message but gets a zero
pointer.
----------------------------------------------------------------------------*/

void *                     /* Returns zero on error */
mallock( 
    size_t size)           /* Size of block requested */
{
  void * retval;

  if (size <= 0)
      ZWarning(ERR_ASSERT, "Block requested with non-positive length");

  retval = malloc(size);
  if ( retval == NULL )
      ZReturnError1(ERR_MALLOC, "Unable to allocate %d bytes.", size);

  return retval;
}



/*---------------------------------------------------------------------------
reallock - Reallocate memory, with error message if not available.

This function calls the standard library routine realloc and generates
a fatal error if it cannot reallocate a block of the requested size.
To help catch memory errors, attempting to reallocate a block with
negative length generates a warning message but gets a zero
pointer.
----------------------------------------------------------------------------*/

void *                  /* Returns zero on error */
reallock( 
    void *ptr,          /* Pointer to existing block */
    size_t size )       /* New size requested */
{
  void * retval;

  if (size < 0) 
      ZWarning(ERR_ASSERT, "Block requested with negative length");

  retval = realloc( ptr, size );
  if ( retval == NULL ) 
      ZReturnError1(ERR_MALLOC, "Unable to reallocate %d bytes.", size);

  return retval;
}
