/****************************************************************************
neighbor.c - Voxel Kit routines for neighborhood operations

Author:  Paul Hughett

Copyright (c) 1998-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  This particular file contains code for neighborhood
operations and related functions.

Neighborhood operations are those in which each output voxel depends
on a fixed neighborhood of the corresponding input voxel.  Some examples
include local mean, local variance, local min/max, erosion, dilation,
and convolution; some less obvious examples include the Gibbs energy,
surface area by counting methods, the median filter, and ICM (iterated
conditional mode).

These functions all face similar algorithmic problems in choosing the
access patterns to use a multi-level cache efficiently.  Several
different algorithms currently coexist in this section, while I
experiment with different trade-offs between implementation time,
readability, and efficiency; the trend over time can be expected to be
toward improving efficiency, even at the cost of implementation time
and readability, since these functions are likely to be major
contributors to the total run time.  A further trend will be to
support neighborhoods of arbitrary shape in images of arbitrary rank.

The vxl_kernel struct is used to define a neighborhood in any number
of dimensions; see its definition for the precise details.  Generally,
it specifies the neighbors of a voxel by giving coordinate deltas;
a convolution coefficient may also be specified for each neighbor.

Note that the algorithms used here are designed to maximize locality
within a multilevel cache architecture, thereby decreasing the run
time.  If you don't understand what that means, you probably shouldn't
touch this code, under penalty of a possible order-of-magnitude
increase in run time.  The code is also optimized for the case that
most voxels are interior rather than boundary voxels.

FIXME: Consider rewriting a lot of the kernel manipulation code as a
Python class.
 
FIXME: The dilabel, dilate, and erode operations are very similar, and
the other convolution-like operations are similar; try to find a way
of sharing the code.

FIXME: Implement the neighborhood operations as yet undone: local
min/max, gray scale morphology, count nonzero neighbors, Gibbs sum,
median/rank filters, ICM, hit-and-miss transform.  Local variance
(without local mean) might be useful for some applications.

FIXME: Can IIR filters be modeled by convolution kernels?
If so, the order of output pixels (and neighbors?) matters.  On
the other hand,they might be closer to scan operations like fill.

FIXME: Decide how to account for physical coordinates in kernels.
Kernels themselves should probably be in index coordinates, but the
kernel constructors may want to consider and correct for spacing.

FIXME: Write functions to generate convolution kernels on demand for
Sobel, lowpass, gaussian, etc, in any number of dimensions.  Ditto for
other neighborhood operations.

FIXME: Implement shrinking for all these neighborhood operations.
*****************************************************************************/

static const char rcsid[] = "$Id: neighbor.c,v 1.24 2006/02/24 17:18:51 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include <stdio.h>
#include <stdlib.h>

#ifdef HAVE_TGMATH_H
#include <tgmath.h>
#else
#include <math.h>
#endif

#include "bips.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"



/*------------------------------------------------------------------------
                        Predefined Kernels

FIXME: The better way to define all these kernels would be to write
functions that create kernels with desired characteristics on demand.

FIXME: These do not initialize the bias to zero, and so cannot be used
in convolution.  The problem is now that the bias is an anytype, that
field will not get set properly in these builtin kernels.  The only
one that currently gets used seems to be neigh3d27; and that one is
used only to define a plain neighborhood for which the bias does not
matter.

FIXME: Implement vxl_box_kernel() and use instead of manually creating
these fixed-rank kernels.

FIXME: Implement functions to create convolution kernels of various
types, or to convert from a voxel array plus a center.  Provide
Pyvox bindings to these functions.

FIXME: Implement functions to create useful morphological
neighborhoods, and provide Pyvox bindings to these functions.
--------------------------------------------------------------------------*/

/* The 9 in-plane neighbors of a voxel in 3d */
static long n3d9ip[][VXL_MAX_RANK] = {
    { 0, -1, -1}, { 0, -1,  0}, { 0, -1, +1},
    { 0,  0, -1}, { 0,  0,  0}, { 0,  0, +1},
    { 0, +1, -1}, { 0, +1,  0}, { 0, +1, +1},
};

struct vxl_kernel neigh3d9ip = {VXL_KERN_MAGIC, 3, 9, 9, INTP_DOUBLE, 
				n3d9ip, NULL};



/* The 27 neighbors of a voxel in 3D */
static long n3d27[][VXL_MAX_RANK] = {
    {-1, -1, -1}, {-1, -1,  0}, {-1, -1, +1},
    {-1,  0, -1}, {-1,  0,  0}, {-1,  0, +1},
    {-1, +1, -1}, {-1, +1,  0}, {-1, +1, +1},
    { 0, -1, -1}, { 0, -1,  0}, { 0, -1, +1},
    { 0,  0, -1}, { 0,  0,  0}, { 0,  0, +1},
    { 0, +1, -1}, { 0, +1,  0}, { 0, +1, +1},
    {+1, -1, -1}, {+1, -1,  0}, {+1, -1, +1},
    {+1,  0, -1}, {+1,  0,  0}, {+1,  0, +1},
    {+1, +1, -1}, {+1, +1,  0}, {+1, +1, +1},
};

/* Coefficients for a 27-element rectangular window */
static double rect27[27] = {
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
   1.0, 1.0, 1.0,
};

struct vxl_kernel neigh3d27 = {VXL_KERN_MAGIC, 3, 27, 27, INTP_DOUBLE,
			       n3d27, rect27};

/* The 13 neighbors of a voxel in 3D which precede it in raster order */
static long n3d13[][VXL_MAX_RANK] = {
    {-1, -1, -1}, {-1, -1,  0}, {-1, -1, +1},
    {-1,  0, -1}, {-1,  0,  0}, {-1,  0, +1},
    {-1, +1, -1}, {-1, +1,  0}, {-1, +1, +1},
    { 0, -1, -1}, { 0, -1,  0}, { 0, -1, +1},
    { 0,  0, -1}, 
};

struct vxl_kernel neigh3d13 = {VXL_KERN_MAGIC, 3, 13, 13, INTP_DOUBLE,
			       n3d13, NULL};



/*------------------------------------------------------------------------
vxl_create_kernel - Create a convolution kernel

This function creates a vxl_kernel structure from given neighborhood
coordinates, coefficients, and bias; the coefficients and bias are
optional.  Some basic consistency checks are made on the requested
kernel.  The function vxl_destroy_kernel can be used to free the
kernel when it is no longer necessary.

There are several possible options for the count, delta, and coef
arguments.  If count == 0, then the delta and coef arguments are
ignored but space is allocated for some number of neighbors determined
by the rank.  If count > 0 and both delta and coef are zero, then
space is allocated for the specified number of neighbors but no actual
neighbors are defined.  If count > 0, and delta and coef are both
nonzero, then space is allocated for count neighbors and populated
from the delta and coef arrays.  Other possibilities are not
permitted, including having only one of delta and coef be
nonzero.  This code does not support neighborhood-only kernels.

For safety, the delta and coef arrays are copied to newly allocated
storage; the originals may be recycled or freed by the caller upon
return from vxl_create_kernel.

FIXME: This function has become more general than is actually needed
by the Python code that calls it; consider simplifying.

FIXME: The status of neighbor-only kernels is currently undecided.  See
TODO for some more information.

FIXME: Provide separate functions to create/allocate and to
create/allocate and define?

FIXME: Provide separate arguments for count and alloc?  If so, how to
indicate that the size should be defaulted?
-------------------------------------------------------------------------*/

struct vxl_kernel *                /* Returns NULL on error */
vxl_create_kernel(
    int rank,                      /* Number of dimensions */
    int count,                     /* Number of neighbors in the kernel */
    int type,                      /* Type of kernel */
    long (*delta)[VXL_MAX_RANK],   /* Coordinate deltas for each neightbor */
    void *coef,                    /* Coefficient array, or zero */
    void *bias)                    /* Constant bias term for convolution */
{
    struct vxl_kernel *kern;       /* Kernel under construction */
    long alloc=0;                  /* How many neighbors to allocate */
    long (*ndelta)[VXL_MAX_RANK];  /* Copy of delta array */
    void *ncoef = 0;               /* Copy of coef array */
    int nbytes=0;                  /* Size of a single coefficient */
    int n;                         /* Index over neighbors */
    int d;                         /* Index over dimensions */
    int flag;

    /* Do some basic consistency checks */
    if (rank < 1 || rank > VXL_MAX_RANK)
	ZReturnError(ERR_ILLOP, "Invalid rank requested");
    if (count < 0 || count > VXL_MAX_NEIGH)
	ZReturnError(ERR_ILLOP, "Invalid neighbor count requested");

    /* Verify count, delta, and coef; decide how much space to allocate */
    /* We assume that 9 neighbors are most common in 2-D and 27 in 3-D. */
    if (count == 0) {
        alloc = (rank < 3) ? 9 : 27;
        delta = 0;
        coef  = 0;
    }
    else if (count > 0) {
        alloc = count;
        if (delta == 0)
            count = 0;
    }
    else
        ZReturnError(ERR_INIT, "Number of neighbors may not be negative");

    /* Only double and double complex kernels are allowed */
    if ( type != INTP_DOUBLE && type != INTP_DCOMPLEX )
        ZReturnError(ERR_ILLOP, "Only double and double complex are allowed");
    nbytes = exim_sizeof_type(type);
    if (nbytes == 0)
        ZReturnError(ERR_ASSERT, "Internal: Type has size zero?!");

    /* Copy the delta array, if any, to newly allocated storage */
    ndelta = mallock(alloc*VXL_MAX_RANK*sizeof(long));
    if (ndelta == 0)    ZReturnLibError();
    if (delta != 0) {
	for (n = 0; n < count; n++) 
	    for (d = 0; d < VXL_MAX_RANK; d++)
	        ndelta[n][d] = (d < rank) ? delta[n][d] : 0;
    }

    /* Copy the coef array, if present, to newly allocated storage */
    ncoef = mallock(alloc*nbytes);
    if (ncoef == 0)    ZReturnLibError();
    if (coef != 0) {
        flag = bips_copy(count, type, ncoef, 1, coef, 1);
        if (flag)    ZReturnLibError();
    }

    /* Create and initialize the kernel */
    kern = mallock(sizeof(struct vxl_kernel));
    if (kern == 0)    ZReturnLibError();
    kern->magic = VXL_KERN_MAGIC;
    kern->rank  = rank;
    kern->alloc = alloc;
    kern->count = count;
    kern->type  = type;
    kern->delta = ndelta;
    kern->coef  = ncoef;

    /* Copy the bias, if any, into the kernel */
    if (bias != 0) {
        flag = bips_copy(1, type, &kern->bias, 0, bias, 0);
        if (flag)    ZReturnLibError();
    }
    else {
        flag = bips_zero(1, type, &kern->bias, 0);
        if (flag)    ZReturnLibError();
    }

    return kern;
}



/*------------------------------------------------------------------------
vxl_destroy_kernel - Free all storage associate with a kernel struct

This function frees the delta array, coef array (if any), and struct
storage associated with a kernel.  It is a serious error to attempt to
destroy a kernel that was not created with vxl_create_kernel or a
similar function.
-------------------------------------------------------------------------*/

int
vxl_destroy_kernel(
    struct vxl_kernel *kern)         /* The kernel to be destroyed */
{
    /* Do some consistency checks to help detect misteaks */
    if ( kern == 0 || kern->magic != VXL_KERN_MAGIC ||  kern->delta == 0 ||
         kern->rank  < 1 || kern->rank  > VXL_MAX_RANK ||
	 kern->count < 0 || kern->count > VXL_MAX_NEIGH )
	NZReturnError(ERR_INIT, "Attempting to destroy an invalid kernel");

    /* Looks okay, so destroy this kernel */
    if (kern->coef != 0)
	free(kern->coef);
    free(kern->delta);
    free(kern);    

    return 0;
}



/*-----------------------------------------------------------------------
vxli_find_kernel_coef - Find index of a given kernel neighbor

This function finds the index (in the delta and coef arrays) of a given
neighbor specified by its delta; it returns -1 if no such neighbor is
found.  This function assumes that the kernel has already been verified
and is for internal use only.

FIXME: Make inlineable?  If so, remember to intern the error strings.
Make a public version for the Pyvox kernel code to use?
------------------------------------------------------------------------*/

long                              /* Returns index or -1 if not found */
vxli_find_kernel_coef(
    struct vxl_kernel *kern,      /* Kernel to examine */
    long *delta)                  /* Desired offset of neighbor */
{
    int fail = 0;                 /* If no match */
    long i, j;

    /* Find the desired neighbor */
    for (i = 0; i < kern->count; i++) {
        fail = 0;
        for (j = 0; j < kern->rank; j++)
            if (delta[j] != kern->delta[i][j])
                fail = 1;
        if (fail == 0)
            return i;
    }
    
    return -1;
}



/*-------------------------------------------------------------------------
vxl_set_kernel_bias - Set bias in an existing kernel

This function sets the bias in the given kernel.  The type of the bias
must be compatible with the type of the kernel itself.  
--------------------------------------------------------------------------*/

int                               /* Returns nonzero on error */
vxl_set_kernel_bias(
    struct vxl_kernel *kern,      /* Kernel to modify */
    int type,                     /* Datatype of bias */
    void *bias)                   /* Value of bias */
{
    int flag;

    /* Check validity of the kernel */
    /* FIXME: Make this into a common function?  Add to PyvoxKernel_Check? */
    if (kern == 0 || kern->magic != VXL_KERN_MAGIC ||
        kern->alloc <= 0 || kern->delta == 0 || kern->coef == 0)
        NZReturnError(ERR_INIT, "Invalid or empty kernel");

    /* Check type compatibility */
    if ( exim_iscomplex(type) && !exim_iscomplex(kern->type) )
        NZReturnError(ERR_INCOMP, "Complex coef requires a complex kernel");

    /* Set the bias */
    flag = bips_cast(1, kern->type, &kern->bias, 0, type, bias, 0);
    if (flag)    NZReturnLibError();

    return 0;
}



/*-------------------------------------------------------------------------
vxl_set_kernel_coef - Set a neighbor coefficient to an existing kernel

This function sets the coefficient of a specified neighbor in the
given kernel; if the neighbor does not yet exist, it is created.
The type of the coefficient must be compatible with the type of the
kernel itself.  The number of neighbors allocated will be increased as
needed; note that the location of the delta and coef arrays may change
when this happens.
--------------------------------------------------------------------------*/

int                               /* Returns nonzero on error */
vxl_set_kernel_coef(
    struct vxl_kernel *kern,      /* Kernel to modify */
    long *delta,                  /* Offset of neighbor */
    int type,                     /* Datatype of coefficient */
    void *coef)                   /* Value of coefficient */
{
    long index;                   /* Index to neighbor */
    long alloc;                   /* Number of neighbors allocated */
    int nbytes;                   /* Size of a single coefficient */
    unsigned char *pcoef;         /* Location of a kernel coefficient */
    int i, flag;

    /* Check validity of the kernel */
    /* FIXME: Make this into a common function?  Add to PyvoxKernel_Check? */
    if (kern == 0 || kern->magic != VXL_KERN_MAGIC ||
        kern->alloc <= 0 || kern->delta == 0 || kern->coef == 0)
        NZReturnError(ERR_INIT, "Invalid or empty kernel");

    /* Check type compatibility */
    if ( exim_iscomplex(type) && !exim_iscomplex(kern->type) )
        NZReturnError(ERR_INCOMP, "Complex coef requires a complex kernel");

    /* Check whether the desired neighbor already exists */
    index = vxli_find_kernel_coef(kern, delta);

    /* If necessary, make room for a new neighbor */
    /* FIXME: If allocation fails, the contents are lost forever */
    if (index < 0 && kern->count >= kern->alloc) {
        alloc  = 2*kern->alloc;
        nbytes = exim_sizeof_intype(kern->type);
        kern->delta = reallock(kern->delta, alloc*VXL_MAX_RANK*sizeof(long));
        kern->coef  = reallock(kern->coef,  alloc*nbytes);
        kern->alloc = alloc;
        if (kern->delta == 0 || kern->coef == 0)
            NZReturnError(ERR_MALLOC, "Memory allocation failed");
    }

    /* If necessary, add the new neighbor */
    if (index < 0) {
        index = kern->count++;
        for (i = 0; i < kern->rank; i++)
            kern->delta[index][i] = delta[i];
    }

    /* Set the coefficient of the (old or new) neighbor */
    pcoef = bips_locate(kern->type, kern->coef, index);
    if (pcoef == 0)
        NZReturnError(ERR_ASSERT, "Internal error: broken kernel?");
    flag = bips_cast(1, kern->type, pcoef, 0, type, coef, 0);
    if (flag != 0)    NZReturnLibError();

    return 0;
}



/*-------------------------------------------------------------------------
vxl_del_kernel_coef - Delete a neighbor from an existing kernel

This function removes a neighbor and its coefficient from an existing
kernel.  It is permissible to remove a neighbor that does not exist.

FIXME: The number of neighbors allocated is NOT decreased, on the grounds
that one only rarely needs to delete kernel coefs, so the wasted space is
trivial in practice.
--------------------------------------------------------------------------*/

int                               /* Returns nonzero on error */
vxl_del_kernel_coef(
    struct vxl_kernel *kern,      /* The kernel to modify */
    long *delta)                  /* The coefficient to remove */
{
    long index, i;                /* Index to neighbor */
    unsigned char *p, *q;         /* Pointers to kernel coefficients */
    int j, flag;

    /* Check validity of the kernel */
    /* FIXME: Make this into a common function? */
    if (kern == 0 || kern->magic != VXL_KERN_MAGIC ||
        kern->alloc <= 0 || kern->delta == 0 || kern->coef == 0)
        NZReturnError(ERR_INIT, "Invalid or empty kernel");

    /* Find the specified neighbor; if none, return immediately */
    index = vxli_find_kernel_coef(kern, delta);
    if (index < 0)
        return 0;

    /* Move following neighbors down one slot */
    kern->count--;
    for (i = index; i < kern->count; i++) {
        p = bips_locate(kern->type, kern->coef, i);
        q = bips_locate(kern->type, kern->coef, i+1);
        if (p == 0 || q == 0)
            NZReturnError(ERR_ASSERT, "Internal error: bogus kernel?");
        flag = bips_copy(1, kern->type, p, 0, q, 0);
        if (flag != 0)
            NZReturnError(ERR_ASSERT, "Internal error: bogus kernel?");
        for (j = 0; j < kern->rank; j++)
            kern->delta[i][j] = kern->delta[i+1][j];
    }

    return 0;            
}



/*--------------------------------------------------------------------------
vxl_lowpass_kernel - Generate a lowpass kernel in N dimensions

This function creates and returns a lowpass kernel in any specified
number of dimensions; the bandpass is chosen so as to exactly cancel
the Nyquist frequency.  The coefficients are 2^(d - |x| - |y| - ...),
where d is the number of dimensions.  The kernel returned is always of
type double (but will work correctly on complex images) and may be
safely destroyed by vxl_destroy_kernel.

FIXME: Modify to use vxl_create_kernel?

FIXME: Add parameters to control bandwidth of the lowpass filter.
-----------------------------------------------------------------------------*/

struct vxl_kernel *             /* Returns NULL on error */
vxl_lowpass_kernel(
    int rank)                   /* Number of dimensions in the kernel */
{
    struct vxl_kernel *kernel = 0;   /* The newly created kernel */
    int count = 0;                   /* Number of neighbors */
    long   (*delta)[VXL_MAX_RANK];   /* Coordinate deltas for each neighbor */
    double *coef  = 0;               /* Coefficient for each neighbor */
    long coord[VXL_MAX_RANK];        /* Coordinates within kernel, plus 1 */
    long dimen[VXL_MAX_RANK];        /* Kernel dimensions: 3, 3, ... */
    double coef1, coef2;             /* Coefficient factors */
    double *cp = 0;                  /* Pointer into coefficient array */
    long   *dp = 0;                  /* Pointer into coordinate delta array */
    int id;                          /* Index over dimensions */

    /*  Do some basic consistency checks */
    if (rank < 1 || rank > VXL_MAX_RANK)
	ZReturnError(ERR_ILLOP, "Invalid rank requested");

    /* Compute parameters dependent on the rank */
    count = 1;
    coef1 = 1.0;
    for (id = 0; id < rank; id++) {
        count *= 3;
	coef1 /= 2.0;
	dimen[id] = 3;
    }

    /* Allocate space for kernel and its parts */
    delta = mallock(count*VXL_MAX_RANK*sizeof(long));
    dp = (long *)delta;
    if (delta == 0)    ZReturnLibError();
    cp = coef = mallock(count*sizeof(double));
    if (coef == 0)    ZReturnLibError();
    kernel = mallock(sizeof(struct vxl_kernel));
    if (kernel == 0)    ZReturnLibError();

    /* Initialize the kernel */
    kernel->magic = VXL_KERN_MAGIC;
    kernel->rank  = rank;
    kernel->alloc = count;
    kernel->count = count;
    kernel->type  = INTP_DOUBLE;
    kernel->delta = delta;
    kernel->coef  = coef;
    kernel->bias.as_double = 0.0;
    
    /* Compute coordinate deltas and coefficient for each neighbor */
    vxli_firstvoxel(rank, coord);
    do {
        coef2 = coef1;
        for (id = 0; id < VXL_MAX_RANK; id++) {
  	    *dp++ = (id < rank) ? (coord[id] - 1) : 0;
	    if (id < rank && coord[id] != 1)
	        coef2 /= 2.0;
	}
	*cp++ = coef2;
    } while ( vxli_nextvoxel(rank, dimen, coord) );

    return kernel;
}



/*--------------------------------------------------------------------------
vxli_rowout - Determine if an image row is outside the image

This function determines if the coordinates of an image row plus delta
coordinates are outside the boundaries of the image; this is a common
operation in the implementation of the neighborhood operations.  Error
checking is assumed to be done by the caller.
---------------------------------------------------------------------------*/

static inline int 
vxli_rowout(
    int rank,                  /* Rank of the voxel array */
    long *dimen,               /* Dimensions of the voxel array */
    long *coords,              /* Row coordinates to be checked */
    long *delta)               /* Offsets to be added before checking */
{
    int i;                     /* Index over dimensions */
    long y;                    /* Coordinate value along some dimension */

    for ( i = rank-2; i >= 0; i-- ) {
	y = coords[i] + delta[i];
	if ( y < 0 || y >= dimen[i] )
	    return 1;
    }

    return 0;
}



/*--------------------------------------------------------------------------
vxl_convolve - Do direct convolution on a image of any rank

This function performs a direct convolution and optional subsampling
on an image of any rank and data type, using a neighborhood and
coefficients defined by a vxl_kernel of the same rank.  The rank must
be one or greater; convolution is not defined on images of rank zero.
The type of the output image is always the same as the type of the
input image.  Convolving a real image by a complex kernel is not
permitted, although convolving a complex image by a real kernel is.
Intermediate calculations are always done in type double or double
complex and rounded back to the final type; if a result voxel cannot
be represented within range of the result type, the result is set to
the appropriate limit of the range.  Incomplete neighborhoods at the
boundary of the image are handled correctly; that is, the missing
voxels are simply omitted from the neighborhood.  The shrink argument
may be set to NULL, in which case no subsampling is done.  In case you
were wondering, the subsampling is done so that the unneeded values
are never computed, which is faster than a plain convolution followed
by a plain subsampling.
---------------------------------------------------------------------------*/

int                            /* Returns non-zero on error */
vxl_convolve(
    voxel_array *dest,         /* Result of the operation */
    voxel_array *src,          /* The input image to be convolved */
    struct vxl_kernel *kern,   /* The convolution kernel to use */
    long *shrink)              /* Shrink factor along each dimension */
{
    unsigned char *dd, *sd;    /* Actual data array for dest and source */
    unsigned char *cdd, *csd;  /* Current row in source and destination */
    unsigned char *sum;        /* One row of convolution sums */
    unsigned char *coef;       /* Coefficient from the convolution kernel */
    int type;                  /* Common type of source and destination */
    int nbytes;                /* Length of an element, in bytes */
    int rank;                  /* Rank of the image */
    int last;                  /* Index to last dimension of the image */
    int mtype;                 /* "Middle" type used for computations */
    int mbytes;                /* Bytes per element of middle type */
    int ktype;                 /* Type of the kernel */
    int kbytes;                /* Bytes per element for the kernel type */
    long ldsrc;                /* Nmbr elements on last dimen of source */
    long lddst;                /* Nmbr elements on last dimen of destination */
    long shrunk[VXL_MAX_RANK]; /* Local array of shrink factors */
    long lshrink;              /* Shrink factor along last dimension */
    long scoord[VXL_MAX_RANK]; /* Coordinates of a source row (scanline) */
    long dcoord[VXL_MAX_RANK]; /* Coordinates of a destination row */
    long *sdimen;              /* Dimensions of the source image */
    long ddimen[VXL_MAX_RANK]; /* Dimensions of the destination image */
    long (*delta)[VXL_MAX_RANK];  /* Coordinate delta for each neighbor */
    long *ldelta;              /* Delta along last dimen, for each neighbor */
    long *offset = 0;          /* Offset (in bytes) to each neighbor */
    long xmin, xmax;           /* Min and max valid column indices */
    long anel;                 /* Adjusted number of elements to process */
    unsigned char *asum;       /* Adjusted starting position in sum array */
    unsigned char *asd;        /* Adjusted starting position in source row */
    int nn;                    /* Number of neighbors */
    int in;                    /* Index over neighbors */
    int err;
    int i;

    /* Check validity of source array; get the dimensions.
       FIXME:  Check validity of dimensions? */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    rank = src->rank;
    if (rank < 1)
	NZReturnError(ERR_ILLOP, 
                      "Convolution is undefined for images of rank 0");
    last   = rank - 1;
    sdimen = src->dimen;
    ldsrc  = sdimen[last];
    type   = src->type;
    sd     = src->data;
    nbytes = exim_sizeof_intype(src->type);

    /* Determine type to use for computations */
    if ( exim_iscomplex(type) )
        mtype = INTP_DCOMPLEX;
    else
        mtype = INTP_DOUBLE;
    mbytes = exim_sizeof_intype(mtype);
    if (mbytes == 0)
        NZReturnError(ERR_ASSERT, "(Internal) Invalid middle type");

    /* Check validity of the kernel */
    if (kern == 0)
	NZReturnError(ERR_ILLOP, "No kernel specified");
    else if (kern->magic != VXL_KERN_MAGIC)
        NZReturnError(ERR_INIT, "Bad magic number in kernel");
    else if (kern->rank != rank)
	NZReturnError(ERR_INCOMP, "Source and kernel must have the same rank");
    else if (kern->coef == 0)
	NZReturnError(ERR_ILLOP, "Convolution kernel has no coefficients");
    ktype  = kern->type;
    kbytes = exim_sizeof_intype(ktype);
    if (kbytes == 0)
        NZReturnError(ERR_ASSERT, "(Internal) Invalid kernel type");

    /* Default to no subsampling */
    if (shrink == 0) {
	shrink = shrunk;
        for (i = 0; i < rank; i++)
	    shrink[i] = 1;
    }
    lshrink = shrink[last];

    /* Check validity of destination; allocate memory */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    for (i = 0; i < rank; i++)
        ddimen[i] = ceil((double)sdimen[i]/shrink[i]);
    lddst = ddimen[last];
    err = vxl_alloc_array(dest, type, rank, ddimen);
    if (err)    NZReturnLibError();
    dd = dest->data;

    /* Set origin and spacing in the destination */
    for (i = 0; i < rank; i++) {
        dest->origin [i] = src->origin [i];
        dest->spacing[i] = src->spacing[i] * shrink[i];
    }

    /* Compute offset in bytes to each neighbor; get delta along last
       dimension for each neighbor */
    nn = kern->count;
    delta = kern->delta;
    offset = mallock(nn*sizeof(long));
    if (offset == 0)    NZReturnLibError();
    ldelta = mallock(nn*sizeof(long));
    if (ldelta == 0)    NZReturnLibError();
    for (in = 0; in < nn; in++) {
	offset[in] = nbytes*vxli_delta(src, delta[in]);
	ldelta[in] = delta[in][last];
    }

    /* Allocate one row of working storage for intermediate sums */
    sum = mallock(lddst*mbytes);
    if (sum == 0)    NZReturnLibError();

    /* Loop over all rows of the destination */
    vxli_firstrow(rank, dcoord);
    do {

	/* Locate current row in source and destination */
        for (i = 0; i < rank; i++)
	    scoord[i] = dcoord[i] * shrink[i];
        csd = vxli_locate(src,  scoord);
        cdd = vxli_locate(dest, dcoord);
        if (csd == 0 || cdd == 0)    NZReturnLibError();

	/* Initialize the intermediate sums to the bias value */
	err = bips_cast(lddst, mtype, sum, 1, ktype, &kern->bias, 0);
	if (err != 0)    NZReturnLibError();

	/* Loop over the neighbors */
	for (in = 0; in < nn; in++) {

	    /* Skip rows that are completely outside the source */
	    if ( vxli_rowout(rank, sdimen, scoord, delta[in]) )
		continue;

	    /* Adjust iteration limits for source and destination limits */
	    xmin = ceil((double)(-ldelta[in])/lshrink);
	    if (xmin < 0)
		xmin = 0;
	    xmax = ceil((double)(ldsrc-ldelta[in])/lshrink);
            if (xmax > lddst)
		xmax = lddst;

	    /* Process into the destination row */
	    anel = xmax - xmin;
            asum = sum + xmin * mbytes;
	    asd  = csd + offset[in] + xmin*lshrink*nbytes;
	    coef = ((unsigned char *)(kern->coef)) + in*kbytes;
	    err  = bips_conv1(anel, mtype, asum, 1, 
                              type, asd, lshrink, ktype, coef);
	    if (err != 0)    NZReturnLibError();

	}

	/* Convert intermediate sums to give result */
        err = bips_round_cast(lddst, type, cdd, 1, mtype, sum, 1);
	if (err != 0)
	    NZReturnError(ERR_MISC, "Error in calling BIPS function");
    }  while ( vxli_nextrow(rank, ddimen, dcoord) ) ;

    /* Free temporary storage */
    free(sum);
    free(ldelta);
    free(offset);

    return 0;
}


/* Old, non-shrinking version of vxl_convolve 
   FIXME: Use this as a guide to modify erode, dilate, and perhaps other 
   neighborhood operations to permit subsampling; then delete it.
   FIXME: Not updated to support complex types. */
int
vxl_convolve_old(
    voxel_array *dest,         /* Result of the operation */
    voxel_array *src,          /* The input image to be convolved */
    struct vxl_kernel *kern)   /* The convolution kernel to use */
{
    unsigned char *dd, *sd;    /* Actual data array for dest and source */
    unsigned char *cdd, *csd;  /* Current row in source and destination */
    unsigned char *asd;        /* Current src row adjusted for boundaries */
    double *sum;               /* One row of convolution sums */
    double *asum;              /* Address adjusted for missing neighbors */
    double *coef;              /* Coefficient from the convolution kernel */
    int type;                  /* (Internal) type of source and destination */
    int rank;                  /* Rank of the image */
    int last;                  /* Index to last dimension of the image */
    long nlast;                /* Number of elements along last dimension */
    long anl;                  /* Adjusted number of elements in row */
    int nbytes;                /* Length of an element, in bytes */
    long coord[VXL_MAX_RANK];  /* Coordinates of a scanline (aka row) */
    long *dimen;               /* Dimensions of the source image */
    long (*delta)[VXL_MAX_RANK];  /* Coordinate delta for each neighbor */
    long *offset = 0;          /* Offset (in bytes) to each neighbor */
    int nn;                    /* Number of neighbors */
    int in;                    /* Index over neighbors */
    int err;

    /* Check validity of source array; get the dimensions.
       FIXME:  Check validity of dimensions? */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    rank = src->rank;
    if (rank < 1)
	NZReturnError(ERR_ILLOP, 
                      "Convolution is undefined for images of rank 0");
    last = rank - 1;
    dimen = src->dimen;
    nlast = dimen[last];
    type  = src->type;
    sd = src->data;
    nbytes = exim_sizeof_intype(src->type);

    /* Check validity of the kernel */
    if (kern == 0)
	NZReturnError(ERR_ILLOP, "No kernel specified");
    else if (kern->magic != VXL_KERN_MAGIC)
        NZReturnError(ERR_INIT, "Bad magic number in kernel");
    else if (kern->rank != rank)
	NZReturnError(ERR_INCOMP, "Source and kernel must have the same rank");
    else if (kern->coef == 0)
	NZReturnError(ERR_ILLOP, "Convolution kernel has no coefficients");

    /* Check validity of destination; allocate memory */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    err = vxl_alloc_array(dest, src->type, src->rank, src->dimen);
    if (err)    NZReturnLibError();
    dd = dest->data;

    /* Compute offset in bytes to each neighbor */
    nn = kern->count;
    delta = kern->delta;
    offset = mallock(nn*sizeof(long));
    if (offset == 0)    NZReturnLibError();
    for (in = 0; in < nn; in++) 
	offset[in] = nbytes*vxli_delta(src, delta[in]);

    /* Allocate one row of working storage for intermediate sums */
    sum = mallock(nlast*sizeof(double));
    if (sum == 0)    NZReturnLibError();

    /* Loop over all but the last dimension of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Locate current row in source and destination */
        csd = vxli_locate(src,  coord);
        cdd = vxli_locate(dest, coord);
        if (csd == 0 || cdd == 0)    NZReturnLibError();

	/* Clear the intermediate sums */
	err = bips_copy(nlast, INTP_DOUBLE, sum, 1, &kern->bias, 0);
	if (err != 0)    NZReturnLibError();

	/* Loop over the neighbors */
	for (in = 0; in < nn; in++) {

	    /* Skip source rows that are outside the image */
	    if ( vxli_rowout(rank, dimen, coord, delta[in]) )
		continue;

	    /* Adjust limits to account for image boundaries */
	    asd  = csd + offset[in];
	    asum = sum;
	    anl  = nlast;
	    if (delta[in][last] < 0) {
		anl  += delta[in][last];
		asum -= delta[in][last];
		asd  -= delta[in][last] * nbytes; }
	    if (delta[in][last] > 0) {
		anl -= delta[in][last]; }

	    /* Process into the destination row */
	    coef = ((double *)(kern->coef)) + in;
	    err = bips_conv1(anl, INTP_DOUBLE, asum, 1, type, asd, 1,
                             INTP_DOUBLE, coef);
	    if (err != 0)    NZReturnLibError();

	}

	/* Convert intermediate sums to give result */
	err = bips_round_cast(nlast, type, cdd, 1, INTP_DOUBLE, sum, 1);
	if (err != 0)    NZReturnLibError();
    }  while ( vxli_nextrow(rank, dimen, coord) ) ;

    /* Free temporary storage */
    free(sum);
    free(offset);

    return 0;
}



/*--------------------------------------------------------------------------
vxl_dilabel - Do morphological dilation on a label image of any rank

This function performs a morphological dilation operation on an
unsigned volume label image of any rank one or more; each voxel of the
output is the smallest nonzero pixel value of the neighbors specified
in the kernel, or zero if all the neighbors are zero.  The type of the
input must be an unsigned integral type; the type of the output is the
same as the type of the input.  Incomplete neighborhoods at the
boundary of the image are handled correctly; that is, voxels outside
the image are simply omitted.  Any bias and coefficients provided in
the kernel are ignored; it is needed just to define the neighborhood
or structuring element of the dilation.

FIXME: Add capability to subsample; see convolve for the algorithm.

FIXME: The dilabel, dilate, and erode operations are identical except
for the choice of operator for combining pixel values and the choice
of initial value for the destination; figure out how to share the
code.  Also consider whether substituting other operations for AND and
OR creates useful operations.  In particular, min and max seem like
they might be useful.
---------------------------------------------------------------------------*/

int
vxl_dilabel(
    voxel_array *dest,         /* Result of the operation */
    voxel_array *src,          /* The input image to be convolved */
    struct vxl_kernel *kern)   /* The convolution kernel to use */
{
    unsigned char *dd,  *sd;   /* Actual data array for dest and source */
    unsigned char *cdd, *csd;  /* Current row in source and destination */
    unsigned char *asd;        /* Current src row adjusted for boundaries */
    unsigned char *add;        /* Current dest row adjusted for boundaries */
    int type;                  /* Type of source and destination */
    int rank;                  /* Rank of the image */
    int last;                  /* Index to last dimension of the image */
    long nlast;                /* Number of elements along last dimension */
    long anl;                  /* Adjusted number of elements in row */
    int nbytes;                /* Length of an element, in bytes */
    long coord[VXL_MAX_RANK];  /* Coordinates of a row */
    long *dimen;               /* Dimensions of the source image */
    long (*delta)[VXL_MAX_RANK];  /* Coordinate delta for each neighbor */
    long *offset = 0;          /* Offset (in bytes) to each neighbor */
    int nn;                    /* Number of neighbors */
    int in;                    /* Index over neighbors */
    int flag;

    /* Check validity of source array; get the dimensions.
       FIXME:  Check validity of dimensions? */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    rank = src->rank;
    if (rank < 1)
	NZReturnError(ERR_ILLOP, 
                      "Dilation is undefined for images of rank 0");
    last   = rank - 1;
    dimen  = src->dimen;
    nlast  = dimen[last];
    type   = src->type;
    sd     = src->data;
    nbytes = exim_sizeof_intype(src->type);

    /* Check validity of the kernel */
    if (kern == 0)
	NZReturnError(ERR_ILLOP, "No kernel specified");
    else if (kern->magic != VXL_KERN_MAGIC)
        NZReturnError(ERR_INIT, "Bad magic number in kernel");
    else if (kern->rank != rank)
	NZReturnError(ERR_INCOMP, "Source and kernel must have the same rank");

    /* Check validity of destination; allocate memory */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    flag = vxl_alloc_array(dest, type, rank, dimen);
    if (flag)    NZReturnLibError();
    vxli_copy_physical(dest, src);
    dd = dest->data;

    /* Compute offset in bytes to each neighbor */
    nn = kern->count;
    delta = kern->delta;
    offset = mallock(nn*sizeof(long));
    if (offset == 0)    NZReturnLibError();
    for (in = 0; in < nn; in++) 
	offset[in] = nbytes*vxli_delta(src, delta[in]);

    /* Loop over all but the last dimension of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Locate current row in source and destination */
        csd = vxli_locate(src,  coord);
        cdd = vxli_locate(dest, coord);
        if (csd == 0 || cdd == 0)    NZReturnLibError();

	/* Clear the destination */
        flag = bips_zero(nlast, type, cdd, 1);
        if (flag)    NZReturnLibError();

	/* Loop over the neighbors */
	for (in = 0; in < nn; in++) {

	    /* Skip source rows that are outside the image */
	    if ( vxli_rowout(rank, dimen, coord, delta[in]) )
		continue;

	    /* Adjust limits to account for image boundaries */
	    asd = csd + offset[in];
	    add = cdd;
	    anl = nlast;
	    if (delta[in][last] < 0) {
		anl += delta[in][last];
		add -= delta[in][last] * nbytes;
		asd -= delta[in][last] * nbytes; }
	    if (delta[in][last] > 0) {
		anl -= delta[in][last]; }

	    /* Process into the destination row */
	    flag = bips_nzmin_set(anl, type, add, 1, asd, 1);
	    if (flag != 0)    NZReturnLibError();

	}

    }  while ( vxli_nextrow(rank, dimen, coord) ) ;

    /* Free temporary storage */
    free(offset);

    return 0;
}



/*--------------------------------------------------------------------------
vxl_dilate - Do morphological dilation on a image of any rank

This function performs a morphological dilation operation on an
unsigned volume image of any rank one or more; each voxel of the
output is the bitwise OR of the neighbors specified in the kernel.
The type of the input must be an unsigned integral type; the type of
the output is the same as the type of the input.  Incomplete
neighborhoods at the boundary of the image are handled correctly; that
is, voxels outside the image are simply omitted from the OR.  Any
bias and coefficients provided in the kernel are ignored; it is needed
just to define the neighborhood or structuring element of the dilation.

NOTE: The generic version of this function seems to be about 1% slower
than the old, specialized 3d version, even with inlining; that is
probably a reasonable price for most applications.  Hence the old
version has been deleted.

FIXME: Add capability to subsample; see convolve for the algorithm.

FIXME: The dilate and erode operations are identical except for the
choice of operator from AND and OR and the choice of initial value for
the destination; figure out how to share the code.  Also consider
whether substituting other operations for AND and OR creates useful
operations.  In particular, min and max seem like they might be
useful.
---------------------------------------------------------------------------*/

int
vxl_dilate(
    voxel_array *dest,         /* Result of the operation */
    voxel_array *src,          /* The input image to be convolved */
    struct vxl_kernel *kern)   /* The convolution kernel to use */
{
    unsigned char *dd,  *sd;   /* Actual data array for dest and source */
    unsigned char *cdd, *csd;  /* Current row in source and destination */
    unsigned char *asd;        /* Current src row adjusted for boundaries */
    unsigned char *add;        /* Current dest row adjusted for boundaries */
    int type;                  /* Type of source and destination */
    int rank;                  /* Rank of the image */
    int last;                  /* Index to last dimension of the image */
    long nlast;                /* Number of elements along last dimension */
    long anl;                  /* Adjusted number of elements in row */
    int nbytes;                /* Length of an element, in bytes */
    long coord[VXL_MAX_RANK];  /* Coordinates of a row */
    long *dimen;               /* Dimensions of the source image */
    long (*delta)[VXL_MAX_RANK];  /* Coordinate delta for each neighbor */
    long *offset = 0;          /* Offset (in bytes) to each neighbor */
    int nn;                    /* Number of neighbors */
    int in;                    /* Index over neighbors */
    int flag;

    /* Check validity of source array; get the dimensions.
       FIXME:  Check validity of dimensions? */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    rank = src->rank;
    if (rank < 1)
	NZReturnError(ERR_ILLOP, 
                      "Dilation is undefined for images of rank 0");
    last = rank - 1;
    dimen = src->dimen;
    nlast = dimen[last];
    type  = src->type;
    sd = src->data;
    nbytes = exim_sizeof_intype(src->type);

    /* Check validity of the kernel */
    if (kern == 0)
	NZReturnError(ERR_ILLOP, "No kernel specified");
    else if (kern->magic != VXL_KERN_MAGIC)
        NZReturnError(ERR_INIT, "Bad magic number in kernel");
    else if (kern->rank != rank)
	NZReturnError(ERR_INCOMP, "Source and kernel must have the same rank");

    /* Check validity of destination; allocate memory */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    flag = vxl_alloc_array(dest, type, rank, dimen);
    if (flag)    NZReturnLibError();
    vxli_copy_physical(dest, src);
    dd = dest->data;

    /* Compute offset in bytes to each neighbor */
    nn = kern->count;
    delta = kern->delta;
    offset = mallock(nn*sizeof(long));
    if (offset == 0)    NZReturnLibError();
    for (in = 0; in < nn; in++) 
	offset[in] = nbytes*vxli_delta(src, delta[in]);

    /* Loop over all but the last dimension of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Locate current row in source and destination */
        csd = vxli_locate(src,  coord);
        cdd = vxli_locate(dest, coord);
        if (csd == 0 || cdd == 0)    NZReturnLibError();

	/* Clear the destination */
        flag = bips_zero(nlast, type, cdd, 1);
        if (flag)    NZReturnLibError();

	/* Loop over the neighbors */
	for (in = 0; in < nn; in++) {

	    /* Skip source rows that are outside the image */
	    if ( vxli_rowout(rank, dimen, coord, delta[in]) )
		continue;

	    /* Adjust limits to account for image boundaries */
	    asd = csd + offset[in];
	    add = cdd;
	    anl = nlast;
	    if (delta[in][last] < 0) {
		anl += delta[in][last];
		add -= delta[in][last] * nbytes;
		asd -= delta[in][last] * nbytes; }
	    if (delta[in][last] > 0) {
		anl -= delta[in][last]; }

	    /* Process into the destination row */
	    flag = bips_or_set(anl, type, add, 1, asd, 1);
	    if (flag != 0)    NZReturnLibError();

	}

    }  while ( vxli_nextrow(rank, dimen, coord) ) ;

    /* Free temporary storage */
    free(offset);

    return 0;
}



/*--------------------------------------------------------------------------
vxl_erode - Do morphological erosion on a image of any rank

This function performs a morphological erosion operation on an
unsigned volume image of any rank one or more; each voxel of the
output is the bitwise AND of the neighbors specified in the kernel.
The type of the input must be an unsigned integral type; the type of
the output is the same as the type of the input.  Incomplete
neighborhoods at the boundary of the image are handled correctly; that
is, voxels outside the image are simply omitted from the neighborhood.
Any bias and coefficients provided in the kernel are ignored; it is
needed just to define the neighborhood or structuring element.

NOTE: The generic version of this function seems to be about 1% slower
than the old, specialized 3d version, even with inlining; that is
probably a reasonable price for most applications.  The old 3d-only
version has therefore been deleted.

FIXME: Add capability to subsample; see convolve for the algorithm.

FIXME: Implement functions to create useful neighborhoods, and provide
Pyvox bindings to these functions.
---------------------------------------------------------------------------*/

int
vxl_erode(
    voxel_array *dest,         /* Result of the operation */
    voxel_array *src,          /* The input image to be convolved */
    struct vxl_kernel *kern)   /* The convolution kernel to use */
{
    unsigned char *dd,  *sd;   /* Actual data array for dest and source */
    unsigned char *cdd, *csd;  /* Current row in source and destination */
    unsigned char *asd;        /* Current src row adjusted for boundaries */
    unsigned char *add;        /* Current dest row adjusted for boundaries */
    int type;                  /* Type of source and destination */
    int rank;                  /* Rank of the image */
    int last;                  /* Index to last dimension of the image */
    long nlast;                /* Number of elements along last dimension */
    long anl;                  /* Adjusted number of elements in row */
    int nbytes;                /* Length of an element, in bytes */
    long coord[VXL_MAX_RANK];  /* Coordinates of a row */
    long *dimen;               /* Dimensions of the source image */
    long (*delta)[VXL_MAX_RANK];  /* Coordinate delta for each neighbor */
    long *offset = 0;          /* Offset (in bytes) to each neighbor */
    int nn;                    /* Number of neighbors */
    int in;                    /* Index over neighbors */
    int err;

    /* Check validity of source array; get the dimensions.
       FIXME:  Check validity of dimensions? */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    rank = src->rank;
    if (rank < 1)
	NZReturnError(ERR_ILLOP, "Erosion is undefined for images of rank 0");
    last = rank - 1;
    dimen = src->dimen;
    nlast = dimen[last];
    type  = src->type;
    sd = src->data;
    nbytes = exim_sizeof_intype(src->type);

    /* Check validity of the kernel */
    if (kern == 0)
	NZReturnError(ERR_ILLOP, "No kernel specified");
    else if (kern->magic != VXL_KERN_MAGIC)
        NZReturnError(ERR_INIT, "Bad magic number in kernel");
    else if (kern->rank != rank)
	NZReturnError(ERR_INCOMP, "Source and kernel must have the same rank");

    /* Check validity of destination; allocate memory */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    err = vxl_alloc_array(dest, type, rank, dimen);
    if (err)   NZReturnLibError();
    vxli_copy_physical(dest, src);
    dd = dest->data;

    /* Compute offset in bytes to each neighbor */
    nn = kern->count;
    delta = kern->delta;
    offset = mallock(nn*sizeof(long));
    if (offset == 0)    NZReturnLibError();
    for (in = 0; in < nn; in++) 
	offset[in] = nbytes*vxli_delta(src, delta[in]);

    /* Loop over all but the last dimension of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Locate current row in source and destination */
        csd = vxli_locate(src,  coord);
        cdd = vxli_locate(dest, coord);
        if (csd == 0 || cdd == 0)    NZReturnLibError();

        /* Initialize the destination to all one bits */
        err = bips_ffff(nlast, type, cdd, 1);
        if (err!= 0)    NZReturnLibError();

	/* Loop over the neighbors */
	for (in = 0; in < nn; in++) {

	    /* Skip source rows that are outside the image */
	    if ( vxli_rowout(rank, dimen, coord, delta[in]) )
		continue;

	    /* Adjust limits to account for image boundaries */
	    asd = csd + offset[in];
	    add = cdd;
	    anl = nlast;
	    if (delta[in][last] < 0) {
		anl += delta[in][last];
		add -= delta[in][last] * nbytes;
		asd -= delta[in][last] * nbytes; }
	    if (delta[in][last] > 0) {
		anl -= delta[in][last]; }

	    /* Process into the destination row */
	    err = bips_and_set(anl, type, add, 1, asd, 1);
	    if (err != 0)    NZReturnLibError();

	}

    }  while ( vxli_nextrow(rank, dimen, coord) ) ;

    /* Free temporary storage */
    free(offset);

    return 0;
}



/*--------------------------------------------------------------------------
vxl_lostat - Compute local (neighborhood) statistics on a image of any rank

This function computes the local mean and standard deviation within
each neighborhood of an image of any rank and returns the results as
local mean and local standard deviation images of the same rank,
dimensions, and type as the original image.  The neighborhood is
defined by a convolution kernel; the kernel coefficients define the
weights assigned to each neighbor.  The rank must be one or greater;
while the local statistics are sort of defined on a rank zero image,
they are not really useful to compute and it simplifies the algorithm
not to have to treat this special case.  To avoid division by zero,
the neighborhood must include the central (0,0,...) voxel.
Intermediate calculations are done in type double or double complex
and rounded back to the final type; if a result voxel cannot be
represented within range of the result type, the result is set to the
appropriate limit of the range.  The standard deviation image is
multiplied by the given scale factor before conversion; this feature
is probably useful only for images of integral type, to match the
expected range of values to the range supported by the type.
Incomplete neighborhoods at the boundary of the image are handled
correctly; that is, the missing voxels are simply omitted from the
neighborhood, so that the mean and standard deviation are computed
correctly for the incomplete neighborhood; note that the standard
error of the mean will be increased accordingly for neighborhoods on
the boundary of the image.

If you want only the local mean, you can use vxl_convolve() instead
with a suitable kernel.

NOTE: This function (with a suitable kernel) replaces the now-omitted
function vxl_lovar_uchar3d.  The newer function will handle images of
any rank, dimensions, or type, and permits the user to specify the
neighborhood; it also handles incomplete neighborhoods at the boundary
of the image correctly.  If you're converting old code, note that
vxl_lostat rounds the standard deviation where vxl_lovar_uchar3d
truncates, and that the scale factor 2.0 is probably better than the
scale factor (255.0 / 127.4125) used in vxl_lovar_uchar3d.  These
differences mean that the output values will be slightly different
between the two algorithms.

FIXME: Possibly provide a way of returning only the variance image,
probably by specifying mean = 0?  There is probably little or no
saving of time, but the space saving might be significant in some
applications that need only one of the return values.

FIXME: Allow the caller to specify the destination type?

FIXME: Is there an application for the local skew or higher moments?

FIXME: Add capability to subsample; see convolve for the algorithm.
---------------------------------------------------------------------------*/

int
vxl_lostat(
    voxel_array *mean,         /* (Output) Local mean image */
    voxel_array *sdev,         /* (Output) Standard deviation image */
    voxel_array *src,          /* Image to take statistics of */
    struct vxl_kernel *kern,   /* The convolution kernel to use */
    double scale)              /* Multiplier for the std dev image */
{
    double *sum, *asum;              /* Raw and adjusted weights */
    unsigned char *sumx, *asumx;     /* Basic and adjusted values */
    double *sumxx, *asumxx;          /* Basic and adjusted squares */
    unsigned char *md, *dd, *sd;     /* Actual data for mean, dev, and src */
    unsigned char *cmd, *cdd, *csd;  /* Current row in each image */
    unsigned char *asd;        /* Current src row adjusted for boundaries */
    double coef;               /* Coefficient from the kernel */
    int type;                  /* Type of source and destination */
    int rank;                  /* Rank of the image */
    int last;                  /* Index to last dimension of the image */
    long nlast;                /* Number of elements along last dimension */
    long anl;                  /* Adjusted number of elements in row */
    int nbytes;                /* Length of a source element, in bytes */
    int wtype;                 /* Working type for computations */
    int wbytes;                /* Bytes per element of working type */
    long coord[VXL_MAX_RANK];  /* Coordinates of a scanline (aka row) */
    long *dimen;               /* Dimensions of the source image */
    long (*delta)[VXL_MAX_RANK];  /* Coordinate delta for each neighbor */
    long *offset = 0;          /* Offset (in bytes) to each neighbor */
    int count;                 /* Number of neighbors */
    int in;                    /* Index over neighbors */
    int err, err1, err2, err3;

    /* Check validity of source array; get the dimensions.
       FIXME:  Check validity of dimensions? */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    rank = src->rank;
    if (rank < 1)
	NZReturnError(ERR_ILLOP, 
                      "Operation is undefined for images of rank 0");
    last  = rank - 1;
    dimen = src->dimen;
    nlast = dimen[last];
    type  = src->type;
    sd = src->data;
    nbytes = exim_sizeof_intype(type);

    /* Check validity of the kernel */
    if (kern == 0)
	NZReturnError(ERR_ILLOP, "No kernel specified");
    if (kern->magic != VXL_KERN_MAGIC)
        NZReturnError(ERR_INIT, "Bad magic number in kernel");
    if (kern->rank != rank)
	NZReturnError(ERR_INCOMP, "Source and kernel must have the same rank");
    if (kern->coef == 0)
	NZReturnError(ERR_ILLOP, "Kernel has no coefficients");
    if ( kern->type != INTP_DOUBLE )
        NZReturnError(ERR_UNIMP, "Not defined for complex kernels");

    /* Determine the working type for computations */
    wtype = INTP_DOUBLE;
    if ( exim_iscomplex(type) )
        wtype = INTP_DCOMPLEX;
    wbytes = exim_sizeof_intype(wtype);

    /* Check validity of destinations; allocate memory */
    if (mean == 0 || mean->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array for mean");
    err = vxl_alloc_array(mean, type, rank, dimen);
    if (err)    NZReturnLibError();
    md = mean->data;
    if (sdev == 0 || sdev->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array for sdev");
    err = vxl_alloc_array(sdev, type, rank, dimen);
    if (err)    NZReturnLibError();
    vxli_copy_physical(mean, src);
    vxli_copy_physical(sdev, src);
    dd = sdev->data;

    /* Compute offset in bytes to each neighbor */
    count = kern->count;
    delta = kern->delta;
    offset = mallock(count*sizeof(long));
    if (offset == 0)    NZReturnLibError();
    for (in = 0; in < count; in++) 
	offset[in] = nbytes*vxli_delta(src, delta[in]);

    /* Allocate one row of working storage for intermediate sums */
    sum   = mallock(nlast*sizeof(double));
    sumx  = mallock(nlast*wbytes);
    sumxx = mallock(nlast*sizeof(double));
    if (sum == 0 || sumx == 0 || sumxx == 0)
        NZReturnLibError();

    /* Loop over all but the last dimension of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Locate current row in source and destination arrays */
        csd = vxli_locate(src,  coord);
        cmd = vxli_locate(mean, coord);
        cdd = vxli_locate(sdev, coord);
        if ( csd == 0 || cmd == 0 || cdd == 0)    NZReturnLibError();

	/* Clear the intermediate sums */
	err1 = bips_zero(nlast, INTP_DOUBLE, sum,   1);
	err2 = bips_zero(nlast, wtype,       sumx,  1);
	err3 = bips_zero(nlast, INTP_DOUBLE, sumxx, 1);
	if (err1 != 0 || err2 != 0 || err3 != 0)
	    NZReturnLibError();

	/* Loop over the neighbors */
	for (in = 0; in < count; in++) {

	    /* Skip source rows that are outside the image */
	    if ( vxli_rowout(rank, dimen, coord, delta[in]) )
		continue;

	    /* Adjust limits to account for image boundaries */
	    anl    = nlast;
	    asum   = sum;
	    asumx  = sumx;
            asumxx = sumxx;
	    asd    = csd + offset[in];
	    if (delta[in][last] < 0) {
		anl    += delta[in][last];
		asum   -= delta[in][last];
		asumx  -= delta[in][last] * wbytes;
		asumxx -= delta[in][last];
		asd    -= delta[in][last] * nbytes; }
	    if (delta[in][last] > 0) {
		anl -= delta[in][last]; }

	    /* Process into the destination row */
            err = exim_get_value(&coef, kern->coef, kern->type, in);
	    if (err != 0)    NZReturnLibError();
	    err = bips_lostat1(anl, asum, asumx, asumxx, type, asd, 1, coef);
	    if (err != 0)    NZReturnLibError();

	}

	/* Convert intermediate sums to give result */
	err = bips_lostat2(nlast, type, cmd, cdd, 1, sum, sumx, sumxx, scale);
	if (err != 0)    NZReturnLibError();
    }  while ( vxli_nextrow(rank, dimen, coord) ) ;

    /* Free temporary storage */
    free(sumxx);
    free(sumx);
    free(sum);
    free(offset);

    return 0;
}
