/****************************************************************************
pyexim.c - Python bindings for the export/import toolkit

Author:  Paul Hughett

Copyright (c) 2000-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This file contains code to implement the data export/import package
exim as a Python extension.  The current version is extremely
experimental and subject to change.  YOU HAVE BEEN WARNED!!!

The only objects defined so far are names for the various internal
and external types; methods may be added later as needed.

FIXME: Is it reasonable to somehow embed the endianness in the type?

FIXME: Define common type for unary and binary coercions?

FIXME: Add the following attributes and methods:

exim.import(???) = import a data stream
exim.export(???) = export a data stream
exim.bigend  = make an extype bigendian?
exim.littleend = make an extype little endian?

For internal types only:
intype.sumtype = Recommended type for summations
intype.rsumtype = Recommended type for real summations

For external types only:
extype.min, max = min/max values, if possible to define

*****************************************************************************/

static const char rcsid[] = "$Id: pyexim.c,v 1.35 2006/02/24 17:29:10 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include "Python.h"

#include "memm.h"
#include "errm.h"
#include "bips.h"
#include "exim.h"
#include "pyvox.h"


/* Stop gcc from grumbling about missing prototypes */
void initexim(void);


/* Pseudo external type for a string */
/* FIXME: Add this to the documentation */
#define EXTP_STRING (1024)    /* Plain string; length defined by format */
#define EXTP_NSTR4  (1025)    /* String prefixed by 4-byte count */



/*------------------------------------------------------------------------
exim.error - Unique error object for Exim to signal an exception
------------------------------------------------------------------------*/

static PyObject *EximError;

#define RaiseError(message) \
    { PyErr_SetString(EximError, message); return NULL; }



/*------------------------------------------------------------------------
ReturnEximError - Set the error string and return a value

These macros set the error string in the error object EximError and
return either zero or one.  The ReturnEximError version should be
used in methods that normally return a Python object but return zero
in case of an error; the XReturnEximError version is used in some
other contexts where a side effect is normal and the function returns
non-zero to indicate an error.  The SetEximError version simply sets
the error message and leaves it to the user to return the appropriate
error value.

On the other hand, the macros SetEximLibError, ReturnEximLibError,
and XReturnEximLibError use the error message previously set by some
library function called by Exim via the set_error function of errm.c
rather than using a explicit message.

The proper usage of these macros has been a bit confused but is
gradually becoming clear.  A top-level function in Exim that is
called directly by the Python interpreter should use set_error_context
and clear_error_context to define the error context; use
ReturnEximError or XReturnEximError to return errors that it finds
itself; and use ReturnEximLibError or XReturnEximLibError to return
errors found by one of the functions that it calls.  The internal
functions, whether part of Exim, the Voxel Kit, or whatever, should
use NZReturnError or ZReturnError (from errm.h) to declare errors. 
It's not clear that any of the other macros serve any useful purpose.

FIXME: Go back through the code and check that these rules are followed.

FIXME: This code duplicates error handling code in pyvox.h except that
it uses Exim rather than Pyvox.  Figure out how to use the code only
once and avoid having to maintain both.

ALGORITHM: The do {...} while 0 format is used so that this macro can
be used wherever a statement is acceptable.

These macros call clear_error_context to make sure that the context is
cleared even on an immediate error return from a function.

FIXME: Convert the error codes into specific Python exceptions rather
than the generic EximError?
--------------------------------------------------------------------------*/

#define ReturnEximError(errcode, message)                  \
    do { set_error(errcode, message);                      \
         PyErr_SetString(EximError, get_error_message());  \
         clear_error_context();                            \
         return 0; }                                       \
    while (0)

#define XReturnEximError(errcode, message)                 \
    do { set_error(errcode, message);                      \
         PyErr_SetString(EximError, get_error_message());  \
         clear_error_context();                            \
         return 1; }                                       \
    while (0)

/* FIXME: Arrange for the errcode to do something useful */
/* FIXME: Use Py_FatalError instead? */
#define FatalEximError(errcode, message) \
    errm_panic(message)


#define ReturnEximLibError()                               \
    do { PyErr_SetString(EximError, get_error_message());  \
         clear_error_context();                            \
         return 0; }                                       \
    while (0)

#define XReturnEximLibError()                              \
    do { PyErr_SetString(EximError, get_error_message());  \
         clear_error_context();                            \
         return 1; }                                       \
    while (0)



/*=====================================================================
                     Pyvox type definitions
======================================================================*/



/*----------------------------------------------------------------------
voxtype_repr - Return representation of a Pyvox type
-----------------------------------------------------------------------*/

static
PyObject *                 /* Returns NULL on error */
voxtype_repr(
    pyvox_type *self)      /* Object to return repr() of */
{
    return PyString_FromString(self->repr);
}



/*----------------------------------------------------------------------
voxtype_getattr - Return attribute of a Pyvox type

FIXME: Add other attributes: sumtype, rsumtype, natural?
-----------------------------------------------------------------------*/

static
PyObject *                 /* Returns NULL on error */
voxtype_getattr(
    pyvox_type *self,      /* Object to return attribute of */
    char *attr)            /* Name of the attribute to get */
{
    int code;              /* Numeric code for type */
    int type;              /* Type of a result */
    anytype value;         /* Value of a result */

    code = self->code;

    if ( strcmp(attr, "code" ) == 0 )
	return PyInt_FromLong(code);

    else if ( strcmp(attr, "complex") == 0)
        return PyvoxType_FromInt(exim_complex_type(code));

    else if ( strcmp(attr, "desc") == 0 )
        return PyString_FromString(exim_describe_type(code));

    else if ( strcmp(attr, "epsilon") == 0 )
        return PyFloat_FromDouble(exim_epsilon(code));

    else if ( strcmp(attr, "extype") == 0)
        return PyvoxType_FromInt(exim_natural_extype(code));

    else if ( strcmp(attr, "intype") == 0)
        return PyvoxType_FromInt(exim_natural_intype(code));

    else if ( strcmp(attr, "iscomplex") == 0 )
        return PyInt_FromLong(exim_iscomplex(code));

    else if ( strcmp(attr, "isfloat") == 0 )
        return PyInt_FromLong(exim_isfloat(code));

    else if ( strcmp(attr, "isint") == 0 )
        return PyInt_FromLong(exim_isint(code));

    else if ( strcmp(attr, "isreal") == 0 )
        return PyInt_FromLong(exim_isreal(code));

    else if ( strcmp(attr, "isunsigned") == 0 )
        return PyInt_FromLong(exim_isunsigned(code));

    else if (strcmp(attr, "max") == 0 ) {
        type = exim_max(code, &value);
        if (type != 0)
            return pyvox_get_value(type, &value, 0);
        else {
            PyErr_SetString(PyExc_AttributeError,
                            "'max' not defined for external types");
            return 0;
        }
    }

    else if (strcmp(attr, "min") == 0 ) {
        type = exim_min(code, &value);
        if (type != 0)
            return pyvox_get_value(type, &value, 0);
        else{
            PyErr_SetString(PyExc_AttributeError,
                            "'min' not defined for external types");
            return 0;
        }
    }

    else if ( strcmp(attr, "name") == 0 )
        return PyString_FromString(exim_nameof_type(code));

    else if ( strcmp(attr, "nbytes") == 0 )
        return PyInt_FromLong(exim_sizeof_type(code));

    else if ( strcmp(attr, "real") == 0)
        return PyvoxType_FromInt(exim_real_type(code));

    /* Provide list of members available from this instance */
    /* FIXME: Create once and cache? If so, make it an (immutable) tuple */
    if (strcmp(attr, "__members__") == 0)
	return Py_BuildValue("[ssssssssssssssss]",
                             "code", "complex", "desc",
                             "epsilon", "extype", "intype", "iscomplex", 
                             "isfloat", "isint", "isunsigned", "isreal", 
                             "max", "min", "name", "nbytes", "real");

    /* return Py_FindMethod(voxtype_methods, self, attr); */
    return 0;
}



/*----------------------------------------------------------------------
voxtype type object
-----------------------------------------------------------------------*/

PyTypeObject voxtype_type = {

    /* Type head */
    PyObject_HEAD_INIT(&PyType_Type)
    0,                              /* ob_size:      Nmbr of items */
    "Pyvox type",                   /* tp_name:      Name of type */
    sizeof(pyvox_type),             /* tp_basicsize: Basic size of type */
    0,                              /* tp_itemsize:  Item size */

    /* Standard methods */
    (destructor)  0,                /* tp_dealloc:   Instance destructor */
    (printfunc)   0,                /* tp_print:     "print x" */
    (getattrfunc) voxtype_getattr,  /* tp_getattr:   "instance.attr" */
    (setattrfunc) 0,                /* tp_setattr:   "instance.attr = foo" */
    (cmpfunc)     0,                /* tp_compare:   compare instances */
    (reprfunc)    voxtype_repr,     /* tp_repr:      string representation */

    /* Method suites for standard classes */
    (PyNumberMethods *)   0,        /* tp_as_number */
    (PySequenceMethods *) 0,        /* tp_as_sequence */
    (PyMappingMethods *)  0,        /* tp_as_mapping */

    /* Other methods */
    (hashfunc)     0,               /* tp_hash:  Hash function of instance */
    (ternaryfunc)  0,               /* tp_call:  "instance()" */
    (reprfunc)     0,               /* tp_str:   "str(instance)" */
    (getattrofunc) 0,               /* tp_getattro: */
    (setattrofunc) 0,               /* tp_setattro: */

    (PyBufferProcs *) 0,            /* tp_as_buffer: access as i/o buffer */
    Py_TPFLAGS_DEFAULT,             /* tp_flags: identify optional features */
    0,                              /* tp_doc:   documentation string */

    /* Following fields were added in Python 2.0 */
#ifdef Py_TPFLAGS_HAVE_INPLACEOPS
    (traverseproc) 0,               /* tp_traverse: find all subobjects*/
    (inquiry) 0,                    /* tp_clear: delete subobjects */
#endif

    /* Following fields added in Python 2.1 */
#ifdef Py_TPFLAGS_HAVE_RICHCOMPARE
    (richcmpfunc) 0,                /* tp_richcompare:    rich compare */
    0,                              /* tp_weaklistoffset: weak refs */
#endif

    /* See Include/object.h for still other fields */
};



/*=======================================================================
         Low-level functions for working with type objects
========================================================================*/



/*----------------------------------------------------------------------
pyvox_types - Table of Pyvox internal and external types
-----------------------------------------------------------------------*/

#define MAXTYPES   32          /* Size of table */

static int numtypes = 0;       /* Number of types actually stored */

static struct {
    int code;           /* Numeric code for type */
    PyObject *ob;       /* Python object defining type */
} pyvox_types[MAXTYPES];



/*-----------------------------------------------------------------------
PyvoxIntype_Check - Is a Python object a valid Pyvox internal type?

If the given object is a valid Pyvox internal type, this function returns
1 and stores the numeric code corresponding to that type in the third
argument; otherwise it sets an appropriate error message and returns 0.

NOTE: The type objects were changed from integers to a new extension
type on 2004-11-15 or so; for the nonce, this routine will also accept
the integers.
------------------------------------------------------------------------*/

int                           /* Returns non-zero if valid */
PyvoxIntype_Check(
    PyObject *ob,             /* Purported Pyvox internal type */
    char *name,               /* Name of argument, for error message */
    int  *type)               /* (Out) The corresponding numeric code */
{
    int code = 0;             /* Numeric code for type */

    /* Convert a type object or an int into a numeric code */
    if ( ob != 0 && ob->ob_type == &voxtype_type ) {
        code = ((pyvox_type *)ob)->code;
    }
    else if ( ob != 0 && PyInt_Check(ob) )  {
        code = PyInt_AsLong(ob);
    }

    /* Accept a numeric code that matches an internal type */
    if ( code != 0 && exim_sizeof_intype(code) != 0 ) {
        if (type != 0)    *type = code;
        return 1;
    }

    /* Reject anything else */
    if (name)
        preset_error(ERR_INIT, "%s is not a valid Pyvox internal type", name);
    else
        preset_error(ERR_INIT, "Argument is not a valid Pyvox internal type");
    return 0;
}



/*-----------------------------------------------------------------------
PyvoxExtype_Check - Is a Python object a valid Pyvox external type?

If the given object is a valid Pyvox external type, this function returns
1 and stores the numeric code corresponding to that type in the third
argument; otherwise it sets an appropriate error message and returns 0.

NOTE: The type objects were changed from integers to a new extension
type on 2004-11-15 or so; for the nonce, this routine will also accept
the integers.
------------------------------------------------------------------------*/

int                           /* Returns non-zero if valid */
PyvoxExtype_Check(
    PyObject *ob,             /* Purported Pyvox external type */
    char *name,               /* Name of argument, for error message */
    int  *type)               /* (Out) The corresponding numeric code */
{
    int code = 0;             /* Numeric code for type */

    /* Convert a type object or an int into a numeric code */
    if ( ob != 0 && ob->ob_type == &voxtype_type ) {
        code = ((pyvox_type *)ob)->code;
    }
    else if ( ob != 0 && PyInt_Check(ob) )  {
        code = PyInt_AsLong(ob);
    }

    /* Accept a numeric code that matches an internal type */
    if ( code != 0 && 
         ( exim_sizeof_extype(code) != 0 ||
           code == EXTP_STRING ||
           code == EXTP_NSTR4 ) ) {
        if (type != 0)    *type = code;
        return 1;
    }

    /* DEBUG: Print out the offending type object */
    if (0) {
        fprintf(stderr, "Invalid type object: ");
        PyObject_Print(ob, stderr, 0);
        fprintf(stderr, "\n");
    }

    /* Reject anything else */
    if (name)
        preset_error(ERR_INIT, "%s is not a valid Pyvox external type", name);
    else
        preset_error(ERR_INIT, "Argument is not a valid Pyvox external type");
    return 0;
}



/*-----------------------------------------------------------------------
PyvoxType_FromInt - Return Pyvox internal/external type for given code

This function returns an owned reference to the Pyvox internal or
external type object corresponding to a given numeric code.  If the
numeric code is zero, None is returned.  If the code is invalid but
not zero, and the report argument is not zero, then the error message
is set.

Prior to 2005-08-30, this function returned a borrowed reference.
-------------------------------------------------------------------------*/

PyObject *                       /* Returns NULL on error */
PyvoxType_FromInt(
    int code)                    /* Numeric code for the type */
{
    int i;
    PyObject *result = 0;

    /* Find the type object matching the given code */
    if (code == 0)
        result = Py_None;
    else
        for (i = 0; i < numtypes; i++)
            if ( code == pyvox_types[i].code )
                result = pyvox_types[i].ob;

    /* Report error if requested */
    if (result == 0)
        set_error(ERR_ASSERT, "%d is not a valid Pyvox type", code);
     
    /* Return the result */
    Py_XINCREF(result);
    return result;
}



/*======================================================================
                     Exim module definitions
=======================================================================*/


/*----------------------------------------------------------------------
"Magic" external types not implemented by exim
-----------------------------------------------------------------------*/

static pyvox_type 
extype_string = {
    PyObject_HEAD_INIT(&voxtype_type)
    EXTP_STRING,
    "exim.string"};

static pyvox_type 
extype_nstr4 = {
    PyObject_HEAD_INIT(&voxtype_type)
    EXTP_NSTR4,
    "exim.nstr4"};



/*-----------------------------------------------------------------------
init_external_formats - Initialize Python names of external formats

This function initializes the names which are used to specify external
data formats, adding the names to the module dictionary; in addition,
a Python tuple containing all the external types (except the EXTP_NONE
type) is added to the dictionary under the name 'extypes'.

NOTE: On 2004-11-12, these type objects were changed from integers to
pyvox_type objects.  Watch out for code that assumes they are still ints.

FIXME:  This code assumes that PyDict_SetItemString() does not modify
the string pointed to by name; this is probably true, but not actually
guaranteed in the documentation.
------------------------------------------------------------------------*/

static void
init_external_formats(
    PyObject *dict)        /* Dictionary to add the names to */
{
    PyObject *format;      /* A format being defined */
    PyObject *exlist;      /* Tuple to contain all external types */
    pyvox_type *extype;    /* Pyvox type struct */
    int nex;               /* Number of external types */
    int *extypes;          /* List of external types */
    const char *name;      /* Name of an external type */
    char *repr = 0;        /* String to return as repr() */
    int code = 0;          /* Numeric code for type */
    int i;
    char initfailed[] = "Can't initialize external types in exim";

    /* Get the list of defined extypes from the exim module */
    nex = exim_all_extypes(0,0);
    extypes = mallock(nex*sizeof(int));
    if (extypes == 0)    Py_FatalError(initfailed);
    (void)exim_all_extypes(extypes, nex);

    /* Allocate a tuple to contain all the defined external types;
       size to include 2 "magic" types and omit 1 "none" type */
    exlist = PyTuple_New(nex+2-1);

    /* Define the magic exim.string and exim.nstr4 types */
    format = (PyObject*)(void*)&extype_string;
    PyDict_SetItemString(dict, "string", format);
    PyTuple_SetItem(exlist, nex-1, format);
    if (numtypes+1 >= MAXTYPES)
        Py_FatalError("Overflow in pyvox_types table");
    pyvox_types[numtypes].code = EXTP_STRING;
    pyvox_types[numtypes].ob   = format;
    numtypes++;
    format = (PyObject*)(void*)&extype_nstr4;
    PyDict_SetItemString(dict, "nstr4", format);
    PyTuple_SetItem(exlist, nex, format);
    pyvox_types[numtypes].code = EXTP_NSTR4;
    pyvox_types[numtypes].ob   = format;
    numtypes++;
    
    /* Add each type to the dictionary, and construct a tuple of all
       types;  skip over the EXTP_NONE type */
    for (i = 1; i < nex; i++) 
	if (extypes[i] != 0) {
	    extype = PyObject_NEW(pyvox_type, &voxtype_type);
	    if ( extype == 0 )    Py_FatalError(initfailed);
            extype->code = code = extypes[i];
	    name = exim_nameof_type(code);
	    repr = mallock(strlen("exim.")+strlen(name)+1);
	    if (repr == 0)    Py_FatalError(initfailed);
	    sprintf(repr, "exim.%s", name);
	    extype->repr = repr;
	    format = (PyObject *)extype;
	    PyDict_SetItemString(dict, (char *)name, format);
	    PyTuple_SetItem(exlist, i-1, format);
            if (numtypes >= MAXTYPES)
                Py_FatalError("Overflow in pyvox_types table");
            pyvox_types[numtypes].code = code;
            pyvox_types[numtypes].ob   = format;
            numtypes++;
	}

    /* Add the tuple of all types to the module dictionary */
    PyDict_SetItemString(dict, "extypes", exlist);

    free(extypes);
    return;
}


/*-----------------------------------------------------------------------
init_internal_formats - Initialize Python names of internal formats

This function initializes the names which are used to specify internal
data formats, adding the names to the module dictionary; in addition,
a Python tuple containing all the internal types (except the INTP_NONE
type) is added to the dictionary under the name 'intypes'.

NOTE: On 2004-11-12, these type objects were changed from integers to
pyvox_type objects.  Watch out for code that assumes they are still ints.

FIXME:  This code assumes that PyDict_SetItemString() does not modify
the string pointed to by name; this is probably true, but not actually
guaranteed in the documentation.
------------------------------------------------------------------------*/

static void
init_internal_formats(
    PyObject *dict)        /* Dictionary to add the names to */
{
    PyObject *format;      /* A format being defined */
    PyObject *inlist;      /* Tuple to contain all the internal types */
    pyvox_type *intype;    /* Internal type struct */
    int nin;               /* Number of internal types */
    int *intypes;          /* List of internal types */
    const char *name;      /* Name of an internal type */
    char *repr = 0;        /* String to return as repr() */
    int code = 0;          /* Numeric code for type */
    int i;
    char initfailed[] = "Can't initialize internal types in exim";

    /* Get the list of defined intypes from the exim module */
    nin = exim_all_intypes(0,0);
    intypes = mallock(nin*sizeof(int));
    if (intypes == 0)    Py_FatalError(initfailed);
    (void)exim_all_intypes(intypes, nin);

    /* Add each type to the dictionary and construct a tuple of all types;
       but skip over the INTP_NONE type. */
    inlist = PyTuple_New(nin-1);
    for (i = 1; i < nin; i++) 
	if (intypes[i] != 0) {
	    intype = PyObject_NEW(pyvox_type, &voxtype_type);
	    if ( intype == 0 )    Py_FatalError(initfailed);
            intype->code = code = intypes[i];
	    name = exim_nameof_type(code);
	    repr = mallock(strlen("exim.")+strlen(name)+1);
	    if (repr == 0)    Py_FatalError(initfailed);
	    sprintf(repr, "exim.%s", name);
	    intype->repr = repr;
	    format = (PyObject *)intype;
	    PyDict_SetItemString(dict, (char *)name, format);
	    PyTuple_SetItem(inlist, i-1, format);
            if (numtypes >= MAXTYPES)
                Py_FatalError("Overflow in pyvox_types table");
            pyvox_types[numtypes].code = code;
            pyvox_types[numtypes].ob   = format;
            numtypes++;
	}

    /* Add the tuple of all types to the module dictionary */
    PyDict_SetItemString(dict, "intypes", inlist);

    free(intypes);
    return;
}



/*----------------------------------------------------------------------
pyexim_export - Convert a Python value into some external format

This internal function converts a numeric Python value into some
external data format contained in a byte string.  Conversions between
real and complex types are not supported.

NOTE that it is the user's responsibility to insure that the receiving
buffer is large enough.

FIXME: The results are unpredictable if the Python value is too large
or small to be represented in the destination type. 

FIXME: Modify to support EXTP_STRING and then use in pyexim_pack().

FIXME: This is intended for use only within pyexim.c but might be
worth making more public.  That might also make testing easier.
------------------------------------------------------------------------*/

static int                           /* Returns non-zero on error */
pyexim_export(
    unsigned char *data,             /* Destination for external datum */
    int extype,                      /* External type to convert to */
    int bigend,                      /* Big-endian? */
    PyObject *value)                 /* Python object to be converted */
{
    long lval;                       /* A long value */
    unsigned long uval;              /* An unsigned long value */
    double dval;                     /* A double value */
    Py_complex cval;                 /* A complex value */
    int flag;                        /* Status flag */

    /* Convert a Python int based on the destination type */
    if ( PyInt_Check(value) ) {
        switch (extype) {

        case EXTP_INT1:
        case EXTP_INT2:
        case EXTP_INT4:
            lval = PyInt_AsLong(value);
            flag = exim_export(1, data, extype, 0, bigend, 
                               &lval, INTP_LONG, 0);
            break;

        case EXTP_UINT1:
        case EXTP_UINT2:
        case EXTP_UINT4:
            lval = PyInt_AsLong(value);
            if (lval < 0)
                XReturnEximError(ERR_RANGE, "Value not representable");
            flag = exim_export(1, data, extype, 0, bigend, 
                               &lval, INTP_ULONG, 0);
            break;

        case EXTP_FLOAT4:
        case EXTP_FLOAT8:
            dval = (double)PyInt_AsLong(value);
            flag = exim_export(1, data, extype, 0, bigend, 
                               &dval, INTP_DOUBLE, 0);
            break;

        default:
            XReturnEximError(ERR_UNIMP, "Export conversion not implemented");
        }

        if (flag)    XReturnEximLibError();
    }

    /* Convert a Python long int based on the destination type */
    else if ( PyLong_Check(value) ) {
        switch (extype) {

        case EXTP_INT1:
        case EXTP_INT2:
        case EXTP_INT4:
            lval = PyLong_AsLong(value);
            flag = exim_export(1, data, extype, 0, bigend, 
                               &lval, INTP_LONG, 0);
            break;

        case EXTP_UINT1:
        case EXTP_UINT2:
        case EXTP_UINT4:
            uval = PyLong_AsUnsignedLong(value);
            flag = exim_export(1, data, extype, 0, bigend, 
                               &uval, INTP_ULONG, 0);
            break;

        case EXTP_FLOAT4:
        case EXTP_FLOAT8:
            dval = PyLong_AsDouble(value);
            flag = exim_export(1, data, extype, 0, bigend, 
                               &dval, INTP_DOUBLE, 0);
            break;

        default:
            XReturnEximError(ERR_UNIMP, "Export conversion not implemented");
        }

        if (flag)    XReturnEximLibError();
    }

    /* Convert a Python float */
    else if ( PyFloat_Check(value) ) {
        switch (extype) {

        case EXTP_INT1:
        case EXTP_INT2:
        case EXTP_INT4:
            lval = (long)PyFloat_AsDouble(value);
            flag = exim_export(1, data, extype, 0, bigend, 
                               &lval, INTP_LONG, 0);
            break;

        case EXTP_UINT1:
        case EXTP_UINT2:
        case EXTP_UINT4:
            uval = (unsigned long)PyFloat_AsDouble(value);
            flag = exim_export(1, data, extype, 0, bigend, 
                               &uval, INTP_ULONG, 0);
            break;

        case EXTP_FLOAT4:
        case EXTP_FLOAT8:
            dval = PyFloat_AsDouble(value);
            flag = exim_export(1, data, extype, 0, bigend, 
                               &dval, INTP_DOUBLE, 0);
            break;

        default:
            XReturnEximError(ERR_UNIMP, "Export conversion not implemented");
        }

        if (flag)    XReturnEximLibError();
    }

    /* Convert a Python complex value */
    else if ( PyComplex_Check(value) ) {

        switch (extype) {

	case EXTP_COMPLEX8  :
	case EXTP_COMPLEX16 :
	    cval = PyComplex_AsCComplex(value);
	    flag = exim_export(1, data, extype, 0, bigend,
			       &cval, INTP_DCOMPLEX, 0);
	    break;

        default:
            XReturnEximError(ERR_UNIMP, "Export conversion not implemented");
        }

        if (flag)    XReturnEximLibError();
    }

    /* Anything else is an error */
    else
        XReturnEximError(ERR_ILLOP, 
                         "No conversion defined for this Python type");

    return 0;
}



/*------------------------------------------------------------------------
pyexim_import - Convert a byte string to a Python value

This internal function converts an external datum of specified type
contained in a byte string into a Python value.  Conversions between
real and complex types are not supported.

FIXME: This function is intended for use only inside pyexim.c but
might be worth making more public.  That might also make testing easier.
-------------------------------------------------------------------------*/

static PyObject *                /* Returns zero on error */
pyexim_import(
    unsigned char *data,         /* The bytes to convert */
    int extype,                  /* The external type of the data */
    int bigend)                  /* Big endian? */
{
    PyObject *value;             /* A Python value */
    long longval;                /* A long value */
    unsigned long ulval;         /* An unsigned long value */
    double dblval;               /* A double value */
    Py_complex cval;             /* A complex value */
    int flag;                    /* Error status */

    /* Switch according to the source type */
    switch (extype) {

        /* All signed integer types to a Python integer */
        case EXTP_INT1:
        case EXTP_INT2:
        case EXTP_INT4:
	    flag = exim_import(1, &longval, INTP_LONG, 1, 
			       data, extype, 1, bigend);
	    if (flag)
	        ReturnEximError(ERR_MISC, "Error importing external data");
	    value = Py_BuildValue("l", longval);
	    if (value == 0)
	        ReturnEximError(ERR_MISC, "Error converting external data");
	    return value;

        /* All unsigned integer types to a Python integer or long */
        case EXTP_UINT1:
        case EXTP_UINT2:
        case EXTP_UINT4:
	    flag = exim_import(1, &ulval, INTP_ULONG, 1, 
			       data, extype, 1, bigend);
	    if (flag)
	        ReturnEximError(ERR_MISC, "Error importing external data");
	    if (ulval > LONG_MAX) 
	        value = PyLong_FromUnsignedLong(ulval);
	    else
	        value = Py_BuildValue("l", ulval);
	    if (value == 0)
	        ReturnEximError(ERR_MISC, "Error converting external data");
	    return value;

	/* All float types convert to a Python float */
        case EXTP_FLOAT4:
        case EXTP_FLOAT8:
	    flag = exim_import(1, &dblval, INTP_DOUBLE, 1, 
			       data, extype, 1, bigend);
	    if (flag)
	        ReturnEximError(ERR_MISC, "Error importing external data");
	    value = Py_BuildValue("d", dblval);
	    if (value == 0)
	        ReturnEximError(ERR_MISC, "Error converting external data");
	    return value;

	/* All complex types convert to a Python complex */
        case EXTP_COMPLEX8:
        case EXTP_COMPLEX16:
	    flag = exim_import(1, &cval, INTP_DCOMPLEX, 1, 
			       data, extype, 1, bigend);
	    if (flag)
	        ReturnEximError(ERR_MISC, "Error importing external data");
	    value = PyComplex_FromCComplex(cval);
	    if (value == 0)
	        ReturnEximError(ERR_MISC, "Error converting external data");
	    return value;

	/* Report an error for unimplemented types */
        default:
	    ReturnEximError(ERR_UNIMP, 
			    "Conversion from given type not implemented");
    }

}



/*----------------------------------------------------------------------
exim.dump - Dump a dict described by a format description

Given a description of the format of an (external) struct and a
dictionary containing the values of the various fields, this method
dumps those values to standard output in a (more or less) humanly
readable format.  The format description is as described below for
exim.pack.  The value None is returned.

FIXME: Print a header giving the type of the struct being dumped?
------------------------------------------------------------------------*/

static PyObject *             /* Returns zero on error */
pyexim_dump(
    PyObject *self,
    PyObject *args)
{
    PyObject *dict;           /* Arg 1: Dictionary: field -> value */
    PyObject *desc;           /* Arg 2: Format description */
    PyObject *pyfile = 0;     /* Arg 3: Optional file to write to */
    PyObject *result;         /* Result: Always None */

    FILE *file = stdout;      /* File to write to */
    int nfields = 0;          /* Number of fields in the description */
    int ifld;                 /* Index over the fields */
    PyObject *fdesc;          /* Field descriptor */
    PyObject *name;           /* Name of field */
    PyObject *value;          /* Value of a field */
    int len, maxlen = 0;      /* Length of current and longest field names */
    int flag;                 /* Status flag */


    set_error_context("exim.dump()");

    /* Fetch the arguments */
    if ( ! PyArg_ParseTuple( args, "OO|O", &dict, &desc, &pyfile) )
        return NULL;

    /* Check that the descriptor is at least a sequence */
    if ( !PySequence_Check(desc) )
        ReturnEximError(ERR_INIT, "Desc argument is not a sequence");
    nfields = PySequence_Length(desc);

    /* Check that the dict is really a dictionary */
    if ( ! PyDict_Check(dict) )
        ReturnEximError(ERR_INIT, "Dict argument is not a dictionary");

    /* Check the file argument, if any */
    if ( pyfile && PyFile_Check(pyfile) )
        file = PyFile_AsFile(pyfile);
    else if ( pyfile )
        ReturnEximError(ERR_INIT, "Invalid file argument");
    else
        file = stdout;

    /* Pass 1: Loop over the fields to determine longest name */
    for (ifld = 0; ifld < nfields; ifld++) {
        fdesc = PySequence_GetItem(desc, ifld);
        if ( ! PySequence_Check(fdesc) )
            ReturnEximError(ERR_INIT, "Field descriptor is not a sequence");
	name = PySequence_GetItem(fdesc, 0);
        len  = PyString_Size(name);
        if (len > maxlen) maxlen = len;
    }

    /* Pass 2: Loop over the fields in the description and print contents */
    for (ifld = 0; ifld < nfields; ifld++) {

        /* Get name and value of each field */
        fdesc = PySequence_GetItem(desc, ifld);
        if ( ! PySequence_Check(fdesc) )
            ReturnEximError(ERR_INIT, "Field descriptor is not a sequence");
	name  = PySequence_GetItem(fdesc, 0);
	value = PyDict_GetItem(dict, name);    /* 0 => absent? */

	/* Print out the name and format */
	flag = PyObject_Print(name, file, Py_PRINT_RAW);
        for (len = PyString_Size(name) ; len < maxlen; len++)
            fprintf(file, " ");
	fprintf(file, " : ");
	if (value)
	  flag = PyObject_Print(value, file, 0);
	else
	  fprintf(file, "(undefined)");
	fprintf(file, "\n");
    }

    /* Return None as the result */
    result = Py_None;
    Py_INCREF(result);

    clear_error_context();
    return result;
}



/*-----------------------------------------------------------------------
exim.pack - Pack contents of a dict into a struct

Given a description of the format of an (external) struct and a
dictionary containing the values of the various fields, this method
constructs the struct in external format and returns it as a Python
string.  The format description is a list or tuple of field
descriptions, each of which is a three- or four-element list or tuple;
each field description contains the name of the field as a string
(which will also be used as the key into the dictionary), its default
value or None, the external type of the field, and the number of
elements in the field.  Each value in the dictionary must be a number
or Pyvox array of a compatible internal type, or a string.  

The number of elements may also be a string matching the name of a
field previously seen in the struct; this means to use the value of that
field.  This feature currently works only for exim.unpack.

There are two special types.  The special type exim.string indicates that 
the value must be a string; in this case, the number of elements is the 
number of characters in the string.  The special type exim.nstr4 indicates 
that the value is a string stored in the NRIA file format; the number of 
elements is the number of such strings.

If the default value is None, then the dictionary must contain a value for 
that field; otherwise the default value will be used when the dictionary 
defines no value for that field.

FIXME: If the platform has 16-bit ints, this code is limited to
structs no longer than 32767 bytes.

FIXME: Fields of type EXTP_NSTR4 are known to work for exim.pack, but
have never been tested for exim.unpack.

FIXME: In some applications it would be useful to write into (or append to) 
a buffer at some specified location, and to return the number of bytes written.
------------------------------------------------------------------------*/

static PyObject *             /* Returns zero on error */
pyexim_pack(
    PyObject *self,
    PyObject *args)
{
    PyObject *dict;           /* Arg 1: Dictionary: field -> value */
    PyObject *desc;           /* Arg 2: Format descriptor */
    int bigend = 1;           /* Arg 3: Big-endian? */
    PyObject *result = 0;     /* Result: Converted values as string */

    int nfields = 0;          /* Number of fields in the descriptor */
    int ifld;                 /* Index over fields */
    long length = 0;          /* Number of bytes needed for output string */
    unsigned char *data = 0;  /* Buffer for the output string */
    long idata = 0;           /* Index to next empty char of buffer */
    PyObject *fdesc;          /* Field descriptor */
    int nfdesc;               /* Length of field descriptor */
    PyObject *name;           /* Name of field */
    PyObject *dflt;           /* Default value for field */
    PyObject *type;           /* External type of field */
    int extype;               /* External type of field */
    long nelem;               /* Number of elements in field */
    long ielem;               /* Index over elements of afield */
    long nbytes;              /* Number of bytes in current element */
    PyObject *value;          /* Value or values to be stored in field */
    PyObject *val;            /* Single value from a sequence */
    char *sval;               /* Value as a string */
    int len;                  /* Length of a string */
    int flag;                 /* Status flag */

    set_error_context("exim.pack()");

    /* Fetch the arguments */
    if ( ! PyArg_ParseTuple( args, "OO|i", &dict, &desc, &bigend) )
        return NULL;

    /* Check that the descriptor is at least a sequence */
    if ( !PySequence_Check(desc) )
        ReturnEximError(ERR_INIT, "Desc argument is not a sequence");
    nfields = PySequence_Length(desc);

    /* Check that the dict is really a dictionary */
    if ( ! PyDict_Check(dict) )
        ReturnEximError(ERR_INIT, "Dict argument is not a dictionary");

    /* Pass 1: Determine total length of output string */
    length = 0;
    for (ifld = 0; ifld < nfields; ifld++) {

        /* Get name, type, and nelem of each field */
        fdesc = PySequence_GetItem(desc, ifld);
        if ( ! PySequence_Check(fdesc) )
            ReturnEximError(ERR_INIT, "Field descriptor is not a sequence");
        nfdesc = PySequence_Length(fdesc);
        if (nfdesc < 3)
            ReturnEximError(ERR_INIT, "Field descriptor is too short");
	name  = PySequence_GetItem(fdesc, 0);
	type  = PySequence_GetItem(fdesc, 2);
        if ( !PyvoxExtype_Check(type, "field type", &extype) )
            ReturnEximLibError();
	nelem = 1;
	if (nfdesc >= 4)
	    nelem = PyInt_AsLong(PySequence_GetItem(fdesc, 3));

        /* Add length of each field to get total length of output string */
	if ( extype == EXTP_STRING )
	    nbytes = 1;
        else if ( extype == EXTP_NSTR4 ) {
            if ( nelem != 1 )
                ReturnEximError(ERR_INIT, "NSTR4 field must have count == 1");
            value = PyDict_GetItem(dict, name);
            if ( !PyString_Check(value) )
              ReturnEximError(ERR_INIT, "Data for NSTR4 field isn't string");
            nbytes = 4 + PyString_Size(value);
        }
	else {
	    nbytes = exim_sizeof_extype(extype);
	    if (nbytes == 0)
	        ReturnEximError(ERR_INIT, "Invalid external type");
	}
        length += nbytes * nelem;
        
    }

    /* Pass 2: Allocate buffer and convert arguments to string */
    data  = mallock(length);
    idata = 0;
    for (ifld = 0; ifld < nfields; ifld++) {

        /* Get name, default, type, and nelem of each field */
        fdesc = PySequence_GetItem(desc, ifld);
        if ( ! PySequence_Check(fdesc) )
            ReturnEximError(ERR_INIT, "Field descriptor is not a sequence");
        nfdesc = PySequence_Length(fdesc);
        if (nfdesc < 3)
            ReturnEximError(ERR_INIT, "Field descriptor is too short");
	name  = PySequence_GetItem(fdesc, 0);
        dflt  = PySequence_GetItem(fdesc, 1);
	type  = PySequence_GetItem(fdesc, 2);
        if ( !PyvoxExtype_Check(type, "type", &extype) )
            ReturnEximLibError();
	nelem = 1;
	if (nfdesc >= 4)
	    nelem = PyInt_AsLong(PySequence_GetItem(fdesc, 3));

        /* Check for valid type and determine bytes needed for each element */
	if ( extype == EXTP_STRING )
	    nbytes = 1;
        else if ( extype == EXTP_NSTR4 ) {
            value = PyDict_GetItem(dict, name);
            if ( !PyString_Check(value) )
              ReturnEximError(ERR_INIT, "Data for NSTR4 field isn't string");
            nbytes = 4 + PyString_Size(value);
        }
	else {
	    nbytes = exim_sizeof_extype(extype);
	    if (nbytes == 0)
	        ReturnEximError(ERR_INIT, "Invalid external type");
	}

        /* Check for buffer overflow (just in case) */
        if ( idata + nelem*nbytes > length)
            ReturnEximError(ERR_ASSERT, "Buffer overflow");

        /* Fetch the current value from the dict, or take the default */
        /* FIXME: Name the field in the "no default allowed" message */
        value = PyDict_GetItem(dict, name);
        if (value == 0)
            value = dflt;
        else if (value == Py_None)
            ReturnEximError(ERR_INIT, "No default allowed for field");

        /* DEBUG: Report name and value */
        if (0) {
            fprintf(stderr, "Field, value, and type: ");
            PyObject_Print(name,  stderr, 0);
            fprintf(stderr, " , ");
            PyObject_Print(value, stderr, 0);
            fprintf(stderr, ", %d\n", extype);
        }

        /* Convert a Python string into a plain external string */
        if ( PyString_Check(value) &&  extype == EXTP_STRING ) {
            sval = PyString_AsString(value);
            len  = strlen(sval);
            if (len == 1 || len == 0)
                for (ielem = 0; ielem < nelem; ielem++)
                    data[idata++] = sval[0];
            else
                for (ielem = 0; ielem < nelem; ielem++)
                    data[idata++] = (ielem < len) ? sval[ielem] : 0;
        }
        
        /* Convert a Python string into an EXTP_NSTR4 string */
        else if ( PyString_Check(value) &&  extype == EXTP_NSTR4 ) {
            sval = PyString_AsString(value);
            len  = strlen(sval);
            flag = exim_export(1, &data[idata], EXTP_UINT4, 0, bigend, 
                               &len, INTP_INT, 0);
            if (flag)    ReturnEximLibError();
            idata += 4;
            for (ielem = 0; ielem < len; ielem++)
                data[idata++] = sval[ielem];
        }

        /* Convert a Python number into one or more elements of a field */
        else if ( PyNumber_Check(value) && 
                  (nbytes = exim_sizeof_extype(extype)) != 0 ) {
            for (ielem = 0; ielem < nelem; ielem++) {
                flag = pyexim_export(&data[idata], extype, bigend, value);
                if (flag)    ReturnEximLibError();
                idata += nbytes;
            }
        }

        /* Convert a sequence of numbers into elements of a field */
        else if ( PySequence_Check(value) &&
                  (nbytes = exim_sizeof_extype(extype)) != 0 ) {
            len = PySequence_Length(value);
            for (ielem = 0; ielem < nelem; ielem++) {
                if ( ielem < len )
                    val = PySequence_GetItem(value, ielem);
                else if ( PyNumber_Check(dflt) )
                    val = dflt;
                else if ( PySequence_Check(dflt) )
                    val = PySequence_GetItem(dflt, ielem);
                else if ( dflt == Py_None )
                    ReturnEximError(ERR_INIT, "Field may not be defaulted");
                else
                    val = 0;
                if (val == 0)
                    ReturnEximError(ERR_INIT, "Invalid default value");
                flag = pyexim_export(&data[idata], extype, bigend, val);
                if (flag)    ReturnEximLibError();
                idata += nbytes;
            }
        }

        /* Otherwise report an impossible conversion */
        else 
            ReturnEximError(ERR_INIT, "Impossible conversion requested");

    }

    /* Prepare and return the result */
    /* NOTE: The call below expects data to be plain char*, but other
       Pyvox calls expect unsigned char *; hence the cast. */
    result = PyString_FromStringAndSize((char *)data, length);

    clear_error_context();
    return result;
}



/*------------------------------------------------------------------------
exim.unpack - Unpack a struct into a dict

Given an (external) struct as a string and a description of its
format, this method unpacks the struct into a dictionary wherein each
field name is a key to the corresponding value.  See exim.pack just
above for details of the format description.  Each element will be
returned as either a number of the indicated type, a Pyvox array, or a
string.  The special type exim.string indicates that the field is to
be returned as a string.

FIXME: The special types exim.string and exim.nstr4 are rather
oddballs; look for ways to regularize them.

FIXME: There are applications in which it would be useful to start at
a given position within a buffer and to report back the number of
bytes read.

FIXME: Here's a bizarre thought.  Create special types for comma,
space, or tab separated values; then you can read those files as well.

FIXME: In the platform has 16-bit ints, this code is limited to
structs no longer than 32767 bytes.
--------------------------------------------------------------------------*/

static PyObject *
pyexim_unpack(
    PyObject *self,
    PyObject *args)
{
    unsigned char *data;         /* Arg 1: Data to unpack */
    int ndata;                   /* ...and its length */
    PyObject *desc;              /* Arg 2: Description of struct format */
    int bigend = 1;              /* Arg 3: Big-endian?  Defaults to yes */

    PyObject *dict;              /* Dictionary containing fields and values */
    PyObject *value = 0;         /* Value of a field, to add to dict */
    PyObject *item;              /* Item in a value which is a list */
    int nfields;                 /* Number of fields in description */
    PyObject *desc1;             /* Description of a single field */
    int ndesc1;                  /* Nmbr elements in desc1 */
    int idata;                   /* Index to next unused byte of data */
    char *name;                  /* Name of current field */
    PyObject *type;              /* External type of current field */
    int extype;                  /* External type of current field */
    long nelem;                  /* Number of elements in current field */
    long ielem;                  /* Index over the elements */
    long nbytes;                 /* Number of bytes in current element */
    int ifld;                    /* Index over the fields */
    unsigned long len;           /* Length of a string */
    int flag;                    /* Error flag */

    /* Error messages used more than once */
    char overrun[] = "Insufficient data for given fields";

    set_error_context("exim.unpack()");

    /* Fetch all the arguments */
    if ( ! PyArg_ParseTuple(args, "s#O|i", &data, &ndata, &desc, &bigend) )
        return NULL;

    /* Create the dict to store results in */
    dict = PyDict_New();

    /* Check that the descriptor is at least a sequence */
    if ( !PySequence_Check(desc) )
        ReturnEximError(ERR_INIT, "Desc argument is not a sequence");
    nfields = PySequence_Length(desc);

    /* Scan the description to unpack the struct into the dict */
    idata = 0;
    for (ifld = 0; ifld < nfields; ifld++) {

        /* Get name, type, and length of the next field; ignore default */
        desc1 = PySequence_GetItem(desc, ifld);
	if ( !PySequence_Check(desc1) )
	    ReturnEximError(ERR_INIT, "Field description is not a sequence");
	ndesc1 = PySequence_Length(desc1);
	if (ndesc1 < 3)
  	    ReturnEximError(ERR_INIT, "Field description is too short");
	name = PyString_AsString(PySequence_GetItem(desc1, 0));
	type = PySequence_GetItem(desc1, 2);
        if ( !PyvoxExtype_Check(type, "type", &extype) )
            ReturnEximLibError();

        /* Get number of elements, if provided; look up a string in the dict */
	nelem = 1;
	if (ndesc1 >= 4) {
            item = PySequence_GetItem(desc1, 3);
            if ( PyString_Check(item) )
                item = PyDict_GetItem(dict, item);
            if ( item != 0  &&  PyInt_Check(item) )
                nelem = PyInt_AsLong(item);
            else 
                ReturnEximError(ERR_INIT, "Invalid number of elements");
        }

        /* DEBUG: Show current name and type */
        if (0) {
          fprintf(stderr, "Name = %s, type = %d\n", name, extype);
        }

	/* Check for valid type and determine number of bytes per element */
	if ( extype == EXTP_STRING )
	    nbytes = 1;
        else if ( extype == EXTP_NSTR4 ) {
            /* Bypass the usual check below for enough bytes, and check 
               during conversion instead */
            nbytes = 0;
        }
	else {
	    nbytes = exim_sizeof_extype(extype);
	    if (nbytes == 0)
	        ReturnEximError(ERR_INIT, "Invalid external type");
	}

        /* Check that we still have enough data left */
	if ( idata + nelem*nbytes > ndata )
	    ReturnEximError(ERR_INCOMP, overrun);

        /* Convert a string into a Python object */
	if (extype == EXTP_STRING) {
	    value = PyString_FromStringAndSize((char *)&data[idata], nelem);
	    if (value == 0)
	        ReturnEximError(ERR_MISC, "Unable to convert string");
	    idata += nelem;
	}

        /* Convert NSTR4 strings into a list of Python strings, or a single
           NSTR4 string to a single Python string */
        else if (extype == EXTP_NSTR4) {
            if (nelem != 1)
                value = PyList_New(nelem);
            for (ielem = 0; ielem < nelem; ielem++) {
                if ( idata + 4 > ndata )
                    ReturnEximError(ERR_INCOMP, overrun); 
                flag = exim_import(1, &len, INTP_ULONG, 0, 
                                   &data[idata], EXTP_UINT4, 0, bigend);
                if (flag)    ReturnEximLibError();
                idata += 4;
                if ( idata + len > ndata )
                    ReturnEximError(ERR_INCOMP, overrun);
	        item = PyString_FromStringAndSize((char *)&data[idata], len);
	        idata += len;
	        if (item == 0)
	            ReturnEximError(ERR_MISC, 
                                    "Unable to convert NSTR4 string");
                if (nelem == 1) {
                    value = item;
                    break;
                }
                flag = PyList_SetItem(value, ielem, item);
		if (flag)
		    ReturnEximError(ERR_ASSERT, "Unable to insert data value");
            }
        }

	/* Convert a singleton element into a Python number */
	else if (nelem == 1) {
  	    value = pyexim_import(&data[idata], extype, bigend);
	    if (value == 0)
	        ReturnEximLibError();
	    idata += nbytes;
	}

	/* Convert more than one element into a list of Python numbers */
	else {
	    value = PyList_New(nelem);
            if (value == 0)
                ReturnEximError(ERR_ASSERT, "Unable to create a list");
	    for (ielem = 0; ielem < nelem; ielem++) {
	        item = pyexim_import(&data[idata], extype, bigend);
		if (item == 0)
		    ReturnEximLibError();
		flag = PyList_SetItem(value, ielem, item);
		if (flag)
		    ReturnEximError(ERR_ASSERT, "Unable to insert data value");
		idata += nbytes;
	    }
	}

	/* Add the converted value to the dict */
	PyDict_SetItemString(dict, name, value);
    }

    clear_error_context();
    return dict;
}



/*-------------------------------------------------------------------------
exim_methods - Methods implemented by the Exim module
-------------------------------------------------------------------------*/

static struct PyMethodDef
exim_methods[] = {
    { "dump",         pyexim_dump,    1,   "Dump contents of a struct" },
    { "pack",         pyexim_pack,    1,   "Pack dict into struct"   },
    { "unpack",       pyexim_unpack,  1,   "Unpack struct into dict" },
    { NULL, NULL }
};



/*--------------------------------------------------------------------------
myinitexim - Initialize the Exim module

This function is called from initexim (in eximC.c) and does all the real 
initialization of the exim module.  Things are done this way so that exim
and pyvox can share internals.
--------------------------------------------------------------------------*/

void
myinitexim()
{
    PyObject *module;        /* The Exim module being created */
    PyObject *dict;          /* ...and its dictionary */

    module = Py_InitModule("exim", exim_methods);
    dict = PyModule_GetDict(module);
    EximError = Py_BuildValue("s", "exim.error");
    PyDict_SetItemString(dict, "error", EximError);
    init_external_formats(dict);
    init_internal_formats(dict);
    if ( PyErr_Occurred() )
        Py_FatalError("Can't initialize exim module");

    /* DEBUG: Show the Pyvox types table */
    if (0) {
        int i;
        fprintf(stderr, "Pyvox types table:\n");
        for (i = 0; i < numtypes; i++) {
            fprintf(stderr, "%4d  ", pyvox_types[i].code);
            PyObject_Print(pyvox_types[i].ob, stderr, 0);
            fprintf(stderr, "\n");
        }
    }

    return;
}
