/****************************************************************************
pyvox.c - Python bindings for the Voxel Kit

Author:  Paul Hughett

Copyright (c) 1999-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This file contains code to provide the Voxel Kit as an extension to
Python, somewhat along the lines of PIL or Numeric Python.  

NOTE that, while this file is named pyvox.c, the module that it
creates is named pyvoxC; the module pyvox refers to a Python-language
upper level that imports the pyvoxC module.

FIXME:  The ordering of declarations and definitions in this file seems
to be rather delicate and in some places counter-intuitive.  Look for
a better way to do things.

FIXME: PyObjects may also be pyvox_arrays or kernel_objects in
disguise; the current practice is to pass arguments as PyObjects and
cast as needed inside the function but this is sometimes confusing and
eliminates some error checking.  Is there a better way?
*****************************************************************************/

static const char rcsid[] = "$Id: pyvox.c,v 1.130 2006/02/24 21:33:41 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include "Python.h"    /* Must precede standard C headers */

#include <stdio.h>
#include <string.h>
#include <math.h>

#ifdef HAVE_COMPLEX_H
#include <complex.h>
#endif

#include "bips.h"
#include "errm.h"
#include "exim.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"       /* For vxli_count & internal kernel functions */
#include "pyvox.h"      /* Must follow voxel.h */




/*************************************************************************
Some basic objects and functions

FIXME: Some of these functions are likely to be useful in more than
just Pyvox and should be considered for promotion to a more general
library file.
*************************************************************************/



/*-----------------------------------------------------------------------
pyvox_dict - Dictionary for the pyvox module

This global variable is set to point to the dictionary of the pyvox
module (the Python level, not the C level in pyvoxC) when that module
is initialized.  It can be used from the C level to find and call
functions and methods that are defined on the Python level.
-------------------------------------------------------------------------*/

PyObject *pyvox_dict = 0;


/*------------------------------------------------------------------------
PyvoxError - Unique error object for Pyvox to signal an exception

FIXME: This is currently defined to be a string; is it useful to turn it
into an exception class instead?
------------------------------------------------------------------------*/

PyObject *PyvoxError = 0;



/*************************************************************************
Argument checking and data conversion functions
**************************************************************************/



/*------------------------------------------------------------------------
pyvox_get_value - Convert a C value to a Python value

This function converts a C value of specified internal type into a
Python value of the most appropriate type.  The normal use is to fetch
a value from a dense homogeneous array such as the data area of a
voxel array, but it is also permissible to fetch a single value from
a C scalar variable or an anytype object with index = 0.

The Python value returned is an owned reference, so it is not normally
necessary to increment its reference count.

See also parray_get_voxel() to fetch the element at specified coordinates 
in a voxel array.
-------------------------------------------------------------------------*/

PyObject *                       /* Returns zero on error */
pyvox_get_value(
    int intype,                  /* Internal type of the C datum */
    void *data,                  /* Start of a densely packed C array */
    long index)                  /* Index to element to be converted */
{
    PyObject *value;             /* A Python value */
    void *datum = 0;             /* Pointer to the element to convert */
    unsigned long ulval;         /* An unsigned long value */

    /* Compute a pointer to the requested datum */
    datum = bips_locate(intype, data, index);
    if (datum == 0)
        ReturnPyvoxError(ERR_ASSERT, "Failed to locate requested datum");

    /* Switch according to the source type */
    switch (intype) {

        /* All signed integer types to a Python integer */
        case INTP_SCHAR:
            value = PyInt_FromLong(*(signed char *)datum);
            break;

        case INTP_SHORT:
            value = PyInt_FromLong(*(short *)datum);
            break;

        case INTP_INT:
            value = PyInt_FromLong(*(int *)datum);
            break;

        case INTP_LONG:
            value = PyInt_FromLong(*(long *)datum);
            break;

        /* All unsigned integer types to a Python integer; except that
           unsigned ints or longs that won't fit convert to a Python long */
        case INTP_UCHAR:
            value = PyInt_FromLong(*(unsigned char *)datum);
            break;

        case INTP_USHORT:
            value = PyInt_FromLong(*(unsigned short *)datum);
            break;

        case INTP_UINT:
	    /* This must work correctly whether int and long are the same
	       size or different.  Think before you change this code. */
            ulval = *(unsigned int *)datum;
	    if (ulval > LONG_MAX) 
	        value = PyLong_FromUnsignedLong(ulval);
	    else
	        value = PyInt_FromLong(ulval);
            break;

        case INTP_ULONG:
            ulval = *(unsigned long *)datum;
	    if (ulval > LONG_MAX) 
	        value = PyLong_FromUnsignedLong(ulval);
	    else
	        value = PyInt_FromLong(ulval);
            break;

	/* All float types convert to a Python float */
        case INTP_FLOAT:
            value = PyFloat_FromDouble(*(float *)datum);
            break;

        case INTP_DOUBLE:
            value = PyFloat_FromDouble(*(double *)datum);
            break;

#ifdef HAVE_COMPLEX_H
	/* All complex types convert to a Python complex */
        case INTP_FCOMPLEX: {
            float  *fp = datum;
	    value = PyComplex_FromDoubles(fp[0], fp[1]);
            break;
	}

        case INTP_DCOMPLEX: {
	    double *dp = datum;
	    value = PyComplex_FromDoubles(dp[0], dp[1]);
            break;
	}
#endif

	/* Report an error for unimplemented types */
        default:
	    ReturnPyvoxError(ERR_UNIMP, 
			     "Conversion from given type not implemented");
    }

    /* Check for valid result; generate error or return the value */
    if (value == 0)
        ReturnPyvoxError(ERR_MISC, "Error converting C to Python");
    return value;
}



/*----------------------------------------------------------------------
pyvox_set_value - Convert a Python value into a C value

This internal function converts a numeric Python value into a C value
of some specified internal type.  The normal use is to set a value in
a dense homogeneous array such as the data area of a voxel_array, but
it is also permissible to set a single value in a C scalar variable or
an anytype object with index = 0.

FIXME: The results are unpredictable if the Python value is too large
or small to be represented in the destination type. 
------------------------------------------------------------------------*/

int                                  /* Returns non-zero on error */
pyvox_set_value(
    int intype,                      /* Internal type to convert to */
    void *data,                      /* Densely packed array of C values */
    long index,                      /* Index to desired element to be set */
    PyObject *value)                 /* Python object to be converted */
{
    void *datum = 0;                 /* Element to be set */
    long lval;                       /* A long value */
    unsigned long uval;              /* An unsigned long value */
    double dval;                     /* A double value */
    int flag;                        /* Status flag */

    /* Convert a Python int based on the destination type */
    datum = bips_locate(intype, data, index);
    if ( PyInt_Check(value) ) {
        switch (intype) {

        case INTP_SCHAR:
        case INTP_SHORT:
        case INTP_INT:
        case INTP_LONG:
        case INTP_UCHAR:
        case INTP_USHORT:
        case INTP_UINT:
        case INTP_ULONG:
        case INTP_FLOAT:
        case INTP_DOUBLE:
        case INTP_FCOMPLEX:
        case INTP_DCOMPLEX:
            lval = PyInt_AsLong(value);
            flag = bips_cast(1, intype, datum, 1, INTP_LONG, &lval, 1);
            break;

        default:
            XReturnPyvoxError(ERR_UNIMP, "Unimplemented Python->C conversion");
        }

        if (flag)    XReturnPyvoxLibError();
    }

    /* Convert a Python long int based on the destination type */
    /* FIXME: I think there are reasonable cases that this will handle
       incorrectly. */
    else if ( PyLong_Check(value) ) {
        switch (intype) {

        case INTP_SCHAR:
        case INTP_SHORT:
        case INTP_INT:
        case INTP_LONG:
            lval = PyLong_AsLong(value);
            flag = bips_cast(1, intype, datum, 1, INTP_LONG, &lval, 1);
            break;

        case INTP_UCHAR:
        case INTP_USHORT:
        case INTP_UINT:
        case INTP_ULONG:
            uval = PyLong_AsUnsignedLong(value);
            flag = bips_cast(1, intype, datum, 1, INTP_ULONG, &uval, 1);
            break;

        case INTP_FLOAT:
        case INTP_DOUBLE:
        case INTP_FCOMPLEX:
        case INTP_DCOMPLEX:
            dval = PyLong_AsDouble(value);
            flag = bips_cast(1, intype, datum, 1, INTP_DOUBLE, &dval, 1);
            break;

        default:
            XReturnPyvoxError(ERR_UNIMP, "Unimplemented Python->C conversion");
        }

        if (flag)    XReturnPyvoxLibError();
    }

    /* Convert a Python float */
    else if ( PyFloat_Check(value) ) {
        switch (intype) {

        case INTP_SCHAR:
        case INTP_SHORT:
        case INTP_INT:
        case INTP_LONG:
        case INTP_UCHAR:
        case INTP_USHORT:
        case INTP_UINT:
        case INTP_ULONG:
        case INTP_FLOAT:
        case INTP_DOUBLE:
        case INTP_FCOMPLEX:
        case INTP_DCOMPLEX:
            dval = PyFloat_AsDouble(value);
            flag = bips_cast(1, intype, datum, 1, INTP_DOUBLE, &dval, 1);
            break;

        default:
            XReturnPyvoxError(ERR_UNIMP, "Unimplemented Python->C conversion");
        }

        if (flag)    XReturnPyvoxLibError();
    }

#ifdef HAVE_COMPLEX_H
    /* Convert a Python complex */
    else if ( PyComplex_Check(value) ) {
        Py_complex cval;

	switch (intype) {

        case INTP_SCHAR:
        case INTP_SHORT:
        case INTP_INT:
        case INTP_LONG:
        case INTP_UCHAR:
        case INTP_USHORT:
        case INTP_UINT:
        case INTP_ULONG:
        case INTP_FLOAT:
        case INTP_DOUBLE:
	case INTP_FCOMPLEX:
	case INTP_DCOMPLEX:
            cval = PyComplex_AsCComplex(value);
            flag = bips_cast(1, intype, datum, 1, INTP_DCOMPLEX,
                             (double complex *)(void *)&cval, 1);
            break;

        default:
            XReturnPyvoxError(ERR_UNIMP, "Unimplemented Python->C conversion");
	}

    }
#endif

    /* Anything else is an error */
    else
        XReturnPyvoxError(ERR_ILLOP, "Python value is not a number");

    return 0;
}



/*------------------------------------------------------------------------
pyvox_list_from_array - Generate a Python list of numbers from a C array

This internal Pyvox function constructs a Python list from a C array
of numbers of some specified type.  The supported types include long,
double, and double complex.

FIXME: Look to see where else this can be used?

FIXME: Generalize to include other C numeric types.
------------------------------------------------------------------------*/

PyObject *                   /* Returns zero on error */
pyvox_list_from_array(
    long nelem,              /* Number of elements in the list */
    int type,                /* Type of array */
    void *data,              /* Location of the source data */
    long stride)             /* Stride in source data, in elements */
{
    PyObject *result = 0;    /* Result */
    PyObject *value = 0;     /* A value to put in the list */
    long n;                  /* Index over list elements */
    char oops[] = "(Internal) PyList_SetItem is unhappy\n";

    /* Construct an empty list with the right number of elements */
    result = PyList_New(nelem);

    /* Switch on the data type */
    switch (type) {

    case INTP_LONG : {

        long *d = data;
        for (n = 0; n < nelem; n++) {
            value = PyInt_FromLong(d[n*stride]);
            if ( PyList_SetItem(result, n, value) )
                ZReturnError(ERR_ASSERT, oops);
        }
        return result;
    }

    case INTP_DOUBLE : {

        double *d = data;
        for (n = 0; n < nelem; n++) {
            value = PyFloat_FromDouble(d[n*stride]);
            if ( PyList_SetItem(result, n, value) )
                ZReturnError(ERR_ASSERT, oops);
        }
        return result;
    }

    case INTP_DCOMPLEX : {

        Py_complex *d = data;
        for (n = 0; n < nelem; n++) {
            value = PyComplex_FromCComplex(d[n*stride]);
            if ( PyList_SetItem(result, n, value) )
                ZReturnError(ERR_ASSERT, oops);
        }
        return result;
    }

    default :
        ZReturnError(ERR_UNIMP, "Unsupported numeric type in list creation");

    }

}



/*-----------------------------------------------------------------------
PyvoxValue_Check - Check if a PyObject is convertible to a Voxel type

This internal Pyvox function attempts to convert a Python object into
a Voxel Kit anytype value.  If successful, it returns 1 and stores the
internal type and value at locations specified by the caller.
Otherwise it returns 0 and sets an error message in the Python error
object.  Note that the calling sequence differs significantly from the
built-in PyNumber_Check functions; this form is more useful for
testing for the right type in if statements.

FIXME: Allow scalar arrays.  Allow Python longs, if within some
reasonable range.  Consider allowing other Python types that claim to
be numbers, by supporting the numeric operations table.
-----------------------------------------------------------------------*/

int                          /* Returns non-zero if successful */
PyvoxValue_Check(
    PyObject *ob,            /* Python object to convert */
    const char *name,        /* Argument name to use in error message */
    int *vtype,              /* (Out) Internal type of value */
    anytype *value)          /* (Out) Value */
{
    if ( PyInt_Check(ob) ) {
	if (value)    value->as_long = PyInt_AsLong(ob);
	if (vtype)    *vtype = INTP_LONG;
	return 1; 
    }
    else if ( PyFloat_Check(ob) ) {
	if (value)    value->as_double = PyFloat_AsDouble(ob);
	if (vtype)    *vtype = INTP_DOUBLE;
	return 1; 
    }
#ifdef HAVE_COMPLEX_H
    else if ( PyComplex_Check(ob) ) {
	Py_complex cval;
	cval = PyComplex_AsCComplex(ob);
	if (value)    value->as_dcomplex = cval.real + I*cval.imag;
	if (vtype)    *vtype = INTP_DCOMPLEX;
	return 1;
    }
#endif 
    else {
        if (name)
            preset_error(ERR_INIT, "%s is not a valid Pyvox value", name);
        else
            preset_error(ERR_INIT, "Argument is not a valid Pyvox value");
	return 0; 
}

}



/*-----------------------------------------------------------------------
PyvoxLong_Check - Check if PyObject is convertible to a C long value

This Pyvox internal function attempts to convert a Python object into
a C long value.  If successful, it returns 1 and stores the value at
the location specified by the caller.  If not, it returns 0 and sets
an error message in the Python error object.  Note that the calling
sequence differs significantly from the built-in PyNumber_Check
functions; this form is more useful for testing for the right type in
if statements.

NOTE: Prior to 2003-08-02, a PyFloat was converted to a C long by
truncation (toward zero) rather than rounding (to nearest int).

FIXME: C99 supports round() rather than floor(x+0.5) and might be
a bit faster.

FIXME: Maybe also convert singleton arrays to numbers?  What about
singleton lists or tuples?

FIXME:  It is not clear what happens if one of the numbers provided is
outside the range that can be represented by a C long; investigate!
It does appear that neither PyInt_AsLong nor PyNumber_Int generates
an error.

FIXME: Set error message if invalid, including argument name?
------------------------------------------------------------------------*/

int                                /* Returns non-zero if successful */
PyvoxLong_Check(
    PyObject *ob,                  /* Python object to convert */
    long *value)                   /* Where to put value, or zero */
{
    PyObject *val;                 /* Value converted to Python int */

    if ( PyInt_Check(ob) ) {
	if (value) 
	    *value = PyInt_AsLong(ob);
	return 1; }
    else if ( PyFloat_Check(ob) ) {
	if (value)
	    *value = (long)floor(PyFloat_AsDouble(ob) + 0.5);
	return 1; }
    else if ( PyNumber_Check(ob) && (val = PyNumber_Int(ob)) ) {
	if (value)
	    *value = PyInt_AsLong(val);
	return 1;
    }
    else {
	return 0; }
}



/*-----------------------------------------------------------------------
PyvoxDouble_Check - Check if PyObject is convertible to a C double value

This Pyvox internal function attempts to convert a Python object into
a C double value.  If successful, it returns 1 and stores the value at
the location specified by the caller.  If not, it returns 0 and sets
an error message in the Python error object.  Note that the calling
sequence differs significantly from the built-in PyNumber_Check
functions; this form is more useful for testing for the right type in
if statements.

FIXME: Maybe also convert singleton arrays to numbers?  What about
singleton lists or tuples?

FIXME: Set error message if invalid, including argument name?
------------------------------------------------------------------------*/

int                                /* Returns non-zero if successful */
PyvoxDouble_Check(
    PyObject *ob,                  /* Python object to convert */
    double *value)                 /* Where to put value, or zero */
{
    PyObject *val;                 /* Value converted to Python float */

    if ( PyInt_Check(ob) ) {
	if (value) 
	    *value = (double)PyInt_AsLong(ob);
	return 1; }
    else if ( PyFloat_Check(ob) ) {
	if (value)
	    *value = PyFloat_AsDouble(ob);
	return 1; }
    else if ( PyNumber_Check(ob) && (val = PyNumber_Float(ob)) ) {
	if (value)
	    *value = PyFloat_AsDouble(val);
	return 1; }
    else {
	return 0; }
}



/*-----------------------------------------------------------------------
PyvoxComplex_Check - Check if PyObject is convertible to a C double complex

This Pyvox internal function attempts to convert a Python object into
a single C double complex value.  If successful, it returns 1 and
stores the value at the location specified by the caller.  If not, it
returns 0 and (not yet implemented) sets an error message in the
Python error object.  Note that the calling sequence differs
significantly from Python's PyNumber_Check functions; this form is
more useful for testing for the right type in if statements.

FIXME: Maybe also convert singleton arrays to numbers?  What about
singleton lists or tuples?

FIXME: Python "numbers" other than PyInt, PyFloat, and PyComplex will
generate an error.  Fortunately, the only possibility seems to be
PyLong, which should be unlikely anyway.

FIXME: Set error message if invalid, including argument name?
------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H
int                                /* Returns non-zero if successful */
PyvoxComplex_Check(
    PyObject *ob,                  /* Python object to convert */
    double complex *value)         /* Where to put value, or zero */
{
    if ( PyInt_Check(ob) ) {
	if (value) 
	    *value = (double)PyInt_AsLong(ob);
	return 1; }

    else if ( PyFloat_Check(ob) ) {
	if (value)
	    *value = PyFloat_AsDouble(ob);
	return 1; }

    else if ( PyComplex_Check(ob) ) {
	if (value)
	    *(Py_complex *)value = PyComplex_AsCComplex(ob);
	return 1; }

    else {
	return 0; }
}
#endif



/*-------------------------------------------------------------------------
PyvoxLongSeq_Check - Check and convert to a vector of longs

This function tests whether the given Python object can reasonably be
interpreted as a sequence of longs.  If so, it converts it to a C
array of longs and returns 1; otherwise it returns 0.  As a special
case, a naked number will be treated as a singleton sequence.  The
length or values arguments may be set to NULL, in which case nothing
is stored.

A Python int or float, a list or tuple of ints or floats, or a Pyvox
array are all acceptable, provided that the count matches the length
specified.

FIXME: Set error message if invalid, also including the argument name?
See PyvoxArray_Check() for an example.  In particular, provide a good
error message if the Python object has the wrong length.
-------------------------------------------------------------------------*/

int                       /* Returns non-zero if successful */
PyvoxLongSeq_Check(
    PyObject *ob,         /* Python object to be checked */
    long minlen,          /* Minimum acceptable length */
    long maxlen,          /* Maximum acceptable length */
    long *length,         /* Actual length */
    long *values)         /* Where to put the values */
{
    long len;             /* Number of elements in the sequence */
    PyObject *elem;       /* An element of the sequence */
    voxel_array *varray;  /* Voxel array used as sequence of longs */
    int type;             /* Type of data in varray */
    unsigned char *data;  /* Start of data in varray */
    int i, flag;

    /* Convert a Pyvox array to a C array of longs */
    if ( PyvoxArray_Check(ob, 0, &varray) ) {
        len = vxl_count(varray);
        if ( len == 0 || len < minlen || len > maxlen )
            return 0;
        if (length)
            *length = len;
        data = varray->data;
        type = varray->type;
        flag = bips_long(len, values, 1, type, data, 1);
        if ( flag )
            return 0;       /* FIXME: Loses message set by bips */
        return 1;
    }

    /* Convert a Python list or tuple to a C array of longs */
    else if ( PySequence_Check(ob) ) {
        len = PyObject_Length(ob);
        if (length)
            *length = len;
        if ( len < minlen || len > maxlen )
            return 0;
        for (i = 0; i < len; i++) {
            elem = PySequence_GetItem(ob, i);
            if ( elem == 0 || ! PyvoxLong_Check(elem, &values[i]) )
                return 0;
        }
        return 1;
    }

    /* Convert a naked number to a C array of longs */
    else if ( minlen <= 1 && maxlen >= 1 && 
              PyvoxLong_Check(ob, &values[0]) ) {
        if (length)
            *length = 1;
        return 1;
    }
    else
        return 0;
}



/*-------------------------------------------------------------------------
PyvoxDoubleSeq_Check - Check and convert to a vector of doubles

This function tests whether the given Python object is a list or tuple
of specified size that contains only Python integers or floats, or a
Pyvox rank-1 vector, row vector, or column vector of the specified
size.  If so, it converts it to a C array of doubles and returns 1;
otherwise it returns 0.  As a special case, a naked number will be
treated as a singleton sequence.  If length is zero, then the sequence
length will not be stored; this is a sensible thing to do only if
minlen == maxlen, so that the actual length is already known.  If the
object cannot be converted, an error message is set and zero is
returned.

See also PyvoxDoubleSeq_AsDouble, which allocates a new double array
rather than filling in an existing one.

FIXME: PyvoxDoubleSeq_Check and PyvoxDoubleSeq_AsDouble really ought
to be merged into a single subroutine to facilitate maintenance.
Alternatively, all callers of PyvoxDoubleSeq_AsDouble should allocate
and free their own double arrays.
-------------------------------------------------------------------------*/

int                       /* Returns non-zero if successful */
PyvoxDoubleSeq_Check(
    PyObject *ob,         /* Python object to be checked */
    char *name,           /* Argument name for error message */
    long minlen,          /* Minimum acceptable length */
    long maxlen,          /* Maximum acceptable length */
    long *length,         /* Actual length */
    double *values)       /* Where to put the values */
{
    voxel_array *vob;     /* Voxel array contained in ob */
    int len;              /* Number of elements in the sequence */
    PyObject *elem;       /* An element of the sequence */
    int i, flag;

    /* Provide default for unspecified arg name */
    if (name == 0)	
	name = "";

    /* Convert a Pyvox vector, row, or column */
    /* NOTE: This must precede the next block (which handles Python
       built-in sequences), since a Pyvox array may also qualify as a
       Python sequence. */
    if ( PyvoxArray_Check(ob, 0, &vob) ) {
        
        /* Allow a rank-1 vector, row vector or column vector */
        len = -1;
        if ( vob->rank == 1 )
            len = vob->dimen[0];
        else if ( vob->rank == 2 && vob->dimen[0] == 1 )
            len = vob->dimen[1];
        else if ( vob->rank == 2 && vob->dimen[1] == 1 )
            len = vob->dimen[0];
        else 
	    ZReturnError1(ERR_INIT, 
			  "Argument %s is not vector, row, or column", name);

        /* Verify that the type is not complex */
        if ( exim_iscomplex(vob->type) )
            ZReturnError1(ERR_INIT,
                          "Argument %s may not be complex", name);

        /* Verify the length; and copy its value to the caller */
        if ( len < minlen || len > maxlen )
            ZReturnError1(ERR_INIT, 
			  "Argument %s has wrong number of elements", name);
        if (length)
            *length = len;

        /* Cast to double and copy to destination */
        flag = bips_double(len, values, 1, vob->type, vob->data, 1);
        if (flag)
            ZReturnLibError();

        return 1;
    }

    /* Convert a list or tuple */
    else if ( PySequence_Check(ob) ) {
        len = PyObject_Length(ob);
        if (length)
            *length = len;
        if ( len < minlen || len > maxlen )
            ZReturnError1(ERR_INIT, 
			  "Argument %s has wrong number of elements", name);
        for (i = 0; i < len; i++) {
            elem = PySequence_GetItem(ob, i);
            if ( PyInt_Check(elem) ) {
                if (values)
                    values[i] = (double)PyInt_AsLong(elem);
            }
            else if (PyFloat_Check(elem)) {
                if (values)
                    values[i] = PyFloat_AsDouble(elem);
            }
            else
                ZReturnError1(ERR_INIT, 
			      "Argument %s has a non-numeric element", name);
        }
        return 1;
    }

    /* Convert a plain number as a singleton list */
    else if ( minlen <= 1 && maxlen >= 1 && PyNumber_Check(ob) ) {
        if (length)
            *length = 1;
        if ( PyInt_Check(ob) ) {
            if (values)
                values[0] = (double)PyInt_AsLong(ob);
        }
        else if (PyFloat_Check(ob)) {
            if (values)
                values[0] = PyFloat_AsDouble(ob);
        }
        else
            ZReturnError1(ERR_INIT, 
			  "Argument %s has a non-numeric element", name);
        return 1;
    }
    else
        ZReturnError1(ERR_INIT, 
		      "Argument %s is not a sequence, array, or number", name);
}



/*-------------------------------------------------------------------------
PyvoxDoubleSeq_AsDouble - Check and convert to a vector of doubles

This function tests whether the given Python object is a list or tuple
of arbitrary size that contains only Python integers or floats.  If
so, it allocates a double array of the necessay size, converts the
Pyvox object to a C array of doubles, and returns the address of the
double array; otherwise it returns NULL.  The length of the array is
returned via a pointer argument.  It is the user's responsibility to
free the double array.  As a special case, a naked number will be
treated as a singleton sequence.  If length is zero, then the sequence
length will not be stored; this is a sensible thing to do only if
minlen == maxlen, so that the actual length is already known.

See also PyvoxDoubleSeq_Check, which fills in an existing double array
rather than allocating new space.
-------------------------------------------------------------------------*/

double *                  /* Returns NULL on failure */
PyvoxDoubleSeq_AsDouble(
    PyObject *ob,         /* Python object to be checked */
    char *name,           /* Argument name for error message */
    long minlen,          /* Minimum acceptable length */
    long maxlen,          /* Maximum acceptable length */
    long *length)         /* Actual length */
{
    int len;              /* Number of elements in the sequence */
    voxel_array *vob;     /* Voxel array inside the Python object */
    double *values = 0;   /* Array of the sequence values */      
    PyObject *elem;       /* An element of the sequence */
    int flag, i;

    /* Provide default for argument name */
    if (name == 0)
	name = "";

    /* Accept a pyvox array of the right shape */
    /* NOTE: This must precede the next block (which handles Python
       built-in sequences), since a Pyvox array may also qualify as a
       Python sequence. */
    if ( PyvoxArray_Check(ob, 0, &vob) ) {
        
        /* Allow a rank-1 vector, row vector or column vector */
        len = -1;
        if ( vob->rank == 1 )
            len = vob->dimen[0];
        else if ( vob->rank == 2 && vob->dimen[0] == 1 )
            len = vob->dimen[1];
        else if ( vob->rank == 2 && vob->dimen[1] == 1 )
            len = vob->dimen[0];
        else 
	    ZReturnError1(ERR_INIT, 
			  "Argument %s is not vector, row, or column", name);

        /* Verify that the array type is not complex */
        if ( exim_iscomplex(vob->type) )
            ZReturnError1(ERR_INIT, "Argument %s may not be complex", name);

        /* Verify the length; and copy its value to the caller */
        if ( len < minlen || len > maxlen )
            ZReturnError1(ERR_INIT, "Argument %s has wrong number of elements",
			  name);
        if (length)
            *length = len;

        /* Allocate space and cast to double */
	values = mallock(len*sizeof(double));
        flag = bips_double(len, values, 1, vob->type, vob->data, 1);
        if (flag)
            ZReturnLibError();

        return values;
    }

    /* Accept a python sequence of numbers */
    else if ( PySequence_Check(ob) ) {
        len = PyObject_Length(ob);
        if ( len < minlen || len > maxlen )
            ZReturnError1(ERR_INIT, "Argument %s has wrong number of elements",
			  name);
	if (length)
	    *length = len;
	values = mallock(len*sizeof(double));
	if (values == 0)
	    ZReturnLibError();
        for (i = 0; i < len; i++) {
            elem = PySequence_GetItem(ob, i);
            if ( PyInt_Check(elem) ) {
		values[i] = (double)PyInt_AsLong(elem);
            }
            else if (PyFloat_Check(elem)) {
		values[i] = PyFloat_AsDouble(elem);
            }
            else
                ZReturnError1(ERR_INIT, 
			      "Argument %s has a non-numeric element", name);
        }
        return values;
    }

    /* Treat a naked number as a singleton list */
    else if ( minlen <= 1 && maxlen >= 1 && PyNumber_Check(ob) ) {
        if (length)
            *length = 1;
	values = mallock(sizeof(double));
        if ( PyInt_Check(ob) ) {
	    values[0] = (double)PyInt_AsLong(ob);
        }
        else if (PyFloat_Check(ob)) {
	    values[0] = PyFloat_AsDouble(ob);
        }
        else
	    ZReturnError1(ERR_INIT, 
			  "Argument %s has a non-numeric element", name);
        return values;
    }

    /* Anything else is an error */
    else
	ZReturnError1(ERR_INIT, 
		     "Argument %s is not an array, sequence, or number", name);

}



/*-------------------------------------------------------------------------
PyvoxComplexSeq_Check - Check and convert to a vector of double complex

This function tests whether the given Python object can reasonably be
converted to a C array of type double complex and specified minimum
and maximum length.  The supported formats include a list or tuple of
number, a Pyvox rank-1 vector, row vector, or column vector.  If so,
it converts it to a C array of doubles and returns 1; otherwise it
returns 0.  As a special case, a naked number will be treated as a
singleton sequence.  If length is zero, then the sequence length will
not be stored; this is a sensible thing to do only if minlen ==
maxlen, so that the actual length is already known.  If the object
cannot be converted, an error message is set and zero is returned.
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H
int                            /* Returns non-zero if successful */
PyvoxComplexSeq_Check(
    PyObject *ob,              /* Python object to be checked */
    char *name,                /* Argument name for error message */
    long minlen,               /* Minimum acceptable length */
    long maxlen,               /* Maximum acceptable length */
    long *length,              /* Actual length */
    double complex *values)    /* Where to put the values */
{
    voxel_array *vob;     /* Voxel array contained in ob */
    int len;              /* Number of elements in the sequence */
    PyObject *elem;       /* An element of the sequence */
    int i, flag;

    /* Provide default for unspecified arg name */
    if (name == 0)	
	name = "";

    /* Convert a Pyvox vector, row, or column */
    /* NOTE: This must precede the next block (which handles Python
       built-in sequences), since a Pyvox array may also qualify as a
       Python sequence. */
    if ( PyvoxArray_Check(ob, 0, &vob) ) {
        
        /* Allow a rank-1 vector, row vector or column vector */
        len = -1;
        if ( vob->rank == 1 )
            len = vob->dimen[0];
        else if ( vob->rank == 2 && vob->dimen[0] == 1 )
            len = vob->dimen[1];
        else if ( vob->rank == 2 && vob->dimen[1] == 1 )
            len = vob->dimen[0];
        else 
	    ZReturnError1(ERR_INIT, 
			  "Argument %s is not vector, row, or column", name);

        /* Verify the length; and copy its value to the caller */
        if ( len < minlen || len > maxlen )
            ZReturnError1(ERR_INIT, 
			  "Argument %s has wrong number of elements", name);
        if (length)
            *length = len;

        /* Cast to double complex and copy to destination */
        flag = bips_dcomplex(len, values, 1, vob->type, vob->data, 1);
        if (flag)      ZReturnLibError();

        return 1;
    }

    /* Convert a list or tuple */
    else if ( PySequence_Check(ob) ) {
        len = PyObject_Length(ob);
        if (length)
            *length = len;
        if ( len < minlen || len > maxlen )
            ZReturnError1(ERR_INIT, 
			  "Argument %s has wrong number of elements", name);
        for (i = 0; i < len; i++) {
            elem = PySequence_GetItem(ob, i);
            if ( PyInt_Check(elem) ) {
                if (values)
                    values[i] = (double)PyInt_AsLong(elem);
            }
            else if (PyFloat_Check(elem)) {
                if (values)
                    values[i] = PyFloat_AsDouble(elem);
            }
            else if (PyComplex_Check(elem)) {
                if (values)
                    ((Py_complex *)values)[i] = PyComplex_AsCComplex(elem);
            }
            else
                ZReturnError1(ERR_INIT, 
			      "Argument %s has a non-numeric element", name);
        }
        return 1;
    }

    /* Convert a plain number as a singleton list */
    else if ( minlen <= 1 && maxlen >= 1 && PyNumber_Check(ob) ) {
        if (length)
            *length = 1;
        if ( PyInt_Check(ob) ) {
            if (values)
                values[0] = (double)PyInt_AsLong(ob);
        }
        else if (PyFloat_Check(ob)) {
            if (values)
                values[0] = PyFloat_AsDouble(ob);
        }
        else if (PyComplex_Check(ob)) {
            if (values)
                ((Py_complex *)values)[0] = PyComplex_AsCComplex(ob);
        }
        else
            ZReturnError1(ERR_INIT, 
			  "Argument %s is not a number", name);
        return 1;
    }
    else
        ZReturnError1(ERR_INIT, 
		      "Argument %s is not a sequence, array, or number", name);
}
#endif



/*------------------------------------------------------------------------
Pyvox_ParseShape - Parse a Pyvox shape description into C

This Pyvox internal function takes a Pyvox shape description--that is,
a tuple or list of numbers--and converts it into the rank and
dimensions that are useful within C.  An empty list is permitted and
indicates a rank-0 or scalar array.  As a special case, a plain number
will be treated as a list containing that one number to specify a rank
1 array.  If the parse is successful, it returns 1; otherwise, it sets
an error message and returns 0.  It is the responsibility of the
caller to return NULL to Python as necessary to indicate that the
operation was successful.
-------------------------------------------------------------------------*/

int                                 /* Returns 1 if successful, else 0 */
Pyvox_ParseShape(
    PyObject *shape,                /* List or tuple to parse */
    int *rank,                      /* (out) Specified rank */
    long *dimen)                    /* (out) Specified dimensions */
{
    PyObject *d;                    /* An element from the shape */
    int i;

    /* A naked number specifies a vector, or rank-1 array */
    if ( PyvoxLong_Check(shape, &dimen[0]) ) {
	*rank = 1;
	return 1;
    }

    /* A list or tuple can specify a multi-dimensional object */
    else if ( PySequence_Check(shape) ) {
	*rank =  PyObject_Length(shape);
	for (i = 0; i < *rank; i++) {
	    d = PySequence_GetItem(shape, i);
	    if ( (d == 0) || ! PyvoxLong_Check(d, &dimen[i]) )
		ReturnPyvoxError(ERR_INIT, 
                                 "Some dimension is not a number");
	    if (dimen[i] <= 0)
		ReturnPyvoxError(ERR_MISC, 
                                 "Some dimension is zero or negative");
	}
	return 1;
    }

    /* Anything else is an error */
    else {
	ReturnPyvoxError(ERR_MISC,
                         "Shape must be a number or list of positive numbers");
    }
}



/*************************************************************************
Undocumented methods

The following methods are exposed only in the pyvoxC module, not the
pyvox module.  They are used either for various internal purposes or
for the debugging and development of Pyvox itself and.
**************************************************************************/



/*-------------------------------------------------------------------------
error_action - Set the action to take in response to an error

This function exposes the set_error_action function from errm as a
function in the pyvoxC package (but not the pyvox package!), and is
intended for debugging Python itself rather than for general use.  The
possible actions are 0 => continue; 1 => call exit; 2 => abort (and
dump core).
---------------------------------------------------------------------------*/

static PyObject *
error_action(
    PyObject *self,
    PyObject *args)
{
    int action=0;            /* Arg 1: Action to take */

    set_error_context("pyvox.error_action()");

    /* Get the argument */
    if ( ! PyArg_ParseTuple(args, "|i", &action) )
        return NULL;

    /* Set the requested action */
    set_error_action(action);

    /* Return the Python object None */
    clear_error_context();
    Py_INCREF(Py_None);
    return Py_None;
}



/*-----------------------------------------------------------------------
feature - Report existence of optional features

This function is used by the Python level to determine whether or not
various optional features have been included in the build.
------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H
#define have_complex  1
#else
#define have_complex  0
#endif

static PyObject *            /* Returns zero on error */
feature(
    PyObject *self,
    PyObject *args)
{
    char *name=0;

    /* Get the argument, which names the feature to test for */
    if ( ! PyArg_ParseTuple(args, "s", &name) )
        return NULL;

    /* Check whether that feature exists */
    if ( strcmp(name, "have_complex") == 0)
        return Py_BuildValue("i", have_complex);
    else
        return 0;
}






/**********************************************************************
Affine and polynomial transform functions

Most of the code for the affine and polynomial transform classes is
written in Python, but a few functions are written in C for
efficiency.  Those functions are provided to the Python code as
functions defined in the PyvoxC module.
************************************************************************/



/*----------------------------------------------------------------------
affine_map - Coordinate map of an affine transform

This function in the PyvoxC module generates the coordinate map of an
affine transform over a notional source image of given dimensions.
The type of the result is always type double and has rank one higher
than the number of dimensions in the transform.
------------------------------------------------------------------------*/

static PyObject *
affine_map(
    PyObject *self,           /* The PyvoxC module? */
    PyObject *args)           /* Tuple of arguments */
{
    PyObject *matrix;         /* Arg 1: Matrix of affine transform */
    PyObject *offset;         /* Arg 2: Offset of affine transform */
    PyObject *shape;          /* Arg 3: Desired dimensions of result */
    pyvox_array *result;      /* Result: Coordinate map */

    voxel_array *vmatrix;     /* Matrix as a voxel array */
    voxel_array *voffset;     /* Offset as a voxel array */
    voxel_array *vresult;     /* Result as a voxel array */

    int rank;                 /* Desired rank of result */
    long dimen[VXL_MAX_RANK]; /* Desired dimensions of result */
    int flag;

    set_error_context("affine.map()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "OOO", &matrix, &offset, &shape) )
	return NULL;

    /* Get the matrix and offset as voxel arrays */
    if  (  ! PyvoxArray_Check(matrix, "matrix", &vmatrix) ||
           ! PyvoxArray_Check(offset, "offset", &voffset) )
        ReturnPyvoxLibError();

    /* Get the desired shape of the notional source */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create the result array */
    result = parray_create(&vresult);
    if (result == 0)    ReturnPyvoxLibError();

    /* Call the Voxel Kit to do the real work */
    /* FIXME: Verify the matrix and offset before taking the data addr;
     or modify vxl_affine_map to take the voxel arrays instead. */
    flag = vxl_affine_map(vresult, rank, dimen, vmatrix->data, voffset->data);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)result;
}



/*----------------------------------------------------------------------
affine_nearest0 - Affine transform and resample with NN interpolation

This function in the PyvoxC module transforms a Pyvox array by an
affine transform, resamples it in the new coordinates using nearest
neighbor interpolation, and returns the result as a new Pyvox array.
Note that the affine transform must be the (reverse)
destination-to-source transform, NOT the forward transform you might
expect.  The type of the result is always the same as the type of the
source.

This version calls the reference implementation in the voxel kit.

FIXME: Allow the destination to be any type, specified as an optional
argument?
------------------------------------------------------------------------*/

static PyObject *
affine_nearest0(
    PyObject *self,           /* The PyvoxC module? */
    PyObject *args)           /* Tuple of arguments */
{
    PyObject *matrix;         /* Arg 1: Matrix of affine transform */
    PyObject *offset;         /* Arg 2: Offset of affine transform */
    PyObject *source;         /* Arg 3: Source array to resample */
    PyObject *shape;          /* Arg 4: Desired dimensions of destination */
    pyvox_array *result;      /* Result: Resampled array */

    voxel_array *vmatrix;     /* Matrix as a voxel array */
    voxel_array *voffset;     /* Offset as a voxel array */
    voxel_array *vsource;     /* Source as a voxel array */
    voxel_array *vresult;     /* Result as a voxel array */

    int rank;                 /* Desired rank of destination */
    long dimen[VXL_MAX_RANK]; /* Desired dimensions of destination */
    int flag;

    set_error_context("affine.nearest0()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "OOOO", &matrix, &offset, &source, &shape) )
	return NULL;

    /* Get the matrix, offset, and source as voxel arrays */
    if  (  ! PyvoxArray_Check(matrix, "matrix", &vmatrix) ||
           ! PyvoxArray_Check(offset, "offset", &voffset) ||
           ! PyvoxArray_Check(source, "source", &vsource) )
        ReturnPyvoxLibError();

    /* Get the desired shape of the destination */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and initialize the destination array */
    result = parray_alloc(&vresult, vsource->type, rank, dimen);
    if (result == 0)    ReturnPyvoxLibError();

    /* Call the Voxel Kit to do the real work */
    /* FIXME: Verify the matrix and offset before taking the data addr;
     or modify vxl_affine_nearest to take the voxel arrays instead. */
    flag = vxl_affine_nearest0(vresult, vsource->type, rank, dimen, vsource,
                               vmatrix->data, voffset->data);    
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)result;
}



/*----------------------------------------------------------------------
affine_nearest - Affine transform and resample with NN interpolation

This function in the PyvoxC module transforms a Pyvox array by an
affine transform, resamples it in the new coordinates using nearest
neighbor interpolation, and returns the result as a new Pyvox array.
Note that the affine transform must be the (reverse)
destination-to-source transform, NOT the forward transform you might
expect.  The type of the result is always the same as the type of the
source.

FIXME: Allow the destination to be any type, specified as an optional
argument?
------------------------------------------------------------------------*/

static PyObject *
affine_nearest(
    PyObject *self,           /* The PyvoxC module? */
    PyObject *args)           /* Tuple of arguments */
{
    PyObject *matrix;         /* Arg 1: Matrix of affine transform */
    PyObject *offset;         /* Arg 2: Offset of affine transform */
    PyObject *source;         /* Arg 3: Source array to resample */
    PyObject *shape;          /* Arg 4: Desired dimensions of destination */
    pyvox_array *result;      /* Result: Resampled array */

    voxel_array *vmatrix;     /* Matrix as a voxel array */
    voxel_array *voffset;     /* Offset as a voxel array */
    voxel_array *vsource;     /* Source as a voxel array */
    voxel_array *vresult;     /* Result as a voxel array */

    int rank;                 /* Desired rank of destination */
    long dimen[VXL_MAX_RANK]; /* Desired dimensions of destination */
    int flag;

    set_error_context("affine.nearest()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "OOOO", &matrix, &offset, &source, &shape) )
	return NULL;

    /* Get the matrix, offset, and source as voxel arrays */
    if  (  ! PyvoxArray_Check(matrix, "matrix", &vmatrix) ||
           ! PyvoxArray_Check(offset, "offset", &voffset) ||
           ! PyvoxArray_Check(source, "source", &vsource) )
        ReturnPyvoxLibError();

    /* Get the desired shape of the destination */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and initialize the destination array */
    result = parray_alloc(&vresult, vsource->type, rank, dimen);
    if (result == 0)    ReturnPyvoxLibError();

    /* Call the Voxel Kit to do the real work */
    /* FIXME: Verify the matrix and offset before taking the data addr;
     or modify vxl_affine_nearest to take the voxel arrays instead. */
    flag = vxl_affine_nearest(vresult, vsource->type, rank, dimen, vsource,
                              vmatrix->data, voffset->data);
    if (flag)    ReturnPyvoxLibError();
    
    clear_error_context();
    return (PyObject *)result;
}



/*----------------------------------------------------------------------
affine_linear0 - Affine transform and resample with linear interpolation

This function in the PyvoxC module transforms a Pyvox array by an
affine transform, resamples it in the new coordinates using linear
interpolation, and returns the result as a new Pyvox array.  Note that
the affine transform must be the (reverse) destination-to-source
transform, NOT the forward transform you might expect.  The type of
the result is always the same as the type of the source.

This version calls the reference implementation in the voxel kit.

FIXME: Allow the destination to be any type, specified as an optional
argument?
------------------------------------------------------------------------*/

static PyObject *
affine_linear0(
    PyObject *self,           /* The PyvoxC module? */
    PyObject *args)           /* Tuple of arguments */
{
    PyObject *matrix;         /* Arg 1: Matrix of affine transform */
    PyObject *offset;         /* Arg 2: Offset of affine transform */
    PyObject *source;         /* Arg 3: Source array to resample */
    PyObject *shape;          /* Arg 4: Desired dimensions of destination */
    pyvox_array *result;      /* Result: Resampled array */

    voxel_array *vmatrix;     /* Matrix as a voxel array */
    voxel_array *voffset;     /* Offset as a voxel array */
    voxel_array *vsource;     /* Source as a voxel array */
    voxel_array *vresult;     /* Result as a voxel array */

    int rank;                 /* Desired rank of destination */
    long dimen[VXL_MAX_RANK]; /* Desired dimensions of destination */
    int flag;

    set_error_context("affine.linear0()");
    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "OOOO", &matrix, &offset, &source, &shape) )
	return NULL;

    /* Get the matrix, offset, and source as voxel arrays */
    if  (  ! PyvoxArray_Check(matrix, "matrix", &vmatrix) ||
           ! PyvoxArray_Check(offset, "offset", &voffset) ||
           ! PyvoxArray_Check(source, "source", &vsource) )
        ReturnPyvoxLibError();

    /* Get the desired shape of the destination */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and initialize the destination array */
    result = parray_alloc(&vresult, vsource->type, rank, dimen);
    if (result == 0)    ReturnPyvoxLibError();

    /* Call the Voxel Kit to do the real work */
    /* FIXME: Verify the matrix and offset before taking the data addr;
     or modify vxl_affine_linear0 to take the voxel arrays instead. */
    flag = vxl_affine_linear0(vresult, vsource->type, rank, dimen, vsource,
                              vmatrix->data, voffset->data);    
    if (flag)    ReturnPyvoxLibError();
    
    clear_error_context();
    return (PyObject *)result;
}



/*----------------------------------------------------------------------
affine_linear - Affine transform and resample with linear interpolation

This function in the PyvoxC module transforms a Pyvox array by an
affine transform, resamples it in the new coordinates using linear
interpolation, and returns the result as a new Pyvox array.  Note that
the affine transform must be the (reverse) destination-to-source
transform, NOT the forward transform you might expect.  The type of
the result is always the same as the type of the source.

This version calls the fast implementation in the Voxel Kit.

FIXME: Allow the destination to be any type, specified as an optional
argument?
------------------------------------------------------------------------*/

static PyObject *
affine_linear(
    PyObject *self,           /* The PyvoxC module? */
    PyObject *args)           /* Tuple of arguments */
{
    PyObject *matrix;         /* Arg 1: Matrix of affine transform */
    PyObject *offset;         /* Arg 2: Offset of affine transform */
    PyObject *source;         /* Arg 3: Source array to resample */
    PyObject *shape;          /* Arg 4: Desired dimensions of destination */
    pyvox_array *result;      /* Result: Resampled array */

    voxel_array *vmatrix;     /* Matrix as a voxel array */
    voxel_array *voffset;     /* Offset as a voxel array */
    voxel_array *vsource;     /* Source as a voxel array */
    voxel_array *vresult;     /* Result as a voxel array */

    int rank;                 /* Desired rank of destination */
    long dimen[VXL_MAX_RANK]; /* Desired dimensions of destination */
    int flag;

    set_error_context("affine.linear()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "OOOO", &matrix, &offset, &source, &shape) )
	return NULL;

    /* Get the matrix, offset, and source as voxel arrays */
    if  (  ! PyvoxArray_Check(matrix, "matrix", &vmatrix) ||
           ! PyvoxArray_Check(offset, "offset", &voffset) ||
           ! PyvoxArray_Check(source, "source", &vsource) )
        ReturnPyvoxLibError();

    /* Get the desired shape of the destination */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and initialize the destination array */
    result = parray_alloc(&vresult, vsource->type, rank, dimen);
    if (result == 0)    ReturnPyvoxLibError();

    /* Call the Voxel Kit to do the real work */
    /* FIXME: Verify the matrix and offset before taking the data addr;
     or modify vxl_affine_linear to take the voxel arrays instead. */
    flag = vxl_affine_linear(vresult, vsource->type, rank, dimen, vsource,
                             vmatrix->data, voffset->data);    
    if (flag)    ReturnPyvoxLibError();
    
    clear_error_context();
    return (PyObject *)result;
}



/*----------------------------------------------------------------------
poly_linear0 - Polynomial transform with linear interpolation

This function in the PyvoxC module transforms a Pyvox array by a
polynomial transform, resamples it in the new coordinates using linear
interpolation, and returns the result as a new Pyvox array.  Note that
the polynomial transform must be the (reverse) destination-to-source
transform, NOT the forward transform you might expect.  The type of
the result is always the same as the type of the source.

This version calls the reference implementation in the voxel kit.

FIXME: Allow the destination to be any type, specified as an optional
argument?

FIXME: Provide better type and dimension checking for the various
arrays that define the transform.
------------------------------------------------------------------------*/

static PyObject *
poly_linear0(
    PyObject *self,           /* The PyvoxC module? */
    PyObject *args)           /* Tuple of arguments */
{
    PyObject *outcor;         /* Arg 1: Indices of output coord per term */
    PyObject *power;          /* Arg 2: Exponents for each term */
    PyObject *coefs;          /* Arg 3: Coefficient for each term */
    PyObject *source;         /* Arg 4: Source array to resample */
    PyObject *shape;          /* Arg 5: Desired dimensions of destination */
    pyvox_array *result;      /* Result: Resampled array */

    voxel_array *voutcor;     /* Output coords as a voxel array */
    voxel_array *vpower;      /* Exponents as a voxel array */
    voxel_array *vcoefs;      /* Coefficients as a voxel array */
    voxel_array *vsource;     /* Source as a voxel array */
    voxel_array *vresult;     /* Result as a voxel array */

    int rank;                 /* Desired rank of destination */
    long dimen[VXL_MAX_RANK]; /* Desired dimensions of destination */
    int nterm;                /* Number of terms in polynomial */
    int flag;

    set_error_context("poly.linear0()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "OOOOO", &outcor, &power, &coefs, 
			   &source, &shape) )
	return NULL;

    /* Get the matrix, offset, and source as voxel arrays */
    if  (  ! PyvoxArray_Check(outcor, "outcor", &voutcor) ||
           ! PyvoxArray_Check(power,  "power",  &vpower ) ||
           ! PyvoxArray_Check(coefs,  "coefs",  &vcoefs ) ||
           ! PyvoxArray_Check(source, "source", &vsource) )
        ReturnPyvoxLibError();

    /* Get the desired shape of the destination */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and initialize the destination array */
    result = parray_alloc(&vresult, vsource->type, rank, dimen);
    if (result == 0)    ReturnPyvoxLibError();

    /* Call the Voxel Kit to do the real work */
    /* FIXME: Verify the transform arrays before taking the data addr;
     or modify vxl_polynomial_linear0 to take the voxel arrays instead. */
    nterm = vxli_count(vcoefs);
    flag  = vxl_polynomial_linear0(vresult, vsource->type, rank, dimen, 
				   vsource, nterm, voutcor->data, 
				   vpower->data, vcoefs->data);
    if (flag)    ReturnPyvoxLibError();
    
    clear_error_context();
    return (PyObject *)result;
}



/*----------------------------------------------------------------------
poly_map - Coordinate map of a polynomial transform

This function in the PyvoxC module generates the coordinate map of a
polynomial transform over a notional source image of given dimensions.
The type of the result is always type double and has rank one higher
than the number of dimensions in the transform.
------------------------------------------------------------------------*/

static PyObject *
poly_map(
    PyObject *self,           /* The PyvoxC module? */
    PyObject *args)           /* Tuple of arguments */
{
    PyObject *outcor;         /* Arg 1: Indices of output coord per term */
    PyObject *power;          /* Arg 2: Exponents for each term */
    PyObject *coefs;          /* Arg 3: Coefficient for each term */
    PyObject *shape;          /* Arg 4: Desired dimensions of result */
    pyvox_array *result;      /* Result: Coordinate map */

    voxel_array *voutcor;     /* Output coords as a voxel array */
    voxel_array *vpower;      /* Exponents as a voxel array */
    voxel_array *vcoefs;      /* Coefficients as a voxel array */
    voxel_array *vresult;     /* Result as a voxel array */

    int rank;                 /* Desired rank of result */
    long dimen[VXL_MAX_RANK]; /* Desired dimensions of result */
    int nterm;                /* Number of terms in polynomial */
    int flag;

    set_error_context("poly.map()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "OOOO", &outcor, &power, &coefs, &shape) )
	return NULL;

    /* Get the transform parameters as voxel arrays */
    if  (  ! PyvoxArray_Check(outcor, "outcor", &voutcor) ||
           ! PyvoxArray_Check(power,  "power",  &vpower ) ||
           ! PyvoxArray_Check(coefs,  "coefs",  &vcoefs ) )
        ReturnPyvoxLibError();

    /* Get the desired shape of the notional source */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create the result array */
    result = parray_create(&vresult);
    if (result == 0)    ReturnPyvoxLibError();

    /* Call the Voxel Kit to do the real work */
    /* FIXME: Verify the transform arrays before taking the data addr;
     or modify vxl_polynomial_map to take the voxel arrays instead. */
    nterm = vxli_count(vcoefs);
    flag  = vxl_polynomial_map(vresult, rank, dimen, nterm, voutcor->data, 
			       vpower->data, vcoefs->data);
    if (flag)    ReturnPyvoxLibError();
    
    clear_error_context();
    return (PyObject *)result;
}




/***********************************************************************
Kernel objects

A kernel object descibes a voxel neighborhood, optionally specifying a
convolution coefficient for each neighbor and a constant bias term
for the convolution sum.  The rank of a kernel is the number of
dimensions; the count is the number of neighbors of any given voxel.
Each neighbor is specified by a set of coordinate deltas relative to
the given voxel.  A kernel can be created with the pyvox.kernel
function; it cannot (currently) be modified once created.

FIXME: Move the kernel code (except maybe pyvox.kernel()) into its own
file.

FIXME:  Provide functions to save a kernel to a file, or read one from
a file.
*************************************************************************/



/*-------------------------------------------------------------------------
PyvoxKernel_Check - Check whether a Python object is a valid Pyvox kernel

If the given Python object is a valid Pyvox kerenl, this function
returns 1 after storing a pointer to the kernel_object struct in the
second argument; otherwise it sets an appropriate error message and
returns 0.

FIXME: Check that this is used all the places that it should be.
------------------------------------------------------------------------*/

int                             /* Returns zero if not valid */
PyvoxKernel_Check(
    PyObject *ob,               /* Python object to check */
    const char *name,           /* Argument name to use in error message */
    struct vxl_kernel **vob)    /* (Out) The Voxkit kernel object itself */
{
    struct vxl_kernel *kern;

    /* Argument name defaults to "argument" */
    if (name == 0)    name = "Argument";

    /* Check for a valid Pyvox kernel object */
    if ( ob == 0 || ob->ob_type != &kernel_type ) {
        preset_error(ERR_INIT, "%s is not a valid (Pyvox) kernel");
        return 0;
    }

    /* Check that it contains a valid Voxkit kernel */
    kern = ((kernel_object *)ob)->kern;
    if ( kern == 0 || kern->magic != VXL_KERN_MAGIC ) {
        preset_error(ERR_INIT, "%s is not a valid (Voxkit) kernel");
	return 0;
    }
      
    /* Appears valid; save a pointer to the Voxkit kernel struct */
    if ( vob != 0 )
        *vob = kern;

    return 1;
}



/*----------------------------------------------------------------------
kernel_coef - Return kernel coefficients

This function returns a list containing the coefficients of the given
kernel object.  Note that the list is a copy; changing it will have no
effect on the kernel.  If no coefficients have been defined, the list
will be empty; this may happen even if the neighborhood is non-empty.

FIXME: Return the coefs as a dict instead?
------------------------------------------------------------------------*/

static PyObject *
kernel_coef(
    PyObject *self)
{
    struct vxl_kernel *kern;      /* The kernel struct itself */
    PyObject *pycoef;             /* List containing the coefficients */
    int count;                    /* Number of neighbors */

    set_error_context("kernel.coef");

    /* Is this really a kernel object? If so, get relevant values. */
    if ( !PyvoxKernel_Check(self, "self", &kern) )
	ReturnPyvoxLibError();
    count = (kern->coef == 0) ? 0 : kern->count;

    /* Construct list with the coefficients */
    pycoef = pyvox_list_from_array(count, kern->type, kern->coef, 1);

    clear_error_context();
    return pycoef;	
}



/*-----------------------------------------------------------------------
kernel_create - Create an instance of a kernel object

This function creates a brand-new instance of a kernel object and
initializes it to contain a kernel with given rank, count, neighbors,
coefficients, and bias.

FIXME: In the new (post 2006-02-20) kernel API, the delta, coef, and
bias arguments are never used except to temporarily support the old
API.  Consider dropping them once the old API is removed.

FIXME:  Should this take the neighbors and coefs as Python objects and
convert them internally?  Which approach is more useful?
------------------------------------------------------------------------*/

static PyObject *
kernel_create(
    int rank,                      /* Number of dimensions */
    int count,                     /* Number of neighbors */
    int type,                      /* Type of kernel coefficients and bias */
    long (*delta)[VXL_MAX_RANK],   /* Coordinate delta to each neighbor */
    void *coef,                    /* Coefficient for each neighbor */
    void *bias)                    /* Constant offset for convolution */
{
    kernel_object *self;           /* Kernel object being created */
    struct vxl_kernel *kern;       /* The kernel struct that it contains */

    /* Create a new kernel struct with given characteristics */
    kern = vxl_create_kernel(rank, count, type, delta, coef, bias);
    if (kern == 0)    ReturnPyvoxLibError();

    /* Create a Python kernel object that contains it */
    self = (kernel_object *)PyObject_NEW(kernel_object, &kernel_type);
    if (self == 0)
        ReturnPyvoxError(ERR_MALLOC, "creation of kernel object failed");
    self->kern = kern;

    /* DEBUG:  Report creation of a new kernel object */
    if (0) {
	fprintf(stderr, "Created new kernel object at %p\n", self);
	if (0) 
	    fprintf(stderr, "...reference count = %d\n",
		    ((PyObject *)self)->ob_refcnt); 
    }

    return (PyObject *)self;
}



/*----------------------------------------------------------------------
kernel_delta - Return coordinate deltas of a kernel

This function returns a list of lists containing the coordinate deltas
of a kernel object.  The outer list contains one entry for each
neighbor; each inner list contains the deltas for each coordinate.
Note that the list is a copy; changing it will have no effect on the
kernel.  It is possible to have an empty neighborhood; if so, the list
returned is empty.
------------------------------------------------------------------------*/

static PyObject *
kernel_delta(
    PyObject *self)
{
    struct vxl_kernel *kern;      /* The kernel struct itself */
    PyObject *outer, *inner;      /* Inner and outer lists */
    int count;                    /* Number of neighbors */
    int rank;                     /* Number of dimensions */
    long (*delta)[VXL_MAX_RANK];  /* Array of coord deltas */
    PyObject *del;                /* A single delta value */
    int n, d;                     /* Index over neighbors and dimensions */

    set_error_context("kernel.delta");

    /* Is this really a kernel object? If so, get useful values. */
    if ( !PyvoxKernel_Check(self, "self", &kern) )
	ReturnPyvoxLibError();
    rank  = kern->rank;
    count = kern->count;
    delta = kern->delta;

    /* Construct the desired list of coordinate deltas */
    outer = PyList_New(count);
    for (n = 0; n < count; n++) {
	inner = PyList_New(rank);
	for (d = 0; d < rank; d++ ) {
	    del = PyInt_FromLong(delta[n][d]);
	    if ( PyList_SetItem(inner, d, del) )
                ReturnPyvoxError(ERR_ASSERT, "PyList_SetItem failed");
	}
	if ( PyList_SetItem(outer, n, inner) )
            ReturnPyvoxError(ERR_ASSERT, "PyList_SetItem failed");
    }

    clear_error_context();
    return outer;	
}



/*-----------------------------------------------------------------------
kernel_destroy - Destroy an instance of a kernel object

This function destroys an instance of a kernel object by freeing the
storage associated with both the kernel struct itself and the Python
object wrapper that refers to it.  It is normally called automatically
by the Python garbage collector when the reference count of the object
goes to zero.
-----------------------------------------------------------------------*/

static void
kernel_destroy(
    kernel_object *self)
{
    struct vxl_kernel *kern;      /* The kernel struct of the object */
    int flag = 0;

    if (0)  /* DEBUG */
        fprintf(stderr, "Destroyed a kernel object at %p\n", (void *)self);

    kern = self->kern;
    flag = vxl_destroy_kernel(kern);
    if (flag != 0)
        FatalPyvoxError(ERR_INIT, "Error while destroying a kernel");
    PyMem_DEL(self);

    return;
}


/*----------------------------------------------------------------------
kernel_getattr - Get an attribute of a kernel object

ALGORITHM: This function is just a dispatcher and assumes that
set_error_context() function will be called in the subsidiary
functions.  There are a few cases where the necessary code is embedded
here rather than called; we hope that they will never generate errors.

FIXME: The methods used below to implement the members of the
instance, and the list of members is not easily expandable to the
number of members that we are likely to get eventually.  In other
words, design and use an attribute table for this purpose.
----------------------------------------------------------------------*/

/* Declare the kernel_methods array; it will be defined later. */
static struct PyMethodDef kernel_methods[];

static PyObject *         /* Returns zero on error */
kernel_getattr(
    PyObject *self,       /* Kernel object to get attribute of */
    char *attr)           /* Name of the attribute to get */
{
    struct vxl_kernel *kern;     /* The kernel struct itself */

    /* Catch things that aren't kernel objects */
    if ( ! PyvoxKernel_Check(self, "self", &kern) )
	ReturnPyvoxLibError();

    /* Return value of any requested member of the instance */
    if (strcmp(attr, "bias") == 0)
	return pyvox_get_value(kern->type, &kern->bias, 0);
    else if (strcmp(attr, "coef") == 0)
	return kernel_coef(self);
    else if (strcmp(attr, "count") == 0)
	return Py_BuildValue("i", kern->count);
    else if (strcmp(attr, "delta") == 0)
	return kernel_delta(self);
    else if (strcmp(attr, "rank") == 0)
	return Py_BuildValue("i", kern->rank);
    else if (strcmp(attr, "type") == 0)
	return PyvoxType_FromInt(kern->type);

    /* Provide list of members available from this instance */
    if (strcmp(attr, "__members__") == 0)
	return Py_BuildValue("[ssssss]",
			     "bias", "coef", "count", "delta", "rank",
                             "type");

    /* Use the Python API to find methods associated with this type */
    return Py_FindMethod(kernel_methods, self, attr);
}



/*------------------------------------------------------------------------
kernel_setattr - Set attribute of a Pyvox kernel

This function sets the value of a given attribute of a Pyvox kernel.

NOTE: This code is just a dispatcher and assumes the set_error_context
will be set by the function dispatched to.  There are a few cases
where the necessary code is embedded here rather than called; we hope
that they will never generate errors.

FIXME: The methods used below to implement the members of the
instance, and the list of members is not easily expandable to the
number of members that we are likely to get eventually.  Convert to a
table-driven approach, perhaps with hashing.
-------------------------------------------------------------------------*/

static
int                              /* Returns nonzero on error */
kernel_setattr(
    PyObject *self,              /* Pyvox kernel to be modified */
    char *attr,                  /* Attribute to set */
    PyObject *value)             /* Value to set to */
{
    struct vxl_kernel *kern=0;   /* Kernel to modify */
    int type=0;                  /* Type of the value */
    anytype val;                 /* The value itself */  
    int flag;

    /* Validate the kernel argument */
    if ( ! PyvoxKernel_Check(self, "self", &kern) )
        XReturnPyvoxLibError();

    /* Set the desired attribute */
    if (strcmp(attr, "bias") == 0) {
        set_error_context("kernel.bias=");
        flag = PyvoxValue_Check(value, "value", &type, &val);
        if (flag == 0)    XReturnPyvoxLibError();
        flag = vxl_set_kernel_bias(kern, type, &val);
        if (flag)    XReturnPyvoxLibError();
    }
    else
        XReturnPyvoxError(ERR_MISC, "Attribute not settable");

    clear_error_context();
    return 0;
}



/*------------------------------------------------------------------------
kernel_print - Print (informal) representation of a kernel object

This function prints out an informal representation of a kernel object
on a specified FILE.  This representation contains enough information
to distinguish different objects, but not to reconstruct the object
when read back in.

FIXME: Also define str() and repr() themselves.

FIXME: Pull out the code to print the bias or coef[n] as a separate
function?  Perhaps generalize to print the nth element of an anytype
or a dense array of given type?
-------------------------------------------------------------------------*/

static 
int                              /* Returns non-zero on error */
kernel_print(
    PyObject *self,              /* The object being printed */
    FILE *out,                   /* Where to print it */
    int flags)                   /* Provide repr() or str()? */
{
    struct vxl_kernel *kern;     /* The Voxkit kernel */
    int n, i;

    set_error_context("print kernel");

    /* Verify kernelness */
    if ( !PyvoxKernel_Check(self, "self", &kern) )
        XReturnPyvoxLibError();

    /* Print dumb value as repr(kernel) */
    if ((flags & Py_PRINT_RAW) == 0)
        fprintf(out, "<Pyvox kernel object at %p>", self);
    
    /* Print contents as str(kernel) */
    else {
        fprintf(out, "\n<Pyvox kernel object at %p:\n", self);
        fprintf(out, "Rank  : %d\n", kern->rank);
        fprintf(out, "Count : %d\n", kern->count);
        fprintf(out, "Bias  : ");
        if (kern->type == INTP_DOUBLE)
            fprintf(out, "%g\n", kern->bias.as_double);
#ifdef HAVE_COMPLEX_H
        else {
            if (cimag(kern->bias.as_dcomplex) < 0)
                fprintf(out, "%g - %gj\n", 
                        creal(kern->bias.as_dcomplex),
                        -cimag(kern->bias.as_dcomplex));
            else
                fprintf(out, "%g + %gj\n", 
                        creal(kern->bias.as_dcomplex),
                        cimag(kern->bias.as_dcomplex));
        }
#endif
        fprintf(out, "Neighbors:\n");
        for (n = 0; n < kern->count; n++) {
            fprintf(out, "  (");
            for (i = 0; i < kern->rank; i++)
                fprintf(out, "%ld,", kern->delta[n][i]);
            fprintf(out, ") : ");
            if (kern->type == INTP_DOUBLE)
                fprintf(out, "%g\n", *((double *)(kern->coef) + n) );
#ifdef HAVE_COMPLEX_H
            else {
                if (cimag(*((double complex *)(kern->coef) + n)) < 0)
                    fprintf(out, "%g - %gj\n", 
                            creal(*((double complex *)(kern->coef) + n)),
                            -cimag(*((double complex *)(kern->coef) + n)) );
                else
                    fprintf(out, "%g + %gj\n", 
                            creal(*((double complex *)(kern->coef) + n)),
                            cimag(*((double complex *)(kern->coef) + n)) );
            }
#endif
        }
        fprintf(out, ">\n");
    }

    clear_error_context();
    return 0;
}



/*=====================================================================
Mapping methods for the Pyvox kernel type
=====================================================================*/



/*----------------------------------------------------------------------
kernel_getitem - Return coefficient for a given neighbor

This function returns the coefficient value corresponding to a
particular neighbor specified by its delta.  If the neighbor does not
exist, then None is returned.

FIXME: This code knows about the internals of the vxl_kernel struct.
Make sure this is really what you want to do.
------------------------------------------------------------------------*/

static
PyObject *                       /* Returns given coef, or None */
kernel_getitem(
    PyObject *self,              /* Kernel to get info from */
    PyObject *pydelta)           /* Delta for desired neighbor */
{
    PyObject *result;            /* Requested coefficient */
    struct vxl_kernel *kern;     /* Kernel struct for Voxkit */
    int rank;                    /* Rank of kernel */
    long delta[VXL_MAX_RANK];    /* Delta for desired neighbor */
    long index;                  /* Index of requested coefficient */

    set_error_context("kernel[]");

    /* Verify the kernel */
    if ( !PyvoxKernel_Check(self, "self", &kern) )
        ReturnPyvoxLibError();
    rank = kern->rank;

    /* Fetch the delta as an array of longs */
    if ( !PyvoxLongSeq_Check(pydelta, rank, rank, 0, delta) )
        ReturnPyvoxLibError();

    /* Find the specified delta */
    index = vxli_find_kernel_coef(kern, delta);

    /* If no such neighbor, return None */
    if (index < 0) {
        Py_INCREF(Py_None);
        result = Py_None;
    }

    /* Otherwise return a Python number for the coefficient value */
    else
        result = pyvox_get_value(kern->type, kern->coef, index);
    
    clear_error_context();
    return result;
}



/*---------------------------------------------------------------------
kernel_length - Return number of neighbors defined for a kernel

This function implements len(kernel) and returns the number of neighbors
defined for a kernel.
----------------------------------------------------------------------*/

static int                      /* Returns number of neighbors in kernel */
kernel_length(
    PyObject *pykern)           /* Kernel to get info from */
{
    struct vxl_kernel *kern;    /* The Voxkit kernel object */
    int result;                 /* Number of neighbors in kernel */

    set_error_context("len(kernel)");

    if ( !PyvoxKernel_Check(pykern, "self", &kern) )
	 ReturnPyvoxLibError();
    result = kern->count;

    clear_error_context();
    return result;
}



/*----------------------------------------------------------------------
kernel_setitem - Set coefficient for a given neighbor

This function sets the coefficient of the neighbor specified by its
delta, adding that neighbor if necessary.  As a special case, "del
kernel[]" calls this function with pyvalue == 0.
-----------------------------------------------------------------------*/

static
int                              /* Returns 0 on error */
kernel_setitem(
    PyObject *self,              /* Kernel to modify */
    PyObject *pydelta,           /* Delta of neighbor to set or add */
    PyObject *pyvalue)           /* Value to set coefficient to */
{
    struct vxl_kernel *kern;     /* Kernel struct for Voxkit */
    int rank;                    /* Rank of kernel */
    int type;                    /* Type of value */
    anytype value;               /* Value as a C value */
    long delta[VXL_MAX_RANK];    /* Delta for desired neighbor */
    int flag;

    set_error_context("kernel[]=");

    /* Verify the kernel and delta arguments */
    /* FIXME: Obscure error message if pydelta has wrong number of elements */
    if ( !PyvoxKernel_Check(self, "self", &kern) )
        XReturnPyvoxLibError();
    rank = kern->rank;
    if ( !PyvoxLongSeq_Check(pydelta, rank, rank, 0, delta) )
        XReturnPyvoxLibError();

    /* If pyvalue is zero or None, delete the indicated neighbor */
    if ( pyvalue == 0 || pyvalue == Py_None ) {
        flag = vxl_del_kernel_coef(kern, delta);
	if (flag)    XReturnPyvoxLibError();
	return 0;
    }

    /* Otherwise validate the value argument */
    if ( !PyvoxValue_Check(pyvalue, "Value", &type, &value) )
        XReturnPyvoxLibError();

    /* Set or add the desired coefficient */
    /* FIXME: Redundant kernel validity checking in called function */
    flag = vxl_set_kernel_coef(kern, delta, type, &value);
    if (flag)    XReturnPyvoxLibError();

    clear_error_context();
    return 0;
}



/*----------------------------------------------------------------------
kernel_as_mapping - Mapping methods for the kernel type
-----------------------------------------------------------------------*/

static PyMappingMethods
kernel_as_mapping = {
    (inquiry)       kernel_length,    /* mp_length:        len(x)   */
    (binaryfunc)    kernel_getitem,   /* mp_subscript:     x[y]     */
    (objobjargproc) kernel_setitem,   /* mp_ass_subscript: x[y] = v */
};




/*------------------------------------------------------------------------
kernel_type - Python type descriptor for the kernel type

FIXME: The kernel_type object is global because it is used in both
pyvox.c and parray.c; this might be fixable if a function to verify
the type of a purported kernel is made available.
------------------------------------------------------------------------*/

/* static */      /* FIXME: Should this be made static? */
PyTypeObject kernel_type = {

    /* Type head */
    PyObject_HEAD_INIT(&PyType_Type)
    0,                              /* Object size */
    "kernel",                       /* Name of type */
    sizeof(kernel_object),          /* Basic size of type */
    0,                              /* Item size */

    /* Standard methods */
    (destructor)  kernel_destroy,   /* Instance destructor */
    (printfunc)   kernel_print,     /* "print x" */
    (getattrfunc) kernel_getattr,   /* "instance.attr" */
    (setattrfunc) kernel_setattr,   /* "instance.attr = foo" */
    (cmpfunc)     0,                /* compare instances */
    (reprfunc)    0,                /* string representation */

    /* Method suites for standard classes */
    (PyNumberMethods *)   0,        /* tp_as_number */
    (PySequenceMethods *) 0,        /* tp_as_sequence */
    (PyMappingMethods *)  &kernel_as_mapping,   /* tp_as_mapping */

    /* Other methods */
    (hashfunc) 0,                   /* Hash function of an instance */
    (ternaryfunc) 0,                /* "instance()" */
    (reprfunc) 0,                   /* "str(instance)" */

    /* See Include/object.h for still other methods */
};



/*--------------------------------------------------------------------------
kernel_methods - Table of kernel instance methods for getattr

(None are actually defined yet.)

FIXME: It's apparently possible to add documentation strings to this
table (and probably to the other method tables) and would be a good
idea in general.
--------------------------------------------------------------------------*/

static struct PyMethodDef
kernel_methods [] = {
    { NULL,      NULL }
};



/*************************************************************************
Methods belonging to the Pyvox Module

The following objects and methods are part of the Pyvox module (rather
than belonging to the Pyvox array type or to instances of the voxel
type.
*************************************************************************/



/*------------------------------------------------------------------------
pyvox_arraym - Create a voxel array containing given data

This method constructs a voxel array with given dimensions and type, and
fills it with data from a list.  The given data may be a scalar value, in
which case the entire array is filled with that value.

ALGORITHM: The function is named pyvox_arraym in C to avoid a conflict
with the pyvox_array struct, but has the name pyvox.array in Python.

FIXME: Allow the data argument to be a file name or file object, in
which case this method should read the file to initialize the array.

FIXME: If the shape is omitted, create a scalar of type double?

FIXME: If the data is None, create the array but don't initialize it?
Warn the user in the strongest possible terms that he uses this option
at his own risk.

FIXME: Allow the first dimension to be -1, to indicate it's to be filled
in later, possibly by read?

FIXME: Leaks memory if the operation fails.
-------------------------------------------------------------------------*/

static PyObject *
pyvox_arraym(
    PyObject *self,
    PyObject *args)
{
    PyObject *shape;          /* Arg 1: Tuple or list of dimensions */
    PyObject *type = 0;       /* Arg 2: Desired internal type */
    PyObject *data = 0;       /* Arg 3: Data to fill array with */

    int intype = INTP_DOUBLE; /* Desired internal type */
    int rank;                 /* Desired rank */
    long dimen[VXL_MAX_RANK]; /* Desired dimensions */
    pyvox_array *result;      /* The new varray  object */
    voxel_array *vres;        /* and its voxel array */
    PyObject *value;          /* Value to be stored in result */
    int vtype = 0;            /* Type of value to be stored */
    anytype val;              /* Value to be stored in result */
    int nelem, n;             /* Number of data values supplied */
    int flag;

    set_error_context("pyvox.array()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "O|OO", &shape, &type, &data) )
        return NULL;

    /* Verify the array type; default to double */
    if ( type != 0 && !PyvoxIntype_Check(type, "type", &intype) )
        ReturnPyvoxLibError();

    /* Extract the desired dimensions */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and initialize the destination array */
    result = parray_alloc(&vres, intype, rank, dimen);
    if (result == 0)    ReturnPyvoxLibError();

    /* Case 1: Data is a single number, or omitted */
    if ( data == 0 )  {
        /* FIXME: Create and use vxl_zero instead */
        val.as_double = 0;
	flag = vxl_const(vres, INTP_DOUBLE, &val);
        if (flag)    ReturnPyvoxLibError();        
    }
    else if ( PyvoxValue_Check(data, "data", &vtype, &val) ) {
	flag = vxl_const(vres, vtype, &val);
        if (flag)    ReturnPyvoxLibError();
    }

    /* Case 2: Data is a sequence of numbers */
    else if ( PySequence_Check(data) ) {
	nelem = PyObject_Length(data);
	if ( nelem != vxli_count(vres) )
	    ReturnPyvoxError(ERR_INCOMP, 
                             "Length of data must match length of array");
	for (n = 0; n < nelem; n++) {
            value = PySequence_GetItem(data, n);
            flag  = pyvox_set_value(intype, vres->data, n, value);
            if (flag)    ReturnPyvoxLibError();
	}
    }

    /* Case 3: Anything else is an error */
    else
	ReturnPyvoxError(ERR_INIT, "Data must be a number or list of numbers");

    /* Return the result */
    clear_error_context();
    return (PyObject *)result;
}



/*-------------------------------------------------------------------------
pyvox_ccmap - Generate list and map of connected components in an image

This Pyvox function scans a 2 or 3D image to generate a list of the
connected components in the image, and a map image showing the
component that each pixel is part of.  The output list contains a
sublist for each component; the sublist contains the component
(region) number, the number of pixels in the component, and a seed
point in the component.
---------------------------------------------------------------------------*/

static PyObject *
pyvox_ccmap(
    PyObject *self,
    PyObject *args)
{
    PyObject *image;         /* Arg 1: Image to find components of */
    PyObject *objects=0;     /* Res 1: List of objects found */
    PyObject *obmap=0;       /* Res 2: Map of objects found */

    voxel_array *vimage=0;               /* Image as a voxel array */
    voxel_array *vobmap=0;               /* Map of components found */
    PyObject *region=0;                  /* Region description */
    PyObject *regno=0;                   /* Region number */
    PyObject *canon=0;                   /* Canonical region number */
    PyObject *count=0;                   /* Pixel count for region */
    PyObject *point=0;                   /* A seed point */
    PyObject *val1=0;                    /* A Python value */
    struct voxel_object_data *oblist=0;  /* List of components found */
    long nobs;                           /* Number of components found */
    int rank;                            /* Rank of the input image */
    char confail[] = "List construction failed";
    int flag, n, i;

    set_error_context("pyvox.ccmap()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "O", &image) )
        return NULL;

    /* Verify the input image */
    if ( !PyvoxArray_CheckSrc(image, "image", &vimage) )
        ReturnPyvoxLibError();
    rank = vimage->rank;
    if ( rank != 2 && rank != 3 )
        ReturnPyvoxError(ERR_INIT, "Image must be rank 2 or 3");
    if ( vimage->type != INTP_UCHAR )
        ReturnPyvoxError(ERR_INIT, "Image must be unsigned char");

    /* Call the Voxel Kit to find the connected components */
    obmap = (PyObject *)parray_create(&vobmap);
    if (obmap == 0)    ReturnPyvoxLibError();
    flag = vxl_find_objects_uchar3d(vobmap, &nobs, &oblist, vimage);
    if (flag)    ReturnPyvoxLibError();

    /* Build object list */
    objects = PyList_New(nobs);
    for (n = 0; n < nobs; n++) {

        /* Build seed point */
        point = PyList_New(rank);
        if (point == 0)
            ReturnPyvoxError(ERR_MALLOC, "Unable to allocate list");
        for (i = 0; i < rank; i++) {
            val1 =  PyInt_FromLong(oblist[n].point[i]);
            if (val1 == 0)
                ReturnPyvoxError(ERR_MISC, confail);
            flag = PyList_SetItem(point, i, val1);
            if (flag)   
                ReturnPyvoxError(ERR_MISC, confail);
        }

        /* Build list entry */
        regno = PyInt_FromLong(oblist[n].ident);
	canon = PyInt_FromLong(oblist[n].canon);
        count = PyInt_FromLong(oblist[n].count);
        if (regno == 0 || canon == 0 || count == 0)
            ReturnPyvoxError(ERR_MISC, confail);
        region = Py_BuildValue("[NNNN]", regno, canon, count, point);
        if (region == 0)
            ReturnPyvoxError(ERR_MISC, confail);
        flag = PyList_SetItem(objects, n, region);
        if (flag)
            ReturnPyvoxError(ERR_MISC, confail);
    }

    /* Return results */
    clear_error_context();
    return Py_BuildValue("[NN]", objects, obmap);
}



/*--------------------------------------------------------------------------
pyvox_fftramp - Create array containing FFT frequencies

This pyvox function creates a voxel array of some specified type,
rank, and dimensions, and fills each voxel with the FFT frequency
corresponding to the value of one of the coordinate axes.  If the type
is omitted, it defaults to long.  If the axis is omitted, it
defaults to the last axis.  This function is not defined for unsigned
types.

FIXME: Implement the negative axis convention?
--------------------------------------------------------------------------*/

static PyObject *
pyvox_fftramp(
    PyObject *self,
    PyObject *args)
{
    PyObject *shape;            /* Arg 1: Tuple or list of dimensions */
    PyObject *type = 0;         /* Arg 2: Desired internal type */
    int axis = -1;              /* Arg 3: Desired axis to store */

    int  intype = INTP_DOUBLE;  /* Desired internal type */
    int  rank;                  /* Desired rank */
    long dimen[VXL_MAX_RANK];   /* Desired dimensions */
    pyvox_array *dest;          /* Result as a Pyvox array */
    voxel_array *vdest;         /* Result as a voxel array */
    int flag;

    set_error_context("pyvox.fftramp()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "O|Oi", &shape, &type, &axis) )
        return NULL;

    /* Verify the array type, or take default*/
    if ( type != 0 && !PyvoxIntype_Check(type, "type", &intype) )
        ReturnPyvoxLibError();

    /* Extract the desired dimensions and axis */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;
    if (axis == -1)
	axis = rank - 1;
    
    /* Create and initialize the destination array */
    dest = parray_alloc(&vdest, intype, rank, dimen);
    if (dest == 0)    ReturnPyvoxLibError();
    flag = vxl_fftramp(dest->varray, axis);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)dest;
}



/*--------------------------------------------------------------------------
pyvox_ramp - Create array containing coordinate values

This pyvox function creates a voxel array of some specified type,
rank, and dimensions, and fills each voxel with the value of one of
the coordinate axes.  If the type is omitted, it defaults to double.
If the axis is omitted, it defaults to the last axis.

FIXME: Implement the negative axis convention?
--------------------------------------------------------------------------*/

static PyObject *
pyvox_ramp(
    PyObject *self,
    PyObject *args)
{
    PyObject *shape;            /* Arg 1: Tuple or list of dimensions */
    PyObject *type = 0;         /* Arg 2: Desired internal type */
    int axis = -1;              /* Arg 3: Desired axis to store */

    int  intype = INTP_DOUBLE;  /* Desired internal type */
    int  rank;                  /* Desired rank */
    long dimen[VXL_MAX_RANK];   /* Desired dimensions */
    pyvox_array *dest;          /* Result as a Pyvox array */
    voxel_array *vdest;         /* Result as a voxel array */
    int flag;

    set_error_context("pyvox.ramp()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "O|Oi", &shape, &type, &axis) )
        return NULL;

    /* Verify the array type, or take default*/
    if ( type != 0 && !PyvoxIntype_Check(type, "type", &intype) )
        ReturnPyvoxLibError();

    /* Extract the desired dimensions */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;
    if (axis == -1)
	axis = rank - 1;
    
    /* Create and initialize the destination array */
    dest = parray_alloc(&vdest, intype, rank, dimen);
    if (dest == 0)    ReturnPyvoxLibError();
    flag = vxl_ramp(dest->varray, axis);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)dest;
}



/*--------------------------------------------------------------------------
pyvox_randi - Create array of random integers

This pyvox function creates a voxel array of type int with some
specified rank and dimensions, and fills it with random choices from
the integers 0, ..., N-1.  This function is NOT thread-safe since the
default generator is used.

FIXME: In this version, the type is always int.

FIXME: Allow caller to specify seed or generator.
--------------------------------------------------------------------------*/

static PyObject *
pyvox_randi(
    PyObject *self,
    PyObject *args)
{
    PyObject *shape;            /* Arg 1: Tuple or list of dimensions */
    int n;                      /* Arg 2: Desired upper bound on choices */

    int  rank;                  /* Desired rank */
    long dimen[VXL_MAX_RANK];   /* Desired dimensions */
    pyvox_array *dest;          /* Result as a pyvox array */
    voxel_array *vdest;         /* ...and as a voxel array */
    int flag;

    set_error_context("pyvox.randi()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "Oi", &shape, &n) )
        return NULL;

    /* Extract the desired dimensions */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and fill the destination array */
    dest = parray_alloc(&vdest, INTP_INT, rank, dimen);
    if (dest == 0)    ReturnPyvoxLibError();
    flag = vxl_randi(dest->varray, 0, n);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)dest;
}



/*--------------------------------------------------------------------------
pyvox_randn - Create array of normal variates with given mean and sdev

This pyvox function creates a new voxel array of type double with
specified rank and dimensions, and fills it with normal random
variates with given mean and variance.  The default generator is
used.  This function is NOT threadsafe.
--------------------------------------------------------------------------*/

static PyObject *
pyvox_randn(
    PyObject *self,
    PyObject *args)
{
    PyObject *shape;            /* Arg 1: Tuple or list of dimensions */
    double mean = 0.0;          /* Arg 2: Desired mean */
    double sdev = 1.0;          /* Arg 3: Desired standard deviation */

    int intype = INTP_DOUBLE;   /* Fixed internal type */
    int  rank;                  /* Desired rank */
    long dimen[VXL_MAX_RANK];   /* Desired dimensions */
    long *state = 0;            /* Use the default random number generator */
    pyvox_array *dest;          /* Result as a pyvox array */
    voxel_array *vdest;         /* ...and as a voxel array */
    int flag;

    set_error_context("pyvox.randn()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "O|dd", &shape, &mean, &sdev) )
        return NULL;

    /* Extract the desired dimensions */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and fill the destination array */
    dest = parray_alloc(&vdest, intype, rank, dimen);
    if (dest == 0)    ReturnPyvoxLibError();
    flag = vxl_randn(dest->varray, state, mean, sdev);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)dest;
}



/*--------------------------------------------------------------------------
pyvox_randu - Create array of uniform variates in given interval

This pyvox function creates a new voxel array of type double with
specified rank and dimensions, and fills it with uniform variates in
the open interval (a,b).  The default random generator is used, so
this function is NOT thread-safe.
--------------------------------------------------------------------------*/

static PyObject *
pyvox_randu(
    PyObject *self,
    PyObject *args)
{
    PyObject *shape;            /* Arg 1: Tuple or list of dimensions */
    double a = 0.0;             /* Arg 2: Lower limit of interval */
    double b = 1.0;             /* Arg 3: Upper limit of interval */

    int intype = INTP_DOUBLE;   /* Fixed internal type */
    int  rank;                  /* Desired rank */
    long dimen[VXL_MAX_RANK];   /* Desired dimensions */
    long *state = 0;            /* Use the default random number generator */
    pyvox_array *dest;          /* Result as a pyvox array */
    voxel_array *vdest;         /* ...and as a voxel array */
    int flag;

    set_error_context("pyvox.randu()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "O|dd", &shape, &a, &b) )
        return NULL;

    /* Extract the desired dimensions */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) ) 
	return NULL;

    /* Create and fill the destination array */
    dest = parray_alloc(&vdest, intype, rank, dimen);
    if (dest == 0)    ReturnPyvoxLibError();
    flag = vxl_randu(dest->varray, state, a, b);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)dest;
}



/*-------------------------------------------------------------------------
pyvox_kernel - Construct a convolution kernel

This Pyvox function constructs a new convolution kernel with a given
rank, type, and initial allocation of neighbors.  The neighbors are
not actually initiallized, since it is assumed that that will be done
later.  If the alloc argument is zero, then it defaults to a value 
depending on the rank.

NOTE: This code was written on 2006-02-20 to support a new and improved
programming interface for kernel construction.  The pyvox_kernel_old()
function still supports the old interface but will go away soon.

FIXME: Add str(kernel) and repr(kernel).  

FIXME: Provide methods for translating kernels into arrays and vice
verse.

FIXME: Consider treating empty kernels as a special case, in which the
rank is always zero and the result of convolving is always zero?  I
think that all the functions that use kernels will handle no neighbors
without any modification.

FIXME: Provide methods for creating various common kernels.
---------------------------------------------------------------------------*/

static PyObject *
pyvox_kernel(
    PyObject *self,
    PyObject *args)
{
    int rank = 0;               /* Arg 1: Desired rank */
    PyObject *pytype = 0;       /* Arg 2: Desired type */
    int alloc = 0;              /* Arg 3: Initial allocation, or zero */

    PyObject *pykern;           /* Kernel being constructed */
    int type = INTP_DOUBLE;     /* Type of the kernel */

    set_error_context("pyvox.kernel()");

    /* Get the arguments */
    if ( ! PyArg_ParseTuple(args, "i|Oi", &rank, &pytype, &alloc) )
        return NULL;
    if ( pytype != 0 && !PyvoxIntype_Check(pytype, "type", &type) )
        ReturnPyvoxLibError();

    /* Construct the kernel itself */
    pykern = kernel_create(rank, alloc, type, 0, 0, 0);

    clear_error_context();
    return pykern;
}



/*-------------------------------------------------------------------------
pyvox_kernel_old - Construct a convolution kernel (Old calling sequence)

This Pyvox function constructs a convolution kernel given the
neighbors (defined by coordinate offsets or deltas), the coefficient
for each neighbor, and the constant bias for the convolution.  The
coefficients are optional; if omitted, the kernel defines a
neighborhood only.  The bias is also optional; if omitted, it defaults
to zero.  Once constructed, the kernel cannot be changed, although
this restriction may be lifted in later versions.  The number of
neighbors is determined by the length of the outer delta list; the
length of the coefficient list must be the same, or zero.  The rank is
determined by the length of the inner delta lists; it is an error if
these are not all the same.  Kernels containing no neighbors are
invalid by decree, since it is not possible to determine the rank in
that case.  The type of the kernel is double complex if any coefficient
is complex; otherwise it is double.  If the coefficients are real, then
the bias must also be real.

NOTE: This code supports the old (pre 2006-02-20) deprecated calling
sequence for pyvox.kernel().  It (or an equivalent) will be kept
around for a while for compatibility but will eventually go away
completely.

FIXME:  Should the nested list conversion loop below be made into
a generic library function?
------------------------------------------------------------------------*/

static PyObject *
pyvox_kernel_old(
    PyObject *self,
    PyObject *args)
{
    PyObject *pydelta;         /* Arg 1: Coordinate deltas */
    PyObject *pycoef = 0;      /* Arg 2: Convolution coefficients */
    PyObject *pybias = 0;      /* Arg 3: Constant bias for convolution */

    PyObject *pykern;              /* Kernel being constructed */
    int rank;                      /* Number of dimensions */
    int count;                     /* Number of neighbors, from delta array */
    int type=0;                    /* Type of kernel coefficients and bias */
    int cnt2 = 0;                  /* Number of neighbors, from coef array */
    anytype biasval;               /* Numeric value of the bias */
    void *bias=0;                  /* Pointer to bias value */
    long (*delta)[VXL_MAX_RANK];   /* Coordinate deltas */
    void *coef = 0;                /* Convolution coefficients */
    PyObject *inner;               /* Inner list of deltas */
    PyObject *del;                 /* A single coordinate delta */
    int n, d;                      /* Index over neighbors and dimensions */

    set_error_context("pyvox.kernel()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "O|OO", &pydelta, &pycoef, &pybias) )
	return NULL;

    /* Determine number of neighbors */
    count = PyObject_Length(pydelta);
    if (count < 1)
	ReturnPyvoxError(ERR_ILLOP, "Empty kernels are not permitted.");
    if (pycoef != 0) {
	cnt2  = PyObject_Length(pycoef);
	if ( cnt2 != count && cnt2 != 0)
	    ReturnPyvoxError(ERR_INCOMP,
                             "Delta and coef arrays have different lengths");
    }

    /* Construct the array of coefficients, if provided */ 
    if (cnt2 != 0) {

        /* Allocate space for coefficients */
        /* FIXME: Allocates twice as much space as is needed for real coefs */
#ifdef HAVE_COMPLEX_H
        coef = mallock(count*sizeof(double complex));
#else
        coef = mallock(count*sizeof(double));
#endif
	if (coef == 0)    ReturnPyvoxLibError();
    

        /* Case 1: Real coefficients */
        if ( PyvoxDoubleSeq_Check(pycoef, "coefs", cnt2, cnt2, 0, coef) ) {
            type = INTP_DOUBLE;
            if (pybias) {
                if ( PyvoxDouble_Check(pybias, &biasval.as_double) )
                    bias = &biasval;
                else
                    ReturnPyvoxError(ERR_INCOMP, 
                                     "Real coefs require real bias");
            }
        }

#ifdef HAVE_COMPLEX_H
        /* Case 2: Complex coefficients */
        else if (PyvoxComplexSeq_Check(pycoef, "coefs", cnt2, cnt2, 0, coef)) {
            type = INTP_DCOMPLEX;
            if (pybias)  {
                if ( PyvoxComplex_Check(pybias, &biasval.as_dcomplex) )
                    bias = &biasval;
                else
                    ReturnPyvoxError(ERR_INCOMP, "Bias is not numeric");
            }
        }
#endif

        /* Case 3: Cannot interpret coefficients as numbers */
        else
            ReturnPyvoxError(ERR_INIT, "Cannot make sense of coefs");
    }

    /* Determine rank; construct delta array */
    delta = mallock(count*VXL_MAX_RANK*sizeof(long));
    if (delta == 0)    ReturnPyvoxLibError();
    inner = PySequence_GetItem(pydelta, 0);
    rank  = PyObject_Length(inner);
    if (rank > VXL_MAX_RANK)
	ReturnPyvoxError(ERR_UNIMP,
                         "Requested rank exceeds max supported rank");
    for (n = 0; n < count; n++) {
	inner = PySequence_GetItem(pydelta, n);
	if (rank != PyObject_Length(inner) )
	    ReturnPyvoxError(ERR_INCOMP,
                             "Inconsistent ranks in coordinate deltas");
	for (d = 0; d < rank; d++) {
	    del = PySequence_GetItem(inner, d);
	    if ( PyInt_Check(del) )
		delta[n][d] = PyInt_AsLong(del);
	    else
		ReturnPyvoxError(ERR_INIT, 
                                 "Some coordinate delta is not an integer");
	}
    }

    /* Construct the kernel itself */
    pykern = kernel_create(rank, count, type, delta, coef, bias);

    /* Free temporary storage */
    free(delta);
    free(coef);

    clear_error_context();
    return pykern;
}



/*-------------------------------------------------------------------------
pyvox_lowpass - Construct a standard lowpass convolution kernel

This Pyvox function constructs a standard lowpass convolution kernel
for any given number of dimensions.
------------------------------------------------------------------------*/

static PyObject *
pyvox_lowpass(
    PyObject *self,
    PyObject *args)
{
    int rank = 0;                  /* Arg 1: Number of dimensions */

    kernel_object *pykern;         /* Pyvox kernel object being created */
    struct vxl_kernel *kern;       /* The kernel struct that it contains */

    set_error_context("pyvox.lowpass()");

    /* Get the arguments */
    if ( !PyArg_ParseTuple(args, "i", &rank) )
	return NULL;

    /* Create a new kernel struct with given characteristics */
    kern = vxl_lowpass_kernel(rank);
    if (kern == 0)    ReturnPyvoxLibError();

    /* Create a Python kernel object that contains it */
    pykern = (kernel_object *)PyObject_NEW(kernel_object, &kernel_type);
    if (pykern == 0)
        ReturnPyvoxError(ERR_MALLOC, "creation of kernel object failed");
    pykern->kern = kern;

    clear_error_context();
    return (PyObject *)pykern;
}



/*------------------------------------------------------------------------
pyvox_rawread - Read raw image data from an external file into a voxel array

This Pyvox function reads raw image data from an external file in some
specified external format, converts it to the most natural internal
format, and returns it as a voxel array.  The arguments are the file
to read from, the desired shape of the array (as a tuple or list), the
external type (which defaults to unsigned char), an optional flag to
mark bigendian (which defaults to bigendian), and the location to seek
to.  The seek argument is ignored for an open file, which is assumed
to have already been positioned at the right location.  Otherwise, if
the seek is positive, it is measured from the beginning of the file;
if negative, from the end; if zero, no seek is done and data is read
from the beginning of the file.

FIXME: The fseek function only supports files up to 2 GB.

FIXME: Let the shape default to rank 1 and however many elements can
be read from the file?

FIXME: Let the outermost dimension (e.g. number of slices) be
determined by how many elements can be read from the file?
------------------------------------------------------------------------*/

static PyObject *
pyvox_rawread(
    PyObject *self,
    PyObject *args)
{
    PyObject *file;            /* Arg 1: A filename or an open file */
    PyObject *shape;           /* Arg 2: Tuple or list giving dimensions */
    PyObject *type = 0;        /* Arg 3: External format (default uchar) */
    int bigend = 1;            /* Arg 4: Bigendian? (defaults to yes) */
    long seek=0;               /* Arg 5: Seek location */

    char *filename = 0;        /* Name of an image file to open */
    FILE *infile;              /* Stream for the image file */
    int extype = EXTP_UCHAR;   /* External format (default uchar) */
    int intype;                /* Natural internal format for the data */
    int rank;                  /* Assumed rank of the image data */
    long dimen[VXL_MAX_RANK];  /* Assumed dimensions of the image data */
    pyvox_array *obj;          /* Result as a Pyvox array */
    voxel_array *vobj;         /* Result as a voxel array */
    int flag;

    set_error_context("pyvox.read()");

    /* Fetch all the arguments */
    if ( !PyArg_ParseTuple(args, "OO|Oil", &file, &shape,
                           &type, &bigend, &seek) )
        return NULL;

    /* Translate a filename or open file into a C FILE pointer */
    if ( PyString_Check(file) ) {
        filename = PyString_AsString(file);
	infile = fopen(filename, "rb");
	if (infile == 0)
	    ReturnPyvoxError(ERR_FILE, "Unable to open the given filename");
        if (seek > 0)
            flag = fseek(infile, seek, SEEK_SET);
        else if (seek < 0)
            flag = fseek(infile, seek, SEEK_END);
        else
            flag = 0;
        if (flag != 0)
            ReturnPyvoxError(ERR_FILE, 
                             "Unable to seek to the requested postion");
    }
    else if ( PyFile_Check(file) )
        infile = PyFile_AsFile(file);
    else
        ReturnPyvoxError(ERR_TYPE,
                         "Argument 1 isn't a filename or an open file");

    /* Determine the assumed rank and dimensions of the data */
    if ( !Pyvox_ParseShape(shape, &rank, dimen) )
	return NULL;

    /* Verify the external format */
    if ( type && !PyvoxExtype_Check(type, "extype", &extype) )
        ReturnPyvoxLibError();

    /* Choose the "natural" internal format for this data. */
    intype = exim_natural_intype(extype);
    if ( intype == 0 )
	ReturnPyvoxError(ERR_MISC, 
                         "Cannot find natural internal format for this data");
    
    /* Create a new Pyvox array */
    obj = parray_create(&vobj);
    if (obj == 0)    ReturnPyvoxLibError();

    /* Read the data into the voxel array */
    flag = vxl_read_raw_data(obj->varray, infile, rank, dimen,
                             intype, extype, bigend);
    if (flag)    ReturnPyvoxLibError();

    /* If originally given a filename, close the open file */
    if (filename)
        fclose(infile);

    /* Return the filled and wrapped voxel array to the caller */
    clear_error_context();
    return (PyObject *)obj;

}



/*------------------------------------------------------------------------
pyvox_read_pgm - Read image data from a PGM file into a voxel array

This Pyvox function reads image data from an external file in the
Portable Gray Map (PGM) format, converts it to the most natural internal
format, and returns it as a voxel array.  The only argument is the file
to read from; all other parameters are specified by the PGM header.

FIXME: Allow spacing and origin as optional arguments?
------------------------------------------------------------------------*/

static PyObject *
pyvox_read_pgm(
    PyObject *self,
    PyObject *args)
{
    PyObject *file;            /* Arg 1: A filename or an open file */

    char *filename = 0;        /* Name of the image file to open */
    pyvox_array *obj;          /* Result as a Pyvox array */
    voxel_array *vobj;         /* Result as a voxel array */
    int flag;

    set_error_context("pyvox.read_pgm()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "O", &file) )
        return NULL;
    filename = PyString_AsString(file);

    /* Create a new Pyvox array */
    obj = parray_create(&vobj);
    if (obj == 0)    ReturnPyvoxLibError();

    /* Read the data into the voxel array */
    flag = vxl_read_pgm_image(obj->varray, filename);
    if (flag)    ReturnPyvoxLibError();

    /* Return the filled and wrapped voxel array to the caller */
    clear_error_context();
    return (PyObject *)obj;
}



/*------------------------------------------------------------------------
pyvox_read_ppm - Read image data from a PPM file into a voxel array

This Pyvox function reads image data from an external file in the
Portable Pixel Map (PPM) format, converts it to the most natural internal
format, and returns it as a voxel array.  The only argument is the file
to read from; all other parameters are specified by the PPM header.

FIXME: Allow spacing and origin as optional arguments?
------------------------------------------------------------------------*/

static PyObject *
pyvox_read_ppm(
    PyObject *self,
    PyObject *args)
{
    PyObject *file;            /* Arg 1: A filename or an open file */

    char *filename = 0;        /* Name of the image file to open */
    pyvox_array *obj;          /* Result as a Pyvox array */
    voxel_array *vobj;         /* Result as a voxel array */
    int flag;

    set_error_context("pyvox.read_ppm()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "O", &file) )
        return NULL;
    filename = PyString_AsString(file);

    /* Create a new Pyvox array */
    obj = parray_create(&vobj);
    if (obj == 0)    ReturnPyvoxLibError();

    /* Read the data into the voxel array */
    flag = vxl_read_ppm_image(obj->varray, filename);
    if (flag)    ReturnPyvoxLibError();

    /* Return the filled and wrapped voxel array to the caller */
    clear_error_context();
    return (PyObject *)obj;
}



/*-----------------------------------------------------------------------
pyvox_refcnt - Return ref count of an Python object

This function returns the current reference count of the Python object
provided as its argument.
------------------------------------------------------------------------*/

static PyObject *
pyvox_refcnt(
    PyObject *self,
    PyObject *args)
{
    PyObject *ob;         /* Arg 1: Python object */
    PyObject *res;        /* Result: Its reference count */

    set_error_context("pyvox.refcnt()");

    /* Get the arguments */
    if ( ! PyArg_ParseTuple(args, "O", &ob) )
        return NULL;

    /* Get the ref count and return as a Python int */
    res = Py_BuildValue("i", ob->ob_refcnt);
    clear_error_context();
    return res;
}



/*-------------------------------------------------------------------------
set_pyvox_dict - Remember the dictionary for the python layer of pyvox

This function is called by the pyvox module as it is being
initialized, and is used to provide the dictionary of the pyvox module
to the Pyvox C level; the dictionary is used to find the callable
objects corresponding to specific functions and methods so that the
corresponding Python objects may be invoked from the C level.

NOTE: The value of self supplied by the caller is 0.  I suppose that
this indicates that this is not being called as a class method but as
a function defined within a modules.

FIXME: Can we also define new functions by inserting them directly in
this dictionary, rather than having to pass calls through the python
level?
--------------------------------------------------------------------------*/

static PyObject *
set_pyvox_dict(
    PyObject *self,
    PyObject *args)
{

    /* Save the dictionary */
    if ( !PyArg_ParseTuple(args, "O", &pyvox_dict) )
	return 0;

    /* Return the Python object None */
    Py_INCREF(Py_None);
    return Py_None;
}



/*-------------------------------------------------------------------------
pyvox_methods - Methods implemented by the Pyvox module

NOTE that the methods listed here are exported from pyvoxC, but are not
automatically exported from pyvox; you must modify pyvox.py to do that.
-------------------------------------------------------------------------*/

static struct PyMethodDef
pyvox_methods[] = {

    /* Public methods of the pyvox module */
    { "array",            pyvox_arraym,    1, "Create new Pyvox array" },
    { "const",            pyvox_arraym,    1, "(Deprecated alias for array)" },
    { "ccmap",            pyvox_ccmap,     1, "Connected component map" },
    { "fftramp",          pyvox_fftramp,   1, "New array of FFT frequencies" },
    { "kernel",           pyvox_kernel,    1, "Create new Pyvox kernel" },
    { "old_kernel",       pyvox_kernel_old, 1, "Create new Pyvox kernel" },
    { "lowpass",          pyvox_lowpass,   1, "Create std lowpass kernel" },
    { "ramp",             pyvox_ramp,      1, "New array filled with ramp" },
    { "randi",            pyvox_randi,     1, "Generate random choices" },
    { "randn",            pyvox_randn,     1, "Generate normal variates" },
    { "randu",            pyvox_randu,     1, "Generate uniform variates" },
    { "refcnt",           pyvox_refcnt,    1, "Ref count of a Python object" },

    /* Undocumented methods for debugging and development use */
    { "error_action",     error_action,    1 },    /* Set action on error */

    /* Hidden methods for the Python layer to use */
    { "affine_linear",    affine_linear,   1 },    /* Fast version */
    { "affine_linear0",   affine_linear0,  1 },    /* Reference version */
    { "affine_map",       affine_map,      1 },
    { "affine_nearest",   affine_nearest,  1 },    /* Fast version */
    { "affine_nearest0",  affine_nearest0, 1 },    /* Reference version */
    { "feature",          feature,         1 },    /* Check optional feature */
    { "poly_linear0",     poly_linear0,    1 },    /* Reference version */
    { "poly_linear",      poly_linear0,    1 },    /* Fast version to be */
    { "poly_map",         poly_map,        1 },
    { "rawread",          pyvox_rawread,   1 },    /* Read raw format data */
    { "read_pgm",         pyvox_read_pgm,  1 },    /* Read a PGM file */
    { "read_ppm",         pyvox_read_ppm,  1 },    /* Read a PPM file */
    { "set_pyvox_dict",   set_pyvox_dict,  1 },
    { NULL, NULL }
};



/*--------------------------------------------------------------------------
myinitpyvoxC - Initialize the Pyvox module

This function initializes the PyvoxC module, after loading the exim
module.  Pretty much standard stuff for a Python extension, but also
sets the errm error manager to continue on error so that Pyvox can
turn it into an exception.  Note that this is called from initpyvoxC
in pyvoxC.c rather than directly from Python; this is done so that the
exim and pyvox modules can share internals.
--------------------------------------------------------------------------*/

void
myinitpyvoxC()
{
    PyObject *module;        /* The Pyvox module being created */
    PyObject *dict;          /* ...and its dictionary */
    PyObject *exim = 0;      /* The exim module */

    /* Import the exim module */
    exim = PyImport_ImportModule("exim");
    if ( exim == 0 )
        Py_FatalError("Failed to import the exim module");

    /* Initialize this module (pyvoxC) */
    module = Py_InitModule("pyvoxC", pyvox_methods);
    dict   = PyModule_GetDict(module);
    set_error_action(0);     /* Continue on Pyvox library error */
    PyvoxError = Py_BuildValue("s", "PyvoxError");
    PyDict_SetItemString(dict, "error", PyvoxError);
    PyDict_SetItemString(dict, "ArrayType",  (PyObject *)(void *)&parray_type);
    PyDict_SetItemString(dict, "KernelType", (PyObject *)(void *)&kernel_type);
    if ( PyErr_Occurred() )
        Py_FatalError("Can't initialize pyvox module");

    return;
}
