/****************************************************************************
reduce.c - Voxel Kit routines for reduction operations

Author:  Paul Hughett

Copyright (c) 1998-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  This particular file contains code for reduction
operations and related functions.

Reduction operations on arrays are defined as operations that process
along selected axes of an array to obtain a scalar or an array of
lower rank; some functions which consist of reduction plus additional
operations are also included here.  Both unary and binary operations
are permitted.  Many of these functions (should) permit reduction
along specified axes only, to yield an array of lower rank rather than
a scalar.  Examples of reduction operations include array sum (as well
as row and column sums), array product, array minimum, array maximum,
any (boolean OR), all (boolean AND), norms, dot product, mean,
variance, covariance, and correlation.

These functions all face similar algorithmic problems in choosing the
access patterns to use a multi-level cache efficiently.  Several
different algorithms may currently coexist in this file, while I
experiment with different trade-offs between implementation time,
readability, and efficiency; the trend over time can be expected to be
toward improving efficiency, even at the cost of implementation time
and readability, since these functions are likely to be major
contributors to the total run time.

Note that the algorithms used here are designed to maximize locality
within a multilevel cache architecture, thereby decreasing the run
time.  If you don't understand what that means, you probably shouldn't
touch this code, under penalty of a possible order-of-magnitude
increase in run time.  

FIXME: Think through the right way to implement these functions to
keep as much common as possible.  Should the result array(s) always be
double, or is it necessary to allow other result types?  Can we assume
that the result is normally sufficiently low dimension that efficient
access is not a serious performance issue?  Note that the expensive
part of dot, norm2, (weighted) mean, variance, covariance, and
correlation consists of sums of powers or products of arrays.  For
example, computing the weighted mean requires the sum of the weights
and the sum of the weights times the data.  So a possible set of core
operations might include sumx, sumxx, sumxy, sumxxy(for weighted
variance), sumxyz, amin, amax.  It is probably convenient if compound
operations such as sumxy also optionally generate sumx and sumy.  Then

    dot      uses  sumxy  or  sumwxy
    norm2    uses  sumxx  or  sumwxx
    mean     uses  sumx   or  sumwx
    var      uses  sumxx  or  sumwxx, plus mean
    cov      uses  sumxy  or  sumwxy, plus mean and var
    corr     uses  cov, var

FIXME: Implement array.any() and array.all().  I think these should
operate on any type but return unsigned char.

FIXME: Modify the Pyvox array methods that call these to allow the
weight and axes arguments to be provided by keyword arguments.
*****************************************************************************/

static const char rcsid[] = "$Id: reduce.c,v 1.38 2005/12/27 19:59:03 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_TGMATH_H
#include <tgmath.h>
#else
#include <math.h>
#endif

#include "bips.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"



/*=========================================================================
Utility routines for handing axis selection in reductions

This function does the required setup for reducing an array over
selected axes, including determining the dimensions of the
destination.  It is used in conjunction with vxli_reduce_next_dest()
and vxli_reduce_next_src() to actually implement the reduction.  As a
performance feature, the first source row to be merged into a
destination row should be handled as a copy; and it may be necessary
to do some finishing up on a destination row after all source rows
have been merged into it.  The following template code shows how to do
this.

    int naxes, *axes;                 // Axes to reduce on
    voxel_array *src, *dest;          // Source and destination arrays
    struct vxli_reduce_state state;   // State of reduction loop
    long nelem;                       // Elements per row
    long dstep, sstep;                // Stride in dest and source
    unsigned char *drow, *srow;       // Current row location in src and dest

    ...check args, general setup, etc
    vxli_reduce_setup(src, naxes, axes, &state);
    nelem  = state.nelem;
    dstep  = state.dstep;
    sstep  = state.sstep;
    vxl_alloc_array(dest, type, state.drank, state.ddimen);
    ...allocate any working buffers, if needed
    do {
        drow = vxli_locate(dest, state.dcoord);
        ...initialize the current destination row, if needed
        do {
            srow = vxli_locate(src, state.scoord);
            // Add source rows to destination row
            flag = bips_foo(nelem, type, drow, dstep, srow, sstep);
            if (flag) ...
            }
        } while (vxli_reduce_next_srow(&state);
        ...finish up the current destination row, if needed
    } while (vxli_reduce_next_drow(&state));
    ...free any working buffers, if needed


In the call to vxli_reduce_setup(), setting naxes to a negative
number means to reduce on all axes; setting naxes to zero means to
reduce on no axes, which is legal but usually devolves into a
expensive copy; setting naxes to the rank or greater is legal but
means that axes contains duplicates.  The elements in axes may be in
any order, and may be repeated without harm.  Setting one of these
elements to a value in [0, rank-1] means to reduce that axis, counting
from the left; setting to [-1, -rank] means to reduce that axis,
counting from the right.  (That is, -1 is equivalent to rank-1, or the
last axis.)  All other values are illegal.


ALGORITHM: The reduction is always vectorized using some bips function
running over the last source axis, with two possible cases.  (1) If
the last axis is being reduced on, then the innermost loop (hidden
inside bips) reads from the last source axis with stride 1 and writes
to a single destination element with stride 0.  (2) Otherwise, the
last axis is not being reduced on, and the innermost loop reads from
the last source axis with stride 1 and writes to the destination row
with stride 1.  The fields dstep and sstep from the state struct give
the strides for this inner loop; nelem is the number of elements
in each row.

The middle loop is updated by vxli_reduce_next_srow() and runs over
the source rows that reduce to a single destination row (or element),
adding (or whatever) each source row t the destination row.  Any
initialization or finalization of that destination row can be done
before or after this loop.  The rmap field in the struct lists the
source axes that must be updated in this loop; the last source axis
is NOT included here, since bips vectorizes over that axis.  The
scoord and dcoord fields give the location of the first element of the
current row in the source and destination, although only scoord needs
to be updated; the last coordinate of scoord and dcoord is always set
to zero to indicate the first element of the row.

The outer loop is updated by vxli_reduce_next_drow() and runs over all
the destination rows.  The dmap field in the struct lists the destination
axes that must be updated in this loop; the axes handled in the middle
and inner loops are not included.  The scoord and dcoord fields give the
location of the first element of the current source and destination rows.

The vxl_asum() function is heavily instrumented for the purpose of
testing and debugging these functions.  Defining DEBUG_AXES to 1 will
turn on instrumentation in these functions and in vxl_asum() for
debugging.  See also vxl_asum0() which is a slow but obviously correct
reference version that can be used for testing.

FIXME: This implementation is pretty good, but is still lacking in a
number of areas: (1) Simply adding (or whatever) each source element
that contributes to a destination element is probably not the right
approach numerically; a hierarchical summation of some sort is
probably better.  (2) Similarly, adding each source element in turn to
the destination element introduces a pipeline dependency and is
probably slower than a hierarchical summation.  (3) The practice here
is to initialize the destination row with the identity element for the
operation.  This adds an extra step relative to copying the first element
and then adding all the remaining elements to it.  (4) Cache coherence
should be reasonably good, but an explicit examination might not be a
bad idea.  (5) The code here assumes that later axes have smaller strides,
but this might not remain true once views are introduced.

There may be enough issues here to justify a paper, but not yet.  I'm
going to get it working first, with solid regression tests, and then
start making it really, really fast.
==========================================================================*/

#define DEBUG_AXES  0


/*--------------------------------------------------------------------------
vxli_reduce_state - Struct to contain state of a reduction loop

This struct contains the state of a rowwise reduction loop set up by
vxli_reduce_setup() and updated by vxli_reduce_next_drow() and 
vxli_reduce_next_drow().  See above for more details.
---------------------------------------------------------------------------*/

struct 
vxli_reduce_state {
    int  srank;                   /* Source rank */
    long sdimen[VXL_MAX_RANK];    /* Source dimensions */
    long sstep;                   /* Source stride */
    int  drank;                   /* Destination rank */
    int  drnk2;                   /* Nmbr of dest axes to iterate over */
    long ddimen[VXL_MAX_RANK];    /* Destination dimensions */
    int  dmap  [VXL_MAX_RANK];    /* Destination axis to source axis */
    long dstep;                   /* Destination stride */
    int  rrank;                   /* Number of reduction rows */
    int  rmap  [VXL_MAX_RANK];    /* Reduction axis to source axis */
    long sperd;                   /* Source elements per destination element */
    long nelem;                   /* Number of elements per (source) row */
    long ndest;                   /* Number of elements per dest row */
    long scoord[VXL_MAX_RANK];    /* Coordinates of current source row */
    long dcoord[VXL_MAX_RANK];    /* Coordinates of current destination row */
};



/*-------------------------------------------------------------------------
vxli_reduce_setup - Initialize a rowwise reduction loop

This function does the required setup for reducing an array over
selected axes, including determining the dimensions of the
destination.  It is used in conjunction with vxli_reduce_next_dest()
and vxli_reduce_next_src() to actually implement the reduction.  See
above for more-detailed usage instructions.

Note that ndest <= nelem, which is often useful in application coding.
--------------------------------------------------------------------------*/

static inline
int                                    /* Returns non-zero on error */
vxli_reduce_setup(
    voxel_array *src,                  /* In: Source array */
    int   naxes,                       /* In: Number of axes to reduce along */
    int  *axes,                        /* In: Axes to reduce along */
    struct vxli_reduce_state *state)   /* Initial state for reduction loop */
{
    int   srank;                       /* Rank of source */
    long *sdimen;                      /* Dimensions of source */
    int reduce[VXL_MAX_RANK];          /* Reduce this axis? */
    int i, iaxis, drank, rrank;
    long sperd;

    /* Copy source arguments into the state */
    srank  = src->rank;
    sdimen = src->dimen;
    state->srank = srank;
    for (i = 0; i < srank; i++)
        state->sdimen[i] = sdimen[i];

    /* Convert axes definition to tabular format */
    if ( naxes < 0 )
	for (i = 0; i < VXL_MAX_RANK; i++)
	    reduce[i] = 1;
    else {
	for (i = 0; i < VXL_MAX_RANK; i++)
	    reduce[i] = 0;
	for (i = 0; i < naxes; i++) {
	    iaxis = axes[i];
	    if  (0 <= iaxis && iaxis < srank)
		reduce[iaxis] = 1;
	    else if ( -srank <= iaxis && iaxis < 0 )
		reduce[srank+iaxis] = 1;
	    else
		NZReturnError(ERR_ILLOP,
                              "Reduction axis exceeds rank of array");
	}
    }

    /* Classify source axes and set up dest and reduce maps */
    drank = 0;
    rrank = 0;
    sperd = 1;
    for (i = 0; i < srank; i++)
	if ( ! reduce[i] ) {
	    state->dmap[drank] = i;
	    state->ddimen[drank++] = src->dimen[i];
	}
        else {
	    sperd *= sdimen[i];
            if (i < srank-1) 
                state->rmap[rrank++] = i;
	}

    /* Vectorize on last source axis */
    state->sstep = 1;
    state->nelem = state->sdimen[srank-1];
    if ( reduce[srank-1] ) {
        state->dstep = 0;
        state->ndest = 1;
        state->drnk2 = drank;
    }
    else {
        state->dstep = 1;
        state->ndest = state->nelem;
        state->drnk2 = drank-1;
    }

    /* Store useful information into the state */
    state->drank = drank;
    state->rrank = rrank;
    state->sperd = sperd;

    /* Initialize source and destination coordinates */
    for (i = 0; i < VXL_MAX_RANK; i++) {
        state->scoord[i] = 0;
        state->dcoord[i] = 0;
    }

    /* DEBUG: Report the setup */
    if (DEBUG_AXES) {
        int j;
        fprintf(stderr, "vxli_reduce_setup:\n");
        fprintf(stderr, "   drank  = %d, drnk2 = %d\n   ddimen = ",
                state->drank, state->drnk2);
        for (j = 0; j < drank; j++) {
            fprintf(stderr, "%ld, ", state->ddimen[j]);
        }
        fprintf(stderr, "\n   dmap   = ");
        for (j = 0; j < state->drnk2; j++)
            fprintf(stderr, "%d, ", state->dmap[j]);
        fprintf(stderr, "\n   rrank  = %d\n   rmap   = ", state->rrank);
        for (j = 0; j < rrank; j++) {
            fprintf(stderr, "%d, ", state->rmap[j]);
        }
        fprintf(stderr, "\n   nelem  = %ld, dstep = %ld, sstep = %ld\n", \
                state->nelem, state->dstep, state->sstep);
    }        

    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxli_reduce_next_srow - Step to next reduction row in source

This function steps to the next source row that contributes to the current
destination row, or returns zero to indicate that the current destination
row is finished.    See above for more-detailed usage instructions.
--------------------------------------------------------------------------*/

static inline
int                             /* Returns non-zero until last row */
vxli_reduce_next_srow(
    struct vxli_reduce_state *state)
{
    int i, isrc;

    for (i = state->rrank-1; i >= 0; i--) {
        isrc = state->rmap[i];
        if (1)      /* DEBUG: FIXME: Make into an assertion? */
            if ( isrc < 0 || isrc >= VXL_MAX_RANK )
                abort();
        if ( ++state->scoord[isrc] < state->sdimen[isrc] )
            return 1;
        else
            state->scoord[isrc] = 0;
    }

    return 0;
}



/*------------------------------------------------------------------------
vxli_reduce_next_drow - Step to the next destination row in a reduction

This function steps to the next destination row in a reduction, or
returns zero to indicate that all destination rows have been
processed.  See above for more-detailed usage instructions.
-------------------------------------------------------------------------*/

static inline
int                             /* Returns non-zero until last row */
vxli_reduce_next_drow(
    struct vxli_reduce_state *state)
{
    int idst, isrc;

    for (idst = state->drnk2-1; idst >= 0; idst--) {
        isrc = state->dmap[idst];
        if (0)   /* DEBUG: FIXME: Make into an assertion? */
            if ( isrc < 0 || isrc >= VXL_MAX_RANK )
                abort();
        if ( ++state->dcoord[idst] < state->ddimen[idst] ) {
            state->scoord[isrc] = state->dcoord[idst];
            return 1;
        }
        else {
            state->dcoord[idst] = 0;
            state->scoord[isrc] = 0;
        }
    }

    return 0;        
        
}



/*==========================================================================
Routines for specific types of reductions
==========================================================================*/



/*-------------------------------------------------------------------------
vxl_amax - Find maximum value in a voxel array

This function find the maximum value along specified axes of a voxel
array, returning the results as an array of the same type but lower
rank.  Setting the number of axes to zero means to reduce along all
axes; that is, to compute the maximum value of the entire array.

FIXME: A possible extension to complex arrays would be to compute the
maximum absolute value.
---------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_amax(
    voxel_array *dest,          /* Result of computation */
    voxel_array *src,           /* Array to reduce */
    int naxes,                  /* Number of axes to reduce on, or zero */
    int *axes)                  /* Axes to reduce on */
{
    int srank, drank;            /* Rank of source and destination */
    int type = 0;                /* Data type */
    long nelem;                  /* Number of elements per source row */
    long sstep, dstep;           /* Stride in source and destination */
    unsigned char *srow, *drow;  /* Location of current src and dest row */
    struct vxli_reduce_state state;
    int i, i2, flag;

    /* Check validity of source */
    if  ( src == 0 || 
          src->magic != VXL_MAGIC ||
          src->data == 0 ||
          exim_sizeof_intype((type = src->type)) == 0 ||
          vxli_count(src) < 1 )
        NZReturnError(ERR_INIT, "Invalid source array");
    srank  = src->rank;

    /* Set up the reduction loop */
    flag = vxli_reduce_setup(src, naxes, axes, &state);
    if (flag)    NZReturnLibError();
    nelem = state.nelem;
    drank = state.drank;
    dstep = state.dstep;
    sstep = state.sstep;

    /* Allocate the destination array; set origin and spacing */
    flag = vxl_alloc_array(dest, type, drank, state.ddimen);
    if (flag)    NZReturnLibError();
    for (i = 0; i < drank; i++) {
        i2 = state.dmap[i];
        dest->origin [i] = src->origin [i2];
        dest->spacing[i] = src->spacing[i2];
    }

    /* Loop over destination rows */
    do {

        /* Initialize the destination row */
	drow = vxli_locate(dest, state.dcoord);
        flag = bips_fill_min(nelem, type, drow, dstep);
        if (flag)    NZReturnLibError();

        /* Loop over source rows for current destination row */
        do {
            srow = vxli_locate(src, state.scoord);
            flag = bips_max_set(nelem, type, drow, dstep, srow, sstep);
            if (flag)    NZReturnLibError();

        } while ( vxli_reduce_next_srow(&state) );

    } while ( vxli_reduce_next_drow(&state) );

    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_amin - Find minimum value in a voxel array

This function find the minimum value along specified axes of a voxel
array, returning the results as an array of the same type but lower
rank.  Setting the number of axes to zero means to reduce along all
axes; that is, to compute the maximum value of the entire array.

FIXME: A possible generalization to complex arrays would be to compute
the minimum absolute value.
---------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_amin(
    voxel_array *dest,          /* Result of computation */
    voxel_array *src,           /* Array to reduce */
    int naxes,                  /* Number of axes to reduce on, or zero */
    int *axes)                  /* Axes to reduce on */
{
    int srank, drank;            /* Rank of source and destination */
    int type = 0;                /* Data type */
    long nelem;                  /* Number of elements per source row */
    long sstep, dstep;           /* Stride in source and destination */
    unsigned char *srow, *drow;  /* Location of current src and dest row */
    struct vxli_reduce_state state;
    int i, i2, flag;

    /* Check validity of source */
    if  ( src == 0 || 
          src->magic != VXL_MAGIC ||
          src->data == 0 ||
          exim_sizeof_intype((type = src->type)) == 0 ||
          vxli_count(src) < 1 )
        NZReturnError(ERR_INIT, "Invalid source array");
    srank  = src->rank;

    /* Set up the reduction loop */
    flag = vxli_reduce_setup(src, naxes, axes, &state);
    if (flag)    NZReturnLibError();
    nelem = state.nelem;
    drank = state.drank;
    dstep = state.dstep;
    sstep = state.sstep;

    /* Allocate the destination array; set origin and spacing */
    flag = vxl_alloc_array(dest, type, drank, state.ddimen);
    if (flag)    NZReturnLibError();
    for (i = 0; i < drank; i++) {
        i2 = state.dmap[i];
        dest->origin [i] = src->origin [i2];
        dest->spacing[i] = src->spacing[i2];
    }

    /* Loop over destination rows */
    do {

        /* Initialize the destination row */
	drow = vxli_locate(dest, state.dcoord);
        flag = bips_fill_max(nelem, type, drow, dstep);
        if (flag)    NZReturnLibError();

        /* Loop over source rows for current destination row */
        do {
            srow = vxli_locate(src, state.scoord);
            flag = bips_min_set(nelem, type, drow, dstep, srow, sstep);
            if (flag)    NZReturnLibError();

        } while ( vxli_reduce_next_srow(&state) );

    } while ( vxli_reduce_next_drow(&state) );

    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_aprod - Compute product of all voxels in a voxel array

This function computes the product of all elements in a voxel array and
returns the result as an array of the same type but lower rank.

FIXME: Compute by rows for better round-off properties?  Or is it only
for summation that this matters?

FIXME: Always promote to double or double complex for better range and
accuracy?
---------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_aprod(
    voxel_array *dest,          /* Result of computation */
    voxel_array *src,           /* Array to reduce */
    int naxes,                  /* Number of axes to reduce on, or zero */
    int *axes)                  /* Axes to reduce on */
{
    int srank, drank;            /* Rank of source and destination */
    int type = 0;                /* Data type */
    long nelem;                  /* Number of elements per source row */
    long sstep, dstep;           /* Stride in source and destination */
    unsigned char *srow, *drow;  /* Location of current src and dest row */
    struct vxli_reduce_state state;
    int i, i2, flag;

    /* Check validity of source */
    if  ( src == 0 || 
          src->magic != VXL_MAGIC ||
          src->data == 0 ||
          exim_sizeof_intype((type = src->type)) == 0 ||
          vxli_count(src) < 1 )
        NZReturnError(ERR_INIT, "Invalid source array");
    srank  = src->rank;

    /* Set up the reduction loop */
    flag = vxli_reduce_setup(src, naxes, axes, &state);
    if (flag)    NZReturnLibError();
    nelem = state.nelem;
    drank = state.drank;
    dstep = state.dstep;
    sstep = state.sstep;

    /* Allocate the destination array; set origin and spacing */
    flag = vxl_alloc_array(dest, type, drank, state.ddimen);
    if (flag)    NZReturnLibError();
    for (i = 0; i < drank; i++) {
        i2 = state.dmap[i];
        dest->origin [i] = src->origin [i2];
        dest->spacing[i] = src->spacing[i2];
    }

    /* Loop over destination rows */
    do {

        /* Initialize the destination row */
	drow = vxli_locate(dest, state.dcoord);
        flag = bips_one(nelem, type, drow, dstep);
        if (flag)    NZReturnLibError();

        /* Loop over source rows for current destination row */
        do {
            srow = vxli_locate(src, state.scoord);
            flag = bips_mul_set(nelem, type, drow, dstep, srow, sstep);
            if (flag)    NZReturnLibError();

        } while ( vxli_reduce_next_srow(&state) );

    } while ( vxli_reduce_next_drow(&state) );

    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_asum - Compute sum of all voxels in a voxel array

This function sums all the voxel values along specified axes of a voxel
array, returning the results as an array of the same type but lower
rank.  Setting the number of axes to zero means to reduce along all
axes; that is, to compute the sum of all voxels in the entire array.

The function vxl_asum0() is a slow but obviously correct reference
version that can be used for testing this function.

ALGORITHM: This particular reduction function is heavily instrumented
so that it can be used to test and debug the generic axis selection
code.  We use addition rather than min or max since it is more
sensitive to changes in the inputs or algoritm.

FIXME: Avoid roundoff error by summing over rows or slices first?

FIXME: Always promote to double or double complex for better range and
accuracy?
---------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_asum(
    voxel_array *dest,          /* Result of computation */
    voxel_array *src,           /* Array to reduce */
    int naxes,                  /* Number of axes to reduce on, or zero */
    int *axes)                  /* Axes to reduce on */
{
    int srank, drank;            /* Rank of source and destination */
    long *sdimen;                /* Dimensions of source */
    int type = 0;                /* Data type */
    long nelem;                  /* Number of elements per source row */
    long sstep, dstep;           /* Stride in source and destination */
    unsigned char *srow, *drow;  /* Location of current src and dest row */
    struct vxli_reduce_state state;
    int i, i2, flag;

    /* Check validity of source */
    if  ( src == 0 || 
          src->magic != VXL_MAGIC ||
          src->data == 0 ||
          exim_sizeof_intype((type = src->type)) == 0 ||
          vxli_count(src) < 1 )
        NZReturnError(ERR_INIT, "Invalid source array");
    srank  = src->rank;
    sdimen = src->dimen;

    /* Set up the reduction loop */
    flag = vxli_reduce_setup(src, naxes, axes, &state);
    if (flag)    NZReturnLibError();
    nelem = state.nelem;
    drank = state.drank;
    dstep = state.dstep;
    sstep = state.sstep;

    /* Allocate the destination array; set origin and spacing */
    flag = vxl_alloc_array(dest, type, drank, state.ddimen);
    if (flag)    NZReturnLibError();
    for (i = 0; i < drank; i++) {
        i2 = state.dmap[i];
        dest->origin [i] = src->origin [i2];
        dest->spacing[i] = src->spacing[i2];
    }

    /* Loop over destination rows */
    do {

        /* Initialize the destination row */
	drow = vxli_locate(dest, state.dcoord);
        flag = bips_zero(nelem, type, drow, dstep);
        if (flag)    NZReturnLibError();

        if (DEBUG_AXES) {
            fprintf(stderr, "dcoord: %ld, %ld\n",
                    state.dcoord[0], state.dcoord[1]);
            fprintf(stderr, "zero:  %p : %p\n", dest->data, drow);
        }

        /* Loop over source rows for current destination row */
        do {
            srow = vxli_locate(src, state.scoord);
            flag = bips_add_set(nelem, type, drow, dstep, srow, sstep);
            if (flag)    NZReturnLibError();

            if (DEBUG_AXES) {
                fprintf(stderr, "scoord: %ld, %ld\n",
                        state.scoord[0], state.scoord[1]);
                fprintf(stderr, "add:  %p: %p <- %p: %p\n",
                        dest->data, drow, src->data, srow);
            }

        } while ( vxli_reduce_next_srow(&state) );

    } while ( vxli_reduce_next_drow(&state) );

    return ERR_OKAY;
}



/*---------------------------------------------------------------------------
vxl_asum0 - Reference version of vxl_asum

This function is a slow but obviously correct version of vxl_asum and is
intended for testing it.
-----------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_asum0(
    voxel_array *dest,          /* Result of computation */
    voxel_array *src,           /* Array to reduce */
    int naxes,                  /* Number of axes to reduce on, or zero */
    int *axes)                  /* Axes to reduce on */
{
    int srank, drank;            /* Rank of source and destination */
    long *sdimen;                /* Dimensions of source */
    long scoord[VXL_MAX_RANK];   /* Current voxel in source */
    int  reduce[VXL_MAX_RANK];   /* Reduce this axis? */
    long ddimen[VXL_MAX_RANK];   /* Dimensions of the destination */
    int  dmap  [VXL_MAX_RANK];   /* Map dest axis to src axis */
    long dcoord[VXL_MAX_RANK];   /* Current voxel in destination */
    int type = 0;                /* Data type */
    long nelem;                  /* Number of elements per source row */
    unsigned char *srow, *drow;  /* Location of current src and dest row */
    int i, iaxis, flag;

    /* Check validity of source */
    if  ( src == 0 || 
          src->magic != VXL_MAGIC ||
          src->data == 0 ||
          exim_sizeof_intype((type = src->type)) == 0 ||
          vxli_count(src) < 1 )
        NZReturnError(ERR_INIT, "Invalid source array");
    srank  = src->rank;
    sdimen = src->dimen;

    /* Convert axes definition to tabular format */
    if ( naxes < 0 )
	for (i = 0; i < VXL_MAX_RANK; i++)
	    reduce[i] = 1;
    else {
	for (i = 0; i < VXL_MAX_RANK; i++)
	    reduce[i] = 0;
	for (i = 0; i < naxes; i++) {
	    iaxis = axes[i];
	    if  (0 <= iaxis && iaxis < srank)
		reduce[iaxis] = 1;
	    else if ( -srank <= iaxis && iaxis < 0 )
		reduce[srank+iaxis] = 1;
	    else
		NZReturnError(ERR_ILLOP,
                              "Reduction axis exceeds rank of array");
	}
    }

    /* Determine dest dimensions and map to source axes */
    drank = 0;
    for (i = 0; i < srank; i++)
	if ( ! reduce[i] ) {
            dmap  [drank]   = i;
	    ddimen[drank++] = src->dimen[i];
	}

    /* Allocate the destination array; set origin and spacing */
    flag = vxl_alloc_array(dest, type, drank, ddimen);
    if (flag)    NZReturnLibError();
    for (i = 0; i < drank; i++) {
        dest->origin [i] = src->origin [dmap[i]];
        dest->spacing[i] = src->spacing[dmap[i]];
    }

    /* Zero the destination */
    nelem = vxli_count(dest);
    flag  = bips_zero(nelem, type, dest->data, 1);

    /* Loop over the source voxels */
    vxli_firstvoxel(srank, scoord);
    do {

        /* Compute corresponding voxel in destination */
        for (i = 0; i < drank; i++)
            dcoord[i] = scoord[dmap[i]];

        /* Compute voxel locations in memory */
        srow = vxli_locate(src,  scoord);
        drow = vxli_locate(dest, dcoord);

        /* Add that voxel */
        flag = bips_add_set(1, type, drow, 0, srow, 0);
        if (flag)    NZReturnLibError();        

    } while ( vxli_nextvoxel(srank, sdimen, scoord) );

    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_dot - Compute weighted vector dot product of two voxel arrays

This function computes the weighted vector dot product, defined as
sum_n w_n x_n* y_n, of two voxel arrays.  If weight is a null pointer,
then a uniformly weighted dot product is computed; otherwise, weight
supplies the weights for each voxel.  If src2 is a null pointer, then
a simple weighted sum is computed.  The src1 argument may not be zero.
The arrays src1, src2, and weight must be the same rank and
dimensions.  The arrays src1, src2, and weight may be of any type and
will be converted to double or double complex for the computation.

FIXME: Add an argument to specify the destination type?

FIXME: Remove the option to set src2 = 0?
--------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_dot(
    voxel_array *dest,          /* Result of calculation */
    voxel_array *src1,          /* First array */
    voxel_array *src2,          /* Second array, or zero */
    voxel_array *weight,        /* Voxel weights, or zero */
    int naxes,                  /* Number of axes to reduce on */
    int *axes)                  /* The axes to reduce one */
{
    int stype1=0, stype2=0;     /* Type of source arrays */
    int wtype=0;                /* Type of weight array */
    int dtype=0;                /* Type of destination array */
    int dbyte=0;                /* Bytes per element in destination */
    int drank, srank;           /* Rank of the destination and source arrays */
    unsigned char *abuf=0, *bbuf=0;    /* Working storage for a row */
    unsigned char *drow=0;      /* Pointer to beginning of each dest row */
    unsigned char *s1=0, *s2=0; /* Pointers to beginning of each source row */
    unsigned char *w=0;         /* Pointer to beginning of each weight row */
    long nelem;                 /* Number of elements in a row */
    long sstep, dstep;          /* Stride in source and destination */
    int flag;                   /* Error indicator */
    struct vxli_reduce_state state;
    long i, i2;

    /* Check validity of source 1 */
    if  (  src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0 ||
           exim_sizeof_intype((stype1 = src1->type)) == 0 ) 
        NZReturnError(ERR_INIT, "Invalid source 1 array");
    srank  = src1->rank;

    /* Check validity of source 2 */
    if ( src2 != 0 ) {
	if  (  src2->magic != VXL_MAGIC || src2->data == 0 ||
               exim_sizeof_intype((stype2 = src2->type)) == 0 ) 
            NZReturnError(ERR_INIT, "Invalid source 2 array");
	if ( !vxli_same_physical(src1, src2) )     NZReturnLibError();
    }

    /* Check validity of weight array */
    if ( weight != 0 ) {
	if  (  weight->magic != VXL_MAGIC || weight->data == 0 ||
               exim_sizeof_intype((wtype = weight->type)) == 0 )
            NZReturnError(ERR_INIT, "Invalid weight array");
	if ( !vxli_same_physical(weight, src1) )    NZReturnLibError();
    }

    /* Select data type for destination */
    if ( stype1 == INTP_FCOMPLEX || stype1 == INTP_DCOMPLEX ||
         stype2 == INTP_FCOMPLEX || stype2 == INTP_DCOMPLEX ||
         wtype  == INTP_FCOMPLEX || wtype  == INTP_DCOMPLEX )
        dtype = INTP_DCOMPLEX;
    else
        dtype = INTP_DOUBLE;
    dbyte = exim_sizeof_intype(dtype);

    /* Set up the reduction loop */
    flag = vxli_reduce_setup(src1, naxes, axes, &state);
    if (flag)    NZReturnLibError();
    nelem = state.nelem;
    drank = state.drank;
    dstep = state.dstep;
    sstep = state.sstep;

    /* Allocate the destination array; set origin and spacing */
    flag = vxl_alloc_array(dest, dtype, drank, state.ddimen);
    if (flag)    NZReturnLibError();
    for (i = 0; i < drank; i++) {
        i2 = state.dmap[i];
        dest->origin [i] = src1->origin [i2];
        dest->spacing[i] = src1->spacing[i2];
    }

    /* Allocate working storage for one row */
    abuf = mallock(2*nelem*dbyte);
    bbuf = abuf + nelem*dbyte;
    if (abuf == 0)    NZReturnLibError();

    /* Loop over destination rows */
    do {

        /* Initialize the destination row */
	drow = vxli_locate(dest, state.dcoord);
        flag = bips_zero(nelem, dtype, drow, dstep);
        if (flag)    NZReturnLibError();

        /* Loop over source rows for current destination row */
        do {

            /* Fetch source 1; conjugate if complex */
            s1 = vxli_locate(src1, state.scoord);
            if (s1 == 0)    NZReturnLibError();
            if (dtype != INTP_DCOMPLEX) {
                flag = bips_cast(nelem, dtype, abuf, 1, stype1, s1, sstep);
            }
#ifdef HAVE_COMPLEX_H
            else {
                flag = bips_cast(nelem, dtype, bbuf, 1, stype1, s1, sstep);
                if (flag)    NZReturnLibError();
                flag = bips_conj(nelem, dtype, abuf, 1, bbuf, 1);
            }
#endif
            if (flag)    NZReturnLibError();

            /* Multiply by source 2, if provided */
            if (src2 != 0) {
                s2 = vxli_locate(src2, state.scoord);
                if (s2 == 0)    NZReturnLibError();
                flag = bips_cast(nelem, dtype, bbuf, 1, stype2, s2, sstep);
                if (flag)    NZReturnLibError();
                flag = bips_mul_set(nelem, dtype, abuf, 1, bbuf, 1);
                if (flag)    NZReturnLibError();
            }
	
            /* Multiply by weights, if provided */
            if (weight != 0) {
                w = vxli_locate(weight, state.scoord);
                if (w == 0)    NZReturnLibError();
                flag = bips_cast(nelem, dtype, bbuf, 1, wtype, w, sstep);
                if (flag)   NZReturnLibError();
                flag = bips_mul_set(nelem, dtype, abuf, 1, bbuf, 1);
                if (flag)    NZReturnLibError();
            }

            /* Add into destination row */
            flag = bips_add_set(nelem, dtype, drow, dstep, abuf, 1);
            if (flag)   NZReturnLibError();

        } while ( vxli_reduce_next_srow(&state) );

    } while ( vxli_reduce_next_drow(&state) );


    /* Free working storage and return */
    free(abuf);
    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_mean - Compute mean along selected axes of an image

This function computes weighted mean along specified axes of a voxel
array, returning the results as a double or double complex array of
lesser dimension.  Setting the number of axes to zero means to reduce
along all axes; that is, to compute the weighted mean of the entire
array.  If the weights sum to zero for some destination element, then
you get whatever your platform provides for division by zero; this
should be nan on IEEE-754 platforms.

FIXME: Division by zero is possible and may cause problems if the platform
does not support IEEE-754 and nans.

FIXME: Provide sum of weights as optional output?

FIXME: Add an argument to specify the destination type?
--------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_mean(
    voxel_array *dest,          /* Result of calculation */
    voxel_array *src1,          /* First array */
    voxel_array *weight,        /* Voxel weights, or zero */
    int naxes,                  /* Number of axes to reduce on */
    int *axes)                  /* The axes to reduce one */
{
    int stype1=0, wtype=0;      /* Type of source and weight arrays */
    int dtype=0;                /* Type of destination array */
    int dbyte=0;                /* Bytes per element in destination */
    int drank, srank;           /* Rank of the destination and source arrays */
    unsigned char *sbuf=0;      /* Buffer for source as double */
    unsigned char *wbuf=0;      /* Buffer for weight as double */
    unsigned char *zbuf=0;      /* Buffer for sum of weights */
    unsigned char *drow=0;      /* Pointer to beginning of each dest row */
    unsigned char *srow=0;      /* Pointer to beginning of each source row */
    unsigned char *wrow=0;      /* Pointer to beginning of each weight row */
    long nelem, ndest;          /* Number of elements in a src or dest row */
    long sstep, dstep;          /* Stride in source and destination */
    int flag;                   /* Error indicator */
    struct vxli_reduce_state state;
    long i, i2;

    /* Check validity of source */
    if  (  src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0 ||
           exim_sizeof_intype((stype1 = src1->type)) == 0 ) 
        NZReturnError(ERR_INIT, "Invalid source 1 array");
    srank  = src1->rank;

    /* Check validity of weight array */
    if ( weight != 0 ) {
	if  (  weight->magic != VXL_MAGIC || weight->data == 0 ||
               exim_sizeof_intype((wtype = weight->type)) == 0 )
            NZReturnError(ERR_INIT, "Invalid weight array");
	if ( !vxli_same_physical(weight, src1) )    NZReturnLibError();
    }

    /* Determine destination type */
    if ( stype1 == INTP_FCOMPLEX || stype1 == INTP_DCOMPLEX ||
         wtype  == INTP_FCOMPLEX || wtype  == INTP_DCOMPLEX )
        dtype = INTP_DCOMPLEX;
    else
        dtype = INTP_DOUBLE;
    dbyte = exim_sizeof_intype(dtype);

    /* Set up the reduction loop */
    flag = vxli_reduce_setup(src1, naxes, axes, &state);
    if (flag)    NZReturnLibError();
    nelem = state.nelem;
    ndest = state.ndest;
    drank = state.drank;
    dstep = state.dstep;
    sstep = state.sstep;

    /* Allocate the destination array; set origin and spacing */
    flag = vxl_alloc_array(dest, dtype, drank, state.ddimen);
    if (flag)    NZReturnLibError();
    for (i = 0; i < drank; i++) {
        i2 = state.dmap[i];
        dest->origin [i] = src1->origin [i2];
        dest->spacing[i] = src1->spacing[i2];
    }

    /* Allocate working storage for one row */
    sbuf = mallock(3*nelem*dbyte);
    if (sbuf == 0)    NZReturnLibError();
    wbuf = sbuf + nelem*dbyte;
    zbuf = wbuf + nelem*dbyte;

    /* Build array of constant weights if not otherwise provided */
    if ( ! weight ) {
        flag = bips_fill_range(0, nelem, dtype, zbuf, 1, (double)state.sperd);
        if (flag)    NZReturnLibError();
    }

    /* Loop over destination rows */
    do {

        /* Initialize the destination row and sum of weights */
	drow = vxli_locate(dest, state.dcoord);
        flag = bips_zero(nelem, dtype, drow, dstep);
        if (flag)    NZReturnLibError();
        if (weight) {
            flag = bips_zero(nelem, dtype, zbuf, 1);
            if (flag)    NZReturnLibError();
        }

        /* Loop over source rows for current destination row */
        do {

            /* Fetch source row */
            srow = vxli_locate(src1, state.scoord);
            if (srow == 0)    NZReturnLibError();
            flag = bips_cast(nelem, dtype, sbuf, 1, stype1, srow, sstep);
            if (flag)    NZReturnLibError();

            /* Process current row of weights, if provided */
            if (weight) {
                
                /* Convert weights to destination type*/
                wrow = vxli_locate(weight, state.scoord);
                if (wrow == 0)    NZReturnLibError();
                flag = bips_cast(nelem, dtype, wbuf, 1, wtype, wrow, sstep);
                if (flag)    NZReturnLibError();

                /* Multiply source by weights */
                flag = bips_mul_set(nelem, dtype, sbuf, 1, wbuf, 1);
                if (flag)    NZReturnLibError();

                /* Sum weights */
                flag = bips_add_set(nelem, dtype, zbuf, dstep, wbuf, 1);
                if (flag)   NZReturnLibError();

            }

            /* Add (possibly) weighted source to dest */
            flag = bips_add_set(nelem, dtype, drow, dstep, sbuf, 1);
            if (flag)    NZReturnLibError();

        } while ( vxli_reduce_next_srow(&state) );

        /* Divide destination row by count or sum of weights */
        flag = bips_div_set(ndest, dtype, drow, dstep, zbuf, dstep);
        if (flag)   NZReturnLibError();

    } while ( vxli_reduce_next_drow(&state) );

    /* Free working storage and return */
    free(sbuf);
    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_minmax - Find minimum and maximum values in a voxel array

This function finds the minimum and maximum values contained in a voxel
array and returns them as a double value.  Either minres or maxres may
be a null pointer, in which case the corresponding value is not computed.
This operation is undefined, and unimplemented, for complex arrays.

FIXME: Assumes that type double is no shorter than any other supported
internal real type.

FIXME: A possible generalization to complex arrays is to compute the
minimum and maximum absolute values.

FIXME: Generalize to return long double as needed.  
---------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_minmax(
    voxel_array *dmin,          /* Minimum values */
    voxel_array *dmax,          /* Maximum values */
    voxel_array *src,           /* Array to reduce */
    int naxes,                  /* Number of axes to reduce on, or zero */
    int *axes)                  /* Axes to reduce on */
{
    int srank, drank;                /* Rank of source and destination */
    int type = 0;                    /* Data type */
    long nelem;                      /* Number of elements per source row */
    long sstep, dstep;               /* Stride in source and destination */
    unsigned char *srow;             /* Location of current src row */
    unsigned char *drow1, *drow2;    /* Location of current dest rows */
    struct vxli_reduce_state state;
    int i, i2, flag;

    /* Check validity of source */
    if  ( src == 0 || 
          src->magic != VXL_MAGIC ||
          src->data == 0 ||
          exim_sizeof_intype((type = src->type)) == 0 ||
          vxli_count(src) < 1 )
        NZReturnError(ERR_INIT, "Invalid source array");
    srank  = src->rank;

    /* Not defined for complex arrays */
    if (type == INTP_FCOMPLEX || type == INTP_DCOMPLEX)
        NZReturnError(ERR_ILLOP, "Not defined for complex arrays");

    /* Set up the reduction loop */
    flag = vxli_reduce_setup(src, naxes, axes, &state);
    if (flag)    NZReturnLibError();
    nelem = state.nelem;
    drank = state.drank;
    dstep = state.dstep;
    sstep = state.sstep;

    /* Allocate the destination arrays; set origin and spacing */
    flag = vxl_alloc_array(dmin, type, drank, state.ddimen);
    if (flag)    NZReturnLibError();
    flag = vxl_alloc_array(dmax, type, drank, state.ddimen);
    if (flag)    NZReturnLibError();
    for (i = 0; i < drank; i++) {
        i2 = state.dmap[i];
        dmin->origin [i] = src->origin [i2];
        dmin->spacing[i] = src->spacing[i2];
        dmax->origin [i] = src->origin [i2];
        dmax->spacing[i] = src->spacing[i2];
    }

    /* Loop over destination rows */
    do {

        /* Initialize the destination rows */
	drow1 = vxli_locate(dmin, state.dcoord);
        flag = bips_fill_max(nelem, type, drow1, dstep);
        if (flag)    NZReturnLibError();
	drow2 = vxli_locate(dmax, state.dcoord);
        flag = bips_fill_min(nelem, type, drow2, dstep);
        if (flag)    NZReturnLibError();

        /* Loop over source rows for current destination row */
        do {
            srow = vxli_locate(src, state.scoord);
            flag = bips_min_set(nelem, type, drow1, dstep, srow, sstep);
            if (flag)    NZReturnLibError();
            flag = bips_max_set(nelem, type, drow2, dstep, srow, sstep);
            if (flag)    NZReturnLibError();

        } while ( vxli_reduce_next_srow(&state) );

    } while ( vxli_reduce_next_drow(&state) );

    return ERR_OKAY;
}



/*------------------------------------------------------------------------
vxl_moments - Compute central moments of an image

This function computes the total mass, center of gravity, and (2nd)
central moments of an image.  The moments are computed in physical
coordinates, as defined by the origin and spacing values.  It is the
caller's responsibility to allocate the necessary space for mass (one
double value), center (rank values), and moments (rank*rank values).

FIXME: Add support for complex types, if you can think of a use case.
Note that the center of gravity is in general complex for a complex
image, which seems rather inconsistent with its interpretation as a
coordinate point.

FIXME: What happens if one of the spacing values is zero?  Or if all
of them are zero?

FIXME: It's not particularly useful to compute moments of a rank-0
array, so we just outlaw it for the moment.  (Pun?  What pun?)

FIXME: Always promote to double or double complex for better accuracy?

FIXME: Generalize to return long double or complex as needed.  

FIXME: Consider generalizing to allow reduction on selected axes only.
But that looks hairy enough that I'm going to postpone until I see a
real use case.
------------------------------------------------------------------------*/

int                      /* Returns non-zero on error */
vxl_moments(
    double *mass,        /* Total mass of the image; length 1 */
    double *center,      /* Center of gravity; length rank */
    double *moments,     /* Central moments; length rank*rank */
    voxel_array *src)    /* Image to take moments of */
{
    int rank;                   /* Rank of the image */
    int type;                   /* Type of the image */
    long *dimen;                /* Dimensions of the image */
    long ncols;                 /* Number of elements along last dimension */
    long coord[VXL_MAX_RANK];   /* Run over coordinates of the image */
    double mom[3];              /* Row moments from bips */
    void *row;                  /* Start address of an image row */
    int flag;                   /* Status returned by lib calls */
    int i, j;

    /* Validate the arguments */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->data == 0 || src->rank < 1)
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    rank  = src->rank;
    type  = src->type;
    dimen = src->dimen;
    ncols = dimen[rank-1];
    if (mass == 0 || center == 0 || moments == 0)
	NZReturnError(ERR_ILLOP, "Null destination for moments");

    /* Not yet implemented for complex arrays */
    if (type == INTP_FCOMPLEX || type == INTP_DCOMPLEX)
        NZReturnError(ERR_UNIMP, "Not yet implemented for complex arrays");

    /* Initialize moment arrays: Use center and moments initially to
       store the first and second non-central moments. */
    *mass = 0.0;
    for (i = 0; i < rank; i++) {
	center[i] = 0.0;
	for (j = 0; j < rank; j++) {
	    moments[i*rank+j] = 0.0;
	}
    }

    /* Sum row moments over the rows of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Compute moments for the current row */
	row = vxli_locate(src, coord);
        if (row == 0)    NZReturnLibError();
	flag = bips_moments(ncols, mom, type, row, 1);
	if (flag != 0)    NZReturnLibError();

	/* Accumulate moments over all rows; by symmetry, we need only
	   the lower triangle of the moments array. */
	*mass += mom[0];
	for (i = 0; i < rank-1; i++) {
	    center[i] += coord[i]*mom[0];
	    for (j = 0; j <= i; j++)
		moments[i*rank+j] += coord[i]*coord[j]*mom[0];
	}
	center[rank-1] += mom[1];
	for (j = 0; j < rank-1; j++)
	    moments[(rank-1)*rank + j] += coord[j]*mom[1];
	moments[rank*rank-1] += mom[2];
    }
    while ( vxli_nextrow(rank, dimen, coord) );

    /* Convert to central moments (in lower triangle only) */
    for (i = 0; i < rank; i++) {
	center[i] /= *mass;
	for (j = 0; j <= i; j++)
	    moments[i*rank+j] -= (*mass)*center[i]*center[j];
    }

    /* Translate to physical coordinates and reflect across diagonal */
    for (i = 0; i < rank; i++) {
	center[i] = src->spacing[i] * center[i] + src->origin[i];
	moments[i*rank+i] *= src->spacing[i] * src->spacing[i];
	for (j = 0; j < i; j++) {
	    moments[j*rank+i] = moments[i*rank+j]
		*= src->spacing[i] * src->spacing[j];
	}
    }

    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_mop - Arbitrary moments of the elementwise product of three arrays

This function computes arbitrary moments of an array or the
elementwise product of two or three arrays.  The arrays must be of the
same rank and shape, but the rank must be at least 2; the arrays may
be of any type or mixture of types.  The moments calculation is always
done in double, regardless of the types of the arrays.  The moments to
be computed are given by an K by M int array, where K is the common
rank of the arrays and M is the number of moments to be computed.
Each row specifies one moment to be computed; each element of that row
gives the power to which the corresponding coordinate is to be raised.
The requested moments are returned in a one-dimensional array, in the
same order as the rows defining them.

FIXME: Add support for complex types, if you can think of a use case.

FIXME: Always promote to double or double complex for better accuracy?

FIXME: Consider generalizing to allow reduction on selected axes only.
But that looks hairy enough that I'm going to postpone until I see a
real use case.
---------------------------------------------------------------------------*/

int                              /* Returns non-zero on error */
vxl_mop(
    voxel_array *result,         /* Values of requested moments */
    voxel_array *moments,        /* Table of desired moments */
    voxel_array *array1,         /* First array */
    voxel_array *array2,         /* Second array, or zero if none */
    voxel_array *array3)         /* Third array, or zero */
{
    int rank;                    /* Common rank of the arrays */
    long *dimen;                 /* Common dimension of the arrays */
    long ncols;                  /* Number of elements along last dimension */
    int type1 = 0;               /* Types of the arrays */
    int type2 = 0, type3 = 0;
    long coord[VXL_MAX_RANK];    /* Run over coordinates of the arrays */
    long momcnt;                 /* Number of moments requested */
    int *momdef;                 /* Actual data area for moments definition */
    int maxpow;                  /* Maximum power requested in x */
    double *work1 = 0;           /* Working storage for two source rows */
    double *work2 = 0;
    double *xramp = 0;           /* Row of x coordinate values */
    double *rowmom = 0;          /* X-moments of a row */
    void *row1, *row2, *row3;    /* Start of a row in each source array */
    double *rdata;               /* Data area for result array */
    double mom;                  /* Partially computed moment value */
    long imom;                   /* Index over requested moments */
    long icol;                   /* Index over elements of a row */
    int  icor;                   /* Index over coordinate axes */
    int  ipow;                   /* Index over powers of a coordinate */
    int  flag;

    /* Validate first source array */
    if ( array1 == 0 || array1->magic != VXL_MAGIC ||
         array1->data == 0 || array1->rank < 2 )
	NZReturnError(ERR_INIT, "Array 1 is invalid");
    rank  = array1->rank;
    type1 = array1->type;
    dimen = array1->dimen;
    ncols = dimen[rank-1];

    /* Not yet implemented for complex arrays */
    if (type1 == INTP_FCOMPLEX || type1 == INTP_DCOMPLEX)
        NZReturnError(ERR_UNIMP, "Not yet implemented for complex arrays");

    /* Validate other source arrays, if provided */
    if (array2 != 0) {
	if ( array2->magic != VXL_MAGIC || array2->data == 0 )
	    NZReturnError(ERR_INIT,    "Array 2 is invalid");
        if ( ! vxli_same_physical(array1, array2) )    NZReturnLibError();
        type2 = array2->type;
    }
    if (array3 != 0) {
	if ( array3->magic != VXL_MAGIC || array3->data == 0 )
	    NZReturnError(ERR_INIT,    "Array 3 is invalid");
        if ( !vxli_same_physical(array1, array3) )    NZReturnLibError();
        type3 = array3->type;
    }

    /* Validate the moments definition array */
    if ( moments == 0 || moments->magic != VXL_MAGIC ||
	 moments->rank != 2 || moments->dimen[1] != rank ||
	 moments->type != INTP_INT )
	NZReturnError(ERR_INIT, "Moments definition array is invalid");
    momcnt = moments->dimen[0];
    momdef = moments->data;

    /* Allocate and initialize result array for moments */
    flag = vxl_alloc_array(result, INTP_DOUBLE, 1, &momcnt);
    if (flag)    NZReturnLibError();
    rdata = result->data;
    for (imom = 0; imom < momcnt; imom++)
	rdata[imom] = 0.0;

    /* Determine highest power requested in x; allocate space for x-moments */
    maxpow = 0;
    for (imom = 0; imom < momcnt; imom++)
	if (maxpow < momdef[imom*rank+rank-1])
	    maxpow = momdef[imom*rank+rank-1];
    rowmom = mallock((maxpow+1)*sizeof(double));
    if (rowmom == 0)    NZReturnLibError();

    /* Allocate three rows of working storage for array product */
    work1 = mallock(3*ncols*sizeof(double));
    if (work1 == 0)    NZReturnLibError();
    work2 = &work1[ncols];
    xramp = &work1[2*ncols];

    /* Construct row of x coordinate values */
    /* FIXME: Use BIPS instead? */
    for (icol = 0; icol < ncols; icol++)
        xramp[icol] = icol;

    /* Loop over the rows of the image */
    vxli_firstrow(rank, coord);
    do {

	/* Convert first source array to double */
	row1 = vxli_locate(array1, coord);
        if (row1 == 0)    NZReturnLibError();
	flag = bips_double(ncols, work1, 1, type1, row1, 1);
	if (flag != 0)    NZReturnLibError();

	/* Multiply by second source array, if present */
	if (array2 != 0) {
	    row2 = vxli_locate(array2, coord);
            if (row2 == 0)    NZReturnLibError();
	    flag = bips_double(ncols, work2, 1, type2, row2, 1);
	    if (flag != 0)    NZReturnLibError();
	    flag = bips_mul_set(ncols, INTP_DOUBLE, work1, 1, work2, 1);
	    if (flag != 0)    NZReturnLibError();
	}

	/* Multiply by third source array, if present */
	if (array3 != 0) {
	    row3 = vxli_locate(array3, coord);
            if (row3 == 0)    NZReturnLibError();
	    flag = bips_double(ncols, work2, 1, type3, row3, 1);
	    if (flag != 0)    NZReturnLibError();
	    flag = bips_mul_set(ncols, INTP_DOUBLE, work1, 1, work2, 1);
	    if (flag != 0)    NZReturnLibError();
	}
	    
	/* Loop over powers of x to compute row moments in x */
	for (ipow = 0; ipow <= maxpow; ipow++)
	    rowmom[ipow] = 0.0;
	flag = bips_add_set(ncols, INTP_DOUBLE, &rowmom[0], 0, work1, 1);
	if (flag != 0)    NZReturnLibError();
	for (ipow = 1; ipow <= maxpow; ipow++) {
	    flag = bips_mul_set(ncols, INTP_DOUBLE, work1, 1, xramp, 1);
	    if (flag != 0)    NZReturnLibError();
	    flag = bips_add_set(ncols, INTP_DOUBLE, 
				&rowmom[ipow], 0, work1, 1);
	    if (flag != 0)    NZReturnLibError();
	}

	/* Loop over the requested moments */
	/* FIXME: Find a faster way to compute the powers?  For example,
	   sort the moments into some order and reshuffle before exit.
	   And/or precompute powers of the outermost coordinate. */
	for (imom = 0; imom < momcnt; imom++) {

	    /* Get appropriate X moment for this row */
	    mom = rowmom[momdef[imom*rank+rank-1]];
	    
	    /* Multiply by powers of remaining coordinates */
	    for (icor = 0; icor < rank-1; icor++) {
		for (ipow = 0; ipow < momdef[imom*rank+icor]; ipow++)
		    mom *= (double)coord[icor];
	    }

	    /* Add this row moment to the accumulated array moments */
	    rdata[imom] += mom;

	}
    }
    while ( vxli_nextrow(rank, dimen, coord) );

    /* Free working storage and return */
    free(work1);
    free(rowmom);
    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_norm1 - Compute weighted vector 1-norm of a voxel array

This function computes the weighted vector 1-norm of the given array,
or of the difference of two arrays.  If src2 is a non-null pointer,
the norm of the difference is computed; otherwise, the norm of src1 is
computed.  If weight is a null pointer, then a uniformly weighted norm
is computed; otherwise, weight supplies the weights for each voxel.
The arrays src1, src2, and weight must be the same rank and dimensions.
The arrays src1 and src2 must be the same type; the weight array may
be of any type and will be converted to double for the computation.
The result is always of type double, even if the input is complex.
--------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_norm1(
    voxel_array *dest,          /* Destination array */
    voxel_array *src1,          /* First source array */
    voxel_array *src2,          /* Second source array, or zero */
    voxel_array *weight,        /* Weight array, or zero */
    int naxes,                  /* Number of axes to reduce on */
    int *axes)                  /* The axes to reduce one */
{
    int stype1=0, stype2=0;     /* Type of source arrays */
    int wtype=0;                /* Type of weight array */
    int dtype=INTP_DOUBLE;      /* Type of destination array */
    int drank, srank;           /* Rank of the destination and source arrays */
    double *sbuf=0;             /* Buffer for source diff as double */
    double *wbuf=0;             /* Buffer for weight as double */
    unsigned char *drow=0;      /* Pointer to beginning of each dest row */
    unsigned char *srow=0;      /* Pointer to beginning of each source 1 row */
    unsigned char *trow=0;      /* Pointer to beginning of each source 2 row */
    unsigned char *wrow=0;      /* Pointer to beginning of each weight row */
    long nelem, ndest;          /* Number of elements in a src or dest row */
    long sstep, dstep;          /* Stride in source and destination */
    int flag;                   /* Error indicator */
    struct vxli_reduce_state state;
    long i, i2;

    /* Check validity of source 1 array */
    if  ( src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0 ||
	  exim_sizeof_intype((stype1 = src1->type)) == 0 ) 
        NZReturnError(ERR_INIT, "Invalid source 1 array");
    srank  = src1->rank;

    /* Check validity of source 2 array */
    if ( src2 != 0 ) {
	if  ( src2->magic != VXL_MAGIC || src2->data == 0 ||
	      exim_sizeof_intype((stype2 = src2->type)) == 0 )
            NZReturnError(ERR_INIT, "Invalid source 2 array");
	if ( !vxli_same_physical(src2, src1) )    NZReturnLibError();
        if ( stype1 != stype2 )
            NZReturnError(ERR_INIT, "Source 1 and 2 differ in type");
    }

    /* Check validity of weight array */
    if ( weight != 0 ) {
	if  ( weight->magic != VXL_MAGIC || weight->data == 0 ||
	      exim_sizeof_intype((wtype = weight->type)) == 0 )
            NZReturnError(ERR_INIT, "Invalid weight array");
	if ( !vxli_same_physical(weight, src1) )    NZReturnLibError();
    }

    /* Set up the reduction loop */
    flag = vxli_reduce_setup(src1, naxes, axes, &state);
    if (flag)    NZReturnLibError();
    nelem = state.nelem;
    ndest = state.ndest;
    drank = state.drank;
    dstep = state.dstep;
    sstep = state.sstep;

    /* Allocate the destination array; set origin and spacing */
    flag = vxl_alloc_array(dest, dtype, drank, state.ddimen);
    if (flag)    NZReturnLibError();
    for (i = 0; i < drank; i++) {
        i2 = state.dmap[i];
        dest->origin [i] = src1->origin [i2];
        dest->spacing[i] = src1->spacing[i2];
    }

    /* Allocate working storage for one row */
    sbuf = mallock(2*nelem*sizeof(double));
    if (sbuf == 0)    NZReturnLibError();
    wbuf = sbuf + nelem;

    /* Loop over destination rows */
    do {

        /* Find the destination row and zero it */
	drow = vxli_locate(dest, state.dcoord);
	if (drow == 0)    NZReturnLibError();
        flag = bips_zero(ndest, dtype, drow, dstep);
        if (flag)    NZReturnLibError();

        /* Loop over source rows for current destination row */
        do {

            /* Compute absolute difference */
            srow = vxli_locate(src1, state.scoord);
            if (srow == 0)    NZReturnLibError();
	    if ( src2 ) {
		trow = vxli_locate(src2, state.scoord);
		if (trow == 0)    NZReturnLibError();
            }
            flag = bips_absdiff(nelem, sbuf, 1, stype1, 
                                srow, sstep, trow, sstep);
            if (flag)    NZReturnLibError();

	    /* Multiply by weights, if provided */
	    if (weight) {
		wrow = vxli_locate(weight, state.scoord);
		if (wrow == 0)    NZReturnLibError();
		flag = bips_double(nelem, wbuf, 1, wtype, wrow, sstep);
		if (flag)    NZReturnLibError();
		flag = bips_mul_set(nelem, dtype, sbuf, 1, wbuf, 1);
		if (flag)    NZReturnLibError();
	    }

            /* Add source elements per destination */
            flag = bips_add_set(nelem, dtype, drow, dstep, sbuf, 1);
            if (flag)    NZReturnLibError();

        } while ( vxli_reduce_next_srow(&state) );

    } while ( vxli_reduce_next_drow(&state) );

    /* Free working storage and return */
    free(sbuf);
    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_norm2 - Compute weighted vector 2-norm of a voxel array

This function computes the weighted vector 2-norm of the given array,
or of the difference of two arrays.  If src2 is a non-null pointer,
the norm of the difference is computed; otherwise, the norm of src1 is
computed.  If weight is a null pointer, then a uniformly weighted norm
is computed; otherwise, weight supplies the weights for each voxel.
The arrays src1, src2, and weight must be the same rank and
dimensions.  The source 1 and 2 arrays must be the same type; the
weight array may be any type and will be converted to double for
computation; the output array is always type double, even if the input
is complex.
--------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_norm2(
    voxel_array *dest,          /* Destination array */
    voxel_array *src1,          /* First source array */
    voxel_array *src2,          /* Second source array, or zero */
    voxel_array *weight,        /* Weight array, or zero */
    int naxes,                  /* Number of axes to reduce on */
    int *axes)                  /* The axes to reduce one */
{
    int stype1=0, stype2=0;     /* Type of source arrays */
    int wtype=0;                /* Type of weight array */
    int dtype=INTP_DOUBLE;      /* Type of destination array */
    int drank, srank;           /* Rank of the destination and source arrays */
    double *sbuf=0;             /* Buffer for source diff as double */
    double *wbuf=0;             /* Buffer for weight as double */
    double *zbuf=0;             /* Buffer for sum of weighted squares */
    unsigned char *drow=0;      /* Pointer to beginning of each dest row */
    unsigned char *srow=0;      /* Pointer to beginning of each source 1 row */
    unsigned char *trow=0;      /* Pointer to beginning of each source 2 row */
    unsigned char *wrow=0;      /* Pointer to beginning of each weight row */
    long nelem, ndest;          /* Number of elements in a src or dest row */
    long sstep, dstep;          /* Stride in source and destination */
    int flag;                   /* Error indicator */
    struct vxli_reduce_state state;
    long i, i2;

    /* Check validity of source 1 array */
    if  ( src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0 ||
	  exim_sizeof_intype((stype1 = src1->type)) == 0 ) 
        NZReturnError(ERR_INIT, "Invalid source 1 array");
    srank  = src1->rank;

    /* Check validity of source 2 array */
    if ( src2 != 0 ) {
	if  ( src2->magic != VXL_MAGIC || src2->data == 0 ||
	      exim_sizeof_intype((stype2 = src2->type)) == 0 )
            NZReturnError(ERR_INIT, "Invalid source 2 array");
	if ( !vxli_same_physical(src2, src1) )    NZReturnLibError();
        if ( stype1 != stype2 )
            NZReturnError(ERR_INIT, "Source 1 and 2 differ in type");
    }

    /* Check validity of weight array */
    if ( weight != 0 ) {
	if  ( weight->magic != VXL_MAGIC || weight->data == 0 ||
	      exim_sizeof_intype((wtype = weight->type)) == 0 )
            NZReturnError(ERR_INIT, "Invalid weight array");
	if ( !vxli_same_physical(weight, src1) )    NZReturnLibError();
    }

    /* Set up the reduction loop */
    flag = vxli_reduce_setup(src1, naxes, axes, &state);
    if (flag)    NZReturnLibError();
    nelem = state.nelem;
    ndest = state.ndest;
    drank = state.drank;
    dstep = state.dstep;
    sstep = state.sstep;

    /* Allocate the destination array; set origin and spacing */
    flag = vxl_alloc_array(dest, dtype, drank, state.ddimen);
    if (flag)    NZReturnLibError();
    for (i = 0; i < drank; i++) {
        i2 = state.dmap[i];
        dest->origin [i] = src1->origin [i2];
        dest->spacing[i] = src1->spacing[i2];
    }

    /* Allocate working storage for one row */
    sbuf = mallock(3*nelem*sizeof(double));
    if (sbuf == 0)    NZReturnLibError();
    wbuf = sbuf + nelem;
    zbuf = wbuf + nelem;

    /* Loop over destination rows */
    do {

        /* Find the destination row */
	drow = vxli_locate(dest, state.dcoord);
	if (drow == 0)    NZReturnLibError();

	/* Initialize buffer for summing terms */
        flag = bips_zero(ndest, dtype, zbuf, dstep);
        if (flag)    NZReturnLibError();

        /* Loop over source rows for current destination row */
        do {

            /* Compute squared absolute differences */
            srow = vxli_locate(src1, state.scoord);
            if (srow == 0)    NZReturnLibError();
	    if ( src2 ) {
		trow = vxli_locate(src2, state.scoord);
		if (trow == 0)    NZReturnLibError();
            }
            flag = bips_sqrdiff(nelem, sbuf, 1, stype1, 
                                srow, sstep, trow, sstep);
            if (flag)    NZReturnLibError();

	    /* Multiply by weights, if provided */
	    if (weight) {
		wrow = vxli_locate(weight, state.scoord);
		if (wrow == 0)    NZReturnLibError();
		flag = bips_double(nelem, wbuf, 1, wtype, wrow, sstep);
		if (flag)    NZReturnLibError();
		flag = bips_mul_set(nelem, dtype, sbuf, 1, wbuf, 1);
		if (flag)    NZReturnLibError();
	    }

            /* Add source elements per destination */
            flag = bips_add_set(nelem, dtype, zbuf, dstep, sbuf, 1);
            if (flag)    NZReturnLibError();

        } while ( vxli_reduce_next_srow(&state) );

        /* Take square root and copy to destination */
        flag = bips_sqrt(ndest, dtype, drow, dstep, zbuf, dstep);
        if (flag)   NZReturnLibError();

    } while ( vxli_reduce_next_drow(&state) );

    /* Free working storage and return */
    free(sbuf);
    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_normsup - Compute weighted vector sup-norm of a voxel array

This function computes the weighted vector sup-norm of the given array,
or of the difference of two arrays.  If src2 is a non-null pointer,
the norm of the difference is computed; otherwise, the norm of src1 is
computed.  If weight is a null pointer, then a uniformly weighted norm
is computed; otherwise, weight supplies the weights for each voxel.
The arrays src1, src2, and weight must be the same rank and dimensions.
The arrays src1 and src2 must be the same type; the weight array may
be of any type and will be converted to double for the computation.
The result is always of type double, even if the input is complex.
--------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_normsup(
    voxel_array *dest,          /* Destination array */
    voxel_array *src1,          /* First source array */
    voxel_array *src2,          /* Second source array, or zero */
    voxel_array *weight,        /* Weight array, or zero */
    int naxes,                  /* Number of axes to reduce on */
    int *axes)                  /* The axes to reduce one */
{
    int stype1=0, stype2=0;     /* Type of source arrays */
    int wtype=0;                /* Type of weight array */
    int dtype=INTP_DOUBLE;      /* Type of destination array */
    int drank, srank;           /* Rank of the destination and source arrays */
    double *sbuf=0;             /* Buffer for source diff as double */
    double *wbuf=0;             /* Buffer for weight as double */
    unsigned char *drow=0;      /* Pointer to beginning of each dest row */
    unsigned char *srow=0;      /* Pointer to beginning of each source 1 row */
    unsigned char *trow=0;      /* Pointer to beginning of each source 2 row */
    unsigned char *wrow=0;      /* Pointer to beginning of each weight row */
    long nelem, ndest;          /* Number of elements in a src or dest row */
    long sstep, dstep;          /* Stride in source and destination */
    int flag;                   /* Error indicator */
    struct vxli_reduce_state state;
    long i, i2;

    /* Check validity of source 1 array */
    if  ( src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0 ||
	  exim_sizeof_intype((stype1 = src1->type)) == 0 ) 
        NZReturnError(ERR_INIT, "Invalid source 1 array");
    srank  = src1->rank;

    /* Check validity of source 2 array */
    if ( src2 != 0 ) {
	if  ( src2->magic != VXL_MAGIC || src2->data == 0 ||
	      exim_sizeof_intype((stype2 = src2->type)) == 0 )
            NZReturnError(ERR_INIT, "Invalid source 2 array");
	if ( !vxli_same_physical(src2, src1) )    NZReturnLibError();
        if ( stype1 != stype2 )
            NZReturnError(ERR_INIT, "Source 1 and 2 differ in type");
    }

    /* Check validity of weight array */
    if ( weight != 0 ) {
	if  ( weight->magic != VXL_MAGIC || weight->data == 0 ||
	      exim_sizeof_intype((wtype = weight->type)) == 0 )
            NZReturnError(ERR_INIT, "Invalid weight array");
	if ( !vxli_same_physical(weight, src1) )    NZReturnLibError();
    }

    /* Set up the reduction loop */
    flag = vxli_reduce_setup(src1, naxes, axes, &state);
    if (flag)    NZReturnLibError();
    nelem = state.nelem;
    ndest = state.ndest;
    drank = state.drank;
    dstep = state.dstep;
    sstep = state.sstep;

    /* Allocate the destination array; set origin and spacing */
    flag = vxl_alloc_array(dest, dtype, drank, state.ddimen);
    if (flag)    NZReturnLibError();
    for (i = 0; i < drank; i++) {
        i2 = state.dmap[i];
        dest->origin [i] = src1->origin [i2];
        dest->spacing[i] = src1->spacing[i2];
    }

    /* Allocate working storage for one row */
    sbuf = mallock(2*nelem*sizeof(double));
    if (sbuf == 0)    NZReturnLibError();
    wbuf = sbuf + nelem;

    /* Loop over destination rows */
    do {

        /* Find the destination row and set it to zero */
	drow = vxli_locate(dest, state.dcoord);
	if (drow == 0)    NZReturnLibError();
        flag = bips_zero(ndest, dtype, drow, dstep);
        if (flag)    NZReturnLibError();

        /* Loop over source rows for current destination row */
        do {

            /* Compute absolute difference */
            srow = vxli_locate(src1, state.scoord);
            if (srow == 0)    NZReturnLibError();
	    if ( src2 ) {
		trow = vxli_locate(src2, state.scoord);
		if (trow == 0)    NZReturnLibError();
            }
            flag = bips_absdiff(nelem, sbuf, 1, stype1, 
                                srow, sstep, trow, sstep);
            if (flag)    NZReturnLibError();

	    /* Multiply by weights, if provided */
	    if (weight) {
		wrow = vxli_locate(weight, state.scoord);
		if (wrow == 0)    NZReturnLibError();
		flag = bips_double(nelem, wbuf, 1, wtype, wrow, sstep);
		if (flag)    NZReturnLibError();
		flag = bips_mul_set(nelem, dtype, sbuf, 1, wbuf, 1);
		if (flag)    NZReturnLibError();
	    }

            /* Take max over source and store in destination */
            flag = bips_max_set(nelem, dtype, drow, dstep, sbuf, 1);
            if (flag)    NZReturnLibError();

        } while ( vxli_reduce_next_srow(&state) );

    } while ( vxli_reduce_next_drow(&state) );

    /* Free working storage and return */
    free(sbuf);
    return ERR_OKAY;
}



/*------------------------------------------------------------------------
vxl_stat - Compute mean and standard deviation for an array

This function computes the mean and standard deviation of the elements
of an array, with optional element weights.  The source and weight
arrays may be of any type or shape, but must be the same shape.

NOTE: There is a little trickery with types here.  The variable cc0
may be passed to bips as either double or double complex, but the
imaginary part is always zero.  Thus we can pass a void pointer
and the right thing will happen.

FIXME: Do the sums rowwise, to reduce the rounding error.  There are
rumors (in itk?) of a numerically stable one-pass algorithm;
investigate and use if true.
-------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_stat(
    voxel_array *mean,          /* Array of mean values */
    voxel_array *sdev,          /* Array of sdev values */
    voxel_array *src1,          /* First array */
    voxel_array *weight,        /* Voxel weights, or zero */
    int naxes,                  /* Number of axes to reduce on */
    int *axes)                  /* The axes to reduce one */
{
    int stype1=0, wtype=0;      /* Type of source and weight arrays */
    int ctype=0, cbyte=0;       /* Type and bytes/elem of complex sum type */
    int rtype=0, rbyte=0;       /* Ditto for corresponding real type */
    int drank, srank;           /* Rank of the destination and source arrays */
    unsigned char *sbuf=0;      /* (Complex) buffer for source */
    unsigned char *sumx=0;      /* (Complex) buffer to sum weighted values */
    unsigned char *qbuf=0;      /* (Real) buffer for squared absolute values */
    unsigned char *wbuf=0;      /* (Real) buffer for weight */
    unsigned char *sumw=0;      /* (Real) buffer for sum of weights */
    unsigned char *sumww=0;     /* (Real) buffer for sum of squared weights */
    unsigned char *sumxx=0;     /* (Real) buffer for sum of weighted squares */
    unsigned char *mrow=0;      /* Pointer to beginning of each mean row */
    unsigned char *drow=0;      /* Pointer to beginning of each sdev row */
    unsigned char *srow=0;      /* Pointer to beginning of each source row */
    unsigned char *wrow=0;      /* Pointer to beginning of each weight row */
    long nelem, ndest;          /* Number of elements in a src or dest row */
    long sstep, dstep;          /* Stride in source and destination */
    int flag;                   /* Error indicator */
    struct vxli_reduce_state state;
#ifdef HAVE_COMPLEX_H
    double complex cc0;
#else
    double cc0;
#endif
    double cc1, v;
    long i, i2;

    /* Check validity of source */
    if  (  src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0 ||
           exim_sizeof_intype((stype1 = src1->type)) == 0 ) 
        NZReturnError(ERR_INIT, "Invalid source 1 array");
    srank  = src1->rank;

    /* Check validity of weight array */
    if ( weight != 0 ) {
	if  (  weight->magic != VXL_MAGIC || weight->data == 0 ||
               exim_sizeof_intype((wtype = weight->type)) == 0 )
            NZReturnError(ERR_INIT, "Invalid weight array");
	if ( !vxli_same_physical(weight, src1) )    NZReturnLibError();
        if (wtype == INTP_FCOMPLEX || wtype == INTP_DCOMPLEX)
            NZReturnError(ERR_INCOMP, "Weights may not be complex");
    }

    /* Determine real and complex destination types */
    if ( stype1 == INTP_FCOMPLEX || stype1 == INTP_DCOMPLEX ||
         wtype  == INTP_FCOMPLEX || wtype  == INTP_DCOMPLEX ) {
        ctype = INTP_DCOMPLEX;
        rtype = INTP_DOUBLE;
    }
    else {
        ctype = INTP_DOUBLE;
        rtype = INTP_DOUBLE;
    }
    cbyte = exim_sizeof_intype(ctype);
    rbyte = exim_sizeof_intype(rtype);

    /* Set up the reduction loop */
    flag = vxli_reduce_setup(src1, naxes, axes, &state);
    if (flag)    NZReturnLibError();
    nelem = state.nelem;
    ndest = state.ndest;
    drank = state.drank;
    dstep = state.dstep;
    sstep = state.sstep;

    /* Allocate the destination arrays; set origin and spacing */
    flag = vxl_alloc_array(mean, ctype, drank, state.ddimen);
    if (flag)    NZReturnLibError();
    flag = vxl_alloc_array(sdev, rtype, drank, state.ddimen);
    if (flag)    NZReturnLibError();
    for (i = 0; i < drank; i++) {
        i2 = state.dmap[i];
        mean->origin [i] = src1->origin [i2];
        mean->spacing[i] = src1->spacing[i2];
        sdev->origin [i] = src1->origin [i2];
        sdev->spacing[i] = src1->spacing[i2];
    }

    /* Allocate working storage for one row.  The sbuf and sumx buffers may
       need to hold complex values; the others hold only real values. */
    sbuf = mallock(2*nelem*cbyte + 5*nelem*rbyte);
    if (sbuf == 0)    NZReturnLibError();
    sumx  = sbuf  + nelem * cbyte;
    qbuf  = sumx  + nelem * cbyte;
    wbuf  = qbuf  + nelem * rbyte;
    sumw  = wbuf  + nelem * rbyte;
    sumww = sumw  + nelem * rbyte;
    sumxx = sumww + nelem * rbyte;

    /* Precompute some constant coefficients */
    if ( weight ) {
        cc0 = 1.0;
    }
    else {
        v = (double)state.sperd;    /* N = # elements */
        cc0 = 1.0/v;                /* cc0 = 1/N        */
        cc1 = v/(v-1.0);            /* cc1 = N/(N-1)  */
    }

    /* Loop over destination rows */
    do {

        /* Initialize buffers for destination row */
        flag = bips_zero(ndest, ctype, sumx, dstep);
        if (flag)    NZReturnLibError();
        flag = bips_zero(ndest, rtype, sumxx, dstep);
        if (flag)    NZReturnLibError();
        if (weight) {
            flag = bips_zero(nelem, rtype, sumw, dstep);
            if (flag)    NZReturnLibError();
            flag = bips_zero(nelem, rtype, sumww, dstep);
            if (flag)    NZReturnLibError();
        }

        /* Loop over source rows for current destination row */
        do {

            /* Convert this row of src1 to type double */
            srow = vxli_locate(src1, state.scoord);
            if (srow == 0)    NZReturnLibError();
            flag = bips_cast(nelem, ctype, sbuf, 1, stype1, srow, sstep);
            if (flag)    NZReturnLibError();

            /* Accumulate sums when weights are provided */
            if (weight) {
                
                /* Convert weights to double */
                wrow = vxli_locate(weight, state.scoord);
                if (wrow == 0)    NZReturnLibError();
                flag = bips_cast(nelem, rtype, wbuf, 1, wtype, wrow, sstep);
                if (flag)    NZReturnLibError();

                /* Accumulate weights and squared weights */
                flag = bips_add_set(nelem, rtype, sumw, dstep, wbuf, 1);
                if (flag)    NZReturnLibError();
                flag = bips_fma_set(nelem, rtype, sumww, dstep, 
                                    rtype, wbuf, 1, wbuf, 1);
                if (flag)    NZReturnLibError();

                /* Accumulate weighted source and squared abs source */
                flag = bips_fma_set(nelem, ctype, sumx, dstep, rtype, wbuf, 1,
                                    sbuf, 1);
                if (flag)    NZReturnLibError();
                flag = bips_sqrdiff(nelem, (double*)qbuf, 1, 
                                    ctype, sbuf, 1, 0, 1);
                if (flag)    NZReturnLibError();
                flag = bips_fma_set(nelem, rtype, sumxx, dstep,
                                    rtype, qbuf, 1, wbuf, 1);
                if (flag)    NZReturnLibError();
            }

            /* Accumulate sums when weights are not provided */
            else {
                flag = bips_add_set(nelem, ctype, sumx, dstep, sbuf, 1);
                if (flag)    NZReturnLibError();
                flag = bips_sqrdiff(nelem, (double*)qbuf, 1, ctype, sbuf, 1, 
                                    0, 1);
                if (flag)    NZReturnLibError();
                flag = bips_add_set(nelem, rtype, sumxx, dstep, qbuf, 1);
                if (flag)    NZReturnLibError();
            }

        } while ( vxli_reduce_next_srow(&state) );

        /* Locate the destination rows */
	mrow = vxli_locate(mean, state.dcoord);
	drow = vxli_locate(sdev, state.dcoord);

        /* Compute weighted mean and standard deviation */
        /* FIXME: Possible division by zero here */
        if (weight) {
            
            /* mrow[i] = sumx[i] / sumw[i] */
            flag = bips_div(ndest, ctype, mrow, dstep, sumx, 1, sumw, 1);
            if (flag)    NZReturnLibError();

            /* sbuf[i] = 1.0 - sumww[i] / (sumw[i] * sumw[i]) */
            flag = bips_mul(ndest, rtype, sbuf, 1, sumw, 1, sumw, 1);
            if (flag)    NZReturnLibError();
            flag = bips_div_set(ndest, rtype, sumww, 1, sbuf, 1);
            if (flag)    NZReturnLibError();
            flag = bips_sub(ndest, rtype, sbuf, 1, &cc0, 0, sumww, 1);
            if (flag)    NZReturnLibError();

            /* sumxx[i] = (sumxx[i] / sumw[i] - |mrow[i]|**2) / sbuf[i] */
            flag = bips_div_set(ndest, rtype, sumxx, 1, sumw, 1);
            if (flag)    NZReturnLibError();
            flag = bips_sqrdiff(ndest, (double*)wbuf, 1, ctype, mrow, dstep, 
                                0, 0);
            if (flag)    NZReturnLibError();
            flag = bips_sub_set(ndest, rtype, sumxx, 1, wbuf, 1);
            if (flag)    NZReturnLibError();
            flag = bips_div_set(ndest, rtype, sumxx, 1, sbuf, 1);
            if (flag)    NZReturnLibError();

            /* drow[i] = sqrt(sumxx[i]) */
            flag = bips_sqrt(ndest, rtype, drow, dstep, sumxx, 1);
            if (flag)    NZReturnLibError();

        }

        /* Compute unweighted mean and standard deviation */
        /* FIXME: Possible division by zero here */
        else {

            /* mrow[i] = sumx[i] / N */
            flag = bips_mul(ndest, ctype, mrow, dstep, sumx, 1, &cc0, 0);
            if (flag)    NZReturnLibError();

            /* sumxx[i] = (sumxx[i] / N - |mrow[i]|**2) * (N/(N-1)) */
            flag = bips_sqrdiff(ndest, (double*)wbuf, 1, ctype, mrow, dstep, 
                                0, 0);
            if (flag)    NZReturnLibError();
            flag = bips_mul_set(ndest, rtype, sumxx, 1, &cc0, 0);
            if (flag)    NZReturnLibError();
            flag = bips_sub_set(ndest, rtype, sumxx, 1, wbuf, 1);
            if (flag)    NZReturnLibError();
            flag = bips_mul_set(ndest, rtype, sumxx, 1, &cc1, 0);
            if (flag)    NZReturnLibError();

            /* drow[i] = sqrt(sumxx[i]) */
            flag = bips_sqrt(ndest, rtype, drow, dstep, sumxx, 1);
            if (flag)    NZReturnLibError();
        }

    } while ( vxli_reduce_next_drow(&state) );

    /* Free working storage and return */
    free(sbuf);
    return ERR_OKAY;
}
