/****************************************************************************
resample.c - Voxel Kit routines for image resampling

Author:  Paul Hughett

Copyright (c) 1998-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  This particular file contains code for image resampling
and related functions.

FIXME: Add code for cubic spline interpolation.  Consider having this
as the default cubic interpolator, with Lagrange interpolation as an
alternative with a different name.

FIXME: Redesign the fast algorithms to allow independent specification of
the transformer, clipper, and interpolator.  Make sure that the new design
also allows noisy interpolation, as in my current (2005-04) ideas for
avoiding interpolation artifacts in image registration.
*****************************************************************************/

static const char rcsid[] = "$Id: resample.c,v 1.35 2005/11/29 15:28:44 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include <float.h>
#include <stdlib.h>

#ifdef HAVE_TGMATH_H
#include <tgmath.h>
#else
#include <math.h>
#endif

#include "bips.h"
#include "errm.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"




/*==========================================================================
Internal utility functions
===========================================================================*/



/*--------------------------------------------------------------------------
BoundaryFuzz - Fudge factor for roundoff error at boundaries

The back-projected source coordinates are computed in several
different places below, using different algorithms; roundoff error can
cause the computed coordinates to be slightly different.  This has
several implications.  First, the reference and fast implementations
can yield different results; this is especially true for
nearest-neighbor interpolation, where a coordinate approximating n +
0.5 can become either n or n+1, yielding a completely different source
value.  Second, a source coordinate near the boundary can be seen as
inside the boundary when the limits of the umbra and penumbra are computed,
but as outside when the source coordinates are computed for interpolation;
this can produce a segmentation fault when the interpolator attempts to
fetch a voxel that isn't there.

The BoundaryFuzz is intended to fix the latter problem.  In the
computation of the umbra and penumbra, a point is counted as inside
only if it is at least BoundaryFuzz units inside the boundary; the
value of BoundaryFuzz is taken as 1000.0 * DBL_EPSILON, which should
be enough to cover roundoff error in coordinate calculations for
coordinate values up to several hundred.
----------------------------------------------------------------------------*/

const double BoundaryFuzz = 1000.0 * DBL_EPSILON;



/*-----------------------------------------------------------------------
vxli_{first,next}_neighbor - Iteration over n-dimensional cube

These functions are used visit all the points of an n-dimensional cube
with increments [0, 1, ..., range-1] along each axis.  The paradigm is
as follows:

    vxli_first_neighbor(rank, delta);
    do {
        ...process one neighbor...
    } while (vxli_next_neighbor(rank, delta, range);


FIXME: Move to vxli2.c and change to inlineable?

FIXME: Write a symmetric version of this?  I.e. the increments are
[-range+1, ..., range-1] instead.
------------------------------------------------------------------------*/

static inline void
vxli_first_neighbor(
    int rank,                /* Number of dimensions */
    long *delta)             /* Relative coordinates to a neighbor */
{
    int i;
    for (i = 0; i < rank; i++)
        delta[i] = 0;
    return;
}


static inline int            /* Returns nonzero if more neighbors */
vxli_next_neighbor(
    int rank,                /* Number of dimensions */
    long *delta,             /* Relative coordinates to a neighbor */
    long range)              /* Number of elements along each side */
{
    int i;

    for (i = rank-1; i >= 0; i--) {
	if ( ++delta[i] < range )
	    return 1;
	else
	    delta[i] = 0; }
    return 0;
}




/*=======================================================================
Generic interpolation functions

These functions implement various interpolation methods via a shared
interface so that different interpolation and transformation methods
can be freely intermixed.

FIXME: Add similar functions for cubic interpolation and perhaps a
truncated sinc interpolation.
=======================================================================*/



/*------------------------------------------------------------------------
vxli_nearest_point - Nearest neighbor interpolation in a voxel array

This internal function evaluates the value of a voxel array at a given
position, by taking the value of the nearest voxel; voxels outside the
array are assumed to be zero.  This version checks the given point
against the source boundaries but assumes that the source has already
been checked for validity and that the step sizes have already been
computed; the function vxl_nearest_point() is recommended for external
use.

FIXME: It is arguable that this function should round when converting
from a float type to an integral type, but bips_round_cast() does not
support all combinations of types.

FIXME: I don't believe bips_round_cast supports casts from complex types
direct to integral types.
---------------------------------------------------------------------------*/

static int                          /* Returns nonzero on error */
vxli_nearest_point(
    int dtype,                      /* Type of destination */
    void *dest,                     /* Destination */
    voxel_array *src,               /* Voxel array to interpolate in */
    double *point,                  /* Point to interpolate at */
    long   *step)                   /* Precomputed step size for each axis */
{
    int type;                       /* Type of the array */
    int rank;                       /* Rank of the array */
    long   *dimen;                  /* Dimensions of the array */
    void   *data;                   /* Data contained in the array */
    long   index[VXL_MAX_RANK];     /* Index of nearest neighbor */
    int i, flag;
    void *datum;

    /* Get some parameters from the source */
    rank  = src->rank;
    dimen = src->dimen;
    type  = src->type;
    data  = src->data;

    /* Compute index of nearest neighbor on each axis, and check for clean
       miss.  NOTE: floor() is very slow on i386, so we play games with
       casting instead. */
    for (i = 0; i < rank; i++) {
        index[i] = (long)(point[i]+1.5) - 1;
        if ( index[i] < 0 || index[i] >= dimen[i] ) {
            flag = bips_zero(1, dtype, dest, 0);
            if (flag)   NZReturnLibError();
            else        return 0;
        }
    }

    /* Copy source voxel to destination, casting as needed */
    datum = vxli_locate(src, index);
    if (datum == 0)
        NZReturnError(ERR_ASSERT, "Impossible error 1 in vxli_nearest_point");
    flag = bips_cast(1, dtype, dest, 0, type, datum, 0);
    if (flag)    NZReturnLibError();

    return 0;
}



/*------------------------------------------------------------------------
vxl_nearest_point - Nearest neighbor interpolation in a voxel array

This function evaluates the value of a voxel array at a given
position, taking the value at the nearest neighbor; voxels outside the
array are assumed to be zero.  This version fully checks its arguments
and is suitable for external use; see vxli_nearest_point() for a faster
version with fewer checks.

This function replaces the old vxl_nearest, but is more general and has
a different calling sequence.
---------------------------------------------------------------------------*/

int                             /* Returns nonzero on error */
vxl_nearest_point(
    int dtype,                  /* Type of destination */
    void *dest,                 /* Destination */
    voxel_array *src,           /* Voxel array to interpolate in */
    double *point)              /* Point to interpolate at */
{
    int type;                   /* Type of the source array */
    int srank;                  /* Rank of the source array */
    long *sdimen;               /* Dimensions of the source array */
    long sstep[VXL_MAX_RANK];   /* Step sizes in source array */
    int i, flag;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    srank  = src->rank;
    sdimen = src->dimen;
    type   = src->type;

    /* Precompute step size along each source coordinate */
    sstep[srank-1] = 1;
    for (i = srank-1; i > 0; i--)
	sstep[i-1] = sstep[i] * sdimen[i];

    /* Call the internal routine to do the actual work */
    flag = vxli_nearest_point(dtype, dest, src, point, sstep);
    
    return flag;
}



/*-----------------------------------------------------------------------
vxl_nearest_map - Nearest neighbor interpolation at points given by a coord map

This function creates a new voxel array whose elements are obtained by
nearest neighbor interpolation in a source array at the coordinate
points given by a coordinate map array.  The source array may be of
any type and shape.  The map array must be of type double and its last
dimension must equal the rank of the source array.  The type of the
result is specified by the caller and defaults to the type of the
source array; the shape is the same as the coordinate map less the
last dimension.  The interpolation calculations are done in double or
double complex.

FIXME: Allow the coordinate array to be type float?
------------------------------------------------------------------------*/

int                             /* Returns nonzero on error */
vxl_nearest_map(
    voxel_array *dest,          /* Destination array */
    int dtype,                  /* Desired destination type, or zero */
    voxel_array *src,           /* Source array */
    voxel_array *map)           /* Coordinate map array */
{
    int srank;                  /* Rank of the source */
    long *sdimen=0;             /* Dimensions of source array */
    long sstep[VXL_MAX_RANK];   /* Step sizes in source array */
    int stype;                  /* Type of the source array */
    int mrank;                  /* Rank of map array */
    int drank;                  /* Rank of destination array */
    long dimen[VXL_MAX_RANK];   /* Dimensions of map and dest arrays */
    long coord[VXL_MAX_RANK];   /* Current row in the map array */
    void *datum=0;              /* Current voxel in the destination */
    void *point=0;              /* Current sample point in the source array */
    int i, flag;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    srank  = src->rank;
    sdimen = src->dimen;
    stype  = src->type;

    /* Destination type defaults to the source type */
    if (dtype == 0)
        dtype = stype;

    /* Precompute step size along each source coordinate */
    sstep[srank-1] = 1;
    for (i = srank-1; i > 0; i--)
	sstep[i-1] = sstep[i] * sdimen[i];

    /* Validate the coordinate array; extract rank and dimensions */
    if ( map == 0 || map->magic != VXL_MAGIC ||
         map->rank < 1 || map->data == 0 )
        NZReturnError(ERR_INIT, "Invalid or empty map array");
    if ( map->type != INTP_DOUBLE )
        NZReturnError(ERR_INIT, "Coord map type must be double");
    mrank = map->rank;
    for (i = 0; i < mrank; i++)
        dimen[i] = map->dimen[i];

    /* Add final dimension of length one to the map if the source is rank
       one and the current final dimension is not one. */
    if (srank == 1 && dimen[mrank-1] != 1)
        dimen[mrank++] = 1;
    drank = mrank - 1;

    /* Verify that the source and map are compatible */
    if ( dimen[mrank-1] != srank )
        NZReturnError(ERR_INIT, 
                      "Last coord map dimen must match source rank");    

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    flag = vxl_alloc_array(dest, dtype, drank, dimen);
    if (flag)    NZReturnLibError();

    /* Loop over the rows of the coordinate array, each of which defines 
       one sample point in the source array and thus a single voxel of the 
       destination. */
    vxli_firstrow(mrank, coord);
    do {

        /* Locate the sample point in the map array and the corresponding
           destination element. */
        point = vxli_locate(map,  coord);
        datum = vxli_locate(dest, coord);
        if (point == 0 || datum == 0)
            NZReturnError(ERR_ASSERT, "Internal error 1 in vxl_nearest_map");
        
        /* Interpolate in source array and save result */
        flag = vxli_nearest_point(dtype, datum, src, point, sstep);
        if (flag)    NZReturnLibError();            

    } while (vxli_nextrow(mrank, dimen, coord));

    return 0;
}



/*------------------------------------------------------------------------
vxli_linear_point - Linear interpolation at a point in a voxel array

This internal function evaluates the value of a voxel array at a given
position, interpolating linearly between sample values; voxels outside
the array are assumed to be zero.  This version checks the given point
against the source boundaries but assumes that the source has already
been checked for validity and that the step sizes have already been
computed; the function vxl_linear_point() is recommended for external
use.

FIXME: Could gather be used to collect all the neighboring voxel values?
Or is there some other fast way to do this?
---------------------------------------------------------------------------*/

static int                          /* Returns nonzero on error */
vxli_linear_point(
    int dtype,                      /* Type of destination */
    void *dest,                     /* Destination */
    voxel_array *src,               /* Voxel array to interpolate in */
    double *point,                  /* Point to interpolate at */
    long   *step)                   /* Precomputed step size for each axis */
{
    int type;                       /* Type of the array */
    int rank;                       /* Rank of the array */
    long   *dimen;                  /* Dimensions of the array */
    void   *data;                   /* Data contained in the array */
    long   index[VXL_MAX_RANK];     /* Index of lowest corner */
    double alpha[VXL_MAX_RANK][2];  /* Fraction part of coordinate */
    long   delta[VXL_MAX_RANK];     /* Index increment to each neighbor */
    long   offset;                  /* Corner location in vectorized array */
    int    valid;                   /* Is corner inside array? */
    double value = 0.0;             /* Value at the given position */
    double v1;                      /* Voxel value at a neighbor */
#ifdef HAVE_COMPLEX_H
    double complex cvalue = 0.0;    /* Value at given position */
    double complex cv1;             /* Voxel value at a neighbor */
#endif
    long coord;                     /* Coordinate value */
    int i, flag=0;
    void *datum;

    /* Get some parameters from the source */
    rank  = src->rank;
    dimen = src->dimen;
    type  = src->type;
    data  = src->data;

    /* Compute index and fraction on each axis, and check for clean miss. 
       The interval (-1,0) is handled specially because floor() is slow and
       casting to long truncates toward zero. */
    for (i = 0; i < rank; i++) {
        if (point[i] >= 0.0 && point[i] < dimen[i]) {
            index[i] = (long)point[i];
            alpha[i][1] = point[i] - (double)index[i];
            alpha[i][0] = 1.0 - alpha[i][1];
        }
        else if (point[i] <= -1.0 || point[i] >= dimen[i]) {
            flag = bips_zero(1, dtype, dest, 0);
            if (flag)   NZReturnLibError();
            else        return 0;
        }
        else {
            alpha[i][1] =  point[i] + 1.0;
            alpha[i][0] = -point[i];
            index[i] = -1;
        }
    }

    /* Loop over the neighbors to compute weighted sum */
    vxli_first_neighbor(rank, delta);
    do {

	/* Compute coordinates of current neighbor; check for boundaries */
	valid  = 1;     /* Assume inside source until proven otherwise */
        offset = 0;
	for (i = rank-1; i >= 0; i--) {
	    coord   = index[i] + delta[i];
            offset += coord * step[i];
	    if ( coord < 0 || coord >= dimen[i] ) {
		valid = 0;
		break;
	    }
	}
	if ( !valid )   continue;

        /* Locate the value for that neighbor */
        datum = bips_locate(type, data, offset);
        if (datum == 0)
                NZReturnError(ERR_ASSERT,
                              "Impossible error 1 in vxli_linear_point");

        /* Case 1: Interpolate in plain double */
        if (type != INTP_FCOMPLEX && type != INTP_DCOMPLEX) {

            /* Fetch voxel value at the neighbor */
            flag  = bips_double(1, &v1, 0, type, datum, 0);
            if (flag)
                NZReturnError(ERR_ASSERT,
                              "Impossible error 3 in vxli_linear_point");
            
            /* Multiply by fraction along each dimension and add to value */
            if ( v1 != 0.0 ) {
                for (i = 0; i < rank; i++)
                    v1 *= alpha[i][delta[i]];
                value += v1;
            }
        }

        /* Case 2: Interpolate in double complex */
#ifdef HAVE_COMPLEX_H
        else {

            /* Fetch voxel value at the neighbor */
            flag = bips_dcomplex(1, &cv1, 0, type, datum, 0);
            if (flag)    
                NZReturnError(ERR_ASSERT,
                              "Impossible error 2 in vxli_linear_point");
            
            /* Multiply by fraction along each dimension and add to value */
            if ( cv1 != 0.0 ) {
                for (i = 0; i < rank; i++)
                    cv1 *= alpha[i][delta[i]];
                cvalue += cv1;
            }
        }
#endif

    } while (vxli_next_neighbor(rank, delta, 2));

    /* Store the final result */
    if (type != INTP_FCOMPLEX && type != INTP_DCOMPLEX)
        flag = bips_round_cast(1, dtype, dest, 0, INTP_DOUBLE, &value, 0);
#ifdef HAVE_COMPLEX_H
    else
        flag = bips_cast(1, dtype, dest, 0, INTP_DCOMPLEX, &cvalue, 0);
#endif
    if (flag)    NZReturnLibError();

    return 0;
}



/*------------------------------------------------------------------------
vxl_linear_point - Linear interpolation at a point in a voxel array

This function evaluates the value of a voxel array at a given
position, interpolating linearly between sample values; voxels outside
the array are assumed to be zero.  This version fully checks its arguments
and is suitable for external use; see vxli_linear_point() for a faster
version with fewer checks.

This function replaces the old vxl_linear, but is more general and has
a different calling sequence.
---------------------------------------------------------------------------*/

int                             /* Returns nonzero on error */
vxl_linear_point(
    int dtype,                  /* Type of destination */
    void *dest,                 /* Destination */
    voxel_array *src,           /* Voxel array to interpolate in */
    double *point)              /* Point to interpolate at */
{
    int type;                   /* Type of the source array */
    int srank;                  /* Rank of the source array */
    long *sdimen;               /* Dimensions of the source array */
    long sstep[VXL_MAX_RANK];   /* Step sizes in source array */
    int i, flag;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    srank  = src->rank;
    sdimen = src->dimen;
    type   = src->type;

    /* Precompute step size along each source coordinate */
    sstep[srank-1] = 1;
    for (i = srank-1; i > 0; i--)
	sstep[i-1] = sstep[i] * sdimen[i];

    /* Call the internal routine to do the actual work */
    flag = vxli_linear_point(dtype, dest, src, point, sstep);
    
    return flag;
}



/*----------------------------------------------------------------------
vxli_linear_row - Linear interpolation along a scanline

NOTE: This function is not yet implemented, and is only a preliminary
attempt at rewriting the resampling code to be both fast and flexible
in the sense that any combination of transformation and interpolation
is possible.

This function does linear interpolation for each destination voxel
along a scanline.  This function is intended to provide high
performance without unnecessarily restricting the permissible types of
coordinate transforms.  The caller is thus expected to verify that the
source and destination are valid, to compute the source coordinates,
and to classify each block of destination voxels as to whether they
completely miss the source, overlap the boundary of the source, or lie
completely within the source.

The 'umbrae' and 'numbrae' arguments specify for each continuous block
of destination voxels whether the neighborhood of each voxel of that
block completely misses the source, overlaps the boundary of the
source (i.e. contains voxels both inside and outside the source), or
lies completely within the source.  The coding of 'umbrae' is as
follows:

      0 => end of list
      1 => clean miss: neighborhood completely outside the source
      2 => mixed: points both inside and outside
      3 => completely inside the source

Cases 1 and 3 are computed more efficiently in this function than case
2; but case 2 may be used by a lazy transform calculating function to
delegate bounday checking to this function.

The first block includes destination voxels 0 up to but not including
numbrae[0]; and the nth block includes voxels numbrae[n-1] up to
numbrae[n].  The end of the list is marked by umbrae[n] == 0.

The 'scoord' argument points to an nelem by srank array which contains
the source coordinates corresponding to each destination voxel.  The
actual source coordinates will be ignored for destination voxels that
completely miss the source, and thus need not be computed by the
caller.

NOTE that this interface is shared with several other interpolation
functions so that interpolation algorithms and transformations can be
freely intermixed.  In short, if you change this interface, you must
be careful to change ALL the uses.

FIXME: Use restrict keyword?  

FIXME: The source step sizes should be computed only once per array.
Perhaps better, they should be precomputed once and stored in the
voxel_array struct.

FIXME: Option to add rather than clear and add?  This would support
overlapped processing of huge arrays.
-----------------------------------------------------------------------*/

#if 0    /* Not yet implemented */
static int                 /* Returns nonzero on error */
vxli_linear_row(
    int nelem,             /* Number of elements in destination row */
    int dtype,             /* Type of destination */
    void *dest,            /* First element of destination */
    long ds,               /* Stride in destination */
    voxel_array *src,      /* Source array */
    int *umbrae,           /* Overlap with umbra */
    long numbrae,          /* Dest pixel indices for umbrae */
    double *scoord)        /* Coordinates of source points */
{
    int stype;                  /* Type of source */
    void *sdata;                /* First element of source */
    int srank;                  /* Rank of source */
    long *sdimen;               /* Dimensions of source array */
    long sstep[VXL_MAX_RANK];   /* Step size along each source dimension */

    /* FIXME: Actually implement this! */

    return 0;
}
#endif



/*-----------------------------------------------------------------------
vxl_linear_map - Linear interpolation at points defined by a coord map

This function creates a new voxel array whose elements are obtained by
interpolating in a source array at the coordinate points given by a
coordinate map array.  The source array may be of any type and shape.
The map array must be of type double and its last dimension must equal
the rank of the source array.  The type of the result is specified by
the caller and defaults to the type of the source array; the shape is
the same as the coordinate map less the last dimension.  The
interpolation calculations are done in double or double complex.

FIXME: Allow the coordinate array to be type float?
------------------------------------------------------------------------*/

int                             /* Returns nonzero on error */
vxl_linear_map(
    voxel_array *dest,          /* Destination array */
    int dtype,                  /* Desired destination type, or zero */
    voxel_array *src,           /* Source array */
    voxel_array *map)           /* Coordinate map array */
{
    int srank;                  /* Rank of the source */
    long *sdimen=0;             /* Dimensions of source array */
    long sstep[VXL_MAX_RANK];   /* Step sizes in source array */
    int stype;                  /* Type of the source array */
    int mrank;                  /* Rank of map array */
    int drank;                  /* Rank of destination array */
    long dimen[VXL_MAX_RANK];   /* Dimensions of map and dest arrays */
    long coord[VXL_MAX_RANK];   /* Current row in the map array */
    void *datum=0;              /* Current voxel in the destination */
    void *point=0;              /* Current sample point in the source array */
    int i, flag;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    srank  = src->rank;
    sdimen = src->dimen;
    stype  = src->type;

    /* Destination type defaults to the source type */
    if (dtype == 0)
        dtype = stype;

    /* Precompute step size along each source coordinate */
    sstep[srank-1] = 1;
    for (i = srank-1; i > 0; i--)
	sstep[i-1] = sstep[i] * sdimen[i];

    /* Validate the coordinate array; extract rank and dimensions */
    if ( map == 0 || map->magic != VXL_MAGIC ||
         map->rank < 1 || map->data == 0 )
        NZReturnError(ERR_INIT, "Invalid or empty map array");
    if ( map->type != INTP_DOUBLE )
        NZReturnError(ERR_INIT, "Coord map type must be double");
    mrank = map->rank;
    for (i = 0; i < mrank; i++)
        dimen[i] = map->dimen[i];

    /* Add final dimension of length one to the map if the source is rank
       one and the current final dimension is not one. */
    if (srank == 1 && dimen[mrank-1] != 1)
        dimen[mrank++] = 1;
    drank = mrank - 1;

    /* Verify that the source and map are compatible */
    if ( dimen[mrank-1] != srank )
        NZReturnError(ERR_INIT, 
                      "Last coord map dimen must match source rank");    

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    flag = vxl_alloc_array(dest, dtype, drank, dimen);
    if (flag)    NZReturnLibError();

    /* Loop over the rows of the coordinate array, each of which defines 
       one sample point in the source array and thus a single voxel of the 
       destination. */
    vxli_firstrow(mrank, coord);
    do {

        /* Locate the sample point in the map array and the corresponding
           destination element. */
        point = vxli_locate(map,  coord);
        datum = vxli_locate(dest, coord);
        if (point == 0 || datum == 0)
            NZReturnError(ERR_ASSERT, "Internal error 1 in vxl_linear_map");
        
        /* Interpolate in source array and save result */
        flag = vxli_linear_point(dtype, datum, src, point, sstep);
        if (flag)    NZReturnLibError();            

    } while (vxli_nextrow(mrank, dimen, coord));

    return 0;
}



/*------------------------------------------------------------------------
vxli_cubic_point - Cubic interpolation at a point in a voxel array

This internal function evaluates the value of a voxel array at a given
position, using cubic Lagrange interpolation on the four nearest
sample values; voxels outside the array are assumed to be zero.  This
version checks the given point against the source boundaries but
assumes that the source has already been checked for validity and that
the step sizes have already been computed; the function
vxl_cubic_point() is recommended for external use.

FIXME: Could gather be used to collect all the neighboring voxel values?
Or is there some other fast way to do this?
---------------------------------------------------------------------------*/

static int                          /* Returns nonzero on error */
vxli_cubic_point(
    int dtype,                      /* Type of destination */
    void *dest,                     /* Destination */
    voxel_array *src,               /* Voxel array to interpolate in */
    double *point,                  /* Point to interpolate at */
    long   *step)                   /* Precomputed step size for each axis */
{
    int type;                       /* Type of the array */
    int rank;                       /* Rank of the array */
    long   *dimen;                  /* Dimensions of the array */
    void   *data;                   /* Data contained in the array */
    double alpha;                   /* Fractional part of requested point */
    long   index[VXL_MAX_RANK];     /* Index of leftmost sample point */
    double coef [VXL_MAX_RANK][4];  /* Sample weights */
    long   delta[VXL_MAX_RANK];     /* Index increment to each neighbor */
    long   offset;                  /* Corner location in vectorized array */
    int    valid;                   /* Is corner inside array? */
    double value = 0.0;             /* Value at the given position */
    double v1;                      /* Voxel value at a neighbor */
#ifdef HAVE_COMPLEX_H
    double complex cvalue = 0.0;    /* Value at given position */
    double complex cv1;             /* Voxel value at a neighbor */
#endif
    long coord;                     /* Coordinate value */
    int i, flag=0;
    void *datum;

    /* Get some parameters from the source */
    rank  = src->rank;
    dimen = src->dimen;
    type  = src->type;
    data  = src->data;

    /* Check whether the requested point is completely outside the source;
       compute index to neighbors and sample weights.  NOTE: floor() is very
       slow on the x86 platform, so we play some funny games with casting
       instead. */
    /* FIXME: Try using alpha + 2.0 instead? */
    for (i = 0; i < rank; i++) {
        if ( -2.0 < point[i] && point[i] < dimen[i] + 1.0 ) {
            index[i] = (long)(point[i] + 2.0) - 2;
            alpha = point[i] - (double)index[i];
            index[i] = index[i] - 1;
            coef[i][0] = -(alpha)*(alpha-1.0)*(alpha-2.0)/6.0;
            coef[i][1] =  (alpha+1.0)*(alpha-1.0)*(alpha-2.0)/2.0;
            coef[i][2] = -(alpha+1.0)*(alpha)*(alpha-2.0)/2.0;
            coef[i][3] =  (alpha+1.0)*(alpha)*(alpha-1.0)/6.0;
        }
        else {
            flag = bips_zero(1, dtype, dest, 0);
            if (flag)   NZReturnLibError();
            else        return 0;
        }
    }

    /* Loop over the neighbors to compute weighted sum */
    vxli_first_neighbor(rank, delta);
    do {

	/* Compute coordinates of current neighbor; check for boundaries */
	valid  = 1;     /* Assume inside source until proven otherwise */
        offset = 0;
	for (i = rank-1; i >= 0; i--) {
	    coord   = index[i] + delta[i];
            offset += coord * step[i];
	    if ( coord < 0 || coord >= dimen[i] ) {
		valid = 0;
		break;
	    }
	}
	if ( !valid )   continue;

        /* Locate the value for that neighbor */
        datum = bips_locate(type, data, offset);
        if (datum == 0)
                NZReturnError(ERR_ASSERT,
                              "Impossible error 1 in vxli_cubic_point");

        /* Case 1: Interpolate in plain double */
        if (type != INTP_FCOMPLEX && type != INTP_DCOMPLEX) {

            /* Fetch voxel value at the neighbor */
            flag  = bips_double(1, &v1, 0, type, datum, 0);
            if (flag)
                NZReturnError(ERR_ASSERT,
                              "Impossible error 2 in vxli_cubic_point");
            
            /* Multiply by weight along each dimension and add to value */
            if ( v1 != 0.0 ) {
                for (i = 0; i < rank; i++)
                    v1 *= coef[i][delta[i]];
                value += v1;
            }
        }

        /* Case 2: Interpolate in double complex */
#ifdef HAVE_COMPLEX_H
        else {

            /* Fetch voxel value at the neighbor */
            flag = bips_dcomplex(1, &cv1, 0, type, datum, 0);
            if (flag)    
                NZReturnError(ERR_ASSERT,
                              "Impossible error 3 in vxli_cubic_point");
            
            /* Multiply by weight along each dimension and add to value */
            if ( cv1 != 0.0 ) {
                for (i = 0; i < rank; i++)
                    cv1 *= coef[i][delta[i]];
                cvalue += cv1;
            }
        }
#endif

    } while (vxli_next_neighbor(rank, delta, 4));

    /* Store the final result */
    if (type != INTP_FCOMPLEX && type != INTP_DCOMPLEX)
        flag = bips_round_cast(1, dtype, dest, 0, INTP_DOUBLE, &value, 0);
#ifdef HAVE_COMPLEX_H
    else
        flag = bips_cast(1, dtype, dest, 0, INTP_DCOMPLEX, &cvalue, 0);
#endif
    if (flag)    NZReturnLibError();

    return 0;
}



/*------------------------------------------------------------------------
vxl_cubic_point - Cubic interpolation at a point in a voxel array

This function evaluates the value of a voxel array at a given
position, using cubic Lagrange interpolation in the four nearest
sample values; voxels outside the array are assumed to be zero.  This
version fully checks its arguments and is suitable for external use;
see vxli_cubic_point() for a faster version with fewer checks.
---------------------------------------------------------------------------*/

int                             /* Returns nonzero on error */
vxl_cubic_point(
    int dtype,                  /* Type of destination */
    void *dest,                 /* Destination */
    voxel_array *src,           /* Voxel array to interpolate in */
    double *point)              /* Point to interpolate at */
{
    int type;                   /* Type of the source array */
    int srank;                  /* Rank of the source array */
    long *sdimen;               /* Dimensions of the source array */
    long sstep[VXL_MAX_RANK];   /* Step sizes in source array */
    int i, flag;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    srank  = src->rank;
    sdimen = src->dimen;
    type   = src->type;

    /* Precompute step size along each source coordinate */
    sstep[srank-1] = 1;
    for (i = srank-1; i > 0; i--)
	sstep[i-1] = sstep[i] * sdimen[i];

    /* Call the internal routine to do the actual work */
    flag = vxli_cubic_point(dtype, dest, src, point, sstep);
    
    return flag;
}



/*-----------------------------------------------------------------------
vxl_cubic_map - Cubic interpolation at points defined by a coord map

This function creates a new voxel array whose elements are obtained by
interpolating in a source array at the coordinate points given by a
coordinate map array.  The source array may be of any type and shape.
The map array must be of type double and its last dimension must equal
the rank of the source array.  The type of the result is specified by
the caller and defaults to the type of the source array; the shape is
the same as the coordinate map less the last dimension.  The
interpolation calculations are done in double or double complex.

FIXME: Allow the coordinate array to be type float?
------------------------------------------------------------------------*/

int                             /* Returns nonzero on error */
vxl_cubic_map(
    voxel_array *dest,          /* Destination array */
    int dtype,                  /* Desired destination type, or zero */
    voxel_array *src,           /* Source array */
    voxel_array *map)           /* Coordinate map array */
{
    int srank;                  /* Rank of the source */
    long *sdimen=0;             /* Dimensions of source array */
    long sstep[VXL_MAX_RANK];   /* Step sizes in source array */
    int stype;                  /* Type of the source array */
    int mrank;                  /* Rank of map array */
    int drank;                  /* Rank of destination array */
    long dimen[VXL_MAX_RANK];   /* Dimensions of map and dest arrays */
    long coord[VXL_MAX_RANK];   /* Current row in the map array */
    void *datum=0;              /* Current voxel in the destination */
    void *point=0;              /* Current sample point in the source array */
    int i, flag;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    srank  = src->rank;
    sdimen = src->dimen;
    stype  = src->type;

    /* Destination type defaults to the source type */
    if (dtype == 0)
        dtype = stype;

    /* Precompute step size along each source coordinate */
    sstep[srank-1] = 1;
    for (i = srank-1; i > 0; i--)
	sstep[i-1] = sstep[i] * sdimen[i];

    /* Validate the coordinate array; extract rank and dimensions */
    if ( map == 0 || map->magic != VXL_MAGIC ||
         map->rank < 1 || map->data == 0 )
        NZReturnError(ERR_INIT, "Invalid or empty map array");
    if ( map->type != INTP_DOUBLE )
        NZReturnError(ERR_INIT, "Coord map type must be double");
    mrank = map->rank;
    for (i = 0; i < mrank; i++)
        dimen[i] = map->dimen[i];

    /* Add final dimension of length one to the map if the source is rank
       one and the current final dimension is not one. */
    if (srank == 1 && dimen[mrank-1] != 1)
        dimen[mrank++] = 1;
    drank = mrank - 1;

    /* Verify that the source and map are compatible */
    if ( dimen[mrank-1] != srank )
        NZReturnError(ERR_INIT, 
                      "Last coord map dimen must match source rank");    

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    flag = vxl_alloc_array(dest, dtype, drank, dimen);
    if (flag)    NZReturnLibError();

    /* Loop over the rows of the coordinate array, each of which defines 
       one sample point in the source array and thus a single voxel of the 
       destination. */
    vxli_firstrow(mrank, coord);
    do {

        /* Locate the sample point in the map array and the corresponding
           destination element. */
        point = vxli_locate(map,  coord);
        datum = vxli_locate(dest, coord);
        if (point == 0 || datum == 0)
            NZReturnError(ERR_ASSERT, "Internal error 1 in vxl_cubic_map");
        
        /* Interpolate in source array and save result */
        flag = vxli_cubic_point(dtype, datum, src, point, sstep);
        if (flag)    NZReturnLibError();            

    } while (vxli_nextrow(mrank, dimen, coord));

    return 0;
}



/*======================================================================
Functions for affine resampling and interpolation
======================================================================*/



/*------------------------------------------------------------------------
vxli_sharp_clip - Clip a line segment against a sharp boundary

This function is used in the fast resampling functions with nearest
neighbor interpolation to clip the image of an ouput row against the
limits of the input image.  The lower and upper limit arguments are
the minimum and maximum values for which source data exists, typically
zero and the array dimension MINUS ONE.  The output array didx
contains indices in the output row corresponding to intersections with
the source array; they are to be interpreted as follows:

     0       <= i < didx[0]    Point i is outside
     didx[0] <= i < didx[1]    Point i is inside
     didx[1] <= i < npoint     Point i is outside

Points outside the source may simply be set to zero; points inside the
source may be interpolated by a fast algorithm that need not check
each vertex against the source limits.  If the segment completely
misses the input image, then both didx[0] and didx[1] are set to
npoint.
-------------------------------------------------------------------------*/

static void
vxli_sharp_clip(
    int rank,                  /* Number of dimensions */
    long npoint,               /* Number of points in output row */
    double *origin,            /* Image of first point */
    double *delta,             /* Increment to next point */
    double *lower,             /* Lower limit in source array */
    double *upper,             /* Upper limit in source array */
    long *didx)                /* (Out) Row intersections with source */
{
    double x0, x1;             /* Row intersections on current axis */
    double y0, y1;             /* Tightest row intersections */
    double ymax;               /* Upper bound on intersection */
    double invdel;
    int i;

    /* Initialize tightest intersections so far */
    y0 = 0.0;
    y1 = ymax = (double)(npoint-1);

    /* Loop over axes and update intersections */
    for (i = 0; i < rank; i++) {

	/* Case 0: segment is constant on this axis */
	if (delta[i] == 0.0) {
	    if  ( origin[i] < lower[i] || origin[i] > upper[i] ) {
		didx[0] = didx[1] = npoint;
		return;
	    }
	}

	/* Cases 1 and 2: segment is not constant */
	else {

	    /* Compute intersections */
	    invdel = 1.0 / delta[i];
	    x0 = (lower[i] - origin[i]) * invdel;
	    x1 = (upper[i] - origin[i]) * invdel;

	    /* Case 1: segment directed toward plus axis */
	    if (delta[i] > 0.0) {
		if (x0 > ymax || x1 < 0.0)  {
		    didx[0] = didx[1] = npoint;
		    return; 
		}
		if ( x0 > y0 ) y0 = x0;
		if ( x1 < y1 ) y1 = x1;
	    }

	    /* Case 2: segment directed toward minus axis */
	    else if (delta[i] < 0.0) {
		if (x1 > ymax || x0 < 0.0)  {
		    didx[0] = didx[1] = npoint;
		    return; 
		}
		if ( x1 > y0 ) y0 = x1;
		if ( x0 < y1 ) y1 = x0;
	    }
	}
    }

    /* Save results and return */
    if ( y0 > ymax || y1 < 0 || y0 > y1 )  {
	didx[0] = didx[1] = npoint;
    }
    else {
	didx[0] = ceil (y0);
	didx[1] = floor(y1) + 1;
    }
    if (0)                                   /* DEBUG */
	fprintf(stderr, "boundaries: %f %f => %ld %ld\n", 
		y0, y1, didx[0], didx[1]);
    return;
}



/*------------------------------------------------------------------------
vxli_fuzzy_clip - Clip a line segment against a fuzzy boundary

This function is used in the fast resampling functions with linear or
higher order interpolation to clip the image of an ouput row against
the limits of the input image; the fuzziness comes in because image
points near the boundary must be handled specially since only some
neighbors exist.  The width must be the half-width of the boundary or
radius (in the Lsup norm) of the interpolation neighborhood.  This
function is believed to work correctly for the exact value of the
width; no error allowance for rounding should be necessary.  The
output array didx contains indices in the output row corresponding to
intersections with the source array; they are to be interpreted as
follows:

     0       <= i < didx[0]    Point i is completely outside (clean miss)
     didx[0] <= i < didx[1]    Point i is near the boundary  (penumbra)
     didx[1] <= i < didx[2]    Point i is completely inside  (umbra)
     didx[2] <= i < didx[3]    Point i is near the boundary  (penumbra)
     didx[3] <= i < npoint     Point i is completely outside (clean miss)

There are two special cases of interest:  If the row cleanly misses both
the umbra and penumbra, then didx[0] = 0 and didx[1] = didx[2] = didx[3]
= npoint.  If the row misses the umbra but hits the penumbra, then didx[0]
is the start of the penumbra and didx[1] = didx[2] = didx[3] are set to
the end of the penumbra.

Points completely outside the source may simply be set to zero; points
completely inside the source may be interpolated by a fast algorithm
that need not check each vertex against the source limits; points near
the boundary must be interpolated by a careful algorithm that actually
checks each vertex against the limits of the source.
-------------------------------------------------------------------------*/

static void
vxli_fuzzy_clip(
    int rank,                  /* Number of dimensions */
    long npoint,               /* Number of points in output row */
    double *origin,            /* Image of first point */
    double *delta,             /* Increment to next point */
    double *lower,             /* Lower limit in source array */
    double *upper,             /* Upper limit in source array */
    double width,              /* Width of fuzzy border */
    long *didx)                /* (Out) Row intersections with source */
{
    double x0, x1, x2, x3;     /* Row intersections on current axis */
    double y0, y1, y2, y3;     /* Tightest row intersections */
    double ymax;               /* Upper limit of row */
    double invdel;             /* Reciprocal of delta[i] */
    int penumbra = 0;          /* Entirely within penumbra for some axis? */
    int miss = 0;              /* Clean miss on some axis? */
    int i;

    /* Initialize tightest intersections */
    ymax = (double)(npoint-1);
    y0 = y1 = 0.0;
    y2 = y3 = ymax;

    /* Loop over axes and update intersections */
    for (i = 0; i < rank; i++) {

	/* Case 0: segment is constant on this axis */
	if (delta[i] == 0.0) {

	    /* Subcase 0.1: Clean miss on this axis => all axes */
	    if  ( origin[i] < lower[i] - width ||
		  origin[i] > upper[i] + width ) {
		miss = 1;
		break;
	    }
	    /* Subcase 0.2: Within penumbra on this axis */
            else if  ( origin[i] < lower[i] + width ||
		  origin[i] > upper[i] - width )
		penumbra = 1;

	    /* Subcase 0.3: Within umbra on this axis */
	    else
		/* Do nothing */ ;
	}

	/* Cases 1 and 2: segment is not constant */
	else {

	    /* Compute intersection, checking for clean misses */
	    invdel = 1.0 / delta[i];
	    x0 = (lower[i] - width - origin[i]) * invdel;
	    x1 = (lower[i] + width - origin[i]) * invdel;
	    x2 = (upper[i] - width - origin[i]) * invdel;
	    x3 = (upper[i] + width - origin[i]) * invdel;

	    /* Case 1: segment directed toward plus axis */
	    if (delta[i] > 0.0) {
		if ( x0 > y0 ) y0 = x0;
		if ( x1 > y1 ) y1 = x1;
		if ( x2 < y2 ) y2 = x2;
		if ( x3 < y3 ) y3 = x3;
	    }

	    /* Case 2: segment directed toward minus axis */
	    else if (delta[i] < 0.0) {
		if ( x3 > y0 ) y0 = x3;
		if ( x2 > y1 ) y1 = x2;
		if ( x1 < y2 ) y2 = x1;
		if ( x0 < y3 ) y3 = x0;
	    }
	}
    }

    /* Convert results to indices:  The rounding of didx[1] and didx[2] 
       is chosen such that an exact integer value does interpolation as 
       part of the umbra */
    if ( miss || y0 > ymax || y3 < 0.0 ) {
	didx[0] = didx[1] = didx[2] = didx[3] = npoint;
    }
    else if ( penumbra || y1 > y2 ) {
	didx[0] = ceil (y0);
	didx[3] = floor(y3) + 1;
	didx[1] = didx[2] = didx[3];
    }
    else {
	didx[0] = ceil (y0);
	didx[1] = ceil (y1);
	didx[2] = floor(y2) + 1;
	didx[3] = floor(y3) + 1;
    }

    return;
}



/*------------------------------------------------------------------------
vxli_affine_exact - Affine transform and resample without interpolation

This function transforms a voxel array using an affine transform,
resamples in the new coordinates without interpolation, and stores the
results as a newly allocated voxel array.  The elements of the affine
matrix and offset are required to be integers, in which case the
coordinates in the source arrays are always integers and interpolation
is never required; in addition, the source and destination must be the
same type.  This special case occurs frequently as image flips,
rotations by multiples of pi/2, and transpositions, and can be handled
more efficiently than the general nearest neighbor or linear
interpolation.  

Since this function is intended to be called from vxl_affine_nearest
or vxl_affine_linear to handle this special case more efficiently, it
is declared as static and has an atypical return value: If the given
affine transform satisfies the requirements, it does the transform and
returns 0 if successful, or >0 for an error; otherwise it returns -1.
Furthermore, it does not check that the source is a valid voxel array,
on the presumption that the caller has already checked this.

The rank of the destination need not match the rank of the source;
this permits projections into a lower dimensional space and liftings
into a higher dimensional one.  However, the rank of the source must
be at least one.  The transform must be specified as the
destination-to-source transformation, since it is used to transform
each destination voxel coordinate back into the source; this is the
opposite of what you might naively expect, so be careful to do it
right.  The number of rows in the matrix is assumed to be the same as
the source rank, and the number of columns is assumed to be the
destination rank; the C ordering of matrix elements is assumed.  The
type of the destination must be the same of the source; this differs
for linear interpolation, in which the types may be different.

The coreps parameter determines how close to integers the coefficients
of the transform must be to use the exact algorithm.  To use this
algorithm only when the transform coefficients are exactly integers,
set coreps = 0.0; otherwise the recommended value is a few times
DBL_EPSILON.  The purpose of this test is to permit a a transform such
as rotation of a multiple of pi/2 to be treated as a fast transform
even though it cannot be computed exactly as integers.

ALGORITHM: This is a fast implementation that clips the back transform
of each row against the limits of the source to save boundary tests in
the inner loop and does the base address calculations in a fast loop.
-------------------------------------------------------------------------*/

#define VXL_MAX_MATRIX (VXL_MAX_RANK*VXL_MAX_RANK)

static int                       /* 0 => success;  >0 => error;  <0 => N/A */
vxli_affine_exact(
    voxel_array *dest,           /* Destination array */
    int type,                    /* Desired type of destination */
    int rank,                    /* Desired rank of destination */
    long *dimen,                 /* Desired dimensions of destination */
    voxel_array *src,            /* Source array to transform and resample */
    double *matrix,              /* Linear part of the affine transform */
    double *offset,              /* Constant part of the transform */
    double coreps)               /* Permissible error in source coords */
{
    int srank;                   /* Rank of the source */
    int stype;                   /* Type of the source */
    long imatrix[VXL_MAX_MATRIX];/* Transform matrix approximated by longs */
    long ioffset[VXL_MAX_RANK];  /* Transform offset approximated by longs */
    double corerr;               /* Error of approximation, per coord */
    long   coord[VXL_MAX_RANK];  /* Voxel coords of destination row */
    long   coors[VXL_MAX_RANK];  /* Voxel coords in source, as long */
    double point[VXL_MAX_RANK];  /* Voxel coords in source, as double */
    double delta[VXL_MAX_RANK];  /* Column increments to source position */
    double lower[VXL_MAX_RANK];  /* Lower bounds in source */
    double upper[VXL_MAX_RANK];  /* Upper bounds in source */
    long sstep;            /* Source stride along a given source axis */
    long sslast;           /* Source stride for unit step in last dest axis */
    long ncol;             /* Number of columns in dest row */
    long didx[2];          /* Clipped limits in dest row */
    int  isrc, idst;       /* Indices over source and destination axes */
    unsigned char *drow;   /* Start of a destination row */
    int  flag;             /* BIPS status returned */
    long nelem;            /* Number of non-zero elements in dest row */
    void *dst0;            /* First element to copy to in dest row */
    void *src0;            /* First element to copy from in src row */
    long i;

    /* Check that source and destination types match */
    srank = src->rank;
    stype = src->type;
    if (stype != type)
	NZReturnError(ERR_INCOMP, "Source and destination types do not match");

    /* Approximate transform by longs and test if good enough; the
       comparison between corerr and coreps handles equality so that
       setting coreps = 0.0 will use the exact algorithm if and only
       if the transform contains only exact integers. */
    for (idst = 0; idst < rank; idst++) {
	for (isrc = 0; isrc < srank; isrc++) {
	    i = isrc*rank + idst;
	    imatrix[i] = floor(matrix[i] + 0.5);
	    corerr = fabs(matrix[i] - (double)imatrix[i]);
	    if (corerr > coreps)
		return -1;
	}
        ioffset[idst] = floor(offset[idst] + 0.5);
	corerr = fabs(offset[idst] - (double)ioffset[idst]);
        corerr = corerr / (double)dimen[idst];
        if (corerr > coreps)
	    return -1;
    }

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    flag = vxl_alloc_array(dest, type, rank, dimen);
    if (flag)    NZReturnLibError();
    ncol = dimen[rank-1];

    /* Prepare point increments along last dimension of destination,
       plus lower and upper limits for clipper */
    for (isrc = 0; isrc < srank; isrc++) {
	delta[isrc] = imatrix[isrc*rank+rank-1];
	lower[isrc] = 0.0;
	upper[isrc] = (double)(src->dimen[isrc] - 1);
    }

    /* Compute the source stride that corresponds to a unit step along 
       the last axis of the destination */
    sstep  = 1;
    sslast = imatrix[srank*rank-1];
    for (isrc = srank-2; isrc >= 0; isrc--) {
	sstep  *= src->dimen[isrc+1];
        sslast += sstep * imatrix[isrc*rank+rank-1];
    }

    /* Loop over the rows of the destination */
    vxli_firstrow(rank, coord);
    do {

	/* Compute the source coordinates that correspond to the first
           point in the current row of the destination */
	for (isrc = 0; isrc < srank; isrc++) {
	    point[isrc] = ioffset[isrc];
	    for (idst = 0; idst < rank-1; idst++) {
		point[isrc] += imatrix[isrc*rank+idst] * coord[idst];
	    }
	}

	/* Clip source coordinates against source boundary */
	vxli_sharp_clip(srank, ncol, point, delta, lower, upper, didx);

	/* Zero destination voxels outside the source */
	drow = vxli_locate(dest, coord);
        if (drow == 0)    NZReturnLibError();
	flag = bips_zero_range(0, didx[0], type, drow, 1);
	if (flag)    NZReturnLibError();
	flag = bips_zero_range(didx[1], ncol, type, drow, 1);
	if (flag)    NZReturnLibError();

        /* Copy desired source pixels, if any, into the destination */
        nelem = didx[1] - didx[0];
        if ( nelem > 0 ) {

            /* Find coords for first voxel actually present in source */
            for (isrc = 0; isrc < srank; isrc++)
                coors[isrc] = point[isrc] +
                    didx[0] * imatrix[isrc*rank+rank-1];

            /* Copy desired source voxels into the destination row */
            dst0  = bips_locate(type, drow, didx[0]);
            src0  = vxli_locate(src, coors);
            if (src0 == 0)    NZReturnLibError();
            if (type == stype)
                flag = bips_copy(nelem, type, dst0, 1, src0, sslast);
            else
                flag = bips_cast(nelem, type, dst0, 1, stype, src0, sslast);
            if (flag)    NZReturnLibError();

        }

    } while (vxli_nextrow(rank, dimen, coord));

    return ERR_OKAY;
}



/*------------------------------------------------------------------------
vxl_affine_map - Generate coordinate map for an affine transform

This function generates a new array which contains the output
coordinates for the given affine transform over a notional destination
image of the given rank and dimensions.  The rank of the result
actually returned is one higher than the notional destination, since
the last dimension is used to range over the output coordinates.  The
result is always of type double.

FIXME: Reimplement using a faster algorithm.
-------------------------------------------------------------------------*/

int                              /* Returns non-zero on error */
vxl_affine_map(
    voxel_array *result,         /* Result array */
    int rank,                    /* Desired rank of notional dest array */
    long *dimen,                 /* Desired dimensions of destination */
    double *matrix,              /* Linear part of the affine transform */
    double *offset)              /* Constant part of the transform */
{
    long  dimen2[VXL_MAX_RANK];  /* Dimensions of result */
    long   coord[VXL_MAX_RANK];  /* Voxel coords in destination */
    double point[VXL_MAX_RANK];  /* Point coords in source */
    double *resdata;             /* Pointer into result data */
    int isrc, idst;              /* Indices over source and destination axes */
    int flag, i;

    /* Compute dimensions of the result array */
    for (i = 0; i < rank; i++)
        dimen2[i] = dimen[i];
    dimen2[i] = rank;

    /* Allocate the result array */
    if (result == 0 || result->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid result array");
    flag = vxl_alloc_array(result, INTP_DOUBLE, rank+1, dimen2);
    if (flag)    NZReturnLibError();

    /* Loop over the destination voxels */
    resdata = result->data;
    vxli_firstvoxel(rank, coord);
    do {

	/* Compute corresponding position in source */
	for (isrc = 0; isrc < rank; isrc++) {
	    point[isrc] = offset[isrc];
	    for (idst = 0; idst < rank; idst++) {
		point[isrc] += matrix[isrc*rank+idst] * (double)coord[idst];
	    }
	}

        /* Save coordinates in result */
        for (i = 0; i < rank; i++)
            *resdata++ = point[i];

    } while (vxli_nextvoxel(rank, dimen, coord) );

    return ERR_OKAY;
}



/*------------------------------------------------------------------------
vxl_affine_nearest0 - Affine transform and resample with NN interpolation

This function transforms a voxel array using an affine transform,
resamples in the new coordinates using nearest neighbor interpolation,
and stores the results as a newly allocated voxel array.  The rank of
the destination need not match the rank of the source; this permits
projections into a lower dimensional space and liftings into a higher
dimensional one.  However, the rank of the source must be at least
one.  The transform must be specified as the destination-to-source
transformation, since it is used to transform each destination voxel
coordinate back into the source; this is the opposite of what you
might naively expect, so be careful to do it right.  The number of
rows in the matrix is assumed to be the same as the source rank, and
the number of columns is assumed to be the destination rank; the C
ordering of matrix elements is assumed.  The type of the destination
need not be the same of the source; this costs nothing extra since all
the intermediate calculations are done in double anyway.

This is a reference implementation.  For simplicity, complex types are
not supported.

FIXME: This reference implementation is arguably obsolete, now that we
can test affine.nearest() by using affine.map() and array.nearest().
Consider deleting it.  But note that pref_affine uses this as a dumb
implementation to compare our better implementation to.
-------------------------------------------------------------------------*/

int                              /* Returns non-zero on error */
vxl_affine_nearest0(
    voxel_array *dest,           /* Destination array */
    int type,                    /* Desired type of destination */
    int rank,                    /* Desired rank of destination */
    long *dimen,                 /* Desired dimensions of destination */
    voxel_array *src,            /* Source array to transform and resample */
    double *matrix,              /* Linear part of the affine transform */
    double *offset)              /* Constant part of the transform */
{
    int srank;                   /* Rank of the source */
    long   coord[VXL_MAX_RANK];  /* Voxel coords in destination */
    double point[VXL_MAX_RANK];  /* Point coords in source */
    double delta[VXL_MAX_RANK];  /* Column increments to source position */
    double *value;               /* Row of interpolated values */
    long ncol;                   /* Number of columns in dest row */
    int isrc, idst;              /* Indices over source and destination axes */
    int icol;                    /* Index over columns of a dest row */
    unsigned char *drow;         /* Start of a destination row */
    int flag;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    srank  = src->rank;

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    flag = vxl_alloc_array(dest, type, rank, dimen);
    if (flag)    NZReturnLibError();

    /* Allocate memory for a row of interpolated values */
    ncol = dimen[rank-1];
    value = mallock(ncol*sizeof(double));
    if (value == 0)    NZReturnLibError();

    /* Prepare point increments along last dimension of destination */
    for (isrc = 0; isrc < srank; isrc++)
	delta[isrc] = matrix[isrc*rank+rank-1];

    /* Loop over the rows of the destination */
    vxli_firstrow(rank, coord);
    do {

	/* Compute source point corresponding to first column in row */
	for (isrc = 0; isrc < srank; isrc++) {
	    point[isrc] = offset[isrc];
	    for (idst = 0; idst < rank-1; idst++) {
		point[isrc] += matrix[isrc*rank+idst] * (double)coord[idst];
	    }
	}

	/* Loop over the columns of the row */
	for (icol = 0; icol < ncol; icol++) {

	    /* Interpolate in source and save in row array */
            flag = vxl_nearest_point(INTP_DOUBLE, &value[icol], src, point);
            if (flag)    NZReturnLibError();

	    /* Update position to next column */
	    for (isrc = 0; isrc < srank; isrc++)
		point[isrc] += delta[isrc];
	}

	/* Copy row to destination, converting to destination type */
	drow = vxli_locate(dest, coord);
        if (drow == 0)    NZReturnLibError();
	flag = bips_cast(ncol, type, drow, 1, INTP_DOUBLE, value, 1);
	if (flag)    NZReturnLibError();

    } while (vxli_nextrow(rank, dimen, coord));

    /* Free temporary storage and return */
    free(value);
    return ERR_OKAY;
}



/*------------------------------------------------------------------------
vxl_affine_nearest - Affine transform and resample with NN interpolation

This function transforms a voxel array using an affine transform,
resamples in the new coordinates using nearest neighbor interpolation,
and stores the results as a newly allocated voxel array.  The rank of
the destination need not match the rank of the source; this permits
projections into a lower dimensional space and liftings into a higher
dimensional one.  However, the rank of the source must be at least
one.  The transform must be specified as the destination-to-source
transformation, since it is used to transform each destination voxel
coordinate back into the source; this is the opposite of what you
might naively expect, so be careful to do it right.  The number of
rows in the matrix is assumed to be the same as the source rank, and
the number of columns is assumed to be the destination rank; the C
ordering of matrix elements is assumed.  The type of the destination
must be the same of the source; this differs for linear interpolation,
in which the types may be different.

This is a fast implementation that clips the back transform
of each row against the limits of the source to save boundary tests in
the inner loop and does the base address calculations in a fast loop.

FIXME: Check the various 0.5 offsets and exact boundary calculations
to make sure they're right.  Check in particular whether the lower and
upper boundary values for vxli_sharp_clip are correct.

FIXME: Allow the source and destination types to differ.  The required
approach would be to gather source elements to a temporary array, and
then cast them to the desination.
-------------------------------------------------------------------------*/

int                              /* Returns non-zero on error */
vxl_affine_nearest(
    voxel_array *dest,           /* Destination array */
    int type,                    /* Desired type of destination */
    int rank,                    /* Desired rank of destination */
    long *dimen,                 /* Desired dimensions of destination */
    voxel_array *src,            /* Source array to transform and resample */
    double *matrix,              /* Linear part of the affine transform */
    double *offset)              /* Constant part of the transform */
{
    int srank;                   /* Rank of the source */
    long   coord[VXL_MAX_RANK];  /* Voxel coords of destination row */
    long   sstep[VXL_MAX_RANK];  /* Source stride along each axis */
    double point[VXL_MAX_RANK];  /* Point coords in source */
    double delta[VXL_MAX_RANK];  /* Column increments to source position */
    double lower[VXL_MAX_RANK];  /* Lower bounds in source */
    double upper[VXL_MAX_RANK];  /* Upper bounds in source */
    long   *index;               /* Source index to each voxel in a row */
    long ncol;                   /* Number of columns in dest row */
    long didx[2];                /* Clipped limits in dest row */
    int isrc, idst;              /* Indices over source and destination axes */
    int icol;                    /* Index over columns of a dest row */
    unsigned char *drow;         /* Start of a destination row */
    int flag;                    /* BIPS status returned */
    long nelem;                  /* Number of non-zero elements in dest row */
    long *idx0;
    void *dst0;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    srank  = src->rank;

    /* Check that source and destination types match */
    /* FIXME: Modify to permit source and dest to have different types? */
    if (src->type != type)
	NZReturnError(ERR_INCOMP, "Source and destination types do not match");

    /* If possible, use exact resampling algorithm; this requires that
       all transform coefficients are close enough to integers */
    flag = vxli_affine_exact(dest, type, rank, dimen,
			     src, matrix, offset, 5.0*DBL_EPSILON);
    if (flag == 0)
	return ERR_OKAY;
    else if (flag > 0)
        NZReturnLibError();

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    flag = vxl_alloc_array(dest, type, rank, dimen);
    if (flag)    NZReturnLibError();

    /* Allocate working memory */
    ncol  = dimen[rank-1];
    index = mallock(ncol*sizeof(long));
    if (index == 0)    NZReturnLibError();

    /* Compute source stride along each axis */
    sstep[srank-1] = 1;
    for (isrc = srank-2; isrc >= 0; isrc--) {
	sstep[isrc] = sstep[isrc+1] * src->dimen[isrc+1];
    }

    /* Prepare point increments along last dimension of destination;
       also lower and upper limits for clipper */
    for (isrc = 0; isrc < srank; isrc++) {
	delta[isrc] = matrix[isrc*rank+rank-1];
	lower[isrc] = 0.0                        + BoundaryFuzz;
	upper[isrc] = (double)(src->dimen[isrc]) - BoundaryFuzz;
    }

    /* Loop over the rows of the destination */
    vxli_firstrow(rank, coord);
    do {

	/* Compute source coordinates of first point in the current row;
	   offset by 0.5 so that truncation will round to nearest neighbor */
	for (isrc = 0; isrc < srank; isrc++) {
	    point[isrc] = offset[isrc] + 0.5;
	    for (idst = 0; idst < rank-1; idst++) {
		point[isrc] += matrix[isrc*rank+idst] * (double)coord[idst];
	    }
	}

	/* Clip source coordinates against source boundary */
	vxli_sharp_clip(srank, ncol, point, delta, lower, upper, didx);

	/* Zero destination voxels outside the source */
	drow = vxli_locate(dest, coord);
        if (drow == 0)    NZReturnLibError();
	flag = bips_zero_range(0, didx[0], type, drow, 1);
	if (flag)    NZReturnLibError();
	flag = bips_zero_range(didx[1], ncol, type, drow, 1);
	if (flag)    NZReturnLibError();

	/* Compute indices into the source for this row; note that clipping
           insures that all coordinates here are non-negative, so it's safe
           to cast to long rather than use floor() */
        for (icol = didx[0]; icol < didx[1]; icol++)
            index[icol] = 0;
	for (isrc = 0; isrc < srank; isrc++) {
	    point[isrc] += didx[0] * delta[isrc];
	    for (icol = didx[0]; icol < didx[1]; icol++) {
		index[icol] += sstep[isrc] * (long)(point[isrc]);
		point[isrc] += delta[isrc];
	    }
	}

	/* Gather desired source voxels into the destination row */
	nelem = didx[1] - didx[0];
	dst0  = bips_locate(type, drow, didx[0]);
	idx0  = index + didx[0];
	flag  = bips_gather(nelem, type, dst0, 1, src->data, idx0);
	if (flag)    NZReturnLibError();

    } while (vxli_nextrow(rank, dimen, coord));

    /* Free temporary storage and return */
    free(index);
    return ERR_OKAY;
}



/*------------------------------------------------------------------------
vxl_affine_linear0 - Affine transform and resample with linear interpolation

This function transforms a voxel array using an affine transform,
resamples in the new coordinates using linear interpolation between
voxels, and stores the results as a newly allocated voxel array.  The
rank of the destination need not match the rank of the source; this
permits projections into a lower dimensional space and liftings into a
higher dimensional one.  However, the rank of the source must be at
least one.  The transform must be specified as the destination-
to-source transformation, since it is used to transform each
destination voxel coordinate back into the source; this is the
opposite of what you might naively expect, so be careful to do it
right.  The number of rows in the matrix is assumed to be the same as
the source rank, and the number of columns is assumed to be the
destination rank; the C ordering of matrix elements is assumed.  The
type of the destination need not be the same of the source; this costs
nothing extra since all the intermediate calculations are done in
double anyway.

This implementation is simple but slow; it is intended primarily as a
reference implementation to verify faster but less comprehensible
implementations.  Given these limited goals, complex types are not
supported.

FIXME: This function appears to be used only in perf_affine as the
default dumb implementation, for comparison with better methods.  Is
that really make it worth it to maintain this version?

FIXME: This function should probably round rather than simply casting
for destinations of integral type.
-------------------------------------------------------------------------*/

int                              /* Returns non-zero on error */
vxl_affine_linear0(
    voxel_array *dest,           /* Destination array */
    int type,                    /* Desired type of destination */
    int rank,                    /* Desired rank of destination */
    long *dimen,                 /* Desired dimensions of destination */
    voxel_array *src,            /* Source array to transform and resample */
    double *matrix,              /* Linear part of the affine transform */
    double *offset)              /* Constant part of the transform */
{
    int srank;                   /* Rank of the source */
    long   coord[VXL_MAX_RANK];  /* Voxel coords in destination */
    double point[VXL_MAX_RANK];  /* Point coords in source */
    void *value=0;               /* Location of a single destination value */
    int isrc, idst;              /* Indices over source and destination axes */
    int flag;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    srank  = src->rank;

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    flag = vxl_alloc_array(dest, type, rank, dimen);
    if (flag)    NZReturnLibError();

    /* Loop over the destination voxels */
    vxli_firstvoxel(rank, coord);
    do {

	/* Compute corresponding position in source */
	for (isrc = 0; isrc < srank; isrc++) {
	    point[isrc] = offset[isrc];
	    for (idst = 0; idst < rank; idst++) {
		point[isrc] += matrix[isrc*rank+idst] * (double)coord[idst];
	    }
	}

	/* Interpolate in source and save in destination */
        value = vxli_locate(dest, coord);
        flag  = vxl_linear_point(dest->type, value, src, point);
        if (flag)    NZReturnLibError();

    } while (vxli_nextvoxel(rank, dimen, coord) );

    return ERR_OKAY;
}



/*------------------------------------------------------------------------
vxl_affine_linear - Affine transform and resample with linear interpolation

This function transforms a voxel array using an affine transform,
resamples in the new coordinates using linear interpolation between
voxels, and stores the results as a newly allocated voxel array.  The
rank of the destination need not match the rank of the source; this
permits projections into a lower dimensional space and liftings into a
higher dimensional one.  However, the rank of the source must be at
least one.  (Allowing rank 0 adds no new capability and would have to
be specially handled.)  The transform must be specified as the
destination-to-source transformation, since it is used to transform
each destination voxel coordinate back into the source; this is the
opposite of what you might naively expect, so be careful to do it
right.  The number of rows in the matrix is assumed to be the same as
the source rank, and the number of columns is assumed to be the
destination rank; the C ordering of matrix elements is assumed.  The
type of the destination need not be the same of the source; this costs
nothing extra since all the intermediate calculations are done in
double or double complex anyway.

This implementation is written to be fast but not necessarily
comprehensible; it is verified against the vxl_affine_linear0
implementation.


ALGORITHM:  You are not expected to understand this.

All voxels in the destination can be divided into three sets: an umbra
of voxels whose interpolation neighborhoods in the source lie
completely inside the boundaries of the source; a penumbra of voxels
whose interpolation neighborhoods straddle the boundary; and all other
voxels (the anumbra), whose neighborhoods completely miss the source.
Voxels in the umbra can be computed with a fast algorithm that does
not need to check each neighbor against the boundary.  Voxels in the
penumbra must be computed with an algorithm that does check against
the boundary; such voxels are relatively uncommon.  All other voxels
can simply be set to zero.

Voxels in the umbra are relatively common and are computed using a
fast iterative vectorized algorithm which simulates the obvious
recursive algorithm for pairwise mixing (along each axis) of intensity
values at corners of the interpolation neighborhood.


FIXME: The interpolated value is arguably guaranteed to be within the
representable range of the destination type, but bips_round_cast checks
against the limits anyway.  The extra cost is about 4% on a Pentium II,
but you could eliminate it if you were really pushing for a little extra
performance.
-------------------------------------------------------------------------*/

int                              /* Returns non-zero on error */
vxl_affine_linear(
    voxel_array *dest,           /* Destination array */
    int type,                    /* Desired type of destination */
    int rank,                    /* Desired rank of destination */
    long *dimen,                 /* Desired dimensions of destination */
    voxel_array *src,            /* Source array to transform and resample */
    double *matrix,              /* Linear part of the affine transform */
    double *offset)              /* Constant part of the transform */
{
    int   srank;                 /* Rank of the source */
    int   stype;                 /* Type of the source */
    void *sdata;                 /* Data array of the source */
    int   wtype;                 /* Working type for interpolation */
    long   coord[VXL_MAX_RANK];  /* Voxel coords in destination */
    long   sstep[VXL_MAX_RANK];  /* Step size along each axis of source */
    double point[VXL_MAX_RANK];  /* Point coords in source */
    double delta[VXL_MAX_RANK];  /* Column increments to source position */
    double lower[VXL_MAX_RANK];  /* Lower limits of source array */
    double upper[VXL_MAX_RANK];  /* Upper limits of source array */
    long   *first;               /* (0,0,0) corner of neighborhood in src */
    double *alpha[VXL_MAX_RANK]; /* Interpolation fractions, per axis */
    void   *stack[VXL_MAX_RANK]; /* Stack of partially mixed rows */
    void   *value;               /* Row of interpolated values */
    int depth;                   /* Number of items in the stack */
    long neigh[VXL_MAX_RANK];    /* Relative indices of current neighbor */
    long  didx[4];               /* Dest indices for clipped source */
    long ncol;                   /* Number of columns in dest row */
    int isrc, idst;              /* Indices over source and destination axes */
    long icol;                   /* Index over columns of a dest row */
    unsigned char *drow;         /* Start of a destination row */
    void   *val, *stk0, *stk1;
    double *alf;
    long idx, nelem, ngh, *fst;
    int flag, i;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    srank = src->rank;
    stype = src->type;
    sdata = src->data;

    /* Verify that source and destination have the same rank */
    if (rank != srank)
        NZReturnError(ERR_INCOMP, 
                      "Source and destination ranks are different");

    /* If possible, use exact resampling algorithm; this requires that
       all transform coefficients are close enough to integers */
    flag = vxli_affine_exact(dest, type, rank, dimen,
			     src, matrix, offset, 5.0*DBL_EPSILON);
    if (flag == 0)
	return ERR_OKAY;
    else if (flag > 0)
        NZReturnLibError();

    /* Determine the working type used for internal computations */
    wtype = INTP_DOUBLE;
    if (exim_iscomplex(stype)) {
        if ( !exim_iscomplex(type) )
            NZReturnError(ERR_INCOMP, 
                          "Complex source requires a complex destination");
        wtype = INTP_DCOMPLEX;
    }

    /* Precompute step size along each source coordinate */
    sstep[rank-1] = 1;
    for (i = rank-1; i > 0; i--)
	sstep[i-1] = sstep[i] * src->dimen[i];

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    flag = vxl_alloc_array(dest, type, rank, dimen);
    if (flag)    NZReturnLibError();

    /* Allocate memory for a stack of partially mixed rows, the
        interpolation fractions, and the (0,0,0) corner of each
        neighborhood */
    ncol  = dimen[rank-1];
    first = mallock(ncol*sizeof(unsigned char *));
    stack[0] = mallock(rank*ncol*exim_sizeof_intype(wtype));
    alpha[0] = mallock(rank*ncol*sizeof(double));
    if (first == 0 || stack[0] == 0 || alpha[0] == 0)    NZReturnLibError();
    value = stack[0];
    for (i = 1; i < rank; i++) {
	stack[i] = stack[i-1] + ncol * exim_sizeof_intype(wtype);
	alpha[i] = alpha[i-1] + ncol;
    }

    /* Prepare point increments along last dimension of destination,
       plus lower and upper limits of source */
    for (isrc = 0; isrc < srank; isrc++) {
	delta[isrc] = matrix[isrc*rank+rank-1];
	lower[isrc] = 0.0                    + BoundaryFuzz;
	upper[isrc] = src->dimen[isrc] - 1.0 - BoundaryFuzz;
    }

    /* Loop over the rows of the destination */
    vxli_firstrow(rank, coord);
    do {

	/* Compute source point corresponding to first column in row */
	for (isrc = 0; isrc < srank; isrc++) {
	    point[isrc] = offset[isrc];
	    for (idst = 0; idst < rank-1; idst++) {
		point[isrc] += matrix[isrc*rank+idst] * (double)coord[idst];
	    }
	}

	/* Clip to determine pixels within the source */
	vxli_fuzzy_clip(srank, ncol, point, delta, lower, upper, 1.0, didx);

	/* Zero points completely outside the source */
        flag = bips_zero_range(0, didx[0], wtype, value, 1);
        if (flag)    NZReturnLibError();
        flag = bips_zero_range(didx[3], ncol, wtype, value, 1);
        if (flag)    NZReturnLibError();

	/* Use boundary-checking method for first penumbra */
	for (isrc = 0; isrc < srank; isrc++)
	    point[isrc] += didx[0] * delta[isrc];
	for (icol = didx[0]; icol < didx[1]; icol++) {
            val  = bips_locate(wtype, value, icol);
            flag = vxli_linear_point(wtype, val, src, point, sstep);
            if (flag)    NZReturnLibError();
	    for (isrc = 0; isrc < srank; isrc++)
		point[isrc] += delta[isrc];
	}

	/* Compute first corner and fractions for the umbra */
	for (icol = didx[1]; icol < didx[2]; icol++)
	    first[icol] = 0;
	for (isrc = 0; isrc < srank; isrc++) {
	    for (icol = didx[1]; icol < didx[2]; icol++) {
		idx = (long)point[isrc];
		first[icol] += sstep[isrc] * idx;
		alpha[isrc][icol] = point[isrc] - idx;
		point[isrc] += delta[isrc];
	    }
	}

        /* Use fancy vectorized algorithm to interpolate rows */
        nelem = didx[2] - didx[1];
        for (isrc = 0; isrc < srank; isrc++)
            neigh[isrc] = 0;
        isrc  = srank - 1;
        depth = 0;
        while (1) {

            if (isrc == srank-1) {

                /* Gather/mix at leaf nodes */
                stk0 = bips_locate(wtype, stack[depth], didx[1]);
                depth++;
                alf  = alpha[isrc]  + didx[1];
                fst  = first + didx[1];
                for (ngh = 0, i = 0; i < srank; i++)
                    ngh += neigh[i] * sstep[i];
                stk1 = bips_locate(stype, sdata, ngh);
                flag = bips_glerp(nelem, wtype, stk0, 1, alf, 1,
                                  stype, stk1, fst);
                if (flag)    NZReturnLibError();

                /* Step to next node to process */
                neigh[isrc] = 0;
                if (isrc == 0)           /* For rank == 1 */
                    break;
                else if ( neigh[isrc-1] == 0 )
                    neigh[isrc-1] = 1;
                else
                    neigh[--isrc]++;
            }

            else if (neigh[isrc] < 2) {
                isrc++;
            }

            else if (neigh[isrc] == 2) {

                /* Mix on an interior node */
                depth--;
                stk1 = bips_locate(wtype, stack[depth  ], didx[1]);
                stk0 = bips_locate(wtype, stack[depth-1], didx[1]);
                alf  = alpha[isrc]    + didx[1];
                flag = bips_lerp(nelem,  wtype, stk0, 1,
                                 alf, 1, wtype, stk1, 1);
                if (flag)    NZReturnLibError();

                /* Step to next node to process */
                if (isrc == 0)      /* Done if just mixed the root */
                    break;
                neigh[isrc] = 0;
                neigh[--isrc]++;
                if (neigh[isrc] < 2)
                    isrc = srank - 1;
            }

            else
                NZReturnError(ERR_ASSERT,
                              "Fast interpolation algorithm is broken");
        }

	/* Use boundary-checking method for second penumbra */
	for (icol = didx[2]; icol < didx[3]; icol++) {
            val  = bips_locate(wtype, value, icol);
            flag = vxli_linear_point(wtype, val, src, point, sstep);
            if (flag)    NZReturnLibError();
	    for (isrc = 0; isrc < srank; isrc++)
		point[isrc] += delta[isrc];
	}

	/* Copy row to destination, converting to destination type */
	drow = vxli_locate(dest, coord);
        if (drow == 0)    NZReturnLibError();
	flag = bips_round_cast(ncol, type, drow, 1, wtype, value, 1);
	if (flag)    NZReturnLibError();

    } while (vxli_nextrow(rank, dimen, coord));

    /* Free temporary storage and return */
    free(alpha[0]);
    free(stack[0]);
    free(first);
    return ERR_OKAY;
}



/*========================================================================
Functions for polynomial transforms and resampling
========================================================================*/



/*------------------------------------------------------------------------
vxl_polynomial_linear0 - Polynomial resampling with linear interpolation

This function transforms a voxel array using an polynomial transform,
resamples in the new coordinates using linear interpolation between
voxels, and stores the results as a newly allocated voxel array.  The
rank of the destination need not match the rank of the source; this
permits projections into a lower dimensional space and liftings into a
higher dimensional one.  However, the rank of the source must be at
least one.  The transform must be specified as the destination-
to-source transformation, since it is used to transform each
destination voxel coordinate back into the source; this is the
opposite of what you might naively expect, so be careful to do it
right.  The out degree of the transform must match the rank of the
source; the in degree must match the rank of the destination.  The
type of the destination need not be the same of the source; this costs
nothing extra since all the intermediate calculations are done in
double anyway.

The outcor agrument is a pointer to an array of nterm elements which
specifies the output coordinate to which that term contributes.  The
power argument is a pointer to an array of nterm*nin elements which
contains for each term the powers to which the input coordinates are
to be raised.  The coefs argument is a pointer to an array of nterm
elements which contains the coefficients for each term.

This implementation is simple but slow; it is intended primarily as a
reference implementation to verify faster but less comprehensible
implementations.  Since this is only a reference implementation, we
don't bother supporting complex types.

FIXME: This function should probably round rather than simply casting
for destinations of integral type.
-------------------------------------------------------------------------*/

int                              /* Returns non-zero on error */
vxl_polynomial_linear0(
    voxel_array *dest,           /* Destination array */
    int type,                    /* Desired type of destination */
    int rank,                    /* Desired rank of destination = nin */
    long *dimen,                 /* Desired dimensions of destination */
    voxel_array *src,            /* Source array to transform and resample */
    int nterm,                   /* Number of terms in polynomial transform */
    int *outcor,                 /* Output coordinate for each term */
    int *power,                  /* Table of powers for each term */
    double *coef)                /* Table of coefficients */
{
    int srank;                   /* Rank of the source = nout */
    long   coord[VXL_MAX_RANK];  /* Voxel coords in destination */
    double point[VXL_MAX_RANK];  /* Point coords in source */
    void *value=0;               /* Location of a single destination value */
    double vterm;                /* Value of a single term */
    int isrc, idst;              /* Indices over source and destination axes */
    int iterm;                   /* Index over terms in polynomial */
    int *ppow;                   /* Pointer over elements of power */
    int ipow;                    /* Index over powers of a variable */
    int flag;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC ||
	src->rank < 1 || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    srank  = src->rank;

    /* Allocate the destination array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    flag = vxl_alloc_array(dest, type, rank, dimen);
    if (flag)    NZReturnLibError();

    /* Loop over the destination voxels */
    vxli_firstvoxel(rank, coord);
    do {

        /* Compute corresponding position in the source */
        for (isrc = 0; isrc < VXL_MAX_RANK; isrc++)
            point[isrc] = 0.0;
        ppow = power;
        for (iterm = 0; iterm < nterm; iterm++) {
            vterm = coef[iterm];
            for (idst = 0; idst < rank; idst++) {
                for (ipow = *ppow++; ipow > 0; ipow--)
                    vterm *= coord[idst];
            }
            isrc  = outcor[iterm];
            point[isrc] += vterm;
        }

	/* Interpolate in source and save in destination */
        value = vxli_locate(dest, coord);
        flag  = vxl_linear_point(dest->type, value, src, point);
        if (flag)    NZReturnLibError();

    } while (vxli_nextvoxel(rank, dimen, coord) );

    return ERR_OKAY;
}



/*------------------------------------------------------------------------
vxl_polynomial_map - Generate coordinate map for a polynomial transform

This function generates a new array which contains the output
coordinates for the given polynomial transform over a notional source
image of the given rank and dimensions.  The rank of the result
actually returned is one higher than the notional source, since the
last dimension is used to range over the output coordinates.  The
result is always of type double.

FIXME: Reimplement using a faster algorithm.
-------------------------------------------------------------------------*/

int                              /* Returns non-zero on error */
vxl_polynomial_map(
    voxel_array *result,         /* Result array */
    int rank,                    /* Desired rank of notional dest array */
    long *dimen,                 /* Desired dimensions of destination */
    int nterm,                   /* Number of terms in the poly transform */
    int *outcor,                 /* Output coordinate of each term */
    int *power,                  /* Power multi-index for each term */
    double *coef)                /* Coefficient of each term */
{
    long  dimen2[VXL_MAX_RANK];  /* Dimensions of result */
    long   coord[VXL_MAX_RANK];  /* Voxel coords in source */
    double point[VXL_MAX_RANK];  /* Point coords in result */
    double *resdata;             /* Pointer into result data */
    int isrc, idst;              /* Indices over source and destination axes */
    double vterm;                /* Value of a single term */
    int iterm;                   /* Index over terms in polynomial */
    int *ppow;                   /* Pointer over elements of power */
    int ipow;                    /* Index over powers of a variable */
    int flag, i;

    /* Compute dimensions of the result array */
    for (i = 0; i < rank; i++)
        dimen2[i] = dimen[i];
    dimen2[i] = rank;

    /* Allocate the result array */
    if (result == 0 || result->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid result array");
    flag = vxl_alloc_array(result, INTP_DOUBLE, rank+1, dimen2);
    if (flag)    NZReturnLibError();

    /* Loop over the destination voxels */
    resdata = result->data;
    vxli_firstvoxel(rank, coord);
    do {

        /* Compute corresponding position in the source */
        for (isrc = 0; isrc < VXL_MAX_RANK; isrc++)
            point[isrc] = 0.0;
        ppow = power;
        for (iterm = 0; iterm < nterm; iterm++) {
            vterm = coef[iterm];
            for (idst = 0; idst < rank; idst++) {
                for (ipow = *ppow++; ipow > 0; ipow--)
                    vterm *= coord[idst];
            }
            isrc  = outcor[iterm];
            point[isrc] += vterm;
        }

        /* Save coordinates in result */
        for (i = 0; i < rank; i++)
            *resdata++ = point[i];

    } while (vxli_nextvoxel(rank, dimen, coord) );

    return ERR_OKAY;
}
