/*****************************************************************************
tkphoto.c - Pyvox wrapper for a Tkinter PhotoImage

Author: Paul Hughett

Copyright (c) 2003-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This file defines a Python type which is a wrapper for the Tkinter
PhotoImage type and which allows fast copying of Pyvox arrays in and
out, and also a Python extension module which allows creation and
manipulation of the new type.

FIXME: Add a TkphotoType object?

FIXME: Add methods to get/put pyvox arrays.
******************************************************************************/

static const char rcsid[] = "$Id: tkphoto.c,v 1.20 2005/12/29 18:33:15 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include <Python.h>      /* Must precede standard C headers */

#include <stdio.h>

#define NO_PROTOTYPES (!PROTOTYPES)   /* For tk.h */
#include <tk.h>

#include "bips.h"
#include "exim.h"
#include "voxel.h"
#include "pyvox.h"


/* Forward declarations for a few static functions and objects */
static PyObject *tkphoto_str(PyObject *self);
static PyTypeObject tkphoto_type;

/* Stop gcc from grumbling about missing prototypes */
void inittkphoto(void);



/*=========================================================================
tkphoto type - A wrapper around a Tk PhotoImage object
=========================================================================*/



/*--------------------------------------------------------------------------
tkphoto_object - Pyvox wrapper for a Tkinter PhotoImage object

This struct is a Pyvox wrapper around the Tkinter PhotoImage object and
supports fast copying in and out of Pyvox arrays.

FIXME: Is it safe to cache the interp, name, and handle here?
----------------------------------------------------------------------------*/

typedef struct {
    PyObject_HEAD               /* Header for all Python objects */
    PyObject *tkphoto;          /* Pointer to a Tkinter PhotoImage */
    Tcl_Interp *interp;         /* Interpreter it belongs to */
    Tk_PhotoHandle handle;      /* Its handle */
    char *name;                 /* Its name */
} tkphoto_object;



/*-------------------------------------------------------------------------
TkPhoto_Check - Test whether a Python object is a tkphoto object
-------------------------------------------------------------------------*/

static inline int
TkPhoto_Check(
    PyObject *ob)
{
    if (ob != 0 && ob->ob_type == &tkphoto_type)
	return 1;
    else
	return 0;
}



/*--------------------------------------------------------------------------
photo_get_info - Get information about a Tk PhotoImage

This function fills in a Tk_PhotoImageBlock structure and other
variables with information about a Tk PhotoImage identified by a
Tkinter PhotoImage object.
----------------------------------------------------------------------------*/

static void
photo_get_info(
    PyObject *tkphoto,             /* Tkinter PhotoImage to get info on */
    Tcl_Interp **interp,           /* Interpreter it belongs to */
    Tk_PhotoHandle *handle,        /* Its handle */
    char **name,                   /* Its name */
    Tk_PhotoImageBlock *info)      /* Info about the PhotoImage */
{
    PyObject *tkapp = 0;           /* Wrapper of the Tcl interpreter */
    PyObject *tkaddr = 0;          /* Address of the Tck interpreter */
    Tcl_Interp *tclinterp = 0;     /* Address of the Tcl interpreter */
    PyObject *tkname = 0;          /* Name of the PhotoImage */
    char *tname = 0;               /* Name of the PhotoImage */
    Tk_PhotoHandle tkhandle = 0;   /* Tk handle of the PhotoImage */

    /* Find the tcl interpreter in which this PhotoImage is defined */
    tkapp = PyObject_GetAttrString(tkphoto, "tk");
    tkaddr = PyObject_CallMethod(tkapp, "interpaddr", "");
    tclinterp = (Tcl_Interp *)PyInt_AsLong(tkaddr);
    if (interp != 0)
        *interp = tclinterp;

    /* Get the name of the PhotoImage */
    tkname = PyObject_GetAttrString(tkphoto, "name");
    tname = PyString_AsString(tkname);
    if (name != 0)
        *name = tname;

    /* Translate the name into a Tk handle */
    tkhandle = Tk_FindPhoto(tclinterp, tname);
    if (handle != 0)
        *handle = tkhandle;

    /* Get the desired information */
    if (info != 0)
        Tk_PhotoGetImage(tkhandle, info);

    return;
}



/*-----------------------------------------------------------------------
tkphoto_destroy - Destroy an instance of a tkphoto object

This function destroys an instance of a tkphoto object by decrementing
the reference count of the Tkinter PhotoImge and freeing the storage
associated with Pyvox Tkphoto wrapper.  It is normally called
automatically by the Python garbage collector when the reference count
of the object goes to zero.
-----------------------------------------------------------------------*/

static void
tkphoto_destroy(
    tkphoto_object *self)
{
    /* Report to the user for debugging */
    if (0)  /* DEBUG */
        fprintf(stderr, "Destroying a tkphoto object at %p\n", (void *)self);

    /* Decrement the ref count of the Tkinter PhotoImage */
    Py_DECREF(self->tkphoto);

    /* Free memory used by the tkphoto object itself */
    PyMem_DEL(self);

    return;
}



/*--------------------------------------------------------------------------
tkphoto_getimage - Get contents of a Tkinter PhotoImage as Pyvox array

This function creates a new Pyvox array and copies the (image)
contents of the Tkinter PhotoImage into it.  The new array is rank 3,
unsigned char, has the same height and width as the PhotoImage, and
contains red, green, blue, and alpha along the last dimension.

FIXME: This is probably not very robust.
---------------------------------------------------------------------------*/

static PyObject *
tkphoto_getimage(
    PyObject *self,
    PyObject *args)
{
    pyvox_array *result = 0;       /* Result: Contents as a Pyvox array */

    tkphoto_object *pytk = (tkphoto_object *)self;
                                   /* Self as a tkphoto object */
    Tk_PhotoImageBlock info;       /* Information on the PhotoImage */
    int rank = 3;                  /* Desired rank */
    long dimen[VXL_MAX_RANK];      /* Desired dimensions */
    long nelem = 0;                /* Number of elements in array */
    voxel_array *vres = 0;         /* Result as a voxel array */
    int flag = 0;

    set_error_context("tkphoto.getimage()");

    /* Parse the arguments */
    if ( !PyArg_ParseTuple(args, "") )
        return NULL;

    /* Get information on the PhotoImage */
    Tk_PhotoGetImage(pytk->handle, &info);

    /* Check for assumed dimensions */
    if (info.pixelSize != 4)
        ReturnPyvoxError(ERR_INCOMP, 
                         "Tkinter PhotoImage has an unexpected shape");

    /* Set desired dimensions */
    dimen[0] = info.height;
    dimen[1] = info.width;
    dimen[2] = 4;
    nelem = info.height * info.width * 4;

    /* Create and allocate a Pyvox array */
    result = parray_alloc(&vres, INTP_UCHAR, rank, dimen);
    if (result == 0)    ReturnPyvoxLibError();

    /* Copy image contents into the Pyvox array */
    flag = bips_copy(nelem, INTP_UCHAR, vres->data, 1, info.pixelPtr, 1);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)result;
}



/*------------------------------------------------------------------------
tkphoto_putimage - Copy Pyvox array into a tkphoto object

This tkphoto method copies the contents of a Pyvox array into the
image of the tkphoto object.  The pyvox array must be unsigned char;
there are four permissible shapes, with corresponding encodings:

   m x n        Gray scale
   m x n x 1    Gray scale
   m x n x 3    Red, green, blue
   m x n x 4    Red, green, blue, alpha

The position defaults to the upper left corner and the size to the
size of the Pyvox array zoomed and subsampled.  The zoom and subsamp
arguments may be either a single number (which applies to both x and
y) or a list or tuple containing the values for each dimension.  Note
that the dimensions are listed in the order (y,x) for vertical and
horizontal to be consistent with the treatment of coordinates
elsewhere in Pyvox.

FIXME: Modify to handle keyword arguments correctly.
-------------------------------------------------------------------------*/

static PyObject *
tkphoto_putimage(
    PyObject *self,
    PyObject *args)
{
    PyObject *array = 0;         /* Arg 1: Pyvox array to copy from */
    PyObject *pos = 0;           /* Arg 2: Position to copy to */
    PyObject *size = 0;          /* Arg 3: Size of region to copy to */
    PyObject *zoom = 0;          /* Arg 4: Zoom factor */
    PyObject *subsamp = 0;       /* Arg 5: Subsampling factor */

    tkphoto_object *pytk = (tkphoto_object *)self;
                                 /* Self as a tkphoto object */
    voxel_array *varray;         /* Pyvox array as a voxel array */
    Tk_PhotoImageBlock info;     /* Description of the source block */
    int rank;                    /* Number of dimensions in source block */
    int nbytes;                  /* Bytes per pixel */
    int xpos,  ypos;             /* Position of destination block */
    int xsize, ysize;            /* Size of destination block */
    int xzoom, yzoom;            /* How much to zoom while copying */
    int xsubsamp, ysubsamp;      /* How much to subsample while copying */
    int i;

    set_error_context("tkphoto.putimage()");

    /* Check that self is valid */
    if ( self == 0 )
      ReturnPyvoxError(ERR_INIT, "Self must not be null");

    /* Parse the arguments */
    if ( ! PyArg_ParseTuple(args, "O|OOOO", 
                            &array, &pos, &size, &zoom, &subsamp) )
        return NULL;

    /* Get information on the Pyvox array; check for correct shape */
    if ( ! PyvoxArray_Check(array, "source", &varray) )
        ReturnPyvoxLibError();
    if ( varray->type != INTP_UCHAR )
        ReturnPyvoxError(ERR_INCOMP, "Source must be type uchar");
    rank   = varray->rank;
    nbytes = varray->dimen[2];
    if ( ! (( rank == 2 ) ||
            ( rank == 3 && (nbytes == 1 || nbytes == 3 || nbytes == 4))) )
        ReturnPyvoxError(ERR_INCOMP, 
                         "Arg 1 is not the right shape for a Tk image");

    /* Set up the Tk info block describing the Pyvox array */
    info.pixelPtr = (unsigned char *)varray->data;
    info.width  = varray->dimen[1];
    info.height = varray->dimen[0];
    info.pitch  = nbytes * varray->dimen[1];
    info.pixelSize = nbytes;
    for (i = 0; i < 4; i++)
        info.offset[i] = i;

    /* Parse the position argument */
    if ( pos == 0 || pos == Py_None ) {
        ypos = xpos = 0;
    }
    else if ( PyTuple_Check(pos) || PyList_Check(pos) ) {
        ypos = PyInt_AsLong(PySequence_GetItem(pos, 0));
        xpos = PyInt_AsLong(PySequence_GetItem(pos, 1));
    }
    else
        ReturnPyvoxError(ERR_INCOMP, 
                         "Arg 2 must be a pair of integers, or omitted");

    /* Parse the zoom argument */
    if ( zoom == 0 || zoom == Py_None ) {
        yzoom = xzoom = 1;
    }
    else if ( PyInt_Check(zoom) ) {
        yzoom = xzoom = PyInt_AsLong(zoom);
    }
    else if ( PyTuple_Check(zoom) || PyList_Check(zoom) ) {
        yzoom = PyInt_AsLong(PySequence_GetItem(zoom, 0));
        xzoom = PyInt_AsLong(PySequence_GetItem(zoom, 1));
    }
    else
        ReturnPyvoxError(ERR_INCOMP, 
                         "Arg 4 must be an int, a pair of ints, or omitted");

    /* Parse the subsamp argument */
    if ( subsamp == 0 || subsamp == Py_None ) {
        ysubsamp = xsubsamp = 1;
    }
    else if ( PyInt_Check(subsamp) ) {
        ysubsamp = xsubsamp = PyInt_AsLong(subsamp);
    }
    else if ( PyTuple_Check(subsamp) || PyList_Check(subsamp) ) {
        ysubsamp = PyInt_AsLong(PySequence_GetItem(subsamp, 0));
        xsubsamp = PyInt_AsLong(PySequence_GetItem(subsamp, 1));
    }
    else
        ReturnPyvoxError(ERR_INCOMP, 
                         "Arg 4 must be an int, a pair of ints, or omitted");

    /* Parse the size argument; done last so that zoom and subsampling
       can be accounted for in the default. */
    if ( size == 0 || size == Py_None ) {
        ysize = varray->dimen[0] * yzoom / ysubsamp;
        xsize = varray->dimen[1] * xzoom / xsubsamp;
    }
    else if ( PyTuple_Check(size) || PyList_Check(size) ) {
        ysize = PyInt_AsLong(PySequence_GetItem(size, 0));
        xsize = PyInt_AsLong(PySequence_GetItem(size, 1));
    }
    else
        ReturnPyvoxError(ERR_INCOMP,
                         "Arg 3 must be a pair of integers, or omitted");

    /* Call Tk to do the real work */
    Tk_PhotoPutZoomedBlock(pytk->handle, &info, xpos, ypos, xsize, ysize, 
                           xzoom, yzoom, xsubsamp, ysubsamp);

    /* Return None */
    Py_INCREF(Py_None);
    clear_error_context();
    return Py_None;
}



/*------------------------------------------------------------------------
tkphoto_size - Get the size of a Pyvox tkphoto object

This function returns a tuple showing the size of the image data contained
in a Pyvox tkphoto object; the size is always height x width x 4, where the
last dimension contains red, green, blue, and alpha.

FIXME: This is probably not very robust.
-------------------------------------------------------------------------*/

static PyObject *
tkphoto_size(
    PyObject *self)
{
    tkphoto_object *pytk;          /* Self as a Pyvox tkphoto */
    Tk_PhotoImageBlock info;       /* Information about the PhotoImage */

    set_error_context("tkphoto.size");

    /* Get description of the data block contained in the PhotoImage */
    pytk = (tkphoto_object *)self;
    Tk_PhotoGetImage(pytk->handle, &info);

    /* Construct and return the size info */
    clear_error_context();
    return Py_BuildValue("(iii)", info.height, info.width, 4);
}



/*----------------------------------------------------------------------
tkphoto_getattr - Get an attribute of a tkphoto object

FIXME: The methods used below to implement the members of the
instance, and the list of members is not easily expandable to the
number of members that we are likely to get eventually.  In other
words, design and use an attribute table for this purpose.

FIXME: Set the error context as needed here.
----------------------------------------------------------------------*/

/* Declare the tkphoto_methods array; it will be defined later. */
static struct PyMethodDef tkphoto_methods[];

static PyObject *
tkphoto_getattr(
    PyObject *self,       /* Tkphoto object to get attribute of */
    char *attr)           /* Name of the attribute to get */
{
    /* Catch things that aren't tkphoto objects */
    if ( ! TkPhoto_Check(self) )
	ReturnPyvoxError(ERR_INIT, "this is not a tkphoto object");

    /* Return value of any requested member of the instance */
    if (strcmp(attr, "handle") == 0)
        return Py_BuildValue("l", ((tkphoto_object *)self)->handle);
    else if (strcmp(attr, "name") == 0)
        return Py_BuildValue("s", ((tkphoto_object *)self)->name);
    else if (strcmp(attr, "size") == 0)
	return tkphoto_size(self);
    else if (strcmp(attr, "tkphoto") == 0)
        return Py_BuildValue("O", ((tkphoto_object *)self)->tkphoto);

    /* Provide list of members available from this instance */
    if (strcmp(attr, "__members__") == 0)
	return Py_BuildValue("[s]", "size");

    /* Use the Python API to find methods associated with this type */
    return Py_FindMethod(tkphoto_methods, self, attr);
}



/*------------------------------------------------------------------------
tkphoto_print - Print (informal) representation of a tkphoto object

This function prints out an informal representation of a tkphoto
object on a specified FILE.  This representation contains enough
information to distinguish different objects, but not to reconstruct
the object when read back in.  The flags argument indicates the object
format to be used: 0 selects repr(), and Py_PRINT_RAW selects str().

FIXME: The repr() format is not actually correct to be read back in.
-------------------------------------------------------------------------*/

static int
tkphoto_print(
    tkphoto_object *self,        /* The object being printed */
    FILE *out,                   /* Where to print it */
    int flags)                   /* Provide repr() or str()? */
{
    set_error_context("print tkphoto");
    fprintf(out, "<Pyvox tkphoto object at %p, wrapping PhotoImage ", self);
    PyObject_Print(tkphoto_str((PyObject *)self), out, flags);
    fprintf(out, " >");
    clear_error_context();
    return 0;
}



/*-------------------------------------------------------------------------
tkphoto_str - Return informal string representation of a tkphoto object

This function returns the str() representation of the enclosed
PhotoImage object.  This makes it possible to supply the tkphoto
object rather than the enclosed PhotoImage as the image argument to
the canvas.create_image method.
---------------------------------------------------------------------------*/

static PyObject *
tkphoto_str(
    PyObject *self)                 /* Object to compute str() of */
{
    return PyObject_Str(((tkphoto_object *)self)->tkphoto);
}



/*------------------------------------------------------------------------
tkphoto_type - Python type descriptor for the tkphoto type

FIXME: This should probably be static unless used elsewhere.
------------------------------------------------------------------------*/

static PyTypeObject 
tkphoto_type = 
{

    /* Type head */
    PyObject_HEAD_INIT(&PyType_Type)
    0,                              /* Object size */
    "tkphoto",                      /* Name of type */
    sizeof(tkphoto_object),         /* Basic size of type */
    0,                              /* Item size */

    /* Standard methods */
    (destructor)  tkphoto_destroy,  /* Instance destructor */
    (printfunc)   tkphoto_print,    /* "print x" */
    (getattrfunc) tkphoto_getattr,  /* "instance.attr" */
    (setattrfunc) 0,                /* "instance.attr = foo" */
    (cmpfunc)     0,                /* compare instances */
    (reprfunc)    0,                /* string representation */

    /* Method suites for standard classes */
    (PyNumberMethods *)   0,        /* tp_as_number */
    (PySequenceMethods *) 0,        /* tp_as_sequence */
    (PyMappingMethods *)  0,        /* tp_as_mapping */

    /* Other methods */
    (hashfunc) 0,                   /* Hash function of an instance */
    (ternaryfunc) 0,                /* "instance()" */
    (reprfunc)    tkphoto_str,      /* "str(instance)" */

    /* See Include/object.h for still other methods */
};



/*--------------------------------------------------------------------------
tkphoto_methods - Table of tkphoto instance methods for getattr
--------------------------------------------------------------------------*/

static struct PyMethodDef
tkphoto_methods [] = {
    { "getimage",   tkphoto_getimage,  1,  "Contents of the image" },
    { "putimage",   tkphoto_putimage,  1,  "Copy array into image" },
    { NULL, NULL }
};



/*===========================================================================
tkphoto module - Python extension module for working with Tk
============================================================================*/



/*------------------------------------------------------------------------
tkinter_dict - Pointer to the method dictionary for Tkinter

This is used to lookup Tkinter methods that need to be called from
this module.
------------------------------------------------------------------------*/

PyObject *tkinter_dict = 0;



/*-----------------------------------------------------------------------
tkphoto_tkphoto - Create an instance of a tkphoto object

This function implements the pyvox.tkphoto() method, which creates a
brand-new instance of a Pyvox tkphoto object which warps the Tkinter
PhotoImage given as argument.  

FIXME: Figure out how to check that the argument is in fact a Tkinter
PhotoImage, and do so.

FIXME: Does set_error_context() work correctly in this context?  Does
this module depend on Pyvox?
------------------------------------------------------------------------*/

static
PyObject *
tkphoto_tkphoto(
    PyObject *self,
    PyObject *args,           /* Positional arguments */
    PyObject *kwargs)         /* Keyword arguments */
{
    PyObject *tkphoto = 0;    /* Arg 1: Tkinter PhotoImage object */

    tkphoto_object *pytk;     /* Pyvox tkphoto object to enwrap it */
    PyObject *tkapp = 0;      /* Wrapper of the Tcl interpreter */
    PyObject *tkaddr = 0;     /* Address of the Tck interpreter */
    PyObject *tkname = 0;     /* Name of the PhotoImage */

    set_error_context("tkphoto.tkphoto()");

    /* Get the positional arguments, if any */
    if ( !PyArg_ParseTuple(args, "|O", &tkphoto) )
	return NULL;

    /* Create a Python tkphoto object */
    pytk = (tkphoto_object *)PyObject_NEW(tkphoto_object, &tkphoto_type);
    if (pytk == 0)
        ReturnPyvoxError(ERR_MALLOC, "creation of tkphoto object failed");

    /* Case 1: Wrap an existing PhotoImage object */
    if (tkphoto) {
        pytk->tkphoto = tkphoto;
        Py_INCREF(tkphoto);
    }

    /* Case 2: Create and wrap a new PhotoImage object */
    /* FIXME: Assumes PyEval_Call... returns an owned reference */
    else {
        PyObject *tkfun=0, *args=0;

        tkfun   = PyDict_GetItemString(tkinter_dict, "PhotoImage");
        args    = Py_BuildValue("()");
        tkphoto = PyEval_CallObjectWithKeywords(tkfun, args, kwargs);
        Py_DECREF(args);
        pytk->tkphoto = tkphoto;
    }

    /* Find the tcl interpreter in which this PhotoImage is defined */
    tkapp = PyObject_GetAttrString(tkphoto, "tk");
    tkaddr = PyObject_CallMethod(tkapp, "interpaddr", "");
    pytk->interp = (Tcl_Interp *)PyInt_AsLong(tkaddr);

    /* Get the name of the PhotoImage */
    tkname = PyObject_GetAttrString(tkphoto, "name");
    pytk->name = PyString_AsString(tkname);

    /* Translate the name into a Tk handle */
    pytk->handle = Tk_FindPhoto(pytk->interp, pytk->name);

    /* DEBUG:  Report creation of a new tkphoto object */
    if (0) {
	fprintf(stderr, "Created new tkphoto object at %p\n", self);
	if (0) 
	    fprintf(stderr, "...reference count = %d\n",
		    ((PyObject *)pytk)->ob_refcnt); 
    }

    clear_error_context();
    return (PyObject *)pytk;
}



/*----------------------------------------------------------------------------
module_methods - Table of methods for the tkphoto module
-----------------------------------------------------------------------------*/

#define VK    (METH_VARARGS | METH_KEYWORDS)
#define PyCF  PyCFunction

static struct PyMethodDef 
module_methods[] = 
{
    {"tkphoto",  (PyCF)tkphoto_tkphoto,  VK, "Create a new tkphoto instance"},
    {NULL, NULL},
};



/*---------------------------------------------------------------------------
inittkphoto - Initialize the tkphoto module
----------------------------------------------------------------------------*/

void
inittkphoto(void)
{
    PyObject *module=0;
    PyObject *tkinter=0;

    /* Import the Tkinter module; remember its dictionary */
    tkinter = PyImport_ImportModule("Tkinter");
    if ( tkinter == 0 )
        Py_FatalError("Failed to import the Tkinter module");
    tkinter_dict = PyModule_GetDict(tkinter);

    /* Initialize the tkphoto module */
    module = Py_InitModule("tkphoto", module_methods);

    return;
}
