/*****************************************************************************
trans1.c - Generic Fourier and other transforms on voxel arrays

Author: Paul Hughett

Copyright (c) 2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit; it contains type-generic functions
for the Fourier transform.  Note that this file is not compiled
directly, but is included in transform.c twice with the macro ITYPE
set appropriately to generate functions for both float complex and
double comples FFT functions.  No header files are included here, since
they are expected to be included by transform.c; similarly, the functions
defined here are static.
******************************************************************************/

#if ITYPE == 3
static const char rcsid1[] = "$Id: trans1.c,v 1.1 2005/07/06 21:45:57 hughett Exp $ Release $Name: pyvox-0-72 $";
#endif



/*----------------------------------------------------------------------------
Define macros appropriate to the type requeste by ITYPE
----------------------------------------------------------------------------*/

#undef  TYPE
#undef  NAME

/* Define macros for ITYPE 3 => float complex */
#if ITYPE == 3
#define TYPE float complex
#define NAME(fcn) fcn##_fcomplex
#endif

/* Define macros for ITYPE 4 => double complex */
#if ITYPE == 4
#define TYPE double complex
#define NAME(fcn) fcn##_dcomplex
#endif



/*----------------------------------------------------------------------------
vxl_fftrow_xxx - Fast Fourier transform one row of a strided array

To compute the inverse FFT instead, set the weight array to its reciprocals,
and fail to divide by N.
-----------------------------------------------------------------------------*/

#undef  vxl_fftrow_xxx
#define vxl_fftrow_xxx   NAME(vxl_fftrow)

static int                    /* Returns nonzero on error */
vxl_fftrow_xxx(
    long nelem,               /* Number of elements; must be power of two */
    int nu,                   /* Log2(nelem) */
    long *permute,            /* Permutation table */
    TYPE *data,               /* Data to transform */
    long ds,                  /* Stride in the data array */
    TYPE *weight)             /* Powers of W; stride always = 1 */
{
    TYPE temp;                /* Temporary for exchanges */
    long n, np;               /* Index and bit-reversed index */
    int inu;                  /* Index over bits of index */
    long nstep;               /* Distance between dual points in butterfly */
    long npow;                /* Increment in power of W */
    long nn;                  /* Index within a block of the butterfly */
    long n0, n1, nw;          /* Indices to dual points and weight */
    TYPE xx;

    /* Permute the given data */
    for (n = 0; n < nelem; n++) {
        np = permute[n];
        if (n < np) {
            n0 =  n*ds;
            n1 = np*ds;
            temp     = data[n0];
            data[n0] = data[n1];
            data[n1] = temp;
        }
    }

    /* Loop over the powers of two */
    for (inu = 0; inu < nu; inu++) {

        /* Compute the butterfly for each power of two */
        nstep = 1<<inu;
        npow  = 1<<(nu-inu-1);
        n     = 0;
        /* fprintf(stderr, "%ld %ld: \n", nstep, npow);  // DEBUG */
        while (n < nelem) {
            for (nn = 0; nn < nstep; nn++) {
                n0 = n*ds;
                n1 = (n+nstep)*ds;
                nw = nn*npow;
                n  = n + 1;
                /* fprintf(stderr, "   %ld %ld %ld\n", n0, n1, nw); // DEBUG */
                xx = weight[nw] * data[n1];
                data[n1] = data[n0] - xx; 
                data[n0] = data[n0] + xx;
            }
            n = n + nstep;
        }
    }

    return 0;
}



/*--------------------------------------------------------------------------
vxl_fft_xxx - Fast Fourier transform of a voxel array

This function computes the fast Fourier transform in place, along
specified axes of a voxel array, returning the results in a new array
of the same rank and dimensions.  The input and output must both be
the complex type specified by the function name.  Setting naxes < 0 is
shorthand for all axes; setting naxes == 0 is a simple copy.  This
function should be called only from vxl_fft and assumes that that
function has already checked the validity of the voxel array and cast
it to the right type.

NOTE: The present code is a straightforward implementation and does
not attempt to achieve the maximum possible speed, except that it
tries to compute the weight and permutation arrays only once for each
axis.  It is generally designed to be efficient for multidimensional
images rather than long one-dimensional signals.

FIXME: Make more efficient.  The best candidate is probably to
transpose the array for each axis to be transformed, rather than using
a strided transform algorithm that uses cache inefficiently.

FIXME: Implement the Winograd algorithm, to support factors of 3, 5, ....

FIXME: Use the special real->complex algorithms when the input data is
real?
---------------------------------------------------------------------------*/

#define vxl_fft_xxx  NAME(vxl_fft)

static int                    /* Returns non-zero on error */
vxl_fft_xxx(
    voxel_array *dest,        /* Result of computation */
    int inverse,              /* Compute inverse transform instead? */
    int naxes,                /* Number of axes to transform on, or -1 */
    int *axes)                /* The axes to transform on */
{
    int drank = 0;                /* Rank of destination */
    long *ddimen = 0;             /* Dimensions of destination */
    int iax;                      /* Index over axes to transform */
    int iaxis;                    /* Current axis being transformed */
    long nelem;                   /* Number of elements on current axis */
    int  nu;                      /* Log2(nelem) */
    long ds;                      /* Stride in destination */
    long n;                       /* Index over elements in a row */
    long *permute = 0;            /* Permutation table for current axis */
    long index;                   /* Index to bit reverse */
    long xedni;                   /* Bit-reversed index */
    int  inu;                     /* Index over bits */
    TYPE *weight = 0;             /* Powers of W = exp(i*2*pi/n) */
    long coord[VXL_MAX_RANK];     /* Coordinates of current row */
    TYPE *drow = 0;               /* Location of current row */
    int flag;

    /* Get rank and dimensions of destination */
    drank  = dest->rank;
    ddimen = dest->dimen;

    /* Verify the axes specification; handle special cases */
    if (naxes == 0)
        return 0;
    if (axes == 0 || naxes  < 0)
        naxes = drank;

    /* Loop over the axes to be transformed */
    for (iax = 0; iax < naxes; iax++) {

        /* Get parameters for this axis */
        if (naxes < drank) 
            iaxis = axes[iax];
        else
            iaxis = iax;
        ds    = vxli_step(dest, iaxis);
        nelem = dest->dimen[iaxis];
        if (nelem < 1)
            NZReturnError(ERR_INIT, "Some axis has length less than 2");
        for (nu = 0; nu < 32; nu++) {
            if ( nelem == (1 << nu) )
                break;
            if ( nelem < (1 << nu) )
                NZReturnError(ERR_INIT, "Some axis is not a power of 2");
        }

        /* Compute the permutation vector */
        permute = mallock(nelem*sizeof(long));
        for (n = 0; n < nelem; n++) {
            index = n;
            xedni = 0;
            for (inu = 0; inu < nu; inu++) {
                xedni = (xedni<<1) | (index & 1);
                index = index >> 1;
            }
            permute[n] = xedni;
        }

        /* Compute the exponential weights */
        weight = mallock(nelem*sizeof(TYPE));
        weight[0] = 1.0;
        if (inverse)
            weight[1] = exp(I*2.0*PI/nelem);
        else
            weight[1] = exp(-I*2.0*PI/nelem);
        for (n = 2; n < nelem; n++)
            weight[n] = weight[n-1]*weight[1];

        /* Loop over rows, transforming each row */
        vxli_firstrow2(drank, coord);
        do {
        
            /* Transform each row */
            drow = (TYPE *)vxli_locate(dest, coord);
            if (drow == 0)    NZReturnLibError();
            flag = vxl_fftrow_xxx(nelem, nu, permute, drow, ds, weight);
            if (flag)    NZReturnLibError();

            /* Divide by N for forward transform only */
            if (!inverse)
                for (n = 0; n < nelem; n++)
                    drow[n*ds] /= nelem;

        } while (vxli_nextrow2(drank, ddimen, iaxis, coord));

        /* Free working storage */
        free(permute);
        free(weight);
    }

    return 0;
}
