/*****************************************************************************
transform.c - Fourier and other transforms on voxel arrays

Author: Paul Hughett

Copyright (c) 2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit and contains functions for various
image and signal transforms, starting with the Fourier transform.
It's not yet entirely clear which operations belong here, but a
tentative definition is image->image functions in which each output
voxel depends, at least potentially, on every input voxel.

FIXME: Given that definition, the chamfer transform belongs here.  But
we might want to include only invertible transforms.

NOTE: This entire file is conditioned on HAVE_COMPLEX_H being defined,
since the FFT is not much use without complex numbers.
******************************************************************************/

static const char rcsid[] = "$Id: transform.c,v 1.9 2005/10/26 18:30:29 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

/* Compile this file only if you have complex numbers */
#ifdef HAVE_COMPLEX_H

#include <complex.h>
#include <stdlib.h>

#ifdef HAVE_TGMATH_H
#include <tgmath.h>
#else
#include <math.h>
#endif

#include "errm.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"

/* FIXME: Move into a common header file? */
#define PI 3.14159265358979



/*--------------------------------------------------------------------------
Define type-specific FFT functions 
--------------------------------------------------------------------------*/

/* Float complex */
#define ITYPE 3
#include "trans1.c"

/* Double complex */
#undef  ITYPE
#define ITYPE 4
#include "trans1.c"



/*--------------------------------------------------------------------------
vxl_fft - Fast Fourier transform of a voxel array

This function computes the fast Fourier transform along specified axes
of a voxel array, returning the results in a new array of the same
rank and dimensions.  Each dimension must be a power of twol The input
may be any type; the type of the result is specified by the dtype
argument and must be either INTP_FCOMPLEX or INTP_DCOMPLEX.  Setting
naxes < 0 is shorthand for all axes; setting naxes == 0 is a simple
copy.

NOTE: The present code is a straightforward implementation and does
not attempt to achieve the maximum possible speed, except that it
tries to compute the weight and permutation arrays only once for each
axis.  It is generally designed to be efficient for multidimensional
images rather than long one-dimensional signals.

FIXME: Use the special real->complex algorithms when the input data is
real?
---------------------------------------------------------------------------*/

int                           /* Returns non-zero on error */
vxl_fft(
    voxel_array *dest,        /* Result of computation */
    int dtype,                /* Desired type of result */
    voxel_array *src,         /* Array to transform */
    int inverse,              /* Compute inverse transform instead? */
    int naxes,                /* Number of axes to transform on, or -1 */
    int *axes)                /* The axes to transform on */
{
    int flag;

    /* Check validity of source */
    /* FIXME: Require some minimum length on each axis? */
    if  ( src == 0 || 
          src->magic != VXL_MAGIC ||
          src->data == 0 ||
          exim_sizeof_intype(src->type) == 0 )
        NZReturnError(ERR_INIT, "Invalid source array");

    /* Case 1: Do FFT in float complex */
    if (dtype == INTP_FCOMPLEX) {
        flag = vxl_fcomplex(dest, src);
	if (flag)   NZReturnLibError();
	flag = vxl_fft_fcomplex(dest, inverse, naxes, axes);
	if (flag)   NZReturnLibError();
    }

    /* Case 2: Do FFT in double complex */
    else if (dtype == INTP_DCOMPLEX) {
        flag = vxl_dcomplex(dest, src);
	if (flag)   NZReturnLibError();
	flag = vxl_fft_dcomplex(dest, inverse, naxes, axes);
	if (flag)   NZReturnLibError();
    }
    /* Error otherwise */
    else
        NZReturnError(ERR_INIT, "FFT type must be complex");

    return 0;
}

#endif
