/****************************************************************************
voxel.c - Voxel Kit routines for volume image processing

Author:  Paul Hughett

Copyright (c) 1998-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  This particular files defines the core operations, plus
various non-core operations that don't yet fit anywhere else.

FIXME:  This file is getting large enough that it might make sense to split
it into separate files.  In any case, some rational order should be imposed
on the functions herein.

FIXME: Clarify exactly when data lengths and locations are specified in
bytes and when in elements.

FIXME:  Identify common validity checking steps and isolate as functions
or macros.  Q:  What are the efficiency tradeoffs between the two?
*****************************************************************************/

static const char rcsid[] = "$Id: voxel.c,v 1.160 2006/02/23 15:13:47 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include <ctype.h>
#include <float.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_TGMATH_H
#include <tgmath.h>
#else
#include <math.h>
#endif

#include "errm.h"
#include "memm.h"
#include "voxel.h"
#include "vxli.h"
#include "bips.h"


#define DEBUG_ALLOC  0     /* Trace array allocation/freeing? */



/**************************************************************************
                     Internal Utility Functions

                 For use within the Voxel Kit itself
**************************************************************************/


/*-------------------------------------------------------------------------
cmp_uchar - Compare two unsigned char values for qsort or bsearch

FIXME:  Consider promoting this to a general-purpose library function
somewhere.

FIXME: Commented out because not currently used anywhere.
-------------------------------------------------------------------------*/

#if 0
static int
cmp_uchar(
    const void *v1,
    const void *v2)
{
    const unsigned char *c1 = v1, *c2 = v2;
    if      (*c1 < *c2) return -1;
    else if (*c1 > *c2) return +1;
    else                return  0;
}
#endif


/*-------------------------------------------------------------------------
cmp_double - Compare two double values for qsort or bsearch

FIXME:  Consider promoting this to a general-purpose library function
somewhere.
-------------------------------------------------------------------------*/

static int
cmp_double(
    const void *v1,
    const void *v2)
{
    const double *c1 = v1, *c2 = v2;
    if      (*c1 < *c2) return -1;
    else if (*c1 > *c2) return +1;
    else                return  0;
}
   



/**************************************************************************
                    Array Structure Manipulations
**************************************************************************/


/*-------------------------------------------------------------------------
vxl_create_array - Create and initialize a voxel array

This function allocates memory for a voxel array, initializes it to
have no dimensions and to contain nothing, and returns a pointer to
the newly created voxel array.  The function vxl_destroy_array should
be used to destroy the voxel array and contents once it is no longer
useful.
---------------------------------------------------------------------------*/

voxel_array *                 /* Returns NULL on error */
vxl_create_array()
{
    voxel_array *dest;

    dest = mallock(sizeof(voxel_array));
    if (dest == 0)    ZReturnLibError();
    vxl_init_array(dest);

    return dest;
}



/*-------------------------------------------------------------------------
vxl_create_scalar - Create a scalar voxel array

This function allocates memory for a voxel array, initializes it to
contain a rank zero array (that is, a scalar array) containing a
specified scalar value of some given type, and returns a pointer to
that voxel array.  The function vxl_destroy_array should be used to
destroy the voxel array and contents once it is no longer useful.

FIXME: Add support for complex types.  However, this function doesn't
seem to be used anywhere in pyvox.
---------------------------------------------------------------------------*/

voxel_array *                  /* Returns zero on error */
vxl_create_scalar(
    int type,                  /* Type of the scalar */
    double value)              /* Value to initialize to */
{
    voxel_array *dest;
    int flag;

    if ( exim_iscomplex(type) )
        ZReturnError(ERR_UNIMP, "Complex types are not supported");
    dest = vxl_create_array();
    if (dest == 0)    ZReturnLibError();
    flag = vxl_alloc_array(dest, type, 0, NULL);
    if (flag)    ZReturnLibError();
    flag = exim_set_value(dest->data, type, 0, value);
    if (flag)    ZReturnLibError();

    return dest;
}




/*--------------------------------------------------------------------------
vxl_destroy_array - Destroy a dynamically allocated voxel array

This function destroys a voxel array created using vxl_create_array,
returning both it and any contents to the heap.  If you attempt to
destroy a voxel array not created by vxl_create_array or to access one
which has been destroyed, the results are unpredictable; if you are
lucky, your program will crash before it destroys any of your data.
The magic number is set to zero, to help detect attempted use after
destruction.

This function generates a warning if the object provided doesn't look
like a voxel array.  Depending on the settings in the error handler,
the warning may be printed out, silently ignored, or convert to an
error; the latter is translated here into a panic.  (The reason why it
is not treated as a normal error (and converted ultimately into a
Python exception) is that voxel arrays are mostly destroyed at times
that are random from the viewpoint of the user; that is, any error
message would usually be unrelated to the statement being executed.)
---------------------------------------------------------------------------*/

void 
vxl_destroy_array(
    voxel_array *dest)          /* Voxel array to be destroyed */
{
    int flag;

    /* If pointer is zero, just return silently */
    if (dest == 0)  return;

    if (dest->magic != VXL_MAGIC) {
	flag = set_warning(ERR_INIT, 
                           "Attempted to destroy something "
                           "that wasn't a voxel array");
        if (flag) errm_panic("Not a voxel array");
    }

    dest->magic = 0;
    if (dest->length > 0 && dest->data != 0)
	free(dest->data);
    free(dest);

    return;
}



/*-------------------------------------------------------------------------
vxl_init_array - Initialize a voxel array

Given a voxel_array structure, this function initializes its
parameters to default values, but does not allocate any memory for its
contents.  Any previous contents of the voxel array are lost forever.

This function differs from vxl_create_array in that the voxel array to be
initialized must already exist; it typically has been declared as voxel_array
in a program.  It is not yet clear which approach is the better.
---------------------------------------------------------------------------*/

void
vxl_init_array(
    voxel_array *dest)         /* Voxel array to initialize */
{
    int i;

    dest->magic    = VXL_MAGIC;
    dest->rank     = 0;
    dest->type     = 0;
    dest->nbytes   = 0;
    dest->length   = 0;
    dest->data     = 0;

    for (i = 0; i < VXL_MAX_RANK; i++) {
	dest->dimen[i] = 0;
        dest->origin [i] = 0.0;
	dest->spacing[i] = 1.0;
    }

    return;
}



/*-------------------------------------------------------------------------
vxl_store_scalar - Store a scalar value as a voxel array

Given a voxel_array structure, this function allocates or reallocates
storage to contain a single scalar of some given type and initializes
it to contain that scalar value.  The rank, type, nbytes, length, and
data fields are set appropriately.  The error field is set to zero.
Note that the voxel array must already have been initialized via
vxl_create_array or vxl_init_array before it is possible to store into
it.

FIXME: Rename as vxl_init_scalar for compatibility?  If so, you might
want to do all the initialization rather than assuming vxl_init_array
has been called.

FIXME: Add support for complex types.  However, this function doesn't
seem to be used anywhere in pyvox.
---------------------------------------------------------------------------*/

int                            /* Returns non-zero on error */
vxl_store_scalar(
    voxel_array *dest,         /* Voxel array to initialize */
    int type,                  /* Type of the scalar */
    double value)              /* Value to initialize to */
{
    int flag;

    if ( exim_iscomplex(type) )
        NZReturnError(ERR_UNIMP, "Complex types are not supported");

    /* Initialize the voxel array to default values */
    flag = vxl_alloc_array(dest, type, 0, NULL);
    if (flag)    NZReturnLibError();

    /* Store the scalar value */
    flag = exim_set_value(dest->data, type, 0, value);
    if (flag)    NZReturnLibError();

    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_alloc_array - Allocate a voxel array

Given a voxel_array structure, this function allocates or reallocates
the necessary data storage for an array of a given type and
dimensions.  The previous data, if any, is lost.  The rank, type,
dimen, nbytes, length, and data fields are set appropriately.  The
error field is set to zero.  The origin and spacing are set to zero
and one, thus making physical coordinates the same as index
coordinates; if other values are wanted, the caller must set them
separately.  

The case of a rank 0 array is handled correctly.  That is, space is
allocated for a single element of the given type; the dimen array is
ignored.

NOTE: It is possible to request an array so large that its byte length
does not fit into a long.  When this happens, the length calculation
overflows, causing less to be allocated than requested and generating
obscure segmentation faults or memory errors.  The cheapest way I've
been able to think of for checking this is to compute the byte length
in parallel using double precision.

FIXME: The error message for violating the above restriction is a
trifle obscure.  But specifying a number requires knowing the platform
limit.

FIXME:  If realloc fails, then the old data is lost forever.
---------------------------------------------------------------------------*/

int                            /* Returns non-zero on error */
vxl_alloc_array(
    voxel_array *dest,         /* Voxel array to initialize */
    int type,                  /* Desired type of array */
    int rank,                  /* Desired rank of array */
    long *dimen)               /* Array of dimensions */
{
    long len;                  /* Total length in bytes */
    double dlen;               /* Same thing, as double */
    int nbytes;                /* Bytes per datum */
    int i;

    /* Check for valid voxel array */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid voxel array");

    /* Allocate the necessary data area */
    nbytes = exim_sizeof_type(type);
    if (nbytes == 0)
	NZReturnError(ERR_UNDEF, "Invalid voxel type requested");
    dest->nbytes = nbytes;
    dlen = len = nbytes;
    for (i = 0; i < rank; i++) {
	dest->dimen[i] = dimen[i];
	len  *= dimen[i];
        dlen *= dimen[i];
        dest->origin [i] = 0.0;
	dest->spacing[i] = 1.0;
    }

    /* Check for impossibly huge request */
    if (dlen > LONG_MAX)
        NZReturnError(ERR_INIT,
                      "Requested array size exceeds platform maximum");

    /* Allocate the requested array */
    dest->length = len;
    dest->data = reallock(dest->data, len);
    if (dest->data == 0)    NZReturnLibError();

    /* Report for debugging */
    if (DEBUG_ALLOC)
        fprintf(stderr, "varray alloc: %p, %ld\n", dest->data, len);

    /* Initialize other fields in the vxl_array structure */
    dest->rank  = rank;
    dest->type  = type;

    return 0;
}




/*--------------------------------------------------------------------------
vxl_free_array - Free the contents of a voxel array

This function frees the contents of a voxel array, making that storage
available for other needs; the voxel_array itself is NOT freed, so
this function is suitable for statically allocated voxel arrays.  It
generates a fatal error if it appears that that voxel_array has never been
initialized.  But if the voxel array is merely empty, then it is left
empty.

This function generates a warning if the object provided doesn't look
like a voxel array.  Depending on the settings in the error handler,
the warning may be printed out, silently ignored, or convert to an
error; the latter is translated here into a panic.  (The reason why it
is not treated as a normal error (and converted ultimately into a
Python exception) is that voxel arrays are mostly destroyed at times
that are random from the viewpoint of the user; that is, any error
message would usually be unrelated to the statement being executed.)

Prior to 2002-09-05, this function would always generate a fatal error
for an object that didn't appear to be a voxel array.

Prior to 1999-03-18, this function would generate a fatal error if no
data area had been previously allocated; that is, if dest->data were
zero.
---------------------------------------------------------------------------*/

void
vxl_free_array(
    voxel_array *dest)        /* Voxel array to be freed */
{
    int flag;

    /* Check for a valid (properly initialized) voxel array */
    if (dest == 0 || dest->magic != VXL_MAGIC) {
	flag = set_warning(ERR_INIT, 
                           "Attempted to free something "
                           "that wasn't a voxel array");
        if (flag) errm_panic("Not a voxel array");
    }

    /* Report for debugging */
    if (DEBUG_ALLOC)
        fprintf(stderr, "varray free : %p\n", dest->data);

    /* Free the data area, if any */
    if (dest->data != 0)   free(dest->data);

    /* Mark this voxel array as empty or unallocated */
    dest->data  = 0;

    return;
}



/*------------------------------------------------------------------------
vxl_count - Compute number of voxels in a voxel array

Given an voxel array, this function computes the total number of
voxels or elements that it contains, after checking that the array is
valid and has been allocated.  If an error is found, the errm error
code and message are set before the function returns zero.

See vxli_count() to compute the number of elements without checking
validity.
--------------------------------------------------------------------------*/

long                    /* Returns number of elements, or zero on error */
vxl_count(
    voxel_array *src)   /* Voxel array to define dimensions */
{
    long count;         /* Count of elements */
    int i;

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
        ZReturnError(ERR_INIT, "Invalid or empty source array");

    /* Compute the desired count */
    count = 1;
    for (i = src->rank - 1; i >= 0; i--) 
	count *= src->dimen[i];

    return count;
}



/*-------------------------------------------------------------------------
vxl_copy_array - Create copy of a voxel array

This function creates an exact copy of a voxel array, the attributes
and contents of which are identical to the original array but which
shares no storage with it.  No error checking is done, except to
verify that the source has the right magic number; if the source
contains garbage, the copy will contain the same garbage.  If the
pointer to the source is zero, then a zero pointer will be returned.

There is another function vxl_copy which copies the contents of a
voxel array into another voxel array but leaves the attributes of the
destination unchanged.
--------------------------------------------------------------------------*/

voxel_array *                /* Returns NULL on error */
vxl_copy_array(
    voxel_array *src)        /* Voxel array to be copied */
{
    voxel_array *dest;       /* Copy of the voxel array struct */
    int flag, i;

    /* Just copy a NULL pointer */
    if (src == 0)
	return 0;

    /* Verify that the source has the right magic number */
    if (src->magic != VXL_MAGIC)
	ZReturnError(ERR_INIT,
                     "Attempted to copy something not a voxel array");

    /* Create a new array struct and copy the old one into it */
    dest = mallock(sizeof(voxel_array));
    if (dest == 0)    ZReturnLibError();
    dest->magic    = VXL_MAGIC;
    dest->rank     = src->rank;
    dest->type     = src->type;
    dest->nbytes   = src->nbytes;
    dest->length   = src->length;

    for (i = 0; i < VXL_MAX_RANK; i++) {
	dest->dimen  [i] = src->dimen[i];
        dest->origin [i] = src->origin[i];
	dest->spacing[i] = src->spacing[i];
    }

    /* Copy the data, if any */
    dest->data = 0;
    if (src->data != 0) {
	dest->data = mallock(src->length);
        if (dest == 0)    ZReturnLibError();
        flag = bips_copy(src->length, INTP_UCHAR, dest->data, 1,
			 src->data, 1);
	if (flag != 0)    ZReturnLibError();
    }

    /* Return a pointer to the copy */
    return dest;
}



/*--------------------------------------------------------------------------
vxl_reshape_array - Change rank and dimensions of a voxel array

This function changes the rank and dimensions of a voxel array while
keeping the data itself unchanged.  As a special case, specifying rank
= -1 will reduce the array to the minimum possible rank by removing
all dimensions along which there is only one element; for example, a 1
x 3 x 1 array would be reduced to a 3-element array with the same
elements.  (The type cannot be portably changed, so this is not
permitted.)  It is an error to try to change the total number of
elements.
---------------------------------------------------------------------------*/

int                       /* Returns non-zero on error */
vxl_reshape_array(
    voxel_array *dest,    /* Destination voxel array */
    int rank,             /* New rank */
    long *dimen)          /* New dimensions */
{
    int i, j;
    long len;

    /* Check validity of arguments */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid voxel array as destination");
    if (dest->data == 0)
	NZReturnError(ERR_ILLOP, "Attempt to reshape empty array");
    if (rank < -1 || rank > VXL_MAX_RANK)
	NZReturnError(ERR_ILLOP, "Invalid rank requested");

    /* Case 1: New dimensions are given explicitly */
    if (rank >= 0) {
   
        /* Check that new length equals old length */
        len = dest->nbytes;
        for (i = 0; i < rank; i++) {
            if (dimen[i] <= 0)
                NZReturnError(ERR_ILLOP, "Invalid dimension requested");
            len *= dimen[i];
        }
        if (len != dest->length)
            NZReturnError(ERR_ILLOP, "Attempted to change number of elements");

        /* Change the shape as requested */
        dest->rank = rank;
        for (i = 0; i < rank; i++)
            dest->dimen[i] = dimen[i];

	/* Set the origin and spacing to the default values */
        vxli_default_physical(dest);
    }

    /* Case 2: Just remove dimensions of length 1 */
    else {

        /* Scan dimensions, copying only those with length > 1 */
        for (i = 0, j = 0; i < dest->rank; i++ ) {
	    if (dest->dimen[i] > 1) {
                dest->dimen  [j] = dest->dimen  [i];
		dest->origin [j] = dest->origin [i];
		dest->spacing[j] = dest->spacing[i];
		j++;
	    }
        }
        dest->rank = j;
    }

    return ERR_OKAY;
}



/**************************************************************************
                        Voxel Access Functions
**************************************************************************/



/*------------------------------------------------------------------------
vxl_locate - Compute pointer to element with given subscripts

Given an voxel array, and subscript values for each coordinate, this
function computes an unsigned char pointer to the element selected by
the given coordinate values.  The length of the coords array must
equal the rank of the voxel array.  If the given coordinates are
outside the voxel array, then the error message is set and a null
pointer is returned.  The voxel array pointer given is checked for
validity, so this version is suitable for users outside the voxel
kit itself.  

SEE ALSO: vxli_locate, which omits to check the array validity and
should be used only within the Voxel Kit itself.
--------------------------------------------------------------------------*/

unsigned char *              /* Returns NULL on error */
vxl_locate(
    voxel_array *src,        /* Voxel array to define dimensions */
    long *coords)            /* Coordinate values */
{
    long step;               /* Step size for current coordinate */
    long offset;             /* Offset computed so far */
    unsigned char *point;    /* Pointer to the desired element */
    int i;
    
    /* Check for valid voxel array */
    if (src == 0 || src->magic != VXL_MAGIC)
	ZReturnError(ERR_INIT, "vxli_locate: Invalid voxel array.");

    /* Compute the desired offset */
    offset = 0;
    step = 1;
    for (i = src->rank - 1; i >= 0; i--) {
	if (coords[i] < 0 || coords[i] >= src->dimen[i])
            ZReturnError(ERR_INDEX, "Voxel array subscript out of range");
	offset += coords[i]*step;
	step *= src->dimen[i];
    }

    point = (unsigned char *)src->data + (src->nbytes)*offset;
    return point;
}



/*--------------------------------------------------------------------------
vxl_get_voxel - Get the value of a specified voxel (as a double)

This function returns the contents of the voxel at specified
coordinates as a double value; it is an error if the given coordinates
are not valid.

The use of this function is strongly discouraged, since double is not
truly a generic type (consider complex, long double).  The better
approach is to use vxl_locate or vxli_locate to generate a pointer to
the destination and call a generic type routine to store the desired
value there.  Thus the code here is commented out to detect any
residual uses without burning my bridges.
---------------------------------------------------------------------------*/

#if 0
int                                /* Returns non-zero on error */
vxl_get_voxel(
    double *value,                 /* (Out) Value from array */
    voxel_array *src,              /* Voxel array to examine */
    long *coords)                  /* Point coordinates */
{
    long offset;                   /* Position in vectorized array */
    int flag;

    /* Check for valid voxel array */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0) 
        NZReturnError(ERR_INIT, "Invalid or empty voxel array.");

    /* Not permitted for complex arrays */
    if ( exim_iscomplex(src->type) )
        NZReturnError(ERR_INCOMP, "Not permitted for complex arrays");

    /* Compute location of desired element; vxli_offset returns a 
       negative value if the coordinates are invalid. */
    offset = vxli_offset(src, coords);
    if (offset < 0)    NZReturnLibError();

    /* Extract and return the right element */
    flag = exim_get_value(value, src->data, src->type, offset);
    if (flag)    NZReturnLibError();
    return ERR_OKAY;
}
#endif



/*--------------------------------------------------------------------------
vxl_set_voxel - Set the value of a specified voxel (from a double)

This function sets the value of a voxel at specified coordinates to
a value given as a double and converted to the type of the voxel array
(using the usual C conversions).  It is an error if the given
coordinates are outside the array.

The use of this function is discouraged, since double is not truly
a generic type (e.g. complex, long double).  The better approach
is to use vxl_locate or vxli_locate to generate a pointer to the
destination and call a generic type routine to store the desired value
there.  Thus the code here is commented out to detect any residual uses 
without burning my bridges.
---------------------------------------------------------------------------*/

#if 0
int                               /* Returns non-zero on error */
vxl_set_voxel(
    voxel_array *dst,             /* Voxel array to modify */
    long *coords,                 /* Location to modify */
    double value)                 /* Value to convert and set */
{
    long offset;                  /* Position in vectorized array */
    int flag;

    /* Compute location of desired element; vxli_offset returns a 
       negative value if the coordinates are invalid. */
    offset = vxli_offset(dst, coords);
    if (offset < 0)
        NZReturnError(ERR_INDEX, "Subscript is out of range");

    /* Set the right element */
    flag = exim_set_value(dst->data, dst->type, offset, value);
    if (flag)    NZReturnLibError();
    
    return ERR_OKAY;
}
#endif



/*************************************************************************
                    Reading and Writing Slices

This section includes functions for reading or writing slices of a
voxel array, where a slice is defined by taking a subset of the
possible index values along each axis.  (Note that a Pyvox slice is
not the same as a Python slice.)  Possible options for each axis
include taking all possible index values, taking a set of specified
values (possibly only a single value), or taking an arithmetic
sequence of index values specified by a count, an initial value, and a
step.  Following the usual convention in C and Python (but not
Fortran, Matlab, or Octave), element indices start at 0.  It is a fatal
error if no indices are selected along any axis.

The nidx argument specifies how many index values to take along each
axis, and the index argument specifies the actual index values along
each axis.  If nidx[i] is positive, it is the actual number of index
values to take along axis i and index[i] points to a list of the
specific index values to take.  If nidx[i] is zero, then all indices
along axis i are to be taken, and index[i] is ignored.  If nidx[i] is
less than -1, then its absolute value specifies the actual number of
index values to take along axis i and index[i][0] and index[i][1]
specify the initial value and increment of an arithmetic sequence of
index values.  It is an error if nidx[i] is exactly -1; this is to
detect uses of the old and incompatible calling sequence described
below.  It is permitted to specify a slice which contains only a
single element, but this may not be efficient (and often requires
special handling in the algorithms).  

NOTE: Prior to 2002-04-29, the nidx and index arguments were defined
as follows.  The nidx argument specifies the number of index values to
take along each axis, and the index argument specifies the actual
index values.  There are two special values for nidx[i]: The value 0
means to take all the indices present along that axis; the
corresponding index array is ignored.  The value -1 means to take an
arithmetic sequence of indices; the corresponding index must contain
the starting value, limit value, and step in that order.  NOTE that
the given limit value is never actually selected; this is consistent
with the usual convention in C and Python but not with Matlab or
Octave.

This code is optimized for the case of large arrays; it will not be
particularly efficient at handling many small arrays.

FIXME: Establish a consistent nomenclature for rank, dimension, axis,
index, coordinate, etc.

The functions vxli_prep_slice and vxli_next_slice are wholly owned
subsidiaries of the public functions vxl_get_slice and vxl_set_slice;
their purpose is to encapsulate in one place some operations
common to the public functions, to make maintenance a little easier.
There is some rather messy code in these functions, to handle all the
possible cases without sacrificing too much efficiency; take care if
you decide to modify these functions.

ALGORITHMS: All of the actual data copying is done in one of the
optimized BIPS functions bips_copy, bips_gather, or bips_scatter; all
of the code that you see here is just to divide the labor into
suitable chunks for BIPS.  For efficiency, the algorithm divides the
axes of the source array into three sets.  The prime axis is the last
axis with more than one index value, and is processed within BIPS for
efficiency.  The leading axes are those which precede the prime axis,
and are handled by looping through the possibilities.  The trailing
axes are those which follow the prime axis and have only a single
index value to extract.  The sets of leading and trailing axes may be
empty.  If every axis selects only a single index, then the prime axis
is taken as the first axis.

Considerable care has been taken to ensure that the following special
cases are handled correctly: the source has rank 0; every axis has
only one index value selected; the set of leading or trailing axes is
empty; there is only one element in the selected slice.  If you modify
this code, make sure that it still handles these special cases.
*************************************************************************/



/*------------------------------------------------------------------------
vxli_prep_slice - Prepare a slice description for processing

This function translates a more or less human-readable slice description
(nidx and index) into the form which permits rapid execution using the
BIPS functions.  It is used by vxl_get_slice and vxl_put_slice and is
probably not useful anywhere else.  The meaning of the nidx and index
arguments is described in detail under "Reading and Writing Slices" above.
This code assumes that the array to be sliced has already been checked for 
validity; the nidx and index are not checked here.  If only one index value 
is specified on each axis or the array being sliced is rank 0, then prime 
is set to -1 to indicate this; note that this case often requires special 
handling in a fast, row-oriented algorithm.

ALGORITHM: Since this function is only used twice, it is made inline
despite its length.
------------------------------------------------------------------------*/

#define PAPRIM  10           /* Pre-allocated space for prime axis */

static inline 
long *                       /* Returns NULL on error */
vxli_prep_slice(
    voxel_array *slc,        /* The voxel array to be sliced */
    long *nidx,              /* What to do with each axis */
    long **index,            /* The indices for each axis */
    long *nmbr,              /* (out) Nmbr elements to xfer for each axis */
    long *init,              /* (out) Initial index for each axis */
    long *step,              /* (out) Index increment for each axis */
    long *scor,              /* (out) Initial coords for sliced array */
    long *ucor,              /* (out) Initial coords for unsliced array */
    int  *prime,             /* (out) Prime axis for BIPS */
    long *sidx)              /* (in/out) Offsets along prime axis */
{
    int rank;                /* Number of dimensions in sliced array */
    int iaxis;               /* Index over axes */
    int paxis;               /* Prime axis */
    long nidxp;              /* The value of nidx(*prime) */
    long ss;                 /* Stride along prime axis */
    long i;

    /* Handle rank 0 as a special case */
    rank = slc->rank;
    if (rank == 0) {
        *prime = -1;
        return sidx; }

    /* Validate arguments */
    if (nidx == 0)
        ZReturnError(ERR_ILLOP, "The nidx argument provided was zero");
    if (index == 0)
        ZReturnError(ERR_ILLOP, "The index argument provided was zero");
    for (iaxis = 0; iaxis < rank; iaxis++)
        if (nidx[iaxis] != 0 && index[iaxis] == 0)
            ZReturnError(ERR_ILLOP, "Null pointer found in the index array");
    if (sidx == 0)
        ZReturnError(ERR_ASSERT, "No preallocated storage was provided");

    /* Compute dimensions of the slice */
    for (iaxis = 0; iaxis < rank; iaxis++) {
        if ( nidx[iaxis] > 0 ) {
            /* Desired indices are given explicitly */
            nmbr[iaxis] = nidx[iaxis];
            init[iaxis] = index[iaxis][0];
            step[iaxis] = 1;
            for (i = 0; i < nmbr[iaxis]; i++)
                if ( index[iaxis][i] < 0 ||
                     index[iaxis][i] >= slc->dimen[iaxis] )
                    ZReturnError(ERR_INDEX, "Index outside sliced array");
        }
        else if ( nidx[iaxis] == 0 ) {
            /* Desired indices constitute the entire axis */
            nmbr[iaxis] = slc->dimen[iaxis];
            init[iaxis] = 0;
            step[iaxis] = 1;
        }
        else if ( nidx[iaxis] < -1 ) {
            /* Desired indices are given as an arithmetic sequence */
            long last;
            nmbr[iaxis] = - nidx[iaxis];
            init[iaxis] = index[iaxis][0];
            step[iaxis] = index[iaxis][1];
            last = init[iaxis] + (nmbr[iaxis] - 1) * step[iaxis] ;
            if ( init[iaxis] < 0 || init[iaxis] >= slc->dimen[iaxis] ||
                 last < 0 || last >= slc->dimen[iaxis] )
                ZReturnError(ERR_INDEX, "Index outside sliced array");
        }
        else
            /* nidx[i] == -1 indicates the old calling sequence */
            ZReturnError(ERR_ASSERT, "Using an obsolete calling sequence");
        if ( nmbr[iaxis] < 1 )
            ZReturnError(ERR_INDEX, "Some slice axis has non-positive count");
    }

    /* Set initial coordinates for sliced and unsliced arrays */
    for (iaxis = 0; iaxis < rank; iaxis++) {
        scor[iaxis] = init[iaxis];
        ucor[iaxis] = 0;
    }

    /* Identify the prime axis */
    for ( paxis = rank - 1;
          paxis >= 0 && nmbr[paxis] == 1;
          paxis-- ) {};
    *prime = paxis;
    if ( paxis < 0 )
        return sidx;

    /* Precompute offsets along prime axis for BIPS.  In addition,
       explicit coordinates on the prime axis use scatter or gather,
       which requires that the starting source coordinate on that axis
       be forced to zero. */
    nidxp = nidx[paxis];
    if (nidxp > 0) {
        scor[paxis] = 0;
        if (nidxp > PAPRIM) { 
            sidx = mallock(nidxp*sizeof(long));
            if (sidx == 0)    ZReturnLibError();
        }
        ss = vxli_step(slc, paxis);
        for (i = 0; i < nidxp; i++)
            sidx[i] = ss * index[paxis][i];
    }
    else {
        sidx[0] = vxli_step(slc, paxis) * step[paxis];
    }

    return sidx;
}



/*------------------------------------------------------------------------
vxli_next_slice - Step to the next set of leading indices

This function increments the current values of the leading coordinates
in the sliced and unsliced arrays; it returns 1 until the last set has
been processed, whereupon it returns 0.  This function does not handle
the case prime == -1, which is normally handled as a special case
outside the loop over rows.

NOTE that this code has been verified to work correctly for the
special cases: prime == 0; and nmbr[prime] == 1.
------------------------------------------------------------------------*/

static inline
int
vxli_next_slice(
    long *nidx,              /* How to scan each axis */
    long **index,            /* Indices for each axis */
    long *nmbr,              /* Number of indices on each axis */
    long *init,              /* Initial index on each axis */
    long *step,              /* Step size on each axis */
    long *scor,              /* Current indices for sliced array */
    long *ucor,              /* Current indices for unsliced array */
    int  prime)              /* Prime axis */
{
    int ilead;               /* Index over leading axes */

    for (ilead = prime-1; ilead >= 0; ilead--) {
        ucor[ilead] += 1;
        if ( ucor[ilead] < nmbr[ilead] ) {
            if ( nidx[ilead] > 0 )
                scor[ilead] = index[ilead][ucor[ilead]];
            else
                scor[ilead] += step[ilead];
            break; }
        else {
            ucor[ilead] = 0;
            scor[ilead] = init[ilead]; }
    }

    return ( ilead > 0 || ucor[0] != 0 );
}



/*------------------------------------------------------------------------
vxli_describe_slice - Dump a slice description for debugging

This function prints out (on stderr) a human-readable description of
the slice defined by a given nidx and index.
-------------------------------------------------------------------------*/

static void
vxli_describe_slice(
    voxel_array *slc,           /* Voxel array being sliced */
    long *nidx,                 /* What to do on each axis */
    long **index)               /* Indices for each axis */
{
    int i, j;

    fprintf(stderr, "Slicing rank-%ld array at %p\n", slc->rank, (void *)slc);

    for (i = 0; i < slc->rank; i++) {
        fprintf(stderr, "   %ld:", nidx[i]);
        if (nidx[i] > 0) {
            for (j = 0; j < nidx[i]; j++)
                fprintf(stderr, "  %ld", index[i][j]);
        }
        else if (nidx[i] < 0) {
            for (j = 0; j < 3; j++)
                fprintf(stderr, "  %ld", index[i][j]);
        }
        else /* if (nidx[i] == 0) */
            {}
        fprintf(stderr, "\n");
    }

    return;
}



/*------------------------------------------------------------------------
vxl_get_slice - Extract a slice from a voxel array

This function creates a new voxel array containing a subset of the
data from some existing voxel array; the data extracted is specified
by giving the index values to be included along each dimension.  The
rank of the output array is the same as the rank of the input, even if
some of the dimensions have length 1.  The meaning of the nidx and
index arguments is described in detail under "Reading and Writing
Slices" above.

Selecting only one element along an axis does not eliminate that axis
and reduce the rank of the result.  In particular, it is permissible
to select only one index value along each axis, but the result is
still an array of rank equal to the source.

If the source array has rank zero, then the destination array is
simply a copy of the source; the slice selection parameters are
ignored.
------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_get_slice(
    voxel_array *dst,           /* Destination array */
    voxel_array *src,           /* Source array */
    long *nidx,                 /* Number of indices along each axis */
    long **index)               /* Actual indices to take along each axis */
{
    int type;                   /* Data type of source and destination */
    int rank;                   /* Rank of source and destination */
    int prime;                  /* Prime axis */
    long nmbr[VXL_MAX_RANK];    /* Dimensions of the destination array */
    long dcor[VXL_MAX_RANK];    /* Current index along each axis in dest */
    long init[VXL_MAX_RANK];    /* Initial value along each source axis */
    long step[VXL_MAX_RANK];    /* Increment along each source axis */
    long scor[VXL_MAX_RANK];    /* Current index along each axis in source */
    long pasidx[PAPRIM];        /* Pre-allocation for a short sidx */
    long *sidx;                 /* Desired indices times appropriate spacing */
    unsigned char *drow;        /* Start of current row in destination */
    unsigned char *srow;        /* Start of current row in source */
    long ddel;                  /* Row-to-row increment in destination */
    long pnidx, pnmbr;          /* nidx[prime] and nmbr[prime] */
    int flag;

    /* (DEBUG) Print out the slice description */
    if (0)
        vxli_describe_slice(src, nidx, index);

    /* Validate the source and destination arrays */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    type = src->type;
    rank = src->rank;
    if (dst == 0 || dst->magic != VXL_MAGIC)
        NZReturnError(ERR_INIT, "Invalid destination array");

    /* Prepare slice description for fast processing */
    sidx = vxli_prep_slice(src, nidx, index, nmbr, init, step, 
                           scor, dcor, &prime, pasidx);
    if (sidx == 0)    NZReturnLibError();

    /* Allocate the destination array */
    flag = vxl_alloc_array(dst, type, rank, nmbr);
    if (flag)    NZReturnLibError();

    /* Handle a single-element slice as a special case */
    if ( prime == -1 ) {
        srow = vxli_locate(src, scor);
        if (srow == 0)    NZReturnLibError();
        drow = dst->data;
        flag = bips_copy(1, type, drow, 1, srow, 1);
        if (flag)    NZReturnLibError();
        if (sidx != pasidx) free(sidx);
        return ERR_OKAY;
    }

    /* Setup for the loop over leading indices */
    pnidx = nidx[prime];
    pnmbr = nmbr[prime];
    drow  = dst->data;
    ddel  = pnmbr*exim_sizeof_intype(type);

    /* Loop over the possible values of the leading indices */
    do {
        /* Copy values using BIPS */
        /* FIXME: Already checked that indices are within the array */
        srow  = vxli_locate(src, scor);
        if (srow == 0)    NZReturnLibError();
        if ( pnidx > 0 ) {
            flag = bips_gather(pnmbr, type, drow, 1, srow, sidx); }
        else {
            flag = bips_copy  (pnmbr, type, drow, 1, srow, sidx[0]); }
        if (flag)    NZReturnLibError();
        drow += ddel;
    }
    while (vxli_next_slice(nidx, index, nmbr, init, step, scor, dcor, prime));

    /* Release temporary array, if used */
    if (sidx != pasidx) 
        free(sidx);

    return ERR_OKAY;
}



/*------------------------------------------------------------------------
vxl_set_slice - Assign values to a slice in a voxel array

This function assigns values from some voxel array to a slice within
another voxel array; to elements to be assigned to are specified data
from some existing voxel array; the data extracted is specified by
giving the index values to be included along each dimension.  The
source may be either a scalar array, in which case the single value
contained therein is broadcast to all the destination elements, or a
non-scalar array, in which case it must contain exactly the number of
elements needed by the destination slice.  The rank and shape of a
non-scalar source array are ignored; the rank and shape of the
destination are unchanged.  The meaning of the nidx and index
arguments is described in detail under "Reading and Writing Slices"
above.
------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_set_slice(
    voxel_array *dst,           /* Destination array */
    long *nidx,                 /* Number of indices along each axis */
    long **index,               /* Actual indices to take along each axis */
    voxel_array *src)           /* Source array */
{
    int  stype, dtype;          /* Data type of source and destination */
    int  srank, drank;          /* Rank of source and destination */
    long slen, dlen;            /* Length of source and destination */
    int  prime;                 /* Prime axis */
    long nmbr[VXL_MAX_RANK];    /* Number of elements to take on each axis */
    long dcor[VXL_MAX_RANK];    /* Current index along each axis in dest */
    long init[VXL_MAX_RANK];    /* Initial value along each source axis */
    long step[VXL_MAX_RANK];    /* Increment along each source axis */
    long scor[VXL_MAX_RANK];    /* Current index along each axis in source */
    long padidx[PAPRIM];        /* Pre-allocation for a short didx */
    long *didx = 0;             /* Element offsets within destination row */
    unsigned char *drow;        /* Start of current row in destination */
    unsigned char *srow;        /* Start of current row in source */
    long ss = 1, ds = 1;        /* Stride within a source or dest row */
    long sdel = 0;              /* Row-to-row increment in source array */
    unsigned char *buff = 0;    /* Row buffer for type conversions */
    long pnidx, pnmbr;          /* nidx[prime] and nmbr[prime] */
    int mode;                   /* Mask for: scatter => +2; cast => +1 */
    int i, flag = 0;

    /* (DEBUG) Print out the slice description */
    if (0)
        vxli_describe_slice(dst, nidx, index);

    /* Validate the source array */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    stype = src->type;
    srank = src->rank;
    slen  = 1;
    for (i = 0; i < srank; i++)
        slen *= src->dimen[i];

    /* Validate the destination array */
    if (dst == 0 || dst->magic != VXL_MAGIC || dst->data == 0)
        NZReturnError(ERR_INIT, "Invalid or empty destination array");
    dtype = dst->type;
    drank = dst->rank;

    /* Prepare slice description for fast processing */
    didx = vxli_prep_slice(dst, nidx, index, nmbr, init, step, 
                           dcor, scor, &prime, padidx);
    if (didx == 0)    NZReturnLibError();

    /* Handle single element destination as a special case */
    if (prime == -1 && slen == 1) {
        drow = vxli_locate(dst, dcor);
        if (drow == 0)    NZReturnLibError();
        srow = src->data;
        flag = bips_cast(1, dtype, drow, 1, stype, srow, 1);
        if (flag)    NZReturnLibError();
        return ERR_OKAY; 
    }
    else if (prime == -1)
        NZReturnError(ERR_INCOMP,
                      "Length mismatch between source and destination slice");

    /* Verify length compatibility; set source stride */
    for (dlen  = 1, i = 0; i < drank; i++)
        dlen *= nmbr[i];
    if (slen == 1) {
        ss = 0; 
        sdel = 0; }
    else if (slen == dlen) {
        ss = 1;
        sdel = nmbr[prime]*exim_sizeof_intype(stype);
    }
    else
        NZReturnError(ERR_INCOMP,
                      "Length mismatch between source and destination slice");

    /* Allocate buffer for type conversions, if needed */
    if (dtype != stype) {
        buff = mallock(nmbr[prime]*exim_sizeof_intype(dtype));
        if (buff == 0)    NZReturnLibError();
    }

    /* Set up remaining parameters for the loop */
    pnidx = nidx[prime];
    pnmbr = nmbr[prime];
    ds    = didx[0];
    mode  = (stype != dtype) ? 1 : 0;
    mode += (pnidx > 0) ? 2 : 0;

    /* Loop over the possible values of the leading indices */
    srow  = src->data;
    do {
        /* Copy values using BIPS */
        /* FIXME: Already checked that indices are within dest */
        drow = vxli_locate(dst, dcor);
        if (drow == 0)    NZReturnLibError();
        switch (mode) {
        case 0:    /* No scatter, no casting */
            flag = bips_copy(pnmbr, stype, drow, ds, srow, ss);
            break;
        case 1:    /* No scatter, type casting */
            flag = bips_cast(pnmbr, dtype, drow, ds, stype, srow, ss);
            break;
        case 2:    /* Scatter, no casting */
            flag = bips_scatter(pnmbr, stype, drow, didx, srow, ss);
            break;
        case 3:    /* Scatter and type casting */
            /* FIXME: Inefficient if ss == 0 */
            flag  = bips_cast(pnmbr, dtype, buff, 1, stype, srow, ss);
            flag |= bips_scatter(pnmbr, dtype, drow, didx, buff, 1);
            break;
        default:
            NZReturnError(ERR_ASSERT, "Invalid mode in vxl_set_slice");
        }
        if (flag)    NZReturnLibError();
        srow += sdel;
    }
    while (vxli_next_slice(nidx, index, nmbr, init, step, dcor, scor, prime));

    /* Release temporary array, if used */
    if (didx != padidx && didx != 0) 
        free(didx);
    if (buff != 0)
        free(buff);

    return ERR_OKAY;
}



/**************************************************************************
                         Vector Operations
***************************************************************************/


/*-------------------------------------------------------------------------
vxl_cross - Cross product of 3-d vectors

This function computes the cross product of two real vectors in three
dimensions.  For convenience, the exact shape of the arguments does
not matter except that each must have exactly three elements.  The
type and shape of the output are the same as the first argument.  Not
defined for complex types.

FIXME: Generalize to dimensions other than 3?

FIXME: Modify to allow single argument in 2-d?
--------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_cross(
    voxel_array *dest,          /* Destination */
    voxel_array *src1,          /* First argument */
    voxel_array *src2)          /* Second argument */
{
    double u[3], v[3], w[3];    /* Elements of src1, src2, and dest */
    int flag;

    /* Check validity of first argument */
    if ( src1 == 0 || 
	 src1->magic != VXL_MAGIC || 
	 src1->data == 0 ||
	 vxli_count(src1) != 3 )
	NZReturnError(ERR_INIT, "Argument 1 is empty or invalid array");

    /* Check validity of second argument */
    if ( src2 == 0 || 
	 src2->magic != VXL_MAGIC || 
	 src2->data == 0 ||
	 vxli_count(src2) != 3 )
	NZReturnError(ERR_INIT, "Argument 2 is empty or invalid array");

    /* Check that arguments are both real */
    if ( exim_iscomplex(src1->type) || exim_iscomplex(src2->type) )
        NZReturnError(ERR_INIT, "Not defined for complex types");

    /* Allocate destination to match first argument */
    flag = vxl_alloc_array(dest, src1->type, src1->rank, src1->dimen);
    if (flag)    NZReturnLibError();

    /* Retrieve argument values as double */
    flag = bips_double(3, u, 1, src1->type, src1->data, 1);
    if (flag)    NZReturnLibError();
    flag = bips_double(3, v, 1, src2->type, src2->data, 1);
    if (flag)    NZReturnLibError();

    /* Compute cross product in double */
    w[0] = u[1]*v[2] - v[1]*u[2];
    w[1] = u[2]*v[0] - v[2]*u[0];
    w[2] = u[0]*v[1] - v[0]*u[1];

    /* Store result in destination */
    flag = bips_cast(3, dest->type, dest->data, 1, INTP_DOUBLE, w, 1);
    if (flag)    NZReturnLibError();

    return 0;
}




/***************************************************************************
                           Image Read and Write
***************************************************************************/


/*---------------------------------------------------------------------------
vxl_read_uchar_data - Read external uchar data and convert to a voxel array

Given a stream, this function reads unsigned char data from that stream
and converts it to a voxel array.  The rank of the voxel array and the
size of each dimension must be specified by the caller.  It is a fatal
error if the stream runs out of data before the voxel array is filled.

In most applications, it will be preferable to use vxl_read_raw_data,
since that function automatically converts the data from external to
internal formats; this function is appropriate only when the data are
known to be in the unsigned char format, or when the external-to-internal
conversion must be done after the data are read.
----------------------------------------------------------------------------*/

int                         /* Returns non-zero on error */
vxl_read_uchar_data(
    voxel_array *dest,      /* Destination for voxel data */
    FILE *src,              /* Stream to read data from */
    int rank,               /* Rank (number of dimensions) */
    long *dimen)            /* Array of dimensions */
{
    long count;             /* Number of bytes read from input file */
    int flag;

    /* Initialize the vxl_array structure; allocate memory */
    flag = vxl_alloc_array(dest, INTP_UCHAR, rank, dimen);
    if (flag)    NZReturnLibError();

    /* Read the data (no conversion is needed) */
    count = fread(dest->data, 1, dest->length, src);
    if (count < dest->length)
	NZReturnError(ERR_EOF, "Premature end of data file");    

    return ERR_OKAY;
}



/*------------------------------------------------------------------------
vxl_write_uchar_data - Write voxel array as external uchar data stream

Given an unsigned char volume image, this function writes the voxels of
that image as a stream of unsigned characters.

In most applications, it will be preferable to use vxl_write_raw_data,
since that function automatically converts the data from internal to
external formats; this function is appropriate only when the data are
known to be in the unsigned char format.
------------------------------------------------------------------------*/

int                      /* Returns non-zero on error */
vxl_write_uchar_data(
    FILE *dest,          /* Stream to write to */
    voxel_array *src)    /* Volume image to write */
{
    long count;          /* Number of voxels in the image */

    /* Validate arguments */
    if (dest == 0)
	NZReturnError(ERR_FILE, "Invalid output stream");
    if (src == 0 || src->magic != VXL_MAGIC || 
        src->data == 0 || src->type != INTP_UCHAR)
	NZReturnError(ERR_INIT, "Invalid or empty source array");

    /* Write out the data */
    count = fwrite(src->data, 1, src->length, dest);
    if (count < src->length)
	NZReturnError(ERR_FILE, "Error while writing data");    
    
    return ERR_OKAY;
}



/*------------------------------------------------------------------------
vxl_read_raw_data - Read raw data file and convert to voxel array

This function reads raw image data from a specified stream in some
specified external format, converts it on the fly to a specified
internal format (or the default "natural" format corresponding to the
external format), and stores it as a voxel array.  The rank and
dimensions of the data are specified by the caller.  If the internal
type is given as INTP_NONE, then the "natural" internal type
corresponding to the external type will be used.  The stream is assumed
to have already been positioned at the right place.  It is a fatal error
if the stream runs out of data before the voxel array is filled.
------------------------------------------------------------------------*/

int                                 /* Returns non-zero on error */
vxl_read_raw_data(
    voxel_array *dest,              /* Where to store the data input */
    FILE *src,                      /* File to read the data from */
    int rank,                       /* Number of dimensions */
    long *dimen,                    /* Dimensions themselves */
    int intype,                     /* Desired internal format */
    int extype,                     /* Format of the external data */
    int bigend)                     /* Big-endian? */
{
    int exbytes;                    /* Bytes per external datum */
    int inbytes;                    /* Bytes per internal datum */
    size_t count;                   /* Number of data to read */
    size_t nbuff = 1024;            /* Number of data per buffer */
    unsigned char *buff;            /* Buffer for input data */
    size_t nget;                    /* Number of data to get next */
    size_t ngot;                    /* Number of data actually gotten */
    unsigned char *data;            /* Data array in destination */
    int i, flag;

    /* Verify the external type; find number of bytes per datum */
    exbytes = exim_sizeof_type(extype);
    if (exbytes == 0)
	NZReturnError(ERR_UNDEF, "Invalid external type");

    /* Determine the internal type; find number of bytes per datum */
    if (intype == INTP_NONE)
	intype = exim_natural_intype(extype);
    inbytes = exim_sizeof_type(intype);
    if (inbytes == 0)
	NZReturnError(ERR_UNDEF, "Invalid internal type");

    /* Initialize the voxel array; allocate memory */
    flag = vxl_alloc_array(dest, intype, rank, dimen);
    if (flag)    NZReturnLibError();
    buff = mallock(nbuff*exbytes);
    if (buff == 0)    NZReturnLibError();

    /* Read and convert the data one bufferfull at a time */
    count = vxli_count(dest);
    data = dest->data;
    while (count > 0) {
	nget = (count < nbuff) ? count : nbuff;
	ngot = fread(buff, exbytes, nget, src);
	if (ngot < nget)
	    NZReturnError(ERR_EOF, "Premature end of data, or I/O error");
	i = exim_import(nget, data, intype, 1, buff, extype, exbytes, bigend);
	if (i != 0)
	    NZReturnError(ERR_MISC,
                          "Error converting from external to internal format");
	count -= nget;
	data  += nget*inbytes;
    }

    free(buff);
    return ERR_OKAY;
}



/*-----------------------------------------------------------------------
vxl_read_pnm_header - Read header info from a PNM file

This function reads the data format, height, width, and maximum value
from the header of a PNM file and leaves the stream positioned at the
first data byte.  If the magic number is not one of those shown below,
the rest of the header is skipped and an error is returned.

    P1 = ASCII Portable Bit Map
    P2 = ASCII Portable Gray Map
    P3 = ASCII Portable Pixel Map
    P4 = Binary Portable Bit Map
    P5 = Binary Portable Gray Map 
    P6 = Binary Portable Pixel Map
------------------------------------------------------------------------*/

#define PNM_HEADER_ERROR   "Unable to read a PBM/PGM/PPM header"
#define PNM_BAD_MAGIC      "Invalid magic number in PBM/PGM/PPM header"

int                             /* Returns non-zero on error */
vxl_read_pnm_header(
    FILE *src,                  /* File to read header info from */
    char *magic,                /* Two-byte magic number from file */
    long *width,                /* Width of image */
    long *height,               /* Height of image */
    long *maxval)               /* Maximum pixel value */
{    
    int c1, c2;                 /* Input characters */
    int narg  = 0;              /* Number of args read so far */
    int state = 0;              /* State of parser */
    long value;                 /* A value extracted from the header */
    char buff[32];              /* Buffer for input data */
    char *bp;                   /* Pointer into buffer */

    /* Extract the magic number */
    magic[0] = c1 = fgetc(src);
    magic[1] = c2 = fgetc(src);
    if (c1 == EOF || c2 == EOF)
        NZReturnError(ERR_MISC, PNM_HEADER_ERROR);

    /* Check for a valid magic number */
    if ( c1 != 'P' || c2 < '1' || c2 > '6' )
        NZReturnError(ERR_MISC, PNM_BAD_MAGIC);
    
    /* Extract the other header info */
    state = 0;
    bp    = buff;
    while (1) {

        /* Get the next character */
        c1 = fgetc(src);
        if (c1 == EOF)
            NZReturnError(ERR_MISC, PNM_HEADER_ERROR);

        /* Parse the header a character at a time */
        if ( state == 0 && c1 == '#' )
            state = 1;
        else if ( state == 0 && isspace(c1) )
            state = 0;
        else if ( ( state == 0 || state == 2) && isdigit(c1) ) {
            state = 2;
            *bp++ = c1;
        }
        else if ( state == 1 && c1 == '\n' )
            state = 0;
        else if ( state == 1 )
            state = 1;
        else if ( state == 2 && isspace(c1) ) {
            state = 0;
            *bp   = 0;
            value = atol(buff);
            bp    = buff;
            narg  = narg + 1;
            if (narg == 1)
                *width = value;
            else if (narg == 2)
                *height = value;
            else if (narg == 3) {
                *maxval = value;
                break;
            }
            else 
                NZReturnError(ERR_ASSERT, "PNM state machine is broken");
        }
        else
            NZReturnError(ERR_MISC,  PNM_HEADER_ERROR);
    }

    return 0;
}



/*-----------------------------------------------------------------------
vxl_read_pgm_image - Read image from a PGM file

This function reads image data from a specified file in Portable Gray
Map (PGM) format, and stores it in a voxel array provided by the
caller.  The rank of the image is set to 2 and the dimensions are set
as specified in the PGM header.

FIXME: Support the plain (ASCII) PGM format?
-------------------------------------------------------------------------*/

int                                 /* Returns non-zero on error */
vxl_read_pgm_image(
    voxel_array *dest,              /* Where to store the data read */
    char *filename)                 /* File to read the data from */
{
    FILE *src;                      /* File to read from */
    char magic[2];                  /* Magic number from pnm file */
    int rank = 2;                   /* Array rank is always 2 */
    long dimen[2] = {0,0};          /* Dimensions of the array */
    long maxval = 0;                /* Maximum pixel value in file */
    int intype = INTP_UCHAR;        /* Internal data type */
    int extype = EXTP_UINT1;        /* External data type */
    int bigend = 1;                 /* PGM is always big endian */
    int flag = 0;                   /* Error status flag */

    /* Open the file */
    src = fopen(filename, "rb");
    if ( ! src )
        NZReturnError(ERR_FILE, "Unable to open PGM input file");

    /* Read the header */
    flag = vxl_read_pnm_header(src, magic, dimen+1, dimen, &maxval);
    if ( flag )
        NZReturnError(ERR_MISC, "Error reading PNM header file");
    if ( strncmp(magic, "P5", 2) != 0 )
        NZReturnError(ERR_MISC, "Magic number in PGM header file is not P5");
    if ( maxval > 255 ) {
        intype = INTP_USHORT;
        extype = EXTP_UINT2;
    }

    /* Read the data into the destination */
    flag = vxl_read_raw_data(dest, src, rank, dimen, intype, extype, bigend);
    if (flag)    NZReturnLibError();

    return 0;
}



/*-----------------------------------------------------------------------
vxl_read_ppm_image - Read image from a PGM file

This function reads image data from a specified file in Portable Pixel
Map (PPM) format, and stores it in a voxel array provided by the
caller.  The rank of the image is set to 3 and the first two
dimensions are set to the height and width specified in the PGM
header; the third dimension is set to 3 for RGB data.

FIXME: Support the plain (ASCII) PGM format?
-------------------------------------------------------------------------*/

int                                 /* Returns non-zero on error */
vxl_read_ppm_image(
    voxel_array *dest,              /* Where to store the data read */
    char *filename)                 /* File to read the data from */
{
    FILE *src;                      /* File to read from */
    char magic[2];                  /* Magic number from pnm file */
    int rank = 3;                   /* Array rank is always 2 */
    long dimen[3] = {0,0,3};        /* Dimensions of the array */
    long maxval = 0;                /* Maximum pixel value in file */
    int intype = INTP_UCHAR;        /* Internal data type */
    int extype = EXTP_UINT1;        /* External data type */
    int bigend = 1;                 /* PGM is always big endian */
    int flag = 0;                   /* Error status flag */

    /* Open the file */
    src = fopen(filename, "rb");
    if ( ! src )
        NZReturnError(ERR_FILE, "Unable to open PPM input file");

    /* Read the header; note that height and width must reverse order */
    flag = vxl_read_pnm_header(src, magic, &dimen[1], &dimen[0], &maxval);
    if ( flag )
        NZReturnError(ERR_MISC, "Error reading PNM header file");
    if ( strncmp(magic, "P6", 2) != 0 )
        NZReturnError(ERR_MISC, "Bad magic number in PPM file");
    if ( maxval > 255 ) {
        intype = INTP_USHORT;
        extype = EXTP_UINT2;
    }

    /* DEBUG: Report header contents */
    if (0) {
      fprintf(stderr, "dimen  = %ld, %ld, %ld\n", 
	      dimen[0], dimen[1], dimen[2]);
      fprintf(stderr, "maxval = %ld\n", maxval);
      fprintf(stderr, "extype = %d\n", extype);
    }

    /* Read the data into the destination */
    flag = vxl_read_raw_data(dest, src, rank, dimen, intype, extype, bigend);
    if (flag)    NZReturnLibError();

    return 0;
}




/*-----------------------------------------------------------------------
vxl_write_raw_data - Convert voxels to external format and write to file

This function takes the contents of a voxel array, converts them to
some specified external data format, and writes them to a file stream.

FIXME: This code ignores the ERR_RANGE error from exim_export; it should
probably generate a warning instead.
------------------------------------------------------------------------*/

int                               /* Returns non-zero on error */
vxl_write_raw_data(
    FILE *dest,                   /* Where to write the data */
    int extype,                   /* External data format to use */
    int bigend,                   /* Write as big endian? */
    voxel_array *src)             /* The voxel array to write */
{
    int exbytes;                    /* Bytes per external datum */
    int intype;                     /* Type of internal data */
    int inbytes;                    /* Bytes per internal datum */
    size_t count;                   /* Number of data to write */
    size_t nbuff = 1024;            /* Number of data per buffer */
    unsigned char *buff;            /* Buffer for output data */
    size_t nput;                    /* Number of data to write next */
    size_t nact;                    /* Number of data actually written */
    unsigned char *data;            /* Data array in source */
    int flag;                       /* Error flag */

    /* Validate the source argument */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	NZReturnError(ERR_INIT, "Invalid or empty source array");

    /* Verify the external type; find number of bytes per datum */
    exbytes = exim_sizeof_type(extype);
    if (exbytes == 0)
	NZReturnError(ERR_UNDEF, "Invalid external type");

    /* Determine the internal type; find number of bytes per datum */
    intype  = src->type;
    inbytes = exim_sizeof_intype(intype);
    if (inbytes == 0)
	NZReturnError(ERR_UNDEF, "Invalid internal type");

    /* Allocate a buffer for i/o */
    buff = mallock(nbuff*exbytes);
    if (buff == 0)    NZReturnLibError();

    /* Convert and write the data one bufferfull at a time */
    count = vxli_count(src);
    data  = src->data;
    while (count > 0) {
	nput = (count < nbuff) ? count : nbuff;
	flag = exim_export(nput, buff, extype, exbytes, bigend, 
			   data, intype, 1);
	if ( flag != 0 && flag != ERR_RANGE )
	    NZReturnError(ERR_MISC,
                          "Error converting from internal to external format");
	nact = fwrite(buff, exbytes, nput, dest);
	if (nact < nput)
	    NZReturnError(ERR_FILE, "I/O error, or disk full");
	count -= nput;
	data  += nput*inbytes;
    }
    
    /* Free buffer and return */
    free(buff);
    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_write_pgm_image - Write a mono2d image as a pgm file

This function takes an mono2d imageof type unsigned char or unsigned
short as input and writes it to a file as raw bits Portable Gray Map
(pgm) image.  More precisely, the image must have exactly two
dimensions that are greater than 1; any dimensions equal to one will
be dropped.

FIXME: Generalize to output a multislice image as multiple images in a
single file?  This is, I believe, permitted by the PGM definition.
-------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_write_pgm_image(
    const char *fname,          /* Name of the new pgm file to write */
    voxel_array *src)           /* The voxel image to write */
{
    FILE *file;                 /* Stream to write the ppm to */
    long dimen[VXL_MAX_RANK];   /* Dimensions of PGM image */
    int intype  = INTP_UCHAR;   /* Internal data format */
    int extype  = EXTP_UINT1;   /* External data format */
    long maxval = 255;          /* Maximum value in external format */
    int  flag;                  /* Error status */
    int i, j;

    /* Verify image properties */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    intype = src->type;
    if (intype == INTP_UCHAR) {
        extype = EXTP_UINT1;
        maxval = 255;
    }
    else if (intype == INTP_USHORT) {
        extype = EXTP_UINT2;
        maxval = 65535;
    }
    else 
	NZReturnError(ERR_ILLOP,
                      "Only unsigned char or short allowed in PGM format");

    /* Determine dimensions of the PGM image */
    j = 0;
    for (i = 0; i < src->rank; i++) {
        if (src->dimen[i] > 1)
	    dimen[j++] = src->dimen[i];
    }
    if (j != 2)
        NZReturnError(ERR_INIT, "Only 2D mono images are supported");

    /* Open the output file */
    file = fopen(fname, "wb");
    if (file == 0)
	NZReturnError(ERR_FILE, "Unable to open ppm output file");

    /* Write the header */
    /* NOTE that Pyvox uses the order height, width whereas PGM uses
       the order width, height. */
    fprintf(file, "P5 %ld %ld %ld\n", dimen[1], dimen[0], maxval);

    /* Write the pixels themselves */
    flag = vxl_write_raw_data(file, extype, 1, src);
    if (flag)     NZReturnLibError();

    fclose(file);
    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_write_ppm_image - Write a rgb2d image as a ppm file

This function takes an rgb2d image of type unsigned char or unsigned
short as input and writes it to a file as raw bits Portable Pixel Map
(ppm) image.  The image must have exactly three dimensions that are
greater than 1 and the last (RGB) must be exactly 3.  Any dimensions equal
to one will be dropped.

FIXME: Generalize to output a multislice image as multiple images in a
single file?  This is, I believe, permitted by the PPM definition.
-------------------------------------------------------------------------*/

int                             /* Returns non-zero on error */
vxl_write_ppm_image(
    const char *fname,          /* Name of the new pgm file to write */
    voxel_array *src)           /* The voxel image to write */
{
    FILE *file;                 /* Stream to write the ppm to */
    long dimen[VXL_MAX_RANK];   /* Dimensions of voxel array */
    int intype  = INTP_UCHAR;   /* Internal data format */
    int extype  = EXTP_UINT1;   /* External data format */
    long maxval = 255;          /* Maximum value in external format */
    int  flag;                  /* Error status */
    int i, j;

    /* Verify image properties */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    intype = src->type;
    if (intype == INTP_UCHAR) {
        extype = EXTP_UINT1;
        maxval = 255;
    }
    else if (intype == INTP_USHORT) {
        extype = EXTP_UINT2;
        maxval = 65535;
    }
    else 
	NZReturnError(ERR_ILLOP,
                      "Only unsigned char or short allowed in writepgm");

    /* Determine dimensions of the PPM image */
    j = 0;
    for (i = 0; i < src->rank; i++) {
        if (src->dimen[i] > 1)
	    dimen[j++] = src->dimen[i];
    }
    if (j != 3 || dimen[2] != 3)
        NZReturnError(ERR_INIT, "Only 2D RGB images are supported");

    /* Open the output file */
    file = fopen(fname, "wb");
    if (file == 0)
	NZReturnError(ERR_FILE, "Unable to open ppm output file");

    /* Write the header */
    fprintf(file, "P6 %ld %ld %ld\n", dimen[1], dimen[0], maxval);

    /* Write the pixels themselves */
    flag = vxl_write_raw_data(file, extype, 1, src);
    if (flag)     NZReturnLibError();

    fclose(file);
    return ERR_OKAY;
}




/************************************************************************
                       Graphical Operations
************************************************************************/


/*----------------------------------------------------------------------
vxl_fill_2d_contour - Fill contour in a 2d image

This function fills the interior of a contour or polyline on a 2-D
image with a specified value.  The given fill value is cast to
whatever type the array is.  The contour is defined by the vertices
given in the array yx, with y and x values appearing alternately.  If
the last coordinate pair is not identical to the first, then a segment
from the last to the first is assumed to close the contour.  The
interior of the contour is defined by the even/odd rule: if a ray to
infinity from a point crosses the boundary contour an odd number of
times, then that point is inside the boundary; otherwise it is
outside.

Integer-valued coordinates are assumed to correspond to the center of
a voxel.  The ambiguity that occurs when a voxel center lies exactly
on a boundary vertex or segment is handled by pretending that the
voxel center is actually located at the given coordinates plus a
positive epsilon; epsilon is chosen large enough to avoid the
ambiguity but small enough to avoid creating another ambiguity.

Not supported for complex types.

FIXME: Add the capability to draw into an RGB or other vector-valued
canvas.
-----------------------------------------------------------------------*/

#define MAX_INTER  256          /* Maximum intersections of row and segment */

int                             /* Returns non-zero on error */
vxl_fill_2d_contour(
    voxel_array *canvas,        /* 2D array to draw into */
    long npoint,                /* Number of points in contour */
    double *yx,                 /* (y,x) pairs defining contour vertices */
    double value)               /* Value to fill with */
{
    int erank;                  /* Effective rank = # of non-unit dimensions */
    long dimen[VXL_MAX_RANK];   /* Effective dimensions of the canvas */
    int  type;                  /* Data type of the canvas */
    unsigned char *drow;        /* Current row of canvas */
    double ymin, ymax;          /* Vertical extent of contour */
    double xmin, xmax;          /* Horizontal extent of canvas */
    double xint[MAX_INTER];     /* Sorted list of intersections */
    int  nint;                  /* Number of intersections in xint */
    long nseg;                  /* Number of segments in boundary */
    long iseg;                  /* Index over segments of contour */
    double y1, x1, y2, x2;      /* Endpoints of current boundary segment */
    double yi, xi;              /* Coordinates of an intersection */
    long first, limit;          /* A range of columns within contour */
    int flag, i, j;

    /* Validate the array to draw into; get useful parameters */
    if ( canvas == 0 || canvas->magic != VXL_MAGIC ||
	 canvas->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty canvas");
    erank = 0;
    for (i = 0; i < canvas->rank; i++)
	if (canvas->dimen[i] > 1)
	    dimen[erank++] = canvas->dimen[i];
    if (erank != 2)
	NZReturnError(ERR_ILLOP, "Canvas is not two-dimensional");
    type  = canvas->type;
    xmin = 0.0;
    xmax = dimen[1] - 1;

    /* Determine vertical extent of the contour */
    if (npoint < 3)
        NZReturnError(ERR_ILLOP, "Fewer than 3 points in some contour");
    ymin = ymax = yx[0];
    for (iseg = 1; iseg < npoint; iseg++) {
	if (yx[2*iseg] < ymin)  ymin = yx[2*iseg];
	if (yx[2*iseg] > ymax)  ymax = yx[2*iseg];
    }

    /* Determine number of segments in the boundary */
    if (yx[0] == yx[2*npoint-2] && yx[1] == yx[2*npoint-1])
	nseg = npoint - 1;
    else
	nseg = npoint;
    if (nseg < 3)    
        NZWarning(ERR_ILLOP, "Fewer than 3 segments in some contour");

    /* Loop over the rows that intersect the contour */
    for (yi = ceil(ymin); yi < ceil(ymax); yi++) {

	/* Loop over the segments in the contour and find intersections */
	nint = 0;
	for (iseg = 0; iseg < nseg; iseg++) {

	    /* Find endpoints of the current segment */
	    y1 = yx[2*iseg];
	    x1 = yx[2*iseg+1];
	    if (iseg < npoint-1) {
		y2 = yx[2*iseg+2];
		x2 = yx[2*iseg+3];
	    }
	    else {
		y2 = yx[0];
		x2 = yx[1];
	    }

	    /* Test for an intersection with the current row */
	    if ( (y1 <= yi && yi < y2) ||
		 (y2 <= yi && yi < y1 ) ) {

		/* If found, clip to left/right edges of array */
		xi = x1 + (x2-x1)*(yi-y1)/(y2-y1);
		if (xi < xmin) xi = xmin;
		if (xi > xmax) xi = xmax;

		/* And add to sorted list of intersections */
                /* FIXME: Error code for overflow of program limits? */
		if (nint == MAX_INTER)
		    NZReturnError(ERR_ASSERT, "Too many intersections");
		for (i = 0; i < nint && xi >= xint[i]; i++) 
		    /* Do nothing */ ;
		for (j = nint; j > i; j--)
		    xint[j] = xint[j-1];
		xint[i] = xi;
		nint++;
	    }
	    
	}

	/* Fill portions of row within the contour */
	if (nint % 2 != 0)
	    NZReturnError(ERR_ASSERT, "Odd number of intersections");
        drow = (unsigned char *)canvas->data +
	       (long)yi * dimen[1] * (canvas->nbytes);
	for (i = 0; i < nint; i += 2) {
	    first = ceil(xint[i]);
	    limit = ceil(xint[i+1]);
	    flag  = bips_fill_range(first, limit, type, drow, 1, value);
	    if (flag)    NZReturnLibError();
	}
    }

    return ERR_OKAY;
}



/*************************************************************************
                 Statistical Processing of Images

Includes voxel classification algorithms.  Because of their similarity
to the other reduction operations, moment calculations are considered
to be reduction operations and are included in reduce.d, though a user-
oriented classification would probably still consider them as statistical
operations.
*************************************************************************/



/*-------------------------------------------------------------------------
vxl_histo_uchar - Compute histogram of an unsigned char/short volume image

This function computes the univariate histogram of an unsigned char or
unsigned short volume image and stores it as a one-dimensional volume
image.  An optional weight array may be provided, and defines the
value to be added to the histogram for each corresponding voxel;
specifying a NULL pointer indicates a weight of 1.  The source and
weight arrays must be unsigned char.  The histogram entries are NOT
checked for overflow, so this function should not be used with large
arrays and weights.

FIXME: Generalize to support other types for array and weights.  Note
that in that case you will need to provide some means of specifying
the number of bins and the bin limits.
-------------------------------------------------------------------------*/

int                          /* Returns non-zero on error */
vxl_histo_uchar(
    voxel_array *hist,       /* Place to store the histogram */
    voxel_array *src,        /* Image to take histogram of */
    voxel_array *weight)     /* Weight image */
{
    long dimen[1];           /* Dimensions of the histogram */
    unsigned long  *hdata;   /* Histogram data itself */
    unsigned char  *wdata;   /* Weight data itself */
    unsigned char  *ucdata;  /* Image data itself, if unsigned char */
    unsigned short *usdata;  /* Image data itself, if unsigned short */
    long count;              /* Count of voxels in the image */
    int nbins;               /* Number of bins in histogram */
    int type;                /* Type of source array */
    long i, flag;

    /* Validate the arguments provided */
    if (hist == 0 || hist->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");
    if ( src == 0 || src->magic != VXL_MAGIC ||  src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");

    /* Validate the source type and set number of bins */
    type = src->type;
    if (type == INTP_UCHAR)
        nbins = UCHAR_MAX + 1;
    else if (type == INTP_USHORT)
        nbins = USHRT_MAX + 1;
    else
        NZReturnError(ERR_INIT, 
                      "Source array must be unsigned char or unsigned short");

    /* Validate the weight argument, if present */
    if (weight != 0) {
	if ( weight->magic != VXL_MAGIC || weight->data == 0 ||
	     weight->type != INTP_UCHAR )
	    NZReturnError(ERR_INIT, "Invalid or empty weight array");
	if ( ! vxli_same_physical(src, weight) )    NZReturnLibError();
    }

    /* Set up the destination histogram; clear to zero */
    dimen[0] = nbins;
    flag = vxl_alloc_array(hist, INTP_ULONG, 1, dimen);
    if (flag)    NZReturnLibError();
    hdata = hist->data;
    for (i = 0; i < nbins; i++)
	hdata[i] = 0;

    /* Compute the histogram */
    count = vxli_count(src);
    if (type == INTP_UCHAR) {
        ucdata = src->data;
        if (weight != 0) {
            wdata = weight->data;
            for (i = 0; i < count; i++)
                hdata[ucdata[i]] += wdata[i];
        }
        else {
            for (i = 0; i < count; i++)
                hdata[ucdata[i]] += 1;
        }
    }
    else  {
        usdata = src->data;
        if (weight != 0) {
            wdata = weight->data;
            for (i = 0; i < count; i++)
                hdata[usdata[i]] += wdata[i];
        }
        else {
            for (i = 0; i < count; i++)
                hdata[usdata[i]] += 1;
        }
    }

    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_bihist_uchar - Compute bivariate histogram of two unsigned char images

This function computes the bivariate histogram of a pair of unsigned
char images and stores it as a two-dimensional unsigned long image; it
also computes the two marginal histograms and stores them as
one-dimensional unsigned long images.  The first source image defines
the row number of the bivariate histogram; the second the column
number; the weight array, if present, determines the count added to
the histogram for the corresponding voxel.  The source and weight
arrays must all be unsigned char; the weight array may be replaced by
a NULL pointer to specify a constant weight 1.  The histogram entries
are NOT checked for overflow, so this function should not be used with
large arrays and weights.
-------------------------------------------------------------------------*/

int                          /* Returns non-zero on error */
vxl_bihist_uchar(
    voxel_array *hist,       /* Place to store the bivariate histogram */
    voxel_array *mar1,       /* Histogram of first image */
    voxel_array *mar2,       /* Histogram of second image */
    voxel_array *src1,       /* First image to take histogram of */
    voxel_array *src2,       /* Second image to take histogram of */
    voxel_array *weight)     /* Weight image, or zero */
{
    long dimen[2];           /* Dimensions of the histogram */
    unsigned long *hdata;    /* Data array of bivariate histogram */
    long hlen;               /* Total elements in bivar histogram */
    long hsp;                /* Spacing between rows of bivar hist */
    unsigned long *m1, *m2;  /* Data array of marginal histograms */
    unsigned char *s1, *s2;  /* Data array of source images */
    unsigned char *w;        /* Data array of weight image */
    unsigned char wi;        /* Weight at a given voxel */
    long count;              /* Count of voxels in the image */
    long i, flag;

    /* Validate the destination arguments provided */
    if (hist == 0 || hist->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array hist");
    if (mar1 == 0 || mar1->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array mar1");
    if (mar2 == 0 || mar2->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array mar2");

    /* Validate the source arguments provided */
    if (src1 == 0 || src1->magic != VXL_MAGIC || 
        src1->data == 0 || src1->type != INTP_UCHAR)
	NZReturnError(ERR_INIT, "Invalid or empty source array src1");
    if (src2 == 0 || src2->magic != VXL_MAGIC || 
        src2->data == 0 || src2->type != INTP_UCHAR)
	NZReturnError(ERR_INIT, "Invalid or empty source array src2");
    if ( ! vxli_same_physical(src1, src2) )    NZReturnLibError();

    /* Validate the weight array, if provided */
    if (weight != 0) {
	if ( weight->magic != VXL_MAGIC || weight->data == 0 ||
	     weight->type != INTP_UCHAR )
	    NZReturnError(ERR_INIT, "Invalid or empty weight array");
	if ( !vxli_same_physical(src1, weight) )    NZReturnLibError();
    }

    /* Set up the destination histograms; clear to zero */
    dimen[0] = UCHAR_MAX + 1;
    dimen[1] = UCHAR_MAX + 1;
    hlen = dimen[0]*dimen[1];
    hsp  = dimen[1];
    flag = vxl_alloc_array(hist, INTP_ULONG, 2, dimen);
    if (flag)    NZReturnLibError();
    flag = vxl_alloc_array(mar1, INTP_ULONG, 1, dimen);
    if (flag)    NZReturnLibError();
    flag = vxl_alloc_array(mar2, INTP_ULONG, 1, dimen);
    if (flag)    NZReturnLibError();
    hdata = hist->data;
    m1    = mar1->data;
    m2    = mar2->data;
    for (i = 0; i <= UCHAR_MAX; i++)
	m1[i] = m2[i] = 0;
    for (i = 0; i < hlen; i++)
	hdata[i] = 0;

    /* Compute the histogram */
    s1 = src1->data;
    s2 = src2->data;
    count = vxli_count(src1);
    if (weight != 0) {
	w = weight->data;
	for (i = 0; i < count; i++) {
	    wi = w[i];
	    m1[s1[i]] += wi;
	    m2[s2[i]] += wi;
	    hdata[s1[i]*hsp+s2[i]] += wi;
	}
    }
    else {
	for (i = 0; i < count; i++) {
	    m1[s1[i]] += 1;
	    m2[s2[i]] += 1;
	    hdata[s1[i]*hsp+s2[i]] += 1;
	}
    }

    return ERR_OKAY;
}



/*--------------------------------------------------------------------------
vxl_kmeans1_histo - Univariate kmeans from a histogram

Given the univariate histogram of an unsigned char image and initial
guesses at class centroids, this function uses the K-means algorithm
to refine the class centroids.  If no initial guesses are available for
the class centroids, they may be initialized to random (but distinct)
values; but note that the results might possibly depend on the particular
random values chosen.  The final centroids will be sorted into ascending
order; no attempt will be made to remove duplicates.

FIXME: Generalize this to allow the histogram to contain any number of
bins?

FIXME:  There seem to be at least some cases in which setting the initial
centroids to all zeros yields two or more final centroids that are zero.
This might be related to the fact that empty centroids are not changed.

FIXME:  It's not clear how to make this work consistently when the
initial means are unknown; perhaps they should be changed to fall
at uniform intervals within the gray level range?

FIXME:  If a class is empty, then its centroid is left unchanged.  Is
this the right thing to do?
--------------------------------------------------------------------------*/

int                          /* Returns non-zero on error */
vxl_kmeans1_histo(
    voxel_array *hist,       /* Histogram as 1d voxel_array */
    int nclass,              /* Nmbr of classes = length of cent array */
    double  *cent)           /* Initial/final centroids */
{
    unsigned long *hd;       /* Actual data array for the histogram */
    long *cnt;               /* Count of gray levels by class */
    double *sum;             /* Sum of gray levels by class */
    int changed = 1;         /* Any class mean changed in last iteration? */
    int i;                   /* Count of iterations */
    int ig;                  /* Index over gray levels in histogram */
    double r, rb;            /* Distance to centroid, and best so far */
    int ic, icb;             /* Index over classes, and best so far */
    double mean;             /* Mean gray value of the current class */

    /* Validate histogram argument */
    if (hist == 0 || hist->magic != VXL_MAGIC || 
        hist->data == 0 || hist->type != INTP_ULONG ||
	hist->rank != 1 || hist->dimen[0] != UCHAR_MAX+1)
	NZReturnError(ERR_INIT, "Invalid or empty histogram array");
    hd = (unsigned long *)hist->data;

    /* Validate centroid arguments:  Lower bound is 2 because it
       doesn't make sense to classify to any fewer classes; upper bound 
       is more or less arbitrary but might catch some blunders. */
    if (nclass <= 1 || nclass > 512)
	NZReturnError(ERR_MISC, "Invalid number of classes");

    /* Allocate working space for class averages */
    cnt = mallock(nclass*sizeof(long));
    sum = mallock(nclass*sizeof(double));    
    if (cnt == 0 || sum == 0)    NZReturnLibError();

    /* Iterate until class means converge.  Never do more than 100
       iterations to prevent a possible endless loop; the value 100
       is just a quick guess unsupported by analysis. */
    changed = 1;
    for (i = 0; changed && i < 100; i++) {

	/* Clear counters for class averages */
	for (ic = 0; ic < nclass; ic++) {
	    cnt[ic] = 0;   sum[ic] = 0.0; }

	/* Assign each gray level to the nearest class centroid */
	for (ig = 0; ig < UCHAR_MAX; ig++) {

	    /* Find the best class for this gray level.  The test for
	       the best centroid is "<=" so that a gray value will be
	       assigned to the last class found if one or more are equally
	       good; this should mean that the final class centroids
	       are in ascending order when the initial values are all
	       zero.  */
	    rb  = HUGE_VAL;
	    icb = -1;
	    for (ic = 0; ic < nclass; ic++) {
		r = (double)ig - cent[ic];
		if (r < 0)  r = -r;
		if (r <= rb) {icb = ic;  rb = r; }
	    }
	    if (icb < 0 || icb >= nclass)
		NZReturnError(ERR_ASSERT, "Failed to find matching class");

	    /* Increment sum for the best class */
	    cnt[icb] += hd[ig];
	    sum[icb] += hd[ig] * ig;
	}

	/* Update class means; check if means have changed */
	changed = 0;
	for (ic = 0; ic < nclass; ic++) {
	    if (cnt[ic] > 0) {
		mean = sum[ic] / (double)cnt[ic];
		if (mean < 0.0 || mean > (double)UCHAR_MAX+0.1)
		    NZReturnError(ERR_ASSERT, "Found invalid mean");
		if ( fabs(mean - cent[ic]) > 0.001 ) {
		    changed = 1;
		    cent[ic] = mean; }
	    }
	}
    }

    /* Sort the centroids into ascending order */
    qsort(cent, nclass, sizeof(double), cmp_double);

    /* Free allocated storage and exit */
    free(cnt);
    free(sum);

    return ERR_OKAY;
}



/*--------------------------------------------------------------------------
vxl_kmeans2_histo - Bivariate kmeans from a histogram

Given the bivariate histogram of two unsigned char images and initial
guesses at class centroids, this function uses the K-means algorithm
to refine the class centroids.  If no initial guesses are available for
the class centroids, they may be initialized to random (but distinct)
values; but note that the results might possibly depend on the particular
random values chosen.  No attempt will be made to remove duplicates in
the final list of class centroids.

FIXME: Generalize this to allow the histogram to contain any number of
bins?

FIXME:  There seem to be at least some cases in which setting the initial
centroids to all zeros yields two or more final centroids that are zero.
This might be related to the fact that empty centroids are not changed.

FIXME:  It's not clear how to make this work consistently when the
initial means are unknown; perhaps they should be changed to fall
at uniform intervals within the gray level range?

FIXME:  If a class is empty, then its centroid is left unchanged.  Is
this the right thing to do?
--------------------------------------------------------------------------*/

int                          /* Returns non-zero on error */
vxl_kmeans2_histo(
    voxel_array *hist,       /* Histogram as 2d voxel_array */
    int nclass,              /* Nmbr of classes = length of cent array */
    double  *cent1,          /* Initial/final centroids for first image */
    double  *cent2)          /* Initial/final centroids for second image */
{
    unsigned long *hd;       /* Actual data array for the histogram */
    long *cnt;               /* Count of gray levels by class */
    double *sum1, *sum2;     /* Sum of gray levels by class, for each image */
    int changed = 1;         /* Any class mean changed in last iteration? */
    int i;                   /* Count of iterations */
    int ig1, ig2;            /* Index over gray levels in histogram */
    double r, rb;            /* Distance to centroid, and best so far */
    int ic, icb;             /* Index over classes, and best so far */
    double mean1, mean2;     /* Mean gray value of the current class */
    double r1, r2;
    long n;

    /* Validate histogram argument */
    if (  hist == 0 || hist->magic != VXL_MAGIC || hist->data == 0 ||
          hist->type != INTP_ULONG || hist->rank != 2 ||
          hist->dimen[0] != UCHAR_MAX+1 ||  hist->dimen[1] != UCHAR_MAX+1 )
	NZReturnError(ERR_INIT, "Invalid or empty histogram array");
    hd = (unsigned long *)hist->data;

    /* Validate centroid arguments:  Lower bound is 2 because it
       doesn't make sense to classify to any fewer classes; upper bound 
       is more or less arbitrary but might catch some blunders. */
    if (nclass <= 1 || nclass > 512)
	NZReturnError(ERR_MISC, "Invalid number of classes");

    /* Allocate working space for class averages */
    cnt  = mallock(nclass*sizeof(long));
    sum1 = mallock(nclass*sizeof(double));    
    sum2 = mallock(nclass*sizeof(double));    
    if (cnt == 0 || sum1 == 0 || sum2 == 0)
        NZReturnLibError();

    /* Iterate until class means converge.  Never do more than 100
       iterations to prevent a possible endless loop; the value 100
       is just a quick guess unsupported by analysis. */
    changed = 1;
    for (i = 0; changed && i < 100; i++) {

	/* Clear counters for class averages */
	for (ic = 0; ic < nclass; ic++) {
	    cnt[ic] = 0;   sum1[ic] = 0.0;   sum2[ic] = 0.0; }

	/* Assign each gray level pair to the nearest class centroid */
	for (ig1 = 0; ig1 < UCHAR_MAX; ig1++) {

            for (ig2 = 0; ig2 < UCHAR_MAX; ig2++) {

                /* Find the best class for this gray level pair.  The test for
                   the best centroid is "<=" so that a gray value will be
                   assigned to the last class found if one or more are equally
                   good; this should mean that the final class centroids
                   are in ascending order when the initial values are all
                   zero.  */
                rb  = HUGE_VAL;
                icb = -1;
                for (ic = 0; ic < nclass; ic++) {
                    r1 = (double)ig1 - cent1[ic];
                    r2 = (double)ig2 - cent2[ic];
                    r = r1*r1 + r2*r2;
                    if (r <= rb) {icb = ic;  rb = r; }
                }
                if (icb < 0 || icb >= nclass)
                    NZReturnError(ERR_ASSERT, "Failed to find matching class");

                /* Increment sum for the best class */
                n = hd[ig1*UCHAR_MAX + ig2];
                cnt[icb]  += n;
                sum1[icb] += n * ig1;
                sum2[icb] += n * ig2;
            }
        }

	/* Update class means; check if means have changed */
	changed = 0;
	for (ic = 0; ic < nclass; ic++) {
	    if (cnt[ic] > 0) {
		mean1 = sum1[ic] / (double)cnt[ic];
		mean2 = sum2[ic] / (double)cnt[ic];
		if (  mean1 < 0.0 || mean1 > (double)UCHAR_MAX+0.1 ||
                      mean2 < 0.0 || mean2 > (double)UCHAR_MAX+0.1 )
		    NZReturnError(ERR_ASSERT, "Found invalid mean");
		if (  fabs(mean1 - cent1[ic]) > 0.001 ||
                      fabs(mean2 - cent2[ic]) > 0.001 )  {
		    changed = 1;
		    cent1[ic] = mean1;
		    cent2[ic] = mean2;
                }
	    }
	}
    }

    /* Free allocated storage and exit */
    free(cnt);
    free(sum1);
    free(sum2);

    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_nnclass1_uchar - Univariate nearest neighbor classification

Given a volume image of any real type and a set of class centroids,
this function finds the nearest class centroid for each voxel position
and outputs the corresponding unsigned char class number.  Different
centroids may be assigned to the same class number.

FIXME: This might be a bit faster if vectorized.

FIXME: It might be useful to allow other types for the class numbers;
in particular, unsigned short and int.  More generally, if the class
numbers are provided as a voxel array, then any type could be used.
-------------------------------------------------------------------------*/

int                         /* Returns non-zero on error */
vxl_nnclass1_uchar(
    voxel_array *dest,      /* Classification image */
    voxel_array *src,       /* First feature image */
    int nclas,              /* Nmbr of classes = length of class array */
    unsigned char *clas,    /* Class id to assign for each centroid */
    double *cent)           /* Class centroids */
{
    int stype;                 /* Type of the source array */
    unsigned char *sd;         /* Actual data for the source images */
    unsigned char *dd;         /* Actual data for the destination image */
    long count;                /* Number of voxels in each image */
    long i;                    /* Index over voxels */
    int ic, icb;               /* Index over classes & best so far */
    double r, rb;              /* Distance to centroid & best so far */
    double v;                  /* Value of a source voxel */
    int flag;

    /* Check validity of source image */
    if ( src == 0 || src->magic != VXL_MAGIC || src->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    stype = src->type;
    if ( exim_iscomplex(stype) )
        NZReturnError(ERR_INIT, "Not defined for complex arrays");

    /* Check validity of other arguments:  Lower bound is 2 because it
       doesn't make sense to classify to any fewer classes; upper bound 
       is more or less arbitrary but should catch some blunders. */
    if (nclas <= 1 || nclas > 512)
	NZReturnError(ERR_MISC, "Invalid number of classes");

    /* Allocate space for the destination */
    flag = vxl_alloc_array(dest, INTP_UCHAR, src->rank, &(src->dimen[0]));
    if (flag)    NZReturnLibError();

    /* Loop over voxels in the images */
    count = vxli_count(src);
    sd    = src->data;
    dd    = dest->data;
    for (i = 0; i < count; i++) {

	/* Find the nearest class and assign it.  If two or more classes
	   are equally close, the first found is chosen.  */
	flag = exim_get_value(&v, sd, stype, i);
	if (flag)    NZReturnLibError();
	rb  = HUGE_VAL;
	icb = -1;
	for (ic = 0; ic < nclas; ic++) {
	    r = v-cent[ic];
	    if (r < 0) r = -r;
	    if (r < rb) {
	        icb = ic;
		rb  = r; }
	}
	if (icb < 0 || icb >= nclas)
	    NZReturnError(ERR_ASSERT, "Failed to find class");
	dd[i] = clas[icb];
    }

    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
vxl_nnclass2_uchar - Bivariate nearest neighbor classification

Given two volume images of the same shape and any type, and a set of
class centroids, this function finds the nearest class centroid for
each voxel position and outputs the corresponding class number.  Different
centroids may be assigned to the same class number.

FIXME: This might be a bit faster if vectorized.

FIXME: It might be useful to allow other types for the class numbers;
in particular, unsigned short and int.  More generally, if the class
numbers are provided as a voxel array, then any type could be used.
-------------------------------------------------------------------------*/

int                         /* Returns non-zero on error */
vxl_nnclass2_uchar(
    voxel_array *dest,      /* Classification image */
    voxel_array *src1,      /* First feature image */
    voxel_array *src2,      /* Second feature image */
    int nclas,              /* Nmbr of classes = length of class array */
    unsigned char *clas,    /* Class id to assign for each centroid */
    double        *cent1,   /* Class centroid in first feature image */
    double        *cent2)   /* Class centroid in second feature image */
{
    int stype1, stype2;        /* Types of the two source images */
    unsigned char *sd1, *sd2;  /* Actual data for the source images */
    unsigned char *dd;         /* Actual data for the destination image */
    long count;                /* Number of voxels in each image */
    long i;                    /* Index over voxels */
    int ic, icb;               /* Index over classes & best so far */
    double r, rb;              /* Distance to centroid & best so far */
    double v1, v2;             /* Values from the two source images */
    double r1, r2;
    int flag;

    /* Check validity and compatibility of source images */
    if ( src1 == 0 || src1->magic != VXL_MAGIC || src1->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array src1");
    if ( src2 == 0 || src2->magic != VXL_MAGIC ||  src2->data == 0 )
	NZReturnError(ERR_INIT, "Invalid or empty source array src2");
    if ( ! vxli_same_physical(src1, src2) )    NZReturnLibError();
    stype1 = src1->type;
    stype2 = src2->type;

    /* Not defined for complex images */
    if ( exim_iscomplex(stype1) || exim_iscomplex(stype2) )
        NZReturnError(ERR_INIT, "Not defined for complex arrays");

    /* Check validity of other arguments:  Lower bound is 2 because it
       doesn't make sense to classify to any fewer classes; upper bound is
       more or less arbitrary but should catch some blunders. */
    if (nclas <= 1 || nclas > 512)
	NZReturnError(ERR_MISC, "Invalid number of classes");

    /* Allocate space for the destination */
    flag = vxl_alloc_array(dest, INTP_UCHAR, src1->rank, &(src1->dimen[0]));
    if (flag)    NZReturnLibError();

    /* Loop over voxels in the images */
    count = vxli_count(src1);
    sd1   = src1->data;
    sd2   = src2->data;
    dd    = dest->data;
    for (i = 0; i < count; i++) {

	/* Find the nearest class and assign it */
	flag = exim_get_value(&v1, sd1, stype1, i);
	if (flag)    NZReturnLibError();
	flag = exim_get_value(&v2, sd2, stype2, i);
	if (flag)    NZReturnLibError();
	rb  = HUGE_VAL;
	icb = -1;
	for (ic = 0; ic < nclas; ic++) {
	    r1 = v1 - cent1[ic];
	    r2 = v2 - cent2[ic];
	    r  = r1*r1 + r2*r2;
	    if (r < rb) {
	        icb = ic;
		rb  = r; }
	}
	if (icb < 0 || icb >= nclas)
	    NZReturnError(ERR_ASSERT, "Failed to find class");
	dd[i] = clas[icb];
    }

    return ERR_OKAY;
}



/***************************************************************************
                     Image Scanning Operations

Image scanning operations scan an image in some data-dependent or
complex fashion and may generate either a new image or scalar results.
Some examples are object extraction, finding the bounding box, and
distance transforms.
***************************************************************************/


/*-------------------------------------------------------------------------
vxl_find_objects_uchar3d - Extract the objects in a volume image

This function scans a 2d or 3d unsigned char volume image and
separates it into objects (which are defined as maximal connected
regions of non-zero voxels).  It returns the number of objects found
(including the background as object 0) and also sets a pointer to a
table containing information on the objects found.  It is possible
that all the pixels in the input image have value 0; in this case,
they are all assigned to the background, which becomes the only
object.  The background (region 0) consists of all the zero pixels and
need not be connected; as a result, the seed point should be ignored.

The algorithm used here is a reference implementation which is
intended to be (relatively!) easy to understand and verify rather than
fast; it will eventually be replaced by a faster algorithm which can
then be tested against this reference implementation.  A 2d image is
handled essentially as one slice of a 3d image; the implementation is
a bit awkward because it was added long after the 3d version was done.

The first stage is to scan the image in raster order, considering each
voxel in turn.  If the voxel is zero, then it is always assigned to
object 0, which is background.  Otherwise the 13 neighbors of that
voxel which preceded it in the scan are examined.  If all the
neighbors are zero, then a new (temporary) object id is allocated and
assigned to the current voxel.  If all the non-zero neighbors have
been assigned the same object id, then the current voxel is assigned
that same id.  If the non-zero neighbors have been assigned two or
more distinct object ids, then the smallest of those neighboring object
ids is taken as canonical and the others are consolidated into it;
then the current voxel is assigned the canonical id.  The object table
is initially allocated to contain a small number of objects and
reallocated when necessary.  The count of voxels in each object is
updated as each voxel is processed; the counts of the non-canonical
objects are merged into the canonical counts in a pass over the object
table once the image has been completely scanned.

The first stage of the algorithm often assigns two or more temporary
object ids to what is later found to be a single object; rather than
go back and change the old idents, it simply chooses the smaller id as
canonical, and indicates in the object table the canonical id of the
redundant ids.  The second stage scans the object table and merges the 
number of pixels from all the temporary objects assigned to the same 
canonical object.

FIXME: The handling of voxels at the boundary of the image is probably
correct but rather inefficient.  How do I handle voxels at the
boundary of the image without excessively complicating the algorithm
or losing performance?  One possible approach is to move the loop over
the neighbors outside the innermost loop over the voxels in a
scanline; then testing for the boundary of the image can be done once
per scanline rather than once per voxel.  A more radical approach (but
not suitable for a reference implementation) might be to focus on
object boundaries as the places where the output object id must
change; the object boundaries can probably be found much faster.

FIXME:  Can this be generalized to arbitrary dimension and structuring
elements using the same methods used for convolution et al?

FIXME: Think about the right terminology for component/region/object 
and use consistently.
-------------------------------------------------------------------------*/

int                           /* Returns number of regions, or zero on error */
vxl_find_objects_uchar3d(
    voxel_array *dest,        /* Volume image containing object labels */
    long *nobjects,           /* Number of objects found */
    struct voxel_object_data  /* Table of objects found */
        *oblist[],
    voxel_array *src)         /* Volume image to extract objects from */
{
    int rank;                    /* Number of dimensions */
    unsigned char  *sd;          /* Actual data array for source */
    unsigned short *dd;          /* Actual data array for destination */
    long nz, ny, nx;             /* Number of slices, rows, columns */
    long iz, iy, ix;             /* Indices over slices, rows, columns */
    long dz, dy, dx;             /* Stride in slices, rows, columns */
    struct voxel_object_data *obl;     /* Object list */
    int lobl;                    /* Allocated length of object table */
    int nobs;                    /* Number of objects found so far */
    long dn[13];                 /* Offsets to neighbors of current voxel */
    int in;                      /* Index over neighbors */
    int nbr;                     /* Object id of a neighbor */
    int nbr1, nbr2;              /* First and second ob ids in a neghborhood */
    int can1, can2;              /* Canonical ids for nbr1 and nbr2 */
    long nbytes;                 /* Number of bytes needed for object list */
    long total;                  /* Total number of voxels in all objects */
    int n0;                      /* Lower limit in dn array */
    int n, nn;
    int flag;

    static char infloop[] = "Infinite loop in object finder";
    static char inflist[] = "Infinite loop in object list";
    static char toomany[] = "Found too many objects for algorithm";
    static char zerocnt[] = "Zero count for a (non-background) canonical object";

    /* Binary code table indicating which neighbors to take when
       iz == 0 (4), iy == 0 (2), or ix == 0 (1).  A 3D array uses all of
       these; a 2D array uses only the last 4. */
    int take[13] = {0, 1, 1, 2, 3, 3, 2, 3, 3, 4, 5, 5, 6};

    /* Set safe object count in case of errors */
    *nobjects = 0;

    /* Check validity of source array; get the dimensions. */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0 || 
	src->type != INTP_UCHAR)
	NZReturnError(ERR_INIT, "Invalid or empty source array");
    sd = src->data;

    /* Get image dimensions; compute step size on each axis; and set
       lower limit of neighbor array dn.  Treat 2D image as slice of a
       3D image. */
    rank = src->rank;
    if (rank == 3) {
        nz = src->dimen[0];
        ny = src->dimen[1];
        nx = src->dimen[2];
        dz = vxli_step(src, 0);
        dy = vxli_step(src, 1);
        dx = vxli_step(src, 2);
        n0 = 0;
    }
    else if (rank == 2) {
        nz = 1;
        ny = src->dimen[0];
        nx = src->dimen[1];
        dz = 0;
        dy = vxli_step(src, 0);
        dx = vxli_step(src, 1);
        n0 = 9;
    }
    else 
        NZReturnError(ERR_INIT, "Not supported except for 2D or 3D");

    /* Check validity of destination */
    if (dest == 0 || dest->magic != VXL_MAGIC)
	NZReturnError(ERR_INIT, "Invalid destination array");

    /* Allocate the destination as an object map; copy origin and spacing */
    flag = vxl_alloc_array(dest, INTP_USHORT, rank, src->dimen);
    if (flag)    NZReturnLibError();
    vxli_copy_physical(dest, src);
    dd = dest->data;

    /* Allocate initial object table; define the background object */
    nobs =   1;
    lobl = 256;
    obl  = mallock(lobl*sizeof(struct voxel_object_data));
    if (obl == 0)    NZReturnLibError();
    obl[0].ident = 0;
    obl[0].canon = 0;
    obl[0].count = 0;
    for (n = 0; n < 3; n++)
        obl[0].point[n] = 0;

    /* Compute offsets to neighbors.  A 3D array uses all of
       these; a 2D array uses only the last 4. */
    dn[ 0] = -dz-dy-dx;
    dn[ 1] = -dz-dy;
    dn[ 2] = -dz-dy+dx;
    dn[ 3] = -dz   -dx;
    dn[ 4] = -dz;
    dn[ 5] = -dz   +dx;
    dn[ 6] = -dz+dy-dx;
    dn[ 7] = -dz+dy;
    dn[ 8] = -dz+dy+dx;
    dn[ 9] =    -dy-dx;
    dn[10] =    -dy;
    dn[11] =    -dy+dx;
    dn[12] =       -dx;

    /* Loop over the voxels in the image */
    /* FIXME:  Probably faster to loop over neighbors outside columns */
    for (iz = 0; iz < nz; iz++)
	for (iy = 0; iy < ny; iy++)
	    for (ix = 0; ix < nx; ix++) {

		nbr1 = 0;

		/* If voxel is non-zero, consider as possible object */
		if (*sd != 0) {


		    /* Loop over the neighbors (omitting those that
		       fall outside the edges of the image) and record
		       the object id in nbr1.  If two different object
		       ids are seen in the neighborhood, mark them as
		       the same in the object list, taking the smaller
		       id as canonical. */
		    for (in = n0; in < 13; in++)
			if ( ( iz > 0 || take[in] & 4 ) &&
			     ( iy > 0 || take[in] & 2 ) && 
			     ( ix > 0 || take[in] & 1 ) &&
			     ((nbr = dd[dn[in]]) != 0 ) ) {

			    /* Check for broken algorithm */
			    if (nbr < 1 || nbr >= nobs)
				NZReturnError(ERR_ASSERT, "Error in object "
                                              "finding algorithm");

			    /* Case 1: The first non-zero neighbor
			     need only be recorded. */
			    if ( nbr1 == 0 ) {
				nbr1 = nbr;
			    }
			    
			    /* Case 2: There are two apparently
			       distinct neighbor ids.  Trace down the
			       canonical ids.  If they are different,
			       make the smaller id canonical. In any
			       case, record the canonical ids in the
			       actual ids */
			    else if ( nbr != nbr1 ) {

				for ( n = 0, can1 = nbr1;
				      n < nobs && can1 != obl[can1].canon;
				      n++, can1 = obl[can1].canon ) /* NIL */;
				if (n >= nobs)
                                  NZReturnError(ERR_ASSERT, infloop);

				for ( n = 0, can2 = nbr2 = nbr;
				      n < nobs && can2 != obl[can2].canon;
				      n++, can2 = obl[can2].canon ) /* NIL */;
				if (n >= nobs)
				    NZReturnError(ERR_ASSERT, infloop);

				if (can2 < can1) {
				    obl[can1].canon = can2;
				    obl[nbr1].canon = can2;
				    obl[nbr2].canon = can2; }
				else if (can1 < can2) {
				    obl[can2].canon = can1;
				    obl[nbr1].canon = can1;
				    obl[nbr2].canon = can1; }
				else {
				    obl[nbr1].canon = can1;
				    obl[nbr2].canon = can2; }
			    }

			}

		    /* If all neighbors are zero, assign new object number,
		       enlarging the object list if necessary. */
		    if (nbr1 == 0) {
			if (nobs >= USHRT_MAX)
			    NZReturnError(ERR_ALGOR, toomany);
			nbr1 = nobs++;
			if (0) 
			    fprintf(stderr, "Creating object #%d\n", nbr1);
			if (nobs > lobl) {
			    lobl *= 2;
			    nbytes = lobl*sizeof(struct voxel_object_data);
			    obl = reallock(obl, nbytes);
                            if (obl == 0)    NZReturnLibError();
			}
			obl[nbr1].ident  = nbr1;
			obl[nbr1].canon  = nbr1;
			obl[nbr1].count  = 0;
			obl[nbr1].point[0] = iz;
			obl[nbr1].point[1] = iy;
			obl[nbr1].point[2] = ix;
		    }

		}

		/* Capture the coordinates of the first zero voxel seen */
                /* FIXME: Not really useful, since the background might
                   not be connected. */
		else if ( *sd == 0 && obl[0].count == 0 ) {
		    obl[0].point[2] = ix;
		    obl[0].point[1] = iy;
		    obl[0].point[0] = iz;
		}

		/* Set label in destination and update voxel count */
		obl[nbr1].count += 1;
		*dd = nbr1;
		
		/* Step to the next voxel in source and destination */
		sd++;
		dd++;
	    }

    /* At this point, obl[n] contains the object description for the
       nth temporary object, and the ident field is equal to n. The
       canon field contains the ident of the temporary object chosen
       as canonical; thus the ident and canon fields are equal iff the
       object is canonical.  The count field contains the pixel count
       for the temporary object only. */

    /* DEBUG: Print out the object table before merging regions */
    if (0) {
        fprintf(stderr, "\nObject table before merging regions\n");
        fprintf(stderr, "  n    id   can     cnt     sz     sy    sz\n");
        for (n = 0; n < nobs; n++) {
            fprintf(stderr, "%3d   %3d   %3d   %5ld   %4ld   %4ld   %4ld\n",
                    n, obl[n].ident, obl[n].canon, obl[n].count,
                    obl[n].point[0], obl[n].point[1], obl[n].point[2]);
        }
    }

    /* Do some inexpensive consistency checks: The ident field is
       equal to the index; every canon field is less than or equal to
       the ident field; the sum of the count fields is equal to the
       total number of voxels in the image. */
    total = 0;
    for (n = 0; n < nobs; n++) {
	if (n != obl[n].ident)
	    NZReturnError(ERR_ASSERT, "Invalid ident field in object finder");
	if (obl[n].canon > obl[n].ident)
	    NZReturnError(ERR_ASSERT, "Invalid canon field in object finder");
	if (n > 0 && obl[n].count <= 0)
	    NZReturnError(ERR_ASSERT, zerocnt);
	total += obl[n].count;
    }
    if (total != nz*ny*nx)
	NZReturnError(ERR_ASSERT, "Objects do not account for all voxels");

    /* Merge all the temporary regions into a canonical region and add all
       their pixel counts to that canonical region.  The canonical region 
       number is the smallest of all the temporary region numbers that it
       contains. */
    for (n = 0; n < nobs; n++ ) {
        if ( obl[n].ident != obl[n].canon ) {

            /* Chase through the canon fields to find the ultimate ancestor */
            for ( nn = 0, can1 = n;
                  nn < nobs && can1 != obl[can1].canon;
                  nn++, can1 = obl[can1].canon ) /* NIL */;
            if (nn >= nobs)
                NZReturnError(ERR_ASSERT, inflist);
            can1 = obl[can1].ident;
	    obl[n].canon = can1;
            
            /* Add counts to the canonical object; zero count in this object */
            obl[can1].count += obl[n].count;
            obl[n].count = 0;
        }
    }

    /* A 2D image has had the seed point coordinates recorded as 0, iy, ix;
       shift them to get iy, ix, 0. */
    if (rank == 2)
        for (n = 0; n < nobs; n++) {
            obl[n].point[0] = obl[n].point[1];
            obl[n].point[1] = obl[n].point[2];
            obl[n].point[2] = 0;
        }

    /* DEBUG: Print out the object table after merging regions */
    if (0) {
        fprintf(stderr, "\nObject table after merging regions\n");
        fprintf(stderr, "  n    id   can     cnt     sz     sy    sz\n");
        for (n = 0; n < nobs; n++) {
            fprintf(stderr, "%3d   %3d   %3d   %5ld   %4ld   %4ld   %4ld\n",
                    n, obl[n].ident, obl[n].canon, obl[n].count,
                    obl[n].point[0], obl[n].point[1], obl[n].point[2]);
        }
    }

    /* Truncate the object list to its actual size */
    obl = reallock(obl, nobs*sizeof(struct voxel_object_data));

    /* Set values for the object list and return */
    *oblist   = obl;
    *nobjects = nobs;
    return ERR_OKAY;
}



/*--------------------------------------------------------------------------
vxl_chamfer - Compute chamfer distance image

This function computes the chamfer distance transform of a given
image.  The chamfer distance image is zero wherever the original image
is non-zero and is otherwise equal to the taxicab (L1) distance to the
nearest non-zero point in the original image.  The type of the output
image may be specified by the caller to either unsigned char or
unsigned short, or allowed to default to whichever type is sufficient
to contain the maximum possible distance.

FIXME: Verify that this works for dimensions that are 0 or 1.

FIXME: Should this code consider and correct for anisotropic sampling?
One approach would be to generalize to support an arbitrary
structuring element (defined by a kernel) with the distances specified
by the coefficients of the neighbors.
----------------------------------------------------------------------------*/

int                              /* Returns non-zero on error */
vxl_chamfer(
    voxel_array *dest,           /* Destination image */
    voxel_array *src,            /* Source image */
    int dtype)                   /* Desired destination type, or zero */
{
    int rank;                    /* Rank of the src and dest arrays */
    int stype;                   /* Type of the source array */
    long *dimen;                 /* Dimensions of the two arrays */
    long coord[VXL_MAX_RANK];    /* Coordinates of current row */
    int idim;                    /* Current scan dimension */
    long nelem;                  /* Number of elements in row for last dimen */
    long nelem1;                 /* Number of elements, minus 1 */
    unsigned char *drow;         /* Start of a row in destination */
    unsigned char *srow;         /* Start of a row in source */
    long doff;                   /* Destination offset to previous voxel */
    long md;                     /* Maximum possible taxicab distance */
    int flag;

    /* Verify the source image and get its properties */
    if ( src == 0 || src->magic != VXL_MAGIC || src->data == 0 ||
	 src->rank < 1 )
	NZReturnError(ERR_INIT, "Source array is invalid");
    rank   = src->rank;
    stype  = src->type;
    dimen  = src->dimen;
    nelem  = dimen[rank-1];
    nelem1 = nelem - 1;

    /* Chose default destination type, if necessary */
    if (dtype == 0) {
        md = 0;
	for (idim = 0; idim < rank; idim++)
	    md += dimen[idim];
	if (md > UCHAR_MAX)
	    dtype = INTP_USHORT;
	else
	    dtype = INTP_UCHAR;
    }
    else if (dtype != INTP_USHORT && dtype != INTP_UCHAR)
	NZReturnError(ERR_MISC, 
                      "Invalid destination type in chamfer transform");

    /* Allocate the destination array */
    flag = vxl_alloc_array(dest, dtype, rank, dimen);
    if (flag)    NZReturnLibError();
    vxli_copy_physical(dest, src);

    /* Initialize destination to largest possible distance and scan
       innermost dimension in positive direction */
    doff = -1;
    vxli_firstrow(rank, coord);
    do {

        /* Initialize the destination row, starting with the first voxel */
        coord[rank-1] = 0;
        drow = vxli_locate(dest, coord);
        if (drow == 0)    NZReturnLibError();
        flag = bips_fill_max(nelem, dtype, drow, 1);
	if (flag)    NZReturnLibError();

	/* Setup and execute the chamfer inner loop, starting with the
           second voxel in the row */
        coord[rank-1] = 1;
	drow = vxli_locate(dest, coord);
	srow = vxli_locate(src,  coord);
        if (drow == 0 || srow == 0)    NZReturnLibError();
	flag = bips_chamfer(nelem1, dtype, drow, 1, doff, stype, srow, 1);
	if (flag)    NZReturnLibError();

    } while ( vxli_nextrow(rank, dimen, coord) );

    /* Scan outer dimensions in positive direction by taking rows in
       forward order */
    for (idim = rank-2; idim >= 0; idim--) {

	/* Loop over the rows along the current direction */
	doff = -vxli_step(dest, idim);
	vxli_firstrow(rank, coord);
	do {

	    /* Skip first row in outer dimensions */
	    if ( coord[idim] == 0 )
		continue;

	    /* Locate starting pixel in the current row */
	    drow = vxli_locate(dest, coord);
            srow = vxli_locate(src,  coord);
            if (drow == 0 || srow == 0)    NZReturnLibError();

	    /* Execute the chamfer inner loop */
	    flag = bips_chamfer(nelem, dtype, drow, 1, doff, stype, srow, 1);
	    if (flag)    NZReturnLibError();

	} while ( vxli_nextrow(rank, dimen, coord) );

    }

    /* Scan innermost dimension in negative direction */
    doff = 1;
    vxli_lastrow(rank, dimen, coord);
    do {

	/* Locate starting pixel at the end of the current row */
        coord[rank-1] = dimen[rank-1] - 2;
	drow = vxli_locate(dest, coord);
	srow = vxli_locate(src,  coord);
        if (drow == 0 || srow == 0)    NZReturnLibError();

	/* Execute the chamfer inner loop */
	flag = bips_chamfer(nelem1, dtype, drow, -1, doff, stype, srow, -1);
	if (flag)    NZReturnLibError();

    } while ( vxli_prevrow(rank, dimen, coord) );

    /* Scan outer dimensions in negative direction by taking rows in
       reverse order */
    for (idim = rank-2; idim >= 0; idim--) {

	/* Initialize dimension-specific parameters */
	doff = vxli_step(dest, idim);

	/* Loop over the rows along the current direction */
	vxli_lastrow(rank, dimen, coord);
	do {

	    /* Skip last row in outer dimensions */
	    if ( coord[idim] == dimen[idim] - 1 )
		continue;

	    /* Locate starting pixel in the current row */
	    coord[rank-1] = 0;
	    drow  = vxli_locate(dest, coord);
            srow  = vxli_locate(src,  coord);
            if (drow == 0 || srow == 0)    NZReturnLibError();

	    /* Execute the chamfer inner loop */
	    flag = bips_chamfer(nelem, dtype, drow, 1, doff, stype, srow, 1);
	    if (flag)    NZReturnLibError();

	} while ( vxli_prevrow(rank, dimen, coord) );
    }

    return ERR_OKAY;
}
