/****************************************************************************
vxli2.c - Inlineable internal functions for the Voxel Kit

Author:  Paul Hughett

Copyright (c) 2001-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  It contains various inlineable functions for use within
the Voxel Kit itself.  An "inlineable" function will be inlined in the
compiler supports inline functions, and treated as an ordinary called
function otherwise.  This is implemented by either including this file
in the using program via the vxli.h header, or including it into the
file vxli.c, which contains other internal functions.

FIXME: Identify other common validity checking steps and isolate as
(inline?) functions or macros.  Q: What are the efficiency tradeoffs
between the two?

FIXME: Add debugging option to add back in the missing validation steps.

NOTE: The rcsid string below will be defined multiple times on compilers 
that support inline.  Unfortunately, there doesn't seem to be any easy way 
to include it only once within the Voxel Kit library, while still including
it in files that use the Voxel Kit but are not actually part of it.
Yes, there are such programs, of which pyvox.c and parray.c are examples.
Strictly speaking, these two files should not use Voxel Kit internals, but
life isn't quie that simple.
*****************************************************************************/

static const char rcsid2[] = "$Id: vxli2.c,v 1.11 2005/12/29 18:33:16 hughett Exp $ Release $Name: pyvox-0-72 $";

/* Assumes that config.h and vxli.h have already been included */



/*------------------------------------------------------------------------
vxli_count - Compute number of voxels in a voxel array

Given an voxel array, this function computes the total number of
voxels or elements that it contains.  This function assumes that the
caller has already checked that the src array is valid and allocated.
For example,

    if (src == 0 || src->magic != VXL_MAGIC || src->data != 0)
        NZReturnError(ERR_TYPE, "Invalid or empty source array");

SEE ALSO vxl_count() to check the validity of the src before computing the
number of elements.
--------------------------------------------------------------------------*/

inlineable 
long                         /* Returns number of elements */
vxli_count(
    voxel_array *src)        /* Voxel array to define dimensions */
{
    long count;              /* Count of elements */
    int i;
   
    /* Compute the desired count */
    count = 1;
    for (i = src->rank - 1; i >= 0; i--) 
	count *= src->dimen[i];

    return count;
}



/*------------------------------------------------------------------------
vxli_delta - Compute element stride for given subscript deltas

Given an voxel array, and subscript deltas for each coordinate, this
function computes the distance in the array which corresponds to the
given change in the coordinates, measured in elements (not bytes).
The length of the coords array must equal the rank of the voxel array.

NOTE that no validity checking is done in this function; the caller is
responsible for checking that the argument source is not NULL, that the
magic number is correct, and that the voxel array has been initialized.
(Validity checking was done prior to 2002-09-02, but all callers already
checked for validity.) 

SEE ALSO: The related routine vxli_offset computes the location of an element
relative to the beginning of the array, AND verifies that the given
coordinates are within the array.  The routine vxli_step and its friends
computes the delta for changing a single coordinate.
--------------------------------------------------------------------------*/

inlineable
long                         /* Returns the desired delta */
vxli_delta(
    voxel_array *src,        /* Voxel array to define dimensions */
    long *coords)            /* Coordinate deltas */
{
    long step;               /* Step size for current coordinate */
    long delta;              /* Delta computed so far */
    int i;
    
    /* Compute the desired delta */
    delta = 0;
    step = 1;
    for (i = src->rank - 1; i >= 0; i--) {
	delta += coords[i]*step;
	step *= src->dimen[i];
    }

    return delta;
}



/*------------------------------------------------------------------------
vxli_offset - Compute element number for given subscripts

Given an voxel array, and subscript values for each coordinate, this
function computes the offset or element number selected by the given
coordinate values, measured in elements (not bytes).  The length of
the coords array must equal the rank of the voxel array.  If the given
coordinates are outside the voxel array, then a negative offset is
returned.  NO validity checking is done, so this version is suitable
only for use within the voxel kit itself.

SEE ALSO: vxli_offset() to compute element number given the rank and
dimensions of a voxel array.
--------------------------------------------------------------------------*/

inlineable
long                         /* Returns NEGATIVE offset on error */
vxli_offset(
    voxel_array *src,        /* Voxel array to define dimensions */
    long *coords)            /* Coordinate values */
{
    long step;               /* Step size for current coordinate */
    long offset;             /* Offset computed so far */
    int i;
    
#if 0
    /* Check for valid voxel array; REMOVED 2002-09-02 */
    if (src == 0 || src->magic != VXL_MAGIC) {
        set_error(ERR_INIT, vxli_errmsg_invalid);
        return -1;
    }
#endif

    /* Compute the desired offset */
    offset = 0;
    step = 1;
    for (i = src->rank - 1; i >= 0; i--) {
	if (coords[i] < 0 || coords[i] >= src->dimen[i]) {
            set_error(ERR_INDEX, vxli_errmsg_subscript);
            return -1;
	}
	offset += coords[i]*step;
	step *= src->dimen[i];
    }

    return offset;
}



/*------------------------------------------------------------------------
vxli_offset2 - Compute element number for given subscripts

Given the rank, dimensions, and subscript values for each coordinate,
this variant of the vxli_offset function computes the offset or
element number selected by the given coordinate values, measured in
elements (not bytes).  The length of the coords array must equal the
rank of the voxel array.  If the given coordinates are outside the
voxel array, then a negative offset is returned.  NO validity checking
is done, so this version is suitable only for use within the voxel kit
itself.

SEE ALSO: vxli_offset() to compute element number given a voxel array.
--------------------------------------------------------------------------*/

inlineable
long                         /* Returns NEGATIVE offset on error */
vxli_offset2(
    long rank,               /* Number of dimensions */
    long *dimen,             /* Number of elements along each dimension */
    long *coords)            /* Coordinate values for each dimension */
{
    long step;               /* Step size for current coordinate */
    long offset;             /* Offset computed so far */
    long i;

#if 0    
    /* Check for valid arguments; REMOVED 2002-09-02  */
    if (rank < 0 || dimen == NULL || coords == NULL) {
        set_error(ERR_INIT, vxli_errmsg_badinfo);
        return -1;
    }
#endif

    /* Compute the desired offset */
    offset = 0;
    step = 1;
    for (i = rank - 1; i >= 0; i--) {
	if (coords[i] < 0 || coords[i] >= dimen[i]) {
            preset_error(ERR_INDEX, vxli_errmsg_subscript);
            return -1;
	}
	offset += coords[i]*step;
	step *= dimen[i];
    }

    return offset;
}



/*-------------------------------------------------------------------------
vxli_step - Compute step size in a given dimension

Given a voxel array, this function computes the step size or stride
that corresponds to an increment of 1 in a given dimension of the
array, measured in elements (not bytes).  NO validity checking is done,
so this should be used only when the array has already been validated.
--------------------------------------------------------------------------*/

inlineable
long
vxli_step(
    voxel_array *src,       /* Voxel array to compute for */
    int dim)                /* Dimension to compute for (0-base) */
{
    long step = 1;          /* Desired step size, so far */
    int i;

#if 0
    /* Basic validity check; REMOVED 2002-09-02 */
    if (src == 0 || src->magic != VXL_MAGIC)
	errm_panic("vxli_step: Invalid voxel array as argument");
#endif

    /* Compute the desired step size */
    for (i = src->rank - 1; i > dim; i--)
	step *= src->dimen[i];

    return step;
}



/*-------------------------------------------------------------------------
vxli_step2 - Compute step size in a given dimension

Given the rank and dimensions of a voxel array, this variant of the
vxli_step function computes the step size or stride that corresponds
to an increment of 1 in a given dimension of the array, measured in
elements (not bytes).  NO validity checking is done, so this should be
used only when the array has already been validated.
--------------------------------------------------------------------------*/

inlineable
long
vxli_step2(
    long rank,              /* Number of dimensions */
    long *dimen,            /* Number of elements along each dimension */
    int dim)                /* Dimension to compute for (0-base) */
{
    long step = 1;          /* Desired step size, so far */
    int i;

#if 0
    /* Basic validity check; REMOVED 2002-09-02 */
    if (rank < 0 || dimen == NULL || dim < 0 || dim >= rank)
	errm_panic("vxli_step2: Invalid arguments");
#endif

    /* Compute the desired step size */
    for (i = rank - 1; i > dim; i--)
	step *= dimen[i];

    return step;
}



/*------------------------------------------------------------------------
vxli_locate - Compute pointer to element with given subscripts

Given an voxel array, and subscript values for each coordinate, this
function computes an unsigned char pointer to the element selected by
the given coordinate values.  The length of the coords array must
equal the rank of the voxel array.  If the given coordinates are
outside the voxel array, then the error message is set and a null
pointer is returned.  The voxel array pointer given is NOT checked for
validity, so this version is suitable only for use within the voxel
kit itself.  In fact, the most common use of this function is to
compute the base address of a scanline or slice to hand off to a bips
function.

SEE ALSO: vxl_locate, which checks the array for validity.

FIXME: Is there a use for a version that assumes both the array and
coords are valid?  Ditto for vxli_offset.
--------------------------------------------------------------------------*/

inlineable
unsigned char *              /* Returns NULL on error */
vxli_locate(
    voxel_array *src,        /* Voxel array to define dimensions */
    long *coords)            /* Coordinate values */
{
    long step;               /* Step size for current coordinate */
    long offset;             /* Offset computed so far */
    unsigned char *point;    /* Pointer to the desired element */
    int i;

    /* Compute the desired offset */
    offset = 0;
    step   = 1;
    for (i = src->rank - 1; i >= 0; i--) {
	if (coords[i] < 0 || coords[i] >= src->dimen[i])
            ZReturnError(ERR_INDEX, vxli_errmsg_subscript);
	offset += coords[i]*step;
	step *= src->dimen[i];
    }

    point = (unsigned char *)src->data + (src->nbytes)*offset;
    return point;
}



/*------------------------------------------------------------------------
vxli_offset_nocheck - Compute element number for given subscripts

Given an voxel array, and subscript values for each coordinate, this
function computes the offset or element number selected by the given
coordinate values, measured in elements (not bytes).  The length of
the coords array must equal the rank of the voxel array.  If the given
coordinates are outside the voxel array, the results are unpredictable,
up to and including a segmentation fault; it is assumed that the
source array and coordinates have already been checked for validity.
--------------------------------------------------------------------------*/

inlineable long
vxli_offset_nocheck(
    voxel_array *src,        /* Voxel array to define dimensions */
    long *coords)            /* Coordinate values */
{
    long step;               /* Step size for current coordinate */
    long offset;             /* Offset computed so far */
    int i;
    
    /* Compute the desired offset */
    offset = 0;
    step = 1;
    for (i = src->rank - 1; i >= 0; i--) {
	offset += coords[i]*step;
	step *= src->dimen[i];
    }

    return offset;
}



/*-------------------------------------------------------------------------
vxli_same_shape - Check whether rank and dimensions match

Given two voxel arrays, this function returns 1 if they have the same
rank and dimensions, or at least one of the arrays has rank 0;
otherwise it presets a suitable error message and returns zero.  The
type is not checked, and it is assumed that the arrays have already be
checked for existence and validity at a higher level, so that this
routine is not suitable to be called directly by the Voxel Kit user.

SEE ALSO vxli_same_physical() to also check that the physical origin
and spacing are consistent.

NOTE: This doesn't seem to be used anywhere (vxli_same_physical() is used
instead) but is being kept around for possible later use.
------------------------------------------------------------------------*/

inlineable
int                              /* Returns zero on mismatch */
vxli_same_shape(
    voxel_array *src1,           /* First array */
    voxel_array *src2)           /* Second array */
{
    int i;

    /* If either array has rank 0, retun 1 */
    if (src1->rank == 0 || src2->rank == 0)
        return 1;

    /* If rank or dimen doesn't match, return 0 */
    if (src1->rank != src2->rank)   {
        preset_error(ERR_INCOMP, vxli_errmsg_diffrank);
        return 0;
    }
    for (i = 0; i < src1->rank; i++)
	if (src1->dimen[i] != src2->dimen[i]) {
            preset_error(ERR_INCOMP, vxli_errmsg_diffshape);
	    return 0;
        }

    /* Otherwise return 1 */
    return 1;
}



/*-------------------------------------------------------------------------
vxli_same_physical - Check whether shape and physical coordinates match

Given two voxel arrays, this function returns 1 if they have the same
rank, dimensions, and physical coordinates, or if at least one of the
voxel arrays has rank 0; otherwise it presets a suitable error message
and returns zero.  The type is not checked, and it is assumed that the
arrays have already been checked for existence and validity at a
higher level, so that this routine is not suitable to be called
directly by the Voxel Kit user.

SEE ALSO vxli_same_shape() to check only the rank and shape.
------------------------------------------------------------------------*/

inlineable
int                              /* Returns zero on mismatch */
vxli_same_physical(
    voxel_array *src1,           /* First array */
    voxel_array *src2)           /* Second array */
{
    int i;

    /* If either array has rank 0, return 1 */
    if (src1->rank == 0 || src2->rank == 0)
        return 1;

    /* If rank doesn't match, return 0 */
    if (src1->rank != src2->rank)  {
        preset_error(ERR_INCOMP, vxli_errmsg_diffrank);
        return 0;
    }

    /* If dimensions, origin, or spacing don't match, return 0 */
    for (i = 0; i < src1->rank; i++) {
	if (src1->dimen[i] != src2->dimen[i]) {
            preset_error(ERR_INCOMP, vxli_errmsg_diffshape);
	    return 0;
        }
	if (src1->origin[i] != src2->origin[i]) {
            preset_error(ERR_INCOMP, vxli_errmsg_difforg);
	    return 0;
        }
	if (src1->spacing[i] != src2->spacing[i]) {
            preset_error(ERR_INCOMP, vxli_errmsg_diffspace);
	    return 0;
        }
    }

    /* Otherwise return 1 */
    return 1;
}



/*-------------------------------------------------------------------------
vxli_default_physical - Set origin and spacing to their default values

This function sets the origin and spacing parameters of an array to
their default values.  Note that vxl_alloc_array() presets the origin
and spacing to their default values, and this function is not needed
in addition.
--------------------------------------------------------------------------*/

inlineable void
vxli_default_physical(
    voxel_array *dest)                  /* Destination array */
{
    int i;

    for (i = 0; i < VXL_MAX_RANK; i++) {
        dest->origin [i] = 0.0;
	dest->spacing[i] = 1.0;
    }

    return;
}



/*-------------------------------------------------------------------------
vxli_set_physical - Set origin and spacing from explicit values

This function sets the origin and spacing parameters of an array from 
explicit values.
--------------------------------------------------------------------------*/

inlineable void
vxli_set_physical(
    voxel_array *dest,                  /* Destination array */
    double *origin,                     /* Origin values */
    double *spacing)                    /* Spacing values */
{
    int i;

    for (i = 0; i < VXL_MAX_RANK; i++) {
        dest->origin [i] = origin [i];
	dest->spacing[i] = spacing[i];
    }

    return;
}



/*-------------------------------------------------------------------------
vxli_copy_physical - Copy origin and spacing from one array to another

This function copies the origin and spacing parameters from one voxel array
to another.
--------------------------------------------------------------------------*/

inlineable void
vxli_copy_physical(
    voxel_array *dest,                  /* Destination array */
    voxel_array *src)                   /* Source array */
{
    int i;

    for (i = 0; i < VXL_MAX_RANK; i++) {
        dest->origin [i] = src->origin [i];
	dest->spacing[i] = src->spacing[i];
    }

    return;
}



/*------------------------------------------------------------------------
vxli_firstvoxel - Initialize iteration over voxels of an image

This function initializes a coordinate array prior to iterating over
the voxels of an image with vxli_nextvoxel; voxels are selected in
memory order, with the last coordinate varying most rapidly.  Only
elementary error checking is done in vxli_firstvoxel and
vxli_nextvoxel; these functions should be called only in contexts
where the array has already been checked for validity.  Note further
that in most applications, it is better to iterate over rows and use
the optimized BIPS functions to process voxels within a row; this
function is intended primarily for reference implementations and
similar contexts where clarity is more important than speed.

Assuming that the special case of an empty array has been excluded, 
the canonical loop over voxels looks like this:

voxel_array *array;             // Array to iterate through
long coord[VXL_MAX_RANK];       // Coords of first voxel in row

vxli_firstvoxel(array->rank, coord);
do {
    ...operate on a voxel...
} 
while (vxli_nextvoxel(array->rank, array->dimen, coord));
--------------------------------------------------------------------------*/

inlineable void
vxli_firstvoxel(
    int rank,                   /* Rank of the array */
    long *coord)                /* Voxel coordinates */
{
    int i;
    for (i = 0; i < rank; i++)
	coord[i] = 0;
    return;
}



/*-------------------------------------------------------------------------
vxli_nextvoxel - Compute coordinates of next row in an image

Given a voxel array and a coordinate array pointing to some particular
voxel in an image, this function updates the coordinate array to point
to the next voxel.  This function returns zero when all rows have been
seen.  See vxli_firstvoxel for usage information.
-------------------------------------------------------------------------*/

inlineable int
vxli_nextvoxel(
    int rank,                 /* Rank of the voxel array */
    long *dimen,              /* Dimensions of the voxel array */
    long *coord)              /* Coordinates of the current row */
{
    int i;

    for (i = rank-1; i >= 0; i--) {
	if ( ++coord[i] < dimen[i] )
	    return 1;
	else
	    coord[i] = 0;
    }

    return 0;
}



/*------------------------------------------------------------------------
vxli_firstrow - Initialize iteration over rows of an image

This function initializes a coordinate array prior to iterating over
the rows of an image with vxli_nextrow, where a row is defined here as
a sequence of voxels all having their first (N-1) coordinates the same
and differing only in their last coordinate.  The rows are guaranteed
to be generated in forward order, with coordinate values increasing;
the index of the last dimension is always set to zero.  Only elementary
error checking is done in vxli_firstrow and vxli_nextrow; these
functions should be called only in contexts where the array has
already been checked for validity.

Iterating over the rows of an image is an important operation in the
Voxel Kit, because most loop nests are organized with the last
dimension as the innermost loop and the outer dimensions as the outer
loops, with other loops put between.  See the implementation of
convolution as an example.  

Assuming that the special case of an empty array has been excluded, 
the canonical loop over rows looks like this:

voxel_array *array;             // Array to iterate through
long coord[VXL_MAX_RANK];       // Coords of first voxel in row

vxli_firstrow(array->rank, coord);
do {
    ...operate on a row...
} 
while (vxli_nextrow(array->rank, array->dimen, coord));
--------------------------------------------------------------------------*/

inlineable void
vxli_firstrow(
    int rank,                   /* Rank of the array */
    long *coord)                /* Voxel coordinates */
{
    int i;
    for (i = 0; i < rank; i++)
	coord[i] = 0;
    return;
}



/*-------------------------------------------------------------------------
vxli_nextrow - Compute coordinates of next row in an image

Given a voxel array and a coordinate array pointing to some particular
row in an image, this function updates the coordinate array to point
to the next row.  This function returns zero when all rows
have been seen.  See vxli_firstrow for usage information.
-------------------------------------------------------------------------*/

inlineable int
vxli_nextrow(
    int rank,                 /* Rank of the voxel array */
    long *dimen,              /* Dimensions of the voxel array */
    long *coord)              /* Coordinates of the current row */
{
    int i;

    for (i = rank-2; i >= 0; i--) {
	if ( ++coord[i] < dimen[i] )
	    return 1;
	else
	    coord[i] = 0;
    }

    return 0;
}



/*------------------------------------------------------------------------
vxli_lastrow - Initialize reversed iteration over rows of an image

This function initializes a coordinate array prior to iterating over
the rows of an image in reverse order using vxli_prevrow.  A row is
defined here as a sequence of voxels all having their first (N-1)
coordinates the same and differing only in their last coordinate.  It
is similar to vxli_firstrow and vxli_nextrow but generates the rows in
reverse order; that is, with coordinate values decreasing.  The index
of the last dimension is set to the last pixel in that row.  Only
elementary error checking is done in vxli_lastrow and vxli_prevrow;
these functions should be called only in contexts where the array has
already been checked for validity.

Assuming that the special case of an empty array has been excluded, 
the canonical loop over rows looks like this:

voxel_array *array;             // Array to iterate through
long coord[VXL_MAX_RANK];       // Coords of last voxel in row

vxli_lastrow(array->rank, array->dimen, coord);
do {
    ...operate on a row...
} 
while (vxli_prevrow(array->rank, array->dimen, coord));
--------------------------------------------------------------------------*/

inlineable void
vxli_lastrow(
    int rank,                   /* Rank of the array */
    long *dimen,                /* Dimensions of the array */
    long *coord)                /* Voxel coordinates */
{
    int i;
    for (i = 0; i < rank; i++)
	coord[i] = dimen[i] - 1;
    return;
}



/*-------------------------------------------------------------------------
vxli_prevrow - Compute coordinates of previous row in an image

Given a voxel array and a coordinate array pointing to some particular
row in an image, this function updates the coordinate array to point
to the previous row.  This function returns zero when all rows
have been seen.  See vxli_lastrow for usage information.
-------------------------------------------------------------------------*/

inlineable int
vxli_prevrow(
    int rank,                 /* Rank of the voxel array */
    long *dimen,              /* Dimensions of the voxel array */
    long *coord)              /* Coordinates of the current row */
{
    int i;

    for (i = rank-2; i >= 0; i--) {
	if ( --coord[i] >= 0 )
	    return 1;
	else
	    coord[i] = dimen[i] - 1;
    }

    return 0;
}



/*------------------------------------------------------------------------
vxli_firstrow2 - Initialize iteration over generalized rows of an image

This function initializes a coordinate array prior to iterating over
the generalized rows of an image with vxli_nextrow2.  A generalized
row is defined here as a sequence of voxels which have N-1 of their
coordinates the same and cover all possible values in the remaining
coordinate.  It generalizes the concept of a row to any dimension of
an array, rather than always the last dimension.  Generalized rows are
used much less often than plain rows but are necessary for some
operations, notably the FFT and chamfer distance transform.  The
generalized rows are guaranteed to be generated in ascending order by
coordinates.  Only elementary error checking is done in vxli_firstrow2
and vxli_nextrow2; these functions should be called only in contexts
where the array has already been checked for validity.

Assuming that the special case of an empty array has been excluded,
the canonical loop over the generalized rows along the ith dimension
looks like this:

voxel_array *array;             // Array to iterate through
long coord[VXL_MAX_RANK];       // Coords of first voxel in row

vxli_firstrow2(array->rank, coord);
...initialize as required...
do {
    ...operate on a generalized row...
} 
while (vxli_nextrow2(array->rank, array->dimen, idim, coord));
--------------------------------------------------------------------------*/

inlineable void
vxli_firstrow2(
    int rank,                   /* Rank of the array */
    long *coord)                /* Voxel coordinates */
{
    int i;
    for (i = 0; i < rank; i++)
	coord[i] = 0;
    return;
}



/*-------------------------------------------------------------------------
vxli_nextrow2 - Compute coordinates of next row in an image

Given a voxel array and a coordinate array pointing to some particular
row in an image, this function updates the coordinate array to point
to the next generalized row along the idim-th direction.  This
function returns zero when all rows have been seen.  See vxli_firstrow2
for usage information.
-------------------------------------------------------------------------*/

inlineable int
vxli_nextrow2(
    int rank,                 /* Rank of the voxel array */
    long *dimen,              /* Dimensions of the voxel array */
    int idim,                 /* Dimension to generate rows along */
    long *coord)              /* Coordinates of the current row */
{
    int i;

    for (i = rank-1; i >= 0; i--) {
	if ( i == idim )
	    continue;
	else if ( ++coord[i] < dimen[i] )
	    return 1;
	else
	    coord[i] = 0;
    }

    return 0;
}
