/****************************************************************************
bips.c - Basic Image Processing Subroutines

Author:  Paul Hughett

Copyright (c) 1999-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This file is part of the Voxel Kit, a set of general-purpose functions
for volume image processing, with a bias toward medical imaging
applications.  It contains the basic image processing subroutines,
which is a relatively small set of functions that operate on
one-dimensional vectors of image data and can reasonably be optimized
by hand for maximum performance and from which higher-level
multidimensional operations can be constructed without losing too much
performance.  They are intended primarily to be called within the Voxel
Kit itself, rather than by a user of the Voxel Kit.

Unless otherwise specified, the results are undefined if the
destintion stride is zero or if the source and destination, or two
destinations, overlap (that is, if they share any elements;
interleaving with a non-zero stride should cause no problems).  The
reason for this and similar restrictions is that a good compiler can
often generate more efficient code if it can assume that pointers
always point to distinct objects; thus we require that you do the
check for aliasing before calling these functions.  The source stride
may generally be zero, but this is only occasionally useful.

Note that, for efficiency, these functions assume that all necessary
error checking has been done by the caller and do only the error
checking that they absolutely must (typically only that the requested
operation is in fact implemented).  This is one of the reasons that
they should not be called by the user directly.

As a general rule, the source and destination addresses are specified
as void or unsigned char pointers, but strides are given in elements
of the specified type.  This mixed specification might seem confusing
but seems to work reasonably well in practice.

The priorities for these functions has been portability and efficiency
when computing large (e.g. 16 million elements) image data sets;
readability has been a poor third.  This means that means that the
code looks rather ugly, since all the decisions have been hoisted
out of the main loop(s), leaving many slight variations on the same
loop, which is often defined as a parameterized macro to save typing.
Note that these macros are not general-purpose and usually intended
for use ONLY in the function that they just precede.

There are a few BIPS operations for which complex types are converted
to real types; for instance, the real(), imag(), abs(), and arg()
functions.  This exim_real_type() function defines the mapping from
input type to output type for these functions.  But note that, for
technical reasons, the BIPS functions themselves do NOT use that
function, so it is important to ensure that they remain consistent.

This file should be compiled as C only, not as C++.  It makes heavy
use of some constructs that cause a C++ compiler to choke; in any
case, the right way to do this under C++ would be to use templates
instead of macros.  There should not be any problems calling these
routines from C++.

FIXME: The "necessary" set of operations supported in bips is
beginning to seem rather large.  Is there a better approach?

FIXME: These functions in most cases simply reproduce the arithmetic
behavior of the underlying C implementation.  Perhaps there should
be some tools for testing for unexpected behaviors and warning the
users (like the paranoia package for floating point, but also checking
for integer anomalies?).

FIXME: Add other operations, e.g. shifts, etc...  Add ternary
operations, like mask/mix, daxpy, fused multiply/add (w += x*y?).  Add
an exchange operation for transpose and related array operations.  Add
an inner loop for global statistics, and IIR (AR or ARMA?)  filters.

FIXME: Add the new types and functions provided by C99, if supported.
These include isnan, isinf, long long, long double.

FIXME: Add the const and restrict qualifiers to formal arguments as
appropriate.

FIXME: Add appropriately scaled versions of the transcedental
functions for integral types, after the fashion of analog computers.
These could be useful for fast but not very accurate computation.  Of
course, you may need to explain the theory of fixed point computation,
since analog computers are so obscure these days.

FIXME: It might possibly make sense to support RGB and RGBA as a data
type, although this might conflict with the method of treating
echo/frequency/band as the last dimension of an array.
*****************************************************************************/

static const char rcsid[] = "$Id: bips.c,v 1.79 2005/12/03 19:07:57 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include <float.h>
#include <limits.h>
#include <stdlib.h>

#ifdef HAVE_TGMATH_H
#include <tgmath.h>
#else
#include <math.h>
#endif

#ifdef HAVE_COMPLEX_H
#include <complex.h>

/* Abbreviations for some long type names */
typedef double complex dcomplex;
typedef float  complex fcomplex;
#endif

#include "bips.h"




/*------------------------------------------------------------------------
Error messages
------------------------------------------------------------------------*/

static char illtype[] = "Illegal data type for BIPS operation";
static char unimp  [] = "Unimplemented BIPS operation";



/**************************************************************************
Internal utility functions

To be defined as needed...
**************************************************************************/

/* Limit and round to an integral destination type */
/* FIXME: floor() is rather slow for gcc on i386 */
#define LIMIT(x,lo,hi)   \
    ( ((x) < (lo)) ? (lo) : ((x) > (hi)) ? (hi) : floor((x) + 0.5) )



/*************************************************************************
Utility functions
*************************************************************************/


/*------------------------------------------------------------------------
bips_locate - Compute address of a vector element

This function returns the address of the vector element numbered i.
No subscript range checking is done.  If the type is invalid, then a
null pointer is returned.  This function should be used only in
circumstances where the caller does most of the validity checking.
-------------------------------------------------------------------------*/

#define LOCATE(type)          \
{                             \
    type *s = (type *) src;   \
    p = (void *) (s + i);     \
}

void *
bips_locate(
    int type,              /* Type of data */
    void *src,             /* Source array */
    long i)                /* Element number to find address of */
{
    void *p;

    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	LOCATE(unsigned char);
	return p;

    case INTP_USHORT:
	LOCATE(unsigned short);
	return p;

    case INTP_UINT:
	LOCATE(unsigned int);
	return p;

    case INTP_ULONG:
	LOCATE(unsigned long);
	return p;

    case INTP_SCHAR:
	LOCATE(signed char);
	return p;

    case INTP_SHORT:
	LOCATE(short);
	return p;

    case INTP_INT:
	LOCATE(int);
	return p;

    case INTP_LONG:
	LOCATE(long);
	return p;

    case INTP_FLOAT:
	LOCATE(float);
	return p;

    case INTP_DOUBLE:
	LOCATE(double);
	return p;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	LOCATE(float complex);
	return p;

    case INTP_DCOMPLEX:
	LOCATE(double complex);
	return p;
#endif

    default:
	/* Drop through to error return */ ;

    }

    return ERR_OKAY;    
}



/*------------------------------------------------------------------------
bips_real_type - Return real type corresponding to given input type

This function was replaced on 2005-09-21 by the exim_real_type() function,
which is its exact equivalent.

There are a few BIPS operations for which complex types are converted to
real types; for instance, the real(), imag(), abs(), and arg() function.
This function defines the mapping from input type to output type for these
functions.

NOTE: For technical reasons, the BIPS functions themselves do NOT use this
function, so it is important to ensure that they remain consistent.
--------------------------------------------------------------------------*/




/***************************************************************************
Nullary point operations
***************************************************************************/



/*------------------------------------------------------------------------
bips_zero - Fill with zeros

NOTE: This operation could be done by bips_copy, but occurs frequently
and it's a nuisance to have to create a zero in the correct type each
time.
-------------------------------------------------------------------------*/

#define ZERO_LOOP(type)             \
{                                   \
    type *d = (type *)dst;          \
    for (i = 0; i < nelem; i++) {   \
	*d = 0;   d += ds;          \
	}                           \
}

int
bips_zero(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds)                 /* Stride in source array */
{
    long i;

    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	ZERO_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_USHORT:
	ZERO_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_UINT:
	ZERO_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_ULONG:
	ZERO_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_SCHAR:
	ZERO_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	ZERO_LOOP(short);
	return ERR_OKAY;

    case INTP_INT:
	ZERO_LOOP(int);
	return ERR_OKAY;

    case INTP_LONG:
	ZERO_LOOP(long);
	return ERR_OKAY;

    case INTP_FLOAT:
	ZERO_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	ZERO_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	ZERO_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	ZERO_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_one - Fill with ones

This function fills the destination with the value 1 or 1+0j for
complex types.

NOTE: This operation could be done by bips_copy, but occurs frequently
and it's a nuisance to have to create a one in the correct type each
time.
-------------------------------------------------------------------------*/

#define ONE_LOOP(type)              \
{                                   \
    type *d = (type *)dst;          \
    for (i = 0; i < nelem; i++) {   \
	*d = 1;   d += ds;          \
	}                           \
}

int
bips_one(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds)                 /* Stride in source array */
{
    long i;

    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	ONE_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_USHORT:
	ONE_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_UINT:
	ONE_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_ULONG:
	ONE_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_SCHAR:
	ONE_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	ONE_LOOP(short);
	return ERR_OKAY;

    case INTP_INT:
	ONE_LOOP(int);
	return ERR_OKAY;

    case INTP_LONG:
	ONE_LOOP(long);
	return ERR_OKAY;

    case INTP_FLOAT:
	ONE_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	ONE_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	ONE_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	ONE_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_ffff - Fill with the all-ones binary value

Supported only for integral types.  

FIXME: Signed integral types might not work right except on two's
complement platforms.
-------------------------------------------------------------------------*/

#define FFFF_LOOP(type)             \
{                                   \
    type *d = (type *)dst;          \
    for (i = 0; i < nelem; i++) {   \
	*d = -1;   d += ds;         \
	}                           \
}

int
bips_ffff(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds)                 /* Stride in source array */
{
    long i;

    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	FFFF_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_USHORT:
	FFFF_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_UINT:
	FFFF_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_ULONG:
	FFFF_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_SCHAR:
	FFFF_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	FFFF_LOOP(short);
	return ERR_OKAY;

    case INTP_INT:
	FFFF_LOOP(int);
	return ERR_OKAY;

    case INTP_LONG:
	FFFF_LOOP(long);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_fill_min - Fill with minimum representable value for destination type

This function fills the given array with the minimum representable
value for the type, not including -inf.  It may be useful for
initializing a max calculation.  This function is not defined for
the complex types, because there is no order of the complex numbers
which is consistent with their arithmetic properties.

NOTE: This operation could be done by bips_copy, but occurs frequently
and it's a nuisance to have to create the min value in the correct
type each time.
-------------------------------------------------------------------------*/

#define FILL_MIN_LOOP(type, minval)   \
{                                     \
    type *d = (type *)dst;            \
    for (i = 0; i < nelem; i++) {     \
	*d = minval;   d += ds;       \
	}                             \
}

int
bips_fill_min(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds)                 /* Stride in source array */
{
    long i;

    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	FILL_MIN_LOOP(unsigned char, 0);
	return ERR_OKAY;

    case INTP_USHORT:
	FILL_MIN_LOOP(unsigned short, 0);
	return ERR_OKAY;

    case INTP_UINT:
	FILL_MIN_LOOP(unsigned int, 0);
	return ERR_OKAY;

    case INTP_ULONG:
	FILL_MIN_LOOP(unsigned long, 0);
	return ERR_OKAY;

    case INTP_SCHAR:
	FILL_MIN_LOOP(signed char, SCHAR_MIN);
	return ERR_OKAY;

    case INTP_SHORT:
	FILL_MIN_LOOP(short, SHRT_MIN);
	return ERR_OKAY;

    case INTP_INT:
	FILL_MIN_LOOP(int, INT_MIN);
	return ERR_OKAY;

    case INTP_LONG:
	FILL_MIN_LOOP(long, LONG_MIN);
	return ERR_OKAY;

    case INTP_FLOAT:
	FILL_MIN_LOOP(float, -FLT_MAX);
	return ERR_OKAY;

    case INTP_DOUBLE:
	FILL_MIN_LOOP(double, -DBL_MAX);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_fill_max - Fill with maximum possible value for destination type

This function fills the given array with the maximum representable
value for the type, not including +inf.  It may be useful for
initializing a max calculation.  This function is not defined for
the complex types, because there is no order of the complex numbers
which is consistent with their arithmetic properties.

NOTE: This operation could be done by bips_copy, but occurs frequently
and it's a nuisance to have to create the max value in the correct
type each time.
-------------------------------------------------------------------------*/

#define FILL_MAX_LOOP(type, maxval)   \
{                                     \
    type *d = (type *)dst;            \
    for (i = 0; i < nelem; i++) {     \
	*d = maxval;   d += ds;       \
	}                             \
}

int
bips_fill_max(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds)                 /* Stride in source array */
{
    long i;

    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	FILL_MAX_LOOP(unsigned char, UCHAR_MAX);
	return ERR_OKAY;

    case INTP_USHORT:
	FILL_MAX_LOOP(unsigned short, USHRT_MAX);
	return ERR_OKAY;

    case INTP_UINT:
	FILL_MAX_LOOP(unsigned int, UINT_MAX);
	return ERR_OKAY;

    case INTP_ULONG:
	FILL_MAX_LOOP(unsigned long, ULONG_MAX);
	return ERR_OKAY;

    case INTP_SCHAR:
	FILL_MAX_LOOP(signed char, SCHAR_MAX);
	return ERR_OKAY;

    case INTP_SHORT:
	FILL_MAX_LOOP(short, SHRT_MAX);
	return ERR_OKAY;

    case INTP_INT:
	FILL_MAX_LOOP(int, INT_MAX);
	return ERR_OKAY;

    case INTP_LONG:
	FILL_MAX_LOOP(long, LONG_MAX);
	return ERR_OKAY;

    case INTP_FLOAT:
	FILL_MAX_LOOP(float, FLT_MAX);
	return ERR_OKAY;

    case INTP_DOUBLE:
	FILL_MAX_LOOP(double, DBL_MAX);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_zero_range - Fill with zeros within limits

This function sets to zero the vector elements numbered first*ds,
(first+1)*ds, ..., (limit-1)*ds.  It's sort of irregular, but useful
in some specialized circumstances.

NOTE: This operation could be done by bips_copy, but it's a nuisance to 
have to create a zero in the correct type each time and to decide exactly 
where to start.

FIXME: Consider integrating this functionality into bips_fill_zero.
-------------------------------------------------------------------------*/

#define ZERO_RANGE(type)                \
{                                       \
    type *d = (type *)dst;              \
    d += first * ds;                    \
    for (i = first; i < limit; i++) {   \
	*d = 0;   d += ds;              \
	}                               \
}

int
bips_zero_range(
    long first,              /* First element to zero */
    long limit,              /* Limit element to zero */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds)                 /* Stride in source array */
{
    long i;

    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	ZERO_RANGE(unsigned char);
	return ERR_OKAY;

    case INTP_USHORT:
	ZERO_RANGE(unsigned short);
	return ERR_OKAY;

    case INTP_UINT:
	ZERO_RANGE(unsigned int);
	return ERR_OKAY;

    case INTP_ULONG:
	ZERO_RANGE(unsigned long);
	return ERR_OKAY;

    case INTP_SCHAR:
	ZERO_RANGE(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	ZERO_RANGE(short);
	return ERR_OKAY;

    case INTP_INT:
	ZERO_RANGE(int);
	return ERR_OKAY;

    case INTP_LONG:
	ZERO_RANGE(long);
	return ERR_OKAY;

    case INTP_FLOAT:
	ZERO_RANGE(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	ZERO_RANGE(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	ZERO_RANGE(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	ZERO_RANGE(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_fill_range - Fill with given value within limits

This function sets the vector elements numbered first*ds, (first+1)*ds, 
..., (limit-1)*ds to a value given as a double.  It's sort of irregular, 
but often useful.

NOTE: This operation could be done by bips_copy, but it's a nuisance to 
have to create a value in the correct type each time and to decide exactly 
where to start.

FIXME: Could this be replaced by another fill function?  Where is it used?

FIXME: Is there a use case for complex-valued fill values?
-------------------------------------------------------------------------*/

#define FILL_RANGE(type)                \
{                                       \
    type  v = (type)value;              \
    type *d = (type *)dst;              \
    d += first * ds;                    \
    for (i = first; i < limit; i++) {   \
	*d = v;   d += ds;              \
	}                               \
}

int
bips_fill_range(
    long first,              /* First element to zero */
    long limit,              /* Limit element to zero */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in source array */
    double value)            /* Value to fill with */
{
    long i;

    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	FILL_RANGE(unsigned char);
	return ERR_OKAY;

    case INTP_USHORT:
	FILL_RANGE(unsigned short);
	return ERR_OKAY;

    case INTP_UINT:
	FILL_RANGE(unsigned int);
	return ERR_OKAY;

    case INTP_ULONG:
	FILL_RANGE(unsigned long);
	return ERR_OKAY;

    case INTP_SCHAR:
	FILL_RANGE(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	FILL_RANGE(short);
	return ERR_OKAY;

    case INTP_INT:
	FILL_RANGE(int);
	return ERR_OKAY;

    case INTP_LONG:
	FILL_RANGE(long);
	return ERR_OKAY;

    case INTP_FLOAT:
	FILL_RANGE(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	FILL_RANGE(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	FILL_RANGE(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	FILL_RANGE(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_ramp - Generate a linear ramp

This BIPS function stores the sequence 0, 1, 2, ... into the
destination.  The count is generated as a (signed) long value and cast
(using the usual C rules) into the destination type; this means that
the count may wrap around or otherwise misbehave if the destination
type cannot contain the count value.  If the number of elements is not
positive, this function returns immediately.  If the destination type is
complex, the real part contains the count and the imaginary part is zero.
-------------------------------------------------------------------------*/

#define RAMP_LOOP(type)            \
{                                  \
    type *d = (type *)dst;         \
    for (i = 0; i < nelem; i++) {  \
        *d = (type)i;              \
        d += ds;    }              \
}

int
bips_ramp(
    long nelem,            /* Number of elements to store */
    int type,              /* Type of the destination array */
    void *dst,             /* Destination array */
    long ds)               /* Stride in the destination */
{
    long i;

    /* Return immediately if negative number of elements */
    if (nelem < 0) return ERR_OKAY;

    /* Switch according to the destination type */
    switch (type) {

    case INTP_UCHAR:
	RAMP_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	RAMP_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	RAMP_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	RAMP_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	RAMP_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	RAMP_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	RAMP_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	RAMP_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	RAMP_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	RAMP_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	RAMP_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	RAMP_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/**************************************************************************
Casts and other type conversions

These functions read a sequence of values in some specified internal
format and convert them to another (internal) format specified either
by the name of the function called or as a parameter.  Casting a type
to itself is permitted for uniformity and is probably almost as fast
as a plain copy.  The source and destination may have arbitrary
strides, except that the destination stride may not be zero.  The
results are unpredictable if thesource and destination arrays overlap.
The details of the conversion (overflow, exceptions, etc) are
inherited from the underlying C implementation.

See the exim library for conversions between internal and external data
formats.
**************************************************************************/


/*------------------------------------------------------------------------
CAST_LOOP - Generic loop for a type conversion operation

This macro defines a generic loop for implementing a type conversion
operation (or copy) where the source and destination arrays may have
arbitrary strides, except that the destination stride may not be zero.
The results are undefined if the source and destination arrays overlap.

FIXME: Provide special loop for the case that ss = 0?  Or assume that
the compiler is smart enough to do it?

FIXME: See if the case nelem == 1 can be handled more efficiently; it
seems to come up fairly often.
------------------------------------------------------------------------*/

#define CAST_LOOP(type)                   \
{                                         \
    type * restrict s = (type *)src;      \
    for (i = 0; i < nelem; i++ ) {        \
        *d = *s;                          \
        d += ds;  s += ss; }              \
}


/*------------------------------------------------------------------------
CMPLX_LOOP - Generic loop for casting real and imaginary parts to complex

This macro defines a generic loop for implementing a type conversion
operation (or copy) where real and imaginary parts are cast to form a
complex number. The source and destination arrays may have arbitrary
strides, except that the destination stride may not be zero.  The
results are undefined if the source and destination arrays overlap.
------------------------------------------------------------------------*/

#define CMPLX_LOOP(type)                   \
{                                          \
    type * restrict s1 = (type *)src1;     \
    type * restrict s2 = (type *)src2;     \
    for (i = 0; i < nelem; i++ ) {         \
        *d = *s1 + I*(*s2);                \
        d += ds;  s1 += ss1; s2 += ss2; }  \
}



/*------------------------------------------------------------------------
bips_uchar - Convert voxel array to unsigned char
-------------------------------------------------------------------------*/

int
bips_uchar(
    long nelem,             /* Number of elements to convert */
    unsigned char *dst,     /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    unsigned char * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CAST_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CAST_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CAST_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	CAST_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CAST_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*------------------------------------------------------------------------
bips_schar - Convert voxel array to signed char
-------------------------------------------------------------------------*/

int
bips_schar(
    long nelem,             /* Number of elements to convert */
    signed char *dst,       /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    signed char * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CAST_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CAST_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CAST_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	CAST_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CAST_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*------------------------------------------------------------------------
bips_ushort - Convert voxel array to unsigned short
-------------------------------------------------------------------------*/

int
bips_ushort(
    long nelem,             /* Number of elements to convert */
    unsigned short *dst,    /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    unsigned short * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CAST_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CAST_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CAST_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	CAST_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CAST_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*------------------------------------------------------------------------
bips_short - Convert voxel array to short
-------------------------------------------------------------------------*/

int
bips_short(
    long nelem,             /* Number of elements to convert */
    short *dst,             /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    short * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CAST_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CAST_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CAST_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	CAST_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CAST_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*------------------------------------------------------------------------
bips_int - Convert voxel array to int
-------------------------------------------------------------------------*/

int
bips_int(
    long nelem,             /* Number of elements to convert */
    int *dst,               /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    int * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CAST_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CAST_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CAST_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	CAST_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CAST_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*------------------------------------------------------------------------
bips_uint - Convert voxel array to unsigned int
-------------------------------------------------------------------------*/

int
bips_uint(
    long nelem,             /* Number of elements to convert */
    unsigned int *dst,      /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    unsigned int * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CAST_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CAST_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CAST_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	CAST_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CAST_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*------------------------------------------------------------------------
bips_long - Convert voxel array to long
-------------------------------------------------------------------------*/

int
bips_long(
    long nelem,             /* Number of elements to convert */
    long *dst,              /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    long * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CAST_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CAST_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CAST_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	CAST_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CAST_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*------------------------------------------------------------------------
bips_ulong - Convert voxel array to unsigned long
-------------------------------------------------------------------------*/

int
bips_ulong(
    long nelem,             /* Number of elements to convert */
    unsigned long *dst,     /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    unsigned long * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CAST_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CAST_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CAST_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	CAST_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CAST_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*------------------------------------------------------------------------
bips_float - Convert voxel array to float
-------------------------------------------------------------------------*/

int
bips_float(
    long nelem,             /* Number of elements to convert */
    float *dst,             /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    float * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CAST_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CAST_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CAST_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	CAST_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CAST_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*------------------------------------------------------------------------
bips_double - Convert voxel array to double
-------------------------------------------------------------------------*/

int
bips_double(
    long nelem,             /* Number of elements to convert */
    double *dst,            /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    double * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CAST_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CAST_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CAST_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	CAST_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CAST_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*------------------------------------------------------------------------
bips_fcomplex - Convert voxel array to float complex
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H

int
bips_fcomplex(
    long nelem,             /* Number of elements to convert */
    float complex *dst,     /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    float complex * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CAST_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CAST_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CAST_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return ERR_OKAY;

    case INTP_FCOMPLEX:
	CAST_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CAST_LOOP(double complex);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}

#endif



/*------------------------------------------------------------------------
bips_fcomplex2 - Convert real and imaginary voxel arrays to float complex
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H

int
bips_fcomplex2(
    long nelem,             /* Number of elements to convert */
    float complex *dst,     /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of both source arrays */
    void *src1,             /* Source 1 is the real part */
    long ss1,               /* Source 1 stride, in elements */
    void *src2,             /* Source 2 is the imaginary part */
    long ss2)               /* Source 2 stride, in elements */
{
    float complex * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CMPLX_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CMPLX_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CMPLX_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CMPLX_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CMPLX_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CMPLX_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CMPLX_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CMPLX_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CMPLX_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CMPLX_LOOP(double);
	return ERR_OKAY;

    case INTP_FCOMPLEX:
	CMPLX_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CMPLX_LOOP(double complex);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}

#endif



/*------------------------------------------------------------------------
bips_dcomplex - Convert voxel array to double complex
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H

int
bips_dcomplex(
    long nelem,             /* Number of elements to convert */
    double complex *dst,    /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    double complex * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CAST_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CAST_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CAST_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CAST_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CAST_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CAST_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CAST_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CAST_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CAST_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CAST_LOOP(double);
	return ERR_OKAY;

    case INTP_FCOMPLEX:
	CAST_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CAST_LOOP(double complex);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}

#endif



/*------------------------------------------------------------------------
bips_dcomplex2 - Convert real and imaginary voxel arrays to double complex
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H

int
bips_dcomplex2(
    long nelem,             /* Number of elements to convert */
    double complex *dst,    /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of both source arrays */
    void *src1,             /* Source 1 is the real part */
    long ss1,               /* Source 1 stride, in elements */
    void *src2,             /* Source 2 is the imaginary part */
    long ss2)               /* Source 2 stride, in elements */
{
    double complex * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	CMPLX_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	CMPLX_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	CMPLX_LOOP(short);
	return ERR_OKAY;

    case INTP_USHORT:
	CMPLX_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	CMPLX_LOOP(int);
	return ERR_OKAY;

    case INTP_UINT:
	CMPLX_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	CMPLX_LOOP(long);
	return ERR_OKAY;

    case INTP_ULONG:
	CMPLX_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	CMPLX_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CMPLX_LOOP(double);
	return ERR_OKAY;

    case INTP_FCOMPLEX:
	CMPLX_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CMPLX_LOOP(double complex);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}

#endif



/*------------------------------------------------------------------------
bips_cast - Convert voxel array to an arbitrary type

This function reads a sequence of voxel values in some specified
internal format and converts them to another internal type specified by
a numeric type code.
-------------------------------------------------------------------------*/

int
bips_cast(
    long nelem,             /* Number of elements to convert */
    int dtype,              /* Desired output type */
    void *dst,              /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    /* Switch according to the destination type */
    switch (dtype) {

    case INTP_UCHAR:
	return bips_uchar(nelem, (unsigned char *)dst, ds, stype, src, ss);

    case INTP_SCHAR:
	return bips_schar(nelem, (signed char *)dst, ds, stype, src, ss);

    case INTP_SHORT:
	return bips_short(nelem, (short *)dst, ds, stype, src, ss);

    case INTP_USHORT:
	return bips_ushort(nelem, (unsigned short *)dst, ds, stype, src, ss);

    case INTP_INT:
	return bips_int(nelem, (int *)dst, ds, stype, src, ss);

    case INTP_UINT:
	return bips_uint(nelem, (unsigned int *)dst, ds, stype, src, ss);

    case INTP_LONG:
	return bips_long(nelem, (long *)dst, ds, stype, src, ss);

    case INTP_ULONG:
	return bips_ulong(nelem, (unsigned long *)dst, ds, stype, src, ss);

    case INTP_FLOAT:
	return bips_float(nelem, (float *)dst, ds, stype, src, ss);

    case INTP_DOUBLE:
	//fprintf(stderr, "value (bips_cast) = %f\n", *(double *)src); // DEBUG
	return bips_double(nelem, (double *)dst, ds, stype, src, ss);

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	return bips_fcomplex(nelem, (float complex *)dst, ds, stype, src, ss);

    case INTP_DCOMPLEX:
	return bips_dcomplex(nelem, (double complex *)dst, ds, stype, src, ss);
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*------------------------------------------------------------------------
bips_round_cast - Round and cast voxel array to an arbitrary type

This function reads a sequence of voxel values in some specified
internal format, rounds and limits appropriately for a given
destination type, and converts to the destination type.  That is, if
the destination is of integral type, the values are rounded to the
nearest integer and limited to the representable range of the
destination type; if the destination is floating point, then no
rounding is done (or needed).  This treatment of floating point
destinations is useful for higher-level functions that compute their
results in double or double complex and must then convert them to
another destination format.  The only source types permitted are
float, double, float complex, and double complex , on the grounds that
a plain bips_cast will handle all the other cases.  Casts from complex
types to real types are not supported.

FIXME: Rename as plain bips_round?

FIXME: There is a trick described in bips_scale_to_uchar() for avoiding
the use of floor for unsigned destination types.  Consider using that
here.

FIXME: Is it actually a useful sanity check to prohibit complex -> real
conversions here?
-------------------------------------------------------------------------*/

/* Round and cast an floating point type to an integral type */
#define RCAST_LOOP(dsttype, srctype, lo, hi) \
{                                            \
    dsttype * restrict d = (dsttype *)dst;   \
    srctype * restrict s = (srctype *)src;   \
    for (i = 0; i < nelem; i++)  {           \
        *d = LIMIT(*s, lo, hi);              \
        d += ds; s+= ss;  }                  \
}

int
bips_round_cast(
    long nelem,             /* Number of elements to convert */
    int dtype,              /* Desired output type */
    void *dst,              /* Buffer for the output data */
    long ds,                /* Destination stride, in elements */
    int stype,              /* Type of the source data */
    void *src,              /* Buffer containing the source data */
    long ss)                /* Source stride, in elements */
{
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_FLOAT:

	/* Switch according to the destination type */
	switch (dtype) {

	case INTP_UCHAR:
	    RCAST_LOOP(unsigned char, float, 0, UCHAR_MAX);
	    return ERR_OKAY;

	case INTP_SCHAR:
	    RCAST_LOOP(signed char, float, SCHAR_MIN, SCHAR_MAX);
	    return ERR_OKAY;

	case INTP_SHORT:
	    RCAST_LOOP(short, float, SHRT_MIN, SHRT_MAX);
	    return ERR_OKAY;

	case INTP_USHORT:
	    RCAST_LOOP(unsigned short, float, 0, USHRT_MAX);
	    return ERR_OKAY;

	case INTP_INT:
	    RCAST_LOOP(int, float, INT_MIN, INT_MAX);
	    return ERR_OKAY;

	case INTP_UINT:
	    RCAST_LOOP(unsigned int, float, 0, UINT_MAX);
	    return ERR_OKAY;

	case INTP_LONG:
	    RCAST_LOOP(long, float, LONG_MIN, LONG_MAX);
	    return ERR_OKAY;

	case INTP_ULONG:
	    RCAST_LOOP(unsigned long, float, 0, ULONG_MAX);
	    return ERR_OKAY;

	case INTP_FLOAT:
	    return bips_float(nelem, (float *)dst, ds, stype, src, ss);

	case INTP_DOUBLE:
	    return bips_double(nelem, (double *)dst, ds, stype, src, ss);

	default:
	    NZReturnError(ERR_ILLOP, illtype);
	}

    case INTP_DOUBLE:

	/* Switch according to the destination type */
	switch (dtype) {

	case INTP_UCHAR:
	    RCAST_LOOP(unsigned char, double, 0, UCHAR_MAX);
	    return ERR_OKAY;

	case INTP_SCHAR:
	    RCAST_LOOP(signed char, double, SCHAR_MIN, SCHAR_MAX);
	    return ERR_OKAY;

	case INTP_SHORT:
	    RCAST_LOOP(short, double, SHRT_MIN, SHRT_MAX);
	    return ERR_OKAY;

	case INTP_USHORT:
	    RCAST_LOOP(unsigned short, double, 0, USHRT_MAX);
	    return ERR_OKAY;

	case INTP_INT:
	    RCAST_LOOP(int, double, INT_MIN, INT_MAX);
	    return ERR_OKAY;

	case INTP_UINT:
	    RCAST_LOOP(unsigned int, double, 0, UINT_MAX);
	    return ERR_OKAY;

	case INTP_LONG:
	    RCAST_LOOP(long, double, LONG_MIN, LONG_MAX);
	    return ERR_OKAY;

	case INTP_ULONG:
	    RCAST_LOOP(unsigned long, double, 0, ULONG_MAX);
	    return ERR_OKAY;

	case INTP_FLOAT:
	    return bips_float(nelem, (float *)dst, ds, stype, src, ss);

	case INTP_DOUBLE:
	    return bips_double(nelem, (double *)dst, ds, stype, src, ss);

	default:
	    NZReturnError(ERR_ILLOP, illtype);
	}

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX :
    case INTP_DCOMPLEX :

      /* Switch according to the destination type */
	switch (dtype) {

	case INTP_FCOMPLEX:
	    return bips_fcomplex(nelem, (float complex *)dst, ds, stype, src, ss);

	case INTP_DCOMPLEX:
	    return bips_dcomplex(nelem, (double complex *)dst, ds, stype, src, ss);

	default:
	    NZReturnError(ERR_ILLOP, illtype);
	}
#endif

    default:
	NZReturnError(ERR_ILLOP, illtype);
    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*-------------------------------------------------------------------------
bips_scale_to_uchar - Scale voxel values to unsigned char

This function reads a sequence of voxel values in some specified
internal format, scales them using a specified offset and gain, and
converts the scaled values to the range 0..255 with values outside the
range being set to the limit.  The scaling equation is

                     y = gain*(x - offset)

for an offset b and a gain g.  This particular operation is provided
as a special case because of the prevalence of the 8-bit intensity
range on display systems.  Any integer or floating point type is
permitted as input, but not a complex type.  The rescaling is given
as gain and offset rather than min and max because this takes slightly
less computation; the higher level routines mostly use min and max
because they are more intutive in most applications.

Complex source types are not supported, because there doesn't seem to
be a use case for that operation.


FIXME: This function might be considered too high-level for bips,
since it could be replaced by a sequences of lower-level bips
operations done blockwise for performance.  But the added efficiency
of this specialized loop might be worth it anyway.  (Although many
modern workstations can overlap float and int computations at no
added cost.)

FIXME: This could be made somewhat more efficient for some special
cases, e.g. offset = 0 and gain = 1; specific single byte of an
unsigned type.  Whether or not it's worth it is another question.

Prior to 2000-03-01, the order of arguments was ...src, stype, ss...;
it was changed for consistency with other functions in this package.
Calls made using the old order will violate the prototype and should
be detected automatically by the compiler.

ALGORITHM: Rounding in this algorithm does not require the use of
floor() since the value being rounded is always non-negative; thus
it doesn't matter if casting a float value to an integral value
truncates toward zero or minus infinity.

FIXME: Does vxl_scale() replace this function?  But that function is
more general and might be a little less efficient.  Consider the trick
noted above for economizing the use of floor().
-------------------------------------------------------------------------*/

#define SCALE_TO_UCHAR(type) {                   \
    type * restrict s = (type *)src;             \
    for (i = 0; i < nelem; i++ ) {               \
	v = gain * ((double)*s - offset) + 0.5;  \
	if      ( v <   0.0 )  v =   0.0;        \
	else if ( v > 255.0 )  v = 255.0;        \
	*d = v;  d += ds;  s += ss; }  }         \


int
bips_scale_to_uchar(
    long   nelem,                /* Number of elements to convert */
    unsigned char  *dst,         /* Buffer for the output data */
    long   ds,                   /* Destination stride, in bytes */
    int    stype,                /* Type of the source data */
    void   *src,                 /* Buffer containing the input data */
    long   ss,                   /* Source stride, in elements */
    double offset,               /* Offset prior to scaling */
    double gain)                 /* Multiplier for scaling */
{
    unsigned char * restrict d = dst;
    double v;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	SCALE_TO_UCHAR(unsigned char);
	return ERR_OKAY;

    case INTP_SCHAR:
	SCALE_TO_UCHAR(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	SCALE_TO_UCHAR(short);
	return ERR_OKAY;

    case INTP_USHORT:
	SCALE_TO_UCHAR(unsigned short);
	return ERR_OKAY;

    case INTP_INT:
	SCALE_TO_UCHAR(int);
	return ERR_OKAY;

    case INTP_UINT:
	SCALE_TO_UCHAR(unsigned int);
	return ERR_OKAY;

    case INTP_LONG:
	SCALE_TO_UCHAR(long);
	return ERR_OKAY;

    case INTP_ULONG:
	SCALE_TO_UCHAR(unsigned long);
	return ERR_OKAY;

    case INTP_FLOAT:
	SCALE_TO_UCHAR(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	SCALE_TO_UCHAR(double);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/**************************************************************************
Unary point operations
**************************************************************************/



/*-------------------------------------------------------------------------
UNOP_LOOP - Generic loop for a unary point operation

This macro defines a generic loop for implementing a unary point
operation with arbitrary stride over a single input array to generate
a single output array.  The input stride may be zero, and the code
will select an optimized loop in this case.  

The UNOP_XLOOP variant must be used when the destination stride may be
zero; the plain UNOP_LOOP may handle the zero-stride case incorrectly
on an optimizing compiler.  The results are undefined if both
destination and source strides are zero; you shouldn't really be using
a loop in this case anyway.

The UNOP_CLOOP variant must be used when the source and destination
have different types; for instance, in taking the real part of a
complex number.

The operation to be performed is given by a macro "asgn" which is
expected to take two arguments and expand into an assignment which
assigns to the first argument the result of the operation on the
second argument; the terminal semicolon must be omitted.  For example,

#define  NOT(d,s)   d = ~(s)

would implement the bitwise not operation.

This macro assumes that the following variable definitions are in
effect:

    void *dst, *src1;       Pointers to destination and source arrays
    long ds, ss1;           Strides in destination and source
    long nelem;             Number of elements to process

FIXME: Consider modifying the using functions to check for errors in
strides.  For the plain UNOP_LOOP, ds == 0 is an error; for the
UNOP_XLOOP, ds == ss == 0 is non-sensical.  An alternative would be to
detect the bad combinations in this macro and return an error code.
Note that the chained ifs could be converted to a switch.
Unfortunately, there doesn't seem to be any fast way to check for
source/destination overlaps.

FIXME: Consider merging the CAST_LOOP and UNOP_CLOOP macros.
--------------------------------------------------------------------------*/

#define UNOP_LOOP(type, asgn)  {       \
    type * restrict d  = (type *)dst;  \
    type * restrict s1 = (type *)src1; \
    long i;                            \
    if (ss1 == 0)                      \
        for (i = 0; i < nelem; i++) {  \
            asgn(*d, *s1);             \
            d += ds; }                 \
    else                               \
        for (i = 0; i < nelem; i++) {  \
            asgn(*d, *s1);             \
            d += ds;  s1 += ss1; }     \
    }


#define UNOP_CLOOP(dtype, stype, asgn)  {   \
    dtype * restrict d  = (dtype *)dst;     \
    stype * restrict s1 = (stype *)src1;    \
    long i;                                 \
    if (ss1 == 0)                           \
        for (i = 0; i < nelem; i++) {       \
            asgn(*d, *s1);                  \
            d += ds; }                      \
    else                                    \
        for (i = 0; i < nelem; i++) {       \
            asgn(*d, *s1);                  \
            d += ds;  s1 += ss1; }          \
    }


#define UNOP_XLOOP(type, asgn)  {      \
    type * restrict d  = (type *)dst;  \
    type * restrict s1 = (type *)src1; \
    long i;                            \
    if (ss1 == 0)                      \
        for (i = 0; i < nelem; i++) {  \
            asgn(*d, *s1);             \
            d += ds; }                 \
    else if (ds == 0)                  \
        for (i = 0; i < nelem; i++) {  \
            asgn(*d, *s1);             \
            s1 += ss1; }               \
    else                               \
        for (i = 0; i < nelem; i++) {  \
            asgn(*d, *s1);             \
            d += ds;  s1 += ss1; }     \
    }



/*------------------------------------------------------------------------
bips_not - Bitwise not (inversion) of an integer array

All integer types are supported, but the effect on signed integers is
defined by the underlying C implementation and is not portable.
-------------------------------------------------------------------------*/

#define NOT(w,x)  w = ~(x)

int
bips_not(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_LOOP(unsigned char, NOT);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_LOOP(unsigned short, NOT);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_LOOP(unsigned int, NOT);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_LOOP(unsigned long, NOT);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_LOOP(signed char, NOT);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_LOOP(short, NOT);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_LOOP(int, NOT);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_LOOP(long, NOT);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_copy - Copy a voxel array
-------------------------------------------------------------------------*/

#define COPY(w,x)  w = (x)

int
bips_copy(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_LOOP(unsigned char, COPY);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_LOOP(unsigned short, COPY);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_LOOP(unsigned int, COPY);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_LOOP(unsigned long, COPY);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_LOOP(signed char, COPY);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_LOOP(short, COPY);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_LOOP(int, COPY);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_LOOP(long, COPY);
	return ERR_OKAY;

    case INTP_FLOAT:
	UNOP_LOOP(float, COPY);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, COPY);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, COPY);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, COPY);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_minus - Elementwise additive inverse of a voxel array
-------------------------------------------------------------------------*/

#define MINUS(w,x)  w = -(x)

int
bips_minus(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_LOOP(unsigned char, MINUS);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_LOOP(unsigned short, MINUS);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_LOOP(unsigned int, MINUS);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_LOOP(unsigned long, MINUS);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_LOOP(signed char, MINUS);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_LOOP(short, MINUS);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_LOOP(int, MINUS);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_LOOP(long, MINUS);
	return ERR_OKAY;

    case INTP_FLOAT:
	UNOP_LOOP(float, MINUS);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, MINUS);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, MINUS);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, MINUS);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_conj - Elementwise complex conjugate of a voxel array

Non-complex types are not supported; this is intended as an error
check.  
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H

#define CONJ(w,x)  w = conj(x)

int
bips_conj(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, CONJ);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, CONJ);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}

#endif



/*------------------------------------------------------------------------
bips_carg - Elementwise complex argument of a voxel array

Note that output from this function is the real type corresponding to
the input complex type.  Non-complex types are not supported; this is
intended as an error check.  The calling sequence is a bit anomalous
in that the type given as an argument must be the *source* type, even
though it is given next to the destination argument.
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H

#define CARG(w,x)  w = carg(x)

int
bips_carg(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FCOMPLEX:
	UNOP_CLOOP(float, float complex, CARG);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_CLOOP(double, double complex, CARG);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}

#endif



/*------------------------------------------------------------------------
bips_imag - Elementwise imaginary part of a voxel array

Note that output from this function is the real type corresponding to
the input complex type.  Non-complex types are not supported; this is
intended as an error check.  The calling sequence is a bit anomalous
in that the type given as an argument must be the *source* type, even
though it is given next to the destination argument.
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H

#define IMAG(w,x)  w = cimag(x)

int
bips_imag(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FCOMPLEX:
	UNOP_CLOOP(float, float complex, IMAG);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_CLOOP(double, double complex, IMAG);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}

#endif



/*------------------------------------------------------------------------
bips_real - Elementwise real part of a voxel array


Note that output from this function is the real type corresponding to
the input complex type.  Non-complex types are supported for
uniformity but are handled simply as copy.  The calling sequence is a
bit anomalous in that the type given as an argument must be the
*source* type, even though it is given next to the destination
argument.
-------------------------------------------------------------------------*/

#define REAL(w,x)   w = creal(x)

int
bips_real(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_LOOP(unsigned char, COPY);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_LOOP(unsigned short, COPY);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_LOOP(unsigned int, COPY);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_LOOP(unsigned long, COPY);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_LOOP(signed char, COPY);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_LOOP(short, COPY);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_LOOP(int, COPY);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_LOOP(long, COPY);
	return ERR_OKAY;

    case INTP_FLOAT:
	UNOP_LOOP(float, COPY);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, COPY);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_CLOOP(float, float complex, REAL);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_CLOOP(double, double complex, REAL);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_abs - Elementwise absolute value of a voxel array

Absolute value of unsigned types is supported for uniformity, but is a
simple copy.  The result type for a complex input is the corresponding
real type.  The calling sequence is a bit anomalous in that the type
given as an argument must be the *source* type, even though it is
given next to the destination argument.
-------------------------------------------------------------------------*/

#define ABS(w,x)   w = abs(x)
#define LABS(w,x)  w = labs(x)
#define FABS(w,x)  w = fabs(x)
#define CABS(w,x)  w = cabs(x)

int
bips_abs(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_LOOP(unsigned char, COPY);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_LOOP(unsigned short, COPY);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_LOOP(unsigned int, COPY);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_LOOP(unsigned long, COPY);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_LOOP(signed char, ABS);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_LOOP(short, ABS);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_LOOP(int, ABS);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_LOOP(long, LABS);
	return ERR_OKAY;

    case INTP_FLOAT:
	UNOP_LOOP(float, FABS);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, FABS);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_CLOOP(float, float complex, CABS);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_CLOOP(double, double complex, CABS);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_floor - Elementwise floor of a voxel array

Not implemented for integral types (because it's trivial) or complex
types (because it's undefined).

FIXME: The gcc/x86 implementation of this is really time-consuming; see
if there is a better way using assembler.  One possibility is to compute
round(x-0.5) which does not require changing the FPU mode.  Better yet,
suggest to the gcc folks that they generate the better assemply code.
-------------------------------------------------------------------------*/

#define FLOOR(w,x)  w = floor(x)

int
bips_floor(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, FLOOR);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, FLOOR);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_ceil - Elementwise ceil of a voxel array

Not implemented for integral types (because it's trivial) or complex
types (because it's undefined).

FIXME: See the comments under bips_floor.
-------------------------------------------------------------------------*/

#define CEIL(w,x)  w = ceil(x)

int
bips_ceil(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, CEIL);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, CEIL);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_exp - Elementwise exp of a voxel array
-------------------------------------------------------------------------*/

#define EXP(w,x)  w = exp(x)

int
bips_exp(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, EXP);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, EXP);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, EXP);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, EXP);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_exp2 - Elementwise exp2 of a voxel array

Not defined for complex types here or in C99, apparently because it
improves precision only in the real floating point format.

FIXME: gcc 3.2 seems to lack this function, even though configure finds it.
-------------------------------------------------------------------------*/

#ifdef HAVE_EXP2
#define EXP2(w,x)  w = exp2(x)

int
bips_exp2(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, EXP2);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, EXP2);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}
#endif /* HAVE_EXP2 */



/*------------------------------------------------------------------------
bips_expm1 - Elementwise expm1 of a voxel array

Not defined for complex types here or in C99, apparently because it
improves precision only in the real floating point format.
-------------------------------------------------------------------------*/

#ifdef HAVE_EXPM1
#define EXPM1(w,x)  w = expm1(x)

int
bips_expm1(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, EXPM1);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, EXPM1);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}
#endif /* HAVE_EXPM1 */



/*------------------------------------------------------------------------
bips_log - Elementwise log of a voxel array
-------------------------------------------------------------------------*/

#define LOG(w,x)  w = log(x)

int
bips_log(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, LOG);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, LOG);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, LOG);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, LOG);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_log10 - Elementwise log10 of a voxel array

Not defined for complex types here or in C99, apparently because it
improves precision only in the real floating point format.
-------------------------------------------------------------------------*/

#define LOG10(w,x)  w = log10(x)

int
bips_log10(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, LOG10);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, LOG10);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_log1p - Elementwise log (1+x) of a voxel array

Not defined for complex types here or in C99, apparently because it
improves precision only in the real floating point format.
-------------------------------------------------------------------------*/

#ifdef HAVE_LOG1P
#define LOG1P(w,x)  w = log1p(x)

int
bips_log1p(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, LOG1P);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, LOG1P);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}
#endif /* HAVE_LOG1P */



/*------------------------------------------------------------------------
bips_log2 - Elementwise log2 of a voxel array

Not defined for complex types here or in C99, apparently because it
improves precision only in the real floating point format.

FIXME: gcc 3.2 seems to lack this function, even though configure finds it.
-------------------------------------------------------------------------*/

#ifdef HAVE_LOG2
#define LOG2(w,x)  w = log2(x)

int
bips_log2(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, LOG2);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, LOG2);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}
#endif /* HAVE_LOG2 */



/*------------------------------------------------------------------------
bips_cbrt - Elementwise cbrt of a voxel array

FIXME: Not defined in C99 for complex types?
-------------------------------------------------------------------------*/

#ifdef HAVE_CBRT
#define CBRT(w,x)  w = cbrt(x)

int
bips_cbrt(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, CBRT);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, CBRT);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}
#endif /* HAVE_CBRT */



/*------------------------------------------------------------------------
bips_sqrt - Elementwise sqrt of a voxel array
-------------------------------------------------------------------------*/

#define SQRT(w,x)  w = sqrt(x)

int
bips_sqrt(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, SQRT);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, SQRT);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, SQRT);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, SQRT);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_square - Elementwise square of a voxel array

See also bips_sqrdiff() which computes the square of the difference
between two arrays.

FIXME: This does not seem to be actually used anywhere in Pyvox; consider
removing it.  In any case, the more useful operation for complex types
would seem to be |z|^2.
-------------------------------------------------------------------------*/

#define SQUARE(w,x)  w = (x)*(x)

int
bips_square(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, SQUARE);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, SQUARE);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, SQUARE);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, SQUARE);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_sin - Elementwise sin of a voxel array
-------------------------------------------------------------------------*/

#define SIN(w,x)  w = sin(x)

int
bips_sin(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, SIN);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, SIN);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, SIN);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, SIN);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_cos - Elementwise cos of a voxel array
-------------------------------------------------------------------------*/

#define COS(w,x)  w = cos(x)

int
bips_cos(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, COS);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, COS);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, COS);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, COS);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_tan - Elementwise tan of a voxel array
-------------------------------------------------------------------------*/

#define TAN(w,x)  w = tan(x)

int
bips_tan(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, TAN);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, TAN);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, TAN);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, TAN);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_asin - Elementwise asin of a voxel array
-------------------------------------------------------------------------*/

#define ASIN(w,x)  w = asin(x)

int
bips_asin(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, ASIN);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, ASIN);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, ASIN);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, ASIN);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_acos - Elementwise acos of a voxel array
-------------------------------------------------------------------------*/

#define ACOS(w,x)  w = acos(x)

int
bips_acos(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, ACOS);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, ACOS);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, ACOS);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, ACOS);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_atan - Elementwise atan of a voxel array

See also bips_atan2 for the four-quadrant arc tanget.
-------------------------------------------------------------------------*/

#define ATAN(w,x)  w = atan(x)

int
bips_atan(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, ATAN);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, ATAN);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, ATAN);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, ATAN);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_sinh - Elementwise sinh of a voxel array
-------------------------------------------------------------------------*/

#define SINH(w,x)  w = sinh(x)

int
bips_sinh(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, SINH);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, SINH);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, SINH);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, SINH);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_cosh - Elementwise cosh of a voxel array
-------------------------------------------------------------------------*/

#define COSH(w,x)  w = cosh(x)

int
bips_cosh(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, COSH);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, COSH);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, COSH);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, COSH);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_tanh - Elementwise tanh of a voxel array
-------------------------------------------------------------------------*/

#define TANH(w,x)  w = tanh(x)

int
bips_tanh(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, TANH);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, TANH);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_LOOP(float complex, TANH);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_LOOP(double complex, TANH);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_erf - Elementwise erf of a voxel array

FIXME: Not defined in C99 for complex types?  Probably not useful anyway.
-------------------------------------------------------------------------*/

#ifdef HAVE_ERF
#define ERF(w,x)  w = erf(x)

int
bips_erf(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, ERF);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, ERF);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}
#endif /* HAVE_ERF */



/*------------------------------------------------------------------------
bips_erfc - Elementwise erfc of a voxel array

Not defined in C99 for complex types, and probably not useful.
-------------------------------------------------------------------------*/

#ifdef HAVE_ERFC
#define ERFC(w,x)  w = erfc(x)

int
bips_erfc(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, ERFC);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, ERFC);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}
#endif /* HAVE_ERFC */



/*------------------------------------------------------------------------
bips_lgamma - Elementwise lgamma of a voxel array

Not defined in C99 for complex types, and probably not useful.
-------------------------------------------------------------------------*/

#ifdef HAVE_LGAMMA
#define LGAMMA(w,x)  w = lgamma(x)

int
bips_lgamma(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, LGAMMA);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, LGAMMA);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}
#endif /* HAVE_LGAMMA */



/*------------------------------------------------------------------------
bips_tgamma - Elementwise tgamma of a voxel array

Not defined in C99 for complex types, and probably not useful.

FIXME: gcc 3.2 seems to lack this function, even though configure finds it.
-------------------------------------------------------------------------*/

#ifdef HAVE_TGAMMA
#define TGAMMA(w,x)  w = tgamma(x)

int
bips_tgamma(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	UNOP_LOOP(float, TGAMMA);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_LOOP(double, TGAMMA);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}
#endif /* HAVE_TGAMMA */



/***************************************************************************
Scatter/gather operations

Read/write elements in random order.
***************************************************************************/


/*-------------------------------------------------------------------------
bips_gather - Read elements in random order

This function reads elements in an order specified by a table of
indices, and copies them to a sequential vector.  The results are
unpredictable if some index value appears more than once, or if the
destination stride is zero.  The indices are NOT checked for range or
validity, since it is assumed that this check was done prior to
calling bips_gather; the vxl_check_index function may be useful for
this purpose.  (This function is often used for slice selection, in
which case it is more efficient to check the indices once and call
bips_gather multiple times with the same index array.)
--------------------------------------------------------------------------*/

#define GATHER_LOOP(type)    {      \
    long i;                         \
    type * restrict d = dst;        \
    type * restrict s = src;        \
    for (i = 0; i < nelem; i++)  {  \
        *d = s[index[i]];           \
        d += ds;  }   }             \

int
bips_gather(
    long nelem,          /* Number of elements to process */
    int  type,           /* Type of data */
    void *dst,           /* Destination array */
    long ds,             /* Stride in the destination array */
    void *src,           /* Source array */
    long *index)         /* Indices into the source array */
{
    /* Switch according to the data type */
    switch (type) {

    case INTP_UCHAR:
        GATHER_LOOP(unsigned char);
        return ERR_OKAY;

    case INTP_SCHAR:
        GATHER_LOOP(signed char);
        return ERR_OKAY;

    case INTP_USHORT:
        GATHER_LOOP(unsigned short);
        return ERR_OKAY;

    case INTP_SHORT:
        GATHER_LOOP(short);
        return ERR_OKAY;

    case INTP_UINT:
        GATHER_LOOP(unsigned int);
        return ERR_OKAY;

    case INTP_INT:
        GATHER_LOOP(int);
        return ERR_OKAY;

    case INTP_ULONG:
        GATHER_LOOP(unsigned long);
        return ERR_OKAY;

    case INTP_LONG:
        GATHER_LOOP(long);
        return ERR_OKAY;

    case INTP_FLOAT:
        GATHER_LOOP(float);
        return ERR_OKAY;

    case INTP_DOUBLE:
        GATHER_LOOP(double);
        return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
        GATHER_LOOP(float complex);
        return ERR_OKAY;

    case INTP_DCOMPLEX:
        GATHER_LOOP(double complex);
        return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}


/*-------------------------------------------------------------------------
bips_scatter - Write elements in random order

This function reads elements from a sequential vector and writes them
in an order specified by a table of indices.  The results are
unpredictable if some index value occurs more than once.  The indices
are NOT checked for range or validity, since it is assumed that this
check was done prior to calling bips_scatter; the vxl_check_index
function may be useful for this purpose.  (This function is often used
for slice selection, in which case it is more efficient to check the
indices once and call bips_scatter multiple times with the same index
array.)  The source stride may be zero, which is useful for stuffing a
constant value into an array slice.
--------------------------------------------------------------------------*/

#define SCATTER_LOOP(type)  {       \
    long i;                         \
    type * restrict d = dst;        \
    type * restrict s = src;        \
    for (i = 0; i < nelem; i++)  {  \
        d[index[i]] = *s;           \
        s += ss;  }  }              \

int
bips_scatter(
    long nelem,          /* Number of elements to process */
    int  type,           /* Type of data */
    void *dst,           /* Destination array */
    long *index,         /* Indices into the destination array */
    void *src,           /* Source array */
    long ss)             /* Stride in the source array */
{
    /* Switch according to the data type */
    switch (type) {

    case INTP_UCHAR:
        SCATTER_LOOP(unsigned char);
        return ERR_OKAY;

    case INTP_SCHAR:
        SCATTER_LOOP(signed char);
        return ERR_OKAY;

    case INTP_USHORT:
        SCATTER_LOOP(unsigned short);
        return ERR_OKAY;

    case INTP_SHORT:
        SCATTER_LOOP(short);
        return ERR_OKAY;

    case INTP_UINT:
        SCATTER_LOOP(unsigned int);
        return ERR_OKAY;

    case INTP_INT:
        SCATTER_LOOP(int);
        return ERR_OKAY;

    case INTP_ULONG:
        SCATTER_LOOP(unsigned long);
        return ERR_OKAY;

    case INTP_LONG:
        SCATTER_LOOP(long);
        return ERR_OKAY;

    case INTP_FLOAT:
        SCATTER_LOOP(float);
        return ERR_OKAY;

    case INTP_DOUBLE:
        SCATTER_LOOP(double);
        return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
        SCATTER_LOOP(float complex);
        return ERR_OKAY;

    case INTP_DCOMPLEX:
        SCATTER_LOOP(double complex);
        return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/***************************************************************************
Compound assignment operations

Such as +=, *=, etc.  These follow the same pattern as unary
operations and are implemented in the same fashion.  The source and
destination must be the same type and are assumed not to overlap.  The
desination stride may be zero, but the source and destination stride
may not both be zero.

These operations may be used for updating an array in place or for
summing (or whatever) over a dimension (an operation also known as
reduction); in either case, it is the caller's responsibility to
initialize the destination to some appropriate value.

The bitwise boolean operators are supported only for integral types; the
user is reminded that bitwise operations on signed integral types are not
portable.
***************************************************************************/


/*-------------------------------------------------------------------------
bips_and_set - Bitwise AND of source and destination

Defined only for integral types.
-------------------------------------------------------------------------*/

#define ANDSET(w,x)   w &= (x)

int
bips_and_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, ANDSET);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, ANDSET);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, ANDSET);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, ANDSET);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, ANDSET);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_XLOOP(short, ANDSET);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_XLOOP(int, ANDSET);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_XLOOP(long, ANDSET);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}


/*-------------------------------------------------------------------------
bips_or_set - Bitwise OR of source and destination

Defined only for integral types.
-------------------------------------------------------------------------*/

#define ORSET(w,x)   w |= (x)

int
bips_or_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, ORSET);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, ORSET);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, ORSET);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, ORSET);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, ORSET);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_XLOOP(short, ORSET);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_XLOOP(int, ORSET);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_XLOOP(long, ORSET);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}


/*-------------------------------------------------------------------------
bips_xor_set - Bitwise XOR of source and destination

Defined only for integral types.
-------------------------------------------------------------------------*/

#define XORSET(w,x)   w ^= (x)

int
bips_xor_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, XORSET);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, XORSET);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, XORSET);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, XORSET);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, XORSET);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_XLOOP(short, XORSET);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_XLOOP(int, XORSET);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_XLOOP(long, XORSET);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}


/*-------------------------------------------------------------------------
bips_min_set - Minimum of source and destination

This function is not defined for the complex types, because there
is no order of the complex numbers which is consistent with their
arithmetic properties.
-------------------------------------------------------------------------*/

#define MINSET(w,x)  if ((x) < (w)) w = (x)

int
bips_min_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, MINSET);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, MINSET);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, MINSET);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, MINSET);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, MINSET);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_XLOOP(short, MINSET);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_XLOOP(int, MINSET);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_XLOOP(long, MINSET);
	return ERR_OKAY;

    case INTP_FLOAT:
	UNOP_XLOOP(float, MINSET);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_XLOOP(double, MINSET);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}


/*-------------------------------------------------------------------------
bips_nzmin_set - Nonzero minimum of source and destination

This function sets each pixel of the destination to the value of the
corresponding source pixel whenever the source is nonzero and the
destination is either zero or larger than the destination.  The net
effect of several passes is to set the destination to the smallest
nonzero source value, or zero if all the source values are zero.  It
is used to implement the vxl_dilabel() function.

This function is not defined for the complex types, because there
is no order of the complex numbers which is consistent with their
arithmetic properties.
-------------------------------------------------------------------------*/

#define NZMINSET(w,x)  if (((x) != 0) && ((w) == 0 || (x) < (w))) w = (x)

int
bips_nzmin_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, NZMINSET);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, NZMINSET);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, NZMINSET);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, NZMINSET);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, NZMINSET);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_XLOOP(short, NZMINSET);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_XLOOP(int, NZMINSET);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_XLOOP(long, NZMINSET);
	return ERR_OKAY;

    case INTP_FLOAT:
	UNOP_XLOOP(float, NZMINSET);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_XLOOP(double, NZMINSET);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}


/*-------------------------------------------------------------------------
bips_max_set - Maximum of source and destination

This function is not defined for the complex types, because there
is no order of the complex numbers which is consistent with their
arithmetic properties.
-------------------------------------------------------------------------*/

#define MAXSET(w,x)  if ((x) > (w)) w = (x)

int
bips_max_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, MAXSET);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, MAXSET);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, MAXSET);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, MAXSET);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, MAXSET);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_XLOOP(short, MAXSET);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_XLOOP(int, MAXSET);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_XLOOP(long, MAXSET);
	return ERR_OKAY;

    case INTP_FLOAT:
	UNOP_XLOOP(float, MAXSET);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_XLOOP(double, MAXSET);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}


/*-------------------------------------------------------------------------
bips_add_set - Add source to destination
-------------------------------------------------------------------------*/

#define ADDSET(w,x)  w += (x)

int
bips_add_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, ADDSET);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, ADDSET);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, ADDSET);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, ADDSET);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, ADDSET);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_XLOOP(short, ADDSET);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_XLOOP(int, ADDSET);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_XLOOP(long, ADDSET);
	return ERR_OKAY;

    case INTP_FLOAT:
	UNOP_XLOOP(float, ADDSET);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_XLOOP(double, ADDSET);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_XLOOP(float complex, ADDSET);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_XLOOP(double complex, ADDSET);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}


/*-------------------------------------------------------------------------
bips_sub_set - Subtract source from destination
-------------------------------------------------------------------------*/

#define SUBSET(w,x)  w -= (x)

int
bips_sub_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, SUBSET);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, SUBSET);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, SUBSET);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, SUBSET);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, SUBSET);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_XLOOP(short, SUBSET);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_XLOOP(int, SUBSET);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_XLOOP(long, SUBSET);
	return ERR_OKAY;

    case INTP_FLOAT:
	UNOP_XLOOP(float, SUBSET);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_XLOOP(double, SUBSET);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_XLOOP(float complex, SUBSET);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_XLOOP(double complex, SUBSET);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}


/*-------------------------------------------------------------------------
bips_mul_set - Multipy destination by source
-------------------------------------------------------------------------*/

#define MULSET(w,x)  w *= (x)

int
bips_mul_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, MULSET);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, MULSET);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, MULSET);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, MULSET);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, MULSET);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_XLOOP(short, MULSET);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_XLOOP(int, MULSET);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_XLOOP(long, MULSET);
	return ERR_OKAY;

    case INTP_FLOAT:
	UNOP_XLOOP(float, MULSET);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_XLOOP(double, MULSET);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_XLOOP(float complex, MULSET);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_XLOOP(double complex, MULSET);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}


/*-------------------------------------------------------------------------
bips_div_set - Divide destination by source
-------------------------------------------------------------------------*/

#define DIVSET(w,x)  w /= (x)

int
bips_div_set(
    long nelem,              /* Number of elements to process */
    int type,                /* Type of data to process */
    void *dst,               /* Destination array */
    long ds,                 /* Stride in destination array */
    void *src1,              /* Source array */
    long ss1)                /* Stride in source array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	UNOP_XLOOP(unsigned char, DIVSET);
	return ERR_OKAY;

    case INTP_USHORT:
	UNOP_XLOOP(unsigned short, DIVSET);
	return ERR_OKAY;

    case INTP_UINT:
	UNOP_XLOOP(unsigned int, DIVSET);
	return ERR_OKAY;

    case INTP_ULONG:
	UNOP_XLOOP(unsigned long, DIVSET);
	return ERR_OKAY;

    case INTP_SCHAR:
	UNOP_XLOOP(signed char, DIVSET);
	return ERR_OKAY;

    case INTP_SHORT:
	UNOP_XLOOP(short, DIVSET);
	return ERR_OKAY;

    case INTP_INT:
	UNOP_XLOOP(int, DIVSET);
	return ERR_OKAY;

    case INTP_LONG:
	UNOP_XLOOP(long, DIVSET);
	return ERR_OKAY;

    case INTP_FLOAT:
	UNOP_XLOOP(float, DIVSET);
	return ERR_OKAY;

    case INTP_DOUBLE:
	UNOP_XLOOP(double, DIVSET);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	UNOP_XLOOP(float complex, DIVSET);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	UNOP_XLOOP(double complex, DIVSET);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}




/***************************************************************************
Binary point operations
***************************************************************************/


/*------------------------------------------------------------------------
BINOP_LOOP - Generic loop for a binary point operation

This macro defines a generic loop for implementing a binary point
operation with arbitrary stride over two input arrays to generate a
third output array.  Either input stride may be zero, and a special
optimized loop will be used in these cases; the case that both source
strides are zero is NOT handled specially, since I can't think of any
use for it.  The BINOP_XLOOP variant must be used when the destination
stride may be zero; the plain BINOP_LOOP may handle this case
incorrectly on an optimizing compiler.  The results are undefined if
the destination and both source strides are zero.

There is also a variant form BINOP_RLOOP which handles the recursive
case correctly, even on an optimizing compiler, but not the case where
the destination stride is zero.  For this case, the user must specify
ds == ss1 and d == s1 + ss1; then the running sum/product/whatever
will be generated.  The second source stride ss2 may be zero.  It is
the caller's responsibility to set d[-ss1] correctly.

The operation to be performed is given by a macro "asgn" which is
expected to take three arguments and expand into an assignment which
assigns to the first argument the result of the operation on the
second and third arguments; the terminal semicolon must be omitted.
For example,

#define  ADD(d,s1,s2)   d = (s1) + (s2)

would implement addition.

This macro assumes the following variable definitions to be in effect:

    void *dst;            Pointer to destination array
    void *src1, *src2;    Pointers to source arrays
    long ds;              Stride in destination array
    long ss1, ss2;        Strides in source arrays
    long nelem;           Number of elements to process

FIXME: Consider modifying the using functions to detect striding
errors.  For LOOP and RLOOP, it is an error if ds == 0; for XLOOP, it
is non-sensical if ds == ss1 == ss2 == 0.  An alternative would be to
detect the bad combinations in this macro and return an error code.
Note that the chained ifs could be converted to a switch.
Unfortunately, there doesn't seem to be any fast way to check for
source/destination overlaps.

FIXME: Look at the assembly code and verify that the restrict
qualifiers are doing the right thing.  Add tests to catch errors in
the restrictions?

FIXME:  Are side parameters supported cleanly?  I think so, but it needs
to be checked; type conversions might be a problem.
-------------------------------------------------------------------------*/

/* No overlaps and dest stride != 0 */
#define BINOP_LOOP(type, asgn)  {            \
    type * restrict d  = (type *)dst;        \
    type * restrict s1 = (type *)src1;       \
    type * restrict s2 = (type *)src2;       \
    long i;                                  \
    if (ss1 == 0)                            \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s2 += ss2; }            \
    else if (ss2 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; }            \
    else                                     \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; s2 += ss2; } \
    }


/* Recursive overlap only and dest stride != 0 */
#define BINOP_RLOOP(type, asgn)  {           \
    type * restrict d  = (type *)dst;        \
    type * restrict s1 = (type *)src1;       \
    type * restrict s2 = (type *)src2;       \
    long i;                                  \
    if (ds == ss1 && d == s1 + ss1 &&        \
        ss2 == 0)                            \
        for (i = 0; i < nelem; i++) {        \
            asgn(*d, *(d-ss1), *s2);         \
            d += ds; }                       \
    else if (ds == ss1 && d == s1 + ss1)     \
        for (i = 0; i < nelem; i++) {        \
            asgn(*d, *(d-ss1), *s2);         \
            d += ds; s2 += ss2; }            \
    else if (ss1 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s2 += ss2; }            \
    else if (ss2 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; }            \
    else                                     \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; s2 += ss2; } \
    }


/* No overlap but dest stride may be zero; allow mixed types */
/* FIXME: Narrowly defined to support bips_fma_set; consider possible
   generalizations for other applications. */
#define BINOP_XLOOP(dtype, stype1, asgn)  {  \
    dtype  * restrict d  = (dtype  *)dst;    \
    stype1 * restrict s1 = (stype1 *)src1;   \
    dtype  * restrict s2 = (dtype  *)src2;   \
    long i;                                  \
    if (ds == 0 && ss1 == 0)                 \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            s2 += ss2; }                     \
    else if (ds == 0 && ss2 == 0)            \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            s1 += ss1; }                     \
    else if (ss1 == 0 && ss2 == 0)           \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; }                       \
    else if (ss1 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s2 += ss2; }            \
    else if (ss2 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; }            \
    else                                     \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; s2 += ss2; } \
    }



/*--------------------------------------------------------------------------
bips_and - Bitwise AND on integer arrays

All integer types are supported, but the effect on signed integers is
defined by the underlying C implementation and is not portable.
---------------------------------------------------------------------------*/

#define AND(w,x,y)  w = (x) & (y)

int
bips_and(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, AND);
	return ERR_OKAY;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, AND);
	return ERR_OKAY;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, AND);
	return ERR_OKAY;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, AND);
	return ERR_OKAY;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, AND);
	return ERR_OKAY;

    case INTP_SHORT:
	BINOP_LOOP(short, AND);
	return ERR_OKAY;

    case INTP_INT:
	BINOP_LOOP(int, AND);
	return ERR_OKAY;

    case INTP_LONG:
	BINOP_LOOP(long, AND);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*--------------------------------------------------------------------------
bips_or - Bitwise OR on integer arrays

All integer types are supported, but the effect on signed integers is
defined by the underlying C implementation and is not portable.
---------------------------------------------------------------------------*/

#define OR(w,x,y)  w = (x) | (y)

int
bips_or(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, OR);
	return ERR_OKAY;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, OR);
	return ERR_OKAY;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, OR);
	return ERR_OKAY;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, OR);
	return ERR_OKAY;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, OR);
	return ERR_OKAY;

    case INTP_SHORT:
	BINOP_LOOP(short, OR);
	return ERR_OKAY;

    case INTP_INT:
	BINOP_LOOP(int, OR);
	return ERR_OKAY;

    case INTP_LONG:
	BINOP_LOOP(long, OR);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*--------------------------------------------------------------------------
bips_xor - Bitwise XOR on integer arrays

All integer types are supported, but the effect on signed integers is
defined by the underlying C implementation and is not portable.
---------------------------------------------------------------------------*/

#define XOR(w,x,y)  w = (x) ^ (y)

int
bips_xor(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, XOR);
	return ERR_OKAY;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, XOR);
	return ERR_OKAY;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, XOR);
	return ERR_OKAY;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, XOR);
	return ERR_OKAY;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, XOR);
	return ERR_OKAY;

    case INTP_SHORT:
	BINOP_LOOP(short, XOR);
	return ERR_OKAY;

    case INTP_INT:
	BINOP_LOOP(int, XOR);
	return ERR_OKAY;

    case INTP_LONG:
	BINOP_LOOP(long, XOR);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_cmove - Conditional move 

This function scans its first source argument for non-zero values and
moves only the corresponding values of the second source array to the
corresponding elements of the destination, leaving unmatched elements
of the destination unchanged.  The first source argument must be unsigned
char; the destination and second source must be the same type, but any
type is permitted.

NOTE that this function is atypical in that the selector is always
unsigned char; as a reminder of this, the argument order is different
than normally used.  A further consequence is that the standard
BINOP_LOOP does not work here because the selector is always unsigned
char; so we have to use a modified loop.

Setting ds or ss1 to zero are not useful in this operation, so we don't
bother to make them efficient.
--------------------------------------------------------------------------*/

#define CMOVE(x,y,z)      if (y != 0)  x = z

#define CMOVE_LOOP(type, asgn)  {                        \
    type * restrict d = (type *)dst;                     \
    unsigned char * restrict s1 = (unsigned char *)src1; \
    type * restrict s2 = (type *)src2;                   \
    long i;                                              \
    if (ss2 == 0)                                        \
        for (i = 0; i < nelem; i++)  {                   \
            asgn(*d, *s1, *s2);                          \
            d += ds; s1 += ss1; }                        \
    else                                                 \
        for (i = 0; i < nelem; i++)  {                   \
            asgn(*d, *s1, *s2);                          \
            d += ds; s1 += ss1; s2 += ss2; }             \
    }

int
bips_cmove(
    long   nelem,                 /* Number of elements to compare */
    int    type,                  /* Common type of dest and source 2 */
    void  *dst,                   /* Destination array */
    long   ds,                    /* Destination stride, in elements */
    unsigned char *src1,          /* First Source: selector array */
    long   ss1,                   /* First source stride, in elements */
    void  *src2,                  /* Second source array */
    long   ss2)                   /* Second source stride, in elements */
{

    /* DEBUG: Show selected arguments */
    if (0) {
	double *d = (double *)dst, *s2 = (double *)src2;
	fprintf(stderr, "dst  = %p\n", dst);
	if (type == INTP_DOUBLE) 
	    fprintf(stderr, "dst[] = %f, %f, ...\n", d[0], d[1]);
	fprintf(stderr, "src1 = %p\n", src1);
	fprintf(stderr, "src1[] = %d, %d, ...\n", src1[0], src1[1]);
	fprintf(stderr, "src2 = %p\n", src2);
	if (type == INTP_DOUBLE) 
	    fprintf(stderr, "src2[] = %f, %f, ...\n", s2[0], s2[1]);
    }

    /* Switch according to the type */
    switch (type) {

    case INTP_UCHAR:
	CMOVE_LOOP(unsigned char, CMOVE);
	return ERR_OKAY;

    case INTP_USHORT:
	CMOVE_LOOP(unsigned short, CMOVE);
	return ERR_OKAY;

    case INTP_UINT:
	CMOVE_LOOP(unsigned int, CMOVE);
	return ERR_OKAY;

    case INTP_ULONG:
	CMOVE_LOOP(unsigned long, CMOVE);
	return ERR_OKAY;

    case INTP_SCHAR:
	CMOVE_LOOP(signed char, CMOVE);
	return ERR_OKAY;

    case INTP_SHORT:
	CMOVE_LOOP(short, CMOVE);
	return ERR_OKAY;

    case INTP_INT:
	CMOVE_LOOP(int, CMOVE);
	return ERR_OKAY;

    case INTP_LONG:
	CMOVE_LOOP(long, CMOVE);
	return ERR_OKAY;

    case INTP_FLOAT:
	CMOVE_LOOP(float, CMOVE);
	return ERR_OKAY;

    case INTP_DOUBLE:
	CMOVE_LOOP(double, CMOVE);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	CMOVE_LOOP(float complex, CMOVE);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	CMOVE_LOOP(double complex, CMOVE);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*------------------------------------------------------------------------
bips_compare - Compare two voxel arrays

This function compares two sequences of voxel values in any
(non-complex) internal format and outputs any of three given unsigned
char values according to whether the voxel in the first sequence is
less than, equal to, or greater than the voxel in the second sequence.
The two source sequences must be of the same type.  This function is not 
defined for the complex types, because there is no order of the complex 
numbers which is consistent with their arithmetic properties.

NOTE that this function is atypical in that the result is always unsigned
char rather than being the same type as the source arrays; as a reminder of
this, the type argument precedes the source arguments rather than the
destination argument.  A consequence of this is that the standard BINOP_LOOP 
does not work here because the destination is always unsigned char; so we 
have to use a modified loop.

See also bips_eq(), bips_gt(), and bips_lt() for two-way comparisons.
Note that bips_eq() is also defined for complex types.
--------------------------------------------------------------------------*/

#define COMP(w,x,y)                \
    if      (x < y)  w = lsval;    \
    else if (x > y)  w = gtval;    \
    else             w = eqval


#define COMP_LOOP(type, asgn)  {             \
    unsigned char * restrict d = (unsigned char *)dst; \
    type * restrict s1 = (type *)src1;       \
    type * restrict s2 = (type *)src2;       \
    long i;                                  \
    if (ss1 == 0)                            \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s2 += ss2; }            \
    else if (ss2 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; }            \
    else                                     \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2);              \
            d += ds; s1 += ss1; s2 += ss2; } \
    }

int
bips_compare(
    long   nelem,                 /* Number of elements to compare */
    unsigned char *dst,           /* Result of comparisons */
    long   ds,                    /* Destination stride, in elements */
    int    stype,                 /* Common source type */
    void  *src1,                  /* First source array */
    long   ss1,                   /* First source stride, in elements */
    void  *src2,                  /* Second source array */
    long   ss2,                   /* Second source stride, in elements */
    unsigned char  lsval,         /* Output if less than */
    unsigned char  eqval,         /* Output if equal to */
    unsigned char  gtval)         /* Output if greater than */
{

    /* Specialized routines are faster for two-way comparisons */
    /* FIXME: Is it possible to easily inline these? */
    if (lsval == gtval)
        return bips_eq(nelem, dst, ds, stype, src1, ss1, src2, ss2, 
                       eqval, lsval);
    else if (lsval == eqval)
        return bips_gt(nelem, dst, ds, stype, src1, ss1, src2, ss2, 
                       gtval, lsval);
    else if (eqval == gtval)
        return bips_lt(nelem, dst, ds, stype, src1, ss1, src2, ss2, 
                       lsval, eqval);

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	COMP_LOOP(unsigned char, COMP);
	return ERR_OKAY;

    case INTP_USHORT:
	COMP_LOOP(unsigned short, COMP);
	return ERR_OKAY;

    case INTP_UINT:
	COMP_LOOP(unsigned int, COMP);
	return ERR_OKAY;

    case INTP_ULONG:
	COMP_LOOP(unsigned long, COMP);
	return ERR_OKAY;

    case INTP_SCHAR:
	COMP_LOOP(signed char, COMP);
	return ERR_OKAY;

    case INTP_SHORT:
	COMP_LOOP(short, COMP);
	return ERR_OKAY;

    case INTP_INT:
	COMP_LOOP(int, COMP);
	return ERR_OKAY;

    case INTP_LONG:
	COMP_LOOP(long, COMP);
	return ERR_OKAY;

    case INTP_FLOAT:
	COMP_LOOP(float, COMP);
	return ERR_OKAY;

    case INTP_DOUBLE:
	COMP_LOOP(double, COMP);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*--------------------------------------------------------------------------
bips_eq - Compare two arrays for equality

This function compares corresponding values from two numeric arrays in
any internal format and outputs one of two given unsigned char values
according as the value in the first array is equal or not equal to the
value in the second.  The two arrays must be the same type.

NOTE that this function is atypical in that the result is always unsigned
char rather than being the same type as the source arrays; as a reminder of
this, the type argument precedes the source arguments rather than the
destination argument.  A consequence of this is that the standard BINOP_LOOP 
does not work here because the destination is always unsigned char; so we 
have to use a modified loop.

See also bips_compare() for three-way comparisons on non-complex types.
----------------------------------------------------------------------------*/

#define EQ(w,x,y)            \
    if (x == y)  w = eqval;  \
    else         w = neval;  


int                               /* Returns nonzero on error */
bips_eq(
    long nelem,                   /* Number of elements to compare */
    unsigned char *dst,           /* Result of comparisons */
    long   ds,                    /* Destination stride, in elements */
    int    stype,                 /* Common source type */
    void  *src1,                  /* First source array */
    long   ss1,                   /* First source stride, in elements */
    void  *src2,                  /* Second source array */
    long   ss2,                   /* Second source stride, in elements */
    unsigned char  eqval,         /* Output if equal */
    unsigned char  neval)         /* Output if not equal */

{
    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	COMP_LOOP(unsigned char, EQ);
	return ERR_OKAY;

    case INTP_USHORT:
	COMP_LOOP(unsigned short, EQ);
	return ERR_OKAY;

    case INTP_UINT:
	COMP_LOOP(unsigned int, EQ);
	return ERR_OKAY;

    case INTP_ULONG:
	COMP_LOOP(unsigned long, EQ);
	return ERR_OKAY;

    case INTP_SCHAR:
	COMP_LOOP(signed char, EQ);
	return ERR_OKAY;

    case INTP_SHORT:
	COMP_LOOP(short, EQ);
	return ERR_OKAY;

    case INTP_INT:
	COMP_LOOP(int, EQ);
	return ERR_OKAY;

    case INTP_LONG:
	COMP_LOOP(long, EQ);
	return ERR_OKAY;

    case INTP_FLOAT:
	COMP_LOOP(float, EQ);
	return ERR_OKAY;

    case INTP_DOUBLE:
	COMP_LOOP(double, EQ);
	return ERR_OKAY;

#if HAVE_COMPLEX_H

    case INTP_FCOMPLEX:
	COMP_LOOP(float complex, EQ);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	COMP_LOOP(double complex, EQ);
	return ERR_OKAY;

#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*--------------------------------------------------------------------------
bips_gt - Compare two arrays for greater than

This function compares corresponding values from two numeric arrays in
any internal format and outputs one of two given unsigned char values
according as the value in the first array is greater than or not
greater than the value in the second.  The two arrays must be the
same type.  Not defined for complex types.

NOTE that this function is atypical in that the result is always unsigned
char rather than being the same type as the source arrays; as a reminder of
this, the type argument precedes the source arguments rather than the
destination argument.  A consequence of this is that the standard BINOP_LOOP 
does not work here because the destination is always unsigned char; so we 
have to use a modified loop.

See also bips_compare() for three-way comparisons on non-complex types.
----------------------------------------------------------------------------*/

#define GT(w,x,y)           \
    if (x > y)  w = gtval;  \
    else        w = leval;  


int                               /* Returns nonzero on error */
bips_gt(
    long nelem,                   /* Number of elements to compare */
    unsigned char *dst,           /* Result of comparisons */
    long   ds,                    /* Destination stride, in elements */
    int    stype,                 /* Common source type */
    void  *src1,                  /* First source array */
    long   ss1,                   /* First source stride, in elements */
    void  *src2,                  /* Second source array */
    long   ss2,                   /* Second source stride, in elements */
    unsigned char  gtval,         /* Output if greater than */
    unsigned char  leval)         /* Output if not greater than */

{
    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	COMP_LOOP(unsigned char, GT);
	return ERR_OKAY;

    case INTP_USHORT:
	COMP_LOOP(unsigned short, GT);
	return ERR_OKAY;

    case INTP_UINT:
	COMP_LOOP(unsigned int, GT);
	return ERR_OKAY;

    case INTP_ULONG:
	COMP_LOOP(unsigned long, GT);
	return ERR_OKAY;

    case INTP_SCHAR:
	COMP_LOOP(signed char, GT);
	return ERR_OKAY;

    case INTP_SHORT:
	COMP_LOOP(short, GT);
	return ERR_OKAY;

    case INTP_INT:
	COMP_LOOP(int, GT);
	return ERR_OKAY;

    case INTP_LONG:
	COMP_LOOP(long, GT);
	return ERR_OKAY;

    case INTP_FLOAT:
	COMP_LOOP(float, GT);
	return ERR_OKAY;

    case INTP_DOUBLE:
	COMP_LOOP(double, GT);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*--------------------------------------------------------------------------
bips_lt - Compare two arrays for less than

This function compares corresponding values from two numeric arrays in
any internal format and outputs one of two given unsigned char values
according as the value in the first array is less than or not less
than the value in the second.  The two arrays must be the same type.
Not defined for complex types.

NOTE that this function is atypical in that the result is always unsigned
char rather than being the same type as the source arrays; as a reminder of
this, the type argument precedes the source arguments rather than the
destination argument.  A consequence of this is that the standard BINOP_LOOP 
does not work here because the destination is always unsigned char; so we 
have to use a modified loop.

See also bips_compare() for three-way comparisons on non-complex types.
----------------------------------------------------------------------------*/

#define LT(w,x,y)           \
    if (x < y)  w = ltval;  \
    else        w = geval;  


int                               /* Returns nonzero on error */
bips_lt(
    long nelem,                   /* Number of elements to compare */
    unsigned char *dst,           /* Result of comparisons */
    long   ds,                    /* Destination stride, in elements */
    int    stype,                 /* Common source type */
    void  *src1,                  /* First source array */
    long   ss1,                   /* First source stride, in elements */
    void  *src2,                  /* Second source array */
    long   ss2,                   /* Second source stride, in elements */
    unsigned char ltval,          /* Output if less than */
    unsigned char geval)          /* Output if not less than */

{
    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	COMP_LOOP(unsigned char, LT);
	return ERR_OKAY;

    case INTP_USHORT:
	COMP_LOOP(unsigned short, LT);
	return ERR_OKAY;

    case INTP_UINT:
	COMP_LOOP(unsigned int, LT);
	return ERR_OKAY;

    case INTP_ULONG:
	COMP_LOOP(unsigned long, LT);
	return ERR_OKAY;

    case INTP_SCHAR:
	COMP_LOOP(signed char, LT);
	return ERR_OKAY;

    case INTP_SHORT:
	COMP_LOOP(short, LT);
	return ERR_OKAY;

    case INTP_INT:
	COMP_LOOP(int, LT);
	return ERR_OKAY;

    case INTP_LONG:
	COMP_LOOP(long, LT);
	return ERR_OKAY;

    case INTP_FLOAT:
	COMP_LOOP(float, LT);
	return ERR_OKAY;

    case INTP_DOUBLE:
	COMP_LOOP(double, LT);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*--------------------------------------------------------------------------
bips_min - Elementwise minimum of two arrays

All non-complex numeric types are supported, but the exact
characteristics and portability are inherited from the underlying C
implementation.  This function is not defined for the complex
types, because there is no order of the complex numbers which is
consistent with their arithmetic properties.
---------------------------------------------------------------------------*/

#define MIN(w,x,y)  w = ((x) < (y)) ? (x) : (y)

int
bips_min(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, MIN);
	return ERR_OKAY;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, MIN);
	return ERR_OKAY;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, MIN);
	return ERR_OKAY;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, MIN);
	return ERR_OKAY;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, MIN);
	return ERR_OKAY;

    case INTP_SHORT:
	BINOP_LOOP(short, MIN);
	return ERR_OKAY;

    case INTP_INT:
	BINOP_LOOP(int, MIN);
	return ERR_OKAY;

    case INTP_LONG:
	BINOP_LOOP(long, MIN);
	return ERR_OKAY;

    case INTP_FLOAT:
	BINOP_LOOP(float, MIN);
	return ERR_OKAY;

    case INTP_DOUBLE:
	BINOP_LOOP(double, MIN);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}




/*--------------------------------------------------------------------------
bips_max - Elementwise maximum of two arrays

All non-complex numeric types are supported, but the exact
characteristics and portability are inherited from the underlying C
implementation.  This function is not defined for the complex
types, because there is no order of the complex numbers which is
consistent with their arithmetic properties.
---------------------------------------------------------------------------*/

#define MAX(w,x,y)  w = ((x) > (y)) ? (x) : (y)

int
bips_max(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, MAX);
	return ERR_OKAY;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, MAX);
	return ERR_OKAY;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, MAX);
	return ERR_OKAY;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, MAX);
	return ERR_OKAY;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, MAX);
	return ERR_OKAY;

    case INTP_SHORT:
	BINOP_LOOP(short, MAX);
	return ERR_OKAY;

    case INTP_INT:
	BINOP_LOOP(int, MAX);
	return ERR_OKAY;

    case INTP_LONG:
	BINOP_LOOP(long, MAX);
	return ERR_OKAY;

    case INTP_FLOAT:
	BINOP_LOOP(float, MAX);
	return ERR_OKAY;

    case INTP_DOUBLE:
	BINOP_LOOP(double, MAX);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}




/*--------------------------------------------------------------------------
bips_add - Sum of two arrays

All numeric types are supported, but the exact characteristics and
portability are inherited from the underlying C implementation.

The destination stride may not be zero; use bips_addset to sum the
elements of a vector.

By special dispensation, bips_add handles the recursive case
correctly, even on an optimizing compiler.  For this case, the user
must specify ds == ss1 and dst == src1 + ss1; then the running sum
will be generated.  The second source stride ss2 may be zero.  It is
the caller's responsibility to set dst[-ss1] correctly.
---------------------------------------------------------------------------*/

#define ADD(w,x,y)  w = (x) + (y)

int
bips_add(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_RLOOP(unsigned char, ADD);
	return ERR_OKAY;

    case INTP_USHORT:
	BINOP_RLOOP(unsigned short, ADD);
	return ERR_OKAY;

    case INTP_UINT:
	BINOP_RLOOP(unsigned int, ADD);
	return ERR_OKAY;

    case INTP_ULONG:
	BINOP_RLOOP(unsigned long, ADD);
	return ERR_OKAY;

    case INTP_SCHAR:
	BINOP_RLOOP(signed char, ADD);
	return ERR_OKAY;

    case INTP_SHORT:
	BINOP_RLOOP(short, ADD);
	return ERR_OKAY;

    case INTP_INT:
	BINOP_RLOOP(int, ADD);
	return ERR_OKAY;

    case INTP_LONG:
	BINOP_RLOOP(long, ADD);
	return ERR_OKAY;

    case INTP_FLOAT:
	BINOP_RLOOP(float, ADD);
	return ERR_OKAY;

    case INTP_DOUBLE:
	BINOP_RLOOP(double, ADD);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	BINOP_RLOOP(float complex, ADD);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	BINOP_RLOOP(double complex, ADD);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}




/*--------------------------------------------------------------------------
bips_sub - Difference of two arrays

All numeric types are supported, but the exact characteristics and
portability are inherited from the underlying C implementation.
---------------------------------------------------------------------------*/

#define SUB(w,x,y)  w = (x) - (y)

int
bips_sub(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, SUB);
	return ERR_OKAY;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, SUB);
	return ERR_OKAY;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, SUB);
	return ERR_OKAY;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, SUB);
	return ERR_OKAY;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, SUB);
	return ERR_OKAY;

    case INTP_SHORT:
	BINOP_LOOP(short, SUB);
	return ERR_OKAY;

    case INTP_INT:
	BINOP_LOOP(int, SUB);
	return ERR_OKAY;

    case INTP_LONG:
	BINOP_LOOP(long, SUB);
	return ERR_OKAY;

    case INTP_FLOAT:
	BINOP_LOOP(float, SUB);
	return ERR_OKAY;

    case INTP_DOUBLE:
	BINOP_LOOP(double, SUB);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	BINOP_LOOP(float complex, SUB);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	BINOP_LOOP(double complex, SUB);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}




/*--------------------------------------------------------------------------
bips_mul - Product of two arrays

All numeric types are supported, but the exact characteristics and
portability are inherited from the underlying C implementation.

The destination stride may not be zero; use bips_mulset to multiply
the elements of a vector.

By special dispensation, bips_mul handles the recursive case
correctly, even on an optimizing compiler.  For this case, the user
must specify ds == ss1 and dst == src1 + ss1; then the running product
will be generated.  The second source stride ss2 may be zero.  It is
the caller's responsibility to set dst[-ss1] correctly.
---------------------------------------------------------------------------*/

#define MUL(w,x,y)  w = (x) * (y)

int
bips_mul(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_RLOOP(unsigned char, MUL);
	return ERR_OKAY;

    case INTP_USHORT:
	BINOP_RLOOP(unsigned short, MUL);
	return ERR_OKAY;

    case INTP_UINT:
	BINOP_RLOOP(unsigned int, MUL);
	return ERR_OKAY;

    case INTP_ULONG:
	BINOP_RLOOP(unsigned long, MUL);
	return ERR_OKAY;

    case INTP_SCHAR:
	BINOP_RLOOP(signed char, MUL);
	return ERR_OKAY;

    case INTP_SHORT:
	BINOP_RLOOP(short, MUL);
	return ERR_OKAY;

    case INTP_INT:
	BINOP_RLOOP(int, MUL);
	return ERR_OKAY;

    case INTP_LONG:
	BINOP_RLOOP(long, MUL);
	return ERR_OKAY;

    case INTP_FLOAT:
	BINOP_RLOOP(float, MUL);
	return ERR_OKAY;

    case INTP_DOUBLE:
	BINOP_RLOOP(double, MUL);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	BINOP_RLOOP(float complex, MUL);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	BINOP_RLOOP(double complex, MUL);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}




/*--------------------------------------------------------------------------
bips_div - Quotient of two arrays

All numeric types are supported, but the exact characteristics and
portability are inherited from the underlying C implementation.
---------------------------------------------------------------------------*/

#define DIV(w,x,y)  w = (x) / (y)

int
bips_div(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, DIV);
	return ERR_OKAY;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, DIV);
	return ERR_OKAY;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, DIV);
	return ERR_OKAY;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, DIV);
	return ERR_OKAY;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, DIV);
	return ERR_OKAY;

    case INTP_SHORT:
	BINOP_LOOP(short, DIV);
	return ERR_OKAY;

    case INTP_INT:
	BINOP_LOOP(int, DIV);
	return ERR_OKAY;

    case INTP_LONG:
	BINOP_LOOP(long, DIV);
	return ERR_OKAY;

    case INTP_FLOAT:
	BINOP_LOOP(float, DIV);
	return ERR_OKAY;

    case INTP_DOUBLE:
	BINOP_LOOP(double, DIV);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	BINOP_LOOP(float complex, DIV);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	BINOP_LOOP(double complex, DIV);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*--------------------------------------------------------------------------
bips_rem - Remainder of two arrays

Computes x - trunc(x/y) * y, where trunc() represents rounding toward zero.

All non-complex numeric types are supported, but the exact
characteristics and portability are inherited from the underlying C
implementation.  Not defined for complex types.
---------------------------------------------------------------------------*/

#define IREM(w,x,y)  w = (x) % (y)
#define FREM(w,x,y)  w = fmod(x,y)

int
bips_rem(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_UCHAR:
	BINOP_LOOP(unsigned char, IREM);
	return ERR_OKAY;

    case INTP_USHORT:
	BINOP_LOOP(unsigned short, IREM);
	return ERR_OKAY;

    case INTP_UINT:
	BINOP_LOOP(unsigned int, IREM);
	return ERR_OKAY;

    case INTP_ULONG:
	BINOP_LOOP(unsigned long, IREM);
	return ERR_OKAY;

    case INTP_SCHAR:
	BINOP_LOOP(signed char, IREM);
	return ERR_OKAY;

    case INTP_SHORT:
	BINOP_LOOP(short, IREM);
	return ERR_OKAY;

    case INTP_INT:
	BINOP_LOOP(int, IREM);
	return ERR_OKAY;

    case INTP_LONG:
	BINOP_LOOP(long, IREM);
	return ERR_OKAY;

    case INTP_FLOAT:
	BINOP_LOOP(float, FREM);
	return ERR_OKAY;

    case INTP_DOUBLE:
	BINOP_LOOP(double, FREM);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*--------------------------------------------------------------------------
bips_pow - Elementwise exponention of two arrays

Only floating point types are supported.

FIXME: It might be useful to directly support exponentiation to small
integral powers.  Also exponentiation of integral types modulo some
integer, though the relevance of this to image processing is dubious.
---------------------------------------------------------------------------*/

#define POW(w,x,y)  w = pow(x,y)

int
bips_pow(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	BINOP_LOOP(float, POW);
	return ERR_OKAY;

    case INTP_DOUBLE:
	BINOP_LOOP(double, POW);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	BINOP_LOOP(float complex, POW);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	BINOP_LOOP(double complex, POW);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*--------------------------------------------------------------------------
bips_atan2 - Elementwise four-quadrant arc tangent of two arrays

Only non-complex floating point types are supported.  This operation is
not defined for complex numbers.
---------------------------------------------------------------------------*/

#define ATAN2(w,x,y)  w = atan2(x,y)

int
bips_atan2(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	BINOP_LOOP(float, ATAN2);
	return ERR_OKAY;

    case INTP_DOUBLE:
	BINOP_LOOP(double, ATAN2);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*--------------------------------------------------------------------------
bips_fma_set - Elementwise in-place fused multiply-add of two arrays

Computes elementwise w = x*y + w.  The destination stride may be zero.
This function is implemented for complex types, but does not use the
fma() function in C99 because that function is not defined for complex
types.  Since real*complex multiplies are very useful, they are
supported directly by this function through the stype1 argument, which
must be either equal to type, or the equivalent real type.

FIXME: Should this go with the other in-place operations?

FIXME: Look to see if this can be used to improve efficiency in
operations such as dot product and moments.
---------------------------------------------------------------------------*/

#ifdef HAVE_FMA
#define FMASET(w,x,y)  w = fma(x,y,w)
#else
#define FMASET(w,x,y)  w = (x)*(y) + (w)
#endif

#define CFMASET(w,x,y)  w = (x)*(y) + (w)

int
bips_fma_set(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    int   stype1,               /* Type of first source */
    void *src1,                 /* First operand array */
    long  ss1,                  /* Stride in first operand array */
    void *src2,                 /* Second operand array */
    long  ss2)                  /* Stride in second operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
        if (stype1 == type) {
            BINOP_XLOOP(float, float, FMASET);
            return ERR_OKAY;
        }
        else
            NZReturnError(ERR_ILLOP, illtype);

    case INTP_DOUBLE:
        if (stype1 == type) {
            BINOP_XLOOP(double, double, FMASET);
            return ERR_OKAY;
        }
        else 
            NZReturnError(ERR_ILLOP, illtype);

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
        if (stype1 == type) {
            BINOP_XLOOP(float complex, float complex, CFMASET);
            return ERR_OKAY;
        }
        else if (stype1 == INTP_FLOAT) {
            BINOP_XLOOP(float, float complex, CFMASET);
            return ERR_OKAY;
        }
        else
            NZReturnError(ERR_ILLOP, illtype); 

    case INTP_DCOMPLEX:
        if (stype1 == type) {
            BINOP_XLOOP(double complex, double complex, CFMASET);
            return ERR_OKAY;
        }
        else if (stype1 == INTP_DOUBLE) {
            BINOP_XLOOP(double, double complex, CFMASET);
            return ERR_OKAY;
        }
        else
            NZReturnError(ERR_ILLOP, illtype); 
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/*--------------------------------------------------------------------------
bips_hypot - Elementwise hypot of two arrays

FIXME: This function is not defined in C99 for complex types.  The useful
definition is probably sqrt(|x|**2 + |y|**2).  If you should decide to
implement it, note that this has a real result even for complex inputs.
---------------------------------------------------------------------------*/

#ifdef HAVE_HYPOT
#define HYPOT(w,x,y)  w = hypot(x,y)

int
bips_hypot(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* Left operand array */
    long  ss1,                  /* Stride in left operand array */
    void *src2,                 /* Right operand array */
    long  ss2)                  /* Stride in right operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	BINOP_LOOP(float, HYPOT);
	return ERR_OKAY;

    case INTP_DOUBLE:
	BINOP_LOOP(double, HYPOT);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}
#endif  /* HAVE_HYPOT */




/***************************************************************************
Ternary point operations
***************************************************************************/


/*------------------------------------------------------------------------
TERNOP_LOOP - Generic loop for a ternary point operation

This macro defines a generic loop for implementing a ternary point
operation with arbitrary strides over three input arrays to generate a
fourth output array.  Any input stride may be zero, but not all three;
the code optimizes for speed in this case.

The operation to be performed is given by a macro "asgn" which is
expected to take four arguments and expand into an assignment which
assigns to the first argument the result of the operation on the
second and third arguments; the terminal semicolon must be omitted.
For example,

#define  ADD(d,s1,s2,s3)   d = (s1) + (s2) + (s3)

would implement ternary addition.

This macro assumes the following variable definitions to be in effect:

    void *dst;                   Pointer to destination array
    void *src1, *src2, *src3;    Pointers to source arrays
    long ds;                     Stride in destination array
    long ss1, ss2, ss3;          Strides in source arrays
    long nelem;                  Number of elements to process

FIXME: Are the ternary equivalents of BINOP_RLOOP and BINOP_XLOOP 
useful?

FIXME: Think about the best ordering of the tests below.  Consider using
a switch instead; it might be faster.

FIXME: Look at the assembly code and verify that the restrict
qualifiers are doing the right thing.  Add tests to catch errors in
the restrictions?

FIXME:  Are side parameters supported cleanly?  I think so, but it needs
to be checked; type conversions might be a problem.
-------------------------------------------------------------------------*/

/* No overlaps and dest stride != 0 */
#define TERNOP_LOOP(type, asgn)  {           \
    type * restrict d  = (type *)dst;        \
    type * restrict s1 = (type *)src1;       \
    type * restrict s2 = (type *)src2;       \
    type * restrict s3 = (type *)src3;       \
    long i;                                  \
    if (ss2 == 0 && ss3 == 0)                \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2, *s3);         \
            d += ds; s1 += ss1; }            \
    else if (ss1 == 0 && ss3 == 0)           \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2, *s3);         \
            d += ds; s2 += ss2; }            \
    else if (ss1 == 0 && ss2 == 0)           \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2, *s3);         \
            d += ds; s3 += ss3; }            \
    else if (ss1 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2, *s3);         \
            d += ds; s2 += ss2; s3 += ss3;}  \
    else if (ss2 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2, *s3);         \
            d += ds; s1 += ss1; s3 += ss3;}  \
    else if (ss3 == 0)                       \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2, *s3);         \
            d += ds; s1 += ss1; s2 += ss2; } \
    else                                     \
        for (i = 0; i < nelem; i++)  {       \
            asgn(*d, *s1, *s2, *s3);         \
            d  += ds;  s1 += ss1;            \
            s2 += ss2; s3 += ss3; }          \
    }



/*--------------------------------------------------------------------------
bips_fma - Elementwise fused multiply-add of two arrays

Computes elementwise w = x*y + z.  This function is implemented for
complex types, but does not use the fma() function in C99 because that
function is not defined for complex types.

FIXME: Modify to support in-place operations, and maybe other
duplicated operands.  Look to see if this can be used to improve
efficiency in operations such as dot product and moments.
---------------------------------------------------------------------------*/

#ifdef HAVE_FMA
#define FMA(w,x,y,z)  w = fma(x,y,z)
#else
#define FMA(w,x,y,z)  w = (x)*(y) + (z);
#endif

#define CFMA(w,x,y,z)  w = (x)*(y) + (z);

int
bips_fma(
    long  nelem,                /* Number of elements to process */
    int   type,                 /* Type of data to process */
    void *dst,                  /* Destination array */
    long  ds,                   /* Stride in destination array */
    void *src1,                 /* First operand array */
    long  ss1,                  /* Stride in first operand array */
    void *src2,                 /* Second operand array */
    long  ss2,                  /* Stride in second operand array */
    void *src3,                 /* Third operand array */
    long  ss3)                  /* Stride in third operand array */
{
    /* Switch according to the source type */
    switch (type) {

    case INTP_FLOAT:
	TERNOP_LOOP(float, FMA);
	return ERR_OKAY;

    case INTP_DOUBLE:
	TERNOP_LOOP(double, FMA);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H

    case INTP_FCOMPLEX:
	TERNOP_LOOP(double, CFMA);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	TERNOP_LOOP(double, CFMA);
	return ERR_OKAY;

#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    
}



/***************************************************************************
Operations with an unusual format

Most of these are inner loops for heavily used functions that could be
done, albeit more expensively, using the simpler operations above.
***************************************************************************/



/*------------------------------------------------------------------------
bips_absdiff - Absolute difference of two arrays

This function computes the absolute value of the difference between
two arrays and converts it to double; the second array may be omitted
(by supplying a zero pointer), in which case the absolute value of the
first array is computed.  Both source arrays must be of the same type;
complex types are supported.  This function is intended to support the 
computation of vector norms, hence the somewhat unusual format.

See also bips_sqrdiff().
-------------------------------------------------------------------------*/

#define ABSDIFF_LOOP(type)                     \
{                                              \
    type * restrict s1 = src1;                 \
    type * restrict s2 = src2;                 \
    if (s2 == 0) {                             \
        for (i = 0; i < nelem; i++)  {         \
            *d = fabs((double)*s1);            \
            d += ds; s1 += ss1; }  }           \
    else {                                     \
        for (i = 0; i < nelem; i++) {          \
            *d = fabs((double)*s1 - *s2);      \
            d += ds; s1 += ss1; s2 += ss2; } } \
}

#define CABSDIFF_LOOP(type)                     \
{                                               \
    type * restrict s1 = src1;                  \
    type * restrict s2 = src2;                  \
    if (s2 == 0) {                              \
        for (i = 0; i < nelem; i++)  {          \
            *d = cabs(*s1);                     \
            d += ds; s1 += ss1; }  }            \
    else {                                      \
        for (i = 0; i < nelem; i++) {           \
            *d = cabs(*s1 - *s2);               \
            d += ds; s1 += ss1; s2 += ss2; } }  \
}


int 
bips_absdiff(
    long nelem,         /* Number of elements to process */
    double *dst,        /* Destination array */
    long ds,            /* Stride in destination array */
    int stype,          /* Type of the source array */
    void *src1,         /* First source array */
    long ss1,           /* Stride in first source array */
    void *src2,         /* Second source array */
    long ss2)           /* Stride in second source array */
{
    double * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	ABSDIFF_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_USHORT:
	ABSDIFF_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_UINT:
	ABSDIFF_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_ULONG:
	ABSDIFF_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_SCHAR:
	ABSDIFF_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	ABSDIFF_LOOP(short);
	return ERR_OKAY;

    case INTP_INT:
	ABSDIFF_LOOP(int);
	return ERR_OKAY;

    case INTP_LONG:
	ABSDIFF_LOOP(long);
	return ERR_OKAY;

    case INTP_FLOAT:
	ABSDIFF_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	ABSDIFF_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX :
        CABSDIFF_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX :
        CABSDIFF_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    


}



/*------------------------------------------------------------------------
bips_sqrdiff - Squared absolute difference of two arrays

This function computes the square of the absolute value of the
difference between two arrays and converts it to double; the second
array may be omitted (by supplying a zero pointer), in which case the
squared absolute value of the first array is computed.  Both source arrays
must be of the same type; complex types are supported.  This function
is intended to support the computation of vector norms, hence the
somewhat different format.

See also bips_absdiff().
-------------------------------------------------------------------------*/

#define SQRDIFF_LOOP(type)                     \
{                                              \
    type * restrict s1 = src1;                 \
    type * restrict s2 = src2;                 \
    type dd;                                   \
    if (s2 == 0) {                             \
        for (i = 0; i < nelem; i++)  {         \
            dd = fabs((double)*s1);            \
            *d = dd*dd;                        \
            d += ds; s1 += ss1; }  }           \
    else {                                     \
        for (i = 0; i < nelem; i++) {          \
            dd = fabs((double)*s1 - *s2);      \
            *d = dd*dd;                        \
            d += ds; s1 += ss1; s2 += ss2; } } \
}

#define CSQRDIFF_LOOP(type)                     \
{                                               \
    type * restrict s1 = src1;                  \
    type * restrict s2 = src2;                  \
    type dd;                                    \
    double dr, di;                              \
    if (s2 == 0) {                              \
        for (i = 0; i < nelem; i++)  {          \
            dr = creal(*s1);  di = cimag(*s1);  \
            *d = dr*dr + di*di;                 \
            d += ds; s1 += ss1; }  }            \
    else {                                      \
        for (i = 0; i < nelem; i++) {           \
            dd = (double)*s1 - *s2;             \
            dr = creal(dd);  di = cimag(dd);    \
            *d = dr*dr + di*di;                 \
            d += ds; s1 += ss1; s2 += ss2; } }  \
}


int 
bips_sqrdiff(
    long nelem,         /* Number of elements to process */
    double *dst,        /* Destination array */
    long ds,            /* Stride in destination array */
    int stype,          /* Type of the source array */
    void *src1,         /* First source array */
    long ss1,           /* Stride in first source array */
    void *src2,         /* Second source array */
    long ss2)           /* Stride in second source array */
{
    double * restrict d = dst;
    long i;

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	SQRDIFF_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_USHORT:
	SQRDIFF_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_UINT:
	SQRDIFF_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_ULONG:
	SQRDIFF_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_SCHAR:
	SQRDIFF_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	SQRDIFF_LOOP(short);
	return ERR_OKAY;

    case INTP_INT:
	SQRDIFF_LOOP(int);
	return ERR_OKAY;

    case INTP_LONG:
	SQRDIFF_LOOP(long);
	return ERR_OKAY;

    case INTP_FLOAT:
	SQRDIFF_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	SQRDIFF_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX :
        CSQRDIFF_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX :
        CSQRDIFF_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    


}



/*------------------------------------------------------------------------
bips_chamfer - Inner loop for chamfer distance

This functions computes the chamfer distance in one direction using
the following algorithm for each destination voxel, where DMAX is the
maximum possible value for the destination type.  If the source voxel
is non-zero, then the destination is set to zero; else if the
destination is less than the offset destination, the destination is
unchanged; else if the offset destination is DMAX, the destination is
set to DMAX; otherwise, the destination is set to the offset
destination plus 1.  Note that the most probable use of this function
requires the caller to initialize the destination to the maximum
possible distance value by hand before calling this function in each
direction starting with the second destination pixel.  If you didn't
understand all that, don't worry; it's unlikely that this code is
useful anywhere other than the vxl_chamfer function.

Not implemented for complex types, since there doesn't seem to be any
compelling use case.
-------------------------------------------------------------------------*/

#define CHAMFER_LOOP(dtype, stype, dmax)    \
{                                           \
    long i;                                 \
    dtype * restrict d = dst;               \
    stype * restrict s = src;               \
    for (i = 0; i < nelem; i++) {           \
        if ( s[0] != 0 )                    \
            d[0] = 0;                       \
        else if ( d[0] <= d[doff] )         \
            /* Do nothing */ ;              \
        else if ( d[doff] == dmax )         \
            d[0] = dmax;                    \
        else                                \
            d[0] = d[doff] + 1;             \
        d += ds;   s += ss;                 \
        }                                   \
                                            \
}

int
bips_chamfer(
    long nelem,              /* Number of elements to process */
    int dtype,               /* Destination type: uchar or ushort */
    void *dst,               /* Destination array */
    long ds,                 /* Destination stride */
    long doff,               /* Offset to prev pixel */
    int stype,               /* Source type */
    void *src,               /* Source array */
    long ss)                 /* Source stride */
{

    /* Switch according to the destination type */
    switch (dtype) {

    case INTP_UCHAR:

        /* Dest is unsigned char; now switch according to the source type */
	switch (stype) {

	case INTP_UCHAR:
	    CHAMFER_LOOP(unsigned char, unsigned char, UCHAR_MAX);
	    return ERR_OKAY;

	case INTP_SCHAR:
	    CHAMFER_LOOP(unsigned char, signed char, UCHAR_MAX);
	    return ERR_OKAY;

	case INTP_USHORT:
	    CHAMFER_LOOP(unsigned char, unsigned short, UCHAR_MAX);
	    return ERR_OKAY;

	case INTP_SHORT:
	    CHAMFER_LOOP(unsigned char, short, UCHAR_MAX);
	    return ERR_OKAY;

	case INTP_UINT:
	    CHAMFER_LOOP(unsigned char, unsigned int, UCHAR_MAX);
	    return ERR_OKAY;

	case INTP_INT:
	    CHAMFER_LOOP(unsigned char, int, UCHAR_MAX);
	    return ERR_OKAY;

	case INTP_ULONG:
	    CHAMFER_LOOP(unsigned char, unsigned long, UCHAR_MAX);
	    return ERR_OKAY;

	case INTP_LONG:
	    CHAMFER_LOOP(unsigned char, long, UCHAR_MAX);
	    return ERR_OKAY;

	case INTP_FLOAT:
	    CHAMFER_LOOP(unsigned char, float, UCHAR_MAX);
	    return ERR_OKAY;

	case INTP_DOUBLE:
	    CHAMFER_LOOP(unsigned char, double, UCHAR_MAX);
	    return ERR_OKAY;

	default:
	    NZReturnError(ERR_ILLOP, illtype);

	}

    case INTP_USHORT:

        /* Dest is unsigned char; now switch according to the source type */
	switch (stype) {

	case INTP_UCHAR:
	    CHAMFER_LOOP(unsigned short, unsigned char, USHRT_MAX);
	    return ERR_OKAY;

	case INTP_SCHAR:
	    CHAMFER_LOOP(unsigned short, signed char, USHRT_MAX);
	    return ERR_OKAY;

	case INTP_USHORT:
	    CHAMFER_LOOP(unsigned short, unsigned short, USHRT_MAX);
	    return ERR_OKAY;

	case INTP_SHORT:
	    CHAMFER_LOOP(unsigned short, short, USHRT_MAX);
	    return ERR_OKAY;

	case INTP_UINT:
	    CHAMFER_LOOP(unsigned short, unsigned int, USHRT_MAX);
	    return ERR_OKAY;

	case INTP_INT:
	    CHAMFER_LOOP(unsigned short, int, USHRT_MAX);
	    return ERR_OKAY;

	case INTP_ULONG:
	    CHAMFER_LOOP(unsigned short, unsigned long, USHRT_MAX);
	    return ERR_OKAY;

	case INTP_LONG:
	    CHAMFER_LOOP(unsigned short, long, USHRT_MAX);
	    return ERR_OKAY;

	case INTP_FLOAT:
	    CHAMFER_LOOP(unsigned short, float, USHRT_MAX);
	    return ERR_OKAY;

	case INTP_DOUBLE:
	    CHAMFER_LOOP(unsigned short, double, USHRT_MAX);
	    return ERR_OKAY;

	default:
	    NZReturnError(ERR_ILLOP, illtype);

	}

    default:
	NZReturnError(ERR_ILLOP, illtype);
    }
}



/*------------------------------------------------------------------------
bips_moments - First and second moments of an array

This function computes the first and second non-central moments of a
one-dimensional vector; the index is taken as the space variable.  The
moments are always returned in double, regardless of the data format.

FIXME: Generalize to support complex types?  The definition is the
same as for real types, but the result type is complex.  I'm inclined
to postpone this until either I need it or get a better idea of how to
handle operations with variable intermediate and output types.
-------------------------------------------------------------------------*/

#define MOM_LOOP(type)             \
{                                  \
    type * restrict s = src;       \
    for (i = 0; i < nelem; i++) {  \
        d0 += (m  = (double)(*s)); \
        d1 += (m *= (double)i);    \
        d2 += (m *= (double)i);    \
        s += ss;  }                \
}

int
bips_moments(
    long nelem,             /* Number of elements in the array */
    double dst[3],          /* Zeroth, first, and second moments */
    int stype,              /* Type of the source array */
    void *src,              /* Source array */
    long ss)                /* Source stride */
{
    double d0, d1, d2;      /* Moments so far */
    double m;               /* Voxel value */
    long i;

    /* Initialize the moments to zero */
    d0 = d1 = d2 = 0.0;

    /* Switch on the source type */
    switch (stype) {

    case INTP_UCHAR:
	MOM_LOOP(unsigned char);
	break;

    case INTP_USHORT:
	MOM_LOOP(unsigned short);
	break;

    case INTP_UINT:
	MOM_LOOP(unsigned int);
	break;

    case INTP_ULONG:
	MOM_LOOP(unsigned long);
	break;

    case INTP_SCHAR:
	MOM_LOOP(signed char);
	break;

    case INTP_SHORT:
	MOM_LOOP(short);
	break;

    case INTP_INT:
	MOM_LOOP(int);
	break;

    case INTP_LONG:
	MOM_LOOP(long);
	break;

    case INTP_FLOAT:
	MOM_LOOP(float);
	break;

    case INTP_DOUBLE:
	MOM_LOOP(double);
	break;

    default:
	NZReturnError(ERR_ILLOP, illtype);
    }

    /* Save the computed moments and return */
    dst[0] = d0;
    dst[1] = d1;
    dst[2] = d2;
    return ERR_OKAY;
}



/*-------------------------------------------------------------------------
bips_conv1 - Inner loop of convolution

This function performs the innermost loop of a convolution.  Each
souce element is converted to the destination type, multiplied by a
fixed coefficient, and added to the destination.  Only certain
combinations of destination and coefficient types are supported, as
shown below.  If the source type is complex, the destination type must
be double complex.

      dest type            coef type
      ---------            ---------
      double               double
      double complex       double
      double complex       double complex


FIXME: Optimize for the case where the stride is one?  Or is there enough
floating point computation here that it doesn't matter?

FIXME: An alternate approach to specifying the coefficient would be to
always pass it as double complex, but use only the real part if the
destination is real.  If this is used, then it should be possible to
have ctype defined implicitly.  See bips_lostat[12] for an example.

NOTE: Prior to 2005-10-05, the dtype and ctype arguments were absent,
the dst argument was double *, and the coef argument was double.
-------------------------------------------------------------------------*/

#define CONV1_LOOP(dtype, stype, ctype)  \
{                                        \
    ctype c = *(ctype *)coef;            \
    dtype * restrict d = (dtype *)dst;   \
    stype * restrict s = (stype *)src;   \
    for (i = 0; i < nelem; i++) {        \
        *d += c*((dtype)*s);             \
        d += ds; s += ss;  }             \
}

int 
bips_conv1(
    long nelem,                   /* Number of elements to process */
    int dtype,                    /* Type of destination and coefficient */
    void *dst,                    /* Destination: convolution sum */
    long ds,                      /* Stride in destination array */
    int stype,                    /* Type of source data */
    void *src,                    /* Source array */
    long ss,                      /* Stride in source array */
    int ctype,                    /* Type of coefficient */
    void *coef)                   /* Coefficient to multiply by */
{
    long i;

    /* Case 1: Destination and coeficient are both double */
    if (dtype == INTP_DOUBLE && ctype == INTP_DOUBLE) {

        /* Switch on the source type */
        switch (stype) {
            
        case INTP_UCHAR :
            CONV1_LOOP(double, unsigned char, double);
            return ERR_OKAY;

        case INTP_USHORT :
            CONV1_LOOP(double, unsigned short, double);
            return ERR_OKAY;

        case INTP_UINT :
            CONV1_LOOP(double, unsigned int, double);
            return ERR_OKAY;

        case INTP_ULONG :
            CONV1_LOOP(double, unsigned long, double);
            return ERR_OKAY;

        case INTP_SCHAR :
            CONV1_LOOP(double, signed char, double);
            return ERR_OKAY;

        case INTP_SHORT :
            CONV1_LOOP(double, short, double);
            return ERR_OKAY;

        case INTP_INT :
            CONV1_LOOP(double, int, double);
            return ERR_OKAY;

        case INTP_LONG :
            CONV1_LOOP(double, long, double);
            return ERR_OKAY;

        case INTP_FLOAT :
            CONV1_LOOP(double, float, double);
            return ERR_OKAY;

        case INTP_DOUBLE :
            CONV1_LOOP(double, double, double);
            return ERR_OKAY;

        default:
            NZReturnError(ERR_ILLOP, illtype);    
        }
    }

#ifdef HAVE_COMPLEX_H
    /* Case 2: Destination is double complex; coefficient is plain double */
    else if (dtype == INTP_DCOMPLEX && ctype == INTP_DOUBLE) {

        /* Switch on the source type */
        switch (stype) {
            
        case INTP_UCHAR :
            CONV1_LOOP(double complex, unsigned char, double);
            return ERR_OKAY;

        case INTP_USHORT :
            CONV1_LOOP(double complex, unsigned short, double);
            return ERR_OKAY;

        case INTP_UINT :
            CONV1_LOOP(double complex, unsigned int, double);
            return ERR_OKAY;

        case INTP_ULONG :
            CONV1_LOOP(double complex, unsigned long, double);
            return ERR_OKAY;

        case INTP_SCHAR :
            CONV1_LOOP(double complex, signed char, double);
            return ERR_OKAY;

        case INTP_SHORT :
            CONV1_LOOP(double complex, short, double);
            return ERR_OKAY;

        case INTP_INT :
            CONV1_LOOP(double complex, int, double);
            return ERR_OKAY;

        case INTP_LONG :
            CONV1_LOOP(double complex, long, double);
            return ERR_OKAY;

        case INTP_FLOAT :
            CONV1_LOOP(double complex, float, double);
            return ERR_OKAY;

        case INTP_DOUBLE :
            CONV1_LOOP(double complex, double, double);
            return ERR_OKAY;

        case INTP_FCOMPLEX :
            CONV1_LOOP(double complex, float complex, double);
            return ERR_OKAY;

        case INTP_DCOMPLEX :
            CONV1_LOOP(double complex, double complex, double);
            return ERR_OKAY;

        default :
            NZReturnError(ERR_ILLOP, illtype);    
        }
    }

    /* Case 3: Destination and coeficient are both double complex */
    else if (dtype == INTP_DCOMPLEX && ctype == INTP_DCOMPLEX) {

        /* Switch on the source type */
        switch (stype) {
            
        case INTP_UCHAR :
            CONV1_LOOP(double complex, unsigned char, double complex);
            return ERR_OKAY;

        case INTP_USHORT :
            CONV1_LOOP(double complex, unsigned short, double complex);
            return ERR_OKAY;

        case INTP_UINT :
            CONV1_LOOP(double complex, unsigned int, double complex);
            return ERR_OKAY;

        case INTP_ULONG :
            CONV1_LOOP(double complex, unsigned long, double complex);
            return ERR_OKAY;

        case INTP_SCHAR :
            CONV1_LOOP(double complex, signed char, double complex);
            return ERR_OKAY;

        case INTP_SHORT :
            CONV1_LOOP(double complex, short, double complex);
            return ERR_OKAY;

        case INTP_INT :
            CONV1_LOOP(double complex, int, double complex);
            return ERR_OKAY;

        case INTP_LONG :
            CONV1_LOOP(double complex, long, double complex);
            return ERR_OKAY;

        case INTP_FLOAT :
            CONV1_LOOP(double complex, float, double complex);
            return ERR_OKAY;

        case INTP_DOUBLE :
            CONV1_LOOP(double complex, double, double complex);
            return ERR_OKAY;

        case INTP_FCOMPLEX :
            CONV1_LOOP(double complex, float complex, double complex);
            return ERR_OKAY;

        case INTP_DCOMPLEX :
            CONV1_LOOP(double complex, double complex, double complex);
            return ERR_OKAY;

        default :
            NZReturnError(ERR_ILLOP, illtype);    
        }
    }
#endif

    else
        NZReturnError(ERR_ILLOP, "Unsupported combination of types");

}



/*-------------------------------------------------------------------------
bips_lostat1 - Inner loop of local statistics

This function performs the innermost loop of a local statistics
computation: summing the weights, values, and squared values over a
neighborhood.  The source stride is specified by the caller; the
destination stride is always 1.

If stype is complex, then the d1 argument is taken as pointer to
double complex; otherwise d1 taken as pointer to double.  The d0 and
d2 arguments are always taken as pointer to double; and the coef
argument is always taken as double.

The destination arrays d0, d1, and d2 must not overlap, or the results
are unpredictable.

SEE ALSO: bips_lostat2().

ALGORITHM: The test (coef == 1.0) below is an exception to the usual
rule that you never test floats for exact equality; if coef is 1.0, it
will almost certainly have been created by assigning the value 1.0 and
will be exactly equal.  (However, computed values or values read in
from a file might not match exactly.)

FIXME: Optimize for the case where the source stride is one?  Or is
there enough floating point computation here that it doesn't matter?
-------------------------------------------------------------------------*/

/* Loop for real types */
#define LOSTAT1_LOOP(stype)                  \
{                                            \
    stype  * restrict s =  (stype  *)src;    \
    double * restrict d1 = (double *)dest1;  \
    double sv;                               \
    if (coef == 1.0)                         \
        for (i = nelem; i > 0; i--) {        \
            sv = (double)*s;                 \
            *d0++ += 1.0;                    \
            *d1++ += sv;                     \
            *d2++ += sv*sv;                  \
            s += ss;  }                      \
    else                                     \
        for (i = 0; i < nelem; i++) {        \
            sv = (double)*s;                 \
            *d0++ += coef;                   \
            *d1++ += coef*sv;                \
            *d2++ += coef*sv*sv;             \
            s += ss;  }                      \
}

/* Loop for complex types */
#ifdef HAVE_COMPLEX_H
#define LOSTAT1_CLOOP(stype)                     \
{                                                \
    stype    * restrict s = (stype     *)src;    \
    dcomplex * restrict d1 = (dcomplex *)dest1;  \
    dcomplex sv;                                 \
    double rp, ip;                               \
    if (coef == 1.0)                             \
        for (i = nelem; i > 0; i--) {            \
            sv = (dcomplex)*s;                   \
            rp = creal(sv);                      \
            ip = cimag(sv);                      \
            *d0++ += 1.0;                        \
            *d1++ += sv;                         \
            *d2++ += rp*rp+ip*ip;                \
            s += ss;  }                          \
    else                                         \
        for (i = 0; i < nelem; i++) {            \
            sv = (dcomplex)*s;                   \
            rp = creal(sv);                      \
            ip = cimag(sv);                      \
            *d0++ += coef;                       \
            *d1++ += coef*sv;                    \
            *d2++ += coef*(rp*rp+ip*ip);         \
            s += ss;  }                          \
}
#endif

int 
bips_lostat1(
    long nelem,                   /* Number of elements to process */
    double *dest0,                /* Sum of weights */
    void   *dest1,                /* Sum of values */
    double *dest2,                /* Sum of squared values */
    int stype,                    /* Type of source data */
    void *src,                    /* Source array */
    long ss,                      /* Stride in source array */
    double coef)                  /* Weight to multiply by */
{
    double * restrict d0 = (double *)dest0;
    double * restrict d2 = (double *)dest2;
    long i;

    /* Switch on the source type */
    switch (stype) {

    case INTP_UCHAR:
	LOSTAT1_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_USHORT:
	LOSTAT1_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_UINT:
	LOSTAT1_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_ULONG:
	LOSTAT1_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_SCHAR:
	LOSTAT1_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	LOSTAT1_LOOP(short);
	return ERR_OKAY;

    case INTP_INT:
	LOSTAT1_LOOP(int);
	return ERR_OKAY;

    case INTP_LONG:
	LOSTAT1_LOOP(long);
	return ERR_OKAY;

    case INTP_FLOAT:
	LOSTAT1_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	LOSTAT1_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
        LOSTAT1_CLOOP(float complex);
        return ERR_OKAY;

    case INTP_DCOMPLEX:
        LOSTAT1_CLOOP(double complex);
        return ERR_OKAY;
#endif

    default:
	NZReturnError(ERR_ILLOP, illtype);    
    }
}



/*-------------------------------------------------------------------------
bips_lostat2 - Finish up a local statistics calculation

This function takes three double arrays containing the summed weights,
values and squared values, computes the corresponding means and
standard deviations, and stores them to the destination arrays.  If
the dtype argument is complex, then the s1 argument is taken as
pointer to double complex; otherwise it is taken as pointer to real
double.  The s0 and s2 arguments are always taken as pointer to
double; the scale argument is always taken as double.

The destination vectors d0 and d1 must not overlap, or the results are
unpredictable.  

The destination stride is specified by the caller; the source stride
is always assumed to be one.

SEE ALSO: bips_lostat1().


FIXME: Allow the option of obtaining only the mean or variance, to
support applications that need only one of them?  The obvious way to
signal this is to use NULL for the appropriate destination.

FIXME: Optimize for the case where the destination stride is one?  Or
is there enough floating point computation here that it doesn't
matter?
-------------------------------------------------------------------------*/

/* Loop for integral destinations: must limit and round */
#define LOSTAT2_LOOP(dtype,lo,hi)           \
{                                           \
    double * restrict s1 = (double *)src1;  \
    dtype  * restrict d1 = (dtype  *)dst1;  \
    dtype  * restrict d2 = (dtype  *)dst2;  \
    double w, m, v;                         \
    for (i = 0; i < nelem; i++) {           \
        w = *s0;                            \
        m = *s1 / w;                        \
        *d1 = (dtype)LIMIT(m,lo,hi);        \
        v = scale*sqrt( *s2/w - m*m );      \
        *d2 = (dtype)LIMIT(v,lo,hi);        \
        s0++; s1++; s2++;                   \
        d1 += ds; d2 += ds; }               \
}

/* Loop for real floating point destination types: no need to round */
#define LOSTAT2_FLOOP(dtype)                \
{                                           \
    double * restrict s1 = (double *)src1;  \
    dtype  * restrict d1 = (dtype  *)dst1;  \
    dtype  * restrict d2 = (dtype  *)dst2;  \
    double w, m;                            \
    for (i = 0; i < nelem; i++) {           \
        w = *s0;                            \
        *d1 = m = *s1 / w ;                 \
        *d2 = scale*sqrt( *s2/w - m*m );    \
        s0++; s1++; s2++;                   \
        d1 += ds; d2 += ds; }               \
}

/* Loop for complex floating point destination types: no need to round */
#ifdef HAVE_COMPLEX_H
#define LOSTAT2_CLOOP(dtype)                    \
{                                               \
    dcomplex * restrict s1 = (dcomplex *)src1;  \
    dtype    * restrict d1 = (dtype    *)dst1;  \
    dtype    * restrict d2 = (dtype    *)dst2;  \
    double w, mr, mi, v;                        \
    dcomplex m;                                 \
    for (i = 0; i < nelem; i++) {               \
        w = *s0;                                \
        *d1 = m = *s1 / w ;                     \
        mr = creal(m);  mi = cimag(m);          \
        v  = *s2/w - mr*mr + mi*mi;             \
        *d2 = scale*sqrt(v);                    \
        s0++; s1++; s2++;                       \
        d1 += ds; d2 += ds; }                   \
}
#endif

int 
bips_lostat2(
    long nelem,                   /* Number of elements to process */
    int dtype,                    /* Type of destination data */
    void *dst1,                   /* Destination for mean values */
    void *dst2,                   /* Destination for standard deviations */
    long ds,                      /* Stride in destination array */
    double *src0,                 /* Source for sums of weights */
    void   *src1,                 /* Source for sums of values */
    double *src2,                 /* Source for sums of squared values */
    double scale)                 /* Multiplier to scale std deviations */
{
    double * restrict s0 = src0;
    double * restrict s2 = src2;
    long i;

    /* Switch on the destination type */
    switch (dtype) {

    case INTP_UCHAR:
	LOSTAT2_LOOP(unsigned char, 0, UCHAR_MAX);
	return ERR_OKAY;

    case INTP_USHORT:
	LOSTAT2_LOOP(unsigned short, 0, USHRT_MAX);
	return ERR_OKAY;

    case INTP_UINT:
	LOSTAT2_LOOP(unsigned int, 0, UINT_MAX);
	return ERR_OKAY;

    case INTP_ULONG:
	LOSTAT2_LOOP(unsigned long, 0, ULONG_MAX);
	return ERR_OKAY;

    case INTP_SCHAR:
	LOSTAT2_LOOP(signed char, SCHAR_MIN, SCHAR_MAX);
	return ERR_OKAY;

    case INTP_SHORT:
	LOSTAT2_LOOP(short, SHRT_MIN, SHRT_MAX);
	return ERR_OKAY;

    case INTP_INT:
	LOSTAT2_LOOP(int, INT_MIN, INT_MAX);
	return ERR_OKAY;

    case INTP_LONG:
	LOSTAT2_LOOP(long, LONG_MIN, LONG_MAX);
	return ERR_OKAY;

    case INTP_FLOAT:
	LOSTAT2_FLOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	LOSTAT2_FLOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	LOSTAT2_CLOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	LOSTAT2_CLOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);    

}



/*--------------------------------------------------------------------------
bips_lookup - Process through lookup table

This function takes each element of the source array (which must be either
unsigned char or unsigned short), uses it as an index into a table of
values (which may be any type), and puts that entry from the table into
the output.  Neither the source nor the destination stride may be zero;
the source and destination may not overlap.  If any index is outside the
table range, then the corresponding output is set to zero, and an error
flag is returned (after all outputs have been set).

NOTE: It is not necessary to check the LUT index against zero, because it
is unsigned and thus cannot be less than zero.  But note that if we ever
implement signed indices, this will have to be added.
--------------------------------------------------------------------------*/

#define LUT_LOOP(type, xtype)                     \
{                                                 \
    type  * restrict d = (type *)dst;             \
    type  * restrict t = (type *)table;           \
    xtype * restrict s = (xtype *)src;            \
    long i;                                       \
    if (ts == 1)                                  \
        for (i = 0; i < nelem; i++) {             \
            if (*s < ntable)                      \
  	        *d = t[*s];                       \
            else {                                \
                *d = 0; flag = ERR_DOMAIN;  }     \
            d += ds; s += ss; }                   \
    else                                          \
        for (i = 0; i < nelem; i++) {             \
            if (*s < ntable)                      \
	        *d = t[(*s)*ts];                  \
             else {                               \
                *d = 0; flag = ERR_DOMAIN;  }     \
             d += ds; s += ss; }                  \
    if (flag == ERR_DOMAIN)                       \
        set_error(flag, indexerr);                \
}

int
bips_lookup(
    long nelem,                 /* Number of elements to process */
    int type,                   /* Type of table and result */
    int xtype,                  /* Type of index array */
    void *dst,                  /* Destination array */
    long ds,                    /* Stride in destination array */
    void *src,                  /* Source array of index values */
    long ss,                    /* Stride in source array */
    void *table,                /* Table of values */
    long ts,                    /* Stride in table array */
    long ntable)                /* Number of elements in the table */
{
    char indexerr[] = "LUT index out of range";
    char striderr[] = "Source or destination stride may not be zero";
    int flag = 0;               /* Error flag */

    /* Check for valid strides */
    if (ds == 0 || ss == 0)
        return set_error(ERR_ILLOP, striderr);

    /* If unsigned char index, switch on the table type */
    if (xtype == INTP_UCHAR)
	switch (type) {
	    
	case INTP_UCHAR:
	    LUT_LOOP(unsigned char, unsigned char);
	    return flag;

	case INTP_SCHAR:
	    LUT_LOOP(signed char, unsigned char);
	    return flag;

	case INTP_USHORT:
	    LUT_LOOP(unsigned short, unsigned char);
	    return flag;

	case INTP_SHORT:
	    LUT_LOOP(short, unsigned char);
	    return flag;

	case INTP_UINT:
	    LUT_LOOP(unsigned int, unsigned char);
	    return flag;

	case INTP_INT:
	    LUT_LOOP(int, unsigned char);
	    return flag;

	case INTP_ULONG:
	    LUT_LOOP(unsigned long, unsigned char);
	    return flag;

	case INTP_LONG:
	    LUT_LOOP(long, unsigned char);
	    return flag;

	case INTP_FLOAT:
	    LUT_LOOP(float, unsigned char);
	    return flag;

	case INTP_DOUBLE:
	    LUT_LOOP(double, unsigned char);
	    return flag;

#ifdef HAVE_COMPLEX_H
	case INTP_FCOMPLEX:
	    LUT_LOOP(float complex, unsigned char);
	    return flag;

	case INTP_DCOMPLEX:
	    LUT_LOOP(double complex, unsigned char);
	    return flag;
#endif

	default:
	    NZReturnError(ERR_UNIMP, unimp);

	}

    /* If unsigned short index, switch on the table type */
    else if (xtype == INTP_USHORT)
	switch (type) {

	case INTP_UCHAR:
	    LUT_LOOP(unsigned char, unsigned short);
	    return flag;

	case INTP_SCHAR:
	    LUT_LOOP(signed char, unsigned short);
	    return flag;

	case INTP_USHORT:
	    LUT_LOOP(unsigned short, unsigned short);
	    return flag;

	case INTP_SHORT:
	    LUT_LOOP(short, unsigned short);
	    return flag;

	case INTP_UINT:
	    LUT_LOOP(unsigned int, unsigned short);
	    return flag;

	case INTP_INT:
	    LUT_LOOP(int, unsigned short);
	    return flag;

	case INTP_ULONG:
	    LUT_LOOP(unsigned long, unsigned short);
	    return flag;

	case INTP_LONG:
	    LUT_LOOP(long, unsigned short);
	    return flag;

	case INTP_FLOAT:
	    LUT_LOOP(float, unsigned short);
	    return flag;

	case INTP_DOUBLE:
	    LUT_LOOP(double, unsigned short);
	    return flag;

#ifdef HAVE_COMPLEX_H
	case INTP_FCOMPLEX:
	    LUT_LOOP(float complex, unsigned short);
	    return flag;

	case INTP_DCOMPLEX:
	    LUT_LOOP(double complex, unsigned short);
	    return flag;
#endif

	default:
	    NZReturnError(ERR_UNIMP, unimp);

	}

    /* All other index types are invalid */
    else
	NZReturnError(ERR_ILLOP, illtype);
}



/*--------------------------------------------------------------------------
bips_alpha - Generate integer and fractional parts of arithmetic sequence

This function generates an arithmetic sequence, separates each element
into its integer and fractional parts, and stores them in two arrays.
It is a useful operation for image resampling.  The integer part is
defined to be the floor, and the fractional part is defined as the
original number minus the floor.  The type of the fractional parts
stored must be either float or double; the type of the integer parts
stored is always long.  The base and increment of the sequence are
always double, since the added precision is almost always necessary
(and the cost is minor--if not negative--on most modern workstations).
The name of this function comes from the common use of the Greek
letter alpha to represent the fractional part that is computed.

FIXME: This doesn't seem to be the most useful form of this function; the 
affine resampling algorithms actually use something else.  Is there a real
use case for this function, or should it be deleted?  A possibly more
useful function is to convert an array of coordinates into their integer
and fractional parts, which is likely to be useful in the resampling
functions.

FIXME: The floor() function is remarkably slow for gcc/i386; could
something faster be used instead?  The trick of adding an offset
greater than the most negative possible value and then truncating
toward zero might be useful here.

FIXME: Write the more general (if less useful?) function to obtain the
integer and fractional parts of the elements of a vector.  Perhaps this
could be used in the resampling functions.
---------------------------------------------------------------------------*/

#define ALPHA_LOOP(type)                 \
{                                        \
    type * restrict df = (type *)dsti;   \
    for (n = 0; n < nelem; n++) {        \
        ival = floor(val);               \
        *di  = (long)ival;               \
        *df  = val - ival;               \
        val += incr;                     \
        df  += dsf;  di += dsi; }        \
}

int
bips_alpha(
    long   nelem,           /* Number of elements to store (in each dest) */
    long  *dsti,            /* Array of integer parts */
    long   dsi,             /* Stride in the integer array */
    int    type,            /* Type to store in the fraction array */
    void  *dstf,            /* Array of fractional parts */
    long   dsf,             /* Stride in the fraction array */
    double base,            /* First value in the arithmetic sequence */
    double incr)            /* Increment in the arithmetic sequence */
{
    double val = base, ival;   /* Value and its integer part */
    long * restrict di = dsti; /* Location in integer array */
    long n;                    /* Index over elements */

    /* Switch according to the destination type */
    switch (type) {

    case INTP_FLOAT:
	ALPHA_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	ALPHA_LOOP(double);
	return ERR_OKAY;

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*--------------------------------------------------------------------------
bips_lerp - Linear interpolation 

This function performs the linear interpolation

   dst[n] += alpha[n]*(src[n]-dst[n]),

including a optional cast from the original type of the source to the
type of the destination.  The destination is NOT initialized; this
turns out to be more convenient for multi-dimensional interpolation.
The destination and alpha arrays must be the same type, either float
or double; the source may be of any type.  The results are
unpredictable if any of the destination, alpha, and source arrays
overlap.

The types of the source and destination arrays must be specified
explicitly; the type of the alpha array is implicit in the destination
type, as shown below:

    dest type           alpha type
    ---------           ----------
    float               float
    double              double
    float complex       float
    double complex      double

--------------------------------------------------------------------------*/

#define LERP_INNER(dtype,atype)              \
{                                            \
    dtype * restrict d = (dtype *)dst;       \
    atype * restrict a = (atype *)alf;       \
    for (n = 0; n < nelem; n++) {            \
        d[0] += a[0]*((dtype)s[0] - d[0]);   \
        d += ds; a += as; s += ss;  }        \
}

#ifdef HAVE_COMPLEX_H
#define LERP_LOOP(stype)                     \
{                                            \
    stype * restrict s = (stype *)src;       \
    if (dtype == INTP_DOUBLE)                \
        LERP_INNER(double, double)           \
    else if (dtype == INTP_FLOAT)            \
        LERP_INNER(float, float)             \
    else if (dtype == INTP_DCOMPLEX)         \
        LERP_INNER(dcomplex, double)         \
    else if (dtype == INTP_FCOMPLEX)         \
        LERP_INNER(fcomplex, float)          \
    else                                     \
        NZReturnError(ERR_ILLOP, illtype);   \
}
#else
#define LERP_LOOP(stype)                     \
{                                            \
    stype * restrict s = (stype *)src;       \
    if (dtype == INTP_DOUBLE)                \
        LERP_INNER(double, double)           \
    else if (dtype == INTP_FLOAT)            \
        LERP_INNER(float, float)             \
    else                                     \
        NZReturnError(ERR_ILLOP, illtype);   \
}
#endif

int
bips_lerp(
    long nelem,            /* Number of elements to process */
    int dtype,             /* Type of the destination and alpha arrays */
    void *dst,             /* Destination array */
    long ds,               /* Destination stride */
    void *alf,             /* Alpha array */
    long as,               /* Alpha stride */
    int stype,             /* Source type */
    void *src,             /* Source array */
    long ss)               /* Source stride */
{
    int n;                 /* Index over elements */

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	LERP_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_USHORT:
	LERP_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_UINT:
	LERP_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_ULONG:
	LERP_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_SCHAR:
	LERP_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	LERP_LOOP(short);
	return ERR_OKAY;

    case INTP_INT:
	LERP_LOOP(int);
	return ERR_OKAY;

    case INTP_LONG:
	LERP_LOOP(long);
	return ERR_OKAY;

    case INTP_FLOAT:
	LERP_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	LERP_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	LERP_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	LERP_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}



/*--------------------------------------------------------------------------
bips_glerp - Gather values and do linear interpolation 

This function gathers neighboring voxels from an n-dimensional
array, casts to double, and performs the linear interpolation

   dst[n] = (1.0-alpha[n])*src[index[n]] + alpha[n]*src[index[n]+1] .

This admittedly odd computation turns out to be useful in a fast
algorithm for N-dimensional linear interpolation (and may in fact be
useful only there!).  The results are unpredictable if any of the
destination, alpha, and source arrays overlap.

The types of the source and destination arrays must be specified
explicitly; the type of the alpha array is implicit in the destination
type, as shown below:

    dest type           alpha type
    ---------           ----------
    float               float
    double              double
    float complex       float
    double complex      double

--------------------------------------------------------------------------*/

#define GLERP_INNER(dtype,atype)          \
{                                         \
    dtype * restrict d = (dtype *)dst;    \
    atype * restrict a = (atype *)alf;    \
    dtype s0, s1;                         \
    for (n = 0; n < nelem; n++) {         \
        si = s + index[n];                \
        s0 = (dtype)si[0];                \
        s1 = (dtype)si[1];                \
        d[0] = s0 + a[0] * (s1 - s0);     \
        d += ds; a += as;  }              \
}

#ifdef HAVE_COMPLEX_H
#define GLERP_LOOP(stype)                    \
{                                            \
    stype * restrict s = (stype *)src;       \
    stype * restrict si;                     \
    if (dtype == INTP_DOUBLE)                \
        GLERP_INNER(double, double)          \
    else if (dtype == INTP_FLOAT)            \
        GLERP_INNER(float, float)            \
    else if (dtype == INTP_DCOMPLEX)         \
        GLERP_INNER(double complex, double)  \
    else if (dtype == INTP_FCOMPLEX)         \
        GLERP_INNER(float complex, float)    \
    else                                     \
        NZReturnError(ERR_ILLOP, illtype);   \
}
#else
#define GLERP_LOOP(stype)                    \
{                                            \
    stype * restrict s = (stype *)src;       \
    stype * restrict si;                     \
    if (dtype == INTP_DOUBLE)                \
        GLERP_INNER(double, double)          \
    else if (dtype == INTP_FLOAT)            \
        GLERP_INNER(float, float)            \
    else                                     \
        NZReturnError(ERR_ILLOP, illtype);   \
}
#endif

int
bips_glerp(
    long nelem,            /* Number of elements to process */
    int dtype,             /* Type of the destination array */
    void *dst,             /* Destination array */
    long ds,               /* Destination stride */
    void *alf,             /* Alpha array */
    long as,               /* Alpha stride */
    int stype,             /* Source type */
    void *src,             /* Source array */
    long *index)           /* Element indices into the source array */
{
    long n;                /* Index over elements */

    /* Switch according to the source type */
    switch (stype) {

    case INTP_UCHAR:
	GLERP_LOOP(unsigned char);
	return ERR_OKAY;

    case INTP_USHORT:
	GLERP_LOOP(unsigned short);
	return ERR_OKAY;

    case INTP_UINT:
	GLERP_LOOP(unsigned int);
	return ERR_OKAY;

    case INTP_ULONG:
	GLERP_LOOP(unsigned long);
	return ERR_OKAY;

    case INTP_SCHAR:
	GLERP_LOOP(signed char);
	return ERR_OKAY;

    case INTP_SHORT:
	GLERP_LOOP(short);
	return ERR_OKAY;

    case INTP_INT:
	GLERP_LOOP(int);
	return ERR_OKAY;

    case INTP_LONG:
	GLERP_LOOP(long);
	return ERR_OKAY;

    case INTP_FLOAT:
	GLERP_LOOP(float);
	return ERR_OKAY;

    case INTP_DOUBLE:
	GLERP_LOOP(double);
	return ERR_OKAY;

#ifdef HAVE_COMPLEX_H
    case INTP_FCOMPLEX:
	GLERP_LOOP(float complex);
	return ERR_OKAY;

    case INTP_DCOMPLEX:
	GLERP_LOOP(double complex);
	return ERR_OKAY;
#endif

    default:
	/* Drop through to error return */ ;

    }

    NZReturnError(ERR_ILLOP, illtype);
}
