/****************************************************************************
parray.c - Pyvox array type for Python

Author:  Paul Hughett

Copyright (c) 1999-2005 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This file contains code to implement the Pyvox array type (often
abbreviated as parray) for the Pyvox module, which provides a Python
interface to the Voxel Kit.

Most functions in this file have the prefix "parray_" to indicate
that they define the Pyvox array type.


FIXME: Implement repr() and str() for Pyvox arrays.

FIXME: Consider creating versions of const, array, read, and ramp (and
add fill from list) that store data into existing Pyvox arrays without
changing their type or shape. Also implement the in-place operations
that have been added in Python 2.x.  Provide a means for creating a
Pyvox array without initializing its contents; this would be useful
for efficiently using the functions just aforesaid.  One possible
method would be to supply None as the data value, and warn the user
that if he tries to access the array without filling it, he might get
anything.  Even better, perhaps, is that pyvox.array() should create
but not allocate and initialize the array (thus varray->data==0
indicates an uninitialized array); then any array operation that fills
the entire array can just do so; but any array operation that affects
only individual pixels must set the entire array to zero first.
(array.eval() should do the fill.)  Actually, there are only a few
operations that are sensible for an uninitialized array; perhaps the
others should just generate an error.

FIXME:  The ordering of declarations and definitions in this file seems
to be rather delicate and in some places counter-intuitive.  Look for
a better way to do things.

FIXME: It would be really nice if parrays were sufficiently like sequences 
that you could simply iterate over their elements, rather than needing ugly 
artifices like

    for i in range(len(foo)) :
        do something with foo[i]

or

    for x in foo.list() :
        do something with x

Unfortunately, Python (1.5.2) doesn't seem to permit that; it insists on
using the mapping methods rather than the sequence methods when they exist
(with some awkward exceptions for which we have to provide sequence methods).

FIXME: PyObjects may also be pyvox_arrays or kernel_objects in
disguise; the current practice is to pass arguments as PyObjects and
cast as needed inside the function but this is sometimes confusing and
eliminates some error checking.  Is there a better way?
*****************************************************************************/

static const char rcsid[] = "$Id: parray.c,v 1.171 2006/02/24 17:18:51 hughett Exp $ Release $Name: pyvox-0-72 $";

#include "config.h"

#include "Python.h"    /* Must precede standard C headers */

#include <limits.h>
#include <stdio.h>

#include "exim.h"
#include "memm.h"
#include "voxel.h"
#include "bips.h"
#include "vxli.h"       /* For vxli_count.  FIXME: Use vxl_count instead? */
#include "pyvox.h"      /* Must follow voxel.h */

#define DEBUG_ALLOC  0     /* Trace array allocation/freeing? */


/* Forward declarations to facilitate logical organization of this file;
   the staticforward keyword is used to avoid certain compiler bugs--see
   the definition of staticforward in (python1.5) object.h for details. */
staticforward struct PyMethodDef  parray_methods[];



/*-----------------------------------------------------------------------
Common error strings

FIXME:  Gather the common error strings into this code block rather than
defining them individually for each function.  Better yet, create nice
Python exceptions for each common error.
------------------------------------------------------------------------*/

static const char cent_error[] =   /* Error: Invalid centroid argument */
	"Initial centroids must be a sequence of values in 0..255";
static const char clid_error[] =   /* Error: Invalid class ids argument */
	"Class ids must be a sequence of values in 0..255";
static const char match_error[] =  /* Error: Clids and centroids don't match */
	"Different numbers of class ids and centroids";
static const char uchar_error[] =  /* Error: Only uchar implemented */
	"Sorry, only unsigned char has been implemented yet.";
static const char voxel_error[] =  /* Error: not a Pyvox array */
	"Input is not a Pyvox array";



/*************************************************************************
             Utility Functions for Pyvox Arrays

FIXME: These should probably be renamed to the standard parray_ prefix;
the attempt to use Python style names seems to have been misguided.
(Misguidoed?)
*************************************************************************/



/*-----------------------------------------------------------------------
PyvoxArray_Dump - Dump description of a Pyvox array for debugging

This function dumps to stderr various useful information about a Pyvox
array and is intended to be used for debugging.
------------------------------------------------------------------------*/

void
PyvoxArray_Dump(
    PyObject *ob,              /* Purported Pyvox array */
    char *name,                /* Name to indicate which array, or zero */
    int die)                   /* Abort after dumping array? */
{
    voxel_array *src;
    
    if (name == 0)  name = "Nameless";

    fprintf(stderr, "Dumping Pyvox array '%s' at %p\n", name, ob);
    fprintf(stderr, "   ob_type = %p  [%p]\n", ob->ob_type, &parray_type); 
    src = ((pyvox_array *)ob)->varray;
    fprintf(stderr, "   varray = %p\n", src);
    if (src != 0) {
	fprintf(stderr, "   magic = %ld   [%d]\n", src->magic, VXL_MAGIC);
	fprintf(stderr, "   data = %p\n", src->data);
    }
    if (die != 0)
	abort();
	
    return;
}



/*------------------------------------------------------------------------
PyvoxArray_Check - Is a Python object a valid Pyvox array?

If the given object is a valid Pyvox array (varray), this function
returns 1 and stores a pointer to the voxel_array struct in the second
argument; otherwise it sets an appropriate error message and returns 0.

FIXME: Be more explicit about exactly what is wrong?
-------------------------------------------------------------------------*/

int                               /* Returns non-zero if valid */
PyvoxArray_Check(
    PyObject *ob,                 /* Purported Pyvox array */
    const char *name,             /* Name of argument it appears as */
    voxel_array **vob)            /* (Out) The voxel array it contains */
{
    if (ob != 0 && ob->ob_type == &parray_type) {
	if (vob != 0)
	    *vob = ((pyvox_array *)ob)->varray;
	return 1;
    }
    else {
        if (name)
            preset_error(ERR_INIT, "%s is not a valid Pyvox array", name);
        else
            preset_error(ERR_INIT, "Argument is not a valid Pyvox array");
        return 0;
    }
}



/*------------------------------------------------------------------------
PyvoxArray_CheckSrc - Is a Python object a valid source array?

If the given object is a valid Pyvox array which contains a voxel
array suitable for use as a source, this function returns 1 and stores
a pointer to the voxel_array struct in the second argument; otherwise
it sets an appropriate error message and returns 0.

FIXME: Be more explicit about exactly what is wrong?
-------------------------------------------------------------------------*/

int                               /* Returns non-zero if valid */
PyvoxArray_CheckSrc(
    PyObject *ob,                 /* Purported Pyvox array */
    const char *name,             /* Name of argument it appears as */
    voxel_array **vob)            /* (Out) The voxel array it contains */
{
    voxel_array *src;

    if  (  ob != 0 && 
           ob->ob_type == &parray_type &&
           (src = ((pyvox_array *)ob)->varray) != 0 &&
           src->magic == VXL_MAGIC &&
           src->data  != 0  )  {
	if (vob != 0)
	    *vob = src;
	return 1;
    }
    else {

	/* Generate error message for Python user */
        if (name)
            preset_error(ERR_INIT, "%s is not a valid Pyvox array", name);
        else
            preset_error(ERR_INIT, "Argument is not a valid Pyvox array");
        return 0;
    }
}



/*------------------------------------------------------------------------
PyvoxArray_CheckDest - Is a Python object is a valid destination array?

If the given object is a valid Pyvox array which contains a voxel
array suitable for use as a destination, this function returns 1 and stores
a pointer to the voxel_array struct in the second argument; otherwise
it sets an appropriate error message and returns 0.

FIXME: Be more explicit about exactly what is wrong?
-------------------------------------------------------------------------*/

int                               /* Returns non-zero if valid */
PyvoxArray_CheckDest(
    PyObject *ob,                 /* Purported Pyvox array */
    const char *name,             /* Name of argument it appears as */
    voxel_array **vob)            /* (Out) The voxel array it contains */
{
    voxel_array *dest;

    if  (  ob != 0 && 
           ob->ob_type == &parray_type &&
           (dest = ((pyvox_array *)ob)->varray) != 0 &&
           dest->magic == VXL_MAGIC &&
           dest->data  != 0  )  {
	if (vob != 0)
	    *vob = dest;
	return 1;
    }
    else {
        if (name)
            preset_error(ERR_INIT, "%s is not a valid Pyvox array", name);
        else
            preset_error(ERR_INIT, "Argument is not a valid Pyvox array");
        return 0;
    }
}



/*-------------------------------------------------------------------------
PyVoxel_GetArray - Verify and get voxel array from a Python object

This function verifies that a given PyObject is in fact a voxel array
object and returns a pointer to the voxel array it refers to.  If the
PyObject is not a voxel array object, zero is returned.

NOTE: This function was commented out on 2002-09-01.  New code should use
PyvoxArray_Check{,Src,Dest} instead.
---------------------------------------------------------------------------*/

#if 0
voxel_array *
PyVoxel_GetArray(
    PyObject *ob)           /* Alleged voxel array object */
{
    if (ob != 0 && ob->ob_type == &parray_type)
	return ((pyvox_array *)ob)->varray;
    else
	return 0;
}
#endif



/*-------------------------------------------------------------------------
PyEmptyObject_Check - Test if a Python object is an empty object

This function returns TRUE if the object is a list, tuple, or voxel
array containing zero elements.  The empty object is treated specially
in some contexts, often selecting a default value.

FIXME: This function is apparently useful but not actually used anywhere
yet.  It has been commented out to prevent a compiler warning.
--------------------------------------------------------------------------*/

#if 0
static int
PyEmptyObject_Check(
    PyObject *ob)
{
    if ( ob == 0 )
        return 0;
    else if ( PyTuple_Check(ob) && PyTuple_Size(ob) == 0 )
        return 1;
    else if ( PyList_Check(ob) && PyList_Size(ob) == 0 )
        return 1;
    else if ( PyvoxArray_Check(ob, 0, 0) && PyVoxel_Count(ob) == 0 )
        return 1;
    else 
        return 0;
}
#endif



/*-------------------------------------------------------------------------
PyScalar_Check - Test whether a Python object is a scalar

This function returns TRUE if the object is either a voxel array object
of rank zero (that is, a scalar) or a number (which can be converted into
a scalar).

FIXME: This function is apparently useful but not actually used anywhere
yet.  It has been commented out to prevent a compiler warning.
-------------------------------------------------------------------------*/

#if 0
static int
PyScalar_Check(
    PyObject *ob)
{
    voxel_array *var;

    if ( ob == 0 )
        return 0;
    else if ( ob->ob_type == &parray_type &&
              (var = ((pyvox_array *)ob)->varray) != 0 &&
              var->rank == 0 )
	return 1;
    else if ( PyNumber_Check(ob) )
	return 1;
    else
	return 0;
}
#endif



/*-------------------------------------------------------------------------
PyNonScalarArray_Check - Test if a Python object is a non-scalar voxel array

This function returns TRUE if the object is a voxel array object
of rank strictly greater than zero.

FIXME: This function is apparently useful but not actually used anywhere
yet.  It has been commented out to prevent a compiler warning.
-------------------------------------------------------------------------*/

#if 0
static int
PyNonScalarArray_Check(
    PyObject *ob)
{
    voxel_array *var;

    if ( ob != 0
	 && ob->ob_type == &parray_type
	 && (var = ((pyvox_array *)ob)->varray) != 0
	 && var->rank > 0 )
	return 1;
    else
	return 0;
}
#endif



/*--------------------------------------------------------------------------
PyVoxel_Type - Return internal type code of a voxel array

Returns zero if the Pyvox array is invalid or empty.
---------------------------------------------------------------------------*/

static int
PyVoxel_Type(
    PyObject *ob)
{
    voxel_array *var;

    if ( ob != 0
	 && ob->ob_type == &parray_type
	 && (var = ((pyvox_array *)ob)->varray) != 0 )
	return var->type;
    else
	return 0;
}



/*-------------------------------------------------------------------------
varray_init_scalar - Initialize scalar voxel array from a Python number

This function sets a voxel array to contain a scalar value of some
specified type converted from a Python number.

NOTE: This code relies on the fact that the Py_complex and C99 complex
types have the same layout in memory.
-------------------------------------------------------------------------*/

static int                    /* Returns non-zero on error */
varray_init_scalar(
    voxel_array *dest,        /* The voxel array to set */
    int type,                 /* Internal type to convert to */
    PyObject *number)         /* The number to convert */
{
    long  lvalue;
    double dvalue;
    int flag;

    if ( PyInt_Check(number) ) {
	flag = vxl_alloc_array(dest, type, 0, NULL);
	if (flag)    NZReturnLibError();
        lvalue = PyInt_AsLong(number);
	flag = bips_cast(1, type, dest->data, 0, INTP_LONG, &lvalue, 0);
	if (flag)    NZReturnLibError();
	return 0;
    }
    else if ( PyFloat_Check(number) ) {
	flag = vxl_alloc_array(dest, type, 0, NULL);
	if (flag)    NZReturnLibError();
        dvalue = PyFloat_AsDouble(number);
	flag = bips_cast(1, type, dest->data, 0, INTP_DOUBLE, &dvalue, 0);
	if (flag)    NZReturnLibError();
	return 0;
    }
#ifdef HAVE_COMPLEX_H
    else if (PyComplex_Check(number) ) {
        Py_complex cvalue;

	flag = vxl_alloc_array(dest, type, 0, NULL);
	if (flag)    NZReturnLibError();
	cvalue = PyComplex_AsCComplex(number);
        flag = bips_cast(1, type, dest->data, 0, INTP_DCOMPLEX, &cvalue, 0);
	if (flag)    NZReturnLibError();
	return 0;
    }
#endif
    else
	NZReturnError(ERR_ILLOP, 
		      "Cannot create scalar array from a non-number");
}



/*--------------------------------------------------------------------------
varray_create_scalar - Create new scalar voxel array from Python number

This function, given a Python number, creates a new scalar voxel array of
a specified type and initializes it from the Python number.

FIXME: Many of the functions that call this should probably allocate a
voxel_array struct on the stack and call varray_init_scalar instead.
---------------------------------------------------------------------------*/

static voxel_array *           /* Returns NULL on error */
varray_create_scalar(
    int type,                  /* Internal type of voxel array to create */
    PyObject *number)          /* Value to initialize from */
{
    voxel_array *dest;         /* Newly created voxel array */
    int flag;

    dest = vxl_create_array();
    if (dest == 0)    ZReturnLibError();
    flag = varray_init_scalar(dest, type, number);
    if (flag) {
      vxl_destroy_array(dest);
      ZReturnLibError();
    }
    return dest;
}



/*--------------------------------------------------------------------------
PyScalar_FromNumber - Create a scalar Pyvox array from a Python number

This function creates a scalar Pyvox array of some specified type from
a Python object of some numeric type; if the conversion is not possible,
it sets the error message and returns NULL.

FIXME: Rename as pyvox_scalar_from_number()?  Add to Reference Manual?

FIXME: Verify that this sets the reference count appropriately.
---------------------------------------------------------------------------*/

static PyObject *             /* Returns NULL on error */
PyScalar_FromNumber(
    PyObject *number,         /* The number to convert */
    int type)                 /* Internal type to convert to */
{
    pyvox_array *ob;          /* Scalar being created */
    voxel_array *vob;         /* Scalar as a voxel array */
    int flag;

    ob = parray_create(&vob);
    if (ob == 0)    ZReturnLibError();
    flag = varray_init_scalar(vob, type, number);
    if (flag)    ZReturnLibError();
    return (PyObject *)ob;
}



/*------------------------------------------------------------------------
Pyvox_CheckAxes - Interpret axes argument for reduction operator

This function checks that the axes argument for a Pyvox reduction
operator is in one of the acceptable forms and converts it to the
(naxes, axes) form expected by the Voxel Kit.  The acceptable forms
are an integer in the range [0, rank-1] or [-1, -rank], meaning the
single axis specified, counting either from the left or the right; a
tuple or list of such integers; or None, meaning all axes.  

This function returns non-zero if the argument is valid, and zero otherwise.

ALGORITHM: The interpretation of the axis arguments themselves is delegated
to the vxli_reduce_setup() function, which will be called by the reduction
function.  Some checks on the range of axis arguments is duplicated here
and in vxli_reduce_setup.
---------------------------------------------------------------------------*/

static 
int                             /* Returns non-zero for valid argument */
Pyvox_CheckAxes(
    PyObject *ob,               /* The purported Pyvox axes argument */
    int rank,                   /* The rank of the array being reduced */
    int *naxes,                 /* (Out) The number of axes to reduce on */
    int *axes)                  /* (Out) The axes to reduce on */
{
    int nax;                    /* Number of elements in a tuple or list */
    PyObject *ax;               /* An axis, as a Python object */
    int n;                      /* An axis, as a C number */
    int i;

    if ( ob == 0 || ob == Py_None ) {
	*naxes = -1;     /* Tells vxl_reduce_setup() to do all axes */
	return 1;
    }
    else if ( PyInt_Check(ob) ) {
	n = PyInt_AsLong(ob);
	if (n >= -rank && n < rank) {
	    *naxes = 1;
	    axes[0] = n;
	    return 1;
	}
	else
            ZReturnError(ERR_INDEX, "Axis is wrong type or out of range");


    }
    else if ( PyList_Check(ob) ) {
	*naxes = nax = PyList_Size(ob);
	for (i = 0; i < nax; i++) {
	    ax = PyList_GetItem(ob, i);
	    if ( !PyInt_Check(ax) )
		return 0;
	    n = PyInt_AsLong(ax);
	    if ( n >= -rank && n < rank)
		axes[i] = n;
	    else
                ZReturnError(ERR_INDEX, "Axis is wrong type or out of range");
	}
	return 1;
    }
    else if ( PyTuple_Check(ob) ) {
	*naxes = nax = PyTuple_Size(ob);
	for (i = 0; i < nax; i++) {
	    ax = PyTuple_GetItem(ob, i);
	    if ( !PyInt_Check(ax) )
		return 0;
	    n = PyInt_AsLong(ax);
	    if ( n >= -rank && n < rank)
		axes[i] = n;
	    else
                ZReturnError(ERR_INDEX, "Axis is wrong type or out of range");
	}
	return 1;
    }
    else
	ZReturnError(ERR_INDEX, "Axes argument is invalid type");
}



/*-----------------------------------------------------------------------
parray_get_voxel - Extract element of a voxel array as a Python number

This function gets some specified element of a voxel array and returns
it as a Python number.  The vxl_locate function checks that the array
is valid, so we don't repeat the check here.

See also pyvox_get_value() to fetch an element from a naked C array
or an anytype struct.
------------------------------------------------------------------------*/

PyObject *                       /* Returns NULL on error */
parray_get_voxel(
    voxel_array *src,            /* Voxel array to fetch from */
    long *coords)                /* Coordinates to fetch from */
{
    unsigned char *elem;         /* Pointer to desired element */
    int type;                    /* Type of the array */

    elem = vxl_locate(src, coords);
    if (elem == 0)    ZReturnLibError();
    type = src->type;
    return pyvox_get_value(type, elem, 0);
}



/*-----------------------------------------------------------------------
pyvox_from_temp_varray - Convert Voxel array to Python object

This function takes a temporary voxel array and converts it to a permanent
(reference-counted) Python object.  If the voxel array is a scalar (rank-0),
then the result is a Python number of the appropriate type, and the original
voxel array is destroyed.  Otherwise the result is a Pyvox array wrapped
around the voxel array.

This function is used to support the very common feature that a scalar
pyvox array decays into a simple number.

See also parray_create(), which creates a pyvox array and its voxel array
from scratch.

FIXME: Add to Reference Manual?  Rename for consistency?

FIXME: This is a rather weird function.  Is there a better way?
-------------------------------------------------------------------------*/

PyObject *                         /* Returns NULL on failure */
pyvox_from_temp_varray(
    voxel_array *src)              /* Voxel array to convert */
{
    PyObject *result;              /* The result */

    /* Validate the voxel array */
    if (src == 0 || src->magic != VXL_MAGIC || src->data == 0)
        ZReturnError(ERR_INIT, "Invalid or empty voxel array");

    /* Convert a rank-0 array to a Python number; destroy the voxel array */
    if (src->rank == 0) {
        result = pyvox_get_value(src->type, src->data, 0);
        vxl_destroy_array(src);
    }

    /* Convert any other array to a Pyvox array */
    else {
        result = (PyObject *)parray_wrap(src);
    }

    return result;
}



/************************************************************************
Attributes of the Pyvox Array Type

The functions must precede the definition of the parray_getattr and
parray_setattr functions.
*************************************************************************/



/*-----------------------------------------------------------------------
parray_header - Get header of a Pyvox array

This function returns the header attribute of Pyvox array.  This is a
dictionary in which the contents of the image file header can be
stored by either the user or the code which actually read the file.
The dict is not actually created until it is needed.
-------------------------------------------------------------------------*/

static PyObject *            /* Returns 0 on error */
parray_header(
    PyObject *self)          /* Pyvox array to get header of */
{
    PyObject *result = 0;
    pyvox_array *pself = (pyvox_array *)self;

    set_error_context("array.header");

    /* Create the header dict, if it does not already exist */
    if ( pself->header == 0 )
        pself->header = PyDict_New();

    /* Get the header and increment its ref count */
    result = pself->header;
    Py_INCREF(result);
    
    clear_error_context();
    return result;
}



/*-----------------------------------------------------------------------
parray_set_header - Set the header attribute of a Pyvox array

This function sets the header attribute of a Pyvox array, which must
be a dict.
-------------------------------------------------------------------------*/

static int                     /* Returns non-zero on error */
parray_set_header(
    PyObject *self,            /* Pyvox array to set header of */
    PyObject *value)           /* Value to set to */
{
    pyvox_array *pself = (pyvox_array *)self;

    set_error_context("array.header=");

    /* The value must be a dict */
    if ( ! PyDict_Check(value) )
        XReturnPyvoxError(ERR_ILLOP, "array.header must be a dict");

    /* Decrement ref count of old contents */
    Py_XDECREF(pself->header);

    /* Install new contents and increment ref count */
    Py_XINCREF(value);
    pself->header = value;

    clear_error_context();
    return 0;
}

	

/*-----------------------------------------------------------------------
parray_origin - Coordinate origin of a Pyvox array

This function returns the physical coordinate origin of a voxel array
object as a list of Python floats.

FIXME:  There is probably some code here that could be shared with
parray_spacing or other methods.
-----------------------------------------------------------------------*/

static PyObject *
parray_origin(
    PyObject *self)           /* Pyvox array to find the origin of */
{
    PyObject *origin;         /* List containing the dimensions */
    PyObject *dim;            /* One of the dimensions */
    voxel_array *varray;      /* The voxel array itself */
    int rank, i;

    set_error_context("array.origin");

    /* Find the voxel array itself */
    if ( !PyvoxArray_Check(self, "self",  &varray) )
        ReturnPyvoxLibError();
    rank = varray->rank;

    /* Construct a list of the origin values */
    /* FIXME:  I _think_ the following code sets the right reference
       counts on the dimensions and the list containing them, but
       the Python API manual is somewhat obscure on this point. */
    origin = PyList_New(rank);
    for (i = 0; i < rank; i++) {
        dim = PyFloat_FromDouble(varray->origin[i]);
	/* FIXME:  The Python C API manual does not specify the value
	   returned by PyList_SetItem; I am assuming that it returns
	   zero on success, and non-zero otherwise. */
	if ( PyList_SetItem(origin, i, dim) )
            ReturnPyvoxError(ERR_ASSERT, "PyList_SetItem failed");

    }

    clear_error_context();
    return origin;
}



/*----------------------------------------------------------------------
parray_set_origin - Set origin of a Pyvox array

This function sets the origin attribute of a voxel array.
-----------------------------------------------------------------------*/

static int                    /* Returns non-zero on error */
parray_set_origin(
    PyObject *self,           /* Object to set origin of */
    PyObject *value)          /* Value to set */
{
    voxel_array *vself;       /* The voxel array itself */
    int rank;                 /* Rank of the voxel array */

    set_error_context("array.origin=");

    /* Find the voxel array itself */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        XReturnPyvoxLibError();
    rank  = vself->rank;

    /* Copy new origin values into array attribute */
    if ( ! PyvoxDoubleSeq_Check(value, "rhs", rank, rank, 0, vself->origin) )
        XReturnPyvoxLibError();

    clear_error_context();
    return 0;
}



/*----------------------------------------------------------------------
parray_set_spacing - Set spacing of a Pyvox array

This function sets the spacing attribute of a voxel array.
-----------------------------------------------------------------------*/

static 
int                           /* Returns non-zero on error */
parray_set_spacing(
    PyObject *self,           /* Object to set spacing of */
    PyObject *value)          /* Value to set */
{
    voxel_array *vself;       /* The voxel array itself */
    int rank;                 /* Rank of the voxel array */

    set_error_context("array.spacing=");

    /* Find the voxel array itself */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
	XReturnPyvoxLibError();
    rank = vself->rank;

    /* Copy new spacing values into array attribute */
    if ( ! PyvoxDoubleSeq_Check(value, "rhs", rank, rank, 0, vself->spacing) )
        XReturnPyvoxLibError();

    clear_error_context();
    return 0;
}

	

/*-----------------------------------------------------------------------
parray_size - Dimensions of a Pyvox array

This function returns the dimensions of a Pyvox array as a list of
integers and is used to implement the instance.size member of a voxel
object.  The rank is obviously the length of this list.
-----------------------------------------------------------------------*/

static PyObject *
parray_size(
    PyObject *self)           /* Pyvox array to find the size of */
{
    PyObject *size;           /* List containing the dimensions */
    PyObject *dim;            /* One of the dimensions */
    voxel_array *varray;      /* The voxel array itself */
    int rank, i;

    set_error_context("array.size");

    /* Find the voxel array itself */
    if ( !PyvoxArray_Check(self, "self", &varray) )
	ReturnPyvoxLibError();
    rank = varray->rank;

    /* Construct a list of the dimensions */
    /* FIXME:  I _think_ the following code sets the right reference
       counts on the dimensions and the list containing them, but
       the Python API manual is somewhat obscure on this point. */
    size = PyList_New(rank);
    for (i = 0; i < rank; i++) {
        dim = PyInt_FromLong(varray->dimen[i]);
	/* FIXME:  The Python C API manual does not specify the value
	   returned by PyList_SetItem; I am assuming that it returns
	   zero on success, and non-zero otherwise. */
	if ( PyList_SetItem(size, i, dim) )
            ReturnPyvoxError(ERR_ASSERT, "PyList_SetItem failed");
    }

    clear_error_context();
    return size;
}

	

/*-----------------------------------------------------------------------
parray_spacing - Coordinate spacings of a Pyvox array

This function returns the physical coordinate spacing of a voxel array
object as a list of Python floats.

FIXME:  There is probably some code here that could be shared with
parray_origin or other methods.
-----------------------------------------------------------------------*/

static PyObject *
parray_spacing(
    PyObject *self)           /* Pyvox array to find the spacing of */
{
    PyObject *spacing;        /* List containing the dimensions */
    PyObject *dim;            /* One of the dimensions */
    voxel_array *varray;      /* The voxel array itself */
    int rank, i;

    set_error_context("array.spacing");

    /* Find the voxel array itself */
    if ( !PyvoxArray_Check(self, "self", &varray) )
	ReturnPyvoxLibError();
    rank = varray->rank;

    /* Construct a list of the spacing values */
    /* FIXME:  I _think_ the following code sets the right reference
       counts on the dimensions and the list containing them, but
       the Python API manual is somewhat obscure on this point. */
    spacing = PyList_New(rank);
    for (i = 0; i < rank; i++) {
        dim = PyFloat_FromDouble(varray->spacing[i]);
	/* FIXME:  The Python C API manual does not specify the value
	   returned by PyList_SetItem; I am assuming that it returns
	   zero on success, and non-zero otherwise. */
	if ( PyList_SetItem(spacing, i, dim) )
            ReturnPyvoxError(ERR_ASSERT, "PyList_SetItem failed");
    }

    clear_error_context();
    return spacing;
}



/**************************************************************************
Standard methods for the Pyvox array type

The following functions implement core methods for the Pyvox array
type and are listed in the type object.  There are also a few closely
related functions that are not actually listed in the type object.
**************************************************************************/



/*-----------------------------------------------------------------------
parray_wrap - Wrap a Pyvox array around an existing voxel array

This function creates a brand-new instance of a Pyvox array and
initializes it to contain a given voxel array.  The reference count of
the created Pyvox array is set to 1.  Note that if Py_BuildValue is
used to created a result to be returned, the "N" code should be used
rather than "O", which increments the reference count.

The header, metadata, and userdata attributes are initialized to zero,
but the dicts will be created automatically if the user ever attempts
to access these attributes.  This is a performance hack so as to not
create dicts that are never actually used.
-------------------------------------------------------------------------*/

pyvox_array *                  /* Returns NULL on error */
parray_wrap(
    voxel_array *varray)       /* Voxel array to be wrapped */
{
    pyvox_array *self;         /* The new Pyvox array */

    /* Create a new Pyvox array */
    self = (pyvox_array *)PyObject_NEW(pyvox_array, &parray_type);
    if (self == 0)
        ZReturnError(ERR_MALLOC, "Creation of Pyvox array failed");

    /* Initialize the Pyvox array to contain the voxel array */
    self->varray   = varray;
    self->header   = 0;
    self->metadata = 0;
    self->userdata = 0;

    /* Report creation of a new Pyvox array */
    if (DEBUG_ALLOC) {
	fprintf(stderr, "parray create:  %p\n", self);
	if (0) 
	    fprintf(stderr, "...reference count = %d\n",
		    ((PyObject *)self)->ob_refcnt); 
    }

    return self;
}



/*-----------------------------------------------------------------------
parray_create - Create an instance of a Pyvox array

This function creates a brand-new owned instance of a Pyvox array and
initializes it to contain a voxel array of rank 0, containing no data.
This function is preferred over parray_alloc when the created array
will promptly be used as a destination by the Voxel Kit, which will do
its own allocation.  The reference count of the created array is set
to 1.  Note that if Py_BuildValue is used to created a result to be
returned, the "N" code should be used rather than "O", which
increments the reference count.

This function returns a pointer to the newly created Pyvox array and,
via a optionally non-zero pointer argument, a pointer to the voxel
array that it contains.  If there is an error, the error code and
message are set, and NULL is returned.  The returned Pyvox array is an
owned reference, and its reference count must be decremented if it is
discarded later.

The header, metadata, and userdata attributes are initialized to zero,
but the dicts will be created automatically if the user ever attempts
to access these attributes.  This is a performance hack so as to not
create dicts that are never actually used.

This function could plausibly return its result as either a PyObject *
or a pyvox_array * pointer.  On the whole, the latter seems to give
the cleanest code in the callers, and a little extra error checking.

Prior to 2002-08-29, the argument varray for the included voxel_array
did not exist.
------------------------------------------------------------------------*/

pyvox_array *                 /* Returns pyvox array, or  NULL on error */
parray_create(
    voxel_array **varray)     /* (Out) The voxel array inside; or zero */
{
    pyvox_array *self;        /* Pyvox array being created */
    voxel_array *vself;       /* The voxel array it contains */

    /* Create a voxel array */
    vself = vxl_create_array();
    if (vself == 0)    ZReturnLibError();    

    /* Wrap it in a Pyvox array */
    self = parray_wrap(vself);

    /* Tell the caller where the voxel array is */
    if (varray != 0)
        *varray = vself;

    /* ...and return the Pyvox array itself */
    return self;
}



/*-----------------------------------------------------------------------
parray_destroy - Destroy an instance of a Pyvox array

This function is called automatically by the Python garbage collector
when the reference count of a Pyvox array object goes to zero and
destroys that instance of a Pyvox array by freeing the storage
associated with both the voxel array itself and the Python object
wrapper that refers to it; the header, metadata, and userdata dicts
have their ref counts decremented but are not actually destroyed, in
case the user is still holding a reference.

FIXME: Does this require set_error_context()?  The situation is a bit
odd because destruction happens implicitly, when the refcount goes to
zero rather than as the result of an explicit call.
-----------------------------------------------------------------------*/

static inline
void
parray_destroy(
    pyvox_array *self)
{
    voxel_array *varray;          /* The voxel array in the object */

    if (DEBUG_ALLOC)
        fprintf(stderr, "parray destroy: %p\n", (void *)self);

    /* Destroy the voxel array */
    varray = self->varray;
    self->varray = 0;
    vxl_destroy_array(varray);

    /* Decrement the ref count of the userdata dict, if any */
    Py_XDECREF(self->userdata);
    self->userdata = 0;

    /* Decrement the ref count of the metadata dict, if any */
    Py_XDECREF(self->metadata);
    self->metadata = 0;

    /* Decrement the ref count of the header dict, if any */
    Py_XDECREF(self->header);
    self->header = 0;

    /* Delete the Pyvox array object itself */
    PyMem_DEL(self);

    return;
}



/*-----------------------------------------------------------------------
parray_alloc - Create and allocate an instance of a voxel array object

This function creates a brand-new instance of a Pyvox array,
initializes it to a given type, rank, and dimensions, and allocates
storage for the contents.  The contents of the array are left
undefined, on the perhaps dubious assumption that the user will
immediately fill them in.  This function should be used only when it
is necessary to actually allocate and initialize the voxel array; if
the new array is used as a destination for a Voxel Kit function which
allocates the array (which is the usual case), then parray_create
should be used instead to avoid a redundant allocation.  The reference
count of the created array is set to 1.  Note that if Py_BuildValue
is used to created a result to be returned, the "N" code should be
used rather than "O", which increments the reference count.

The header, metadata, and userdata attributes are initialized to zero,
but the dicts will be created automatically if the user ever attempts
to access those attributes.  This is a performance hack so as to not
create userdata dicts that are never actually used.

This function returns a pointer to the newly created and allocated
Pyvox array and, via a optionally non-zero pointer argument, a pointer
to the voxel array that it contains.  If there is an error, the error
code and message are set, and NULL is returned.

This function could plausibly return its result as either a PyObject *
or a pyvox_array * pointer.  On the whole, the latter seems to give
the cleanest code in the callers, and a little extra error checking.

Prior to 2002-08-29, the argument varray for the included voxel_array
did not exist.
------------------------------------------------------------------------*/

pyvox_array *                /* Returns new pyvox array, or NULL on error */
parray_alloc(
    voxel_array **varray,    /* (Out) New voxel array within result */
    int type,                /* Type of voxel array to create */
    int rank,                /* Rank of voxel array to create */
    long *dimen)             /* Dimensions of voxel array */
{
    pyvox_array *self;       /* Pyvox array being created */
    voxel_array *vself;      /* The voxel array that it contains */
    int flag;

    /* Create and allocate a voxel array */
    vself = vxl_create_array();
    if (vself == 0)    ZReturnLibError();
    flag = vxl_alloc_array(vself, type, rank, dimen);
    if (flag)    ZReturnLibError();

    /* Create a Pyvox array that contains it */
    self = (pyvox_array *)PyObject_NEW(pyvox_array, &parray_type);
    if (self == 0)
        ZReturnError(ERR_MALLOC, "Creation of Pyvox array failed");
    self->varray   = vself;
    self->header   = 0;
    self->metadata = 0;
    self->userdata = 0;

    /* Report creation of a new Pyvox array */
    if (DEBUG_ALLOC) {
	fprintf(stderr, "parray alloc:   %p\n", self);
	if (0) 
	    fprintf(stderr, "...reference count = %d\n",
		    ((PyObject *)self)->ob_refcnt); 
    }

    /* Return results */
    if (varray != 0)
        *varray = vself;
    return self;
}



/*------------------------------------------------------------------------
parray_copy - Create a copy of a Pyvox array

This Pyvox array method creates a new Pyvox array with the same
contents and attributes of an existing array, but which shares no
storage with it.  Shallow copies of the header, metadata, and userdata
dicts are made, so there may actually be some sharing of their
contents.

FIXME: Should I make shallow or deep copies of the dicts?  Or not
copy them at all?
-------------------------------------------------------------------------*/

static 
PyObject *                          /* Returns NULL on error */
parray_copy(
    PyObject *self, 
    PyObject *args)
{
    pyvox_array *in,  *out;         /* Input and output as Pyvox arrays */
    voxel_array *vin, *vout;        /* Input and output as voxel arrays */
    PyObject    *din, *dout;        /* Dictionaries in input and output */

    set_error_context("array.copy()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;


    /* Create a new Pyvox array to contain it */
    out = (pyvox_array *)PyObject_NEW(pyvox_array, &parray_type);
    if (out == 0)
        ReturnPyvoxError(ERR_MALLOC, "Creation of Pyvox array failed");

    /* Find the input voxel array itself and copy into the new Pyvox array */
    if ( !PyvoxArray_Check(self, "self", 0) )
	ReturnPyvoxLibError();
    in  = (pyvox_array *)self;
    vin = in->varray;
    vout = vxl_copy_array(vin);
    if (vout == 0)    ReturnPyvoxLibError();
    out->varray   = vout;

    /* Shallow copy the header dict, if any */
    din  = in->header;
    dout = 0;
    if ( din != 0 ) {
        dout = PyObject_CallMethod(din, "copy", "");
        if (dout == 0)
            ReturnPyvoxError(ERR_MISC, "Failed to copy header dictionary");
    }
    out->header = dout;

    /* Shallow copy the metadata dict, if any */
    din  = in->metadata;
    dout = 0;
    if ( din != 0 ) {
        dout = PyObject_CallMethod(din, "copy", "");
        if (dout == 0)
            ReturnPyvoxError(ERR_MISC, "Failed to copy metadata dictionary");
    }
    out->metadata = dout;

    /* Shallow copy the userinto dict, if any */
    din  = in->userdata;
    dout = 0;
    if ( din != 0 ) {
        dout = PyObject_CallMethod(din, "copy", "");
        if (dout == 0)
            ReturnPyvoxError(ERR_MISC, "Failed to copy userdata dictionary");
    }
    out->userdata = dout;

    /* Return the new Pyvox array */
    clear_error_context();
    return (PyObject *)out;
}



/*------------------------------------------------------------------------
parray_print - Print (informal) representation of a Pyvox array

This function prints out an informal representation of a Pyvox array
on a specified FILE.  This representation contains enough information
to distinguish different objects, but not to reconstruct the object
when read back in.  (It was thought undesireable to print out a 10 MB
object in full, without at least warning the poor unsuspecting user).

FIXME: The format could use a lot of improvement yet.

FIXME: Most of the array attributes, including the header, metadata,
and userdata dicts, are not shown.

FIXME: Figure out a good way to nicely format a 2-D array.

FIXME: Provide a means for printing out the whole array, for those who
really, really want it.

FIXME: The flags argument should be either 0 or Py_PRINT_RAW to select 
either the repr() or str() representations.
-------------------------------------------------------------------------*/

static 
int                              /* Returns non-zero on error */
parray_print(
    PyObject *self,              /* The object being printed */
    FILE *out,                   /* Where to print it */
    int flags)                   /* Provide repr() or str()? */
{
    PyObject *elem;              /* An element from the array */
    voxel_array *vself;          /* Voxel array struct */
    int rank;                    /* Rank of the voxel array */
    long *dimen;                 /* Dimensions of the array */
    long coord[VXL_MAX_RANK];    /* Index into the array */
    long count;                  /* Number of elements in array */
    const char *name;            /* Short name of type */
    int i, j, flag;

    set_error_context("print array");

    /* DEBUG: Report contents of Pyvox array */
    if (0) PyvoxArray_Dump(self, "print source", 0);

    /* Get characteristics of the array */
    if ( ! PyvoxArray_CheckSrc(self, "self", &vself) )
        XReturnPyvoxLibError();
    rank  = vself->rank;
    dimen = vself->dimen;
    name  = exim_nameof_type(vself->type);
    count = vxli_count(vself);

    /* Print summary characteristics */
    fprintf(out, "<Pyvox voxel array: %s (", name);
    for (i = 0; i < rank-1; i++)
	fprintf(out, "%ld,", vself->dimen[i]);
    if (rank > 0)
	fprintf(out, "%ld", vself->dimen[rank-1]);
    fprintf(out, ") at %p>", self);

    /* If short, print all elements */
    coord[0] = coord[1] = 0;
    fprintf(out, "\n");
    if (rank == 0) {
        elem = parray_get_voxel(vself, coord);
        if (elem == 0)    XReturnPyvoxLibError();
        flag = PyObject_Print(elem, out, flags);
        if (flag)
            XReturnPyvoxError(ERR_MISC, "Error in PyObject_Print");
    }
    else if(rank == 1 && count <= 100) {
	for (i = 0; i < dimen[0]; i++) {
	    coord[0] = i;
            elem = parray_get_voxel(vself, coord);
            if (elem == 0)    XReturnPyvoxLibError();
            flag = PyObject_Print(elem, out, flags);
            if (flag)
                XReturnPyvoxError(ERR_MISC, "Error in PyObject_Print");
            fprintf(out, "  ");
	}
    }
    else if (rank == 2 && count <= 100) {
	for (i = 0; i < dimen[0]; i++) {
	    coord[0] = i;
	    for (j = 0; j < dimen[1]; j++) {
		coord[1] = j;
                elem = parray_get_voxel(vself, coord);
                if (elem == 0)    XReturnPyvoxLibError();
                flag = PyObject_Print(elem, out, flags);
                if (flag)
                    XReturnPyvoxError(ERR_MISC, "Error in PyObject_Print");
                fprintf(out, "  ");
	    }
	    fprintf(out, "\n");
	}
    }
    else {
	fprintf(out, 
		"  [%ld elements not printed to save electrons]",
		count);
    }
    fprintf(out, "\n");

    clear_error_context();
    return 0;
}



/*----------------------------------------------------------------------
parray_getattr - Get an attribute of a Pyvox array

This function gets the value of a given attribute of a Pyvox array.

ALGORITHM: This function is just a dispatcher and assumes that
set_error_context() function will be called in the subsidiary
functions.  There are a few cases where the necessary code is embedded
here rather than called; we hope that they will never generate errors.

FIXME: The methods used below to implement the members of the
instance, and the list of members is not easily expandable to the
number of members that we are likely to get eventually.  Convert to a
table-driven approach, perhaps with hashing.
----------------------------------------------------------------------*/

static 
PyObject *                   /* Returns NULL on error */
parray_getattr(
    PyObject *self,          /* Pyvox array to get attribute of */
    char *attr)              /* Name of the attribute to get */
{
    voxel_array *varray;     /* The voxel array itself */

    /* Check for some errors */
    if ( !PyvoxArray_Check(self, "self", &varray) )
	ReturnPyvoxLibError();

    /* Return value of any requested member of the instance */
    if (strcmp(attr, "header") == 0)
        return parray_header(self);
    else if (strcmp(attr, "metadata") == 0) {
        pyvox_array *pself = (pyvox_array *)self;
        if (pself->metadata == 0)
            pself->metadata = PyDict_New();
        Py_INCREF(pself->metadata);
        return pself->metadata;
    }
    else if (strcmp(attr, "origin") == 0)
	return parray_origin(self);
    else if (strcmp(attr, "rank") == 0)
	return Py_BuildValue("i", varray->rank);
    else if (strcmp(attr, "spacing") == 0)
	return parray_spacing(self);
    else if (strcmp(attr, "size") == 0)
	return parray_size(self);
    else if (strcmp(attr, "type") == 0)
        return PyvoxType_FromInt(varray->type);
    else if (strcmp(attr, "userdata") == 0) {
        pyvox_array *pself = (pyvox_array *)self;
        if (pself->userdata == 0)
            pself->userdata = PyDict_New();
        Py_INCREF(pself->userdata);
        return pself->userdata;
    }

    /* Provide list of members available from this instance */
    if (strcmp(attr, "__members__") == 0)
	return Py_BuildValue("[ssssssss]", "header", "metadata", "origin",  
                             "rank", "spacing", "size", "type", "userdata");

    /* Use the Python API to find methods associated with this type */
    return Py_FindMethod(parray_methods, self, attr);
}



/*----------------------------------------------------------------------
parray_setattr - Set an attribute of a Pyvox array

This method sets a given attribute of a Pyvox array.

ALGORITHM: This function is just a dispatcher and assumes that
set_error_context() function will be called in the subsidiary
functions.  There are a few cases where the necessary code is embedded
here rather than called; we hope that they will never generate errors.

FIXME: The methods used below to implement the members of the
instance, and the list of members is not easily expandable to the
number of members that we are likely to get eventually.  Convert to a
table-driven approach, perhaps with hashing.

FIXME:  Does this handle inherited attributes correctly?

FIXME: Provide the capability to replace the dict of the metadata and
userdata dictionaries?
----------------------------------------------------------------------*/

static 
int                          /* Returns non-zero on error */
parray_setattr(
    PyObject *self,          /* Pyvox array to set attribute of */
    char *attr,              /* Name of the attribute to set */
    PyObject *value)         /* Value to set attribute to */
{
    voxel_array *varray;     /* The voxel array itself */

    /* Check for some errors */
    if ( !PyvoxArray_Check(self, "self", &varray) )
	XReturnPyvoxLibError();

    /* Set value of any requested member of the instance */
    if (strcmp(attr, "header") == 0)
        return parray_set_header(self, value);
    else if (strcmp(attr, "origin") == 0)
	return parray_set_origin(self, value);
    else if (strcmp(attr, "spacing") == 0)
	return parray_set_spacing(self, value);
    else
	XReturnPyvoxError(ERR_MISC, "Attribute not settable");
}



/***********************************************************************
Standard Mapping Methods for the Pyvox Array Type
***********************************************************************/



/*-----------------------------------------------------------------------
parray_convert_slice - Convert a Python slice object to C values

This function is a wholly owned subsidiary of vxl_get_slice and
vxl_set_slice; it takes a Python slice object and converts to count,
initial, and step values as C long values; for compatibility with the
conventions used for nidx in vxl_get_slice and vxl_set_slice, the
negative of the count is returned, unless it's one.  The step defaults
to 1.  If the step is positive, then the start and stop values default
to 0 and deflimit; if the step is negative, the start and stop default
to -1 and -deflimit-1; it is an error if the step is zero.  If the
initial or limit value is negative, then deflimit is added to it.
The resulting initial and limit values are then bounded by 0 and
deflimit-1, so that non-existing elements are simply omitted.
This function returns 0 for success and non-zero for error.  It is an
error if any component of the slice is neither None nor a number.

NOTE: The strange default values when the step is negative are
carefully chosen to yield the sequence deflimit-1, deflimit-2, ..., 0
when the step is -1 and after the negative index rule is applied.
------------------------------------------------------------------------*/

static inline 
int                              /* Returns non-zero on error */
parray_convert_slice(
    PyObject *ob,                /* Python slice object to convert */
    long deflimit,               /* Default value for limit */
    long *nelem,                 /* (out) minus number of elements */
    long *init,                  /* (out) initial value from slice ob */
    long *step)                  /* (out) step value from slice ob */
{
    PyObject *val;
    long vinit  = 0;             /* Initial value */
    long vlimit = deflimit;      /* Limit value */
    long vstep  = 1;             /* Step value */
    long nstep;                  /* Number of steps */

    /* Assume initially that nstep is zero, or user will ignore error flag */
    *nelem = *init = *step = 0;

    /* Convert Python step value; set defaults for start and limit */
    val = PyObject_GetAttrString(ob, "step");
    if ( val == Py_None )
	vstep = 1;
    else if ( PyInt_Check(val) )
	vstep = PyInt_AsLong(val);
    else
        NZReturnError(ERR_TYPE, "Slice step must be integer or None");
    if (vstep < 0) {
        vinit  = -1;
        vlimit = -deflimit-1;
    }

    /* Convert Python start value */
    val = PyObject_GetAttrString(ob, "start");
    if ( PyInt_Check(val) )
	vinit = PyInt_AsLong(val);
    else if ( val != Py_None )
        NZReturnError(ERR_TYPE, "Slice initial value must be integer or None");

    if ( vinit <= -deflimit )
	vinit = 0;
    else if ( vinit < 0 )
	vinit += deflimit;
    else if ( vinit > deflimit - 1 )
	vinit = deflimit - 1;

    /* Convert Python limit value */
    val = PyObject_GetAttrString(ob, "stop");
    if ( PyInt_Check(val) ) {
	vlimit = PyInt_AsLong(val);
    }
    else if ( val != Py_None )
        NZReturnError(ERR_TYPE, "Slice final value must be integer or None");

    if ( vlimit <= -deflimit-1 )
	vlimit = -1;
    else if ( vlimit < 0 )
	vlimit += deflimit;
    else if ( vlimit > deflimit )
	vlimit = deflimit;

    /* Compute number of elements in sequence */
    /* (C division is portable only for positive operands) */
    if (vstep > 0 && vlimit > vinit) 
        nstep = (vlimit - vinit + vstep - 1) / vstep;
    else if (vstep < 0 && vlimit < vinit) {
        nstep = (-vlimit + vinit - vstep - 1) / (-vstep);
    }
    else if (vstep == 0)
        NZReturnError(ERR_INIT, "Step size in slice may not be zero");
    else
        NZReturnError(ERR_INIT, "Requested slice contains no elements");

    /* Store converted values and return */
    if (nstep <= 0)
        return 1;
    else if (nstep == 1)
        *nelem = 1;
    else
        *nelem = -nstep;
    *init  = vinit;
    *step  = vstep;
    return 0;
}



/*------------------------------------------------------------------------
parray_prep_slice - Convert slice description from Python to Voxel Kit

This function takes a slice description in the style used within
Python (or Pyvox) and converts it to the style used within the Voxel
Kit.  The subscr argument may be a naked number, which is interpreted
as a (single) index value for a rank-1 array; a naked slice object,
which applies to the only axis of a rank-1 array; a naked ellipsis,
which is interpreted to mean that the desired slice is (a copy of) the
entire array; or a list or tuple of subscripts.  Each subscript may be
a number; a slice object; a tuple or list of numbers; or an ellipsis.
As a special case, a rank-0 (scalar) array will accept (and ignore)
any number of subscripts, returning its single element as a number;
this is mostly to work around the fact that Python syntax does not
allow you to provide an empty list of subscripts to an array.
Attempting to access a non-existent voxel raises an error.

There are three possible values for the returned value.  If the
conversion is unsuccessful, the Python error message is set and zero
is returned.  If the pre-allocated space provided as the temp argument
is sufficiently large or not used at all, its value is returned and
must not be freed.  If the pre-allocated space is inadequate, or temp
was zero, then a pointer to space allocated by this function is
returned, and must be freed by the caller.

NOTE: The usual assignment "bar = foo" in Python does a shallow copy,
thus causing bar and foo to refer to the same array.  To copy the
contents (a deep copy), you must use the form "bar = foo[...]".

FIXME: Extend this function to permit things like foo[:] to return all
the elements as a vector; and NewAxis as used in NumPy.

ALGORITHM: The possibility of ellipses means that the number of items
in the subsrc argument and the number of dimensions in the self array
may not be equal.  To keep them straight, we refer to the first as
subscripts and the second as coordinates.  The actual values to be used
as coordinates are referred to as indices.

The subscript list is scanned twice, once to determine how much temporary
storage is needed and once to actually process it.  This appears to yield
the simplest code and should not be unreasonably inefficient compared to
the time taken to actually extract the desired slice.

FIXME: Enumerate all the cases supposedly handled here and add test
cases to test them.
------------------------------------------------------------------------*/

#define PATEMP 64      /* Size of pre-allocated space */

#if (PATEMP < VXL_MAX_RANK) || (PATEMP < 3)
#error "Defined value of PATEMP violates its constraints"
#endif

static 
long *                          /* Temp space to be freed after use, or NULL */
parray_prep_slice(
    voxel_array *sarray,        /* Voxel array to slice from */
    PyObject *subscr,           /* Subscripts for desired voxel or slice */
    long *nidx,                 /* (out) How to handle each axis */
    long **index,               /* (out) Index values along each axis */
    long *temp)                 /* Preallocated temp space, length PATEMP */
{
    int rank;                   /* Rank of the voxel array */
    PyObject *sub;              /* Current subscript */
    int nsub, isub;             /* Number of items in subscr, and index */
    int icor;                   /* Index over coordinates in voxel array */
    int nell;                   /* Number of ellipses seen */
    long ntemp = 0;             /* Temp space needed for coords and slices */
    long *p, nelem, s;
    int i, flag;

    /* DEBUG: Show input subscript */
    if (0)
        PyObject_Print(subscr, stderr, 0);

    /* Handle rank 0 as a special case */
    rank = sarray->rank;
    if (rank == 0) {
        return temp; }

    /* Handle naked ellipsis as a special case */
    if ( subscr == Py_Ellipsis ) {
	for (i = 0; i < rank; i++)
	    nidx[i] = 0;
	return temp;
    }

    /* Handle naked number for rank-1 array as a special case */
    if ( rank == 1 && PyvoxLong_Check(subscr, &temp[0]) ) {
	nidx[0]  = 1;
	index[0] = &temp[0];
	if ( temp[0] < 0 )
	    temp[0] += sarray->dimen[0];
	return temp;
    }

    /* Handle naked slice object for rank-1 array as special case */
    if ( rank == 1 && PySlice_Check(subscr) ) {
	flag = parray_convert_slice(subscr, sarray->dimen[0],
                                    &nidx[0], &temp[0], &temp[1]);
        if (flag)    ZReturnLibError();
	index[0] = &temp[0];
	return temp;
    }

    /* Otherwise the subscr argument must be a sequence object */
    if ( !PySequence_Check(subscr) )
	ZReturnError(ERR_INDEX, "Invalid subscripts to an array");

    /* Prescan the subscripts to determine the total number of
       ellipsis and non-ellipsis subscripts, and the space that must
       be allocated for indices. */
    nsub  = PyObject_Length(subscr); 
    nell  = 0;
    ntemp = 0;
    for (isub = 0; isub < nsub; isub++) {
        sub = PySequence_GetItem(subscr, isub);
        if (sub == 0)
            ZReturnError(ERR_INDEX, "Invalid array subscripts");
        if ( PyNumber_Check(sub) ) {
            /* Subscript is a number */
            ntemp += 1;
        }
        else if ( PySequence_Check(sub) ) {
            /* Subscript is a list of numbers */
            nelem = PyObject_Length(sub);
            if ( nelem == 0 )
                ZReturnError(ERR_INDEX, 
                             "Empty list of indices is not permitted");
            for (i = 0; i < nelem; i++)
                if ( ! PyNumber_Check(PySequence_GetItem(sub, i)) )
                     ZReturnError(ERR_INDEX, "Index value is not a number");
            ntemp += nelem;
        }
        else if ( PySlice_Check(sub) ) {
            /* Subscript is a slice object */
            ntemp += 2;
        }
        else if ( sub == Py_Ellipsis ) {
            /* Subscript is an ellipsis */
            if (nell > 0)
                ZReturnError(ERR_INDEX,
                             "More than one ellipsis in subscript list");
            nell = 1;
        }
        else {
            ZReturnError(ERR_INDEX, "Illegal object in subscript list");
        }
    }
    if ( (nell == 0 && nsub != rank) || (nell > 0 && nsub > rank) )
        ZReturnError(ERR_INDEX, "Too many or too few subscripts");

    /* Allocate space for index arrays, if needed */
    if ( ntemp > PATEMP ) {  
        temp = mallock(ntemp*sizeof(long));
        if (temp == 0)    ZReturnLibError();
    }

    /* Convert slice objects to sequences, scalars and lists to index
       arrays, and a single ellipsis to the missing coordinates */
    icor = 0;
    p = temp;
    for ( isub = 0; isub < nsub; isub++ ) {
        sub = PySequence_GetItem(subscr, isub);
	if ( PyvoxLong_Check(sub, &s) ) {
            /* Subscript is a number */
	    if ( s < 0 )
		s += sarray->dimen[icor];
            nidx [icor] = 1;
            index[icor] = p;
            *p++ = s;
	    icor++;
        }
        else if ( PySequence_Check(sub) ) {
            /* Subscript is a list of numbers */
            nelem = PyObject_Length(sub);
            nidx [icor] = nelem;
            index[icor] = p;
            if ( ! PyvoxLongSeq_Check(sub, nelem, nelem, 0, p) )
                ZReturnError(ERR_INDEX, "Invalid sequence as subscript");
	    for (i = 0; i < nelem; i++)
		if ( p[i] < 0 )
		    p[i] += sarray->dimen[icor];
            p += nelem;
	    icor++;
        }
        else if ( PySlice_Check(sub) ) {
            /* Subscript is a slice object */
	    flag = parray_convert_slice(sub, sarray->dimen[icor],
                                        &nidx[icor], p, p+1);
            if (flag)    ZReturnLibError();
            index[icor] = p;
	    p += 2;
            icor++;
        }
        else if ( sub == Py_Ellipsis ) {
            /* Subscript is an ellipsis */
            for (i = nsub-1; i < rank; i++)
                nidx[icor++] = 0;
        }
        else 
            ZReturnError(ERR_TYPE, "Unexpected type in subscripts");
    }
    if ( icor > rank || p > temp + ntemp )
        errm_panic("Internal error: overran allocated space");

    return temp;
}



/*------------------------------------------------------------------------
parray_getitem - Return the voxel value(s) at given coordinates

This function returns the voxel values at a given position or in a
given slice.  See parray_prep_slice for a description of the subscript
list.  Attempting to retrieve a non-existent voxel raises an error, as
does attempting to retrieve a slice with no elements.  If the
subscript is a plain number, it is interpreted as an index into the
vectorized array with the negative index convention. But note that a
list of plain numbers is interpreted as coordinates for each axis, not
as a request for the specified indices from the vectorized array.

FIXME: Remove axes for which the subscript is a naked number?
------------------------------------------------------------------------*/

static 
PyObject *
parray_getitem(
    PyObject *self,             /* Voxel array object to fetch from */
    PyObject *subscr)           /* Subscripts for desired voxel or slice */
{
    voxel_array *sarray;        /* Voxel array indicated by self */
    int rank;                   /* Rank of self as voxel array */
    void *value;                /* Value in voxel array */
    pyvox_array *result;        /* The extracted subarray */
    voxel_array *rarray;        /* ...and its voxel array */
    long *dimen = 0;            /* Dimensions of the voxel array */
    long subval[VXL_MAX_RANK];  /* Subscript values */
    long sub0, count;           /* Singleton subscript & nmbr of elements */
    long patemp[PATEMP];        /* Pre-allocated temp space */
    long *temp = patemp;        /* Actual temp space used */
    long nidx[VXL_MAX_RANK];    /* Methods arg for vxl_get_slice */
    long *index[VXL_MAX_RANK];  /* Indices arg for vxl_get_slice */
    int i, flag;

    /* DEBUG: print the arguments */
    if (0) {
        fprintf(stderr, "Calling parray_getitem:\n");
	PyObject_Print(self,  stderr, Py_PRINT_RAW);
	fprintf(stderr, "\n");
	PyObject_Print(subscr, stderr, Py_PRINT_RAW);
	fprintf(stderr, "\n");
    }

    set_error_context("array[]");

    /* Get the voxel array itself, plus its rank */
    if ( ! PyvoxArray_CheckSrc(self, "Pyvox array",  &sarray) )
        ReturnPyvoxLibError();
    rank   = sarray->rank;
    dimen  = sarray->dimen;

    /* If subscript is a single number, return number from vectorized array */
    if ( PyvoxLong_Check(subscr, &sub0) ) {
	count = vxli_count(sarray);
	if ( sub0 < 0 )
	    sub0 += count;
	if ( sub0 < 0 || sub0 >= count )
	    ReturnPyvoxError(ERR_INDEX, "Pyvox array subscript out of range");
	return pyvox_get_value(sarray->type, sarray->data, sub0);
    }

    /* If array is rank 0 or all subscripts are numbers, return a number */
    if ( rank == 0 || PyvoxLongSeq_Check(subscr, rank, rank, 0, subval) ) {
	for (i = 0; i < rank; i++) {
	    if ( subval[i] < 0 )
		subval[i] += dimen[i];
	    if ( subval[i] < 0 || subval[i] >= dimen[i] )
		ReturnPyvoxError(ERR_INDEX, 
                                 "Pyvox array subscript out of range");
	}
	value = vxl_locate(sarray, subval);
	if (value == 0)
  	    ReturnPyvoxError(ERR_INIT, "Invalid subscripts");
	return pyvox_get_value(sarray->type, value, 0);
    }

    /* Convert slice description for use by Voxel Kit */
    temp = parray_prep_slice(sarray, subscr, nidx, index, patemp);
    if (temp == 0)    ReturnPyvoxLibError();

    /* Allocate space for the result */
    result = parray_create(&rarray);
    if (result == 0)    ReturnPyvoxLibError();

    /* Call the Voxel Kit to do the actual work */
    flag = vxl_get_slice(rarray, sarray, nidx, index);
    if (flag)    ReturnPyvoxLibError();

    /* Discard temporary storage and return the result */
    if (temp != patemp)   free(temp);
    clear_error_context();
    return (PyObject *)result;
}



/*------------------------------------------------------------------------
parray_length - Return the length of a Pyvox array

This function returns the length of a Pyvox array and is used to
implement the Pythen len(x) function.

FIXME:  Python expects this function to return int, but array lengths
can be long instead; this will be a bug on platforms for which int and
long are different.  Perhaps check that the count is <= INT_MAX and
abort if not.

FIXME: Python appears to provide no way to indicate an error detected
by this function; if it finds one, it happens to return zero.
------------------------------------------------------------------------*/

static 
int                            /* Returns length */
parray_length(
    PyObject *self)
{
    voxel_array *vself;
    int result;

    set_error_context("len(array)");
    if ( ! PyvoxArray_CheckSrc(self, "self", &vself) )
        ReturnPyvoxLibError();
    result = vxli_count(vself);
    clear_error_context();
    return result;
}



/*------------------------------------------------------------------------
parray_setitem - Set voxels in a given slice

This function sets the values of the voxels at a given position or in
a given slice.  The source may be either a Python numeric value or
rank 0 voxel array, which is converted to the type of the voxel array
according to the rules of the underlying C implementation and
broadcast to all the voxels contained in the slice, or a voxel array
containing exactly the same number of elements as the specified slice,
which are converted to the type of the destination array and copied
into it.  See parray_prep_slice for a description of the subscript
list.  Attempting to set a non-existent voxel raises an error.  If the
subscript is a plain number, it is interpreted as an index into the
vectorized array with the negative index convention. But note that a
list of plain numbers is interpreted as coordinates for each axis, not
as a request for the specified indices from the vectorized array.

NOTE: Contrary to the usual practice in Python, this function is
expected to return 0 for success and non-zero for failure.


FIXME: Extend this to work for the case "foo[...] = [...]"; that is,
for assigning the contents of a sequence into a subarray, or the
entire array without changing its shape or type.

FIXME: Also support sequences (that is, Python slice objects) of
numbers as the value to be assigned?
------------------------------------------------------------------------*/

static 
int                             /* Returns non-zero on error */
parray_setitem(
    PyObject *self,             /* Voxel array object to set values in */
    PyObject *subscr,           /* Subscripts for desired voxel or slice */
    PyObject *value)            /* Value(s) to set voxel or slice to */
{
    voxel_array *darray;        /* Voxel array destination indicated by self */
    int drank;                  /* Rank of destination */
    long *ddimen = 0;           /* Dimensions of destination array */
    voxel_array *sarray;        /* Source array */
    long patemp[PATEMP];        /* Pre-allocated temp space */
    long *temp = patemp;        /* Actual temp space used */
    long coords[VXL_MAX_RANK];  /* Coordinates of a single element */
    long sub0, count;           /* Singleton subscript & nmbr of elements */
    void *datum;                /* Single element to be set */
    long nidx[VXL_MAX_RANK];    /* Methods arg for vxl_set_slice */
    long *index[VXL_MAX_RANK];  /* Indices arg for vxl_set_slice */
    int free1=0;                /* Free sarray at end of function? */
    int flag, i;

    /* DEBUG: print the arguments */
    if (0) {
        fprintf(stderr, "Calling parray_setitem:\n");
	PyObject_Print(self,  stderr, Py_PRINT_RAW);
	fprintf(stderr, "\n");
	PyObject_Print(subscr, stderr, Py_PRINT_RAW);
	fprintf(stderr, "\n");
	PyObject_Print(value, stderr, Py_PRINT_RAW);
	fprintf(stderr, "\n");
    }

    set_error_context("array[]=");

    /* Get the destination voxel array itself, plus its rank */
    darray = ((pyvox_array *)self)->varray;
    drank  = darray->rank;
    ddimen = darray->dimen;

    /* If subscript is a single number, set in vectorized array */
    if ( PyvoxLong_Check(subscr, &sub0) ) {
	count = vxli_count(darray);
	if ( sub0 < 0 )
	    sub0 += count;
	if ( sub0 < 0 || sub0 >= count )
	    XReturnPyvoxError(ERR_INDEX, "Pyvox array subscript out of range");
	flag = pyvox_set_value(darray->type, darray->data, sub0, value);
        if (flag)    XReturnPyvoxLibError();
        return 0;
    }

    /* Handle the case where a single element is being set */
    if ( PyvoxLongSeq_Check(subscr, drank, drank, 0, coords) ) {
        for (i = 0; i < drank; i++) {
            if ( coords[i] < 0 )
                coords[i] += ddimen[i];
            if ( coords[i] < 0 || coords[i] >= ddimen[i] )
                XReturnPyvoxError(ERR_INDEX, 
				  "Pyvox array subscript out of range");
        }
        datum = vxl_locate(darray, coords);
        if (datum == 0)    XReturnPyvoxLibError();
        flag = pyvox_set_value(darray->type, datum, 0, value);
        if (flag)    XReturnPyvoxLibError();
        return 0;
    }

    /* Get the value(s) to assign, wrapping in a scalar array if needed */
    if ( PyvoxArray_Check(value, "source", &sarray) ) {
        free1 = 0;
    }
    else if ( PyNumber_Check(value) ) {
        sarray = varray_create_scalar(darray->type, value);
	if (sarray == 0)    XReturnPyvoxLibError();
        free1 = 1;
    }
    else {
        XReturnPyvoxError(ERR_ILLOP,
			  "Can assign to slice only from array or number");
    }

    /* Call the Voxel Kit to do the real work */
    temp = parray_prep_slice(darray, subscr, nidx, index, patemp);
    if (temp == 0)    XReturnPyvoxLibError();
    flag = vxl_set_slice(darray, nidx, index, sarray);
    if (flag)    XReturnPyvoxLibError();

    /* Free temporary objects and return */
    if (free1)    vxl_destroy_array(sarray);
    if (temp != patemp)  free(temp);

    clear_error_context();
    return 0;
}



/*-----------------------------------------------------------------------
parray_as_mapping - Mapping protocol functions for the voxel array type
-----------------------------------------------------------------------*/

static PyMappingMethods
parray_as_mapping = {
    (inquiry)       parray_length,    /* mp_length:        len(x)   */
    (binaryfunc)    parray_getitem,   /* mp_subscript:     x[y]     */
    (objobjargproc) parray_setitem,   /* mp_ass_subscript: x[y] = v */
};



/************************************************************************
Standard sequence methods for the Pyvox array type

The sequence methods for Pyvox arrays are provided solely to work
arount a bug in Python 1.5.2.  The constructs "x[i:j]" and "x[i:j] =
v" always call the sequence method rather than the mapping method, and
fail with an error message if no sequence method exists; the similar
constructs "x[i]", "x[i] = v", "x[i:j:k]" and "x[slice(i,j)]" all work
fine, calling the mapping method.

FIXME: See if this error has been fixed in later versions of Python.
If not, get this error fixed in Python, and get rid of the sequence
methods defined here.
************************************************************************/



/*------------------------------------------------------------------------
parray_newslice - Create a new slice object

This function creates a new slice object given its start, limit, and
step values.  It is probably used only in parray_getslice and
parray_setslice to work around some limits of the Python
implementation.
-------------------------------------------------------------------------*/

static PyObject *
parray_newslice(
    long start,           /* Start value */
    long limit,           /* Limit value */
    long step)            /* Step  value */
{
    PyObject *pystart, *pylimit, *pystep;   /* Values as Python objects */

    pystart = PyInt_FromLong(start);
    pylimit = PyInt_FromLong(limit);
    pystep  = PyInt_FromLong(step);

    return PySlice_New(pystart, pylimit, pystep);
}



/*------------------------------------------------------------------------
parray_getslice - Get given slice of a 1-d array

This function returns a Pyvox array containing the voxel values in a
given slice of a 1-dimensional array; arrays with other numbers of
dimensions are handled through the mapping methods.  Attempting to
retrieve a non-existent voxel raises an error.

ALGORITHM: We take the easy way out here, and just construct a slice
object and pass it to parray_getitem to do the real work.  The
argument is that we are just working around a bug in Python.
------------------------------------------------------------------------*/

static PyObject *
parray_getslice(
    PyObject *self,             /* Voxel array object to fetch from */
    int start,                  /* Index of first element in slice */
    int limit)                  /* Index of limit element in slice */
{
    PyObject *subscr;           /* Indices as a slice object */
    PyObject *result;           /* Resulting slice */  

    set_error_context("array[slice]");

    /* Convert indices into a slice object and call parray_getitem */
    subscr = parray_newslice(start, limit, 1);
    result = parray_getitem(self, subscr);

    Py_DECREF(subscr);
    clear_error_context();
    return result;
}



/*------------------------------------------------------------------------
parray_setslice - Set given slice of a 1-d array

This function sets the contents of a given slice of a 1-dimensional
array from another 1-d array; arrays with other numbers of dimensions
are handled through the mapping methods.  Attempting to set a
non-existent voxel raises an error.

ALGORITHM: We take the easy way out here, and just construct a slice
object and pass it to parray_getitem to do the real work.  The
argument is that we are just working around a bug in Python.
------------------------------------------------------------------------*/

static int
parray_setslice(
    PyObject *self,             /* Voxel array object to fetch from */
    int start,                  /* Index of first element in slice */
    int limit,                  /* Index of limit element in slice */
    PyObject *value)            /* Values to be set */
{
    PyObject *subscr;           /* Indices as a slice object */
    int result;                 /* Result returned from parray_setitem */

    set_error_context("array[slice]=");

    /* Convert indices into a slice object and call parray_getitem */
    subscr = parray_newslice(start, limit, 1);
    result = parray_setitem(self, subscr, value);

    Py_DECREF(subscr);
    clear_error_context();
    return result;
}



/*----------------------------------------------------------------------
parray_as_sequence - Table of sequence methods for the Pyvox array type

FIXME: Verify that the mapping methods support slices, and then remove
the sequence methods?  Will it still work properly for 1-D arrays?
----------------------------------------------------------------------*/

static PySequenceMethods 
parray_as_sequence = {
    (inquiry)          0,                  /* sq_length:     len(x)     */
    (binaryfunc)       0,                  /* sq_concat:     x + y      */
    (intargfunc)       0,                  /* sq_repeat:     n * x      */
    (intargfunc)       0,                  /* sq_item:       x[i]       */
    (intintargfunc)    parray_getslice,    /* sq_slice:      x[i:j]     */
    (intobjargproc)    0,                  /* sq_ass_item:   x[i] = v   */
    (intintobjargproc) parray_setslice,    /* sq_ass_slice:  x[i:j] = v */
};



/************************************************************************
Standard Numeric Methods for the Pyvox Array Type
************************************************************************/


/*----------------------------------------------------------------------
parray_coerce - Coerce voxel arrays to common type if possible

This function takes two Python objects that may be voxel arrays,
scalars, or numbers and coerces them to a common type so that
arithmetic may be performed upon them.  The converted objects replace
the original objects.  Zero is returned if conversion is successful,
and a non-zero value if it fails.  Checking for compatibility
of types and sizes is done by the arithmetic routines themselves.
The references passed by the caller are assumed to be borrowed, and
the reference counts will be incremented if the object is retained
without conversion.

See also setup_vxl_binpt, which does coercion for methods that Python
does not automatically classify as numeric.

FIXME: Modify all parray functions to do their own coercion?  There is
a flag Py_TPFLAGS_CHECKTYPES in tp_flags of the type object to specify
this.

FIXME: Generalize to convert two voxel arrays of different types when
needed.  Better yet, have the Voxel Kit functions convert the inputs
by blocks on the fly to save cache refills.

FIXME: The exact semantics of this function don't seem to be defined
in the Python documentation, and I have guessed at it from studying a
similar routine in Numerical Python.  Some verification might be in
order.  Some questions:  Can the case "voxel sees voxel" be used to
coerce incompatible voxel types?  Is the case "number sees voxel"
actually used, or does Python convert automatically to the symmetric
case "voxel sees number"?  Are the reference counts managed correctly?
------------------------------------------------------------------------*/

static 
int                          /* Returns non-zero on error */
parray_coerce(
    PyObject **src1,         /* First object */
    PyObject **src2)         /* Second object */
{
    int intype;              /* Internal type of voxel array */

    if ( PyvoxArray_Check(*src1, 0, 0) && PyvoxArray_Check(*src2, 0, 0) ) {
	if (0) fprintf(stderr, "Voxel sees voxel\n");
	Py_INCREF(*src1);
	Py_INCREF(*src2);
	return 0; }
    else if ( PyvoxArray_Check(*src1, 0, 0) && PyNumber_Check(*src2) ) {
	if (0) fprintf(stderr, "Voxel sees number\n");
	intype = PyVoxel_Type(*src1);
	*src2 = PyScalar_FromNumber(*src2, intype);
	Py_INCREF(*src1);
	return 0; }
    else if ( PyNumber_Check(*src1) && PyvoxArray_Check(*src2, 0, 0) ) {
	if (0) fprintf(stderr, "Number sees voxel\n");
	intype = PyVoxel_Type(*src2);
	*src1 = PyScalar_FromNumber(*src1, intype);
	Py_INCREF(*src2);
	return 0; }
    else
	return -1;          /* Cannot make compatible */
	
}



/*-------------------------------------------------------------------------
parray_vxl_unfunc - Common function wrapper for unary voxel kit function

This function defines a wrapper for a common class of unary Pyvox
array functions that are implemented using unary voxel kit functions.
This function assumes that the error context has already been set, but
clears it before returning.
---------------------------------------------------------------------------*/

static inline
PyObject *                   /* Returns result, or zero for error */
parray_vxl_unfunc(
    vxl_unop  *vxl_func,     /* Voxel Kit function to use */
    PyObject  *pysrc1)       /* Source array */
{
    voxel_array  *src1;      /* Operand as voxel array */
    pyvox_array  *result;    /* Result as a Pyvox array */
    voxel_array  *rarray;    /* Result's voxel array */
    int flag;

    /* Check that the argument is a voxel array */
    if ( !PyvoxArray_Check(pysrc1, "source", &src1) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the actual computation */
    result = parray_create(&rarray);
    if (result == 0)    ReturnPyvoxLibError();
    flag = vxl_func(result->varray, src1);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)result;
}


/*-------------------------------------------------------------------------
parray_abs1 - Elementwise absolute value of a voxel array

This deprecated function is invoked by the Python abs(array); the
recommended syntax array.abs() is consistent with other univariate
array operations in Pyvox and is implemented by parray_abs2.
--------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_abs1(
    PyObject *pysrc1)       /* Arg 1: Source */
{
    set_error_context("abs(array)");
    return parray_vxl_unfunc(vxl_abs, pysrc1);
}



/*-------------------------------------------------------------------------
parray_minus - Elementwise additive inverse of a voxel array
--------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_minus(
    PyObject *pysrc1)       /* Arg 1: Source */
{
    set_error_context("-array");
    return parray_vxl_unfunc(vxl_minus, pysrc1);
}



/*-------------------------------------------------------------------------
parray_not - Bitwise NOT of a voxel array
--------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_not(
    PyObject *pysrc1)       /* Arg 1: Source */
{
    set_error_context("~array");
    return parray_vxl_unfunc(vxl_not, pysrc1);
}



/*-------------------------------------------------------------------------
parray_plus - Complement to additive inverse of a voxel array

This function is invoked to compute "+V" when V is a voxel array;
it does nothing except to pass through its argument, without even
copying it or doing any error checking.  It is provided only for
symmetry with "-V".
--------------------------------------------------------------------------*/

static PyObject *           /* Result: combination of operands */
parray_plus(
    PyObject *pysrc1)       /* Arg 1: Source */
{
    return pysrc1;
}



/*-------------------------------------------------------------------------
parray_vxl_binfunc - Common function wrapper for binary voxel kit functions

This function defines a wrapper for a common class of binary Pyvox
array functions that are implemented using binary voxel kit functions.
It assumes the error context has already been set but clears it before
returning.
---------------------------------------------------------------------------*/

static inline
PyObject *                       /* Returns result, or zero if error */
parray_vxl_binfunc(
    vxl_binop  *vxl_func,        /* Voxel Kit function to use */
    PyObject   *pysrc1,          /* First source array */
    PyObject   *pysrc2)          /* Second source array */
{
    voxel_array  *src1, *src2;   /* Operands as voxel arrays */
    pyvox_array  *result;        /* Result as a Pyvox array */
    voxel_array  *rarray;        /* Result's voxel array */
    int flag;

    /* Check that both arguments have been coerced to voxel arrays */
    if ( !PyvoxArray_Check(pysrc1, "source 1", &src1) )
	ReturnPyvoxLibError();
    if ( !PyvoxArray_Check(pysrc2, "source 2", &src2) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the actual computation */
    result = parray_create(&rarray);
    if (result == 0)    ReturnPyvoxLibError();
    flag = vxl_func(result->varray, src1, src2);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)result;
}



/*-------------------------------------------------------------------------
parray_and - Bitwise AND of two voxel arrays
-------------------------------------------------------------------------*/

static PyObject *parray_and(PyObject *pysrc1, PyObject *pysrc2)
{
    set_error_context("array & array");
    return parray_vxl_binfunc(vxl_and, pysrc1, pysrc2);
}



/*-------------------------------------------------------------------------
parray_or - Bitwise OR of two voxel arrays
-------------------------------------------------------------------------*/

static PyObject *parray_or(PyObject *pysrc1, PyObject *pysrc2)
{
    set_error_context("array | array");
    return parray_vxl_binfunc(vxl_or, pysrc1, pysrc2);
}



/*-------------------------------------------------------------------------
parray_xor - Bitwise XOR of two voxel arrays
-------------------------------------------------------------------------*/

static PyObject *parray_xor(PyObject *pysrc1, PyObject *pysrc2)
{
    set_error_context("array ^ array");
    return parray_vxl_binfunc(vxl_xor, pysrc1, pysrc2);
}



/*-------------------------------------------------------------------------
parray_add -  Sum of two voxel arrays
-------------------------------------------------------------------------*/

static PyObject *parray_add(PyObject *pysrc1, PyObject *pysrc2)
{
    set_error_context("array + array");
    return parray_vxl_binfunc(vxl_add, pysrc1, pysrc2);
}



/*-------------------------------------------------------------------------
parray_sub - Difference of two voxel arrays
-------------------------------------------------------------------------*/

static PyObject *parray_sub(PyObject *pysrc1, PyObject *pysrc2)
{
    set_error_context("array - array");
    return parray_vxl_binfunc(vxl_sub, pysrc1, pysrc2);
}



/*-------------------------------------------------------------------------
parray_mul - Product of two voxel arrays
-------------------------------------------------------------------------*/

static PyObject *parray_mul(PyObject *pysrc1, PyObject *pysrc2)
{
    set_error_context("array * array");
    return parray_vxl_binfunc(vxl_mul, pysrc1, pysrc2);
}



/*-------------------------------------------------------------------------
parray_div - Quotient of two voxel arrays
-------------------------------------------------------------------------*/

static PyObject *parray_div(PyObject *pysrc1, PyObject *pysrc2)
{
    set_error_context("array / array");
    return parray_vxl_binfunc(vxl_div, pysrc1, pysrc2);
}



/*-------------------------------------------------------------------------
parray_div - Remainder of two voxel arrays

Computes x - trunc(x/y)*y where trunc() represents rounding toward zero.
-------------------------------------------------------------------------*/

static PyObject *parray_rem(PyObject *pysrc1, PyObject *pysrc2)
{
    set_error_context("array % array");
    return parray_vxl_binfunc(vxl_rem, pysrc1, pysrc2);
}



/*----------------------------------------------------------------------
parray_as_number - Number protocol functions for the voxel array type

FIXME: Arrange for float(), int(), and long() to return a number for
singleton arrays and an error otherwise.  This will go partway toward
making singleton arrays act like numbers in the appropriate contexts.

FIXME: Several more of these could usefully be implemented, e.g. divmod,
power, nonzero?, lshift, rshift.
-----------------------------------------------------------------------*/

static PyNumberMethods
parray_as_number = {
	(binaryfunc)  parray_add,       /* nb_add */
	(binaryfunc)  parray_sub,       /* nb_subtract */
	(binaryfunc)  parray_mul,       /* nb_multiply */
	(binaryfunc)  parray_div,       /* nb_divide */
	(binaryfunc)  parray_rem,       /* nb_remainder */
	(binaryfunc)  0,                /* nb_divmod */
	(ternaryfunc) 0,                /* nb_power */
	(unaryfunc)   parray_minus,     /* nb_negative */
	(unaryfunc)   parray_plus,      /* nb_positive */
	(unaryfunc)   parray_abs1,      /* nb_absolute */
	(inquiry)     0,                /* nb_nonzero */
	(unaryfunc)   parray_not,       /* nb_invert */
	(binaryfunc)  0,                /* nb_lshift */
	(binaryfunc)  0,                /* nb_rshift */
	(binaryfunc)  parray_and,       /* nb_and */
	(binaryfunc)  parray_xor,       /* nb_xor */
	(binaryfunc)  parray_or,        /* nb_or */
	(coercion)    parray_coerce,    /* nb_coerce */
	(unaryfunc)   0,                /* nb_int */
	(unaryfunc)   0,                /* nb_long */
	(unaryfunc)   0,                /* nb_float */
	(unaryfunc)   0,                /* nb_oct */
	(unaryfunc)   0,                /* nb_hex */
};



/**************************************************************************
Simple Unary Point Operations on Pyvox Arrays

These operations correspond directly to Voxel Kit operations and can
be handled via simple wrapper functions.

FIXME: This is a somewhat elastic category and may need some refinement.
Note also that some methods may still need to be modified to use the
standard wrapper functions.
**************************************************************************/



/*-------------------------------------------------------------------------
parray_vxl_unop - Generic unary operation on voxel arrays

This wrapper function is used to translate a standard unary Voxel Kit
function into a Pyvox array method.  More precisely, this function
applies a given function of type vxl_unop from the Voxel Kit to (the
voxel array contained in) self and returns the resulting voxel array
as a Pyvox array.  The Pyvox method takes no arguments.  This function
assumes the error context has already been set, but clears it before
exit.
-------------------------------------------------------------------------*/

static inline
PyObject *
parray_vxl_unop(
    vxl_unop *funct,          /* Voxel Kit function to apply */
    PyObject *self,           /* Self given to wrapper function */
    PyObject *args)           /* Args given to wrapper function */
{
    voxel_array *vin, *vout;  /* Input and output as voxel arrays */
    pyvox_array *out;         /* Output as (Pyvox) Pyvox arrays */
    int flag;

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;
    if ( !PyvoxArray_Check(self, "self", &vin) )
	ReturnPyvoxLibError();
  
    /* Compute and return the result */
    out = parray_create(&vout);
    if (out == 0)    ReturnPyvoxLibError();
    flag = funct(vout, vin);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)out;
}



/*-----------------------------------------------------------------------
Type conversions

These Pyvox array methods convert self, which is a Pyvox array any
type into a new Pyvox array of the specified type.  The details of the
conversion are inherited from the underlying C implementation.  A
Pyvox array can be converted to another Pyvox array of the same type;
this is simply a copy (and is probably almost as fast as a plain
copy).

The code for parray_uchar is treated specially, to provide the capability
to scale before conversion, since this operation is used so heavily.

The code for parray_fcomplex and parray_dcomplex is treated specially,
to allow an optional second argument that specifies the imaginary part.
-------------------------------------------------------------------------*/

static PyObject * parray_schar(PyObject *self, PyObject *args)
{   
    set_error_context("array.schar()");
    return parray_vxl_unop(vxl_schar, self, args); 
}

static PyObject * parray_ushort(PyObject *self, PyObject *args)
{   
    set_error_context("array.ushort()");
    return parray_vxl_unop(vxl_ushort, self, args); 
}

static PyObject * parray_short(PyObject *self, PyObject *args)
{   
    set_error_context("array.short()");
    return parray_vxl_unop(vxl_short, self, args); 
}

static PyObject * parray_uint(PyObject *self, PyObject *args)
{   
    set_error_context("array.uint()");
    return parray_vxl_unop(vxl_uint, self, args); 
}

static PyObject *parray_int(PyObject *self, PyObject *args)
{   
    set_error_context("array.int()");
    return parray_vxl_unop(vxl_int, self, args); 
}

static PyObject *parray_ulong(PyObject *self, PyObject *args)
{   
    set_error_context("array.ulong()");
    return parray_vxl_unop(vxl_ulong, self, args); 
}

static PyObject *parray_long(PyObject *self, PyObject *args)
{   
    set_error_context("array.long()");
    return parray_vxl_unop(vxl_long, self, args); 
}

static PyObject *parray_float(PyObject *self, PyObject *args)
{   
    set_error_context("array.float()");
    return parray_vxl_unop(vxl_float, self, args); 
}

static PyObject *parray_double(PyObject *self, PyObject *args)
{   
    set_error_context("array.double()");
    return parray_vxl_unop(vxl_double, self, args); 
}



/*-------------------------------------------------------------------------
parray_cast - Convert to specified type

This Pyvox function converts self, which is a Pyvox array of any type
into a Pyvox array of some type specified by the caller.
-------------------------------------------------------------------------*/

static PyObject *
parray_cast(
    PyObject *self,
    PyObject *args)
{
    PyObject *ptype;          /* Desired type as a Python object */
    int type;                 /* Desired type as a number */

    voxel_array *vin, *vout;  /* Input and output as voxel arrays */
    pyvox_array *out;         /* Output as a Pyvox array */
    int flag;

    set_error_context("array.cast()");

    /* Fetch the arguments */
    if ( !PyvoxArray_Check(self, "self", &vin) )
	ReturnPyvoxLibError();
    if ( !PyArg_ParseTuple(args, "O", &ptype) )
	return NULL;
    if ( !PyvoxIntype_Check(ptype, "type", &type) )
        ReturnPyvoxLibError();

    /* Compute and return the result */
    out = parray_create(&vout);
    if (out == 0)    ReturnPyvoxLibError();
    flag = vxl_cast(vout, type, vin);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)out;
}



/*-------------------------------------------------------------------------
parray_uchar - Convert to unsigned char, with optional rescaling

This Pyvox function converts self, which is a Pyvox array of any type
into a Pyvox array of unsigned char.  The conversion is done the same
way as the underlying C implementation; one consequence is that casting
from an integral type is done modulo the size of the destination.

The caller may optionally specify the lower and upper limits of a
range to be linearly compressed to fit the 0..255 range of unsigned
char voxels.  If this is done, then the results are limited to fit the
range rather than cast modulo the destination size; note that this
happens even if the limits happen to be 0 and 255.

NOTE: Prior to 2003-08-25, the results were limited to the destination
min/max even if no lower and upper limits were specified.

FIXME: The operation of scaling, limiting, and casting to an integral
type is no doubt useful, but I begin to wonder if it should have its
own method rather than piggy-backing on casting.  Alternatively,
should it be extended to the other integral types?
-------------------------------------------------------------------------*/

static PyObject *
parray_uchar(
    PyObject *self,
    PyObject *args)
{
    int nargs;                /* Number of arguments provided */
    double minval =   0.0;    /* Arg 1: Lower limit of input range */
    double maxval = 255.0;    /* Arg 2: Upper limit of input range */

    voxel_array *vin, *vout;  /* Input and output as voxel arrays */
    pyvox_array *out;         /* Output as a Pyvox array */
    int flag;

    set_error_context("array.uchar()");

    /* Check that the number of args is either 0 or 2 */
    nargs = PyObject_Length(args);
    if ( nargs != 0 && nargs != 2)
        ReturnPyvoxError(ERR_ILLOP, "Must have either 0 or 2 arguments");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "|dd", &minval, &maxval) )
	return NULL;
    if ( !PyvoxArray_Check(self, "self", &vin) )
	ReturnPyvoxLibError();
  
    /* Compute and return the result */
    out = parray_create(&vout);
    if (out == 0)    ReturnPyvoxLibError();
    if (nargs == 0)
        flag = vxl_uchar(vout, vin);
    else
       flag = vxl_scale_to_uchar(vout, vin, minval, maxval);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)out;
}



/*-------------------------------------------------------------------------
parray_fcomplex - Convert to float complex, with optional imaginary part

This Pyvox function converts self, which is a Pyvox array of any type
into a Pyvox array of float complex; an optional second argument
specifies the imaginary part.
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H
static PyObject *
parray_fcomplex(
    PyObject *self,
    PyObject *args)
{
    PyObject *imag=0;             /* Arg 1 : Imaginary part */

    voxel_array *vreal, *vimag;   /* Inputs as voxel arrays */
    voxel_array *vout;            /* Output as a voxel array */
    pyvox_array *out;             /* Output as a Pyvox array */
    int flag;

    set_error_context("array.fcomplex()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "|O", &imag) )
	return NULL;
    if ( !PyvoxArray_Check(self, "self", &vreal) )
	ReturnPyvoxLibError();
    if ( imag != 0 && !PyvoxArray_Check(imag, "imag", &vimag) )
        ReturnPyvoxLibError();
  
    /* Compute and return the result */
    out = parray_create(&vout);
    if (out == 0)    ReturnPyvoxLibError();
    if (imag == 0)
        flag = vxl_fcomplex(vout, vreal);
    else
        flag = vxl_fcomplex2(vout, vreal, vimag);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)out;
}
#endif



/*-------------------------------------------------------------------------
parray_dcomplex - Convert to double complex, with optional imaginary part

This Pyvox function converts self, which is a Pyvox array of any type
into a Pyvox array of double complex; an optional second argument
specifies the imaginary part.
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H
static PyObject *
parray_dcomplex(
    PyObject *self,
    PyObject *args)
{
    PyObject *imag=0;             /* Arg 1 : Imaginary part */

    voxel_array *vreal, *vimag;   /* Inputs as voxel arrays */
    voxel_array *vout;            /* Output as a voxel array */
    pyvox_array *out;             /* Output as a Pyvox array */
    int flag;

    set_error_context("array.dcomplex()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "|O", &imag) )
	return NULL;
    if ( !PyvoxArray_Check(self, "self", &vreal) )
	ReturnPyvoxLibError();
    if ( imag != 0 && !PyvoxArray_Check(imag, "imag", &vimag) )
        ReturnPyvoxLibError();
  
    /* Compute and return the result */
    out = parray_create(&vout);
    if (out == 0)    ReturnPyvoxLibError();
    if (imag == 0)
        flag = vxl_dcomplex(vout, vreal);
    else
        flag = vxl_dcomplex2(vout, vreal, vimag);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)out;
}
#endif



/*------------------------------------------------------------------------
Unary operations on voxel arrays, in standard format

The following functions all take a single Pyvox voxel array as input,
generate a single Pyvox voxel array as output, and call upon a function
of type vxl_unop from the Voxel Kit.  This uniformity means that theyd
can all be implemented in the same fashion using the wrapper function
parray_vxl_unop.  Each definition *could* be reduced to a single line
by using an appropriate macro, but this would make them harder to find
using emacs tags and to understand.

FIXME: Define array.abs(other) to compute and return the absolute value
of the difference?  Note the possibility of overflow for integral types.
Note also that bips_absdiff returns double and cannot be used for this
purpose.
-------------------------------------------------------------------------*/

static PyObject *parray_abs2(PyObject *self, PyObject *args)
{   
    set_error_context("array.abs()");
    return parray_vxl_unop(vxl_abs, self, args); 
}

static PyObject *parray_acos(PyObject *self, PyObject *args)
{   
    set_error_context("array.acos()");
    return parray_vxl_unop(vxl_acos, self, args); 
}

static PyObject *parray_asin(PyObject *self, PyObject *args)
{   
    set_error_context("array.asin()");
    return parray_vxl_unop(vxl_asin, self, args); 
}

static PyObject *parray_atan(PyObject *self, PyObject *args)
{   
    set_error_context("array.atan()");
    return parray_vxl_unop(vxl_atan, self, args); 
}

#ifdef HAVE_CBRT
static PyObject *parray_cbrt(PyObject *self, PyObject *args)
{   
    set_error_context("array.cbrt()");
    return parray_vxl_unop(vxl_cbrt, self, args); 
}
#endif

static PyObject *parray_ceil(PyObject *self, PyObject *args)
{   
    set_error_context("array.ceil()");
    return parray_vxl_unop(vxl_ceil, self, args); 
}

#ifdef HAVE_COMPLEX_H
static PyObject *parray_carg(PyObject *self, PyObject *args)
{
    set_error_context("array.carg()");
    return parray_vxl_unop(vxl_carg, self, args);
}

static PyObject *parray_conj(PyObject *self, PyObject *args)
{
    set_error_context("array.conj()");
    return parray_vxl_unop(vxl_conj, self, args);
}

static PyObject *parray_imag(PyObject *self, PyObject *args)
{
    set_error_context("array.imag()");
    return parray_vxl_unop(vxl_imag, self, args);
}

static PyObject *parray_real(PyObject *self, PyObject *args)
{
    set_error_context("array.real()");
    return parray_vxl_unop(vxl_real, self, args);
}
#endif

static PyObject *parray_cos(PyObject *self, PyObject *args)
{   
    set_error_context("array.cos()");
    return parray_vxl_unop(vxl_cos, self, args); 
}

static PyObject *parray_cosh(PyObject *self, PyObject *args)
{   
    set_error_context("array.cosh()");
    return parray_vxl_unop(vxl_cosh, self, args); 
}

#ifdef HAVE_ERF
static PyObject *parray_erf(PyObject *self, PyObject *args)
{   
    set_error_context("array.erf()");
    return parray_vxl_unop(vxl_erf, self, args); 
}
#endif

#ifdef HAVE_ERFC
static PyObject *parray_erfc(PyObject *self, PyObject *args)
{   
    set_error_context("array.erfc()");
    return parray_vxl_unop(vxl_erfc, self, args); 
}
#endif

static PyObject *parray_exp(PyObject *self, PyObject *args)
{   
    set_error_context("array.exp()");
    return parray_vxl_unop(vxl_exp, self, args); 
}

#ifdef HAVE_EXPM1
static PyObject *parray_expm1(PyObject *self, PyObject *args)
{   
    set_error_context("array.expm1()");
    return parray_vxl_unop(vxl_expm1, self, args); 
}
#endif

#ifdef HAVE_EXP2
static PyObject *parray_exp2(PyObject *self, PyObject *args)
{   
    set_error_context("array.exp2()");
    return parray_vxl_unop(vxl_exp2, self, args); 
}
#endif

static PyObject *parray_floor(PyObject *self, PyObject *args)
{   
    set_error_context("array.floor()");
    return parray_vxl_unop(vxl_floor, self, args); 
}

#ifdef HAVE_LGAMMA
static PyObject *parray_lgamma(PyObject *self, PyObject *args)
{   
    set_error_context("array.lgamma()");
    return parray_vxl_unop(vxl_lgamma, self, args); 
}
#endif

static PyObject *parray_log(PyObject *self, PyObject *args)
{   
    set_error_context("array.log()");
    return parray_vxl_unop(vxl_log, self, args); 
}

static PyObject *parray_log10(PyObject *self, PyObject *args)
{   
    set_error_context("array.log10()");
    return parray_vxl_unop(vxl_log10, self, args); 
}

#ifdef HAVE_LOG1P
static PyObject *parray_log1p(PyObject *self, PyObject *args)
{   
    set_error_context("array.log1p()");
    return parray_vxl_unop(vxl_log1p, self, args); 
}
#endif

#ifdef HAVE_LOG2
static PyObject *parray_log2(PyObject *self, PyObject *args)
{   
    set_error_context("array.log2()");
    return parray_vxl_unop(vxl_log2, self, args); 
}
#endif

static PyObject *parray_sin(PyObject *self, PyObject *args)
{   
    set_error_context("array.sin()");
    return parray_vxl_unop(vxl_sin, self, args); 
}

static PyObject *parray_sinh(PyObject *self, PyObject *args)
{   
    set_error_context("array.sinh()");
    return parray_vxl_unop(vxl_sinh, self, args); 
}

static PyObject *parray_sqrt(PyObject *self, PyObject *args)
{   
    set_error_context("array.sqrt()");
    return parray_vxl_unop(vxl_sqrt, self, args); 
}

static PyObject *parray_tan(PyObject *self, PyObject *args)
{   
    set_error_context("array.tan()");
    return parray_vxl_unop(vxl_tan, self, args); 
}

static PyObject *parray_tanh(PyObject *self, PyObject *args)
{   
    set_error_context("array.tanh()");
    return parray_vxl_unop(vxl_tanh, self, args); 
}

#ifdef HAVE_TGAMMA
static PyObject *parray_tgamma(PyObject *self, PyObject *args)
{   
    set_error_context("array.tgamma()");
    return parray_vxl_unop(vxl_tgamma, self, args); 
}
#endif



/*-------------------------------------------------------------------------
parray_logcomp - Log-compress an unsigned long volume image

This Pyvox array method compresses the dynamic range of an unsigned
long Pyvox array using the transformation y = A*log(1+x), where A is
chosen such that the largest voxel value actually present in the image
is converted to the value 255; the compressed image is returned as
unsigned char.
-------------------------------------------------------------------------*/

static 
PyObject *
parray_logcomp(
    PyObject *self,
    PyObject *args)
{
    pyvox_array *out;         /* Output as a Pyvox array */
    voxel_array *vin, *vout;  /* Input and output as voxel arrays */
    int flag;

    set_error_context("array.logcomp()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;
    if ( !PyvoxArray_Check(self, "self", &vin) )
	ReturnPyvoxLibError();
  
    /* Compute and return the result */
    out = parray_create(&vout);
    if (out == 0)    ReturnPyvoxLibError();
    flag = vxl_logcomp_ulong(vout, vin);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)out;
}



/*-------------------------------------------------------------------------
parray_lookup - Process self through a lookup table

This Pyvox array method takes each voxel in self and uses it as the
index into a lookup table, saving the result of the lookup as the
destination image.  Self must be either unsigned char or unsigned
short but may be of any shape.  The table image must be rank 1 and
contain at least as many entries as the largest voxel value in the
source image; but it (and thus the destination) may be of any type.
-------------------------------------------------------------------------*/

static 
PyObject *                  /* Returns result, or NULL on error */
parray_lookup(
    PyObject *self,         /* Self: Source image */
    PyObject *args)
{
    PyObject *pysrc2;       /* Arg 1: Lookup table */

    voxel_array *src1, *src2;         /* Sources as voxel arrays */
    pyvox_array *vodest;              /* Result as a Pyvox array */
    voxel_array *dest;                /* Result as a voxel array */
    int flag;

    set_error_context("array.lookup()");

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "O", &pysrc2) )
	return NULL;
    if ( !PyvoxArray_Check(self, "source", &src1) )
	ReturnPyvoxLibError();
    if ( !PyvoxArray_Check(pysrc2, "lookup table", &src2) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the actual computation */
    vodest = parray_create(&dest);
    if (vodest == 0)    ReturnPyvoxLibError();
    flag = vxl_lookup(dest, src1, src2);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)vodest;
}




/**************************************************************************
Simple Binary Point Operations on Pyvox Arrays 

These operations correspond directly to Voxel Kit operations and can
be handled via simple wrapper functions.

FIXME: This is a somewhat elastic category and may need some refinement.
Note also that some methods may still need to be modified to use the
standard wrapper functions.
**************************************************************************/



/*-------------------------------------------------------------------------
setup_vxl_binpt - Set up a binary point operation in the Voxel Kit

Given two Python objects which may be either numbers or voxel array
objects, this function promotes numbers into scalar voxel array
objects so that the Voxel Kit may be called in all cases.  It is an
error to provide two voxel arrays of different types; casting is NOT
automatic as it is in C.  A number is promoted to a scalar array of
the same type as the other object; if two numbers are provided, then
they are both promoted to a default type specified as an argument.
The default type will be promoted to complex if either number is
complex in the case number meets number.
Shape compatibility will be checked within the Voxel Kit itself.
Promoting a number yields a temporary scalar array which must be freed
once the operation is complete; the free1 and free2 flags indicate
which must be freed.  This function returns a pointer to the first
Python object on success (which is meaningless but non-zero), and zero
on failure.  

NOTE that this function is inappropriate for those arithmetic
operations for which the Python type coercion method is automatically
invoked; see parray_and for an example.  See also parray_coerce, which
does coercion for methods that Python automatically classifies as
numeric.

FIXME: The use of a default type in the number + number case will probably
be incorrect when the inputs are complex numbers; on the other hand,
it's not clear that this case will ever be executed.

FIXME: Combining a real array and a complex number (or vice versa)
will not work correctly.  In fact, check through all the possibilities
involving complex arrays and/or numbers.

FIXME: When used in a pyvox class method, this function permits either
operand to be a scalar.  This is no longer possible when used as a
Pyvox array method.  Do I need to provide a means to compute binary
operations when the first operand is a number?  If so, how?  One idea
is to add a method pyvox.scalar which take a number and converts it to
a scalar array; however, this might not coerce to the right type.
The usage might be something like "pyvox.scalar(0.39).atan2(array)".
Another option might be for the methods to take either one or two
arguments: if one arg, then self and arg 1 are combined; if two args,
then args 1 and 2 are combined.
------------------------------------------------------------------------*/

static inline
PyObject *                       /* Returns NULL on error */
setup_vxl_binpt(
    PyObject *pysrc1,            /* Left operand, as Python object */
    PyObject *pysrc2,            /* Right operand, as Python object */
    int deftype,                 /* Default type if both numbers */
    voxel_array **src1,          /* Left operand, as voxel array */
    int *free1,                  /* Should src1 be freed after use? */
    voxel_array **src2,          /* Right operand, as voxel array */
    int *free2)                  /* Should src2 be freed after use? */
{
    int type = 0;                /* Type of voxel arrays */

    /* Case 1: Both arguments are Pyvox arrays; set up to call Voxel Kit */
    if ( PyvoxArray_Check(pysrc1, 0, src1) &&
	 PyvoxArray_Check(pysrc2, 0, src2) ) {
	if (0) fprintf(stderr, "Voxel meets voxel\n");
	*free1 = *free2 = 0;
    }

    /* Case 2: First argument is Pyvox array, second is number; convert
       second to scalar array and set up to call Voxel Kit */
    else if ( PyvoxArray_Check(pysrc1, 0, src1) &&
	      PyNumber_Check(pysrc2) ) {
	if (0) fprintf(stderr, "Voxel meets number\n");
	type = (*src1)->type;
	*src2 = varray_create_scalar(type, pysrc2);
	if (*src2 == 0)    ReturnPyvoxLibError();
	*free1 = 0;
	*free2 = 1;
    }

    /* Case 3: First argument is a number, second is Pyvox array; convert
       first to scalar array and set up to call Voxel Kit */
    else if ( PyNumber_Check(pysrc1) &&
	      PyvoxArray_Check(pysrc2, 0, src2) ) {
	if (0) fprintf(stderr, "Number meets voxel\n");
	type = (*src2)->type;
	*src1 = varray_create_scalar(type, pysrc1);
	if (*src1)    ReturnPyvoxLibError();
	*free1 = 1;
	*free2 = 0;
    }

    /* Case 4: Both arguments are numbers; convert both to default type */
    else if ( PyNumber_Check(pysrc1) && PyNumber_Check(pysrc2) ) {
	if (0) fprintf(stderr, "Number meets number\n");
	type = deftype;
        if ( PyComplex_Check(pysrc1) || PyComplex_Check(pysrc2) )
            type = exim_complex_type(deftype);
	*src1 = varray_create_scalar(type, pysrc1);
        if (*src1)    ReturnPyvoxLibError();
	*src2 = varray_create_scalar(type, pysrc2);
        if (*src2)    ReturnPyvoxLibError();
	*free1 = *free2 = 1;
    }

    /* Case 5:  Anything else is an error */
    else
	ReturnPyvoxError(ERR_INCOMP, 
                         "Incompatible arguments to voxel array computation");

    return pysrc1;
}



/*-------------------------------------------------------------------------
parray_vxl_binop - Pyvox array method wrapper for binary Voxel Kit function

This wrapper function is used to translate many common binary Voxel
Kit functions into a one-argument method for a Pyvox array.  This
function assumes the error context has already been set but clears it
before returning.
--------------------------------------------------------------------------*/

static inline
PyObject *                        /* Returns result, or NULL on error */
parray_vxl_binop(
    vxl_binop  *vxl_func,         /* Voxel Kit function to apply */
    int        deftype,           /* Default type for coercion */
    PyObject   *self,             /* Self: First operand */
    PyObject   *args)
{
    PyObject *arg1;               /* Arg 1: Second operand */

    voxel_array *src1=0, *src2=0; /* Operands as voxel arrays */
    int free1 = 0, free2 = 0;     /* Should src1 & src2 be freed? */
    pyvox_array  * dest;          /* Result as a Pyvox array */
    voxel_array  *vdest;          /* Result as a voxel array */
    int flag;

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "O", &arg1) )
	return NULL;

    /* Promote numeric arguments to scalar arrays if necessary */
    if ( !setup_vxl_binpt(self, arg1, INTP_DOUBLE,
			  &src1, &free1, &src2, &free2) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the actual computation */
    dest  = parray_create(&vdest);
    if (dest == 0)    ReturnPyvoxLibError();
    flag = vxl_func(vdest, src1, src2);
    if (flag)    ReturnPyvoxLibError();

    /* Free the scalar arrays, if needed */
    if (free1 != 0) vxl_destroy_array(src1);
    if (free2 != 0) vxl_destroy_array(src2);

    clear_error_context();
    return (PyObject *)dest;
}



/*-------------------------------------------------------------------------
parray_atan2 - Four-quadrant arc tangent of two voxel arrays

This Pyvox function computes the elementwise four-quadrant arc tangent
of two numeric voxel arrays; that is, it computes dest = atan2(x, y)
for each element.  The first array is self; the second is passed as
the method argument.
-------------------------------------------------------------------------*/

static PyObject *
parray_atan2(PyObject *self, PyObject *args)
{
    set_error_context("array.atan2()");
    return parray_vxl_binop(vxl_atan2, INTP_DOUBLE, self, args);
}



/*-----------------------------------------------------------------------
parray_cross - Cross product of vectors in 3-d

This Pyvox function computes the vector cross product of two voxel
arrays considered as vectors in 3-d.  As long as both arguments have
exactly three elements, their shape does not matter.  The type and
shape of the result are set to match the first argument.
------------------------------------------------------------------------*/

static PyObject *
parray_cross(PyObject *self, PyObject *args)
{
    PyObject *arg1;               /* Second operand as Pyvox array */

    voxel_array *src1, *src2;     /* Operands as voxel arrays */
    pyvox_array *dest;            /* Result as a Pyvox array */
    voxel_array *vdest;           /* Result as a voxel array */
    int flag;

    set_error_context("array.cross()");

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "O", &arg1) )
	return NULL;

    /* Extract the voxel arrays themselves */
    if ( ! PyvoxArray_Check(self, "self",  &src1) || 
	 ! PyvoxArray_Check(arg1, "other", &src2) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the actual computation */
    dest = parray_create(&vdest);
    if (dest == 0)    ReturnPyvoxLibError();
    flag = vxl_cross(vdest, src1, src2);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)dest;
}



/*-------------------------------------------------------------------------
parray_hypot - Compute sqrt(x*x+y*y) of two voxel arrays

This Pyvox function computes the hypot function of two numeric voxel
arrays.  The first array is self; the second is passed as the method
argument.
-------------------------------------------------------------------------*/

#ifdef HAVE_HYPOT
static PyObject *
parray_hypot(PyObject *self, PyObject *args)
{
    set_error_context("array.hypot()");
    return parray_vxl_binop(vxl_hypot, INTP_DOUBLE, self, args);
}
#endif



/*-------------------------------------------------------------------------
parray_cmove - Conditional move from one array to another

   array.cmove(select, source)

This Pyvox array method takes a destination, selector, and source
array as operands.  For each element of the selector that is non-zero,
the corresponding element of the source is copied to the corresponding
element of the destination.  The destination, selector, and source
must have the same rank, dimensions, origin, and spacing.  The
selector must be unsigned char; the destination and source may be any
type but must be the same type.  Note that this function is atypical
in that the destination must already exist and be compatible with the
source array.  The modified destination array is returned.

FIXME: Here and elsewhere, it might not be necessay to free a
temporary scalar array if the array header is put on the stack and
just points to the existing scalar value.
-------------------------------------------------------------------------*/

static 
PyObject *                  /* Returns modified destination or NULL on error */
parray_cmove(
    PyObject *self,         /* Self: Destination to be modified in place */
    PyObject *args)
{
    PyObject *pysel;        /* Arg 1: Selector array */
    PyObject *pysrc;        /* Arg 2: Source array */

    voxel_array *dest = 0;               /* Destination as a voxel array */
    voxel_array *src1 = 0, *src2 = 0;    /* Sources as voxel arrays */
    int type;                            /* Type of destination */
    int free2 = 0;                       /* Should src2 be freed? */
    int flag;

    set_error_context("array.cmove()");

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "OO", &pysel, &pysrc) )
	return NULL;

    /* DEBUG: Show the argument arrays */
    if (0) {
	PyvoxArray_Dump(self,  "cmove dest",   0);
	PyvoxArray_Dump(pysel, "cmove select", 0);
	PyvoxArray_Dump(pysrc, "cmove source", 0);
    }

    /* Get voxel arrays underneath the arguments; promote a numeric
       source 2 to a scalar array if necessary */
    if ( ! PyvoxArray_Check(self, "Array", &dest) )
	ReturnPyvoxLibError();
    if ( ! PyvoxArray_Check(pysel, "Selector", &src1) )
	ReturnPyvoxLibError();
    if ( PyvoxArray_Check(pysrc, "Source", &src2) )
	/* Do nothing */ ;
    else if ( PyNumber_Check(pysrc) ) {
	type = dest->type;
        src2 = varray_create_scalar(type, pysrc);
        if (src2 == 0)    ReturnPyvoxLibError();
	free2 = 1;
    }
    else 
	ReturnPyvoxError(ERR_INIT, "Source must be an array or a number");

    /* Call Voxel Kit to do the actual work */
    flag = vxl_cmove(dest, src1, src2);

    /* Free the scalar array, if needed */
    if (free2 != 0) vxl_destroy_array(src2);

    /* Return result or error to caller */
    if (flag)    ReturnPyvoxLibError();
    clear_error_context();
    Py_INCREF(self);
    return self;
}



/*-------------------------------------------------------------------------
parray_compare - Compare self with another Pyvox array

This Pyvox array method compares self element by element with another
Pyvox array and returns a new unsigned char voxel array object
containing the results of the comparison; each element of the result
is set to one of three given values according to whether the result of
the comparison is less than, equal to, are greater than.

FIXME: Generalize to let lsval, eqval, and gtval be arrays as well as
numbers?  Or should this function do just comparison, with another
function for selection?

FIXME: Consider adding a method array.equal() for equality comparison
only, which can also support complex arrays.  Alternatively, allow for
complex types, but only if lsval = gtval; if so, make the more argument
optional.

FIXME: This method is much less useful now that we have rich
comparisons; consider deprecating or deleting?
-------------------------------------------------------------------------*/

static 
PyObject *                  /* Returns result, or NULL on error */
parray_compare(
    PyObject *self,         /* Self: First image to compare */
    PyObject *args)
{
    PyObject *pysrc2;       /* Arg 1: Second image to compare */
    int lsval;              /* Arg 2: Output value if less than */
    int eqval;              /* Arg 3: Output value if equal */
    int gtval;              /* Arg 4: Output value if greater than */

    voxel_array  *src1, *src2;           /* Sources as voxel arrays */
    int free1 = 0, free2 = 0;            /* Should src1 & src2 be freed? */
    pyvox_array *vocomp;                 /* Result as a Pyvox array */
    voxel_array  *comp;                  /* Result as a voxel array */
    int flag;

    set_error_context("array.compare()");

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "Oiii", &pysrc2, 
			  &lsval, &eqval, &gtval) )
	return NULL;

    /* Promote numeric arguments to scalar arrays if necessary */
    if ( !setup_vxl_binpt(self, pysrc2, INTP_DOUBLE,
			  &src1, &free1, &src2, &free2) )
	ReturnPyvoxLibError();

    /* Not defined for complex types */
    if ( exim_iscomplex(src1->type) || exim_iscomplex(src2->type) )
        ReturnPyvoxError(ERR_ILLOP, "Not defined for complex types");

    /* Call Voxel Kit to do the actual comparison */
    vocomp = parray_create(&comp);
    if (vocomp == 0)    ReturnPyvoxLibError();
    flag = vxl_compare(comp, src1, src2, lsval, eqval, gtval);

    /* Free the scalar arrays, if needed */
    if (free1 != 0) vxl_destroy_array(src1);
    if (free2 != 0) vxl_destroy_array(src2);

    /* Return result or error to caller */
    if (flag)
        ReturnPyvoxLibError();
    clear_error_context();
    return (PyObject *)vocomp;
}



/*-------------------------------------------------------------------------
parray_max - Elementwise maximum of self and another Pyvox array

This Pyvox array method computes the elementwise maximum of self and
another Pyvox arrays; that is, it computes dest = max(x, y) for each
element.
-------------------------------------------------------------------------*/

static PyObject *
parray_max(PyObject *self, PyObject *args)
{
    set_error_context("array.max()");
    return parray_vxl_binop(vxl_max, INTP_DOUBLE, self, args);
}



/*-------------------------------------------------------------------------
parray_min - Elementwise minimum of self and another Pyvox array

This Pyvox array method computes the elementwise minimum of self and
another Pyvox array; that is, it computes dest = min(x, y) for each
element.
-------------------------------------------------------------------------*/

static PyObject *
parray_min(PyObject *self, PyObject *args)
{
    set_error_context("array.min()");
    return parray_vxl_binop(vxl_min, INTP_DOUBLE, self, args);
}



/*-------------------------------------------------------------------------
parray_pow - Power function (exponentiation) of two voxel arrays

This Pyvox array method computes the elementwise power function of
self and another Pyvox array; that is, it computes dest = pow(x, y)
for each element.
-------------------------------------------------------------------------*/

static PyObject *
parray_pow(PyObject *self, PyObject *args)
{
    set_error_context("array.pow()");
    return parray_vxl_binop(vxl_pow, INTP_DOUBLE, self, args);
}



/*---------------------------------------------------------------------
parray_richcmp - Elementwise comparison of arrays

This function implements the rich comparison of Pyvox arrays.  The
result is an unsigned char array the same shape as the arguments,
each element of which contains 1 if the comparison of the corresponding
elements is true, and 0 otherwise.

FIXME: Implement and use bips_lt() and bips_gt().
----------------------------------------------------------------------*/

#ifdef Py_TPFLAGS_HAVE_RICHCOMPARE
static
PyObject *                       /* Returns NULL on error */
parray_richcmp(
     PyObject *pysrc1,           /* First comparand */
     PyObject *pysrc2,           /* Second comparand */
     int op)                     /* Relationship to test */
{
    voxel_array  *src1, *src2;   /* Operands as voxel arrays */
    pyvox_array  *result;        /* Result as a Pyvox array */
    voxel_array  *rarray;        /* Result's voxel array */
    vxl_comp2 *funct=0;          /* Comparison function */
    int iftrue;                  /* Sense of comparison */
    int free1=0, free2=0;        /* Free scalar arrays after use? */
    int flag;

    /* Decode relationship type to necessary parameters */
    switch (op) {

    case Py_EQ :
        set_error_context("array == array");
        funct  = vxl_equal;
        iftrue = 1;
        break;

    case Py_NE :
        set_error_context("array != array");
        funct  = vxl_equal;
        iftrue = 0;
        break;

    case Py_LT :
        set_error_context("array < array");
        funct  = vxl_less;
        iftrue = 1;
        break;

    case Py_LE :
        set_error_context("array <= array");
        funct  = vxl_greater;
        iftrue = 0;
        break;

    case Py_GE :
        set_error_context("array >= array");
        funct  = vxl_less;
        iftrue = 0;
        break;

    case Py_GT :
        set_error_context("array > array");
        funct  = vxl_greater;
        iftrue = 1;
        break;

    default :
        ReturnPyvoxError(ERR_ASSERT, "Invalid op for array comparison");

    }

    /* Promote numeric arguments to scalar arrays if necessary */
    if ( ! setup_vxl_binpt(pysrc1, pysrc2, INTP_DOUBLE,
                           &src1, &free1, &src2, &free2) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the actual computation */
    result = parray_create(&rarray);
    if (result == 0)    ReturnPyvoxLibError();
    flag = funct(result->varray, src1, src2, iftrue, !iftrue);
    if (flag)    ReturnPyvoxLibError();

    /* Free the scalar arrays, if needed */
    if (free1 != 0) vxl_destroy_array(src1);
    if (free2 != 0) vxl_destroy_array(src2);

    clear_error_context();
    return (PyObject *)result;
}
#endif



/**************************************************************************
Simple Ternary Point Operations on Pyvox Arrays 

These operations correspond directly to Voxel Kit operations and can
be handled via simple wrapper functions.

FIXME: This is a somewhat elastic category and may need some refinement.
Note also that some methods may still need to be modified to use the
standard wrapper functions.
**************************************************************************/



/*-------------------------------------------------------------------------
setup_vxl_terpt - Set up a ternary point operation in the Voxel Kit

Given three Python objects which may be either numbers or voxel array
objects, this function promotes numbers into scalar voxel array
objects so that the Voxel Kit may be called in all cases.  It is an
error to provide two voxel arrays of different types; casting is NOT
automatic as it is in C.  A number is promoted to a scalar array of
the same type as the other object; if three numbers are provided, then
they are both promoted to a default type specified as an argument.
Shape compatibility will be checked within the Voxel Kit itself.
Promoting a number yields a temporary scalar array which must be freed
once the operation is complete; the free1, free2, and free3 flags
indicate which must be freed.  This function returns non-zero on
failure.

NOTE that this function is inappropriate for those arithmetic
operations for which the Python type coercion method is automatically
invoked; see parray_and for an example.

FIXME: Review for improvements that could be made here or in
setup_vxl_binpt().

FIXME: Figure out whether the case of all numbers is ever used, and, if
so, how it should be handled.

FIXME: When used in a pyvox class method, this function permits either
operand to be a scalar.  This is no longer possible when used as a
Pyvox array method.  Do I need to provide a means to compute binary
operations when the first operand is a number?  If so, how?  One idea
is to add a method pyvox.scalar which take a number and converts it to
a scalar array; however, this might not coerce to the right type.
The usage might be something like "pyvox.scalar(0.39).atan2(array)".
Another option might be for the methods to take either one or two
arguments: if one arg, then self and arg 1 are combined; if two args,
then args 1 and 2 are combined.
------------------------------------------------------------------------*/

static inline
int                              /* Returns non-zero on error */
setup_vxl_terpt(
    PyObject *pysrc1,            /* First operand, as Python object */
    PyObject *pysrc2,            /* Second operand, as Python object */
    PyObject *pysrc3,            /* Third operand, as Python object */
    int deftype,                 /* Default type if both numbers */
    voxel_array **src1,          /* First operand, as voxel array */
    int *free1,                  /* Should src1 be freed after use? */
    voxel_array **src2,          /* Second operand, as voxel array */
    int *free2,                  /* Should src2 be freed after use? */
    voxel_array **src3,          /* Third operand, as voxel array */
    int *free3)                  /* Should src3 be freed after use? */
{
    int type = 0;                /* Common type of voxel arrays */
    voxel_array *src;            /* A voxel array */

    /* Determine the common type, or take the default */
    if  (  PyvoxArray_Check(pysrc1, 0, &src) || 
           PyvoxArray_Check(pysrc2, 0, &src) || 
           PyvoxArray_Check(pysrc3, 0, &src)   )
        type = src->type;
    else
        type = deftype;

    /* Set up the first argument */
    /* FIXME: Make into a macro; look for further improvements */
    if ( PyvoxArray_Check(pysrc1, 0, src1) ) {
        *free1 = 0;
    }
    else if ( PyNumber_Check(pysrc1) ) {
	*src1= varray_create_scalar(type, pysrc1);
        if (*src1)    NZReturnLibError();
        *free1 = 1;
    }
    else {
        NZReturnError(ERR_INIT, "Argument must be array or number");
    }

    /* Set up the second argument */
    if ( PyvoxArray_Check(pysrc2, 0, src2) ) {
        *free2 = 0;
    }
    else if ( PyNumber_Check(pysrc2) ) {
	*src2 = varray_create_scalar(type, pysrc2);
        if (*src2)    ReturnPyvoxLibError();
        *free2 = 1;
    }
    else {
        NZReturnError(ERR_INIT, "Argument must be array or number");
    }

    /* Set up the third argument */
    if ( PyvoxArray_Check(pysrc3, 0, src3) ) {
        *free3 = 0;
    }
    else if ( PyNumber_Check(pysrc3) ) {
        *src3 = varray_create_scalar(type, pysrc3);
        if (*src3)    ReturnPyvoxLibError();
        *free3 = 1;
    }
    else {
        NZReturnError(ERR_INIT, "Argument must be array or number");
    }

    return 0;
}



/*-------------------------------------------------------------------------
parray_vxl_ternop - Pyvox array method wrapper for ternary Voxel Kit function

This wrapper function is used to translate many common ternary Voxel
Kit functions into a two-argument method for a Pyvox array.  This
function assumes the error context has already been set but clears it
before returning.

FIXME: Could the scalar args be predefined as automatic variables so that
they do not have to be allocated and destroyed?
--------------------------------------------------------------------------*/

static inline
PyObject *                        /* Returns result, or NULL on error */
parray_vxl_ternop(
    vxl_ternop  *vxl_func,        /* Voxel Kit function to apply */
    int        deftype,           /* Default type for coercion */
    PyObject   *self,             /* Self: First operand */
    PyObject   *args)
{
    PyObject *arg1;               /* Arg 1: Second operand */
    PyObject *arg2;               /* Arg 2: Third operand */

    voxel_array *src1=0, *src2=0, *src3=0; /* Operands as voxel arrays */
    int free1 = 0, free2 = 0, free3 = 0;   /* Should src1 & src2 be freed? */
    pyvox_array  * dest;                   /* Result as a Pyvox array */
    voxel_array  *vdest;                   /* Result as a voxel array */
    int flag;

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "OO", &arg1, &arg2) )
	return NULL;

    /* Promote numeric arguments to scalar arrays if necessary */
    if ( setup_vxl_terpt(self, arg1, arg2, INTP_DOUBLE,
                         &src1, &free1, &src2, &free2, &src3, &free3) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the actual computation */
    dest = parray_create(&vdest);
    if (dest == 0)    ReturnPyvoxLibError();
    flag = vxl_func(vdest, src1, src2, src3);
    if (flag)    ReturnPyvoxLibError();

    /* Free the scalar arrays, if needed */
    if (free1 != 0) vxl_destroy_array(src1);
    if (free2 != 0) vxl_destroy_array(src2);
    if (free3 != 0) vxl_destroy_array(src3);

    clear_error_context();
    return (PyObject *)dest;
}



/*-------------------------------------------------------------------------
parray_fma - Fused multiply-add

This Pyvox function computes the elementwise fused multiply-add for
three numeric voxel arrays; that is, it computes dest = x*y + z for
each element.  The first array is self; the second and third are
passed as the method arguments.
-------------------------------------------------------------------------*/

static PyObject *
parray_fma(PyObject *self, PyObject *args)
{
    set_error_context("array.fma()");
    return parray_vxl_ternop(vxl_fma, INTP_DOUBLE, self, args);
}




/*************************************************************************
Other Instance Methods for Pyvox Arrays

FIXME: Add functions to change the type of a voxel array.
This is, of course, not guaranteed to be portable.
*************************************************************************/



/*-------------------------------------------------------------------------
parray_amax - Find maximum value in voxel array


This method finds the largest value in the voxel array.

See also array.max() to find the elementwise maximum of two arrays.
-------------------------------------------------------------------------*/

static PyObject *
parray_amax(
    PyObject *self,             /* Self: array to take means of */
    PyObject *args)
{
    PyObject *paxes  = 0;       /* Arg 1: axes to reduce on */
    PyObject *result = 0;       /* Result: Reduced array */

    voxel_array *vres  = 0;     /* Result as a voxel array */
    voxel_array *vself = 0;     /* Self as a voxel array */
    int naxes = -1;             /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];     /* Specific axes to reduce on */
    int flag;

    set_error_context("array.amax()");

    /* Parse the Python arguments into C values */
    if  ( ! PyArg_ParseTuple(args, "|O", &paxes) )
	return NULL;

    /* Extract voxel array */
    if  ( ! PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxLibError();

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, vself->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work, and convert to a Python object */
    vres = vxl_create_array();
    if (vres == 0)    ReturnPyvoxLibError();
    flag = vxl_amax(vres, vself, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    result = pyvox_from_temp_varray(vres);
    if (result == 0)    ReturnPyvoxLibError();

    /* Return the result */
    clear_error_context();
    return result;
}



/*-------------------------------------------------------------------------
parray_amin - Find minimum value in a voxel array

This method finds the smallest value in the voxel array.

See also array.min() to find the elementwise minimum of two arrays.
-------------------------------------------------------------------------*/

static PyObject *
parray_amin(
    PyObject *self,             /* Self: array to take means of */
    PyObject *args)
{
    PyObject *paxes  = 0;       /* Arg 1: axes to reduce on */
    PyObject *result = 0;       /* Result: Reduced array */

    voxel_array *vres  = 0;     /* Result as a voxel array */
    voxel_array *vself = 0;     /* Self as a voxel array */
    int naxes = -1;             /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];     /* Specific axes to reduce on */
    int flag;

    set_error_context("array.amin()");

    /* Parse the Python arguments into C values */
    if  ( ! PyArg_ParseTuple(args, "|O", &paxes) )
	return NULL;

    /* Extract voxel array */
    if  ( ! PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxLibError();

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, vself->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work */
    vres = vxl_create_array();
    if (vres == 0)    ReturnPyvoxLibError();
    flag = vxl_amin(vres, vself, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    result = pyvox_from_temp_varray(vres);
    if (result == 0)    ReturnPyvoxLibError();

    /* Return the result */
    clear_error_context();
    return result;
}



/*-------------------------------------------------------------------------
parray_aprod - Find sum of elements in voxel array

This method finds the product of the element values in the voxel array.

FIXME: Return a double result for integral arrays?
-------------------------------------------------------------------------*/

static PyObject *
parray_aprod(
    PyObject *self,             /* Self: array to reduce */
    PyObject *args)
{
    PyObject *paxes  = 0;       /* Arg 1: axes to reduce on */
    PyObject *result = 0;       /* Result: Reduced array */

    voxel_array *vres  = 0;     /* Result as a voxel array */
    voxel_array *vself = 0;     /* Self as a voxel array */
    int naxes = -1;             /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];     /* Specific axes to reduce on */
    int flag;

    set_error_context("array.aprod()");

    /* Parse the Python arguments into C values */
    if  ( ! PyArg_ParseTuple(args, "|O", &paxes) )
	return NULL;

    /* Extract voxel array */
    if  ( ! PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxLibError();

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, vself->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work */
    vres = vxl_create_array();
    if (vres == 0)    ReturnPyvoxLibError();
    flag = vxl_aprod(vres, vself, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    result = pyvox_from_temp_varray(vres);
    if (result == 0)    ReturnPyvoxLibError();

    /* Return the result */
    clear_error_context();
    return result;
}



/*-------------------------------------------------------------------------
parray_asum - Find sum of elements in voxel array

This method finds the sum of the element values in the voxel array.

FIXME: Return a double result for integral arrays?
-------------------------------------------------------------------------*/

static PyObject *
parray_asum(
    PyObject *self,             /* Self: array to reduce */
    PyObject *args)
{
    PyObject *paxes  = 0;       /* Arg 1: axes to reduce on */
    PyObject *result = 0;       /* Result: Reduced array */

    voxel_array *vres  = 0;     /* Result as a voxel array */
    voxel_array *vself = 0;     /* Self as a voxel array */
    int naxes = -1;             /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];     /* Specific axes to reduce on */
    int flag;

    set_error_context("array.asum()");

    /* Parse the Python arguments into C values */
    if  ( ! PyArg_ParseTuple(args, "|O", &paxes) )
	return NULL;

    /* Extract voxel array */
    if  ( ! PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxLibError();

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, vself->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work */
    vres = vxl_create_array();
    if (vres == 0)    ReturnPyvoxLibError();
    flag = vxl_asum(vres, vself, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    result = pyvox_from_temp_varray(vres);
    if (result == 0)    ReturnPyvoxLibError();

    /* Return the result */
    clear_error_context();
    return result;
}



/*-------------------------------------------------------------------------
parray_asum0 - Find sum of elements in voxel array

This is a slow but obviously correct version of parray_asum used for
testing.
-------------------------------------------------------------------------*/

static PyObject *
parray_asum0(
    PyObject *self,             /* Self: array to reduce */
    PyObject *args)
{
    PyObject *paxes  = 0;       /* Arg 1: axes to reduce on */
    PyObject *result = 0;       /* Result: Reduced array */

    voxel_array *vres  = 0;     /* Result as a voxel array */
    voxel_array *vself = 0;     /* Self as a voxel array */
    int naxes = -1;             /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];     /* Specific axes to reduce on */
    int flag;

    set_error_context("array.asum0()");

    /* Parse the Python arguments into C values */
    if  ( ! PyArg_ParseTuple(args, "|O", &paxes) )
	return NULL;

    /* Extract voxel array */
    if  ( ! PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxLibError();

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, vself->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work */
    vres = vxl_create_array();
    if (vres == 0)    ReturnPyvoxLibError();
    flag = vxl_asum0(vres, vself, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    result = pyvox_from_temp_varray(vres);
    if (result == 0)    ReturnPyvoxLibError();

    /* Return the result */
    clear_error_context();
    return result;
}



/*-------------------------------------------------------------------------
parray_bigob - Remove all but the biggest object in a binary image

This instance method for Pyvox arrays aggregates all the non-zero
voxels of self into objects (maximal connected sets of non-zero voxels),
counts the number of voxels in each object, and returns an unsigned
char mask covering only the largest object.  (FIXME: Suppose there are
two or more biggest objects?)

This function was deprecated on 2005-11-23.  Use pyvox.ccmap(image).bigob() 
in new code, or take advantage of additional functionality in the ccmap object.

NOTE:  This is a temporary version for testing and interim use.  It will
probably be replaced by several functions for doing individual steps of
the process, once I figure out exactly how to map object lists and 
operations on them into Python.
-------------------------------------------------------------------------*/

static PyObject *        /* Result: Volume image of biggest object only */
parray_bigob(
    PyObject *self,      /* Self: Volume image to process */
    PyObject *args)
{
    int lab1 = 255;      /* Arg 1: (Optional) label for biggest object */
    int lab2 = 0;        /* Arg 2: (Optional) label for all other objects */

    pyvox_array *in, *out;    /* Input and output as Pyvox arrays */
    voxel_array  *vin, *vout; /* Input and output as voxel arrays */
    voxel_array  *labels;     /* Image of voxel labels */
    voxel_array  *table;      /* Look up table to convert labels to mask */
    unsigned char *td;        /* Data contained in table */
    struct voxel_object_data *oblist;   /* Table of objects found */
    long nobs;                /* Number of objects found */
    int ibrain;               /* Canonical object id of brain */
    long vbrain;              /* Number of voxels in brain object */
    int i;
    int flag;
#if 0      /* Needed only for debugging output */
    int ncan;                 /* Number of canonical objects found */
    int n;
#endif

    set_error_context("array.bigob()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "|ii", &lab1, &lab2) )
	return NULL;

    /* Find and verify the input voxel array itself */
    if ( !PyvoxArray_Check(self, "self", &vin) )
	ReturnPyvoxLibError();
    in = (pyvox_array *)self;

    /* Scan the input to do preliminary labels and create the object table */
    labels = vxl_create_array();
    if (labels == 0)    ReturnPyvoxLibError();
    flag = vxl_find_objects_uchar3d(labels, &nobs, &oblist, vin);
    if (flag != 0)    ReturnPyvoxLibError();

    /* DEBUG: Report the first 50 objects found */
#if 0
    fprintf(stderr, "\nFound %ld objects in this image\n", nobs);
    for (i = 0; i < 50 && i < nobs; i++) {
	fprintf(stderr, "%6hu (%6hu): %10ld voxels at (%4ld, %4ld, %4ld)\n",
		oblist[i].ident, oblist[i].canon, oblist[i].count,
		oblist[i].point[0], oblist[i].point[1], oblist[i].point[2]);
    }
#endif

    /* DEBUG: Report the first 50 large canonical objects found. */
#if 0
    ncan = 0;
    for (i = 0; i < nobs; i++)  
	if (oblist[i].canon == i)  ncan ++;
    fprintf(stderr, "\nFound %d canonical objects in this image\n",
	    ncan);
    for ( n = 0, i = 0;
	  n < 50 && i < nobs;
	  i++ ) {
	if (oblist[i].canon == i && oblist[i].count > 100) {
	    n++;
	    fprintf(stderr, 
		    "%6hu (%6hu): %10ld voxels at (%4ld, %4ld, %4ld)\n",
		    oblist[i].ident, oblist[i].canon, oblist[i].count,
		    oblist[i].point[0], oblist[i].point[1],
		    oblist[i].point[2]);
	}
    }
#endif

    /* Find the largest non-background object and call it brain.
       FIXME:  Suppose there are *two* largest objects? */
    ibrain = 0;
    vbrain = 0;
    for (i = 1; i < nobs; i++)
	if (oblist[i].count > vbrain) {
	    ibrain = oblist[i].canon;
	    vbrain = oblist[i].count;
	}
    if (ibrain == 0)
	ReturnPyvoxError(ERR_ALGOR, 
                         "Couldn't find any non-background objects");
    if (0)  /* DEBUG */
	fprintf(stderr, "The brain object has index %d\n", ibrain);

    /* Create a lookup table to convert brain to 255, others to 0 */
    table = vxl_create_array();
    if (table == 0)    ReturnPyvoxLibError();
    flag = vxl_alloc_array(table, INTP_UCHAR, 1, &nobs);
    if (flag)    ReturnPyvoxLibError();
    td = table->data;
    td[0] = 0;
    for (i = 1; i < nobs; i++) {
	if (oblist[i].canon == ibrain)
	    td[i] = lab1;
	else
	    td[i] = lab2;
    }

    /* DEBUG:  Print first 50 entries of the lookup table */
#if 0
    for (i = 0; i < 50; i++)
	fprintf(stderr, "%5d -> %3d\n", i, td[i]);
#endif
    
    /* Create mask by processing labels image through the lookup table */
    out = parray_create(&vout);
    if (out == 0)    ReturnPyvoxLibError();
    flag = vxl_lookup(vout, labels, table);
    if (flag)    ReturnPyvoxLibError();

    /* Free temporary storage */
    vxl_destroy_array(table);
    free(oblist);
    vxl_destroy_array(labels);

    clear_error_context();
    return (PyObject *)out;
}



/*-------------------------------------------------------------------------
parray_bihist - Compute bivariate histogram of two unsigned char images

This instance method for Pyvox arrays computes the bivariate histogram
and marginal histograms of self and another Pyvox array, both of which
must be the same shape and type unsigned char.  An optional unsigned
char weight image may be provided, which must the the same shape as
the source image.  The bivariate histogram is returned as a
two-dimensional unsigned long voxel arrays; the two marginal
histograms (which are the univariate histograms of the individual
images) are returned as one-dimensional unsigned long voxel arrays.
These results are returned as a tuple containing the bivariate
histogram and the two univariate histograms.
-------------------------------------------------------------------------*/

static PyObject *
parray_bihist(
    PyObject *self,             /* Self: First image */
    PyObject *args)
{
    PyObject *src2;             /* Arg 1: Second image */
    PyObject *weight = 0;       /* Arg 2: Weight image */

    pyvox_array *hist;          /* Result 1: Bivariate histogram */
    pyvox_array *mar1;          /* Result 2: Univariate histogram of 1st img */
    pyvox_array *mar2;          /* Result 3: Univariate histogram of 2nd img */

    voxel_array *vhist, *vmar1, *vmar2;  /* Results as voxel arrays */
    voxel_array *s1, *s2, *w = 0;  /* Arguments as voxel arrays */
    PyObject *result;              /* Python tuple containing the results */
    int flag;

    set_error_context("array.bihist()");

    /* Fetch and verify the arguments */
    /* FIXME: Several distinct errors are gathered into each message here */
    if ( !PyArg_ParseTuple(args, "O|O", &src2, &weight) )
	return NULL;
    if ( !PyvoxArray_Check(self, "source 1", &s1) || s1->type != INTP_UCHAR )
	ReturnPyvoxError(ERR_INIT, 
                         "Self is not an unsigned char voxel array");
    if ( !PyvoxArray_Check(src2, "source 2", &s2) || s2->type != INTP_UCHAR)
	ReturnPyvoxError(ERR_INIT, 
                         "Source 2 is not an unsigned char voxel array");
    if ( weight != 0 && weight != Py_None ) {
	if ( !PyvoxArray_Check(weight, "weight", &w) || w->type != INTP_UCHAR )
	    ReturnPyvoxError(ERR_INIT,
                             "Weight is not an unsigned char voxel array");
    }

    /* Create Pyvox arrays for the output */
    hist = parray_create(&vhist);
    mar1 = parray_create(&vmar1);
    mar2 = parray_create(&vmar2);
    if (hist == 0 || mar1 == 0 || mar2 == 0)    ReturnPyvoxLibError();

    /* Compute the histograms */
    flag = vxl_bihist_uchar(vhist, vmar1, vmar2, s1, s2, w);
    if (flag)    ReturnPyvoxLibError();

    /* Form results into a tuple and return them */
    result = Py_BuildValue("(NNN)", hist, mar1, mar2);
    clear_error_context();
    return result;
}



/*-------------------------------------------------------------------------
parray_chamfer - Compute chamfer distance transform of an image

This instance method computes the chamfer distance image of the given
image.  The chamfer distance is defined to be zero wherever the original
image is non-zero, and to be equal to the taxicab (L1) distance to the
nearest non-zero voxel otherwise.  The original image may be of any
rank greater than zero, and any size and type.  The distance image may
be either unsigned char or unsigned short, as specified by the caller, or
may be defaulted to the shortest type able to contain the maximum possible
distance.
---------------------------------------------------------------------------*/

static PyObject *
parray_chamfer(
    PyObject *self,             /* Self: the original image */
    PyObject *args)
{
    PyObject *type = 0;         /* Arg 1: Requested destination type */
    pyvox_array *result;        /* Result: Chamfer distance image */

    int dtype = INTP_NONE;      /* Requested destination type */
    voxel_array *vself;         /* Self as a voxel array */
    voxel_array *vres;          /* Result as a voxel array */
    int flag;

    set_error_context("array.chamfer()");

    /* Fetch the argument, if any */
    if ( !PyArg_ParseTuple(args, "|O", &type) )
	return NULL;

    /* Find the input voxel array */
    if ( !PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxLibError();

    /* Verify the desired type of the result */
    if ( type && !PyvoxIntype_Check(type, "result type", &dtype) )
        ReturnPyvoxLibError();

    /* Compute and return the result */
    result = parray_create(&vres);
    if (result == 0)    ReturnPyvoxLibError();
    flag = vxl_chamfer(vres, vself, dtype);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)result;
}



/*------------------------------------------------------------------------
parray_cholesky - Compute Cholesky decomposition of a matrix

This instance method computes the Cholesly decomposition of a
symmetric or Hermitian positive definite square matrix represented as
a Pyvox array of floating-point type.
-------------------------------------------------------------------------*/

static
PyObject *
parray_cholesky(
    PyObject *self,         /* Pyvox array */
    PyObject *args)         /* Arguments (none) */
{
    voxel_array *vself;     /* Self as a voxel array */
    pyvox_array *fact;      /* Factor as a Pyvox array */
    voxel_array *vfact;     /* Factpr as a voxel array */
    int flag;

    set_error_context("array.cholesky()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        ReturnPyvoxLibError();

    /* Create Pyvox array for result */
    fact = parray_create(&vfact);
    if (fact == 0)    ReturnPyvoxLibError();

    /* Compute the Cholesky factorization */
    flag = matrix_cholesky(vfact, vself);
    if (flag)    ReturnPyvoxLibError();

    /* Return the result as a Pyvox array */
    clear_error_context();
    return Py_BuildValue("N", fact);
}



/*------------------------------------------------------------------------
parray_conjtrans - Conjugate transpose of a matrix

This instance method computes the complex conjugate (Hermitian)
transpose of a matrix represented as a Pyvox array of any type.

FIXME: Generalize to allow any rank, any permutation of axes, or reversal
of indices along an axis.  Find or invent a fast algorithm for this.
-------------------------------------------------------------------------*/

static
PyObject *
parray_conjtrans(
    PyObject *self,        /* Pyvox array */
    PyObject *args)        /* Arguments (none) */
{
    voxel_array *vself;     /* Self as a voxel array */
    pyvox_array *res;       /* Result as a Pyvox array */
    voxel_array *vres;      /* Result as a voxel array */
    int flag;               /* Error flag */

    set_error_context("array.trans()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        ReturnPyvoxLibError();

    /* Create Pyvox array for result */
    res = parray_create(&vres);
    if (res == 0)    ReturnPyvoxLibError();

    /* Compute the transpose */
    flag = matrix_transpose(vres, vself, 1);
    if (flag)    ReturnPyvoxLibError();

    /* Return the result as a Pyvox array */
    clear_error_context();
    return Py_BuildValue("N", res);
}



/*-------------------------------------------------------------------------
parray_convolve - Convolve an image with a convolution kernel

This instance method for Pyvox arrays convolves self, which may be of
any non-zero rank, dimensions, or type, with a convolution kernel of
the same rank, and optionally subsamples the resulting image using a
given integral shrink factor along each dimension.  (Pixel values not
included in the subsampled image are never calculated, of course.)
The calculations are done in double or double precision and converted
back to the original image type; if the original type cannot represent
the values, the nearest representable value is used instead.
Convolving a real image with a complex kernel is not supported.

FIXME: Add some other useful filters, such as high pass, gradient
(Sobel) and Laplacian.  (Note that the derivative filters require a
bias that depends on the data type.)  Add functions to construct
various useful kernels.
-------------------------------------------------------------------------*/

static PyObject *
parray_convolve(
    PyObject *self,             /* Self: Volume image to process */
    PyObject *args)
{
    PyObject *pykern;           /* Arg 1: Convolution kernel to use */
    PyObject *pyshrink=0;       /* Arg 2: Shrink factors */

    pyvox_array *output;        /* Result as a Pyvox array */
    voxel_array *in, *out;      /* The voxel arrays themselves */
    struct vxl_kernel *kern;    /* The kernel struct itself */
    long shrink[VXL_MAX_RANK];  /* Shrink factors */
    int i, flag;

    set_error_context("array.convolve()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "O|O", &pykern, &pyshrink) )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, "self", 0) )
	ReturnPyvoxLibError();
    in = ((pyvox_array *)self)->varray;

    /* Find the kernel struct itself */
    if ( !PyvoxKernel_Check(pykern, "kernel", &kern) )
        ReturnPyvoxLibError();
    if (in->rank != kern->rank)
	ReturnPyvoxError(ERR_INCOMP, "Image and kernel have different ranks");

    /* Find the shrink factors, if any */
    if ( pyshrink == 0 ) {
        for (i = 0; i < in->rank; i++)
            shrink[i] = 1;
    }
    else if ( PyvoxLong_Check(pyshrink, &shrink[0]) ) {
        for (i = 1; i < in->rank; i++)
            shrink[i] = shrink[0];
    }
    else if ( PyvoxLongSeq_Check(pyshrink, in->rank, in->rank, 0, shrink) )
        /* Do nothing */ ;
    else
        ReturnPyvoxError(ERR_INIT, "Unable to interpret shrink argument");
  
    /* Compute and return the result */
    output = parray_create(&out);
    if (output == 0)    ReturnPyvoxLibError();
    flag = vxl_convolve(out, in, kern, shrink);
    if (flag != 0)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)output;
}



/*-----------------------------------------------------------------------
parray_count - Total number of voxels in a Pyvox array

This Pyvox array method returns the total number of voxels in a Pyvox
array.
-------------------------------------------------------------------------*/

static
PyObject *
parray_count(
    PyObject *self,          /* Pyvox array */
    PyObject *args)          /* Arguments (none) */
{
    voxel_array *vself;      /* Self as a voxel array */
    long count;              /* Number of voxels */

    set_error_context("array.count()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_CheckSrc(self, "self", &vself) )
        ReturnPyvoxLibError();

    /* Determine and return the count */
    count = vxli_count(vself);
    clear_error_context();
    return Py_BuildValue("l", count);
}



/*------------------------------------------------------------------------
parray_cubic - Cubic interpolation in a voxel array

This instance method for Pyvox arrays returns a new array whose
elements are computed by cubic Lagrange interpolation in the voxels of
self at the sample points defined by the map argument.
-------------------------------------------------------------------------*/

static PyObject *
parray_cubic(
    PyObject *self,                 /* Pyvox array to interpolate in */
    PyObject *args)                 /* Other arguments */
{
    PyObject *map;                  /* Arg 1: Points to evaluate at */
    PyObject *type = 0;             /* Arg 2: Desired type of the result */
    PyObject *result = 0;           /* Result */

    voxel_array *vself;             /* Self as voxel array */
    int rank;                       /* Rank of self */
    int stype;                      /* Type of self */
    voxel_array *vmap;              /* Points to evaluate at */
    voxel_array *vres;              /* Result as voxel array */
    int dtype;                      /* Desired type of result */
    double point[VXL_MAX_RANK];     /* A single point to evaluate at */
    anytype value;                  /* Value at the given point */
    int flag;

    set_error_context("array.cubic()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_CheckSrc(self, "self", &vself) )
        ReturnPyvoxLibError();
    rank  = vself->rank;
    stype = vself->type;

    /* Parse the arguments */
    if ( ! PyArg_ParseTuple(args, "O|O", &map, &type) )  
	return NULL;

    /* Verify the array type; default to type of source */
    if ( type == 0 )
        dtype = stype;
    else if ( !PyvoxIntype_Check(type, "type", &dtype) )
        ReturnPyvoxLibError();

    /* Case 1: Return a plain number */
    if ( PyvoxDoubleSeq_Check(map, "map", rank, rank, 0, point) ) {
        flag = vxl_cubic_point(dtype, &value, vself, point);
        if (flag)    ReturnPyvoxLibError();
        result = pyvox_get_value(dtype, &value, 0);
        if (result == 0)    ReturnPyvoxLibError();
    }

    /* Case 2: Return an array */
    else if ( PyvoxArray_CheckSrc(map, "map", &vmap) ) {
        result = (PyObject *)parray_create(&vres);
        if (result==0)    ReturnPyvoxLibError();
        flag = vxl_cubic_map(vres, dtype, vself, vmap);
        if (flag)    ReturnPyvoxLibError();
    }

    /* Case 3: Otherwise report an invalid map argument */
    else
        ReturnPyvoxError(ERR_INIT, "Invalid map argument");

    /* Return the result as a Python float */
    clear_error_context();
    return result;
}



/*------------------------------------------------------------------------
parray_det - Determinant of a Pyvox array (considered as a matrix)

This Pyvox array method returns the determinant of a Pyvox array
considered as a matrix.
-------------------------------------------------------------------------*/

static
PyObject *
parray_det(
    PyObject *self,            /* Pyvox array */
    PyObject *args)            /* Arguments (none) */
{
    PyObject *result;          /* Result as Python object */
    voxel_array *vself;        /* Self as a voxel array */
    anytype det;               /* Determinant of self */
    int type;                  /* Data type of determinant */
    int flag;                  /* Error flag */

    set_error_context("array.det()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        ReturnPyvoxLibError();

    /* Compute determinant and convert to Python number */
    flag = matrix_determinant(&type, &det, vself);
    if (flag)    ReturnPyvoxLibError();
    result = pyvox_get_value(type, &det, 0);

    clear_error_context();
    return result;
}



/*------------------------------------------------------------------------
parray_diag - Extract diagnonal elements of a rank-2 array

This instance method constructs a new rank-1 Pyvox array containing
the diagonal elements of the given rank-2 array; the type of the
result is the same as the type of the original array.  The inverse
operation of constructing a diagonal matrix from its diagonal elements
may be done using pyvox.diag().

NOTE: Prior to 2001-10-23 this function constructed a diagonal matrix
from a vector of the elements to be placed on the diagonal.  This is
an incompatible change, so beware of old users; the argument types,
however, are incompatible, so old uses should be detected automatically.
-------------------------------------------------------------------------*/

static
PyObject *
parray_diag(
    PyObject *self,        /* Pyvox array */
    PyObject *args)        /* Arguments (none) */
{
    voxel_array *vself;     /* Self as a voxel array */
    pyvox_array *res;       /* Result as a Pyvox array */
    voxel_array *vres;      /* Result as a voxel array */
    int flag;

    set_error_context("array.diag()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        ReturnPyvoxLibError();

    /* Create Pyvox array for result */
    res = parray_create(&vres);
    if (res == 0)    ReturnPyvoxLibError();

    /* Construct the diagonal matrix */
    flag = matrix_get_diag(vres, vself);
    if (flag)    ReturnPyvoxLibError();

    /* Return the result as a Pyvox array */
    clear_error_context();
    return Py_BuildValue("N", res);
}



/*-------------------------------------------------------------------------
parray_dilabel - Morphological dilation of a label image

This instance method for Pyvox arrays performs a morphological
dilation on self, which must be a volume image of non-zero rank and
any unsigned integral type.  Each pixel of the result is the smallest
nonzero pixel value within the neighborhood defined by the kernel; if
all pixels in the neighborhood are zero, then the result is zero.  The
neighborbood is specified by a kernel object and defaults to a
centered 3x3x3 neighborhood.  The bias and coefficients of the kernel,
if any, are ignored.

FIXME:  Default to a neighborhood of the correct rank, computing it as
needed.
-------------------------------------------------------------------------*/

static PyObject *
parray_dilabel(
    PyObject *self,               /* Self: Volume image to process */
    PyObject *args)
{
    PyObject *pykern = 0;         /* Arg 1: Kernel specifying neighborhood */

    pyvox_array *out;             /* Output as Pyvox arrays */
    voxel_array *vin, *vout;      /* Input and output as voxel arrays */
    struct vxl_kernel *kern;      /* The kernel struct itself */
    int flag;

    set_error_context("array.dilabel()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "|O", &pykern) )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, "self", &vin) )
	ReturnPyvoxLibError();

    /* Find the kernel struct, if specified */
    kern = &neigh3d27;
    if ( pykern != 0 && !PyvoxKernel_Check(pykern, "kernel", &kern) )
        ReturnPyvoxLibError();
    if (kern->rank != vin->rank)
	ReturnPyvoxError(ERR_INCOMP, 
                         "Image and kernel must have the same rank");

    /* Compute and return the result */
    out = parray_create(&vout);
    if (out == 0)    ReturnPyvoxLibError();
    flag = vxl_dilabel(vout, vin, kern);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)out;
}



/*-------------------------------------------------------------------------
parray_dilate - Morphological dilation of a binary image

This instance method for Pyvox arrays performs a morphological
dilation on self, which must be a volume image of non-zero rank and
any unsigned integral type.  Each pixel of the result is the bitwise
OR of the pixel values in the neighborhood defined by the kernel.  The
neighborbood is specified by a kernel object and defaults to a
centered 3x3x3 neighborhood.  The bias and coefficients of the kernel,
if any, are ignored.

FIXME:  Default to a neighborhood of the correct rank, computing it as
needed.
-------------------------------------------------------------------------*/

static PyObject *
parray_dilate(
    PyObject *self,               /* Self: Volume image to process */
    PyObject *args)
{
    PyObject *pykern = 0;         /* Arg 1: Kernel specifying neighborhood */

    pyvox_array *out;             /* Output as Pyvox arrays */
    voxel_array *vin, *vout;      /* Input and output as voxel arrays */
    struct vxl_kernel *kern;      /* The kernel struct itself */
    int flag;

    set_error_context("array.dilate()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "|O", &pykern) )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, "self", &vin) )
	ReturnPyvoxLibError();

    /* Find the kernel struct, if specified */
    kern = &neigh3d27;
    if ( pykern != 0 && !PyvoxKernel_Check(pykern, "kernel", &kern) )
        ReturnPyvoxLibError();
    if (kern->rank != vin->rank)
	ReturnPyvoxError(ERR_INCOMP, 
                         "Image and kernel must have the same rank");

    /* Compute and return the result */
    out = parray_create(&vout);
    if (out == 0)    ReturnPyvoxLibError();
    flag = vxl_dilate(vout, vin, kern);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)out;
}



/*--------------------------------------------------------------------------
parray_dot - Compute weighted vector dot product of two arrays

This function computes the vector dot product of two arrays,
optionally weighted by a pixelwise weight.  The source and weight
arrays may be of any type and are converted to double for this
computation; they must, however, be the same shape.  The weight array
is optional, defaulting to None.
---------------------------------------------------------------------------*/

static PyObject *
parray_dot(
    PyObject *self,          /* Self: First array */
    PyObject *args)
{
    PyObject *other = 0;     /* Arg 1: Other array */
    PyObject *weight = 0;    /* Arg 2: Weight array */
    PyObject *paxes = 0;     /* Arg 3: Axes to reduce on */
    PyObject *result = 0;    /* Result as a Python object */

    voxel_array *slf  = 0;   /* Self array as a voxel array */
    voxel_array *oth  = 0;   /* Other array as a voxel array */
    voxel_array *wate = 0;   /* Weight array as a voxel array */
    voxel_array *vres = 0;   /* Result as a voxel array */
    int naxes = -1;          /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];  /* The axes to reduce on */
    int flag;

    set_error_context("array.dot()");

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "|OOO", &other, &weight, &paxes) )
	return NULL;

    /* Extract voxel arrays */
    if ( ! PyvoxArray_Check(self, "self", &slf) )
	ReturnPyvoxLibError();
    if ( other != 0 && 
         other != Py_None &&
         ! PyvoxArray_Check(other, "other", &oth) )
	ReturnPyvoxLibError();
    if ( weight != 0 &&
         weight != Py_None &&
         ! PyvoxArray_Check(weight, "weight", &wate) )
	ReturnPyvoxLibError();

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, slf->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work */
    vres = vxl_create_array();
    if (vres == 0)    ReturnPyvoxLibError();
    flag = vxl_dot(vres, slf, oth, wate, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    result = pyvox_from_temp_varray(vres);
    if (result == 0)    ReturnPyvoxLibError();

    /* Return the result */
    clear_error_context();
    return result;
}



/*-------------------------------------------------------------------------
parray_erode - Morphological erosion of a volume image

This instance method for Pyvox arrays performs a morphological erosion
on self, which must be a volume image of any non-zero rank and any
unsigned integral type, using an optional neighborhood specified as a
kernel object.  If omitted, the neighborhood defaults to a centered
3x3x3 neighborhood.  The bias and coefficients of the kernel, if any,
are ignored.

FIXME:  Default to a neighborhood of the correct rank, computing it as
needed.
-------------------------------------------------------------------------*/

static PyObject *
parray_erode(
    PyObject *self,                /* Self: Volume image to process */
    PyObject *args)
{
    PyObject *pykern = 0;          /* Arg 1: Neighborhood to use */

    pyvox_array *out;              /* Output as Pyvox arrays */
    voxel_array *vin, *vout;       /* Input and output as voxel arrays */
    struct vxl_kernel *kern;       /* The kernel struct */
    int flag;

    set_error_context("array.erode()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "|O", &pykern) )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, "self", &vin) )
	ReturnPyvoxLibError();
  
    /* Find the kernel struct, if specified */
    kern = &neigh3d27;
    if ( pykern != 0 && !PyvoxKernel_Check(pykern, "kernel", &kern) )
        ReturnPyvoxLibError();
    if (kern->rank != vin->rank)
	ReturnPyvoxError(ERR_INCOMP,
                         "Image and kernel must have the same rank");
  
    /* Compute and return the result */
    out = parray_create(&vout);
    if (out == 0)    ReturnPyvoxLibError();
    flag = vxl_erode(vout, vin, kern);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)out;
}



/*----------------------------------------------------------------------
parray_eigsy - Solve real symmetric eigensystem

This Pyvox array method returns a pair (val, vec) containing the
eigenvalues and eigenvectors of the Pyvox array considered as a real
symmetric matrix or a complex Hermitian matrix.
-----------------------------------------------------------------------*/

static 
PyObject *
parray_eigsy(
    PyObject *self,            /* Pyvox array */
    PyObject *args)            /* Arguments (none) */
{
    voxel_array *vself;        /* Self as a voxel array */
    PyObject *val;             /* Eigenvalues */
    voxel_array *vval;         /* Eigenvalues as voxel array */
    PyObject *vec;             /* Eigenvectors */
    voxel_array *vvec;         /* Eigenvectors as voxel array */
    int flag;                  /* Error flag */

    set_error_context("array.eigsy()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        ReturnPyvoxLibError();

    /* Create voxel arrays for values and vectors */
    val = (PyObject *)parray_create(&vval);
    vec = (PyObject *)parray_create(&vvec);
    if (val == 0 || vec == 0)    ReturnPyvoxLibError();

    /* Solve the eigensystem */
    flag = matrix_eigen_symmetric(vval, vvec, vself);
    if (flag)    ReturnPyvoxLibError();

    /* Return pair of eigenvalues and eigenvectors */
    clear_error_context();
    return Py_BuildValue("(NN)", val, vec);
}



/*-------------------------------------------------------------------------
parray_fft - Compute fast Fourier transform of a Pyvox array

FIXME: Allow caller to choose result type?  Ditto for parray_ifft().
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H
static PyObject *
parray_fft(
    PyObject *self,             /* Self: array to transform */
    PyObject *args)
{
    PyObject *paxes  = 0;       /* Arg 1: axes to transform on */
    PyObject *result = 0;       /* Result: transformed array */

    voxel_array *vres  = 0;     /* Result as a voxel array */
    voxel_array *vself = 0;     /* Self as a voxel array */
    int inverse = 0;            /* Compute inverse FFT instead? */
    int naxes = -1;             /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];     /* Specific axes to reduce on */
    int dtype = 0;              /* Destination type */
    int flag;

    set_error_context("array.fft()");

    /* Parse the Python arguments into C values */
    if  ( ! PyArg_ParseTuple(args, "|O", &paxes) )
	return NULL;

    /* Extract voxel array */
    if  ( ! PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxLibError();

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, vself->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Choose suitable destination type */
    dtype = exim_complex_type(vself->type);
    if (dtype == 0)
        ReturnPyvoxError(ERR_INIT, "Can't determine dest type");

    /* Call Voxel Kit to do the real work */
    vres = vxl_create_array();
    if (vres == 0)    ReturnPyvoxLibError();
    flag = vxl_fft(vres, dtype, vself, inverse, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    result = pyvox_from_temp_varray(vres);
    if (result == 0)    ReturnPyvoxLibError();

    /* Return the result */
    clear_error_context();
    return result;
}
#endif



/*------------------------------------------------------------------------
parray_fill2d - Fill contour in a 2D image

This instance method for two-dimensional Pyvox arrays fills with a
specified value all the voxels that lie within a contour given by a
set of vertices.

FIXME: The format for the vertices is currently very restrictive; generalize
it to include a variety of useful formats.
-------------------------------------------------------------------------*/

static PyObject *
parray_fill2d(
    PyObject *self,           /* Self: Array to draw into */
    PyObject *args)
{
    PyObject *ppoints;        /* Arg 1: Vertices defining the contour */
    PyObject *pvalue;         /* Arg 2: Value to fill with */

    voxel_array *canvas;      /* Voxel array to draw into */
    voxel_array *points;      /* Vertices defining the contour */
    long npoints;             /* Number of vertices */
    double value;             /* Value to fill with */
    int flag;

    set_error_context("array.fill2d()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "OO", &ppoints, &pvalue) )
	return NULL;

    /* Find the voxel array to draw into */
    if ( !PyvoxArray_Check(self, "self", &canvas) )
	ReturnPyvoxLibError();

    /* Find the vertices defining the contour */
    if  ( !PyvoxArray_Check(ppoints, "points", &points) ||
	  points == 0 || points->magic != VXL_MAGIC ||
	  points->data == 0 || points->type != INTP_DOUBLE ||
	  points->rank != 2 || points->dimen[1] != 2
	)
	ReturnPyvoxError(ERR_INIT,
                         "Vertices must be given as an N by 2 Pyvox array");
    npoints = points->dimen[0];

    /* Find the value to fill with */
    if ( !PyvoxDouble_Check(pvalue, &value) )
	ReturnPyvoxError(ERR_INCOMP, "Fill value is incompatible type");

    /* Call the Voxel Kit to do the real work */
    flag = vxl_fill_2d_contour(canvas, npoints, points->data, value);
    if (flag)    ReturnPyvoxLibError();

    /* Return the modified array */
    Py_INCREF(self);
    clear_error_context();
    return self;
}



/*-------------------------------------------------------------------------
parray_histo - Compute univariate histogram of an unsigned char image

This instance method for Pyvox arrays computes the univariate
histogram of self, which must be an unsigned char image, and returns
it as a one-dimensional unsigned long voxel array.  An optional weight
array may be provided, which must be the same shape and type as self.

FIXME:  Implemented only for unsigned char/short images.
-------------------------------------------------------------------------*/

static PyObject *
parray_histo(
    PyObject *self,             /* Self: Image to take histogram of */
    PyObject *args)
{
    PyObject *weight = 0;       /* Arg 1: Weight image */

    pyvox_array *hist;          /* Result: Univariate histogram */
    voxel_array *vself;         /* Self as a voxel array */
    voxel_array *vweight = 0;   /* Weight as a voxel array */
    voxel_array *vhist;         /* Histogram as a voxel array */
    int flag;

    set_error_context("array.histo()");

    /* Fetch and verify the arguments */
    if ( !PyArg_ParseTuple(args, "|O", &weight) )
	return NULL;
    if ( !PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxError(ERR_UNIMP, "Self is not a valid Pyvox array");
    if ( weight != 0 && weight != Py_None )
	if ( !PyvoxArray_Check(weight, "weight", &vweight) ||
	     vweight->type != INTP_UCHAR )
	    ReturnPyvoxError(ERR_UNIMP,
                             "Weight is not an unsigned char Pyvox array");

    /* Compute and return the histogram */
    hist = parray_create(&vhist);
    if (hist == 0)    ReturnPyvoxLibError();
    flag = vxl_histo_uchar(vhist, vself, vweight);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)hist;
}



/*-------------------------------------------------------------------------
parray_i2p - Compute the index-to-physical transform

This function returns the affine transform which converts from index to
physical coordinates for this array, based on the origin and spacing
attributes.

ALGORITHM: This function makes an upcall to the Python level, where the
affine transform class is defined.
-------------------------------------------------------------------------*/

static PyObject *
parray_i2p(
    PyObject *self,             /* Array to get transform of */
    PyObject *args)             /* Args: None used */
{
    voxel_array *vself;         /* Self as a voxel array */
    int rank;                   /* Rank of the array */
    long dimen[2];              /* Dimensions of matrix and offset */
    pyvox_array *mtx;           /* Matrix of the transform */
    voxel_array *vmtx;          /* Matrix, as a voxel array */
    double *mdata;              /* Data of the matrix */
    pyvox_array *off;           /* Offset of the transform */
    voxel_array *voff;          /* Offset, as a voxel array */
    double *odata;              /* Data of the offset */
    PyObject *affine;           /* Affine class object */
    PyObject *result;           /* The transform to be returned as result */
    int i;

    set_error_context("array.i2p()");

    /* Fetch and verify the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the array itself */
    if ( !PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxLibError();
    rank = vself->rank;

    /* Construct arrays for the matrix and offset of the transform */
    dimen[0] = dimen[1] = rank;
    mtx = parray_alloc(&vmtx, INTP_DOUBLE, 2, dimen);
    if (mtx == 0)    ReturnPyvoxLibError();
    mdata = (double *)vmtx->data;
    dimen[1] = 1;
    off = parray_alloc(&voff, INTP_DOUBLE, 2, dimen);
    if (off == 0)    ReturnPyvoxLibError();
    odata = (double *)voff->data;

    /* Fill matrix and offset with the right values */
    for (i = 0; i < rank*rank; i++)
	mdata[i] = 0.0;
    for (i = 0; i < rank; i++) {
	odata[i] = vself->origin[i];
	mdata[i*rank+i] = vself->spacing[i];
    }

    /* Construct and return the affine transform */
    affine = PyDict_GetItemString(pyvox_dict, "affine");
    result = PyObject_CallFunction(affine, "iOO", rank, mtx, off);

    Py_INCREF(result);
    clear_error_context();
    return (PyObject *)result;
}



/*-------------------------------------------------------------------------
parray_ifft - Compute inverse fast Fourier transform of a Pyvox array

FIXME: Allow caller to choose result type?
-------------------------------------------------------------------------*/

#ifdef HAVE_COMPLEX_H
static PyObject *
parray_ifft(
    PyObject *self,             /* Self: array to transform */
    PyObject *args)
{
    PyObject *paxes  = 0;       /* Arg 1: axes to transform on */
    PyObject *result = 0;       /* Result: transformed array */

    voxel_array *vres  = 0;     /* Result as a voxel array */
    voxel_array *vself = 0;     /* Self as a voxel array */
    int inverse = 1;            /* Compute inverse FFT instead? */
    int naxes = -1;             /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];     /* Specific axes to reduce on */
    int dtype = 0;              /* Destination type */
    int flag;

    set_error_context("array.ifft()");

    /* Parse the Python arguments into C values */
    if  ( ! PyArg_ParseTuple(args, "|O", &paxes) )
	return NULL;

    /* Extract voxel array */
    if  ( ! PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxLibError();

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, vself->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Choose suitable destination type */
    dtype = exim_complex_type(vself->type);
    if (dtype == 0)
        ReturnPyvoxError(ERR_INIT, "Can't determine dest type");

    /* Call Voxel Kit to do the real work */
    vres = vxl_create_array();
    if (vres == 0)    ReturnPyvoxLibError();
    flag = vxl_fft(vres, dtype, vself, inverse, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    result = pyvox_from_temp_varray(vres);
    if (result == 0)    ReturnPyvoxLibError();

    /* Return the result */
    clear_error_context();
    return result;
}
#endif



/*------------------------------------------------------------------------
parray_inverse - Compute inverse of a matrix

This instance method computes the inverse of a non-singular square
matrix represented as a Pyvox array of any floating-point type.
-------------------------------------------------------------------------*/

static
PyObject *
parray_inverse(
    PyObject *self,         /* Pyvox array */
    PyObject *args)         /* Arguments (none) */
{
    voxel_array *vself;     /* Self as a voxel array */
    pyvox_array *inv;       /* Inverse as a Pyvox array */
    voxel_array *vinv;      /* Inverse as a voxel array */
    int flag;

    set_error_context("array.inverse()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        ReturnPyvoxLibError();

    /* Create Pyvox array for result */
    inv = parray_create(&vinv);
    if (inv == 0)    ReturnPyvoxLibError();

    /* Compute the inverse */
    flag = matrix_inverse(vinv, vself);
    if (flag)    ReturnPyvoxLibError();

    /* Return the result as a Pyvox array */
    clear_error_context();
    return Py_BuildValue("N", inv);
}



/*--------------------------------------------------------------------------
parray_kmeans1 - Compute univariate K-means from a histogram

Assuming that self is the univariate histogram of an unsigned char
image and given initial guesses at class centroids, this instance
method for Pyvox arrays uses the K-means algorithm to refine the class
centroids.  The centroids thus computed can be used later by the
array.nnclass1 function to do the actual segmentation.  If initial
guesses are not available for the centroids, they should all be set to
zero.  The number of classes found is set by the number of centroids
provided.

FIXME: The initial centroids could be made optional, but only if I
provide some other means of specifying the number of classes to be
found.  One possibility is to consider a naked number rather than as
list as the number of classes to be found.  Another possibility is to
have the centroids and nclass as keyword arguments.

FIXME:  It's probably reasonable to allow the histogram to be something
other than a Pyvox array.  A numeric sequence, for instance.  For that
matter, a rank-1 Pyvox array could also reasonably be regarded as a
sequence and support all the usual sequence operations.
---------------------------------------------------------------------------*/

static PyObject *
parray_kmeans1(
    PyObject *self,                 /* Self: Image histogram */
    PyObject *args)
{
    PyObject *centin = 0;           /* Arg 1: Initial centroids */
    PyObject *centout;              /* Result: Final centroids */

    pyvox_array  *vohist;           /* hist as a Pyvox array */
    voxel_array  *vahist;           /* Voxel array that hist refers to */
    double  *cent = 0;              /* Working copy of centroids */
    long nclass;                    /* Number of classes to find */
    int i, flag;
    PyObject *ob;

    set_error_context("array.keans1()");

    /* Parse the arguments */
    if ( !PyArg_ParseTuple(args, "O", &centin) )  
	return NULL;

    /* Find the voxel array that the hist argument refers to */
    vohist = (pyvox_array *)self;
    if ( !PyvoxArray_Check(self, "self", 0) )
	ReturnPyvoxLibError();
    vahist = vohist->varray;

    /* Convert the centroids to a format usable in C */
    /* The upper limit 512 on the number of classes is
       essentially arbitrary but might catch some bugs. */
    cent = PyvoxDoubleSeq_AsDouble(centin, "centroids", 2, 512, &nclass);
    if ( cent == 0 )
	ReturnPyvoxLibError();

    /* Run the K-means algorithm */
    flag = vxl_kmeans1_histo(vahist, nclass, cent);
    if (flag)    ReturnPyvoxLibError();

    /* Convert the updated centroids to a Python object and return it. */
    /* FIXME:  I _think_ the following code sets the right reference
       counts on the dimensions and the tuple containing them, but
       the Python API manual is somewhat obscure on this point. */
    centout = PyList_New(nclass);
    for (i = 0; i < nclass; i++) {
        ob = PyFloat_FromDouble(cent[i]);
	PyList_SetItem(centout, i, ob); }

    free(cent);
    clear_error_context();
    return centout;
}



/*--------------------------------------------------------------------------
parray_kmeans2 - Compute univariate K-means from a bivariate histogram

Assuming that self is the bivariate histogram of two unsigned char
images and given initial guesses at class centroids, this instance
method for Pyvox arrays uses the K-means algorithm to refine the class
centroids.  The centroids thus computed can be used later by the
array.nnclass2 function to do the actual segmentation.  If initial
guesses are not available for the centroids, they should all be set to
zero.  The number of classes found is set by the number of centroids
provided.

FIXME: The initial centroids could be made optional, but only if I
provide some other means of specifying the number of classes to be
found.  One possibility is to consider a naked number rather than as
list as the number of classes to be found.  Another possibility is to
have the centroids and nclass as keyword arguments.
---------------------------------------------------------------------------*/

static PyObject *
parray_kmeans2(
    PyObject *self,               /* Self: Image histogram */
    PyObject *args)
{
    PyObject *pycent1 = 0;        /* Arg 1: Initial centroids for 1st image */
    PyObject *pycent2 = 0;        /* Arg 2: Initial centroids for 2nd image */
    PyObject *result = 0;         /* Result: Final centroids */

    pyvox_array  *vohist;         /* hist as a Pyvox array */
    voxel_array  *vahist;         /* Voxel array that hist refers to */
    double  *cent1=0, *cent2=0;   /* Working copy of centroids */
    long nclass;                  /* Number of classes to find */
    int i, flag;
    long n1, n2;
    PyObject *ob;

    set_error_context("array.keans2()");

    /* Parse the arguments */
    if ( !PyArg_ParseTuple(args, "OO", &pycent1, &pycent2) )  
	return NULL;

    /* Find the voxel array that the hist argument refers to */
    vohist = (pyvox_array *)self;
    if ( !PyvoxArray_Check(self, "self", 0) )
	ReturnPyvoxLibError();
    vahist = vohist->varray;

    /* Convert the centroids to a format usable in C */
    /* The upper limit 512 on the number of classes is
       essentially arbitrary but might catch some bugs. */
    cent1 = PyvoxDoubleSeq_AsDouble(pycent1, "centroids", 2, 512, &n1);
    if ( cent1 == 0 )
	ReturnPyvoxLibError();
    cent2 = PyvoxDoubleSeq_AsDouble(pycent2, "centroids", 2, 512, &n2);
    if ( cent2 == 0 )
	ReturnPyvoxLibError();
    if (n1 != n2) 
        ReturnPyvoxError(ERR_INCOMP, 
                         "Both centroid lists must be the same length");
    nclass = n1;

    /* Run the K-means algorithm */
    flag = vxl_kmeans2_histo(vahist, nclass, cent1, cent2);
    if (flag)    ReturnPyvoxLibError();

    /* Convert the updated centroids to a Python object and return it. */
    /* FIXME:  I _think_ the following code sets the right reference
       counts on the dimensions and the tuple containing them, but
       the Python API manual is somewhat obscure on this point. */
    pycent1 = PyList_New(nclass);
    pycent2 = PyList_New(nclass);
    for (i = 0; i < nclass; i++) {
        ob = PyFloat_FromDouble(cent1[i]);
	PyList_SetItem(pycent1, i, ob);
        ob = PyFloat_FromDouble(cent2[i]);
	PyList_SetItem(pycent2, i, ob);
    }
    result = PyList_New(2);
    PyList_SetItem(result, 0, pycent1);
    PyList_SetItem(result, 1, pycent2);

    free(cent1);
    free(cent2);
    clear_error_context();
    return result;
}



/*------------------------------------------------------------------------
parray_linear - Linear interpolation in a voxel array

This instance method for Pyvox arrays returns a new array whose elements
are computed by linear interpolation in the voxels of self at the sample
points defined by the map argument.
-------------------------------------------------------------------------*/

static PyObject *
parray_linear(
    PyObject *self,                 /* Pyvox array to interpolate in */
    PyObject *args)                 /* Other arguments */
{
    PyObject *map;                  /* Arg 1: Points to evaluate at */
    PyObject *type = 0;             /* Arg 2: Desired type of the result */
    PyObject *result = 0;           /* Result */

    voxel_array *vself;             /* Self as voxel array */
    int rank;                       /* Rank of self */
    int stype;                      /* Type of self */
    voxel_array *vmap;              /* Points to evaluate at */
    voxel_array *vres;              /* Result as voxel array */
    int dtype;                      /* Desired type of result */
    double point[VXL_MAX_RANK];     /* A single point to evaluate at */
    anytype value;                  /* Value at the given point */
    int flag;

    set_error_context("array.linear()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_CheckSrc(self, "self", &vself) )
        ReturnPyvoxLibError();
    rank  = vself->rank;
    stype = vself->type;

    /* Parse the arguments */
    if ( ! PyArg_ParseTuple(args, "O|O", &map, &type) )  
	return NULL;

    /* Verify the array type; default to type of source */
    if ( type == 0 )
        dtype = stype;
    else if ( !PyvoxIntype_Check(type, "type", &dtype) )
        ReturnPyvoxLibError();

    /* Case 1: Return a plain number */
    if ( PyvoxDoubleSeq_Check(map, "map", rank, rank, 0, point) ) {
        flag = vxl_linear_point(dtype, &value, vself, point);
        if (flag)    ReturnPyvoxLibError();
        result = pyvox_get_value(dtype, &value, 0);
        if (result == 0)    ReturnPyvoxLibError();
    }

    /* Case 2: Return an array */
    else if ( PyvoxArray_CheckSrc(map, "map", &vmap) ) {
        result = (PyObject *)parray_create(&vres);
        if (result==0)    ReturnPyvoxLibError();
        flag = vxl_linear_map(vres, dtype, vself, vmap);
        if (flag)    ReturnPyvoxLibError();
    }

    /* Case 3: Otherwise report an invalid map argument */
    else
        ReturnPyvoxError(ERR_INIT, "Invalid map argument");

    /* Return the result as a Python float */
    clear_error_context();
    return result;
}



/*--------------------------------------------------------------------------
parray_list - Return array elements as a list

This instance method for Pyvox arrays returns the elements of the array
(in row-major order) as a Python list of floats, ints, or longs.  See also 
parray_tuple, which returns a tuple rather than a list.
--------------------------------------------------------------------------*/

static PyObject *
parray_list(
    PyObject *self,            /* Pyvox array to enlist */
    PyObject *args)            /* Arguments (none in this case) */
{
    PyObject *result;          /* List to return as result */
    voxel_array *vself;        /* Self as a voxel array */
    int type;                  /* Type of voxel array */
    void *data;                /* Data contained in voxel array */
    long count;                /* Number of elements in array */
    PyObject *value;           /* Value of an element */
    long i;

    set_error_context("array.list()");

    /* Find voxel array of self; count number of elements */
    if ( ! PyvoxArray_CheckSrc(self, "self", &vself) )
	ReturnPyvoxLibError();
    type  = vself->type;
    data  = vself->data;
    count = vxli_count(vself);

    /* Check other arguments; there should be none */
    if ( !PyArg_ParseTuple(args, "") )  
	return NULL;

    /* Construct the output list */
    result = PyList_New(count);
    for (i = 0; i < count; i++) {
        value = pyvox_get_value(type, data, i);
        if (value == 0)    ReturnPyvoxLibError();
        PyList_SetItem(result, i, value);
    }

    clear_error_context();
    return result;
}



/*-------------------------------------------------------------------------
parray_lostat - Compute local statistics of a volume image

This instance method for Pyvox arrays computes the local mean and
standard deviation within each 3x3x3 neighborhood of self and returns
them as a list containing two new images of the same shape and type;
the first is the local mean and the second is the local standard
deviation.  Boundary voxels are handled correctly.  The standard
deviation is multiplied by 2.0 for images of type exim.uchar, to
better match the possible range of values to the range supported by
unsigned char; the scale is left at 1.0 for all other data types.

The array is allowed to be complex, but not the kernel coefficients.

FIXME: Add support for a user-supplied kernel?  Add scale to the
kernel and use to define the scale parameter here?

FIXME: The underlying voxel kit function supports any type, rank,
kernel, and scale; generalize this function to access those
capabilities.

FIXME: Provide means to support users who want only the local mean or
local variance?  This might not be particularly useful, though.  The
local mean could be computed as a convolution, but the local variance
requires that the local mean be computed (but perhaps never stored?).
-------------------------------------------------------------------------*/

static PyObject *
parray_lostat(
    PyObject *self,           /* Self: Volume image to take statistics of */
    PyObject *args)
{
    PyObject *result;         /* Result: List containing two volume images */

    pyvox_array *mean, *sdev; /* The mean and std dev images */
    voxel_array *in, *m, *d;  /* Voxel arrays inside the Pyvox objects */

    struct vxl_kernel *kern;  /* Neighborhood definition */
    double scale;             /* Scale factor for std dev */
    int flag;

    set_error_context("array.lostat()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, "self",  0) )
	ReturnPyvoxLibError();
    in = ((pyvox_array *)self)->varray;
    if (in->rank != 3)
	ReturnPyvoxError(ERR_UNIMP,
                         "Sorry, only 3d has been implemented yet.");
  
    /* Create Pyvox arrays for the output */
    mean = parray_create(&m);
    sdev = parray_create(&d);
    if (mean == 0 || sdev == 0)    ReturnPyvoxLibError();

    /* Set parms needed by Voxel Kit but not yet supported by Pyvox */
    kern = &neigh3d27;
    if (in->type == INTP_UCHAR)
	scale = 2.0;
    else
	scale = 1.0;

    /* Compute and return the result */
    flag = vxl_lostat(m, d, in, kern, scale);
    if (flag)    ReturnPyvoxLibError();
    result = Py_BuildValue("[NN]", mean, sdev);

    clear_error_context();
    return result;
}



/*-------------------------------------------------------------------------
parray_lowpass - Lowpass filter a volume image

This instance method for Pyvox arrays lowpass filters and optionally
subsamples self, which must be a 3d volume image of any type.  The
kernel is a 3x3x3 convolution kernel of the form 2^-(3+|x|+|y|+|z|);
this kernel will completely surpress the Nyquist frequency along any of
the three principal axes.  If the shrink argument is provided, then
pixel values are computed for only the subsample with that spacing.

FIXME: Make the default bandwidth depend on the subsampling interval.

FIXME: Allow the user to specify the bandwidth?

FIXME: Should this function be removed, in lieu of a function that
generates a lowpass kernel in any specified number of dimensions?
-------------------------------------------------------------------------*/

static PyObject *
parray_lowpass(
    PyObject *self,             /* Self: Volume image to process */
    PyObject *args)
{
    PyObject *pyshrink = 0;      /* Arg 1: Desired shrink factors */

    pyvox_array *out;            /* Output as Pyvox arrays */
    voxel_array *vin, *vout;     /* Input and output as voxel array */
    long shrink[VXL_MAX_RANK];   /* Shrink factors */
    struct vxl_kernel *lowpass = 0;     /* Lowpass kernel */
    int i, flag;

    set_error_context("array.lowpass()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "|O", &pyshrink) )
	return NULL;

    /* Find the input voxel array itself */
    if ( !PyvoxArray_Check(self, "self", &vin) )
	ReturnPyvoxLibError();
  
    /* Find the shrink factors, if any */
    if ( pyshrink == 0 ) {
        for (i = 0; i < vin->rank; i++)
            shrink[i] = 1;
    }
    else if ( PyvoxLong_Check(pyshrink, &shrink[0]) ) {
        for (i = 1; i < vin->rank; i++)
            shrink[i] = shrink[0];
    }
    else if ( PyvoxLongSeq_Check(pyshrink, vin->rank, vin->rank, 0, shrink) )
        /* Do nothing */ ;
    else
        ReturnPyvoxError(ERR_MISC, "Unable to interpret shrink argument");
  
    /* Compute and return the result */
    lowpass = vxl_lowpass_kernel(vin->rank);
    out  = parray_create(&vout);
    if (out == 0)    ReturnPyvoxLibError();
    flag = vxl_convolve(vout, vin, lowpass, shrink);
    if (flag != 0)    ReturnPyvoxLibError();

    /* Clean up and return result */
    vxl_destroy_kernel(lowpass);
    clear_error_context();

    return (PyObject *)out;
}



/*--------------------------------------------------------------------------
parray_mean - Compute mean along specified axes

This function computes the mean along specified axes to yield an array or
scalar of reduced rank.  The source and weight arrays may be of any type;
the result is always of type double or double complex.
---------------------------------------------------------------------------*/

static PyObject *
parray_mean(
    PyObject *self,             /* Self: array to take means of */
    PyObject *args)
{
    PyObject *pweight = 0;      /* Arg 1: weight array */
    PyObject *paxes   = 0;      /* Arg 2: axes to reduce on */
    PyObject *result  = 0;      /* Result: mean of array */

    voxel_array *vmean;         /* Result as a voxel array */
    voxel_array *vself  = 0;    /* Self as a voxel array */
    voxel_array *weight = 0;    /* Weight as a voxel array */
    int naxes = -1;             /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];     /* Specific axes to reduce on */
    int flag;

    set_error_context("array.mean()");

    /* Parse the Python arguments into C values */
    if  ( ! PyArg_ParseTuple(args, "|OO", &pweight, &paxes) )
	return NULL;

    /* Extract voxel arrays */
    if  ( ! PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxLibError();
    if  ( ! (pweight == 0) &&
	  ! (pweight == Py_None) && 
	  ! PyInt_Check(pweight) &&
	  ! PyFloat_Check(pweight) &&
	  ! PyvoxArray_Check(pweight, "weight", &weight) )
	ReturnPyvoxLibError();

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, vself->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work and return result */
    vmean = vxl_create_array();
    if (vmean == 0)    ReturnPyvoxLibError();
    flag = vxl_mean(vmean, vself, weight, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    result = pyvox_from_temp_varray(vmean);
    if (result == 0)    ReturnPyvoxLibError();

    /* Return results */
    clear_error_context();
    return result;
}



/*-------------------------------------------------------------------------
parray_metacopy - Copy selected metadata from another array

This instance method copies selected metadata from another
Pyvox array; but all the real work is done in the Python function
pyvox.metacopy().
----------------------------------------------------------------------------*/

static PyObject *
parray_metacopy(
     PyObject *self,           /* Self as a Pyvox array */
     PyObject *args,           /* Positional arguments of this call */
     PyObject *kwargs)         /* Keyword arguments of this call */
{
    PyObject *other = 0;       /* Arg 1: Array to copy from */

    PyObject *funct   = 0;     /* pyvox.metacopy() function */
    PyObject *newargs = 0;     /* Modified arguments for it */
    PyObject *result  = 0;     /* Result of calling it */

    set_error_context("array.metacopy()");

    /* Parse argument list */
    if ( !PyArg_ParseTuple(args, "O", &other) )
        ReturnPyvoxLibError();

    /* Paste self onto front of argument tuple. */
    newargs = Py_BuildValue("(OO)", self, other);

    /* Find the pyvox.write_array function in the pyvox dictionary */
    funct = PyDict_GetItemString(pyvox_dict, "metacopy");
    if (funct == 0)
        ReturnPyvoxError(ERR_ASSERT, "Could not find pyvox.metacopy()");

    /* Call it and return the result it gives */
    result = PyEval_CallObjectWithKeywords(funct, newargs, kwargs);

    /* Decrement ref count of the newargs tuple; the ref counts of its
       elements will be automatically decremented also. */
    Py_XDECREF(newargs);

    clear_error_context();
    return result;
}



/*-------------------------------------------------------------------------
parray_minmax - Find minimum and maximum values in a voxel array

This method finds the smallest and largest values in the voxel array
and returns a list of those two values, optionally working along
selected axes only.  Defined for real types only.
-------------------------------------------------------------------------*/

static PyObject *
parray_minmax(
    PyObject *self,                  /* Voxel array to find min/max of */
    PyObject *args)
{
    PyObject *paxes=0;               /* Arg 1: axes to reduce on */
    PyObject *result=0;              /* Result: list [min, max] */
    PyObject *res1=0, *res2=0;       /* Minima and maxima as Python objects */

    voxel_array *vself;              /* Input volume image */
    int naxes = -1;                  /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];          /* Specific axes to reduce on */
    voxel_array *vres1=0, *vres2=0;  /* Minima and maxima as voxel arrays */
    int flag;

    set_error_context("array.minmax()");

    /* Fetch the arguments */
    if ( ! PyArg_ParseTuple(args, "|O", &paxes) )
	return NULL;

    /* Find the input voxel array itself */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        ReturnPyvoxLibError();
  
    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, vself->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Call the Voxel Kit to compute the result */
    vres1 = vxl_create_array();
    vres2 = vxl_create_array();
    if (vres1 == 0 || vres2 == 0)
        ReturnPyvoxLibError();
    flag = vxl_minmax(vres1, vres2, vself, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    res1 = pyvox_from_temp_varray(vres1);
    res2 = pyvox_from_temp_varray(vres2);
    if (res1 == 0 || res2 == 0)
        ReturnPyvoxLibError();

    /* Return the result */
    result = Py_BuildValue("[NN]", res1, res2);
    clear_error_context();
    return result;
}



/*----------------------------------------------------------------------
parray_mmul - Multiply self by another matrix

This Pyvox array method returns a new array which is the product of
self and another matrix.  As a special case, if the other matrix is a
rank-1 vector of the appropriate length, it will be treated as a
column vector.

FIXME: The current version requires that both arrays be the same type.
-----------------------------------------------------------------------*/

static 
PyObject *
parray_mmul(
    PyObject *self,            /* Voxel array */
    PyObject *args)            /* Arguments */
{
    PyObject *other;           /* Arg 1: Matrix to multiply self by */
    pyvox_array *result;       /* Result: Product of matrices */

    voxel_array *vself;        /* Self as a voxel array */
    voxel_array *vother;       /* Other as a voxel array */
    voxel_array *vresult;      /* Result as a voxel array */
    int flag;

    /* Set context for error messages */
    set_error_context("array.mmul()");

    /* Find voxel array of self */
    if ( !PyvoxArray_Check(self, "left factor", &vself) )
	ReturnPyvoxLibError();

    /* Get arguments; find voxel array of other */
    if ( !PyArg_ParseTuple(args, "O", &other) )
	return NULL;
    if ( !PyvoxArray_Check(other, "right factor", &vother) )
	ReturnPyvoxLibError();
    
    /* Create voxel array for result */
    result = parray_create(&vresult);
    if (result == 0)    ReturnPyvoxLibError();

    /* Multiply the matrices */
    flag = matrix_multiply(vresult, vself, vother);
    if (flag)    ReturnPyvoxLibError();

    /* Return product of matrices */
    clear_error_context();
    return Py_BuildValue("N", result);
}



/*-------------------------------------------------------------------------
parray_moments - Compute moments of a volume image

This Pyvox function computes the total mass, center of gravity, and
second central moments of a volume image; the value returned is a tuple
containing the total mass (as a Python float), the center of gravity
(as a voxel array), and the moments (as a voxel array).  The center of 
gravity and moments are in physical units, as defined by the origin and 
spacing of the volume image.

Not defined for complex types, since the cg could be complex and it's not
clear what meaning should attach to this.
-------------------------------------------------------------------------*/

static PyObject *
parray_moments(
    PyObject *self,           /* Voxel array to find moments of */
    PyObject *args)           /* Arguments: none expected */
{
    PyObject *result;         /* Result: List of mass, cg, and moments */

    voxel_array *vself;       /* Input volume image */
    pyvox_array *center;      /* Center of gravity */
    pyvox_array *moments;     /* Central moments */
    voxel_array *vcg, *vcm;   /* Center of gravity and moments */
    double mass, *cg, *cm;    /* Total mass, center of gravity, moments */
    int rank;                 /* Rank of the input voxel array */
    long dim[2];              /* Dimensions of moments array */
    int flag;

    set_error_context("array.moments()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the input voxel array itself */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        ReturnPyvoxLibError();
    rank   = vself->rank;
    dim[0] = dim[1] = rank;
  
    /* Create Pyvox arrays for the output */
    center  = parray_alloc(&vcg, INTP_DOUBLE, 1, dim);
    moments = parray_alloc(&vcm, INTP_DOUBLE, 2, dim);
    if (center == 0 || moments == 0)    ReturnPyvoxLibError();
    cg = vcg->data;
    cm = vcm->data;

    /* Compute and return the result */
    flag = vxl_moments(&mass, cg, cm, vself);
    if (flag)    ReturnPyvoxLibError();
    result = Py_BuildValue("[dNN]", mass, center, moments);
    clear_error_context();
    return result;
}



/*-----------------------------------------------------------------------
parray_mop - Arbitrary moments of a product of arrays

This instance methed computes arbitrary moments of the elementwise product
of one, two, or three arrays.

FIXME: Add support for complex types, though it's not clear that there is
any use case for it.
-------------------------------------------------------------------------*/

static PyObject *
parray_mop(
    PyObject *self,             /* Self: First array */
    PyObject *args)             /* Other arguments */
{
    PyObject *pmoments=0;       /* Arg 1: moments requested */
    PyObject *parray2=0;        /* Arg 2: second array */
    PyObject *parray3=0;        /* Arg 3: third array */

    pyvox_array *presult=0;     /* Result: computed moments */

    voxel_array *array1=0;      /* First array */
    voxel_array *array2=0;      /* Second array */
    voxel_array *array3=0;      /* Third array */
    voxel_array *moments=0;     /* Moments requested */
    voxel_array *result=0;      /* Result */
    int flag;

    set_error_context("array.mop()");

    /* Find the voxel array contained in self */
    if ( !PyvoxArray_Check(self, "self", &array1) )
	ReturnPyvoxLibError();

    /* Parse the argument tuple */
    if ( !PyArg_ParseTuple(args, "O|OO", &pmoments, &parray2, &parray3) )
	return NULL;

    /* Extract the voxel arrays contained in the arguments */
    if ( !PyvoxArray_Check(pmoments, "moments", &moments) )
	ReturnPyvoxLibError();
    if ( parray2 != 0 && !PyvoxArray_Check(parray2, "array 2", &array2) )
	ReturnPyvoxLibError();
    if ( parray3 != 0 && !PyvoxArray_Check(parray3, "array 3", &array3) )
	ReturnPyvoxLibError();

    /* Create a new Pyvox array for the result */
    presult = parray_create(&result);
    if (presult == 0)    ReturnPyvoxLibError();

    /* Call the Voxel Kit to do the real work */
    flag = vxl_mop(result, moments, array1, array2, array3);
    if (flag)    ReturnPyvoxLibError();

    /* Return the computed moments */
    clear_error_context();
    return (PyObject *)presult;
}



/*------------------------------------------------------------------------
parray_nearest - Nearest neighbor interpolation in a voxel array

This instance method for Pyvox arrays returns a new array whose
elements are computed by nearest neighbor interpolation in the voxels
of self at the sample points defined by the map argument.  Pixel
values outside the array are assumed to be zero.
-------------------------------------------------------------------------*/

static PyObject *
parray_nearest(
    PyObject *self,                 /* Pyvox array to interpolate in */
    PyObject *args)                 /* Other arguments */
{
    PyObject *map;                  /* Arg 1: Points to evaluate at */
    PyObject *type = 0;             /* Arg 2: Desired type of the result */
    PyObject *result = 0;           /* Result */

    voxel_array *vself;             /* Self as voxel array */
    int rank;                       /* Rank of self */
    int stype;                      /* Type of self */
    voxel_array *vmap;              /* Points to evaluate at */
    voxel_array *vres;              /* Result as voxel array */
    int dtype;                      /* Desired type of result */
    double point[VXL_MAX_RANK];     /* A single point to evaluate at */
    anytype value;                  /* Value at the given point */
    int flag;

    set_error_context("array.nearest()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_CheckSrc(self, "self", &vself) )
        ReturnPyvoxLibError();
    rank  = vself->rank;
    stype = vself->type;

    /* Parse the arguments */
    if ( ! PyArg_ParseTuple(args, "O|O", &map, &type) )  
	return NULL;

    /* Verify the array type; default to type of source */
    if ( type == 0 )
        dtype = stype;
    else if ( !PyvoxIntype_Check(type, "type", &dtype) )
        ReturnPyvoxLibError();

    /* Case 1: Return a plain number */
    if ( PyvoxDoubleSeq_Check(map, "map", rank, rank, 0, point) ) {
        flag = vxl_linear_point(dtype, &value, vself, point);
        if (flag)    ReturnPyvoxLibError();
        result = pyvox_get_value(dtype, &value, 0);
        if (result == 0)    ReturnPyvoxLibError();
    }

    /* Case 2: Return an array */
    else if ( PyvoxArray_CheckSrc(map, "map", &vmap) ) {
        result = (PyObject *)parray_create(&vres);
        if (result==0)    ReturnPyvoxLibError();
        flag = vxl_nearest_map(vres, dtype, vself, vmap);
        if (flag)    ReturnPyvoxLibError();
    }

    /* Case 3: Otherwise report an invalid map argument */
    else
        ReturnPyvoxError(ERR_INIT, "Invalid map argument");

    /* Return the result as a Python float */
    clear_error_context();
    return result;
}



/*------------------------------------------------------------------------
parray_nnclass1 - Univariate nearest neighbor classification

Assuming that self is an unsigned char volume image and given a set of
class centroids, this instance method for Pyvox arrays finds the
nearest class centroid for each voxel position and outputs the
corresponding class id to yield a class image.  Different centroids
may be assigned to the same class number.  There must be exactly as
many class ids as class centroids.

FIXME: The current calling sequence was adopted because it was fairly
easy to implement quickly, but it's not clear that it's the best
option; so it might change in the future.
-------------------------------------------------------------------------*/

static PyObject *
parray_nnclass1(
    PyObject *self,             /* Self: Feature image to classify */
    PyObject *args)
{
    PyObject *pyclid;           /* Arg 1: Class ids as Python sequence */
    PyObject *pycent;           /* Arg 2: Class centroids as Python sequence */

    const char clid_error[] =   /* Error: Invalid class ids argument */
	"Class ids must be a sequence of values in 0..255";

    pyvox_array  *vosrc;        /* Feature image as a Pyvox array */
    voxel_array  *vasrc;        /* Feature image as a voxel array */
    pyvox_array  *voclass;      /* Class image as a Pyvox array */
    voxel_array  *vaclass;      /* Class image as a voxel array */
    int nclass;                 /* Number of classes */
    unsigned char *clid;        /* Array of class ids */
    double *cent;               /* Array of class centroids */

    int i, flag;
    long val;
    PyObject *ob;

    set_error_context("array.nnclass1()");

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "OO", &pyclid, &pycent) )
	return NULL;

    /* Verify the source image */
    if ( !PyvoxArray_Check(self, "self", 0) )
	ReturnPyvoxLibError();
    vosrc = (pyvox_array *)self;
    vasrc = vosrc->varray;
    if ( vasrc->type != INTP_UCHAR)
	ReturnPyvoxError(ERR_UNIMP,
                         "Sorry, only uchar has been implemented yet.");

    /* Convert the class ids */
    /* FIXME:  Should this be a general-purpose utility? */
    if ( !PySequence_Check(pyclid) )
	ReturnPyvoxError(ERR_MISC, clid_error);
    nclass = PyObject_Length(pyclid);
    clid = mallock(nclass*sizeof(unsigned char));
    if (clid == 0)    ReturnPyvoxLibError();
    for (i = 0; i < nclass; i++) {
	ob = PySequence_GetItem(pyclid, i);
	if ( !PyvoxLong_Check(ob, &val) )
	    ReturnPyvoxError(ERR_MISC, clid_error);
	if ( val < 0 || val > 255)
	    ReturnPyvoxError(ERR_MISC, clid_error);
	clid[i] = val;
    }

    /* Convert the class centroids */
    cent = PyvoxDoubleSeq_AsDouble(pycent, "centroids", nclass, nclass, 0);
    if ( cent == 0 )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work */
    voclass = parray_create(&vaclass);
    if (voclass == 0)    ReturnPyvoxLibError();
    flag = vxl_nnclass1_uchar(vaclass, vasrc, nclass, clid, cent);
    if (flag)    ReturnPyvoxLibError();

    /* Free working storage and return the result */
    free(clid);
    free(cent);
    clear_error_context();
    return (PyObject *)voclass;
}



/*------------------------------------------------------------------------
parray_nnclass2 - Bivariate nearest neighbor classification

Assuming that self is a unsigned char volume image and given another
unsigned char volume image and a set of class centroids, this instance
method for Pyvox arrays finds the nearest class centroid for each
voxel position and outputs the corresponding class id to yield a class
image.  Different centroids may be assigned to the same class id.
There must be exactly as many class ids as class centroids.

The current calling sequence was adopted because it was fairly easy to
implement quickly, but it's not clear that it's the best option; so it
might change in the future.
-------------------------------------------------------------------------*/

static PyObject *
parray_nnclass2(
    PyObject *self,            /* Arg 1: First feature image */
    PyObject *args)
{
    PyObject *pysrc2;           /* Arg 2: Second feature image */
    PyObject *pyclid;           /* Arg 3: Class ids as Python sequence */
    PyObject *pycent1;          /* Arg 4: Class centroids in 1st image */
    PyObject *pycent2;          /* Arg 5: Class centroids in 2nd image */

    pyvox_array  *vosrc1, *vosrc2;   /* Feature images as Pyvox arrays */
    voxel_array  *vasrc1, *vasrc2;   /* Feature images as a voxel array */
    pyvox_array  *voclass;           /* Class image as a Pyvox array */
    voxel_array  *vaclass;           /* Class image as a voxel array */
    int nclass;                      /* Number of classes */
    unsigned char *clid;             /* Array of class ids */
    double *cent1, *cent2;           /* Arrays of class centroids */

    int i, flag;
    long val;
    PyObject *ob;
    
    set_error_context("array.nnclass2()");

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "OOOO", &pysrc2, 
			   &pyclid, &pycent1, &pycent2) )
	return NULL;

    /* Verify the source images */
    if ( !PyvoxArray_Check(self, "self", 0) )
	ReturnPyvoxLibError();
    vosrc1 = (pyvox_array *)self;
    vasrc1 = vosrc1->varray;
    if ( vasrc1->type != INTP_UCHAR)
	ReturnPyvoxError(ERR_UNIMP, uchar_error);

    if ( !PyvoxArray_Check(pysrc2, "source 2", 0) )
	ReturnPyvoxLibError();
    vosrc2 = (pyvox_array *)pysrc2;
    vasrc2 = vosrc2->varray;
    if ( vasrc2->type != INTP_UCHAR)
	ReturnPyvoxError(ERR_UNIMP, uchar_error);

    /* Convert the class ids */
    /* FIXME:  Should this be a general-purpose utility? */
    if ( !PySequence_Check(pyclid) )
	ReturnPyvoxError(ERR_INIT, clid_error);
    nclass = PyObject_Length(pyclid);
    clid = mallock(nclass*sizeof(unsigned char));
    if (clid == 0)    ReturnPyvoxLibError();
    for (i = 0; i < nclass; i++) {
	ob = PySequence_GetItem(pyclid, i);
	if ( !PyvoxLong_Check(ob, &val) )
	    ReturnPyvoxError(ERR_INIT, clid_error);
	if ( val < 0 || val > 255)
	    ReturnPyvoxError(ERR_INIT, clid_error);
	clid[i] = val;
    }

    /* Convert the class centroids for first feature image */
    cent1 = PyvoxDoubleSeq_AsDouble(pycent1, "centroids1", nclass, nclass, 0);
    if (cent1 == 0)
	ReturnPyvoxLibError();

    /* Convert the class centroids for second feature image */
    cent2 = PyvoxDoubleSeq_AsDouble(pycent2, "centroids2", nclass, nclass, 0);
    if (cent2 == 0)
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work */
    voclass = parray_create(&vaclass);
    if (voclass == 0)    ReturnPyvoxLibError();
    flag = vxl_nnclass2_uchar(vaclass, vasrc1, vasrc2, 
                              nclass, clid, cent1, cent2);
    if (flag)    ReturnPyvoxLibError();

    /* Free working storage and return the result */
    free(clid);
    free(cent1);
    free(cent2);
    clear_error_context();
    return (PyObject *)voclass;
}



/*--------------------------------------------------------------------------
parray_norm1 - Compute vector 1-norm of an array

This function computes the vector 1-norm of an array or the difference
between two arrays, optionally weighted by a pixelwise weight.  The
two arrays must be the same type, but the weight array may be of any
type.  The other and weight arrays are optional, defaulting to zero
and one respectively; if provided, they must be the same shape as the
first array.  As a special case, a scalar zero may be provided as the
value of other; otherwise, scalars may not be provided as arguments.
---------------------------------------------------------------------------*/

static PyObject *
parray_norm1(
    PyObject *self,          /* Self: First array */
    PyObject *args)
{
    PyObject *other  = 0;    /* Arg 1: Other array */
    PyObject *weight = 0;    /* Arg 2: Weight array */
    PyObject *paxes  = 0;    /* Arg 3: Axes to reduce on */
    PyObject *result = 0;    /* Result: Weighted norm of difference */

    voxel_array *slf  = 0;   /* Self array as a voxel array */
    voxel_array *oth  = 0;   /* Other array as a voxel array */
    voxel_array *wate = 0;   /* Weight array as a voxel array */
    voxel_array *vres = 0;   /* Result as a voxel array */
    int naxes = -1;          /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];  /* Specific axes to reduce on */
    int flag;

    set_error_context("array.norm1()");

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "|OOO", &other, &weight, &paxes) )
	return NULL;

    /* Extract voxel array for self */
    if ( !PyvoxArray_Check(self, "self", &slf) )
	ReturnPyvoxLibError();

    /* Extract voxel array for other */
    if ( other == 0 )
	/* Do nothing */ ;
    else if ( PyvoxArray_Check(other, 0, &oth) )
	/* Do nothing */ ;
    else if ( PyInt_Check(other) && PyInt_AsLong(other) == 0 )
        /* Do nothing */ ;
    else if ( PyFloat_Check(other) && PyFloat_AsDouble(other) == 0.0 )
        /* Do nothing */ ;
    else
	ReturnPyvoxError(ERR_INIT, "Other must be a Pyvox array");

    /* Extract voxel array for weights */
    if ( weight == 0 || weight == Py_None )
        wate = 0;
    else if ( PyInt_Check(weight) && PyInt_AsLong(weight) == 1 )
        wate = 0;
    else if ( PyFloat_Check(weight) && PyFloat_AsDouble(weight) == 1.0 )
        wate = 0 ;
    else if ( ! PyvoxArray_Check(weight, "weight", &wate) )
	ReturnPyvoxLibError();

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, slf->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work */
    vres = vxl_create_array();
    if (vres == 0)    ReturnPyvoxLibError();
    flag = vxl_norm1(vres, slf, oth, wate, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    result = pyvox_from_temp_varray(vres);
    if (result == 0)    ReturnPyvoxLibError();

    clear_error_context();
    return result;
}



/*--------------------------------------------------------------------------
parray_norm2 - Compute vector 2-norm of an array

This function computes the vector 2-norm of an array or the difference
between two arrays, optionally weighted by a pixelwise weight.  The
two arrays must be the same type, but the weight array may be of any
type.  The other and weight arrays are optional, defaulting to zero
and one respectively; if provided, they must be the same shape as the
first array.  As a special case, a scalar zero may be provided as the
value of other; otherwise, scalars may not be provided as arguments.
---------------------------------------------------------------------------*/

static PyObject *
parray_norm2(
    PyObject *self,          /* Self: First array */
    PyObject *args)
{
    PyObject *other = 0;     /* Arg 1: Other array */
    PyObject *weight = 0;    /* Arg 2: Weight array */
    PyObject *paxes = 0;     /* Arg 3: Axes to reduce on */
    PyObject *result = 0;    /* Result: Weighted norm of difference */

    voxel_array *slf  = 0;   /* Self array as a voxel array */
    voxel_array *oth  = 0;   /* Other array as a voxel array */
    voxel_array *wate = 0;   /* Weight array as a voxel array */
    voxel_array *vres = 0;   /* Result as a voxel array */
    int naxes = -1;          /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];  /* Specific axes to reduce on */
    int flag;

    set_error_context("array.norm2()");

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "|OOO", &other, &weight, &paxes) )
	return NULL;

    /* Extract voxel array for self */
    if ( !PyvoxArray_Check(self, "self", &slf) )
	ReturnPyvoxLibError();

    /* Extract voxel array for other */
    if ( other == 0 )
	/* Do nothing */ ;
    else if ( PyvoxArray_Check(other, 0, &oth) )
	/* Do nothing */ ;
    else if ( PyInt_Check(other) && PyInt_AsLong(other) == 0 )
        /* Do nothing */ ;
    else if ( PyFloat_Check(other) && PyFloat_AsDouble(other) == 0.0 )
        /* Do nothing */ ;
    else
	ReturnPyvoxError(ERR_INIT, "Other must be a Pyvox array");

    /* Extract voxel array for weights */
    if ( weight == 0 || weight == Py_None )
        wate = 0;
    else if ( PyInt_Check(weight) && PyInt_AsLong(weight) == 1 )
        wate = 0;
    else if ( PyFloat_Check(weight) && PyFloat_AsDouble(weight) == 1.0 )
        wate = 0 ;
    else if ( ! PyvoxArray_Check(weight, "weight", &wate) )
	ReturnPyvoxLibError();

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, slf->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work */
    vres = vxl_create_array();
    if (vres == 0)    ReturnPyvoxLibError();
    flag = vxl_norm2(vres, slf, oth, wate, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    result = pyvox_from_temp_varray(vres);
    if (result == 0)    ReturnPyvoxLibError();

    clear_error_context();
    return result;
}



/*--------------------------------------------------------------------------
parray_normsup - Compute vector sup-norm of an array

This function computes the vector sup-norm of an array or the
difference between two arrays, optionally weighted by a pixelwise
weight.  The two arrays must be the same type; the weight array may be
of any type.  The other and weight arrays are optional, defaulting to
zero and one respectively; if provided, they must be the same shape as
the first array.  As a special case, a scalar zero may be provided as
the value of other; otherwise, scalars may not be provided as
arguments.

FIXME: Allow the 'other' argument to be a plain number?
---------------------------------------------------------------------------*/

static PyObject *
parray_normsup(
    PyObject *self,          /* Self: First array */
    PyObject *args)
{
    PyObject *other = 0;     /* Arg 1: Other array */
    PyObject *weight = 0;    /* Arg 2: Weight array */
    PyObject *paxes = 0;     /* Arg 3: Axes to reduce on */
    PyObject *result = 0;    /* Result: Weighted norm of difference */

    voxel_array *slf  = 0;   /* Self array as a voxel array */
    voxel_array *oth  = 0;   /* Other array as a voxel array */
    voxel_array *wate = 0;   /* Weight array as a voxel array */
    voxel_array *vres = 0;   /* Result as a voxel array */
    int naxes = -1;          /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];  /* Specific axes to reduce on */
    int flag;

    set_error_context("array.normsup()");

    /* Parse the Python arguments */
    if ( !PyArg_ParseTuple(args, "|OOO", &other, &weight, &paxes) )
	return NULL;

    /* Extract voxel array for self */
    if ( !PyvoxArray_Check(self, "self", &slf) )
	ReturnPyvoxLibError();

    /* Extract voxel array for other */
    if ( other == 0 )
	/* Do nothing */ ;
    else if ( PyvoxArray_Check(other, 0, &oth) )
	/* Do nothing */ ;
    else if ( PyInt_Check(other) && PyInt_AsLong(other) == 0 )
        /* Do nothing */ ;
    else if ( PyFloat_Check(other) && PyFloat_AsDouble(other) == 0.0 )
        /* Do nothing */ ;
    else
	ReturnPyvoxError(ERR_INIT, "Other must be a Pyvox array");

    /* Extract voxel array for weights */
    if ( weight == 0 || weight == Py_None )
        wate = 0;
    else if ( PyInt_Check(weight) && PyInt_AsLong(weight) == 1 )
        wate = 0;
    else if ( PyFloat_Check(weight) && PyFloat_AsDouble(weight) == 1.0 )
        wate = 0 ;
    else if ( ! PyvoxArray_Check(weight, "weight", &wate) )
	ReturnPyvoxLibError();

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, slf->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work */
    vres = vxl_create_array();
    if (vres == 0)    ReturnPyvoxLibError();
    flag = vxl_normsup(vres, slf, oth, wate, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    result = pyvox_from_temp_varray(vres);
    if (result == 0)    ReturnPyvoxLibError();

    clear_error_context();
    return result;
}



/*-------------------------------------------------------------------------
parray_p2i - Compute the physical-to-index transform

This function returns the affine transform which converts from
physical to index coordinates for this array, based on the origin and
spacing attributes.

ALGORITHM: This function makes an upcall to the Python level, where the
affine transform class is defined.
-------------------------------------------------------------------------*/

static PyObject *
parray_p2i(
    PyObject *self,             /* Array to get transform of */
    PyObject *args)             /* Args: None used */
{
    voxel_array *vself;         /* Self as a voxel array */
    int rank;                   /* Rank of the array */
    long dimen[2];              /* Dimensions of matrix and offset */
    pyvox_array *mtx;           /* Matrix of the transform */
    voxel_array *vmtx;          /* Matrix, as a voxel array */
    double *mdata;              /* Data of the matrix */
    pyvox_array *off;           /* Offset of the transform */
    voxel_array *voff;          /* Offset, as a voxel array */
    double *odata;              /* Data of the offset */
    PyObject *affine;           /* Affine class object */
    PyObject *result;           /* The transform to be returned as result */
    int i;

    set_error_context("array.p2i()");

    /* Fetch and verify the arguments */
    if ( !PyArg_ParseTuple(args, "") )
	return NULL;

    /* Find the array itself */
    if ( !PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxLibError();
    rank = vself->rank;

    /* Construct arrays for the matrix and offset of the transform */
    dimen[0] = dimen[1] = rank;
    mtx = parray_alloc(&vmtx, INTP_DOUBLE, 2, dimen);
    if (mtx == 0)    ReturnPyvoxLibError();
    mdata = (double *)vmtx->data;
    dimen[1] = 1;
    off = parray_alloc(&voff, INTP_DOUBLE, 2, dimen);
    if (off == 0)    ReturnPyvoxLibError();
    odata = (double *)voff->data;

    /* Fill matrix and offset with the right values */
    for (i = 0; i < rank*rank; i++)
	mdata[i] = 0.0;
    for (i = 0; i < rank; i++) {
	mdata[i*rank+i] = 1.0 / vself->spacing[i];
	odata[i] = - vself->origin[i] / vself->spacing[i];
    }

    /* Construct and return the affine transform */
    affine = PyDict_GetItemString(pyvox_dict, "affine");
    result = PyObject_CallFunction(affine, "iOO", rank, mtx, off);

    Py_INCREF(result);
    clear_error_context();
    return result;
}



/*----------------------------------------------------------------------
parray_prinaxes - Principal axes transformation of real symmetric matrix

This Pyvox array method returns a pair (val, vec) containing the
eigenvalues and eigenvectors of the Pyvox array considered as a real
symmetric matrix.  The matrix of eigenvalues is guaranteed to be proper.
Note that this function is not defined for complex Hermitian matrices.
-----------------------------------------------------------------------*/

static 
PyObject *
parray_prinaxes(
    PyObject *self,            /* Pyvox array */
    PyObject *args)            /* Arguments (none) */
{
    voxel_array *vself;        /* Self as a voxel array */
    pyvox_array *val;          /* Eigenvalues */
    voxel_array *vval;         /* Eigenvalues as voxel array */
    pyvox_array *vec;          /* Eigenvectors */
    voxel_array *vvec;         /* Eigenvectors as voxel array */
    int flag;

    set_error_context("array.prinaxes()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        ReturnPyvoxLibError();

    /* Create voxel arrays for values and vectors */
    val = parray_create(&vval);
    vec = parray_create(&vvec);
    if (val == 0 || vec == 0)    ReturnPyvoxLibError();

    /* Compute principal axes transformation */
    flag = matrix_principal_axes(vval, vvec, vself);
    if (flag)    ReturnPyvoxLibError();

    /* Return pair of eigenvalues and eigenvectors */
    clear_error_context();
    return Py_BuildValue("(NN)", val, vec);
}



/*----------------------------------------------------------------------
parray_reshape - Change the shape (dimensions) of a Pyvox array

This instance method changes the shape of a Pyvox array without
modifying the values contained therein.  As a special case, omitting
the new shape will simply eliminate any dimensions along which there
is only a single element; for example, a 3 x 1 x 5 x 1 array would be
reduced to a 3 x 5 array with the same elements.  It is not permitted
to change the total number of elements by this method; that is, the
product of the new dimensions must equal the product of the old
dimensions.  Nor is it permitted to change the type of the data.

FIXME: Another common need is to flatten an array into a vector.  The
pyvox.array() function doesn't quite do this.
-----------------------------------------------------------------------*/

static PyObject *
parray_reshape(
    PyObject *self,              /* Pyvox array to reshape */
    PyObject *args)              /* New dimensions */
{
    PyObject *shape = 0;         /* Arg 1: Tuple or list with new dimens */

    voxel_array *vself;          /* Self as a voxel array */
    int  newrank = -1;           /* New rank */
    long newdim[VXL_MAX_RANK];   /* New dimensions */
    long oldcnt, newcnt;         /* Old and new number of elements */
    int i, flag;

    set_error_context("array.reshape()");

    /* Convert self to a voxel array */
    vself = ((pyvox_array *)self)->varray;

    /* Fetch the arguments */
    if ( ! PyArg_ParseTuple(args, "|O", &shape) )
        return NULL;

    /* Verify that the number of elements is unchanged */
    /* FIXME: Is some of this testing duplicated in the Voxel Kit? */
    if ( shape ) {
        if ( ! Pyvox_ParseShape(shape, &newrank, newdim) )
            return NULL;
        newcnt = 1;
        for (i = 0; i < newrank; i++)
            newcnt *= newdim[i];
        oldcnt = 1;
        for (i = 0; i < vself->rank; i++)
            oldcnt *= vself->dimen[i];
        if ( newcnt != oldcnt )
            ReturnPyvoxError(ERR_ILLOP, 
                             "Reshape may not change number of elements");
    }

    /* Call Voxel Kit to actually change the shape */
    flag = vxl_reshape_array(vself, newrank, newdim);
    if (flag)    ReturnPyvoxLibError();

    /* Return the modified array */
    Py_INCREF(self);
    clear_error_context();
    return self;
}



/*-----------------------------------------------------------------------
parray_rgba2d - Extract 2D RGBA slice from uninterpreted raw data

This function is intended to support interactive display programs that
must quickly change the assumed external format of displayed images in
response to user commands.  It assumes that the external file has been
read as one-dimensional array of unsigned char bytes, reinterprets
those bytes as an array of given dimensions and external type, selects
a specified 2-dimensional slice, further selects specified input
channels as red, green, blue, and alpha, intensity windows the input
data, and converts into an unsigned char array suitable for display;
all this is, of course, done efficiently so that the user doesn't have
to wait too long.  The result is always a rank three array, although
the last dimension may be one.  The following keyword arguments
specify the details of the transform:

   dimen       List of dimensions.
   extype      External type; defaults to unsigned char.
   bigend      Big endian?  defaults to 1.
   coords      List of coordinates to select slice; defaults to zero.
               NOTE that the echo dimension, if any, is included here.
   axes        Original axes corresponding to output y, x, and band
               of a rank 3 RGB(A) image, or y,x of a rank 2 L image;
               defaults to (0,1).
   chan        Input channel corresponding to each output channel; 
               defaults to (0,0,0), which generates an RGB image from
               a luminance image.

FIXME: This function arguably combines too many functions at once.
Consider providing an import function which does just data import,
slice selection, and channel selection.  More generally, think about
how the various functions can be subdivided into distinct modules
without costing too much efficiency.

FIXME: Check if slice coords are outside the given dimensions; check
if product of dimensions exceeds the available data.  Check that the
dimensions are all positive.  Check that the type is a valid
*external* type.

FIXME: Could provide more complete argument checking for dimen, type,
bigend, slice, and echo.

FIXME: Provide capability to reverse the scan direction along an axis,
or to select a block or subsample.  Provide zoom or subsample along
each axis.

FIXME: Implement intensity windowing and perhaps color table lookup.

FIXME: Provide capability to fill alpha channel with a fixed value, or
to provide 0 if RGB==0 and 255 otherwise?
------------------------------------------------------------------------*/

static PyObject *
parray_rgba2d(
    PyObject *self,
    PyObject *args,
    PyObject *kw)
{
    char *argnames[] =           /* Argument keyword names */
        {"dimen", "extype", "bigend", "coords", "axes", "chan", NULL};
    PyObject *dimen = 0;             /* Arg 1: Dimensions */
    PyObject *type = 0;              /* Arg 2: Type of raw data */
    int bigend = 1;                  /* Arg 3: Big endian? */
    PyObject *coords = 0;            /* Arg 4: Slice coordinates */
    PyObject *axes = 0;              /* Arg 5: Axes to select */
    PyObject *chan = 0;              /* Arg 6: Channels to select */
    pyvox_array *result = 0;         /* Result: Interpreted slice */

    voxel_array *vresult;            /* Result as a voxel array */
    voxel_array *vself = 0;          /* Input data as a voxel array */
    int extype = EXTP_UINT1;         /* Type of raw data */
    struct vxl_raw_format rawform;   /* Describes format of raw data */
    long rank  = 0;                  /* Assumed rank of data */
    long nchan = 0;                  /* Assumed number of channels */
    int  i, flag;

    set_error_context("array.rgba2d()");

    /* Parse the Python arguments */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        ReturnPyvoxLibError();
    if ( ! PyArg_ParseTupleAndKeywords(args, kw, "O|iiOOO", argnames,
                                       &dimen, &extype, &bigend,
                                       &coords, &axes, &chan) )
        return NULL;

    /* DEBUG: Check keyword argument handling by printing out the values */
    if (0) {
        fprintf(stderr, "dimen = ");
        PyObject_Print(dimen, stderr, 0);
        fprintf(stderr, "\nextype, bigend = %d, %d\n", extype, bigend);
        fprintf(stderr, "coords = ");
        PyObject_Print(coords, stderr, 0);
        fprintf(stderr, "\naxes = ");
        PyObject_Print(axes, stderr, 0);
        fprintf(stderr, "\nchan = ");
        PyObject_Print(chan, stderr, 0);
        fprintf(stderr, "\n");
    }

    /* Initialize the raw format descriptor */
    flag = vxl_init_raw_format(&rawform);
    if (flag)
        ReturnPyvoxLibError();

    /* Parse the dimen argument */
    /* FIXME: Create and use PyvoxDimen_Check()? */
    if ( ! PyvoxLongSeq_Check(dimen, 1, VXL_MAX_RANK, 
                              &rank, rawform.dimen) )
        ReturnPyvoxError(ERR_ILLOP, "Invalid dimensions");
    rawform.rank = rank;

    /* Check and copy the type and bigend arguments */
    if ( ! PyvoxExtype_Check(type, "type", &extype) )
        ReturnPyvoxLibError();
    rawform.type   = extype;
    rawform.bigend = bigend;
    
    /* Parse the coords argument; provide default if needed */
    /* FIXME: Create and use PyvoxCoords_Check()? */
    if ( coords == 0 )
        for (i = 0; i < VXL_MAX_RANK; i++)
            rawform.coords[i] = 0;
    else if ( ! PyvoxLongSeq_Check(coords, rank, rank, 0, rawform.coords) )
        ReturnPyvoxError(ERR_ILLOP, "Invalid current slice coordinates");

    /* Parse the axes argument, or provide default */
    /* FIXME: Create and use PyvoxAxes_Check()? */
    rawform.axes[2] = rank;
    if ( axes == 0 || axes == Py_None ) {
        rawform.axes[0] = 0;
        rawform.axes[1] = 1;
    }        
    else if ( ! PyvoxLongSeq_Check(axes, 2, 3, 0, rawform.axes) )
        ReturnPyvoxError(ERR_ILLOP, "Invalid axes selected");

    /* Parse the chan argument, or provide default */
    if ( chan == 0 || chan == Py_None ) {
        nchan = 3;
        for (i = 0; i < nchan; i++)
            rawform.chan[i] = 0;
    }        
    else if ( ! PyvoxLongSeq_Check(chan, 0, VXL_MAX_RGBA, 
                                   &nchan, rawform.chan) )
        ReturnPyvoxError(ERR_ILLOP, "Invalid choice of echoes");
    rawform.nchan = nchan;

    /* DEBUG: Print out contents of the raw format descriptor */
    if (0)
        vxl_print_raw_format(stderr, &rawform);

    /* Create the output array and call Voxel Kit to do the real work */
    result = parray_create(&vresult);
    if (result == 0)    ReturnPyvoxLibError();
    flag = vxl_raw_to_rgba2d(vresult, vself, &rawform);
    if (flag)    ReturnPyvoxLibError();

    clear_error_context();
    return (PyObject *)result;
}



/*----------------------------------------------------------------------
parray_scale - Scale without changing type

This instance method multiples each element by a given constant, adds
another constant, and returns the results at a new array of the same
type as the original, limiting to the range of the destination as
necessary.

FIXME:  Allow user to specify the result type.  Permit keyword args?
------------------------------------------------------------------------*/

static 
PyObject *
parray_scale(
    PyObject *self,             /* Pyvox array */
    PyObject *args)             /* Arguments */
{
    PyObject *gain = 0;         /* Arg 1: Constant to multiply by */
    PyObject *bias = 0;         /* Arg 2: Constant to add */
    pyvox_array *result;        /* Result: New scaled array */

    voxel_array *vself;         /* Self as a voxel array */
    voxel_array *vresult;       /* Result as a voxel array */
    int gtype, btype;           /* Types of gain and bias */
    int rtype;                  /* Type of result */
    anytype gv, bv;             /* Values of gain and bias */
    int flag;

    set_error_context("array.scale()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        ReturnPyvoxLibError();

    /* Convert the gain and bias arguments */
    if ( !PyArg_ParseTuple(args, "|OO", &gain, &bias) )
	return NULL;
    if ( gain && !PyvoxValue_Check(gain, "gain", &gtype, &gv) )
        ReturnPyvoxLibError();
    if ( bias && !PyvoxValue_Check(bias, "bias", &btype, &bv) )
        ReturnPyvoxLibError();

    /* Create voxel array for result */
    result = parray_create(&vresult);
    if (result == 0)    ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work */
    rtype = vself->type;
    flag = vxl_scale(rtype, vresult, vself, gtype, &gv, btype, &bv);
    if (flag)  ReturnPyvoxLibError();
    
    /* Return the result */
    clear_error_context();
    return Py_BuildValue("N", result);
}



/*------------------------------------------------------------------------
parray_solve - Solve the linear system AX = B

This instance method solves the linear system AX = B for a square
matrix A and general matrix B.  Both A and B must be the same
floating-point type.
-------------------------------------------------------------------------*/

static
PyObject *
parray_solve(
    PyObject *self,         /* Pyvox array = A */
    PyObject *args)         /* Arguments */
{
    PyObject *rhs;          /* Arg 1: Right hand side = B */

    voxel_array *vself;     /* Self A as a voxel array */
    voxel_array *vrhs;      /* RHS B as a voxel array */
    pyvox_array *soln;      /* Solution X as a Pyvox array */
    voxel_array *vsoln;     /* Solution X as a voxel array */
    int flag;               /* Error flag */

    set_error_context("array.solve()");

    /* Parse arguments */
    if ( !PyArg_ParseTuple(args, "O", &rhs) )
	return NULL;

    /* Find voxel array of self and rhs */
    if ( !PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxLibError();
    if ( !PyvoxArray_Check(rhs, "right-hand side", &vrhs) )
	ReturnPyvoxLibError();

    /* Create Pyvox array for result */
    soln = parray_create(&vsoln);
    if (soln == 0)    ReturnPyvoxLibError();

    /* Compute the inverse */
    flag = matrix_solve(vsoln, vself, vrhs);
    if (flag)    ReturnPyvoxLibError();

    /* Return the result as a Pyvox array */
    clear_error_context();
    return Py_BuildValue("N", soln);
}



/*-----------------------------------------------------------------------
parray_stat - Compute mean and standard deviation of array elements

This function computes the mean and standard deviation of the elements
of the array, optionally weighted by the elements of another array.
The results are returned as a tuple containing the mean and standard
deviation.
------------------------------------------------------------------------*/

static PyObject *
parray_stat(
    PyObject *self,             /* Self: Source array */
    PyObject *args)
{
    PyObject *weight = 0;       /* Arg 1: Optional weight array */
    PyObject *paxes  = 0;       /* Arg 2: Axes to reduce on */

    PyObject *mean = 0;         /* Result 1: Mean values */
    PyObject *sdev = 0;         /* Result 2: Standard deviation values */

    voxel_array *vself = 0;     /* Self as a voxel array */
    voxel_array *vweight = 0;   /* Weight as a voxel array */
    voxel_array *vmean = 0;     /* Mean as voxel array */
    voxel_array *vsdev = 0;     /* Sdev as voxel array */
    int naxes;                  /* Number of axes to reduce on */
    int axes[VXL_MAX_RANK];     /* The axes to reduce on */
    int flag;

    set_error_context("array.stat()");

    /* Parse the Python arguments and extract voxel arrays */
    if ( !PyArg_ParseTuple(args, "|OO", &weight, &paxes) )
	return NULL;

    /* Extract voxel arrays */
    if ( !PyvoxArray_Check(self, "self", &vself) )
	ReturnPyvoxLibError();
    if ( weight == 0 || weight == Py_None )
        vweight = 0;
    else if ( !PyvoxArray_Check(weight, "weight", &vweight) )
	ReturnPyvoxLibError();

    /* Extract axes to reduce on */
    if  ( ! Pyvox_CheckAxes(paxes, vself->rank, &naxes, axes) )
	ReturnPyvoxLibError();

    /* Call Voxel Kit to do the real work */
    vmean = vxl_create_array();
    vsdev = vxl_create_array();
    if (vmean == 0 || vsdev == 0)
        ReturnPyvoxLibError();
    flag = vxl_stat(vmean, vsdev, vself, vweight, naxes, axes);
    if (flag)    ReturnPyvoxLibError();
    mean = pyvox_from_temp_varray(vmean);
    sdev = pyvox_from_temp_varray(vsdev);
    if (mean == 0 || sdev == 0)
        ReturnPyvoxLibError();

    /* Return results */
    clear_error_context();
    return Py_BuildValue("(NN)", mean, sdev);
}



/*------------------------------------------------------------------------
parray_svd - Compute singular value decomposition of a matrix

This instance method computes the singular value decomposition of a
matrix represented as a Pyvox array of floating-point type.
-------------------------------------------------------------------------*/

static
PyObject *
parray_svd(
    PyObject *self,         /* Pyvox array */
    PyObject *args)         /* Arguments (none) */
{
    voxel_array *vself;     /* Self as a voxel array */
    pyvox_array *left;      /* Left factor as a Pyvox array */
    voxel_array *vleft;     /* Left factor as a voxel array */
    pyvox_array *values;    /* Singular values as a Pyvox array */
    voxel_array *vval;      /* Singular values as a voxel array */
    pyvox_array *right;     /* Right factor as a Pyvox array */
    voxel_array *vright;    /* Right factor as a voxel array */
    int flag;

    set_error_context("array.svd()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        ReturnPyvoxLibError();

    /* Create Pyvox arrays for results */
    left = parray_create(&vleft);
    if (left == 0)    ReturnPyvoxLibError();
    values = parray_create(&vval);
    if (values == 0)    ReturnPyvoxLibError();
    right = parray_create(&vright);
    if (right == 0)    ReturnPyvoxLibError();

    /* Compute the SVD factorization */
    flag = matrix_svd(vleft, vval, vright, vself);
    if (flag)    ReturnPyvoxLibError();

    /* Return the result as a tuple of Pyvox arrays */
    clear_error_context();
    return Py_BuildValue("(NNN)", left, values, right);
}



/*------------------------------------------------------------------------
parray_trans - Transpose of a matrix

This instance method computes the transpose of a matrix represented as
a Pyvox array of any type.

FIXME: Generalize to allow any rank, any permutation of axes, or reversal
of indices along an axis.  Find or invent a fast algorithm for this.
-------------------------------------------------------------------------*/

static
PyObject *
parray_trans(
    PyObject *self,        /* Pyvox array */
    PyObject *args)        /* Arguments (none) */
{
    voxel_array *vself;     /* Self as a voxel array */
    pyvox_array *res;       /* Result as a Pyvox array */
    voxel_array *vres;      /* Result as a voxel array */
    int flag;               /* Error flag */

    set_error_context("array.trans()");

    /* Find voxel array of self */
    if ( ! PyvoxArray_Check(self, "self", &vself) )
        ReturnPyvoxLibError();

    /* Create Pyvox array for result */
    res = parray_create(&vres);
    if (res == 0)    ReturnPyvoxLibError();

    /* Compute the transpose */
    flag = matrix_transpose(vres, vself, 0);
    if (flag)    ReturnPyvoxLibError();

    /* Return the result as a Pyvox array */
    clear_error_context();
    return Py_BuildValue("N", res);
}



/*--------------------------------------------------------------------------
parray_tuple - Return array elements as a tuple

This instance method for Pyvox arrays returns the elements of the array
(in row-major order) as a Python tuple of floats, ints, or longs.  See also 
parray_list, which returns a list rather than a tuple.
--------------------------------------------------------------------------*/

static PyObject *
parray_tuple(
    PyObject *self,            /* Pyvox array to enlist */
    PyObject *args)            /* Arguments (none in this case) */
{
    PyObject *result;          /* Tuple to return as result */
    voxel_array *vself;        /* Self as a voxel array */
    int type;                  /* Type of voxel array */
    void *data;                /* Data contained in voxel array */
    long count;                /* Number of elements in array */
    PyObject *value;           /* Value of an element */
    long i;

    set_error_context("array.tuple()");

    /* Find voxel array of self; count number of elements */
    if ( ! PyvoxArray_CheckSrc(self, "self", &vself) )
        ReturnPyvoxLibError();
    type  = vself->type;
    data  = vself->data;
    count = vxli_count(vself);

    /* Check other arguments; there should be none */
    if ( !PyArg_ParseTuple(args, "") )  
	return NULL;

    /* Construct the output tuple */
    result = PyTuple_New(count);
    for (i = 0; i < count; i++) {
        value = pyvox_get_value(type, data, i);
        if (value == 0)    ReturnPyvoxLibError();
        PyTuple_SetItem(result, i, value);
    }

    clear_error_context();
    return result;
}



/*-------------------------------------------------------------------------
parray_write - Write contents of a Pyvox array in specified external format

This instance method writes the contents of a Pyvox array to an external
file in some specified external format such as raw, pgm, ppm, etc.  For
simplicity and consistency with pyvox.read(), it upcalls to the Python
function pyvox.write_array(), which decides which format is required and
usually calls back down to one of the functions parray_writeraw, etc
in this file.

NOTE that this method also serves as an example of an upcall from C to
Python.

FIXME: Deprecate and remove the optional positional arguments format
and bigend, making them available as keyword arguments only.  Make the
matching changes to pyvox.py.
----------------------------------------------------------------------------*/

static PyObject *
parray_write(
     PyObject *self,           /* Self as a Pyvox array */
     PyObject *args,           /* Positional arguments of this call */
     PyObject *kwargs)         /* Keyword arguments of this call */
{
    PyObject *filename = 0;    /* Arg 1: Filename */
    PyObject *format   = 0;    /* Arg 2: Format */
    PyObject *bigend   = 0;    /* Arg 3: Bigendian? */

    PyObject *funct   = 0;     /* pyvox.write_array() function */
    PyObject *newargs = 0;     /* Modified arguments for it */
    PyObject *result  = 0;     /* Result of calling it */

    set_error_context("array.write()");

    /* Parse argument list */
    if ( !PyArg_ParseTuple(args, "O|OO", &filename, &format, &bigend) )
        ReturnPyvoxLibError();

    /* DEBUG: Print ref counts of arguments */
    if (0) {
      printf("Refcnts 1: %d, %d\n", self->ob_refcnt, filename->ob_refcnt);
    }

    /* Paste self onto front of argument tuple.  FIXME: This is really
       klutzy, but there doesn't seem to an easier way. */
    if ( bigend != 0 )
        newargs = Py_BuildValue("(OOOO)", self, filename, format, bigend);
    else if ( format != 0 )
        newargs = Py_BuildValue("(OOO)", self, filename, format);
    else
        newargs = Py_BuildValue("(OO)", self, filename);

    /* DEBUG: Print ref counts of arguments */
    if (0) {
      printf("Refcnts 2: %d, %d, %d\n", self->ob_refcnt, filename->ob_refcnt,
	     newargs->ob_refcnt);
    }

    /* Find the pyvox.write_array function in the pyvox dictionary */
    funct = PyDict_GetItemString(pyvox_dict, "write_array");
    if (funct == 0)
        ReturnPyvoxError(ERR_ASSERT, "Could not find pyvox.write_array()");

    /* Call it and return the result it gives */
    result = PyEval_CallObjectWithKeywords(funct, newargs, kwargs);

    /* Decrement ref count of the newargs tuple; the ref counts of its
       elements will be automatically decremented also. */
    Py_XDECREF(newargs);

    /* DEBUG: Print ref counts of arguments */
    if (0) {
      printf("Refcnts 3: %d, %d, %d\n", self->ob_refcnt, filename->ob_refcnt,
	     newargs->ob_refcnt);
    }

    clear_error_context();
    return result;
}



/*--------------------------------------------------------------------------
parray_writeraw - Write contents of a Pyvox array to an external file

This instance method writes the contents of a Pyvox array to an
external file in some specified external format.  If a string is given
as the file argument, then that file is opened for writing, the data
are written to it, and the file is closed; if a file object is given,
then the data are written to the file starting at its current position
but it is not closed nor repositioned after writing.  If the string
given for the file name ends in ".gz" or ".Z", the image file is
gzipped or compressed after being written.  If no external data type
is specified, the data are written in the "natural" data type
corresponding to the type of the voxel array; note that careless use
of this feature can produce output files that you don't know how to
read back in.
---------------------------------------------------------------------------*/

static PyObject *
parray_writeraw(
    PyObject *self,           /* Pyvox array to write */
    PyObject *args)           /* Arguments describing the output */
{
    PyObject *file;           /* Arg 1: Filename or open file */
    PyObject *type = 0;       /* Arg 2: External type (optional) */
    int bigend = 1;           /* Arg 3: Bigendian? (default to yes) */

    int extype = EXTP_NONE;   /* External format */
    voxel_array *varray;      /* Voxel array that self contains */
    char *filename = 0;       /* Name of the image file for output, or zero */
    char *basename = 0;       /* Name of uncompressed image file, or zero */
    char *cmd = 0;            /* Command to compress after writing */
    FILE *outfile;            /* Open file for output */
    int  fnlen;               /* Length of filename */
    int  flag;                /* Error status flag */
    
    set_error_context("array.writeraw()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "O|Oi", &file, &type, &bigend) )
	return NULL;

    /* Convert a filename or open file into a C FILE pointer; set up
       for compression if requested */
    if ( PyString_Check(file) ) {
        filename = PyString_AsString(file);
        fnlen    = strlen(filename);
        basename = mallock(fnlen+1);
        if (basename == 0)   ReturnPyvoxLibError();
        strcpy(basename, filename);
        if ( strcmp(filename+fnlen-3, ".gz") == 0 ) {
            basename[fnlen-3] = 0;
            cmd = mallock(fnlen + 9);
            if (cmd == 0)    ReturnPyvoxLibError();
            strcpy(cmd, "gzip -f ");
            strcat(cmd, basename);
        }
        else if ( strcmp(filename+fnlen-2, ".Z") == 0 ) {
            basename[fnlen-2] = 0;
            cmd = mallock(fnlen + 13);
            if (cmd == 0)    ReturnPyvoxLibError();
            strcpy(cmd, "compress -f ");
            strcat(cmd, basename);
        }        
	outfile  = fopen(basename, "wb");
	if (outfile == 0)
	    ReturnPyvoxError(ERR_FILE, "Unable to open the given filename");
    }
    else if ( PyFile_Check(file) )
        outfile = PyFile_AsFile(file);
    else
        ReturnPyvoxError(ERR_INIT, 
                         "Argument 1 isn't a filename or an open file");

    /* Choose the external data type and write the data */
    if ( type && !PyvoxExtype_Check(type, "format", &extype) )
        ReturnPyvoxLibError();
    varray = ((pyvox_array *)self)->varray;
    if (extype == EXTP_NONE)
	extype = exim_natural_extype(varray->type);
    flag = vxl_write_raw_data(outfile, extype, bigend, varray);
    if (flag)    ReturnPyvoxLibError();

    /* Close the file, if specified as a filename; maybe compress or gzip */
    if ( filename ) {
	fclose(outfile);
    }
    if ( cmd) {
        flag = system(cmd);
        if (flag) 
            ReturnPyvoxError(ERR_MISC, "Image file compression failed");
    }

    /* Free any temporary storage allocated */
    if (basename) free(basename);
    if (cmd)      free(cmd);

    /* Return the Python object None */
    Py_INCREF(Py_None);
    clear_error_context();
    return Py_None;
}



/*--------------------------------------------------------------------------
parray_writepgm - Write contents of a Pyvox array as a pgm file

This instance method writes the contents of a Pyvox array to an
external file in the binary Portable Gray Map format.  The array
must be unsigned char and rank 3.

FIXME: Write a test case for this function.  And one for ppm, if you
haven't already.

FIXME: Arrange to ignore extra dimensions of length 1.
---------------------------------------------------------------------------*/

static PyObject *
parray_writepgm(
    PyObject *self,              /* Pyvox array to write */
    PyObject *args)              /* Arguments describing the output */
{
    PyObject *file;              /* Arg 1: Filename */

    voxel_array *varray;         /* Voxel array that self contains */
    char *filename;              /* Name of the image file for output */
    int flag;
    
    set_error_context("array.writepgm()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "O", &file) )
	return NULL;
    if ( ! PyString_Check(file) )
        ReturnPyvoxError(ERR_INIT, "Argument isn't a filename");
    filename = PyString_AsString(file);

    /* Choose the external format and write the data */
    varray = ((pyvox_array *)self)->varray;
    flag = vxl_write_pgm_image(filename, varray);
    if (flag)    ReturnPyvoxLibError();

    /* Return the Python object None */
    Py_INCREF(Py_None);
    clear_error_context();
    return Py_None;
}



/*--------------------------------------------------------------------------
parray_writeppm - Write contents of a Pyvox array as a ppm file

This instance method writes the contents of a Pyvox array to an
external file in the binary Portable Pixel Map format.  The array
must be unsigned char, rank 3, and the last dimension must be 3.

FIXME: Arrange to ignore extra dimensions of length 1.
---------------------------------------------------------------------------*/

static PyObject *
parray_writeppm(
    PyObject *self,              /* Pyvox array to write */
    PyObject *args)              /* Arguments describing the output */
{
    PyObject *file;              /* Arg 1: Filename */

    voxel_array *varray;         /* Voxel array that self contains */
    char *filename;              /* Name of the image file for output */
    int flag;
    
    set_error_context("array.writeppm()");

    /* Fetch the arguments */
    if ( !PyArg_ParseTuple(args, "O", &file) )
	return NULL;
    if ( ! PyString_Check(file) )
        ReturnPyvoxError(ERR_INIT, "Argument isn't a filename");
    filename = PyString_AsString(file);

    /* Choose the external format and write the data */
    varray = ((pyvox_array *)self)->varray;
    flag = vxl_write_ppm_image(filename, varray);
    if (flag)    ReturnPyvoxLibError();

    /* Return the Python object None */
    Py_INCREF(Py_None);
    clear_error_context();
    return Py_None;
}



/**************************************************************************
Method Table for Pyvox Array Type
**************************************************************************/



/*--------------------------------------------------------------------------
parray_methods - Table of instance methods for getattr
--------------------------------------------------------------------------*/

#define V    METH_VARARGS
#define VK ( METH_VARARGS | METH_KEYWORDS )

static struct PyMethodDef
parray_methods [] = {
    { "abs",      (PyCFunction)parray_abs2,     V, "Absolute value" },
    { "acos",     (PyCFunction)parray_acos,     V, "Arc cosine" },
    { "amax",     (PyCFunction)parray_amax,     V, "Maximum in array" },
    { "amin",     (PyCFunction)parray_amin,     V, "Minimum in array" },
    { "aprod",    (PyCFunction)parray_aprod,    V, "Product of elements" },
    { "asin",     (PyCFunction)parray_asin,     V, "Arc sine" },
    { "asum",     (PyCFunction)parray_asum,     V, "Sum of elements" },
    { "asum0",    (PyCFunction)parray_asum0,    V, "Sum of elements (Ref)" },
    { "atan",     (PyCFunction)parray_atan,     V, "Arc tangent" },
    { "atan2",    (PyCFunction)parray_atan2,    V, "Arc tangent of two args" },
    { "bigob",    (PyCFunction)parray_bigob,    V, "Biggest object in image" },
    { "bihist",   (PyCFunction)parray_bihist,   V, "Bivariate histogram" },
#ifdef HAVE_COMPLEX_H
    { "carg",     (PyCFunction)parray_carg,     V, "Complex argument" },
#endif
    { "cast",     (PyCFunction)parray_cast,     V, "Cast to specified type" },
#ifdef HAVE_CBRT
    { "cbrt",     (PyCFunction)parray_cbrt,     V, "Cube root" },
#endif
    { "ceil",     (PyCFunction)parray_ceil,     V, "Ceiling" },
    { "chamfer",  (PyCFunction)parray_chamfer,  V, "Chamfer distance" },
    { "cholesky", (PyCFunction)parray_cholesky, V, "Cholesky decomposition" },
    { "cmove",    (PyCFunction)parray_cmove,    V, "Conditional move" },
    { "compare",  (PyCFunction)parray_compare,  V, "Comparison" },
#ifdef HAVE_COMPLEX_H
    { "conj",     (PyCFunction)parray_conj,     V, "Complex conjugate" },
#endif
    { "convolve", (PyCFunction)parray_convolve, V, "Convolution" },
    { "copy",     (PyCFunction)parray_copy,     V, "Copy of array" },
    { "cos",      (PyCFunction)parray_cos,      V, "Cosine" },
    { "cosh",     (PyCFunction)parray_cosh,     V, "Hyperbolic cosine" },
    { "count",    (PyCFunction)parray_count,    V, "Count of voxels in image"},
    { "cross",    (PyCFunction)parray_cross,    V, "Cross product"},
    { "cubic",    (PyCFunction)parray_cubic,    V, "Cubic interpolation"},
#ifdef HAVE_COMPLEX_H
    { "dcomplex", (PyCFunction)parray_dcomplex, V, "Cast to double complex" },
#endif
    { "det",      (PyCFunction)parray_det,      V, "Determinant" },
    { "diag",     (PyCFunction)parray_diag,     V, "Diagonal of matrix" },
    { "dilabel",  (PyCFunction)parray_dilabel,  V, "Dilate label image" },
    { "dilate",   (PyCFunction)parray_dilate,   V, "Dilate binary image" },
    { "dot",      (PyCFunction)parray_dot,      V, "Dot product" },
    { "double",   (PyCFunction)parray_double,   V, "Cast to double" },
    { "eigsy",    (PyCFunction)parray_eigsy,    V, "Symmetric eigensystem" },
#ifdef HAVE_ERF
    { "erf",      (PyCFunction)parray_erf,      V, "Error function" },
#endif
#ifdef HAVE_ERFC
    { "erfc",     (PyCFunction)parray_erfc,     V, "Comp error function" },
#endif
    { "erode",    (PyCFunction)parray_erode,    V, "Morphological erosion" },
    { "exp",      (PyCFunction)parray_exp,      V, "Exponential" },
#ifdef HAVE_EXPM1
    { "expm1",    (PyCFunction)parray_expm1,    V, "Exp(x) - 1" },
#endif
#ifdef HAVE_EXP2
    { "exp2",     (PyCFunction)parray_exp2,     V, "2**x" },
#endif
    { "fill2d",   (PyCFunction)parray_fill2d,   V, "Fill 2D contour" },
#ifdef HAVE_COMPLEX_H
    { "fcomplex", (PyCFunction)parray_fcomplex, V, "Cast to float complex" },
    { "fft",      (PyCFunction)parray_fft,      V, "Fast Fourier transform" },
#endif
    { "float",    (PyCFunction)parray_float,    V, "Cast to float" },
    { "floor",    (PyCFunction)parray_floor,    V, "Floor" },
    { "fma",      (PyCFunction)parray_fma,      V, "Fused multiply-add x*y+z"},
    { "H",        (PyCFunction)parray_conjtrans,V, "Conjugate transpose" },
    { "histo",    (PyCFunction)parray_histo,    V, "Histogram" },
#ifdef HAVE_HYPOT
    { "hypot",    (PyCFunction)parray_hypot,    V, "Hypotenuse" },
#endif
    { "i2p",      (PyCFunction)parray_i2p,      V, "Index-to-physical xform" },
#ifdef HAVE_COMPLEX_H
    { "ifft",     (PyCFunction)parray_ifft,     V, "Fast Fourier transform" },
    { "imag",     (PyCFunction)parray_imag,     V, "Imaginary part" },
#endif
    { "int",      (PyCFunction)parray_int,      V, "Cast to int" },
    { "inverse",  (PyCFunction)parray_inverse,  V, "Inverse of matrix" },
    { "kmeans1",  (PyCFunction)parray_kmeans1,  V, "K-means in 1D" },
    { "kmeans2",  (PyCFunction)parray_kmeans2,  V, "K-means in 2D" },
    { "linear",   (PyCFunction)parray_linear,   V, "Linear interpolation" },
    { "list",     (PyCFunction)parray_list,     V, "List of elements" },
#ifdef HAVE_LGAMMA
    { "lgamma",   (PyCFunction)parray_lgamma,   V, "Log of gamma function" },
#endif
    { "log",      (PyCFunction)parray_log,      V, "Logarithm base e" },
    { "log10",    (PyCFunction)parray_log10,    V, "Logarithm base 10" },
#ifdef HAVE_LOG1P
    { "log1p",    (PyCFunction)parray_log1p,    V, "Logarithm (1+x)" },
#endif
#ifdef HAVE_LOG2
    { "log2",     (PyCFunction)parray_log2,     V, "Logarithm base 2" },
#endif
    { "logcomp",  (PyCFunction)parray_logcomp,  V, "Logarithmic compression" },
    { "long",     (PyCFunction)parray_long,     V, "Cast to long" },
    { "lookup",   (PyCFunction)parray_lookup,   V, "Lookup table" },
    { "lostat",   (PyCFunction)parray_lostat,   V, "Local mean and variance" },
    { "lowpass",  (PyCFunction)parray_lowpass,  V, "Lowpass filter" },
    { "max",      (PyCFunction)parray_max,      V, "Elementwise maximum" },
    { "mean",     (PyCFunction)parray_mean,     V, "Mean along given axes" },
    { "metacopy", (PyCFunction)parray_metacopy,VK, "Copy metadata" },
    { "min",      (PyCFunction)parray_min,      V, "Elementwise minimum" },
    { "minmax",   (PyCFunction)parray_minmax,   V, "Min/max value in array" },
    { "mmul",     (PyCFunction)parray_mmul,     V, "Matrix multiplication" },
    { "moments",  (PyCFunction)parray_moments,  V, "Image moments" },
    { "mop",      (PyCFunction)parray_mop,      V, "Moments of image product"},
    { "nearest",  (PyCFunction)parray_nearest,  V, "NN interpolation" },
    { "nnclass1", (PyCFunction)parray_nnclass1, V, "NN classifier in 1D" },
    { "nnclass2", (PyCFunction)parray_nnclass2, V, "NN classifier in 2D" },
    { "norm1",    (PyCFunction)parray_norm1,    V, "Vector 1-norm" },
    { "norm2",    (PyCFunction)parray_norm2,    V, "Vector 2-norm" },
    { "normsup",  (PyCFunction)parray_normsup,  V, "Vector sup-norm" },
    { "p2i",      (PyCFunction)parray_p2i,      V, "Physical-to-index xform" },
    { "pow",      (PyCFunction)parray_pow,      V, "Power function" },
    { "prinaxes", (PyCFunction)parray_prinaxes, V, "Principal axes" },
#ifdef HAVE_COMPLEX_H
    { "real",     (PyCFunction)parray_real,     V, "Real part" },
#endif
    { "reshape",  (PyCFunction)parray_reshape,  V, "Change shape of array" },
    { "rgba2d",   (PyCFunction)parray_rgba2d,  VK, "Convert raw to rgba2d" },
    { "scale",    (PyCFunction)parray_scale,    V, "Scale by gain and bias" },
    { "schar",    (PyCFunction)parray_schar,    V, "Cast to signed char" },
    { "short",    (PyCFunction)parray_short,    V, "Cast to short" },
    { "sin",      (PyCFunction)parray_sin,      V, "Sine" },
    { "sinh",     (PyCFunction)parray_sinh,     V, "Hyperbolic sine" },
    { "solve",    (PyCFunction)parray_solve,    V, "Solve linear system" },
    { "sqrt",     (PyCFunction)parray_sqrt,     V, "Square root" },
    { "stat",     (PyCFunction)parray_stat,     V, "Mean and std deviation" },
    { "svd",      (PyCFunction)parray_svd,      V, "SVD"},
    { "T",        (PyCFunction)parray_trans,    V, "Matrix transpose" },
    { "tan",      (PyCFunction)parray_tan,      V, "Tangent" },
    { "tanh",     (PyCFunction)parray_tanh,     V, "Hyperbolic tangent" },
#ifdef HAVE_TGAMMA
    { "tgamma",   (PyCFunction)parray_tgamma,   V, "Gamma function" },
#endif
    { "trans",    (PyCFunction)parray_trans,    V, "Matrix transpose" },
    { "tuple",    (PyCFunction)parray_tuple,    V, "Tuple of elements" },
    { "uchar",    (PyCFunction)parray_uchar,    V, "Cast to unsigned char" },
    { "uint",     (PyCFunction)parray_uint,     V, "Cast to unsigned int" },
    { "ulong",    (PyCFunction)parray_ulong,    V, "Cast to unsigned long" },
    { "ushort",   (PyCFunction)parray_ushort,   V, "Cast to unsigned short" },
    { "write",    (PyCFunction)parray_write,   VK, "Generic array write" },
    { "writeraw", (PyCFunction)parray_writeraw, V, "Write in raw format" },
    { "writepgm", (PyCFunction)parray_writepgm, V, "Write in PGM format"},
    { "writeppm", (PyCFunction)parray_writeppm, V, "Write in PPM format"},
    { NULL,       NULL }
};



/*************************************************************************
Pyvox array type object
*************************************************************************/



/*------------------------------------------------------------------------
parray_type - Python type descriptor for the voxel array type

See Include/object.h for more information on this struct, including
more methods.

FIXME: Implement repr() and str() methods.
------------------------------------------------------------------------*/

/* FIXME: Always use Py-TPFLAGS_DEFAULT instead? */
#ifdef Py_TPFLAGS_HAVE_RICHCOMPARE
#define PARRAY_TPFLAGS  Py_TPFLAGS_HAVE_RICHCOMPARE
#else
#define PARRAY_TPFLAGS Py_TPFLAGS_DEFAULT
#endif

PyTypeObject parray_type = {

    /* Type head */
    PyObject_HEAD_INIT(&PyType_Type)
    0,                              /* ob_size:      Nmbr of items */
    "Pyvox_array",                  /* tp_name:      Name of type */
    sizeof(pyvox_array),            /* tp_basicsize: Basic size of type */
    0,                              /* tp_itemsize:  Item size */

    /* Standard methods */
    (destructor)  parray_destroy,   /* tp_dealloc:   Instance destructor */
    (printfunc)   parray_print,     /* tp_print:     "print x" */
    (getattrfunc) parray_getattr,   /* tp_getattr:   "instance.attr" */
    (setattrfunc) parray_setattr,   /* tp_setattr:   "instance.attr = foo" */
    (cmpfunc)     0,                /* tp_compare:   compare instances */
    (reprfunc)    0,                /* tp_repr:      string representation */

    /* Method suites for standard classes */
    (PyNumberMethods *)   &parray_as_number,    /* tp_as_number */
    (PySequenceMethods *) &parray_as_sequence,  /* tp_as_sequence */
    (PyMappingMethods *)  &parray_as_mapping,   /* tp_as_mapping */

    /* Other methods */
    (hashfunc) 0,                   /* tp_hash:  Hash function of instance */
    (ternaryfunc) 0,                /* tp_call:  "instance()" */
    (reprfunc) 0,                   /* tp_str:   "str(instance)" */
    (getattrofunc) 0,               /* tp_getattro: */
    (setattrofunc) 0,               /* tp_setattro: */

    (PyBufferProcs *) 0,            /* tp_as_buffer: access as i/o buffer */
    PARRAY_TPFLAGS,                 /* tp_flags: identify optional features */
    0,                              /* tp_doc:   documentation string */

    /* Following fields were added in Python 2.0 */
#ifdef Py_TPFLAGS_HAVE_INPLACEOPS
    (traverseproc) 0,               /* tp_traverse: find all subobjects*/
    (inquiry) 0,                    /* tp_clear: delete subobjects */
#endif

    /* Following fields added in Python 2.1 */
#ifdef Py_TPFLAGS_HAVE_RICHCOMPARE
    (richcmpfunc) parray_richcmp,   /* tp_richcompare:    rich compare */
    0,                              /* tp_weaklistoffset: weak refs */
#endif

    /* See Include/object.h for still other methods */
};
